;; gimml-mode.el. Major mode for editing Gimml.
;; Copyright (C) 1995-1999 Jean Goubault-Larrecq and Bull S.A.
;; Copyright (C) 2000-2004, 2011, Jean Goubault-Larrecq
;; and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.
;; Copyright (C) 1995-1999, Jean Goubault-Larrecq.
;; Copyright (C) 1989, Lars Bo Nielsen.
;; Copyright (C) 1989, Free Software Foundation, Inc.
;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;; AUTHOR	Lars Bo Nielsen
;;		Aalborg University
;;		Computer Science Dept.
;;		9000 Aalborg
;;		Denmark
;;
;; EMAIL	lbn@iesd.auc.dk
;;		or: ...!mcvax!diku!iesd!lbn
;;		or: mcvax!diku!iesd!lbn@uunet.uu.net
;;
;; Modified by Jean Goubault-Larrecq for use with Gimml.
;;		Jean Goubault-Larrecq
;;		LSV, ENS Cachan
;;		61 avenue du président Wilson
;;		F-94230 Cachan
;;
;; EMAIL	goubault@lsv.ens-cachan.fr
;;
;; Please let me know if you come up with any ideas, bugs, or fixes.
;; ----------------------------------------------------------------


(provide 'gimml-mode)

(defconst gimml-mode-version-string
  "Gimml-MODE V1.1 (September 1996) (Jean.Goubault@inria.fr),
derived from SML-MODE, Version 3.0 (May 1990) (lbn@iesd.auc.dk)")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; CONSTANTS CONTROLLING THE MODE.
;;;
;;; These are the constants you might want to change
;;; 

(defvar gimml-indent-level 4
  "*Indentation of blocks in ml.")

(defvar gimml-pipe-indent -2
  "*Extra (negative) indentation for lines beginning with |.")

(defvar gimml-case-indent nil
  "*How to indent case-of expressions.
 If t:   case expr              If nil:   case expr of
           of exp1 => ...                     exp1 => ...
            | exp2 => ...                   | exp2 => ...

The first seems to be the standard in NJ-SML, but the second
is the default.")

(defvar gimml-nested-if-indent t
  "*If set to t, nested if-then-else expression will have the same
indentation as:
                 if exp1 then exp2
                 else if exp3 then exp4
                 else if exp5 then exp6
                      else exp7")

(defvar gimml-type-of-indent t
  "*How to indent `let' `struct' etc.
 If t:  fun foo bar = let              If nil:  fun foo bar = let
                          val p = 4                 val p = 4
                      in                        in
                          bar + p                   bar + p
                      end                       end

Will not have any effect if the starting keyword is first on the line.")

(defvar gimml-electric-semi-mode nil
  "*If t, a `\;' will insert itself, reindent the line, and perform a newline.
If nil, just insert a `\;'. (To insert while t, do: C-q \;).")

(defvar gimml-paren-lookback 2000
  "*Determines how far back (in chars) the indentation algorithm
should look for open parenthesis. High value means slow indentation
algorithm. A value of 2000 (being the equivalent of 40-60 lines) should
suffice most uses. (A value of nil, means do not look at all)")

(defvar gimml-prog-name "gimml"
  "*Name of program to run as ml.")

(defvar gimml-prog-name-ask-p nil
  "*If t, you will be asked which program to run when the inferior
shell starts up.")

(defvar gimml-use-left-delim "\""
  "*The left delimiter for the filename when using \"use\".
 To be set to `\"[\\\"\"' for Edinburgh SML, and `\"\\\"\"' for New Jersey SML.
 Corresponds to `gimml-use-right-delim'.")

(defvar gimml-use-right-delim "\""
  "*The right delimiter for the filename when using \"use\".
 To be set to `\"\\\"]\"' for Edinburgh SML, and `\"\\\"\"' for New Jersey SML.
 Corresponds to `gimml-use-left-delim'.")

(defvar gimml-shell-prompt-pattern "^\\(>\\|\\((debug)\\)[ \\t]*\\)+"
  "*The prompt pattern for the inferior shell running ml.")

(defvar gimml-tmp-template "/tmp/gimml.tmp."
  "*Template for the temporary files, created when a region is send to
the inferior process running ml (by gimml-simulate-send-region).")

(defvar gimml-strip-path t
  "*If t, then when sending a `use file' to the
inferior ml process, the leading path of the filename is stripped of,
iff the file is in the same directory as the directory ml was started
in.
If you change the working directory of the inferior gimml, after it has
started up you are in trouble. To get out of this trouble, set this
variable to nil.")

(defvar gimml-message-buffer-name "*Gimml message*"
  "*Name of the buffer where all Gimml messages will be printed.")

(defvar gimml-message-buffer-height 7
  "*Normal height of Gimml's message buffer.")

(defvar gimml-shell-cd-regexp "cd"
  "*Regexp to match Gimml commands equivalent to cd.")

(defvar gimml-shell-dirtrackp t
  "Non-nil in a Gimml buffer means directory tracking is enabled.")

;;;
;;; END OF CONSTANTS CONTROLLING THE MODE.
;;;
;;; If you change anything below, you are on your own.
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defvar gimml-mode-map nil "The mode map used in gimml-mode.")
(defvar gimml-mode-abbrev-table nil "The table of abbrevs for gimml-mode.")

(defvar gimml-emacs19-p (>= (string-to-number (substring emacs-version 0 2)) 19))
(defvar gimml-emacs20-p (>= (string-to-number (substring emacs-version 0 2)) 20))

(if gimml-emacs19-p ; emacs 19: then install a menu
    (defun gimml-create-mode-menu (mode-map shellp)
      (let ((map (make-sparse-keymap "Gimml")))
	(define-key mode-map [menu-bar] (make-sparse-keymap))
	(define-key mode-map [menu-bar gimml]
	  (cons "Gimml" map))
	(if shellp
	    (progn
	      (define-key map [next-error] '("Next Error" . gimml-next-error))
	      (define-key map [prev-error] '("Previous Error" . gimml-prev-error))
	      (define-key map [separator-format] '("--"))
	      )
	  )
	(define-key map [send-buffer] '("Send Buffer" . gimml-send-buffer))
	(define-key map [send-region] '("Send Region" . gimml-send-region))
	(define-key map [send-function] '("Send Function" . gimml-send-function))
	(define-key map [eval] '("Eval..." . gimml-eval))
	(define-key map [run-on-file] '("Run on File..." . gimml-run-on-file))
	(if shellp
	    ()
	  (define-key map [separator-format] '("--"))
	  (define-key map [electric-pipe] '("Electric Pipe" . gimml-electric-pipe))
	  (define-key map [indent-region] '("Indent Region" . gimml-indent-region))
	  (define-key map [indent-line] '("Indent Line" . gimml-indent-line))
	  (put 'indent-region 'menu-enable 'mark-active))))
  (defun gimml-create-mode-menu (mode-map shellp))
  )

(if gimml-mode-map
    ()
  (setq gimml-mode-map (if gimml-emacs19-p
			   (make-sparse-keymap "Gimml")
			 (make-sparse-keymap)))
  ;(define-key gimml-mode-map "\C-c\C-'" 'gimml-next-error)
  (define-key gimml-mode-map "\C-c\C-n" 'gimml-next-error)
  (define-key gimml-mode-map "\C-c\C-p" 'gimml-prev-error)
  (define-key gimml-mode-map "\C-c\C-v" 'gimml-mode-version)
  (define-key gimml-mode-map "\C-c\C-u" 'gimml-save-buffer-use-file)
  (define-key gimml-mode-map "\C-c\C-s" 'gimml-pop-to-shell)
  (define-key gimml-mode-map "\C-c\C-r" 'gimml-send-region)
  (define-key gimml-mode-map "\C-c\C-e" 'gimml-eval)
  (define-key gimml-mode-map "\C-c\C-m" 'gimml-region)
  (define-key gimml-mode-map "\C-c\C-f" 'gimml-run-on-file)
  (define-key gimml-mode-map "\C-c\C-c" 'gimml-send-function)
  (define-key gimml-mode-map "\C-c\C-b" 'gimml-send-buffer)
  (define-key gimml-mode-map "\e|" 'gimml-electric-pipe)
  (define-key gimml-mode-map "\e\t" 'gimml-back-to-outer-indent)
  (define-key gimml-mode-map "\C-j" 'reindent-then-newline-and-indent)
  (define-key gimml-mode-map "\177" 'backward-delete-char-untabify)
  (define-key gimml-mode-map "\;" 'gimml-electric-semi)
  (define-key gimml-mode-map "\C-c\t" 'gimml-indent-region)
  (define-key gimml-mode-map "\t" 'gimml-indent-line)
  (gimml-create-mode-menu gimml-mode-map nil)
  )

(defvar gimml-mode-syntax-table nil "The syntax table used in gimml-mode.")
(if gimml-mode-syntax-table
    ()
  (setq gimml-mode-syntax-table (make-syntax-table))
  ;; Set everything to be "." (punctuation) except for [A-Za-z0-9],
  ;; which will default to "w" (word-constituent).
  (let ((i 0))
    (while (< i ?0)
      (modify-syntax-entry i "." gimml-mode-syntax-table)
      (setq i (1+ i)))
    (setq i (1+ ?9))
    (while (< i ?A)
      (modify-syntax-entry i "." gimml-mode-syntax-table)
      (setq i (1+ i)))
    (setq i (1+ ?Z))
    (while (< i ?a)
      (modify-syntax-entry i "." gimml-mode-syntax-table)
      (setq i (1+ i)))
    (setq i (1+ ?z))
    (while (< i 128)
      (modify-syntax-entry i "." gimml-mode-syntax-table)
      (setq i (1+ i))))

  ;; Now we change the characters that are meaningful to us.
  (modify-syntax-entry ?\(      "()1"   gimml-mode-syntax-table)
  (modify-syntax-entry ?\)      ")(4"   gimml-mode-syntax-table)
  (modify-syntax-entry ?\[      "(]"    gimml-mode-syntax-table)
  (modify-syntax-entry ?\]      ")["    gimml-mode-syntax-table)
  (modify-syntax-entry ?{       "(}"    gimml-mode-syntax-table)
  (modify-syntax-entry ?}       "){"    gimml-mode-syntax-table)
  (modify-syntax-entry ?\*      ". 23"  gimml-mode-syntax-table)
  (modify-syntax-entry ?\"      "\""    gimml-mode-syntax-table)
  (modify-syntax-entry ?        " "     gimml-mode-syntax-table)
  (modify-syntax-entry ?\t      " "     gimml-mode-syntax-table)
  (modify-syntax-entry ?\n      " "     gimml-mode-syntax-table)
  (modify-syntax-entry ?\f      " "     gimml-mode-syntax-table)
  (modify-syntax-entry ?\'      "w"     gimml-mode-syntax-table)
  (modify-syntax-entry ?\_      "w"     gimml-mode-syntax-table)
  (modify-syntax-entry ?| "w" gimml-mode-syntax-table)
  (modify-syntax-entry ?\\ "\\" gimml-mode-syntax-table))


(if gimml-emacs19-p ; emacs 19: then this is comment-indent-function
					; instead of comment-indent-hook
    (defun set-comment-hook ()
      (make-local-variable 'comment-indent-function)
      (setq comment-indent-function 'gimml-comment-indent))
  (defun set-comment-hook ()
    (make-local-variable 'comment-indent-hook)
    (setq comment-indent-hook 'gimml-comment-indent)))

(defun gimml-mode ()
  "Major mode for editing Gimml code.
Tab indents for Gimml code.
Comments are delimited with (* ... *).
Paragraphs are separated by blank lines only.
Delete converts tabs to spaces as it moves back.

Default key bindings:
=====================

TAB      -  Indent current line.
C-c TAB  -  Indent region.
LFD      -  Reindent line, newline and indent.
ESC TAB  -  Back line out one indentation.
ESC |    -  Insert a \"|\". Insert function name, \"=>\" etc.
C-c RET  -  Insert a common used structure.
C-c C-n  -  Find the next error.
C-c C-p  -  Find previous error.
C-c C-s  -  Pop to the gimml window.
C-c C-u  -  Save the buffer, and send a \"use file\".
C-c C-r  -  Send region (point and mark) to gimml.
C-c C-f  -  Send a \"use file\" to gimml.
C-c C-c  -  Send function/region to gimml.
C-c C-b  -  Send whole buffer to gimml.
C-c C-v  -  Get the version of gimml-mode.


Variables controlling the indentation
=====================================

gimml-indent-level (default 4)
    The indentation of a block of code.

gimml-pipe-indent (default -2)
    Extra indentation of a line starting with \"|\".

gimml-case-indent (default nil)
    Determine the way to indent case-of expression.
    If t:   case expr              If nil:   case expr of
              of exp1 => ...                     exp1 => ...
               | exp2 => ...                   | exp2 => ...

    The first seems to be the standard in NJ-SML. The second is the default.

gimml-nested-if-indent (default t)
    If set to t, nested if-then-else expression will have the same
    indentation as:
                     if exp1 then exp2
                     else if exp3 then exp4
                     else if exp5 then exp6
                          else exp7

gimml-type-of-indent (default t)
    How to indent `let' `struct' etc.

    If t:  fun foo bar = let                If nil:  fun foo bar = let
                             val p = 4                   val p = 4
                         in                          in
                             bar + p                     bar + p
                         end                         end

    Will not have any effect if the starting keyword is first on the line.

gimml-electric-semi-mode (default nil)
    If t, a `\;' will reindent line, and perform a newline.

gimml-paren-lookback (default 2000)
    Determines how far back (in chars) the indentation algorithm
    should look for open parenthesis. High value means slow indentation
    algorithm. A value of 2000 (being the equivalent of 40-60 lines) should
    suffice most uses. (A value of nil, means do not look at all)

Mode map
========
\\{gimml-mode-map}
See gimml-shell for further information.

Runs gimml-mode-hook if non nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map gimml-mode-map)
  (setq major-mode 'gimml-mode)
  (setq mode-name "Gimml")
  (define-abbrev-table 'gimml-mode-abbrev-table ())
  (setq local-abbrev-table gimml-mode-abbrev-table)
  (set-syntax-table gimml-mode-syntax-table)
  ;; A paragraph is seperated by blank lines or ^L only.
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^[\t ]*$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'gimml-indent-line)
  (make-local-variable 'require-final-newline) ; Always put a new-line
  (setq require-final-newline t)	       ; in the end of file
  (make-local-variable 'comment-start)
  (setq comment-start "(* ")
  (make-local-variable 'comment-end)
  (setq comment-end " *)")
  (make-local-variable 'comment-column)
  ;; Start of comment in this column. Changed from origional 39.
  ;; I still don't know where I got the 39 from !! (891117 lbn) 
  (setq comment-column 40)		
  (make-local-variable 'comment-start-skip)
  ;; This matches a start of comment (I sure hope!)
  (setq comment-start-skip "(\\*+[ \t]?")
  (set-comment-hook)
  ;;
  ;; Adding these will fool the matching of parens. I really don't
  ;; know why. It would be nice to have comments treated as
  ;; white-space
  ;; 
  ;; (make-local-variable 'parse-sexp-ignore-comments)
  ;; (setq parse-sexp-ignore-comments t)
  ;;
  (run-hooks 'gimml-mode-hook))		; Run the hook

(defconst gimml-pipe-matchers-reg
  "\\bcase\\b\\|\\bfn\\b\\|\\bfun\\b\\|\\bmemofun\\b\\|\\bmemofn\\b\\|\
\\bhandle\\b\
\\|\\bdatatype\\b\\|\\babstype\\b\\|\\band\\b\\|{\\|\\[\
\\|\\ball\\b\\|\\bexists\\b\\|\\bsome\\b\\|\\biterate\\b\
\\|\\bwith\\b"
  "The keywords a `|' can follow.")

(defconst gimml-openings
  (vector "\\blet\\b\\|\\blocal\\b\\|\\bsig\\b\\|\\bstruct\\b\
\\|\\ball\\b\\|\\bexists\\b\\|\\bsome\\b\\|\\biterate\\b"
	  "{"
	  "\\["
	  "\\bif\\b"
	  "(")
  "The vector of opening regular expressions")

(defconst gimml-closings
  (vector "\\bend\\b"
	  "}"
	  "]"
	  "\\belse\\b"
	  ")")
  "The vector of closing regular expressions")

(defconst gimml-closing-quotes
  (vector "end"
	  "}"
	  "]"
	  "else"
	  "*)"
	  ")")
  "The vector of texts corresponding to closing regular expressions")

(defconst gimml-in-regexp
  "in\\b\
[ \\t\\n]*\\($\\|\\B\\|\\<\\([^sl]\\|s$\\|s[^e]\\|se$\\|se[^t]\\|set\\B\
\\|l$\\|l[^i]\\|li$\\|li[^s]\\|lis$\\|lis[^t]\\|list\\B\\)\\)"
  "The regular expression matching 'in' but neither 'in set' nor 'in list'")

(defun gimml-electric-pipe ()
  "Insert a \"|\". Depending on the context insert the name of
function, a \"=>\" etc."
  (interactive)
  (let ((case-fold-search nil)		; Case sensitive
	;(here (point))
	(match (save-excursion
		 (gimml-find-matching-starter gimml-pipe-matchers-reg)
		 (point)))
	(tmp "  => ")
	(case-or-handle-exp t))
    (if (/= (save-excursion (beginning-of-line) (point))
	    (save-excursion (skip-chars-backward "\t ") (point)))
	(insert "\n"))
    (insert "|")
    (save-excursion
      (goto-char match)
      (cond
       ;; It was a function, insert the function name
       ((looking-at "fun\\b")
	(setq tmp (concat " " (buffer-substring
			       (progn (forward-char 3)
				      (gimml-skip-chars-forward "\t\n ") (point))
			       (progn (forward-word 1) (point))) " "))
	(setq case-or-handle-exp nil))
       ((looking-at "with\\b")
	(setq tmp (concat " " (buffer-substring
			       (progn (forward-char 4)
				      (gimml-skip-chars-forward "\t\n ") (point))
			       (progn (forward-word 1) (point))) " "))
	(setq case-or-handle-exp nil))
       ((looking-at "memofun\\b")
	(setq tmp (concat " " (buffer-substring
			       (progn (forward-char 7)
				      (gimml-skip-chars-forward "\t\n ") (point))
			       (progn (forward-word 1) (point))) " "))
	(setq case-or-handle-exp nil))
       ;; It was a datatype, insert nothing
       ((looking-at "datatype\\b\\|abstype\\b\\|{\\|\\[\\|all\\b\\|exists\\b\\|some\\b\\|iterate\\b")
	(setq tmp " ") (setq case-or-handle-exp nil))
       ;; If it is an and, then we have to see what is was
       ((looking-at "and\\b")
	(let (isfun)
	  (save-excursion
	    (condition-case ()
		(progn
		  (re-search-backward "datatype\\b\\|abstype\\b\\|fun\\b")
		  (setq isfun (looking-at "fun\\b")))
	      (error (setq isfun nil))))
	  (if isfun
	      (progn
		(setq tmp
		      (concat " " (buffer-substring
				   (progn (forward-char 3)
					  (gimml-skip-chars-forward "\t\n ") (point))
				   (progn (forward-word 1) (point))) " "))
		(setq case-or-handle-exp nil))
	    (setq tmp " ") (setq case-or-handle-exp nil))))))
    (insert tmp)
    (gimml-indent-line)
    (back-to-indentation)
    (forward-char (1+ (length tmp)))
    (if case-or-handle-exp
	(forward-char -4))))

(defun gimml-electric-semi ()
  "Inserts a \;.
If gimml-electric-semi-mode is t, indent the current line, and newline."
  (interactive)
  (insert "\;")
  (if gimml-electric-semi-mode
      (reindent-then-newline-and-indent)))

(defun gimml-mode-version ()
  "Message the version of gimml-mode."
  (interactive)
  (message gimml-mode-version-string))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; SHORT CUTS (gimml-region)
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst gimml-region-alist
  '(("let") ("local") ("signature") ("structure") ("datatype")
    ("case") ("functor") ("abstype") ("abstraction"))
  "The list of regions to auto-insert.")

(defun gimml-region ()
  "Interactive short-cut. Insert a common used structure in gimml."
  (interactive)
  (let ((newline nil)			; Did we insert a newline
	(name (completing-read "Region to insert: (default let) "
			       gimml-region-alist nil t nil)))
    ;; default is "let"
    (if (string= name "") (setq name "let"))
    ;; Insert a newline if point is not at empty line
    (gimml-indent-line)			; Indent the current line
    (if (save-excursion (back-to-indentation) (eolp))
	()
      (setq newline t)
      (insert "\n"))
    (condition-case ()
	(cond
	 ((string= name "let") (gimml-let))
	 ((string= name "local") (gimml-local))
	 ((string= name "structure") (gimml-structure))
	 ((string= name "signature") (gimml-signature))
	 ((string= name "abstraction") (gimml-abstraction))
	 ((string= name "functor") (gimml-functor))
	 ((string= name "case") (gimml-case))
	 ((string= name "abstype") (gimml-abstype))
	 ((string= name "datatype") (gimml-datatype)))
      (quit (if newline 
		(progn
		  (delete-char -1)
		  (beep)))))))

(defun gimml-let () 
  "Insert a `let in end'."
  (interactive) (gimml-let-local "let"))

(defun gimml-local ()
  "Insert a `local in end'."
  (interactive) (gimml-let-local "local"))

(defun gimml-signature ()
  "Insert a `signature ??? = sig end', prompting for name."
  (interactive) (gimml-structure-signature "signature"))

(defun gimml-structure ()
  "Insert a `structure ??? = struct end', prompting for name."
  (interactive) (gimml-structure-signature "structure"))

(defun gimml-case ()
  "Insert a case, prompting for case-expresion."
  (interactive)
  (let (indent (expr (read-string "Case expr: ")))
    (insert (concat "case " expr))
    (gimml-indent-line)
    (setq indent (current-indentation))
    (end-of-line)
    (if gimml-case-indent
	(progn
	  (insert "\n")
	  (indent-to (+ 2 indent))
	  (insert "of "))
      (insert " of\n")
      (indent-to (+ indent gimml-indent-level)))
    (save-excursion (insert " => "))))

(defun gimml-let-local (starter)
  (let (indent)
    (insert starter)
    (gimml-indent-line)
    (setq indent (current-indentation))
    (end-of-line)
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "\n") (indent-to indent)
    (insert "in\n") (indent-to (+ gimml-indent-level indent))
    (insert "\n") (indent-to indent)
    (insert "end") (previous-line 3) (end-of-line)))

(defun gimml-structure-signature (which)
  (let (indent
	(name (read-string (concat "Name of " which ": "))))
    (insert (concat which " " name " ="))
    (gimml-indent-line)
    (setq indent (current-indentation))
    (end-of-line)
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert (if (string= which "signature") "sig\n" "struct\n"))
    (indent-to (+ (* 2 gimml-indent-level) indent))
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "end") (previous-line 1) (end-of-line)))

(defun gimml-functor ()
  "Insert a `functor ??? () : ??? = struct end', prompting for name and type."
  (let (indent
	(name (read-string "Name of functor: "))
	(signame (read-string "Signature type of functor: ")))
    (insert (concat "functor " name " () : " signame " ="))
    (gimml-indent-line)
    (setq indent (current-indentation))
    (end-of-line)
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "struct\n")
    (indent-to (+ (* 2 gimml-indent-level) indent))
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "end") (previous-line 1) (end-of-line)))

(defun gimml-abstraction ()
  "Insert a `abstraction ??? : ??? = struct end', prompting for name and type."
  (let (indent
	(name (read-string "Name of abstraction: "))
	(signame (read-string "Signature type of abstraction: ")))
    (insert (concat "abstraction " name " : " signame " ="))
    (gimml-indent-line)
    (setq indent (current-indentation))
    (end-of-line)
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "struct\n")
    (indent-to (+ (* 2 gimml-indent-level) indent))
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "end") (previous-line 1) (end-of-line)))

(defun gimml-datatype ()
  "Insert a `datatype ??? =', prompting for name."
  (let (indent 
	(type (read-string (concat "Type of datatype (default none): ")))
	(name (read-string (concat "Name of datatype: "))))
    (insert (concat "datatype "
		    (if (string= type "") "" (concat type " "))
		    name " ="))
    (gimml-indent-line)
    (setq indent (current-indentation))
    (end-of-line) (insert "\n") (indent-to (+ gimml-indent-level indent))))

(defun gimml-abstype ()
  "Insert an `abstype 'a ??? = with ... end'"
  (let (indent
	(typevar (read-string "Name of typevariable (default 'a): "))
	(type (read-string "Name of abstype: ")))
    (if (string= typevar "")
	(setq typevar "'a"))
    (insert (concat "abstype " typevar " " type " ="))
    (gimml-indent-line)
    (setq indent (current-indentation))
    (insert "\n") (indent-to (+ gimml-indent-level indent))
    (insert "\n") (indent-to indent)
    (insert "with\n") (indent-to (+ gimml-indent-level indent))
    (insert "\n") (indent-to indent)
    (insert "end")
    (previous-line 3)
    (end-of-line)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; PARSING ERROR MESSAGES (NOTE: works only with Gimml)
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; The reg-expression used when looking for errors.
;; An example of error message is:
;; Error: a.ml, line 2(3)-2(5) : can only apply functions

;; On the Mac (before MacOSX), Gimml uses error messages like:
;;  File "a.ml"; Line 2!3:2!5 # can only apply functions
;; the file and line descriptors follow the MPW Shell conventions;
;; moreover there is no difference between errors and warnings.
(defvar gimml-error-regexp
  "\\(Error:\\|Warning:\\|[^\n;]*;in\\)[ \t]*\\([^,\n]+\\),[ ]*lines?[ ]*\\([0-9()-]+\\)"
  "Regular expression for matching error. If you change this beware of the
second to fourth parenthesized expressions. gimml-next-error uses these to
get the filename, and the line and character offset.")

(defvar gimml-hilit-face
  (if (and (fboundp 'make-face) (fboundp 'set-face-background) (fboundp 'copy-face))
      (progn
	(if (or (and (fboundp 'x-color-display-p) (x-color-display-p)) ; Lucid Emacs 19
		(and (fboundp 'x-display-color-p) (x-display-color-p))) ; GNU Emacs 19
	    (progn
	      (make-face 'gimml-hilit-face)
	      (set-face-background 'gimml-hilit-face "lightblue"))
	  (if (and (fboundp 'facep) (facep 'highlight))
	      (copy-face 'highlight 'gimml-hilit-face)
	    (copy-face 'bold 'gimml-hilit-face)))
	'gimml-hilit-face)
    nil ;;otherwise, we cannot highlight
    )
  "Face used to highlight regions of text when displaying errors through gimml-display-error.")

(defvar gimml-hilit-extent nil
  "The current highlighted extent when displaying errors through gimml-display-error.")

(if (fboundp 'defalias)
    ()
  (defun defalias (sym func)
    (fset sym func)))

(if gimml-emacs19-p ; emacs 19: then
    ; build a compatibility library with Lucid for highlighting
    ; Lucid's extents are mapped to GNU's overlays.
    (if (and (fboundp 'make-overlay)
	     (fboundp 'delete-overlay)
	     (fboundp 'overlay-buffer)
	     (fboundp 'overlayp))
	(progn
	  (if (fboundp 'make-extent)
	      ()
	    (defalias 'make-extent (symbol-function 'make-overlay)))
	  (if (fboundp 'delete-extent)
	      ()
	    (defalias 'delete-extent (symbol-function 'delete-overlay)))
	  (if (fboundp 'set-extent-face)
	      ()
	    (defun set-extent-face (extent face)
	      (overlay-put extent 'face face)))
	  (if (fboundp 'extent-buffer)
	      ()
	    (defalias 'extent-buffer (symbol-function 'overlay-buffer)))
	  (if (fboundp 'extentp)
	      ()
	    (defalias 'extentp (symbol-function 'overlayp)))
	  )
      )
  )

(defun gimml-hilit-p ()
  (and gimml-hilit-face (fboundp 'set-extent-face) (fboundp 'delete-extent)))

(defun gimml-parse-error ()
  ; parses susbtrings given as match-beginning/match-end pairs
  ; after recognition of gimml-error-regexp, and returns
  ; (filename (line1 . pos1) . (line2 . pos2))
  (let ((filename (buffer-substring (match-beginning 2) (match-end 2)) )
	(positions (buffer-substring (match-beginning 3) (match-end 3)) )
	line1 line2 (pos1 0) (pos2 0) )
    (string-match "^\\([0-9]+\\)" positions) ;; should not fail
    (setq line1 (string-to-int (substring positions (match-beginning 1) (match-end 1))))
    (setq line2 line1)
    (setq positions (substring positions (match-end 0)))
    (if (string-match "^(\\([0-9]+\\))" positions)
	(progn (setq pos1 (string-to-int (substring positions
						    (match-beginning 1)
						    (match-end 1))))
	       (setq positions (substring positions (match-end 0)))
	       (setq pos2 pos1)) )
    (if (string-match "^-\\([0-9]+\\)" positions)
	(progn (setq line2 (string-to-int (substring positions
						     (match-beginning 1)
						     (match-end 1))))
	       (setq positions (substring positions (match-end 0)))
	       (setq pos2 0)) )
    (if (string-match "^(\\([0-9]+\\))" positions)
	(progn (setq pos2 (string-to-int (substring positions
						    (match-beginning 1)
						    (match-end 1)))) ))
    (cons filename (cons (cons line1 pos1) (cons line2 pos2))) ) )

(defun gimml-delete-extent ()
  (if (gimml-hilit-p)
      (and (extentp gimml-hilit-extent) (buffer-name (extent-buffer gimml-hilit-extent))
	   (progn
	     (delete-extent gimml-hilit-extent)
	     (setq gimml-hilit-extent nil)))
    (setq overlay-arrow-string nil)))

(defun gimml-display-error-1 (file line1 pos1 line2 pos2 msg)
  (let* ((buffer (find-file-noselect file))
	 (window (display-buffer buffer t))
	 start end)
    (save-excursion
      (set-buffer buffer)
      (save-restriction
	(widen)
	(goto-line line1)
	(beginning-of-line)
	(if (gimml-hilit-p)
	    () ;; only display overlay arrow if we cannot hilite
	  ;; hiliting is better visually; moreover, the overlay arrow
	  ;; does not appear under Lucid Emacs 19.4.3 (at least)
	  (setq overlay-arrow-string "=>")
	  (or overlay-arrow-position
	      (setq overlay-arrow-position (make-marker)))
	  (set-marker overlay-arrow-position (point) (current-buffer)) )
	(forward-char pos1)
	(setq start (point))
	(save-excursion
	  (goto-line line2)
	  (forward-char pos2)
	  (setq end (point)))
	(if (gimml-hilit-p)
	    (progn ;; delete previous extent if it is an extent and its buffer is still alive
	      (gimml-delete-extent)
	      (setq gimml-hilit-extent (make-extent start end))
	      (set-extent-face gimml-hilit-extent gimml-hilit-face)) )
	)
      (cond ((or (< start (point-min)) (> end (point-max)))
	     (widen)
	     (goto-char start)))
      (if msg
	  (let* ((buf (get-buffer-create gimml-message-buffer-name))
		 (win (get-buffer-window buf)))
	    (if (null win)
		(setq win (split-window window (- (window-height)
						  gimml-message-buffer-height))) )
	    (set-window-buffer win buf)
	    (set-buffer buf)
	    (erase-buffer)
	    (insert msg) )
	(let ((buf (get-buffer gimml-message-buffer-name)))
	  (if buf
	      (let ((win (get-buffer-window buf)))
		(if win
		    (delete-window win))))) )
	)
    (set-window-point window start)
    ))

(defun gimml-display-error (position)
  (let* ((pos (car position))
	 (msg (cdr position))
	 (file (car pos))
	 (pos (cdr pos))
	 (line1 (car (car pos)))
	 (pos1 (cdr (car pos)))
	 (pos (cdr pos))
	 (line2 (car pos))
	 (pos2 (cdr pos)))
    (gimml-display-error-1 file line1 pos1 line2 pos2 msg)))

(defun gimml-get-msg ()
  (if (looking-at " :") ;; this is after an error or a warning
      (progn
	(skip-chars-forward "[ :\t\n]+")
	(buffer-substring (point)
			  (progn 
			    (re-search-forward (concat "\\("
						       gimml-shell-prompt-pattern
						       "\\)\\|^Error:\\|^Warning:")
					       (point-max)
					       'gimml-next-error
					       1)
			    (beginning-of-line)
			    (point))) )
    (forward-line)
    nil))

(defun gimml-find-next-error ()
  (condition-case ()
      (progn
	(beginning-of-line)
	(re-search-forward gimml-error-regexp) ; Search for error
	(save-excursion
	  (cons (gimml-parse-error) (gimml-get-msg))))
    (error ())))

(defun gimml-find-prev-error ()
  (condition-case ()
      (progn
	(re-search-backward gimml-error-regexp) ; Search for error
	(cons (gimml-parse-error)
	      (save-excursion
		(re-search-forward gimml-error-regexp)
		(gimml-get-msg))))
    (error ())))
 
(defun gimml-next-error (arg)
  "Find the next error under a Gimml toplevel by parsing the current buffer.
A non-nil argument (prefix arg, if interactive) means reparse the
error messages from the last erroneous toplevel command down and
start at the first error."
  (interactive "P")
  (let ((case-fold-search nil)
	position)
    (if arg
	(while (and (re-search-backward gimml-shell-prompt-pattern
					(point-min)
					'gimml-next-error 1)
		    (let ((p (point)))
		      (setq position (gimml-find-next-error))
		      (if position
			  nil
			(progn (goto-char p) t))) ) )
      (beginning-of-line)
      (if (looking-at gimml-error-regexp) ;;skip current error if we are looking at one
	  (forward-line))
      (setq position (gimml-find-next-error)) )
    (if position
	(gimml-display-error position)
      (gimml-delete-extent)
      (message "No more errors")	; Or have we passed all errors
      (beep))) )

(defun gimml-prev-error (arg)
  "Find the previous error under a Gimml toplevel by parsing the current buffer.
A non-nil argument (prefix arg, if interactive) means reparse
the error messages from the bottom of the buffer up and start
at the last error."
  (interactive "P")
  (let ((case-fold-search nil)
	position)
    (if arg
	(goto-char (point-max)) )
    (setq position (gimml-find-prev-error))
    (if position
	(gimml-display-error position)
      (message "No more errors")	; Or have we passed all errors
      (beep))) )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; HIGHLIGHTING
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(if gimml-emacs20-p
    ()
(if gimml-emacs19-p ; emacs 19
    (progn
      (require 'hilit19)
      (let ((comments     '(("(\\*" "\\*)" comment)))
	    (strings      '((hilit-string-find ?' string)))
	    (preprocessor '(("^#[ \t]*\\(undef\\|define\\).*$" "[^\\]$" define)
			    ("^#.*$" nil include))))

	(hilit-set-mode-patterns
	 'gimml-mode
	 (append
	  comments strings preprocessor
	  '(
	    ;; declarations
	    ("\\b\\(fun\\|memofun\\|with\\)\\b[^\n=]*[\n=]" nil defun)
	    ("\\b\\(extern\\|val\\|and\\|exception\\|type\\|datatype\\|abstype\\|withtype\\)\\b[^\n=]*[\n=]" nil decl)
	    ;; key words
	    ("[^_]\\(\\<\\(\\|if\\|then\\|else\\|fn\\|memofn\\|case\\|of\\|while\\|do\\|all\\|exists\\|some\\|iterate\\|let\\|end\\|handle\\)\\>\\|=>\\|{}\\|{\\|}\\||\\[\\]|\\||\\[\\|\\]|\\|\\[\\]\\|\\[\\|\\]\\||\\)[^_]" 1 keyword)
      ))))))
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; INDENTATION
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun gimml-indent-region (begin end)
  "Indent region of gimml code."
  (interactive "r")
  (message "Indenting region...")
  (save-excursion
    (goto-char end) (setq end (point-marker)) (goto-char begin)
    (while (< (point) end)
      (skip-chars-forward "\t\n ")
      (gimml-indent-line)
      (end-of-line))
    (move-marker end nil))
  (message "Indenting region... done"))

(defun gimml-indent-line1 ()
  (let ((indent (gimml-calculate-indentation)))
    (if (/= (current-indentation) indent)
	(save-excursion			;; Added 890601 (point now stays)
	  (let ((beg (progn (beginning-of-line) (point))))
	    (back-to-indentation)
	    (delete-region beg (point))
	    (indent-to indent))))
    ;; If point is before indentation, move point to indentation
    (if (< (current-column) (current-indentation))
	(back-to-indentation))))

(defun gimml-indent-line ()
  "Indent current line of gimml code."
  (interactive)
  (if (eq major-mode 'gimml-shell)
      (if (save-excursion
	    (re-search-backward (concat gimml-shell-prompt-pattern " ?")
				(point-min) t))
	  (save-restriction
	    (narrow-to-region (match-end 0) (point-max))
	    (gimml-indent-line1))
	(gimml-indent-line1))
    (gimml-indent-line1)))

(defun gimml-back-to-outer-indent ()
  "Unindents to the next outer level of indentation."
  (interactive)
  (save-excursion
    (back-to-indentation)
    (let ((start-column (current-column))
	  (indent (current-column)))
      (if (> start-column 0)
	  (progn
	    (save-excursion
	      (while (>= indent start-column)
		(if (re-search-backward "^[^\n]" nil t)
		    (setq indent (current-indentation))
		  (setq indent 0))))
	    (backward-delete-char-untabify (- start-column indent)))))))

(defconst gimml-indent-starters-reg
  (concat
   "abstraction\\b\\|abstype\\b\\|and\\b\\|case\\b\\|datatype\\b\
\\|else\\b\\|fun\\b\\|memofun\\b\\|fn\\b\\|memofn\\b\\|functor\\b\\|if\\b\\|sharing\\b\\|in\\b\\|sub[ \t]map\\b\
\\|infix\\b\\|infixr\\b\\|let\\b\\|local\\b\
\\|nonfix\\b\\|open\\b\\|raise\\b\\|sig\\b\\|signature\\b\
\\|struct\\b\\|structure\\b\\|then\\b\\|\\btype\\b\\|val\\b\
\\|while\\b\\|with\\b\\|withtype\\b\\|extern\\b\\|such[ \t]that\\b\
\\|[{[(]\\|\\ball\\b\\|\\bsome\\b\\|\\bexists\\b\\|\\biterate\\b")
  "The indentation starters. The next line, after one starting with
one of these, will be indented.")

(defconst gimml-starters-reg
  "\\babstraction\\b\\|\\babstype\\b\\|\\bdatatype\\b\
\\|\\bexception\\b\\|\\bfun\\b\\|\\bmemofun\\b\\|\
\\bfunctor\\b\\|\\blocal\\b\
\\|\\binfix\\b\\|\\binfixr\\b\\|\\bsharing\\b\
\\|\\bnonfix\\b\\|\\bopen\\b\\|\\bsignature\\b\\|\\bstructure\\b\
\\|\\btype\\b\\|\\bval\\b\\|\\bwithtype\\b\\|\\bwith\\b\\|\\bextern\\b\
\\|[{[(]\\|\\ball\\b\\|\\bsome\\b\\|\\bexists\\b\\|\\biterate\\b"
  "The starters of new expressions.")

(defconst gimml-decl-starters-reg
  "\\babstraction\\b\\|\\babstype\\b\\|\\bdatatype\\b\
\\|\\bexception\\b\\|\\bfun\\b\\|\\bmemofun\\b\\|\
\\bfunctor\\b\\|\\blocal\\b\
\\|\\binfix\\b\\|\\binfixr\\b\\|\\bsharing\\b\
\\|\\bnonfix\\b\\|\\bopen\\b\\|\\bsignature\\b\\|\\bstructure\\b\
\\|\\btype\\b\\|\\bval\\b\\|\\bwithtype\\b\\|\\bwith\\b\\|\\bextern\\b"
  "The starters of new declarations.")

(defconst gimml-end-starters-reg
  "\\blet\\b\\|\\blocal\\b\\|\\bsig\\b\\|\\bstruct\\b\\|\\bwith\\b\
\\|\\ball\\b\\|\\bexists\\b\\|\\bsome\\b\\|\\biterate\\b"
  "Matching reg-expression for the \"end\" keyword.")

(defconst gimml-starters-indent-after
  "let\\b\\|local\\b\\|struct\\b\\|in\\b\\|sig\\b\\|[{[(]"
  "Indent after these.")

(defconst gimml-starters-for-pipe
  "\\bfun\\b\\|\\bfn\\b\\|\\bmemofun\\b\\|\\bmemofn\\b\\|\\band\\b\
\\|\\bhandle\\b\\|\\bcase\\b\\|\\bdatatype\\b\\|\\bwith\\b\
\\|{\\|\\[\\|\\ball\\b\\|\\bexists\\b\\|\\bsome\\b\\|\\biterate\\b"
  "Matching reg-expression for the | separator.")

(defconst gimml-pipe-regexp
  "|\\($\\|[^[]\\)"
  "Regexp matching the pipe |, but not the open record delimiter |[.")

(defun gimml-looking-at-match (prefix)
  (and (looking-at prefix)
       (save-excursion
	 (goto-char (match-end 0))
	 (let ((start (point))
	       (matchp t))
	   (while (and matchp (not (looking-at "[ \t]*=>")))
	     (setq matchp (not (or (looking-at "[ \t]*=")
				   (gimml-forward-sexp)
				   (eobp)))))
	   (and matchp
		(progn
		  (goto-char start)
		  (and (gimml-find-matching-starter gimml-starters-for-pipe)
		       (looking-at "fn\\b\\|memofn\\|handle\\b\\|case\\b\\|datatype\\b")
		       (point)
		       ))) )) ))

(defun gimml-indentation-at-starter (offset)
  (cond
   ((looking-at "fun\\b")
    (+ (current-column) 4))
   ((looking-at "memofun\\b")
    (+ (current-column) 8))
   ((looking-at "fn\\b") (+ (current-column) 3))
   ((looking-at "memofn\\b") (+ (current-column) 7))
   ((looking-at "and\\b") (+ (current-column) 4))
   ((looking-at "handle\\b") (+ (current-column) 7))
   ((looking-at "with") (+ (current-column) 5))
   ((looking-at "datatype") (+ (current-column) 9))
   ((looking-at "[[{]\\|some\\b\\|exists\\b\\|all\\|iterate\\b")
    (+ (current-column) 2))
   (t (+ (current-column) offset))))

(defun gimml-calculate-indentation ()
  (save-excursion
    (let ((case-fold-search nil) match)
      (beginning-of-line)
      (if (bobp)			; Beginning of buffer
	  0				; Indentation = 0
	(back-to-indentation)
	(if (gimml-inside-comment-or-string-p)
	    (gimml-get-paren-indent)
	  (cond
	   ;; Indentation for comments alone on a line, matches the
	   ;; proper indentation of the next line. Search only for the
	   ;; next "*)", not for the matching.
	   ((looking-at "(\\*")
	    (if (not (search-forward "*)" nil t))
		(error "Comment not ended."))
	    (end-of-line)
	    (skip-chars-forward "\n\t ")
	    ;; If we are at eob, just indent 0
	    (if (eobp) 0 (gimml-calculate-indentation)))
	   ;; Continued string ? (Added 890113 lbn)
	   ((looking-at "\\\\")
	    (save-excursion
	      (if (save-excursion (previous-line 1)
				  (beginning-of-line)
				  (looking-at "[\t ]*\\\\"))
		  (progn (previous-line 1) (current-indentation))
		(if (re-search-backward "[^\\\\]\"" nil t)
		    (1+ (current-indentation))
		  0))))
	   ;; Are we looking at an alternative (| pattern =>)?
	   ((setq match (gimml-looking-at-match gimml-pipe-regexp))
	    (save-excursion
	      (goto-char match)
	      (cond
	       ((looking-at "fn\\b") (1+ (current-column)))
	       ((looking-at "memofn\\b\\|handle\\b") (+ 5 (current-column)))
	       ((looking-at "case\\b") (+ 4 (current-column) gimml-pipe-indent))
	       (t (+ (current-column) gimml-pipe-indent)))))
	   ;; Are we looking at the first alternative of a case expression?
	   ((and (save-excursion
		   (gimml-backward-sexp)
		   (looking-at "of\\b"))
		 (setq match (gimml-looking-at-match "")))
	    (save-excursion
	      (goto-char match)
	      (+ 4 (current-column))))
	   ((looking-at "and\\b")
	    (if (gimml-find-matching-starter gimml-starters-reg)
		(current-column)
	      0))
	   ((looking-at gimml-in-regexp)	; Match the beginning let/local
	    (gimml-find-match-indent "in" 0 nil "\\blocal\\b\\|\\blet\\b"))
	   ((looking-at "end\\b")		; Match the beginning
	    (gimml-find-match-indent "end" 0 nil))
	   ((and gimml-nested-if-indent (looking-at "else\\b")
		 (setq match
		       (save-excursion (and (gimml-re-search-backward "\\bif\\b\\|\\belse\\b")
					    (back-to-indentation)
					    (looking-at "\\bif\\b\\|\\belse[\t ]*if\\b")
					    (current-indentation)))))
	    (or match (gimml-find-match-indent "else" 3 t)))
	   ((looking-at "else\\b")	; Match the if
	    (gimml-find-match-indent "else" 3 t))
	   ((looking-at "then\\b")	; Match the if or else if + extra indentation
	    (+ (or (and gimml-nested-if-indent
			(save-excursion
			  (and (gimml-re-search-backward "\\bif\\b\\|\\belse\\b")
			       (back-to-indentation)
			       (looking-at "\\bif\\b\\|\\belse[\t ]*if\\b")
			       (current-indentation))))
		   (gimml-find-match-indent "then" 3 t))
	       gimml-indent-level))
	   ((and gimml-case-indent (looking-at "of\\b"))
	    (gimml-re-search-backward "\\bcase\\b")
	    (+ (current-column) 2))
	   ((looking-at "such[ \t]that\\b")
	    (while (and (not (bobp))
			(gimml-find-matching-starter gimml-pipe-regexp)
			(gimml-looking-at-match gimml-pipe-regexp)) )
	    (+ (if gimml-type-of-indent
		   (current-column)
		 (current-indentation))
	       gimml-indent-level))
	   ((looking-at gimml-decl-starters-reg)
	    (let ((start (point)))
	      (if (gimml-find-matching-starter gimml-starters-reg)
		  (if (and (looking-at gimml-starters-indent-after)
			   (/= start (point)))
		      (+ (if gimml-type-of-indent
			     (current-column)
			   (if (progn (back-to-indentation)
				      (looking-at "|\\($\\|[^[]\\)"))
			       (- (current-indentation) gimml-pipe-indent)
			     (current-indentation)))
			 gimml-indent-level)
		    (let ((c (current-column)))
		      (back-to-indentation)
		      (if (looking-at "extern\\b")
			  (current-column)
			c)))
		0)))
	   (t
	    (let ((indent (gimml-get-indent)))
	      (cond
	       ((looking-at gimml-pipe-regexp)
		;; Lets see if it is the follower of a function definition
		(if (gimml-find-matching-starter gimml-starters-for-pipe)
		    (+ (gimml-indentation-at-starter 0) gimml-pipe-indent)
		  (+ indent gimml-pipe-indent)))
	       (t
		(if gimml-paren-lookback	; Look for open parenthesis ?
		    (max indent (gimml-get-paren-indent))
		  indent)))))))))))

(defun gimml-get-indent ()
  (save-excursion
    (let ((case-fold-search nil))
      (beginning-of-line)
      (gimml-skip-chars-backward "\t\n; ")
      (if (looking-at ";") (gimml-backward-sexp))
      (cond
       ((save-excursion (gimml-backward-sexp) (looking-at "end\\b"))
	;(- (current-indentation) gimml-indent-level)
	(current-indentation)
	)
       ;((save-excursion (back-to-indentation)
	;		(looking-at gimml-pipe-regexp))
	;(- (current-indentation) gimml-pipe-indent))
       (t
	(while (and (/= (current-column) (current-indentation))
		    (if (gimml-backward-sexp)
			(not (looking-at gimml-indent-starters-reg))
		      (not (looking-at gimml-decl-starters-reg)))))
	;(gimml-skip-chars-forward "\t |")
	(let ((indent (current-indentation)))
	  (gimml-skip-chars-forward "\t ")
	  (cond
	   ;; Started val/fun/structure...
	   ((looking-at "||")
	    (+ indent 3))
	   ((looking-at "|")
	    (+ indent 2))
	   ((looking-at "[([{]")
	    indent)
	   ((looking-at gimml-indent-starters-reg)
	    (gimml-indentation-at-starter gimml-indent-level))
	   ;; else keep the same indentation as previous line
	   (t indent))))))))

(defun gimml-get-paren-indent ()
  (save-excursion
    (let ((levelpar 0)			; Level of "()"
          (levelcurl 0)                 ; Level of "{}"
          (levelsqr 0)                  ; Level of "[]"
          (backpoint (max (- (point) gimml-paren-lookback) (point-min))))
      (catch 'loop
	(while (and (/= levelpar 1) (/= levelsqr 1) (/= levelcurl 1))
	  (if (re-search-backward "[][{}()]" backpoint t)
	      (if (not (gimml-inside-comment-or-string-p))
		  (cond
		   ((looking-at "(") (setq levelpar (1+ levelpar)))
		   ((looking-at ")") (setq levelpar (1- levelpar)))
		   ((looking-at "\\[") (setq levelsqr (1+ levelsqr)))
		   ((looking-at "\\]") (setq levelsqr (1- levelsqr)))
		   ((looking-at "{") (setq levelcurl (1+ levelcurl)))
		   ((looking-at "}") (setq levelcurl (1- levelcurl)))))
	    (throw 'loop 0)))		; Exit with value 0
	(save-excursion
	  (forward-char 1)
	  (if (looking-at gimml-indent-starters-reg)
	      (gimml-indentation-at-starter gimml-indent-level)
	    (current-column)))))))

;; This is too slow
;;
;; (defun gimml-inside-comment-or-string-p ()
;;  (let ((state (parse-partial-sexp (point-min) (point))))
;;    (or (nth 4 state) (nth 3 state))))

(defun gimml-inside-comment-or-string-p ()
  (let ((start (point)))
    (if (save-excursion
	  (condition-case ()
	      (progn
		(search-backward "(*")
		(search-forward "*)")
		(forward-char -1)	; A "*)" is not inside the comment
		(> (point) start))
	    (error nil)))
	t
      (let ((numb 0))
	(save-excursion
	  (save-restriction
	    (narrow-to-region (progn (beginning-of-line) (point)) start)
	    (condition-case ()
		(while t
		  (search-forward "\"")
		  (setq numb (1+ numb)))
	      (error (if (and (not (zerop numb))
			      (not (zerop (% numb 2))))
			 t nil)))))))))

(defun gimml-skip-block ()
  (let ((case-fold-search nil))
    (gimml-backward-sexp)
    (if (looking-at "end\\b")
	(progn
	  (goto-char (gimml-find-match-backward "end" 0))
	  (gimml-skip-chars-backward "\n\t "))
      ;; Here we will need to skip backward past if-then-else
      ;; and case-of expression. Please - tell me how !!
      )))

(defun gimml-find-match-backward (unquoted-this this-index &optional start)
  (save-excursion
    (let ((case-fold-search nil)
	  (levels (make-vector (length gimml-openings) 0))
	  (goal-levels (make-vector (length gimml-openings) 0))
	  (pattern (concat (gimml-make-pattern gimml-openings) "\\|"
			   (gimml-make-pattern gimml-closings))))
      (aset levels this-index 1)
      (if start (goto-char start))
      (while (not (equal levels goal-levels))
	(if (gimml-re-search-backward pattern)
	    (let ((i (gimml-indexed-looking-at gimml-openings)))
	      (if i
		  (aset levels i (1- (aref levels i)))
		(let ((i (gimml-indexed-looking-at gimml-closings)))
		  (aset levels i (1+ (aref levels i))))))
	  ;; The right match couldn't be found
	  (error (concat "Unbalanced: " unquoted-this))))
      (point))))

(defun gimml-make-pattern (l)
  (if (zerop (length l))
      ""
    (concat (aref l 0)
	    (gimml-make-pattern-1 l 1 (length l)))))

(defun gimml-make-pattern-1 (l i n)
  (if (< i n)
      (concat "\\|" (aref l i)
	      (gimml-make-pattern-1 l (1+ i) n))
    ""))

(defun gimml-indexed-looking-at (l)
  (gimml-indexed-looking-at-1 l 0 (length l)))

(defun gimml-indexed-looking-at-1 (l i n)
  (if (< i n)
      (if (looking-at (aref l i))
	  i
	(gimml-indexed-looking-at-1 l (1+ i) n))
    nil))

(defun gimml-find-match-indent (unquoted-this this-index indented &optional check)
  (save-excursion
    (goto-char (gimml-find-match-backward unquoted-this this-index))
    (if (and check (not (looking-at check)))
	(error "Unbalanced: %s" unquoted-this))
    (if (or gimml-type-of-indent indented)
	(current-column)
      (if (progn
	    (back-to-indentation)
	    (looking-at "|\\($\\|[^[]\\)"))
	  (- (current-indentation) gimml-pipe-indent)
	(current-indentation)))))

(defun gimml-find-matching-starter (regexp)
  (let ((case-fold-search nil)
	(new-regexp (concat regexp "\\|" (gimml-make-pattern gimml-closings))))
    (gimml-re-search-backward-1 new-regexp)))

(defun gimml-re-search-backward-1 (regexp)
  (let ((cont t) (found t))
    (while cont
      (if (gimml-re-search-backward regexp)
	  (let ((i (gimml-indexed-looking-at gimml-closings)))
	    (if i
		(goto-char (gimml-find-match-backward (aref gimml-closing-quotes i)
						   i))
	      (setq cont nil)))
	(setq cont nil found nil)))
    found))

(defun gimml-point-inside-let-etc ()
  (let ((case-fold-search nil) (last nil) (loop t) (found t) (start (point)))
    (save-excursion
      (while loop
	(condition-case ()
	    (progn
	      (re-search-forward "\\bend\\b")
	      (while (gimml-inside-comment-or-string-p)
		(re-search-forward "\\bend\\b"))
	      (forward-char -3)
	      (setq last (gimml-find-match-backward "end" 0 last))
	      (if (< last start)
		  (setq loop nil)
		(forward-char 3)))
	  (error (progn (setq found nil) (setq loop nil)))))
      (if found
	  last
	0))))

(defun gimml-re-search-backward (regexpr)
  (let ((case-fold-search nil) (found t))
    (if (re-search-backward regexpr nil t)
	(progn
	  (condition-case ()
	      (while (gimml-inside-comment-or-string-p)
		(re-search-backward regexpr))
	    (error (setq found nil)))
	  found)
      nil)))

(defun gimml-up-list ()
  (save-excursion
    (condition-case ()
	(progn
	  (up-list 1)
	  (point))
      (error 0))))

(defun gimml-backward-sexp ()
  (condition-case ()
      (progn
	(let ((start (point)))
	  (backward-sexp 1)
	  (while (and (/= start (point)) (looking-at "(\\*"))
	    (setq start (point))
	    (backward-sexp 1))))
    (error (forward-char -1) t)))

(defun gimml-forward-sexp ()
  (condition-case ()
      (progn
	(let ((start (point)))
	  (forward-sexp 1)
	  (while (and (>= (point) (- start 2))
		      (save-excursion
			(backward-char 2)
			(looking-at "\\*)")))
	    (setq start (point))
	    (forward-sexp 1))))
    (error t)))

(defun gimml-comment-indent ()
  (if (looking-at "^(\\*")		; Existing comment at beginning
      0					; of line stays there.
    (save-excursion
      (skip-chars-backward " \t")
      (max (1+ (current-column))	; Else indent at comment column
	   comment-column))))		; except leave at least one space.

(defun gimml-skip-chars-forward (chars)
  (skip-chars-forward chars)
  (if (looking-at "(\\*")
      (progn (forward-sexp 1)
	     (gimml-skip-chars-forward chars))))

(defun gimml-skip-chars-backward (chars)
  (skip-chars-backward chars)
  (if (save-excursion
	(backward-char 2)
	(looking-at "\\*)"))
      (progn (backward-sexp 1)
	     (gimml-skip-chars-backward chars))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; INFERIOR SHELL
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar gimml-shell-working-dir ""
  "The working directory of gimml-shell")

(defvar gimml-process-name nil
  "The name of the Gimml process")

(defvar gimml-shell-map nil
  "The mode map for gimml-shell.")

(defun gimml-startup-ready ()
  (save-excursion
    (goto-char (point-min))
    (re-search-forward gimml-shell-prompt-pattern nil t)))


(require 'shell)

(if (fboundp 'make-comint)
    ()
  (defalias 'make-comint (symbol-function 'make-shell)))

(if (fboundp 'copy-keymap)
    ()
  (defalias 'copy-keymap (symbol-function 'copy-sequence)))

(if gimml-emacs19-p
    (defun gimml-cd-command (dir)
      (format "cd %S" (expand-file-name dir)))
  (defun gimml-cd-command (dir)
    (concat "cd " (prin1-to-string (expand-file-name dir)))))

(defun gimml-shell (&optional dir)
  "Inferior shell invoking Gimml.
Like the shell mode with the additional command:

\\[gimml-run-on-file]\t Runs gimml on the file.
\\{gimml-shell-map}
Variables controlling the mode:

gimml-prog-name (default \\[gimml-prog-name\\])
    The string used to invoke the gimml program.

gimml-prog-name-ask-p (default \\[gimml-prog-name-ask-p\\])
    If t, you will be asked which program to run when the inferior
    shell starts up.

gimml-use-right-delim (default \\[gimml-use-right-delim\\])
gimml-use-left-delim  (default \\[gimml-use-left-delim\\])
    The left and right delimiter used by your version of gimml, for
    \"use file-name\".

gimml-shell-prompt-pattern (default \\[gimml-shell-prompt-pattern\\])
    The prompt pattern.

gimml-tmp-template (default \\[gimml-tmp-template\\])
    Template for the temporary files, created when a region is send
    to the inferior process running gimml (by gimml-simulate-send-region).

`cd' commands given to the Gimml process are watched by Emacs to
keep this buffer's default directory the same as the shell's default directory.
While directory tracking is enabled, the shell's working directory is displayed
by \\[list-buffers] or \\[mouse-buffer-menu] in the `File' field.
\\[gimml-dirs] queries the shell and resyncs Emacs' idea of what the current 
    directory stack is.
\\[gimml-dirtrack-toggle] turns directory tracking on and off.

Runs gimml-shell-hook if not nil."
  (interactive "DGimml in directory: ")
  (if (and gimml-process-name (eq (process-status gimml-process-name) 'run))
      (if (or (null dir) (string= dir default-directory))
	  ()
	(gimml-eval (gimml-cd-command dir)))
    (message "Starting Gimml...")	; start up a new process
    (save-excursion			; Process is not running
      (and gimml-prog-name-ask-p
	   (setq gimml-prog-name
		 (read-file-name
		  (concat "Gimml (default " gimml-prog-name "): ")
		  (file-name-directory (buffer-file-name))
		  gimml-prog-name)))
      (setq gimml-shell-working-dir default-directory)
      (if gimml-process-name
	  ()
	(setq gimml-process-name
	      (concat
	       ""
	       (substring gimml-prog-name
			  (string-match "[^/]*$" gimml-prog-name)
			  (string-match "$" gimml-prog-name)))))
      (set-buffer
       (apply (function make-comint)
	      gimml-process-name
	      (if (= (string-to-char gimml-prog-name) ?~)
		  (expand-file-name gimml-prog-name)
		gimml-prog-name)
	      nil
	      "-pwd-prompt" "|%s|%s"
	      (if dir
		  (list "-init" (gimml-cd-command dir))
		nil)))
      (erase-buffer)		; Erase the buffer if a previous
      (if gimml-shell-map		; process died in there
	  ()
	(setq gimml-shell-map (copy-keymap shell-mode-map))
	(define-key gimml-shell-map "\C-c\C-f" 'gimml-run-on-file)
	(define-key gimml-shell-map "\C-c\C-e" 'gimml-eval)
	(define-key gimml-shell-map "\C-c\C-r" 'gimml-send-region)
	(define-key gimml-shell-map "\C-c\C-p" 'gimml-prev-error)
	(define-key gimml-shell-map "\C-c\C-n" 'gimml-next-error)
	(define-key gimml-shell-map "\C-c\C-m" 'gimml-region)
	(define-key gimml-shell-map "\e|" 'gimml-electric-pipe)
	(define-key gimml-shell-map "\e\t" 'gimml-back-to-outer-indent)
	(define-key gimml-shell-map "\C-j" 'reindent-then-newline-and-indent)
	(define-key gimml-shell-map "\177" 'backward-delete-char-untabify)
	(define-key gimml-shell-map "\;" 'gimml-electric-semi)
	(define-key gimml-shell-map "\t" 'gimml-indent-line)
	(gimml-create-mode-menu gimml-shell-map t))
      (use-local-map gimml-shell-map)
      (make-local-variable 'shell-prompt-pattern)
      (setq shell-prompt-pattern gimml-shell-prompt-pattern)
      (make-local-variable 'comint-prompt-regexp)
      (setq comint-prompt-regexp gimml-shell-prompt-pattern)
      (make-local-variable 'list-buffers-directory)
      (setq list-buffers-directory (expand-file-name default-directory))
      (setq major-mode 'gimml-shell)
      (setq mode-name "Gimml-shell")
      (set-process-filter
       (get-process gimml-process-name) 'gimml-startup-filter)
      (run-hooks 'gimml-shell-hook))
    ))

(defun gimml-spy-cd (str)
  (if (string-match "|\\([^|]*\\)|" str)
      (let* ((start (match-beginning 0))
	     (end (match-end 0))
	     (wd (substring str (match-beginning 1) (match-end 1)))
	     (newstr (concat (substring str 0 start)
			     (substring str end))))
	(if (string= gimml-shell-working-dir wd)
	    ()
	  (setq gimml-shell-working-dir wd)
	  (save-excursion
	    (set-buffer (process-buffer (get-process gimml-process-name)))
	    (cd wd)
	    (setq list-buffers-directory (file-name-as-directory
					  (expand-file-name wd)))
	    (message "Gimml directory: %s" wd))
	  )
	newstr)
    str))

(defun gimml-startup-filter (proc str)
;  (save-excursion
;    (set-buffer "*scratch*")
;    (insert "``" str "''"))
  (let ((cur (selected-window))
	(pop-up-windows t)
	(process (concat "*" gimml-process-name "*")))
    (pop-to-buffer process)
    (goto-char (point-max))
    (insert (gimml-spy-cd str))
    (if (gimml-startup-ready)
	(progn
	  (message "Starting Gimml... done.")
	  (set-process-filter (get-process gimml-process-name)
			      'gimml-process-filter))
      )
    (set-marker (process-mark proc) (point-max))
    (select-window cur)))

(defun gimml-process-filter (proc str)
;  (save-excursion
;    (set-buffer "*scratch*")
;    (insert "``" str "''"))
  (let ((cur (selected-window))
	(pop-up-windows t)
	(process (concat "*" gimml-process-name "*")))
    (pop-to-buffer process)
    (goto-char (point-max))
    (insert (gimml-spy-cd str))
    (set-marker (process-mark proc) (point-max))
    (select-window cur)))

(defun gimml-pop-to-shell ()
  "Pop to the buffer running gimml."
  (interactive)
  (gimml-shell)
  (pop-to-buffer (concat "*" gimml-process-name "*")))

(defun gimml-run-on-file (file)
  "Send a use FILE to the inferior shell running gimml."
  (interactive "FUse file: ")
  (gimml-shell)
  (setq file (expand-file-name file))
  (if gimml-strip-path
      (if (string= (substring file 0 (string-match "[^/]*$" file))
		   gimml-shell-working-dir)
	  (setq file (substring file (string-match "[^/]*$" file)
				(string-match "$" file)))))
  (save-some-buffers)
  ;(gimml-skip-errors)
  (send-string gimml-process-name
	       (concat "use " gimml-use-left-delim file
		       gimml-use-right-delim ";\n")))

(defun gimml-save-buffer-use-file ()
  "Save the buffer, and send a `use file' to the inferior shell
running gimml."
  (interactive)
  (let (file)
    (if (setq file (buffer-file-name))	; Is the buffer associated
	(progn				; with file ?
	  (save-buffer)
	  (gimml-shell)
	  ;(gimml-skip-errors)
	  (if gimml-strip-path
	      (if (string= (substring file 0 (string-match "[^/]*$" file))
			   gimml-shell-working-dir)
		  (setq file (substring file (string-match "[^/]*$" file)
					(string-match "$" file)))))
	  (message (concat "use " gimml-use-left-delim file gimml-use-right-delim))
	  (send-string gimml-process-name
		       (concat "use " gimml-use-left-delim
			       file
			       gimml-use-right-delim ";\n")))
      (error "Buffer not associated with file."))))

(defvar gimml-tmp-files-list nil
  "List of all temporary files created by gimml-simulate-send-region.
Each element in the list is a list with the format:
\n   (tmp-filename file-name start-line)")

(defvar gimml-simulate-send-region-called-p nil
  "Has gimml-simulate-send-region been called previously.")

(defvar gimml-old-kill-emacs-hook nil
  "Old value of kill-emacs-hook")

(defvar gimml-tmp-bug "")

(defun gimml-simulate-send-region (point1 point2)
  "Simulate send region. As send-region only can handle what ever the
system sets as the default, we have to make a temporary file.
Updates the list of temporary files (gimml-tmp-files-list)."
  (let ((file (expand-file-name
	       (make-temp-name (concat gimml-tmp-template gimml-tmp-bug)))))
    ;; Remove temporary files when we leave emacs
    (if (not gimml-simulate-send-region-called-p)
	(progn
	  (setq gimml-old-kill-emacs-hook kill-emacs-hook)
	  (setq kill-emacs-hook 'gimml-remove-tmp-files)
	  (setq gimml-simulate-send-region-called-p t)))
    ;; As make-temp-name can only make 26 unique file names with the
    ;; same template (bug in Un*x function mktemp), we add a new
    ;; letter to gimml-tmp-template.
    (if (zerop (% (1+ (length gimml-tmp-files-list)) 25))
	(setq gimml-tmp-bug (concat gimml-tmp-bug "A")))
    (save-excursion
      (goto-char point1)
      (setq gimml-tmp-files-list
	    (cons (list file
			(buffer-file-name)
			(save-excursion	; Calculate line no.
			  (beginning-of-line)
			  (1+ (count-lines 1 (point)))))
		  gimml-tmp-files-list)))
    (write-region point1 point2 file nil 'dummy)
    (gimml-shell)
    (message "Using temporary file: %s" file)
    (send-string
     gimml-process-name
     ;; string to send: use file;
     (concat "use " gimml-use-left-delim file gimml-use-right-delim ";\n"))))

(defun gimml-remove-tmp-files ()
  "Remove the temporary files, created by gimml-simulate-send-region, if
they still exist. Only files recorded in gimml-tmp-files-list are removed.
This function is put into kill-emacs-hook if any temporary files are
used."
  (message "Removing temporary files created by gimml-mode...")
  (while gimml-tmp-files-list
    (condition-case ()
 	(delete-file (car (car gimml-tmp-files-list)))
      (error ()))
    (setq gimml-tmp-files-list (cdr gimml-tmp-files-list)))
  (message "Removing temporary files created by gimml-mode... done.")
  (run-hooks 'gimml-old-kill-emacs-hook))

(defun gimml-send-region ()
  "Send region to inferior shell running gimml."
  (interactive)
  (gimml-shell)
  ;(gimml-skip-errors)
  (let (start end)
    (save-excursion
      (setq end (point))
      (exchange-point-and-mark)
      (setq start (point)))
    (gimml-simulate-send-region start end)))

(defun gimml-eval (cmd)
  (interactive "sGimml: ")
  (gimml-shell)
  ;(gimml-skip-errors)
  (let* ((s (concat cmd ";\n"))
	 (proc (get-process gimml-process-name))
	 (buf (process-buffer proc)))
    (save-excursion
      (set-buffer buf)
      (goto-char (point-max))
      (insert s))
    (send-string gimml-process-name s)))
  
(defun gimml-send-function ()
  "Does *not* send the function, but the paragraph, to inferior shell
running gimml (sorry)."
  (interactive)
  (gimml-shell)
  ;(gimml-skip-errors)
  (let (start end)
    (save-excursion
      (condition-case ()
	  (progn
	    (backward-paragraph)
	    (setq start (point)))
	(error (setq start (point-min))))
      (condition-case ()
	  (progn
	    (forward-paragraph)
	    (setq end (point)))
	(error (setq end (point-max)))))
    (gimml-simulate-send-region start end)))

(defun gimml-send-buffer ()
  "Send the buffer, to inferior shell running gimml."
  (interactive)
  (gimml-shell)
  ;(gimml-skip-errors)
  (gimml-simulate-send-region (point-min) (point-max)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; END OF GimML-MODE
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
