/* gen - actual generation (writing) of glex scanners */

/*-
 * Copyright (C) 2021 Jean Goubault-Larrecq
 * and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
 * Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.
 * This code is derived from flex version 2.5.2.
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Vern Paxson.
 * 
 * The United States Government has rights in this work pursuant
 * to contract no. DE-AC03-76SF00098 between the United States
 * Department of Energy and the University of California.
 *
 * Jean Goubault-Larrecq, as the author of the modifications to
 * flex that eventually produced glex, also has rights in this
 * work, as well as Université Paris-Saclay, CNRS, ENS Paris-Saclay,
 * Laboratoire Méthodes Formelles.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "glexdef.h"


/* declare functions that have forward references */

void genecs PROTO((void));
void indent_put2s PROTO((char [], char []));
void indent_puts PROTO((char []));


static int indent_level = 0; /* each level is 8 spaces */

#define indent_up() (++indent_level)
#define indent_down() (--indent_level)
#define set_indent(indent_val) indent_level = indent_val

/* Almost everything is done in terms of arrays starting at 1, so provide
 * a null entry for the zero element of all C arrays.  (The exception
 * to this is that the fast table representation generally uses the
 * 0 elements of its arrays, too.)
 */

static char ML_int_decl[] = "val %s = (* %d entries *) \n    [   0,\n";

/* Indent to the current level. */

void do_indent()
	{
	register int i = indent_level * 8;

	while ( i >= 8 )
		{
		outc( '\t' );
		i -= 8;
		}

	while ( i > 0 )
		{
		outc( ' ' );
		--i;
		}
	}


/* Generate the code to keep backing-up information. */

/* Generate equivalence-class tables. */

void genecs()
{
  register int i, j;
  int numrows;

  out_str_dec( ML_int_decl, "yy_ec", csize );

  for ( i = 1; i < csize; ++i )
    {
      if ( caseins && (i >= 'A') && (i <= 'Z') )
	ecgroup[i] = ecgroup[clower( i )];

      ecgroup[i] = ABS( ecgroup[i] );
      mkdata( ecgroup[i] );
    }

  dataend();

  if ( trace )
    {
      fputs( _( "\n\nEquivalence Classes:\n\n" ), stderr );

      numrows = csize / 8;

      for ( j = 0; j < numrows; ++j )
	{
	  for ( i = j; i < csize; i = i + numrows )
	    {
	      fprintf( stderr, "%4s = %-2d",
		       readable_form( i ), ecgroup[i] );

	      putc( ' ', stderr );
	    }

	  putc( '\n', stderr );
	}
    }
}

/* gentabs - generate data statements for the transition tables */

void gentabs()
{
  int i, j, k, *accset, nacc, *acc_array, total_states;
  int end_of_buffer_action = num_rules + 1;

  acc_array = allocate_integer_array( current_max_dfas );
  nummt = 0;

  {
    dfaacc[end_of_buffer_state].dfaacc_state = end_of_buffer_action;
    for ( i = 1; i <= lastdfa; ++i )
      acc_array[i] = dfaacc[i].dfaacc_state;
    /* add accepting number for jam state */
    acc_array[i] = 0;
  }

  /* Spit out "yy_accept" array.  It'll be actual
   * accepting numbers.  We just dump the numbers.
   */
  /* "lastdfa + 2" is the size of "yy_accept"; includes room for C arrays
   * beginning at 0 and for "jam" state.
   */
  k = lastdfa + 2;
  out_str_dec( ML_int_decl, "yy_accept", k );
  for ( i = 1; i <= lastdfa; ++i )
    {
      mkdata( acc_array[i] );
      if ( trace && acc_array[i] )
	fprintf( stderr, _( "state # %d accepts: [%d]\n" ),
		 i, acc_array[i] );
    }
  /* Add entry for "jam" state. */
  mkdata( acc_array[i] );
  dataend();

  genecs();

  {
    /* Write out meta-equivalence classes (used to index
     * templates with).
     */
    if ( trace )
      fputs( _( "\n\nMeta-Equivalence Classes:\n" ),
	     stderr );
    out_str_dec( ML_int_decl, "yy_meta", numecs + 1 );
    for ( i = 1; i <= numecs; ++i )
      {
	if ( trace )
	  fprintf( stderr, "%d = %d\n",
		   i, ABS( tecbck[i] ) );
	mkdata( ABS( tecbck[i] ) );
      }
    dataend();
  }

  total_states = lastdfa + numtemps;
  if (tblend >= MAX_SHORT)
    lerrif ( _("yy_base table too large: %d elements, should be at most 32767 elements\n"),
	     tblend);
  out_str_dec( ML_int_decl, "yy_base", total_states + 1 );
  for ( i = 1; i <= lastdfa; ++i )
    {
      register int d = def[i];

      if ( base[i] == JAMSTATE )
	base[i] = jambase;
      if ( d == JAMSTATE )
	def[i] = jamstate;
      else if ( d < 0 )
	{
	  /* Template reference. */
	  ++tmpuses;
	  def[i] = lastdfa - d + 1;
	}
      mkdata( base[i] );
    }
  /* Generate jam state's base index. */
  mkdata( base[i] );
  for ( ++i /* skip jam state */; i <= total_states; ++i )
    {
      mkdata( base[i] );
      def[i] = jamstate;
    }
  dataend();

  if (total_states >= MAX_SHORT)
    lerrif ( _("yy_def table too large: %d elements, should be at most 32767 elements\n"),
	     total_states);
  out_str_dec( ML_int_decl,
	       "yy_def", total_states + 1 );
  for ( i = 1; i <= total_states; ++i )
    mkdata( def[i] );
  dataend();

  if (total_states >= MAX_SHORT)
    lerrif ( _("yy_nxt table too large: %d elements, should be at most 32767 elements\n"),
	     total_states);
  out_str_dec( ML_int_decl,
	       "yy_nxt", tblend + 1 );
  for ( i = 1; i <= tblend; ++i )
    {
      /* Note, the order of the following test is important.
       * If chk[i] is 0, then nxt[i] is undefined.
       */
      if ( chk[i] == 0 || nxt[i] == 0 )
	nxt[i] = jamstate;	/* new state is the JAM state */
      mkdata( nxt[i] );
    }
  dataend();

  if (total_states >= MAX_SHORT)
    lerrif ( _("yy_chk table too large: %d elements, should be at most 32767 elements\n"),
	     total_states);
  out_str_dec( ML_int_decl,
	       "yy_chk", tblend + 1 );
  for ( i = 1; i <= tblend; ++i )
    {
      if ( chk[i] == 0 )
	++nummt;
      mkdata( chk[i] );
    }
  dataend();
}


/* Write out a formatted string (with a secondary string argument) at the
 * current indentation level, adding a final newline.
 */

void indent_put2s( fmt, arg )
char fmt[], arg[];
	{
	do_indent();
	out_str( fmt, arg );
	outn( "" );
	}


/* Write out a string at the current indentation level, adding a final
 * newline.
 */

void indent_puts( str )
char str[];
	{
	do_indent();
	outn( str );
	}


void out_with_special_chars( str )
     const char str[];
{
  register const char *s;
  register char c;

  for (s = str; (c = *s++)!=0; )
    {
      switch (c)
	{
	case '\001':
	  {
	    int n = 0;

	    while (c = *s++, c>='0' && c<='9')
	      n = 10*n+(c-'0');
	    fprintf (stdout, "%d", n+num_rules+1);
	  }
	  /*FALLTHROUGH*/
	default:
	  putc (c, stdout);
	  break;
	}
    }
  out_line_count( str );
}

/* make_tables - generate transition tables and finishes generating output file
 */

void make_tables()
{
  register int i;

  set_indent( 0 );

  out( &action_array[defs1_offset] );
  line_directive_out( stdout, 0 );

  outn ("local");
  set_indent (2);
  do_indent ();
  outn ("val return = SOME;");
  do_indent ();
  outn ("val continue = NONE;");
  if (provide_yyterminate)
    {
      do_indent ();
      outn ("val yyterminate = (fn _ => SOME 0 (* end of file reached *));");
    }
  do_indent ();
  outn ("infix 0 glex_then;");
  do_indent ();
  outn ("fun f glex_then g = fn yyd => (f yyd; g yyd);");
  /*outn ("fun glex_set (a, i, v) = (a.(i) .:= v; v);");*/
  for (i=1; i<=lastsc; i++)
    {
      do_indent ();
      out_str_dec ("val %s = %d;\n", scname[i], i-1);
    }
  outn ("");
  do_indent ();
  out_dec ("val yy_end_of_buffer=%d;\n", num_rules+1);
  do_indent ();
  outn ("val action_0 = fn _ => continue;");
  /* "array (yy_end_of_buffer+%d, fn _ => continue);\n", lastsc+1);*/
  gentabs();
  if ( nultrans )
    {
      out_str_dec( ML_int_decl, "yy_NUL_trans", lastdfa + 1 );
      for ( i = 1; i <= lastdfa; ++i )
	{
	  mkdata( nultrans[i] );
	}
      dataend();
    }
  else out_str ("val %s = nil;\n", "yy_NUL_trans");

  /* Copy prolog to output file. */
  out( &action_array[prolog_offset] );
  line_directive_out( stdout, 0 );
  set_indent( 2 );
  out_with_special_chars ( &action_array[action_offset] );
  line_directive_out( stdout, 0 );
  out ("val action = [ action_0");
  for (i=1; i<=num_rules+1 /*yy_end_of_buffer*/; i++)
    out_dec (",\n               action_%d", action_get_bit (2*i)?i:0);
  for (i=1; i<=lastsc; i++)
    if (action_get_bit (2*i+1))
      out_dec (",\n               action_%d", num_rules+1+i);
    else out (",\n               yyterminate");
  outn ("]");
  out ("val yytables =");

#ifdef OBSOLETE
  /* generate cases for any missing EOF rules */
  for ( i = 1; i <= lastsc; ++i )
    if ( ! sceof[i] )
      {
	char s[MAXLINE];

	do_indent();
	out_str ("action.(yy_end_of_buffer+%s+1) .:= (fn _ => continue);\n",
		 /* used to call some yyterminate() function */
		 scname[i] );
      }
#endif
  printf ("\n  glex_tables (action, %d, yy_end_of_buffer, yy_accept, yy_ec, yy_meta, yy_base, yy_def, yy_nxt, yy_chk, yy_NUL_trans, %d, %d, %d)\nin\n  val %slex = glex yytables\nend;\n",
	  num_rules, jambase, jamstate, NUL_ec, prefix);

  /* Copy remainder of input to output. */
  line_directive_out( stdout, 1 );

  if ( sectnum == 3 )
    (void) glexscan(); /* copy remainder of input to output */
}
