/* glex - tool to generate reasonably fast lexical analyzers for GimML */

/*-
 * Copyright (C) 2021 Jean Goubault-Larrecq
 * and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
 * Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.
 * This code is derived from flex version 2.5.2.
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Vern Paxson.
 * 
 * The United States Government has rights in this work pursuant
 * to contract no. DE-AC03-76SF00098 between the United States
 * Department of Energy and the University of California.
 *
 * Jean Goubault-Larrecq, as the author of the modifications to
 * flex that eventually produced glex, also has rights in this
 * work, as well as Université Paris-Saclay, CNRS, ENS Paris-Saclay,
 * Laboratoire Méthodes Formelles.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * that: (1) source distributions retain this entire copyright notice and
 * comment, and (2) distributions including binaries display the following
 * acknowledgement:  ``This product includes software developed by the
 * University of California, Berkeley and its contributors'' in the
 * documentation or other materials provided with the distribution and in
 * all advertising materials mentioning features or use of this software.
 * Neither the name of the University nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1990 The Regents of the University of California.\n     Copyright (C) 2021 Jean Goubault-Larrecq\n     and Université Paris-Saclay, CNRS, ENS Paris-Saclay,\n     Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.\n     All rights reserved.\n";
#endif /* not lint */

#include <stdlib.h>

#include "glexdef.h"
#include "version.h"

static char glex_version[] = GLEX_VERSION;


/* declare functions that have forward references */

void glexinit PROTO((int, char**));
void readin PROTO((void));
void set_up_initial_allocations PROTO((void));

#ifdef NEED_ARGV_FIXUP
extern void argv_fixup PROTO((int *, char ***));
#endif


/* these globals are all defined and commented in glexdef.h */
int printstats, syntaxerror, eofseen, trace, nowarn, spprdflt;
int caseins;
int gen_line_dirs;
int csize;
int continued_action, in_rule;
int datapos, dataline, linenum, out_linenum;
int skel_ind = 0;
char *action_array;
int action_size, defs1_offset, prolog_offset, action_offset, action_index;
char *infilename = NULL, *outfilename = NULL;
char *prefix = "yy";
int did_outfilename;
int provide_yyterminate;
int onestate[ONE_STACK_SIZE], onesym[ONE_STACK_SIZE];
int onenext[ONE_STACK_SIZE], onedef[ONE_STACK_SIZE], onesp;
int current_mns, current_max_rules;
int num_rules, num_eof_rules, default_rule, lastnfa;
int *firstst, *lastst, *finalst, *transchar, *trans1, *trans2;
int *accptnum, *assoc_rule, *state_type;
int *rule_linenum, *rule_useful;
int current_state_type;
int numtemps, numprots, protprev[MSP], protnext[MSP], prottbl[MSP];
int protcomst[MSP], firstprot, lastprot, protsave[PROT_SAVE_SIZE];
int numecs, nextecm[CSIZE + 1], ecgroup[CSIZE + 1], nummecs, tecfwd[CSIZE + 1];
int tecbck[CSIZE + 1];
int lastsc, *scset, *scbol, *scxclu, *sceof;
int current_max_scs;
char **scname;
int current_max_dfa_size, current_max_xpairs;
int current_max_template_xpairs, current_max_dfas;
int lastdfa, *nxt, *chk, *tnxt;
int *base, *def, *nultrans, NUL_ec, tblend, firstfree, **dss, *dfasiz;
union dfaacc_union *dfaacc;
int *accsiz, *dhash, numas;
int numsnpairs, jambase, jamstate;
int lastccl, *cclmap, *ccllen, *cclng, cclreuse;
int current_maxccls, current_max_ccl_tbl_size;
Char *ccltbl;
char nmstr[MAXLINE];
int sectnum, nummt, hshcol, dfaeql, numeps, eps2, num_reallocs;
int tmpuses, totnst, peakpairs, numuniq, numdup, hshsave;
int end_of_buffer_state;
char **input_files;
int num_input_files;

/* Make sure program_name is initialized so we don't crash if writing
 * out an error message before getting the program name from argv[0].
 */
char *program_name = "glex";

static char *outfile_template = "lex%s.%s";

#ifdef MS_DOS
extern unsigned _stklen = 16384;
#endif

static char outfile_path[MAXLINE];
static int outfile_created = 0;


int main( argc, argv )
int argc;
char **argv;
	{
	int i;

#ifdef THINK_C
	argc = ccommand( &argv );
#endif
#ifdef NEED_ARGV_FIXUP
	argv_fixup( &argc, &argv );
#endif

	glexinit( argc, argv );

	readin();

	ntod();

	for ( i = 1; i <= num_rules; ++i )
		if ( ! rule_useful[i] && i != default_rule )
			line_warning( _( "rule cannot be matched" ),
					rule_linenum[i] );

	if ( spprdflt && rule_useful[default_rule] )
		line_warning(
			_( "-s option given but default rule can be matched" ),
			rule_linenum[default_rule] );

	/* Generate the C state transition tables from the DFA. */
	make_tables();

	/* Note, glexend does not return.  It exits with its argument
	 * as status.
	 */
	glexend( 0 );

	return 0;	/* keep compilers/lint happy */
	}


/* check_options - check user-specified options */

void check_options()
	{
	int i;

	{ /* Set up doubly-linked equivalence classes. */
	  /* We loop all the way up to csize, since ecgroup[csize] is
	   * the position used for NUL characters.
	   */
	  ecgroup[1] = NIL;
	  for ( i = 2; i <= csize; ++i )
	    {
	      ecgroup[i] = i - 1;
	      nextecm[i - 1] = i;
	    }
	  nextecm[csize] = NIL;
	}

	{
	  FILE *prev_stdout;

	  if ( ! did_outfilename )
	    {
	      sprintf( outfile_path, outfile_template, "yy", "ml" );
	      outfilename = outfile_path;
	    }

	  prev_stdout = freopen( outfilename, "w", stdout );

	  if ( prev_stdout == NULL )
	    lerrsf( _( "could not create %s" ), outfilename );

	  outfile_created = 1;
	}

	if ( did_outfilename )
	  line_directive_out( stdout, 0 );

#ifdef OBSOLETE
	{
	  int len;
	  char dotmlfilename[MAXLINE];
	  FILE *f;

	  len = strlen (outfilename);
	  if (len>=3 && outfilename[len-3]=='.' && outfilename[len-2]=='m'
	      && outfilename[len-1]=='l')
	    len -= 3;
	  memcpy (prefix, outfilename, len);
	  prefix[len] = 0;
	  memcpy (dotmlfilename, outfilename, len);
	  strcpy (dotmlfilename+len, "_h.ml");
	  f = fopen (dotmlfilename, "w");
	  if (f==NULL)
	    lerrsf ( _("could not create %s" ), dotmlfilename);
	  fprintf (f, "\nopen \"%s_tab_h\";\n\
(* This file should define the type of tokens %s_token returned by %s_glex.\n\
   If you use hyacc, this will be defined for you as type %s_glex = int;\n\
 *)\n\
extern val %s_glex : glex_data -> %s_token;\n",
		   prefix, prefix, prefix, prefix, prefix, prefix);
	  fclose (f);

	  out ("(* A lexical scanner generated by glex *)\n\n");
	  out ("open \"");
	  out (dotmlfilename);
	  outn ("\";\n\n");
	}
#endif
	}


/* glexend - terminate glex
 *
 * note
 *    This routine does not return.
 */

void glexend( exit_status )
int exit_status;

	{
	int tblsiz;
	int unlink();

	if ( exit_status != 0 && outfile_created )
		{
		if ( ferror( stdout ) )
			lerrsf( _( "error writing output file %s" ),
				outfilename );

		else if ( fclose( stdout ) )
			lerrsf( _( "error closing output file %s" ),
				outfilename );

		else if ( unlink( outfilename ) )
			lerrsf( _( "error deleting output file %s" ),
				outfilename );
		}

	if ( printstats )
		{
		fprintf( stderr, _( "%s version %s usage statistics:\n" ),
			program_name, glex_version );

		fprintf( stderr, _( "  scanner options: -" ) );

		if ( caseins )
			putc( 'i', stderr );
		if ( spprdflt )
			putc( 's', stderr );
                if (!provide_yyterminate)
                        putc ('t', stderr);
		if ( printstats )
			putc( 'v', stderr );	/* always true! */
		if ( nowarn )
			putc( 'w', stderr );
		if ( gen_line_dirs )
			putc( 'L', stderr );
		if ( trace )
			putc( 'T', stderr );

		if ( did_outfilename )
			fprintf( stderr, " -o%s", outfilename );

		putc( '\n', stderr );

		fprintf( stderr, _( "  %d/%d NFA states\n" ),
			lastnfa, current_mns );
		fprintf( stderr, _( "  %d/%d DFA states (%d words)\n" ),
			lastdfa, current_max_dfas, totnst );
		fprintf( stderr, _( "  %d rules\n" ),
		num_rules + num_eof_rules - 1 /* - 1 for def. rule */ );

		fprintf( stderr, _( "  %d/%d start conditions\n" ), lastsc,
			current_max_scs );
		fprintf( stderr,
			_( "  %d epsilon states, %d double epsilon states\n" ),
			numeps, eps2 );

		if ( lastccl == 0 )
			fprintf( stderr, _( "  no character classes\n" ) );
		else
			fprintf( stderr,
_( "  %d/%d character classes needed %d/%d words of storage, %d reused\n" ),
				lastccl, current_maxccls,
				cclmap[lastccl] + ccllen[lastccl],
				current_max_ccl_tbl_size, cclreuse );

		fprintf( stderr, _( "  %d state/nextstate pairs created\n" ),
			numsnpairs );
		fprintf( stderr, _( "  %d/%d unique/duplicate transitions\n" ),
			numuniq, numdup );
		{
		  tblsiz = 2 * (lastdfa + numtemps) + 2 * tblend;
		  fprintf( stderr,
			   _( "  %d/%d base-def entries created\n" ),
			   lastdfa + numtemps, current_max_dfas );
		  fprintf( stderr,
			   _( "  %d/%d (peak %d) nxt-chk entries created\n" ),
			   tblend, current_max_xpairs, peakpairs );
		  fprintf( stderr,
			   _( "  %d/%d (peak %d) template nxt-chk entries created\n" ),
			   numtemps * nummecs,
			   current_max_template_xpairs,
			   numtemps * numecs );
		  fprintf( stderr, _( "  %d empty table entries\n" ),
			   nummt );
		  fprintf( stderr, _( "  %d protos created\n" ),
			   numprots );
		  fprintf( stderr,
			   _( "  %d templates created, %d uses\n" ),
			   numtemps, tmpuses );
		}

		{
		  tblsiz = tblsiz + csize;
		  fprintf( stderr,
			   _( "  %d/%d equivalence classes created\n" ),
			   numecs, csize );
		}
		{
		  tblsiz = tblsiz + numecs;
		  fprintf( stderr,
			   _( "  %d/%d meta-equivalence classes created\n" ),
			   nummecs, csize );
		}

		fprintf( stderr,
		_( "  %d (%d saved) hash collisions, %d DFAs equal\n" ),
			hshcol, hshsave, dfaeql );
		fprintf( stderr, _( "  %d sets of reallocations needed\n" ),
			num_reallocs );
		fprintf( stderr, _( "  %d total table entries needed\n" ),
			tblsiz );
		}

	exit( exit_status );
	}


/* glexinit - initialize glex */

void glexinit( argc, argv )
int argc;
char **argv;
	{
	int i;
	char *arg, *mktemp();

	printstats = syntaxerror = trace = spprdflt = caseins = false;
	nowarn = continued_action = false;
	in_rule = false;
	csize = CSIZE; /* always 8-bit scanner */
	gen_line_dirs = false;
	did_outfilename = 0;
	provide_yyterminate = true;


	/* Initialize dynamic array for holding the rule actions. */
	action_size = 2048;	/* default size of action array in bytes */
	action_array = allocate_character_array( action_size );
	defs1_offset = prolog_offset = action_offset = action_index = 0;
	action_array[0] = '\0';

	program_name = argv[0];

	/* read flags */
	for ( --argc, ++argv; argc ; --argc, ++argv )
		{
		arg = argv[0];

		if ( arg[0] != '-' || arg[1] == '\0' )
			break;

		if ( arg[1] == '-' )
			{ /* --option */
			if ( ! strcmp( arg, "--help" ) )
				arg = "-h";

			else if ( ! strcmp( arg, "--version" ) )
				arg = "-V";
			else if ( ! strcmp( arg, "--" ) )
				{ /* end of options */
				--argc;
				++argv;
				break;
				}
			}

		for ( i = 1; arg[i] != '\0'; ++i )
			switch ( arg[i] )
				{
				case '?':
				case 'h':
					usage();
					exit( 0 );

				case 'i':
					caseins = true;
					break;

				case 'L':
					gen_line_dirs = true;
					break;

				case 'o':
					if ( i != 1 )
						glexerror(
				_( "-o flag must be given separately" ) );

					outfilename = arg + i + 1;
					did_outfilename = 1;
					goto get_next_arg;

				case 'P':
				  if (i!=1)
				    glexerror (_( "-P flag must be given separately"));
                                  prefix = arg+i+1;
                                  goto get_next_arg;

				case 's':
					spprdflt = true;
					break;

                                case 't': provide_yyterminate = false; break;
 
				case 'T':
					trace = true;
					break;

				case 'v':
					printstats = true;
					break;

				case 'V':
					printf( _( "%s version %s\n" ),
						program_name, glex_version );
					exit( 0 );

				case 'w':
					nowarn = true;
					break;

				default:
					fprintf( stderr,
		_( "%s: unknown flag '%c'.  For usage, try\n\t%s --help\n" ),
						program_name, (int) arg[i],
						program_name );
					exit( 1 );
				}

		/* Used by -C, -S, -o, and -P flags in lieu of a "continue 2"
		 * control.
		 */
		get_next_arg: ;
		}

	num_input_files = argc;
	input_files = argv;
	set_input_file( num_input_files > 0 ? input_files[0] : NULL );

	lastccl = lastsc = lastdfa = lastnfa = 0;
	num_rules = num_eof_rules = default_rule = 0;
	numas = numsnpairs = tmpuses = 0;
	numecs = numeps = eps2 = num_reallocs = hshcol = dfaeql = totnst = 0;
	numuniq = numdup = hshsave = eofseen = datapos = dataline = 0;
	onesp = numprots = 0;

	out_linenum = linenum = sectnum = 1;
	firstprot = NIL;

	/* Used in mkprot() so that the first proto goes in slot 1
	 * of the proto queue.
	 */
	lastprot = 1;

	set_up_initial_allocations();
	}


/* readin - read in the rules section of the input file(s) */

void readin()
{
  line_directive_out( (FILE *) 0, 1 );
  if ( yyparse() )
    {
      pinpoint_message( _( "fatal parse error" ) );
      glexend( 1 );
    }

  if ( syntaxerror )
    glexend( 1 );

  numecs = cre8ecs( nextecm, ecgroup, csize );

  /* Now map the equivalence class for NUL to its expected place. */
  ecgroup[0] = ecgroup[csize];
  NUL_ec = ABS( ecgroup[0] );

  ccl2ecl();
}


/* set_up_initial_allocations - allocate memory for internal tables */

void set_up_initial_allocations()
	{
	current_mns = INITIAL_MNS;
	firstst = allocate_integer_array( current_mns );
	lastst = allocate_integer_array( current_mns );
	finalst = allocate_integer_array( current_mns );
	transchar = allocate_integer_array( current_mns );
	trans1 = allocate_integer_array( current_mns );
	trans2 = allocate_integer_array( current_mns );
	accptnum = allocate_integer_array( current_mns );
	assoc_rule = allocate_integer_array( current_mns );
	state_type = allocate_integer_array( current_mns );

	current_max_rules = INITIAL_MAX_RULES;
	rule_linenum = allocate_integer_array( current_max_rules );
	rule_useful = allocate_integer_array( current_max_rules );

	current_max_scs = INITIAL_MAX_SCS;
	scset = allocate_integer_array( current_max_scs );
	scbol = allocate_integer_array( current_max_scs );
	scxclu = allocate_integer_array( current_max_scs );
	sceof = allocate_integer_array( current_max_scs );
	scname = allocate_char_ptr_array( current_max_scs );

	current_maxccls = INITIAL_MAX_CCLS;
	cclmap = allocate_integer_array( current_maxccls );
	ccllen = allocate_integer_array( current_maxccls );
	cclng = allocate_integer_array( current_maxccls );

	current_max_ccl_tbl_size = INITIAL_MAX_CCL_TBL_SIZE;
	ccltbl = allocate_Character_array( current_max_ccl_tbl_size );

	current_max_dfa_size = INITIAL_MAX_DFA_SIZE;

	current_max_xpairs = INITIAL_MAX_XPAIRS;
	nxt = allocate_integer_array( current_max_xpairs );
	chk = allocate_integer_array( current_max_xpairs );

	current_max_template_xpairs = INITIAL_MAX_TEMPLATE_XPAIRS;
	tnxt = allocate_integer_array( current_max_template_xpairs );

	current_max_dfas = INITIAL_MAX_DFAS;
	base = allocate_integer_array( current_max_dfas );
	def = allocate_integer_array( current_max_dfas );
	dfasiz = allocate_integer_array( current_max_dfas );
	accsiz = allocate_integer_array( current_max_dfas );
	dhash = allocate_integer_array( current_max_dfas );
	dss = allocate_int_ptr_array( current_max_dfas );
	dfaacc = allocate_dfaacc_union( current_max_dfas );

	nultrans = (int *) 0;
	}


void usage()
	{
	FILE *f = stdout;

	fprintf( f,
_( "%s [-histvwLTV? -ooutput -Pprefix]\n" ),
		program_name );
	fprintf( f, _( "\t[--help --version] [file ...]\n" ) );

	fprintf( f, _( "\t-h  produce this help message\n" ) );
	fprintf( f, _( "\t-i  generate case-insensitive scanner\n" ) );
	fprintf( f,
		_( "\t-s  suppress default rule to print unmatched text\n" ) );
	fprintf (f,
		 _("\t-t  suppress default definition of yyterminate function\n"));

	if ( ! did_outfilename )
		{
		sprintf( outfile_path, outfile_template,
			"yy", "ml");
		outfilename = outfile_path;
		}

	fprintf( f,
		_( "\t-v  write summary of scanner statistics to f\n" ) );
	fprintf( f, _( "\t-w  do not generate warnings\n" ) );
	fprintf( f, _( "\t-L  add #line directives to scanner\n" ) );
	fprintf( f, _( "\t-T  %s should run in trace mode\n" ), program_name );
	fprintf( f, _( "\t-V  report %s version\n" ), program_name );
	fprintf( f, _( "\t-?  produce this help message\n" ) );
	fprintf( f, _( "\t-o  specify output filename\n" ) );
	fprintf( f, _( "\t-P  specify alternate prefix\n" ) );
	fprintf( f, _( "\t--help     produce this help message\n" ) );
	fprintf( f, _( "\t--version  report %s version\n" ), program_name );
	}
