/* Input parser for gyacc
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.
   Copyright (C) 1984, 1986, 1989 Free Software Foundation, Inc.

This file is part of Gyacc, the GNU Compiler Compiler.

Gyacc is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Gyacc is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Gyacc; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


/* read in the grammar specification and record it in the format described in gram.h.
  All guards are copied into the fguard file and all actions into faction,
  in each case forming the body of a C function (yyguard or yyaction)
  which contains a switch statement to decide which guard or action to execute.

The entry point is reader().  */

#include <stdio.h>
#include <ctype.h>
#include "system.h"
#include "files.h"
#include "new.h"
#include "symtab.h"
#include "lex.h"
#include "gram.h"
#include "machine.h"

/* Number of slots allocated (but not necessarily used yet) in `rline'  */
int rline_allocated;

extern char *program_name;
extern int nolinesflag;
extern bucket *symval;
extern int numval;
extern int failure;
extern int expected_conflicts;
extern char *token_buffer;
extern char *tmpactfile;

extern void init_lex();
extern void tabinit();
extern void output_headers();
extern void output_trailers();
extern void free_symtab();
extern void fatal();
extern void fatals();
extern void unlex();
extern void done();
extern FILE *tryopen ();

extern int skip_white_space();
extern int parse_percent_token();
extern int lex();

void read_declarations();
void copy_definition();
void copy_to_header();
void parse_token_decl();
void parse_start_decl();
void parse_type_decl();
void parse_assoc_decl();
void parse_union_decl();
void parse_expect_decl();
void copy_action();
void readgram();
void record_rule_line();
void packsymbols();
void output_token_defines();
void packgram();
int read_signed_integer();
int get_type();

typedef
  struct symbol_list
    {
      struct symbol_list *next;
      bucket *sym;
      bucket *ruleprec;
    }
  symbol_list;



int lineno;
symbol_list *grammar;
int start_flag;
bucket *startval;
char **tags;

/* Nonzero if components of semantic values are used, implying
   they must be unions.  */
static int value_components_used;

static int typed;  /* nonzero if %union has been seen.  */

static int lastprec;  /* incremented for each %left, %right or %nonassoc seen */

static int gensym_count;  /* incremented for each generated symbol */

static bucket *errtoken;

extern char *version_string;

void
reader()
{
  start_flag = 0;
  startval = NULL;  /* start symbol not specified yet. */

  nsyms = 1;
  nvars = 0;
  nrules = 0;
  nitems = 0;
  rline_allocated = 10;
  rline = NEW2(rline_allocated, short);

  typed = 0;
  lastprec = 0;

  gensym_count = 0;

  grammar = NULL;

  init_lex();
  lineno = 1;

  /* initialize the symbol table.  */
  tabinit();
  /* construct the error token */
  errtoken = getsym("error");
  errtoken->class = STOKEN;
  errtoken->user_token_number = 256; /* Value specified by posix.  */
  /* construct a token that represents all undefined literal tokens. */
  /* it is always token number 2.  */
  getsym("$illegal.")->class = STOKEN;
  /* Read the declaration section.  Copy %{ ... %} groups to ftable and fdefines file.
     Also notice any %token, %left, etc. found there.  */
  fprintf(ftable, "\n(*  A Gyacc parser, made from %s", infile);
  fprintf(ftable, " with gyacc version %s  *)\n\n", version_string);
  {
    int len = strlen (defsfile)-3;
    char hold = defsfile[len];

    defsfile[len] = 0;
    fprintf(ftable, "open \"%s\";\n", defsfile);
    defsfile[len] = hold;
  }
  read_declarations();
  /* start writing the guard and action files, if they are needed.  */
  output_headers();
  /* read in the grammar, build grammar in list form.  write out guards and actions.  */
  readgram();
  /* write closing delimiters for actions and guards.  */
  output_trailers();
  /* assign the symbols their symbol numbers.
     Write #defines for the token symbols into fdefines if requested.  */
  packsymbols();
  /* convert the grammar into the format described in gram.h.  */
  packgram();
  /* free the symbol table data structure
     since symbols are now all referred to by symbol number.  */
  free_symtab();
}



/* read from finput until %% is seen.  Discard the %%.
Handle any % declarations,
and copy the contents of any %{ ... %} groups to fattrs.  */

void
read_declarations ()
{
  register int c;
  register int tok;

  for (;;)
    {
      c = skip_white_space();

      if (c == '%')
	{
	  tok = parse_percent_token();

	  switch (tok)
	    {
	    case TWO_PERCENTS:
	      return;

	    case PERCENT_LEFT_CURLY:
	      copy_definition();
	      break;

	    case TOKEN:
	      parse_token_decl (STOKEN, SNTERM);
	      break;
	
	    case NTERM:
	      parse_token_decl (SNTERM, STOKEN);
	      break;
	
	    case TYPE:
	      parse_type_decl();
	      break;
	
	    case START:
	      parse_start_decl();
	      break;

	    case HEADER:
	      copy_to_header();
	      break;
	      
	    case UNION:
	      parse_union_decl();
	      break;
	
	    case EXPECT:
	      parse_expect_decl();
	      break;
	
	    case LEFT:
	      parse_assoc_decl(LEFT_ASSOC);
	      break;

	    case RIGHT:
	      parse_assoc_decl(RIGHT_ASSOC);
	      break;

	    case NONASSOC:
	      parse_assoc_decl(NON_ASSOC);
	      break;

	    default:
	      fatal("junk after `%%' in definition section");
	    }
	}
      else if (c == EOF)
        fatal("no input grammar");
      else if (c >= 040 && c <= 0177)
	fatals ("unknown character `%c' in declaration section", c);
      else
	fatals ("unknown character with code 0x%x in declaration section", c);
    }
}


/* copy the contents of a %{ ... %} into the definitions file,
or of a %header { ... %} into the header file.
The %{ or %header{ has already been read.  Return after reading the %}.  */

void copy_definition_to_file (f, decl)
     FILE *f;
     char *decl;
{
  register int c;
  register int match;
  register int after_percent;  /* -1 while reading a character if prev char was % */
  char errbuf[128];

  if (!nolinesflag)
    fprintf(f, "# %d \"%s\"\n", lineno, infile);

  after_percent = 0;

  c = getc(finput);

  for (;;)
    {
      switch (c)
	{
	case '\n':
	  putc(c, f);
	  lineno++;
	  break;

	case '%':
          after_percent = -1;
	  break;
	      
	case '"':
	  match = c;
	  putc(c, f);
	  c = getc(finput);

	  while (c != match)
	    {
	      if (c == EOF || c == '\n')
		fatal("unterminated string");

	      putc(c, f);
	      
	      if (c == '\\')
		{
		  c = getc(finput);
		  if (c == EOF)
		    fatal("unterminated string");
		  putc(c, f);
		  if (c == '\n')
		    lineno++;
		}

	      c = getc(finput);
	    }

	  putc(c, f);
	  break;

	case '(':
	  putc(c, f);
	  c = getc(finput);
	  if (c != '*')
	    continue;

	  putc(c, f);
	  c = getc(finput);

	  {
	    int nest = 1; /* current nesting level inside comments */
	    int state = 0; /* 0=in comment;
			      1=seen '(', waiting for '*';
			      2=seen '*', waiting for ')';
			   */

	    while (nest!=0)
	      switch (state)
		{
		case 0:
		  switch (c)
		    {
		    case '(': state = 1; break;
		    case '*': state = 2; break;
		    case '\n': lineno++; break;
		    case EOF:
		    comment_eof:
		      sprintf (errbuf,
			       "unterminated comment in `%s' definition", decl);
		      fatal(errbuf); break;
		    }
		comment_refill:
		  putc (c, f);
		  c = getc (finput);
		  break;
		case 1:
		  switch (c)
		    {
		    case '*': nest++; break;
		    case '\n': lineno++; break;
		    case EOF: goto comment_eof;
		    }
		  state = 0;
		  goto comment_refill;
		case 2:
		  switch (c)
		    {
		    case ')': state = 0; putc (c, f); if (--nest!=0) c = getc (finput); break;
		    case EOF: goto comment_eof;
		    case '\n': lineno++; /*FALLTHROUGH*/
		    default: state = 0; goto comment_refill;
		    }
		  break;
		}
	  }
	  break;

	case EOF:
	  sprintf (errbuf, "unterminated `%s' definition", decl);
	  fatal(errbuf);
	  break;

	default:
	  putc(c, f);
	  break;
	}

      c = getc(finput);

      if (after_percent)
	{
	  if (c == '}')
	    return;
	  putc('%', f);
	}
      after_percent = 0;
    }
}

void copy_definition ()
{
  copy_definition_to_file (fattrs, "%{");
}

void copy_to_header ()
{
  copy_definition_to_file (fdefines, "%header{");
}


/* parse what comes after %token or %nterm.
For %token, what_is is STOKEN and what_is_not is SNTERM.
For %nterm, the arguments are reversed.  */

void
parse_token_decl (what_is, what_is_not)
     int what_is, what_is_not;
{
/*   register int start_lineno; JF */
  register int token = 0;
  register int prev;
  register char *typename = 0;
  int k;

/*   start_lineno = lineno; JF */

  for (;;)
    {
      if (ungetc(skip_white_space(), finput) == '%')
	return;

/*      if (lineno != start_lineno)
	return; JF */

      /* we have not passed a newline, so the token now starting is in this declaration */
      prev = token;

      token = lex();
      if (token == COMMA)
	continue;
      if (token == TYPENAME)
	{
	  k = strlen(token_buffer);
	  typename = NEW2(k + 1, char);
	  strcpy(typename, token_buffer);
	  value_components_used = 1;
	}
      else if (token == IDENTIFIER)
	{
	  int oldclass = symval->class;

	  if (symval->class == what_is_not)
	    fatals("symbol %s redefined", symval->tag);
	  symval->class = what_is;
	  if (what_is == SNTERM && oldclass != SNTERM)
	    symval->value = nvars++;

	  if (typename)
	    {
	      if (symval->type_name == NULL)
		symval->type_name = typename;
	      else
		fatals("type redeclaration for %s", symval->tag);
	    }
	}
      else if (prev == IDENTIFIER && token == NUMBER)
        {
	  symval->user_token_number = numval;
        }
      else
	fatal("invalid text in %token or %nterm declaration");
    }

}



/* parse what comes after %start */

void
parse_start_decl ()
{
  if (start_flag)
    fatal("multiple %start declarations");
  start_flag = 1;
  if (lex() != IDENTIFIER)
    fatal("invalid %start declaration");
  startval = symval;
}



/* read in a %type declaration and record its information for get_type_name to access */

void
parse_type_decl ()
{
  register int k;
  register char *name;
/*   register int start_lineno; JF */

  if (lex() != TYPENAME)
    fatal("ill-formed %type declaration");

  k = strlen(token_buffer);
  name = NEW2(k + 1, char);
  strcpy(name, token_buffer);

/*   start_lineno = lineno; */

  for (;;)
    {
      register int t;

      if(ungetc(skip_white_space(), finput) == '%')
	return;

/*       if (lineno != start_lineno)
	return; JF */

      /* we have not passed a newline, so the token now starting is in this declaration */

      t = lex();

      switch (t)
	{

	case COMMA:
	case SEMICOLON:
	  break;

	case IDENTIFIER:
	  if (symval->type_name == NULL)
	    symval->type_name = name;
	  else
	    fatals("type redeclaration for %s", symval->tag);

	  break;

	default:
	  fatal("invalid %type declaration");
	}
    }
}



/* read in a %left, %right or %nonassoc declaration and record its information.  */
/* assoc is either LEFT_ASSOC, RIGHT_ASSOC or NON_ASSOC.  */

void
parse_assoc_decl (assoc)
int assoc;
{
  register int k;
  register char *name = NULL;
/*  register int start_lineno; JF */
  register int prev = 0;	/* JF added = 0 to keep lint happy */

  lastprec++;  /* Assign a new precedence level, never 0.  */

/*   start_lineno = lineno; */

  for (;;)
    {
      register int t;

      if(ungetc(skip_white_space(), finput) == '%')
	return;

      /* if (lineno != start_lineno)
	return; JF */

      /* we have not passed a newline, so the token now starting is in this declaration */

      t = lex();

      switch (t)
	{

	case TYPENAME:
	  k = strlen(token_buffer);
	  name = NEW2(k + 1, char);
	  strcpy(name, token_buffer);
	  break;

	case COMMA:
	  break;

	case IDENTIFIER:
	  if (symval->prec != 0)
	    fatals("redefining precedence of %s", symval->tag);
	  symval->prec = lastprec;
	  symval->assoc = assoc;
	  if (symval->class == SNTERM)
	    fatals("symbol %s redefined", symval->tag);
	  symval->class = STOKEN;
	  if (name)
	    { /* record the type, if one is specified */
	      if (symval->type_name == NULL)
		symval->type_name = name;
	      else
		fatals("type redeclaration for %s", symval->tag);
	    }
	  break;

	case NUMBER:
	  if (prev == IDENTIFIER)
            {
	      symval->user_token_number = numval;
            }
          else	  
	    fatal("invalid text in association declaration");
	  break;

	case SEMICOLON:
	  return;

	default:
	  fatal("malformatted association declaration");
	}

      prev = t;

    }
}



/* copy the union declaration into fdefines,
   where it is made into the
   definition of YYSTYPE, the type of elements of the parser value stack.  */

void
parse_union_decl()
{
  register int c;
  register int count;

  if (typed)
    fatal("multiple %union declarations");

  typed = 1;
  fprintf(fdefines, "datatype ");

  count = 0;

  c = getc(finput);

  while (c != EOF)
    {
      switch (c)
	{
	case '\n':
	  putc(c, fdefines);
	  lineno++;
	  break;

	case '(':
	  putc(c, fdefines);
	  c = getc(finput);
	  if (c != '*')
	    continue;
	  putc(c, fdefines);
	  c = getc(finput);

	  {
	    int nest = 1; /* current nesting level inside comments */
	    int state = 0; /* 0=in comment;
			      1=seen '(', waiting for '*';
			      2=seen '*', waiting for ')';
			   */

	    while (nest!=0)
	      switch (state)
		{
		case 0:
		  switch (c)
		    {
		    case '(': state = 1; break;
		    case '*': state = 2; break;
		    case '\n': lineno++; break;
		    case EOF:
		    comment_eof:
		    fatal("unterminated comment"); break;
		    }
		comment_refill:
		  putc (c, fdefines);
		  c = getc (finput);
		  break;
		case 1:
		  switch (c)
		    {
		    case '*': nest++; break;
		    case '\n': lineno++; break;
		    case EOF: goto comment_eof;
		    }
		  state = 0;
		  goto comment_refill;
		case 2:
		  switch (c)
		    {
		    case ')': state = 0; putc (c, fdefines); if (--nest!=0) c = getc (finput); break;
		    case EOF: goto comment_eof;
		    case '\n': lineno++; /*FALLTHROUGH*/
		    default: state = 0; goto comment_refill;
		    }
		  break;
		}
	  }
	  break;

	case '{':
	  if (++count==1)
	    {
	      fprintf (fdefines, " %s_value = ", spec_name_prefix);
	    }
	  else fatal ("illegal open-brace ('{')");
	  break;

	case '}':
	  if (count == 0)
	    fatal ("unmatched close-brace (`}')");
	  count--;
	  if (count == 0)
	    {
	      fprintf(fdefines, ";\n");
	      /* JF don't choke on trailing semi */
	      c=skip_white_space();
	      if(c!=';') ungetc(c,finput);
	      return;
	    }
	default:
	  putc (c, fdefines);
	  break;
	}

      c = getc(finput);
    }
}

/* parse the declaration %expect N which says to expect N
   shift-reduce conflicts.  */

void
parse_expect_decl()
{
  register int c;
  register int count;
  char buffer[20];

  c = getc(finput);
  while (c == ' ' || c == '\t')
    c = getc(finput);

  count = 0;
  while (c >= '0' && c <= '9')
    {
      if (count < 20)
	buffer[count++] = c;
      c = getc(finput);
    }
  buffer[count] = 0;

  ungetc (c, finput);

  expected_conflicts = atoi (buffer);
}

/* that's all of parsing the declaration section */

/* Get the data type (alternative in the union) of the value for symbol n in rule rule.  */

char *
get_type_name(n, rule)
int n;
symbol_list *rule;
{
  static char *msg = "invalid $ value";

  register int i;
  register symbol_list *rp;

  if (n < 0)
    fatal(msg);

  rp = rule;
  i = 0;

  while (i < n)
    {
      rp = rp->next;
      if (rp == NULL || rp->sym == NULL)
	fatal(msg);
      i++;
    }

  return (rp->sym->type_name);
}


/* Assuming that a { has just been seen, copy everything up to the matching }
into the actions file.
stack_offset is the number of values in the current rule so far,
which says where to find $0 with respect to the top of the stack.  */

static int n_copy_actions = 0;

void
copy_action(rule, stack_offset)
symbol_list *rule;
int stack_offset;
{
  register int c;
  register int n;
  register int count;
  register int match;
  register char *type_name;
  FILE *flocal;
  int letvaldone = 0, letlocdone = 0;
  long valsdone = 0;
  long locsdone = 0;
  char *spaces = "";
  char *tab = "    ";

  /* offset is always 0 if parser has already popped the stack pointer */

  for (; n_copy_actions<nrules; n_copy_actions++)
    fprintf (faction, ",\nyydefault");
  n_copy_actions++;
  fprintf (faction, ",\n(fn hyd => (\n");

  count = 1;
  c = getc(finput);

  flocal = tryopen (tmpactfile, "w+");

  while (count > 0)
    {
      while (c != '}')
        {
          switch (c)
	    {
	    case '\n':
	      putc(c, flocal);
	      lineno++;
	      break;

	    case '{':
	      putc(c, flocal);
	      count++;
	      break;

	      /*case '\'':*/
	    case '"':
	      match = c;
	      putc(c, flocal);
	      c = getc(finput);

	      while (c != match)
		{
		  if (c == EOF || c == '\n')
		    fatal("unterminated string");

		  putc(c, flocal);

		  if (c == '\\')
		    {
		      c = getc(finput);
		      if (c == EOF)
			fatal("unterminated string");
		      putc(c, flocal);
		      if (c == '\n')
			lineno++;
		    }

		  c = getc(finput);
		}

	      putc(c, flocal);
	      break;

	    case '(':
	      putc(c, flocal);
	      c = getc(finput);
	      if (c != '*')
		continue;

	      putc(c, flocal);
	      c = getc(finput);

	      {
		int nest = 1; /* current nesting level inside comments */
		int state = 0; /* 0=in comment;
				  1=seen '(', waiting for '*';
				  2=seen '*', waiting for ')';
			       */

		while (nest!=0)
		  switch (state)
		    {
		    case 0:
		      switch (c)
			{
			case '(': state = 1; break;
			case '*': state = 2; break;
			case '\n': lineno++; break;
			case EOF:
			comment_eof:
			fatal("unterminated comment"); break;
			}
		    comment_refill:
		      putc (c, flocal);
		      c = getc (finput);
		      break;
		    case 1:
		      switch (c)
			{
			case '*': nest++; break;
			case '\n': lineno++; break;
			case EOF: goto comment_eof;
			}
		      state = 0;
		      goto comment_refill;
		    case 2:
		      switch (c)
			{
			case ')': state = 0; putc (c, flocal); if (--nest!=0) c = getc (finput); break;
			case EOF: goto comment_eof;
			case '\n': lineno++; /*FALLTHROUGH*/
			default: state = 0; goto comment_refill;
			}
		      break;
		    }
	      }
	      break;

	    case '$':
	      c = getc(finput);
	      type_name = NULL;

	      if (c=='$') /* acts as a return, i.e. we write '$$ result', not '$$ = result;' as in C. */
		{
		  if (!type_name) type_name = rule->sym->type_name;
		  if(!type_name && typed)	/* JF */
		    fprintf(stderr,"%s:%d:  warning:  $$ of '%s' has no declared type.\n",infile,lineno,rule->sym->tag);
		  if (type_name)
		    fprintf (flocal, "(SOME o %s) ", type_name); /* we wish to write SOME (type arg), however this
								    requires that we know where the argument arg
								    ends; instead we use the roundabout notation
								    (SOME o type) arg, which does the same thing. */
		  else fprintf (flocal, "SOME ");
		}
	      else if (isdigit(c) || c == '-')
		{
		  int m;

		  ungetc (c, finput);
		  n = read_signed_integer(finput);
		  m = n - stack_offset;
		  c = getc(finput);

		  if (!type_name && n > 0)
		    type_name = get_type_name(n, rule);
		  if(!type_name && typed)	/* JF */
		    fprintf(stderr,"%s:%d:  warning:  $%d of '%s' has no declared type.\n",infile,lineno,n,rule->sym->tag);
		  if (!(letvaldone || letlocdone))
		    fputs ("let ", faction);
		  if (!letvaldone)
		    {
		      letvaldone = 1;
		      fprintf (faction, "%sval %s_value = gyacc_value hyd\n", spaces, spec_name_prefix);
		      spaces = tab;
		    }
		  if (n>8*sizeof(long) || (valsdone & (1L << n))==0)
		    {
		      valsdone |= 1L << n;
		      if (type_name)
			fprintf (faction, "%sval %s %s_val%d = %s_value ~%d\n",
				 spaces, type_name, spec_name_prefix, -m, spec_name_prefix, -m);
		      else fprintf (faction, "%sval %s_val%d = %s_value ~%d\n",
				    spaces, spec_name_prefix, -m, spec_name_prefix, -m); /* then all values are of the
											  same type */
		      /* if n>8*sizeof(long)=32, then the same yy_val value can be generated more than once;
			 we estimate that it should be rare. */
		    }
		  fprintf (flocal, "%s_val%d", spec_name_prefix, -m);
		  continue;
		}
	      else
		{
#ifdef OBSOLETE
		  fatals("$%c is invalid",c);	/* JF changed format */
#endif
		  putc ('$', flocal); /* allow for GimML symbols based on $. */
		  putc (c, flocal);
		}
	      break;

	    case '@':
	      c = getc(finput);
	      if (isdigit(c) || c == '-')
		{
		  int m;

		  ungetc (c, finput);
		  n = read_signed_integer(finput);
		  m = n - stack_offset;
		  c = getc(finput);
		  if (!(letvaldone || letlocdone))
		    fputs ("let ", faction);
		  if (!letlocdone)
		    {
		      letlocdone = 1;
		      fprintf (faction, "%sval %s_location = gyacc_location hyd\n", spaces, spec_name_prefix);
		      spaces = tab;
		    }
		  if (n>8*sizeof(long) || (locsdone & (1L << n))==0)
		    {
		      locsdone |= 1L << n;
		      fprintf (faction, "%sval %s_loc%d = %s_location ~%d\n",
			       spaces, spec_name_prefix, -m, spec_name_prefix, -m);
		      /* if n>8*sizeof(long)=32, then the same yy_loc value can be generated more than once;
			 we estimate that it should be rare. */
		    }
		  fprintf(flocal, "%s_loc%d", spec_name_prefix, -m);
		}
	      else
		{
#ifdef OBSOLETE
		  fatal("invalid @-construct");
#endif
		  putc ('@', flocal);
		  putc (c, flocal);
		}

	      continue;

	    case EOF:
	      fatal("unmatched '{'");

	    default:
	      putc(c, flocal);
	    }

          c = getc(finput);
        }

      /* above loop exits when c is '}' */

      if (--count)
        {
	  putc(c, flocal);
	  c = getc(finput);
	}
    }

  if (!nolinesflag) /* need to put #line directive here,
		       because it is parsed by GimML as a declaration;
		       the "in" keyword uses one additional line,
		       so we adjust by subtracting 1 from lineno. */
    {
      if (!(letvaldone || letlocdone))
	fputs ("let ", faction);
      fprintf(faction, "# %d \"%s\"\n", lineno-1, infile);
    }
  if (!nolinesflag || letvaldone || letlocdone)
    fputs ("in\n", faction);
  rewind (flocal);
  {
    register int c;

    c = '\n';
    while (c!=EOF)
      {
	c = getc (flocal);
	if (c!=EOF)
	  {
	    fputs (spaces, faction);
	    while (c!=EOF && (putc (c, faction), c!='\n'))
	      c = getc (flocal);
	  }
      }
  }
  fclose (flocal);
  if (!nolinesflag || letvaldone || letlocdone)
    fputs ("\nend", faction);
  fprintf(faction, "))");
}



/* generate a dummy symbol, a nonterminal,
whose name cannot conflict with the user's names. */

bucket *
gensym()
{
  register bucket *sym;

  sprintf (token_buffer, "@%d", ++gensym_count);
  sym = getsym(token_buffer);
  sym->class = SNTERM;
  sym->value = nvars++;
  return (sym);
}

/* Parse the input grammar into a one symbol_list structure.
Each rule is represented by a sequence of symbols: the left hand side
followed by the contents of the right hand side, followed by a null pointer
instead of a symbol to terminate the rule.
The next symbol is the lhs of the following rule.

All guards and actions are copied out to the appropriate files,
labelled by the rule number they apply to.  */

void
readgram()
{
  register int t;
  register bucket *lhs;
  register symbol_list *p;
  register symbol_list *p1;
  register bucket *bp;

  symbol_list *crule;	/* points to first symbol_list of current rule.  */
			/* its symbol is the lhs of the rule.   */
  symbol_list *crule1;  /* points to the symbol_list preceding crule.  */

  p1 = NULL;

  t = lex();

  while (t != TWO_PERCENTS && t != ENDFILE)
    {
      if (t == IDENTIFIER || t == BAR)
	{
	  register int actionflag = 0;
	  int rulelength = 0;  /* number of symbols in rhs of this rule so far  */
	  int xactions = 0;	/* JF for error checking */
	  bucket *first_rhs = 0;

	  if (t == IDENTIFIER)
	    {
	      lhs = symval;
    
	      t = lex();
	      if (t != COLON)
		fatal("ill-formed rule");
	    }

	  if (nrules == 0)
	    {
	      if (t == BAR)
		fatal("grammar starts with vertical bar");

	      if (!start_flag)
		startval = lhs;
	    }

	  /* start a new rule and record its lhs.  */

	  nrules++;
	  nitems++;

	  record_rule_line ();

	  p = NEW(symbol_list);
	  p->sym = lhs;

	  crule1 = p1;
	  if (p1)
	    p1->next = p;
	  else
	    grammar = p;

	  p1 = p;
	  crule = p;

	  /* mark the rule's lhs as a nonterminal if not already so.  */

	  if (lhs->class == SUNKNOWN)
	    {
	      lhs->class = SNTERM;
	      lhs->value = nvars;
	      nvars++;
	    }
	  else if (lhs->class == STOKEN)
	    fatals("rule given for %s, which is a token", lhs->tag);

	  /* read the rhs of the rule.  */

	  for (;;)
	    {
	      t = lex();

	      if (! (t == IDENTIFIER || t == LEFT_CURLY)) break;

	      /* If next token is an identifier, see if a colon follows it.
		 If one does, exit this rule now.  */
	      if (t == IDENTIFIER)
		{
		  register bucket *ssave;
		  register int t1;

		  ssave = symval;
		  t1 = lex();
		  unlex(t1);
		  symval = ssave;
		  if (t1 == COLON) break;

		  if(!first_rhs)	/* JF */
		    first_rhs = symval;
		  /* Not followed by colon =>
		     process as part of this rule's rhs.  */
		}

	      /* If we just passed an action, that action was in the middle
		 of a rule, so make a dummy rule to reduce it to a
		 non-terminal.  */
	      if (actionflag)
		{
		  register bucket *sdummy;

		  /* Since the action was written out with this rule's */
		  /* number, we must write give the new rule this number */
		  /* by inserting the new rule before it.  */

		  /* Make a dummy nonterminal, a gensym.  */
		  sdummy = gensym();

		  /* Make a new rule, whose body is empty,
		     before the current one, so that the action
		     just read can belong to it.  */
		  nrules++;
		  nitems++;
		  record_rule_line ();
		  p = NEW(symbol_list);
		  if (crule1)
		    crule1->next = p;
		  else grammar = p;
		  p->sym = sdummy;
		  crule1 = NEW(symbol_list);
		  p->next = crule1;
		  crule1->next = crule;

		  /* insert the dummy generated by that rule into this rule.  */
		  nitems++;
		  p = NEW(symbol_list);
		  p->sym = sdummy;
		  p1->next = p;
		  p1 = p;

		  actionflag = 0;
		}

	      if (t == IDENTIFIER)
		{
		  nitems++;
		  p = NEW(symbol_list);
		  p->sym = symval;
		  p1->next = p;
		  p1 = p;
		}
	      else /* handle an action.  */
		{
		  copy_action(crule, rulelength);
		  actionflag = 1;
		  xactions++;	/* JF */
		}
	      rulelength++;
	    }

	  /* Put an empty link in the list to mark the end of this rule  */
	  p = NEW(symbol_list);
	  p1->next = p;
	  p1 = p;

	  if (t == PREC)
	    {
	      t = lex();
	      crule->ruleprec = symval;
	      t = lex();
	    }
	  if (t == LEFT_CURLY)
	    {
	      if (actionflag) fatal("two actions at end of one rule");
	      copy_action(crule, rulelength);
	      t = lex();
	    }
	  if (t == SEMICOLON)
	    t = lex();
	}
      /* these things can appear as alternatives to rules.  */
      else if (t == TOKEN)
	{
	  parse_token_decl(STOKEN, SNTERM);
	  t = lex();
	}
      else if (t == NTERM)
	{
	  parse_token_decl(SNTERM, STOKEN);
	  t = lex();
	}
      else if (t == TYPE)
	{
	  t = get_type();
	}
      else if (t == UNION)
	{
	  parse_union_decl();
	  t = lex();
	}
      else if (t == EXPECT)
	{
	  parse_expect_decl();
	  t = lex();
	}
      else if (t == START)
	{
	  parse_start_decl();
	  t = lex();
	}
      else
	fatal("invalid input");
    }

  if (nsyms > MAXSHORT)
    fatals("too many symbols (tokens plus nonterminals); maximum %d",
	   MAXSHORT);
  if (nrules == 0)
    fatal("no input grammar");

#ifdef OBSOLETE
  if (typed == 0	/* JF put out same default YYSTYPE as YACC does */
      && !value_components_used)
    {
      /* We used to use `unsigned long' as YYSTYPE on MSDOS,
	 but it seems better to be consistent.
	 Most programs should declare their own type anyway.  */
      fprintf(fattrs, "#ifndef YYSTYPE\n#define YYSTYPE int\n#endif\n");
      fprintf(fdefines, "#ifndef YYSTYPE\n#define YYSTYPE int\n#endif\n");
    }
#endif

  /* Report any undefined symbols and consider them nonterminals.  */

  for (bp = firstsymbol; bp; bp = bp->next)
    if (bp->class == SUNKNOWN)
      {
	fprintf(stderr, "symbol %s used, not defined as token, and no rules for it\n",
			bp->tag);
	failure = 1;
	bp->class = SNTERM;
	bp->value = nvars++;
      }

  /* added by jgl 11 Jul 2003, otherwise could lead to core dump if last rules did not
     return anything (not declared with %type <...>) */
  for (; n_copy_actions<nrules; n_copy_actions++)
    fprintf (faction, ",\nyydefault");

  ntokens = nsyms - nvars;
}


void
record_rule_line ()
{
  /* Record each rule's source line number in rline table.  */

  if (nrules >= rline_allocated)
    {
      rline_allocated = nrules * 2;
      rline = (short *) xrealloc (rline,
				  rline_allocated * sizeof (short));
    }
  rline[nrules] = lineno;
}


/* read in a %type declaration and record its information for get_type_name to access */

int
get_type()
{
  register int k;
  register int t;
  register char *name;

  t = lex();

  if (t != TYPENAME)
    fatal("ill-formed %type declaration");

  k = strlen(token_buffer);
  name = NEW2(k + 1, char);
  strcpy(name, token_buffer);

  for (;;)
    {
      t = lex();

      switch (t)
	{
	case SEMICOLON:
	  return (lex());

	case COMMA:
	  break;

	case IDENTIFIER:
	  if (symval->type_name == NULL)
	    symval->type_name = name;
	  else
	    fatals("type redeclaration for %s", symval->tag);

	  break;

	default:
	  return (t);
	}
    }
}



/* assign symbol numbers, and write definition of token names into fdefines.
Set up vectors tags and sprec of names and precedences of symbols.  */

void
packsymbols()
{
  register bucket *bp;
  register int tokno = 1;
  register int i;
  register int last_user_token_number;

  /* int lossage = 0; JF set but not used */

  tags = NEW2(nsyms + 1, char *);
  tags[0] = "$";

  sprec = NEW2(nsyms, short);
  sassoc = NEW2(nsyms, short);

  max_user_token_number = 256;
  last_user_token_number = 256;

  for (bp = firstsymbol; bp; bp = bp->next)
    {
      if (bp->class == SNTERM)
	{
	  bp->value += ntokens;
	}
      else
	{
	  if (!bp->user_token_number)
	    bp->user_token_number = ++last_user_token_number;
	  if (bp->user_token_number > max_user_token_number)
	    max_user_token_number = bp->user_token_number;
	  bp->value = tokno++;
	}

      tags[bp->value] = bp->tag;
      sprec[bp->value] = bp->prec;
      sassoc[bp->value] = bp->assoc;

    }

  {
    register int i;

    token_translations = NEW2(max_user_token_number+1, short);

    /* initialize all entries for literal tokens to 2,
       the internal token number for $illegal., which represents all invalid inputs.  */
    for (i = 0; i <= max_user_token_number; i++)
      token_translations[i] = 2;      
  }

  for (bp = firstsymbol; bp; bp = bp->next)
    {
      if (bp->value >= ntokens) continue;
      if (token_translations[bp->user_token_number] != 2)
	{
	  /* JF made this a call to fatals() */
	  fatals( "tokens %s and %s both assigned number %d",
		  tags[token_translations[bp->user_token_number]],
		  bp->tag,
		  bp->user_token_number);
	}
      token_translations[bp->user_token_number] = bp->value;
    }

  error_token_number = errtoken->value;

  /*output_token_defines(ftable);*/

  if (startval->class == SUNKNOWN)
    fatals("the start symbol %s is undefined", startval->tag);
  else if (startval->class == STOKEN)
    fatals("the start symbol %s is a token", startval->tag);

  start_symbol = startval->value;

  output_token_defines(fdefines);
}
      

void
output_token_defines(file)
FILE *file;
{
  bucket *bp;

  for (bp = firstsymbol; bp; bp = bp->next)
    {
      if (bp->value >= ntokens) continue;

      /* For named tokens, but not literal ones, define the name.  */
      /* The value is the user token number.  */

      if ('\'' != *tags[bp->value] && bp != errtoken)
	{
	  register char *cp = tags[bp->value];
	  register char c;

	  /* Don't #define nonliteral tokens whose names contain periods.  */

	  while ((c = *cp++) && c != '.');
	  if (!c)
	    {
              fprintf(file, "val %s = %d\n", tags[bp->value],
		      bp->user_token_number);
	    }
	}
    }

  putc('\n', file);
}



/* convert the rules into the representation using rrhs, rlhs and ritems.  */

void
packgram()
{
  register int itemno;
  register int ruleno;
  register symbol_list *p;
/*  register bucket *bp; JF unused */

  bucket *ruleprec;

  ritem = NEW2(nitems + 1, short);
  rlhs = NEW2(nrules, short) - 1;
  rrhs = NEW2(nrules, short) - 1;
  rprec = NEW2(nrules, short) - 1;
  rprecsym = NEW2(nrules, short) - 1;
  rassoc = NEW2(nrules, short) - 1;

  itemno = 0;
  ruleno = 1;

  p = grammar;
  while (p)
    {
      rlhs[ruleno] = p->sym->value;
      rrhs[ruleno] = itemno;
      ruleprec = p->ruleprec;

      p = p->next;
      while (p && p->sym)
	{
	  ritem[itemno++] = p->sym->value;
	  /* A rule gets by default the precedence and associativity
	     of the last token in it.  */
          if (p->sym->class == STOKEN)
	    {
	      rprec[ruleno] = p->sym->prec;
	      rassoc[ruleno] = p->sym->assoc;
	    }
	  if (p) p = p->next;
	}

      /* If this rule has a %prec,
	 the specified symbol's precedence replaces the default.  */
      if (ruleprec)
	{
          rprec[ruleno] = ruleprec->prec;
          rassoc[ruleno] = ruleprec->assoc;
	  rprecsym[ruleno] = ruleprec->value;
	}

      ritem[itemno++] = -ruleno;
      ruleno++;

      if (p) p = p->next;
    }

  ritem[itemno] = 0;
}

/* Read a signed integer from STREAM and return its value.  */

int
read_signed_integer (stream)
     FILE *stream;
{
  register int c = getc(stream);
  register int sign = 1;
  register int n;

  if (c == '-')
    {
      c = getc(stream);
      sign = -1;
    }
  n = 0;
  while (isdigit(c))
    {
      n = 10*n + (c - '0');
      c = getc(stream);
    }

  ungetc(c, stream);

  return n * sign;
}
