
open "intsets_h.ml"

datatype 'a IntSet =
    INTSET of 'a list;

(* Used to do some Bus errors in alloctmp(), due to an alloctmp(yylen)
 in yyparse() where yylen was != strlen(yytext).  The syntax error at
 line 79 is normal (you may correct it if you want). *)

(* an IntSet is a list of data representing integers.
 Assuming I have two functions <= (provably less than or equal to)
 and = (provably equal), let != denote not = (not provably equal), then:
     INTSET [i1, i2, ..., ik]
     is valid when i1 <= i2 <= ... <= ik and i1 != i2 != ... != ik
     and denotes the set [i1, i2[ U [i3, i4[ U ... [i(k-1), ik[ if k is even
     and [i1, i2[ U [i3, i4[ U ... [ik, +infty[ if k is odd.
 In particular, INTSET nil = empty set, and
 INTSET (i::l) = [i, +inf) - INTSET l.
*)

fun intset_funs (pleq : 'a * 'a -> bool, pless : 'a * 'a -> bool, zero : 'a) =
    (* interval_minus (i, l) computes [i,+inf) - l *)
    let fun interval_minus (i, nil) = [i]
	  | interval_minus (i, ll as i'::l) =
	    if pleq (i',i)
		then interval_inter (i, l)
	    else if pleq (i,i')
		     then i::ll
		 else raise IntSet
	and interval_inter (i, nil) = nil
	  | interval_inter (i, ll as i'::l) =
	    if pleq (i,i')
		then ll
	    else if pleq (i',i)
		     then interval_minus (i,l)
		 else raise IntSet
	fun negate (INTSET l) = INTSET (interval_minus (zero, l))

	(*
	(* interval_union (i, l) computes l U [i,+inf) *)
	fun interval_union (i, n) = [i]
	  | interval_union (i, i'::l) =
	    if pleq (i, i')
		then [i]
	    else if pleq (i', i)
		     then (* [i,+inf) U ([i', +inf) - l) =
			   [i',+inf) - (l - [i,+inf)) *)
			 i'::minus_interval (i, l)
		 else raise IntSet
	(* minus_interval (i, l) computes l - [i,+inf) *)
	and minus_interval (i, nil) = nil
	  | minus_interval (i, i'::l) =
	    if pleq (i, i')
		then nil
	    else if pleq (i', i)
		     then (* [i',+inf)-l - [i,+inf) = [i',+inf) - ([i,+inf) U l) *)
			 i'::interval_union (i, l)
		 else raise IntSet
		     *)

	fun inter (nil, _) = nil (* nil = empty set *)
	  | inter (_, nil) = nil
	  | inter (i1::l1, i2::l2) =
	    if pleq (i1, i2)
		then (* ([i1,+inf)-l1) & ([i2,+inf)-l2) = [i2,+inf) - (l1 U l2) *)
		    interval_minus (i2, union (l1, l2))
	    else if pleq (i2, i1)
		     then interval_minus (i1, union (l1, l2))
		 else raise IntSet
	and union (nil, l) = l
	  | union (l, nil) = l
	  | union (ll1 as i1::l1, ll2 as i2::l2) =
	    if pleq (i1, i2)
		then (* ([i1,+inf)-l1) U ll2 = [i1,+inf) - (l1 - ll2) *)
		    interval_minus (i1, diff (l1, ll2))
	    else if pleq (i2, i1)
		     then interval_minus (i2, diff (l2, ll1))
	and and diff (nil, _) = nil
	  | diff (l, nil) = l
	  | diff (ll1 as i1::l1, ll2 as i2::l2) =
	    if pleq (i1, i2)
		then i1 :: union (l1, ll2)
	    else if pleq (i2, i1)
		     then inter (ll1, l2)
		 else raise IntSet

	fun wrap2 f (INTSET l1, INTSET l2) = INTSET (f (l1, l2))

	fun must_be_in (_, nil) = false
	  | must_be_in (x, i::l) = pleq (i,x) andalso must_not_be_in (x, l)
	and must_not_be_in (_, nil) = true
	  | must_not_be_in (x, i::l) = pless (x,i) orelse must_be_in (x, l)

	fun wrapb f (x, INTSET l) = f (x, l)

    in
	|[
	   inter = wrap2 (inter),
	   union = wrap2 (union),
	   diff = wrap2 (diff),
	   negate = negate,
	   must_be_in = wrapb must_be_in,
	   must_not_be_in = wrapb must_not_be_in,
	   may_not_be_in = not o wrapb must_be_in,
	   may_be_in = not o wrapb must_not_be_in,
	   open_interval = (fn i => INTSET [i]),
	   interval = (fn (i, j) => INTSET [i,j]),
	   shift = (fn f => let val m = map f
			    in
				fn INTSET l => INTSET (m l)
			    end),
	   empty = (fn (INTSET nil) => true | _ => false),
	   equal = op = : 'a IntSet * 'a IntSet -> bool
	 ]|
    end;

