(* Generic congruence closure implementation.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* use "ufind.ml"
 or else you'll get a syntax error after UFIND below. *)

infixr  5 %%

datatype ''var bintree = (* the type of generic terms *)
    NULL
  | LEAF of ''var
  | %% of ''var bintree * ''var bintree

local
  type ''a heap = ''a bintree
  fun heapConcat (NULL,h) = h
    | heapConcat (h,NULL) = h
    | heapConcat (h1,h2) = h1 %% h2

  type (''var,'other) info = |[parents : num * ''var bintree heap, others : 'other]|

  type (''var,'other) sigtable =
    ''var bintree * ''var bintree -m> ''var bintree color

in
  datatype (''var,'other) congruence =
    CONGRUENCE of |[find : ''var bintree -> ''var bintree color,
		    oneof : ''var bintree color -> ''var bintree,
		    info : ''var bintree color -> 'other,
		    setinfo : ''var bintree color * 'other -> unit,
		    colored : ''var bintree -> bool,
		    copy : ('other -> 'other) -> (''var,'other) congruence,
		    colorize : (''var bintree -> 'other) *
				(''var bintree color * ''var bintree color
				 * 'other * 'other -> 'other)
				-> ''var bintree -> ''var bintree color,
		    equalize : (''var bintree -> 'other) *
				(''var bintree color * ''var bintree color
				 * 'other * 'other -> 'other)
				-> (''var bintree * ''var bintree) set -> unit
		    ]|
  local
  fun new (ufind as (UFIND |[find,oneof,info,setinfo,colored,copy,newcolor,merge,...]|),
	   sigr) =
      let fun merge1 mergeother =
	      merge (fn (c1,c2,
			 |[parents = (n1,bt1),others = o1]|,
			 |[parents = (n2,bt2),others = o2]|) =>
		     |[parents = (n1+n2,heapConcat (bt1,bt2)),
		       others = mergeother (c1,c2,o1,o2)
		       ]|)
	  fun add1parent (c,p) =
	      let val |[parents = (n,bt),others = ot]| = info c
	      in
		  setinfo (c,|[parents = (n+1,heapConcat (LEAF p,bt)),others = ot]|)
	      end
	  fun colorize (newother,mergeother) =
	      let val merge2 = merge1 mergeother
		  fun colorize t =
		  if colored t
		      then find t
		  else (case t of
			   LEAF x => newcolor (t,|[parents = (0,NULL),
						   others = newother t]|)
			 | t1%%t2 =>
			   let val c1 = colorize t1 and c2 = colorize t2
			       val csig = (c1,c2)
			       val c' = newcolor (t,|[parents = (0,NULL),
						      others = newother t]|)
			   in
			      if csig inset !sigr
				 then merge2 (?(!sigr)csig,c')
			      else (add1parent (c1,t);
				    (if c1=c2
					then ()
				     else add1parent (c2,t));
				     c')
			   end)
	      in
		 colorize
	      end
      in CONGRUENCE
	  |[find = find,
	    oneof = oneof,
	    info = #others o info,
	    setinfo = fn (c,ot) => setinfo (c,(info c) ++|[others = ot]|),
	    colored = colored,
	    copy = fn othercopy =>
		new (copy (fn |[parents,others]| =>
				|[parents = parents,others = othercopy others]|),
		     ref (!sigr)),
	    colorize = colorize,
	    equalize = fn (fns as (newother,mergeother)) =>
			  let val combine = ref nil and pending = ref nil
			      val colorize = colorize fns
			      val merge2 = merge1 mergeother
			      fun killmerge (cv,cw,btw) =
				  let fun kill NULL = ()
					| kill (LEAF (t as t1%%t2)) =
					  let val c1 = find t1 and c2 = find t2
					      val csig = (c1,c2)
					  in
					      sigr := {csig} <-| !sigr;
					      pending := t :: !pending
					  end
					| kill (bt1%%bt2) =
					  (kill bt1; kill bt2)
				  in
				     kill btw;
				     merge2 (cv,cw)
				  end
			  in
			     fn eqns =>
				(combine := [eqn | eqn in set eqns];
				 while !combine <> nil do
				   (while !combine <> nil do
				      let val (v,w)::rest = !combine
					  val cv = colorize v and cw = colorize w
				      in
					  combine := rest;
					  if cv=cw
					     then cv
					  else let val (nv,btv) = #parents (info cv)
						   and (nw,btw) = #parents (info cw)
					       in
						  if nv<nw
						     then killmerge (cw,cv,btv)
						  else killmerge (cv,cw,btw)
					       end
				      end;
				    while !pending <> nil do
				      let val (t as t1%%t2) :: rest = !pending
					  val c1 = find t1 and c2 = find t2
					  val csig = (c1,c2)
				      in
					 pending := rest;
					 if csig inset !sigr
					    then combine := (t,oneof (?(!sigr)csig))
							    :: !combine
					 else sigr := !sigr ++ {csig => find t}
				      end))
			  end
	    ]|
      end
  in
  fun congruenceNew (newother : ''_var bintree -> '_other) =
      let val uf as (UFIND |[newcolor,...]|) = ufindNew ()
      in
	  newcolor (NULL,|[parents = (0,NULL),others = newother NULL]|);
	  new (uf,ref {}) : (''_var,'_other) congruence
      end
  end
end

(*
  in STD_ML_NJ_TYVARS:
NULL : ''a bintree
LEAF : ''a -> ''a bintree
op %% : ''a bintree * ''a bintree -> ''a bintree

CONGRUENCE : |[colored : ''a bintree -> bool,
	       oneof : ''a bintree color -> ''a bintree,
	       copy : ('b -> 'b) -> (''a,'b) congruence,
	       info : ''a bintree color -> 'b,
	       colorize : (''a bintree -> 'b)
	                    * (''a bintree color * ''a bintree color * 'b * 'b -> 'b)
	                   -> ''a bintree -> ''a bintree color,
	       find : ''a bintree -> ''a bintree color,
	       setinfo : ''a bintree color * 'b -> unit,
	       equalize : (''a bintree -> 'b)
	                    * (''a bintree color * ''a bintree color * 'b * 'b -> 'b)
	                   -> (''a bintree * ''a bintree) set -> unit]|
	      -> (''a,'b) congruence

congruenceNew : (''1a bintree -> '1b) -> (''1a,'1b) congruence
*)
