(* Lazy evaluation examples, based on promises.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

datatype 'a lazylist = lazynil | lazycons of 'a * 'a lazylist promise

	(* lazymap : ('a lazylist -> 'b lazylist) * 'a lazylist -> 'b lazylist *)
fun lazymap f lazynil = lazynil
|   lazymap f (lazycons(x,l)) = lazycons(f x,delay lazymap f (force l))

	(* lazyappend : 'a lazylist * 'a lazylist promise -> 'a lazylist *)
fun lazyappend(lazynil,l2) = force l2
|   lazyappend(lazycons(x,l1),l2) = lazycons(x,delay lazyappend(force l1,l2))

	(* eqlist : 'a lazylist * 'a lazylist -> bool *)
fun eqlist(lazynil,lazynil) = true
|   eqlist(lazycons(x1,p1),lazycons(x2,p2)) = x1=x2 andalso eqlist(force p1,force p2)
|   eqlist(_,_) = false

val rec ones = lazycons(1,delay ones)   (* infinite list of ones *)
	(* produces a warning, as the recursive definition is not safe *)

datatype 'a bintree =   Lf of 'a
	|   /\ of ('a bintree * 'a bintree) promise

	(* 'a bintree -> 'a lazylist *)
fun leavesof (Lf x) = lazycons(x,delay lazynil)
|   leavesof (/\ p) = let val (left,right)=force p
	in lazyappend(leavesof left,delay leavesof right)
	end

fun eqleaves(t1,t2) = eqlist(leavesof t1,leavesof t2)

	(* list of natural numbers *)

	(* from : int -> int lazylist *)
fun from n = lazycons(n,delay from (n+1))
val N = from 0  (* N : int lazylist *)

	(* list of prime numbers *)

fun nextfind(a,b,l as lazycons(c,x)) =  (* nextfind : int * int * int lazylist -> int lazylist *)
	if c<b
	then lazycons(c,delay nextfind(a,b,force x))
	else if c=b
	then nextfind(a,a+b,force x)
	else nextfind(a,a+b,l)
fun sift(a,x) = nextfind(a,a+a,x)
	(* sift : int * int lazylist -> int lazylist *)
fun sieve (lazycons(a,x)) = lazycons(a,delay sieve(sift(a,force x)))
	(* sieve : int lazylist -> int lazylist *)
val primes = sieve (from 2) (* primes : int lazylist *)

	(* hamming numbers (2**p*3**q*5**r, for p,q,r integers) *)

fun merge2(m as lazycons(a,x),n as lazycons(b,y)) =
	(* merge2 : int lazylist * int lazylist -> int lazylist *)
	if a<b
	then lazycons(a,delay merge2(force x,n))
	else if a>b
	then lazycons(b,delay merge2(m,force y))
	else lazycons(a,delay merge2(force x,force y))
fun merge3(x,y,z) = merge2(x,merge2(y,z))
	(* merge3 : int lazylist * int lazylist * int lazylist -> int lazylist *)
val rec hamming = lazycons(1,   (* hamming : int lazylist *)
	delay merge3(   lazymap (fn x=>2*x) hamming,
	lazymap (fn x=>3*x) hamming,
	lazymap (fn x=>5*x) hamming))
	(* produces a warning, as the recursive definition is not safe *)

	(* the replacemin function; problem due to Bird(1984) *)
	(* would work similarly without promises *)

	(* fmin : int bintree -> ((int -> int bintree) * int) *)
fun fmin (Lf a) = (fn m => Lf m,a)
  | fmin (/\ p) =
    let val (t1,t2)=force p
	val (f1,min1)=fmin t1 and (f2,min2)=fmin t2
    in
	(fn m => /\ (delay (f1 m,f2 m)),min(min1,min2))
    end
fun replacemin bt = let val (f,m)=fmin bt in f m end	(* replacemin : int bintree -> int bintree *)
