(* Prompts.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* Prompts, as described in
 "A Generalization of Exceptions and Control in ML-like Languages"
 Carl A. Gunter, Didier Remy, Jon G. Riecke
 FPCA'95.
*)
(* with some slight changes, to suit my personal tastes. *)

abstype
    'a prompt =
    PROMPT of
    ((bool * 'a cont) -> exn) (* In *) *
    (((bool * 'a cont) -> 'a cont * exn list)
     -> (exn -> 'a control)
     -> (exn -> 'a control)) (* Out *)
withtype
    'a control = 'a cont * exn list
with
    exception Uncaught_prompt
    exception Core_dumped

    fun new_prompt () =
	let exception Sharp of (bool * '_a cont)
	in
	    PROMPT (Sharp,
		    fn yes => fn no =>
		    (fn Sharp p => yes p
		      | z => no z) )
	end

    local
	val stack = ref ([] : exn list)
	fun push pc = stack := pc :: !stack
	fun pop () =
	    case !stack of
		nil => raise Uncaught_prompt
	      | pc::rest => (stack := rest;
			     pc)

	fun pop_control (PROMPT (In,Out) : 'a prompt) =
	    let fun pop_it control =
		Out
		(fn (b,c) =>
		 if b
		     then (c,control)
		 else pop_it (In (b,c) :: control))
		(fn pc => pop_it (pc::control))
		(pop ())
	    in
		pop_it nil
	    end
	fun push_control (pc::control) =
	    (push pc;
	     push_control control)
	  | push_control nil = ()
    in
	fun set (PROMPT (In,Out)) e =
	    callcc (fn normal_cont =>
		    (push (In (true, normal_cont));
		     let val v = e()
			 val (eff_cont,_) = Out
			     (fn (b,c) => (c,nil))
			     (fn sc => raise Core_dumped)
			     (pop ())
		     in
			 throw eff_cont v
		     end))

	    fun cupto (p as PROMPT (In,Out)) f =
		let val (abort,control) = pop_control p
		in
		    callcc (fn x =>
			    throw abort
			    (f (fn v =>
				callcc (fn after =>
					(push (In (false,after));
					 push_control control;
					 throw x v)))))
		end
    end
end

(* The following are examples drawn from the paper: *)

(*
 val p = new_prompt () : int prompt;
 5 > (set p (fn () => 1+(cupto p (fn k => 2+ k 3))));
*)
(* should yield false *)

(* Traversing a tree [Felleisen] *)
(*
 datatype tree = Null | Cell of int | Pair of tree * tree;
 datatype answer = None | Some of int;
 local
 val toplevel = new_prompt() : answer prompt
 in
 val resume = ref (fn (x:answer) => x)
 fun start f = set toplevel f
 fun suspend v = cupto toplevel (fn k => ((resume := k); v))
 end;
 fun walk Null = None
   | walk (Cell i) = suspend (Some i)
   | walk (Pair (t1,t2)) = (walk t1; walk t2);
 fun get_first t = start (fn () => walk t);
 fun get_next () = start (fn () => !resume None);

 (* for example: *)
 val tree = Pair (Pair (Cell 1,Null), Pair (Cell 2,Cell 3));
     get_next();
     get_first tree;
     get_next();
     get_next();
     get_next();
*)
(* should return: None, Some 1, Some 2, Some 3 and None. *)

(* Coroutines *)
(*
 (* first, we define simple-minded queues: *)
 exception Dequeue;
 type '_a queue = '_a list ref;
 fun mkQueue () = ref [];
 fun dequeue (q as ref (a::l)) = (q := l; a)
   | dequeue _ = raise Dequeue;
 fun enqueue (q as ref l, x) = q := l @ [x];
 
 val p = new_prompt() : unit prompt;
 val readyQ : (unit -> unit) queue = mkQueue();
 fun coroutine f = set p f;
 fun dispatch () = (dequeue readyQ ()) handle Dequeue => ();
 fun exit () = cupto p (fn k => coroutine dispatch);
 fun yield () = cupto p (fn k => (enqueue (readyQ,k);
				  coroutine dispatch));
 fun fork f = enqueue (readyQ, fn () => (f(); exit()));
*)

