/* Support functions for compiler to interpreted form, gimmlrun version.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "compile.h"
#include "type.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "cmpproto.h"
#include "memproto.h"
#include "errproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rcompaux.h"
#ifdef HAS_PROTOTYPES
#include "scompaux.h"
#endif

#undef DEBUG

EXTERN SHORT generation_count;

PUBLIC COUNTER nLocals(assum)
IN CONST ExprPtr assum;
{
  return C_NLOCALS(assum);
}

PUBLIC void setNLocals(assum,n)
IN CONST ExprPtr assum;
IN COUNTER n;
{
  C_NLOCALS(assum) = n;
}

PRIVATE struct compile_assum *allocCompileAssums=(struct compile_assum *)0L;
PRIVATE struct compile_assum *oldCompileAssums=(struct compile_assum *)0L;

PUBLIC ExprPtr MakeCompAssum(ids,
			     level,ilevel,nlocals,cons,tystate,tytags,imports,
			     fundef,dbgcnt
			     )
IN CONST ExprPtr ids;
IN COUNTER level,ilevel;
IN COUNTER nlocals;
IN CONST ExprPtr cons;
IN CONST ExprPtr tystate;
IN CONST MapPtr tytags;
IN CONST MapPtr imports;
IN CONST ExprPtr fundef;
IN COUNTER dbgcnt;
{
  struct compile_assum *new;

  new = (struct compile_assum *)MALLOC(sizeof(struct compile_assum));
  new->kind = COMPILE_ASSUM_KIND;
  new->mark = 0;
  new->flags = 0;
  /*new->gcmark = 0;*/
  new->next = allocCompileAssums;
  allocCompileAssums = new;
  C_IDS(new) = ids;
  C_LEVEL(new) = level;
  C_ILEVEL(new) = ilevel;
  C_NLOCALS(new) = nlocals;
  C_CONS(new) = cons;
  C_TYSTATE(new) = tystate;
  C_TYTAGS(new) = tytags;
  C_IMPORTS(new) = imports;
  C_FUNDEF(new) = fundef;
  C_DBGCNT(new) = dbgcnt;
  return KEEP((ExprPtr)new);
}

PUBLIC ExprPtr NewCompAssum()
{
  return MAKE_TABLE(NewMap(),NewMap(),NewMap(),(COUNTER)0,(COUNTER)0,(COUNTER)0,NewMap(),NewMap(),
		    NewMap(),NIL,NIL,0);
}

PUBLIC ExprPtr DupCompAssum(assum)
IN CONST ExprPtr assum;
{
  ExprPtr new;

  new = MAKE_TABLE(C_IDS(assum),C_STRUCTS(assum),C_SIGS(assum),
		   (COUNTER)C_LEVEL(assum),(COUNTER)C_ILEVEL(assum),
		   (COUNTER)C_NLOCALS(assum),C_CONS(assum),
		   C_TYSTATE(assum),C_TYTAGS(assum),C_IMPORTS(assum),
		   C_FUNDEF(assum),C_DBGCNT(assum));
  C_FLAGS(new) = C_FLAGS(assum);
  return new;
}

PUBLIC void FreeCompileAssums()
{
  struct compile_assum **assump,*assum;

  for (assump = &allocCompileAssums;
       (assum = *assump)!=(struct compile_assum *)0L; )
    {
      if (GC_MARKED(assum))
	{
	  GC_UNMARK(assum);
	  assump = &assum->next;
	}
      else
	{
	  *assump = assum->next;
	  MARKFREE(assum);
	}
    }
}

PUBLIC void FreeCompileAssum(assum)
IN struct compile_assum *assum;
{
  struct compile_assum **assump;

  for (assump = &allocCompileAssums; *assump!=assum; assump = &(*assump)->next);
  /* should not be too slow, since there are few compile assums */
  assum->mark = 0;
  *assump = assum->next;
  MARKFREE(assum);
}

PRIVATE struct compile_assum *CAReserve[MAX_GENERATIONS];

PUBLIC void cutoffInitialCompileAssums()
{
  REGISTER struct compile_assum *assum;

  for (assum = allocCompileAssums; assum!=(struct compile_assum *)0L; assum=assum->next)
    {
      GC_MARK(assum);
    }
  CAReserve[0] = oldCompileAssums = allocCompileAssums;
  allocCompileAssums = (struct compile_assum *)0L;
}

#ifdef GC_SLIDING
PUBLIC void slideOldCompileAssums()
{
  COUNTER i;

  for (i=1; i<generation_count; i++)
    CAReserve[i] = CAReserve[i+1];
}
#endif

PUBLIC void cutoffOldCompileAssums()
{
  REGISTER struct compile_assum **assump,*assum;

  CAReserve[generation_count] = oldCompileAssums;
  for (assump = &allocCompileAssums; (assum = *assump)!=(struct compile_assum *)0L;
       assump = &assum->next)
    GC_MARK(assum);
  *assump = oldCompileAssums;
  oldCompileAssums = allocCompileAssums;
  allocCompileAssums = (struct compile_assum *)0L;
}

PUBLIC void refreshOldCompileAssums()
{
  REGISTER struct compile_assum **assump,*assum,*last;

  last = CAReserve[generation_count];
  for (assump = &oldCompileAssums; assum = *assump, assum!=last; assump = &assum->next)
    {
      GC_UNMARK(assum);
    }
  *assump = allocCompileAssums;
  allocCompileAssums = oldCompileAssums;
  oldCompileAssums = assum;
}

PUBLIC void unmarkOldCompileAssums()
{
  REGISTER struct compile_assum *assum;

  for (assum = oldCompileAssums; assum!=(struct compile_assum *)0L; assum = assum->next)
    GC_UNMARK(assum);
}

PUBLIC MapPtr IdLocs(assum)
IN CONST ExprPtr assum;
{
  return C_IDS(assum);
}

PUBLIC MapPtr TyTags(assum)
IN CONST ExprPtr assum;
{
  return C_TYTAGS(assum);
}

PUBLIC Boolean hasChoice(pat)
IN CONST ExprPtr pat;
{
again:
  switch (pat->kind)
    {
    case E_VARINTRO:
    case E_NIL: case E_TUPLE_0: case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
    case E_QUOTE: case E_VARREF: case E_PAT_ANY:
      return false;
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto again;
    case E_PATCHOICE:
      return true;
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_PAT_REVCONS: case E_PAT_CONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_2:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_APPL: case E_PAT_AS:
    case E_PAT_RECORD_MAPADD: case E_PAT_MAPADD:
      if (hasChoice(CAR(pat)))
	return true;
      pat = CDR(pat);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1: case E_REF:
      pat = REFVAL(pat);
      goto again;
    case E_DATA:
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
      pat = CDR(pat);
      goto again;
    case E_PAT_CHECK:
      pat = PATCHECKVAL(pat);
      goto again;
    case E_PAT_PACK:
      pat = PACKVAL(pat);
      goto again;
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
      pat = COUNTEDVAL(pat);
      goto again;
    default:
      kinderror(message(665),pat->kind);
      /*NOTREACHED*/
      return false;
    }
}
