/* Compiler to interpreted form, gimmlrun version.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "gram.h"
#include "compile.h"
#include "type.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "grmproto.h"
#include "evlproto.h"
#include "cmpproto.h"
#include "errproto.h"
#include "prtproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rcompile.h"
#ifdef HAS_PROTOTYPES
#include "scompile.h"
#endif

#undef DEBUG
#undef DEBUGMOD
#undef DBGINLINE

#define EP(expr) valPrint(expr,alpha,do_printfile,(char *)hstderr,0)

EXTERN ExprPtr basisEnv;
#if DEBUGGING
EXTERN ExprPtr currentFile;
#endif
#if PROFILING
PUBLIC ExprPtr profileExprs;
#endif

#define CP(expr,assum) cp(expr,assum)

PRIVATE ExprPtr Cons(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return z;
}

PRIVATE ExprPtr CONS(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return KEEP(z);
}

PRIVATE ExprPtr wrapDebugName1(type,name,ex,assum,consp,excp)
IN CONST ExprPtr type;
IN CONST ExprPtr name;
IN CONST ExprPtr ex;
IN CONST ExprPtr assum;
IN Boolean consp,excp;
{
  ExprPtr wrap,aux;
  ENTER;

  aux = MakePair(excp?E_GOTO:(consp?E_APPL:E_LET),name,type);
  (void) KEEP(aux);
  aux = MakeCounted(E_COUNTED,aux,(INT)VARRANK(ex)
#ifdef OLDDEBUG
		    +1
#endif
		    );
  wrap = MakeDebugName(ex,KEEP(aux));
  KEEPONLY(wrap);
  return wrap;
}

	/*****  compilation *****/

PUBLIC ExprPtr getTopEnv()
{
  ExprPtr env;

  env = basisEnv;
  while (ENEXT(env)!=NIL)
    env = ENEXT(env);
  return env;
}

PUBLIC COUNTER itIndex()
{
  return ORIGINAL_IT;
}

PRIVATE ExprPtr cpError(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER gto;
{
  crash_error(ZONE(expr),(CHAR *)message(425));
  return NIL;
}

PRIVATE ExprPtr mkIL(i,l)
IN COUNTER i,l;
{
  ExprPtr i_l;
  ENTER;

  i_l = KINT(i);
  if (l!=0)
    i_l = KONS(i_l,KINT(l));
  KEEPONLY(i_l);
  return i_l;
}

PRIVATE ExprPtr cpDeclList1(args,assum)
IN CONST ExprPtr args;
INOUT ExprPtr assum;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
  REGISTER ExprPtr e,b,last;

  _x.p[0] = NIL;
  _x.n = 1;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  last = NIL;
  for (; args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      b = CP(e,assum);
      if (b!=NIL)
	{
	  if (last!=NIL)
	    {
	      SETCDR(last,b);
	    }
	  else
	    _x.p[0] = b;
	  while (last=b, b=CDR(b), b!=NIL);
#if DEBUGGING!=0
	  if (C_FLAGS(assum) & C_NODEBUG)
	    ;
	  else
	    {
	      ExprPtr info;
	      ENTER;

	      info = KONS(NIL,KONS(currentFile,NIL));
	      info = ZBuild(ZONE(e),A_CODEPOINT,info);
	      (void) KEEP(info);
	      info = Kref(E_DEBUGDECL, info);
	      (void) KEEP(info);
	      b = MakePair(E_CONS,info,NIL);
	      SETCDR(last,b);
	      last = b;
	      QUIT;
	      /* New debugging scheme, not understood by eval.c */
	    }
#endif
	}
    }
  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE ExprPtr cpDeclList(expr,assum,gto)
	/* A_DECL_LIST, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN COUNTER gto;
{
  ExprPtr ex;
  ENTER;

  ex = cpDeclList1(ARGS(expr),assum);
  KEEPONLY(ex);
  return ex;
}

PRIVATE void cpExtVal1(sym,expr,assum,excp,type)
IN CONST ExprPtr sym;
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN Boolean excp;
IN CONST ExprPtr type;
{
  ExprPtr i_l,ex,aux;
  COUNTER i,l;
  Boolean found;
  ENTER;

  (void) MapGet(sym,C_IDS(assum),&found);
  if (found) /* already known externs are resolved at compile-time */
    return;
  /*otherwise, reserve a slot for the variable, and create a binding x = ?, where ? is a dummy. */
  l = C_ILEVEL(assum);
  i = C_NLOCALS(assum)++;
  i_l = mkIL(i,l);
  setMapPut(sym,i_l,&C_IDS(assum));
  ex = Kvarintro(i);
  (void) KEEP(ex);
  ex = CONS(ex,sym); /* sym is a dummy, to be patched at link-time */
  aux = Cons(ex,C_IMPORTS(assum));
  SETQ(C_IMPORTS(assum),aux);
  QUIT;
}

PRIVATE ExprPtr resolveImports(zone,imports,assum) /* return reversed list of bindings */
IN CONST ExprPtr zone;
INOUT ExprPtr imports;
IN CONST ExprPtr assum;
{
  ExprPtr binding,i_l,pat,bindings,bind,ll;
  Boolean found,excp;
  COUNTER i,l,level;
  ENTER;

  bindings = NIL;
  level = C_LEVEL(assum)+2; /* the (local) module environment is allocated just below
			       a (fake) copy of GLOBENV, which is again just below the
			       current environment. */
  for (ll = imports; ll!=NIL; ll = CDR(ll))
    {
      binding = CAR(ll);
      i_l = MapGet(CDR(binding),C_IDS(assum),&found);
      if (!found) /* then insert an implicit 'extern x:type' declaration */
	{
	  /* we have to estimate whether is has to be treated as an exception
	     constructor by the debugger, if we have one. We can only know
	     this if the module was compiled with debug options, i.e. when
	     the pattern contains a E_DEBUGNAME cell. */
	  l = C_ILEVEL(assum);
	  i = C_NLOCALS(assum);
	  pat = CAR(binding);
	  excp = (pat->kind==(CHAR)E_DEBUGNAME && DNINFO(pat)->kind==(CHAR)E_GOTO);
	  cpExtVal1(CDR(binding),zone,assum,excp,LETBODY(COUNTEDVAL(DNINFO(pat))));
	}
      else
	{
	  if (i_l->kind==(CHAR)E_CONS)
	    {
	      i = IVAL(CAR(i_l));
	      l = IVAL(CDR(i_l));
	    }
	  else
	    {
	      i = IVAL(i_l);
	      l = 0;
	    }
	}
      bind = Kvarref(i,level-l);
      (void) KEEP(bind);
      bind = CONS(CAR(binding),bind);
      bindings = Cons(bind,bindings);
      KEEPONLY(bindings);
    }
  KEEPONLY(bindings);
  return bindings;
}

MAPSWEEPFUNC(do_export)
{
  struct export *data = MAPDATA(struct export *);
  ExprPtr var,val,i_l,binding,bindings,cons;
  COUNTER i,l,maxl;
  Boolean found,consp,excp;
  ENTER;

  if (y->kind==(CHAR)E_CONS)
    {
      i = IVAL(CAR(y));
      l = IVAL(CDR(y));
    }
  else
    {
      i = IVAL(y);
      l = 0;
    }
  maxl = C_LEVEL(data->basis->basisTable);
  if (l<=maxl) /* C_LEVEL should be 1, variables with higher levels
		  are locals (and it is probably a bug if l>1). */
    {
      val = Kvarref(i,maxl-l);
      (void) KEEP(val);
      l = C_ILEVEL(data->assum);
      i = C_NLOCALS(data->assum)++;
      i_l = mkIL(i,l);
      setMapPut(x,i_l,&C_IDS(data->assum));
      cons = MapGet(y,C_CONS(data->basis->basisTable),&found);
      if (found)
	{
	  switch (cons->kind)
	    {
	    case E_COUNTED: /* case E_POLYFUN: case E_PROFILE: */
	      /* !!! We should shiftExpr(cons,IVAL(CDR(y)) - l) here,
		 however this means shiftExpr(cons,-1):
		 all references to varref(rank, level = 0) should
		 become references to where resolveImports have relocated
		 them, all references to varref(rank, level = 1) should
		 become references to toplevel primitives, and all other
		 references should be shifted down by 1.
		 For now, I simply disallow inlining of imported functions,
		 this will be simpler (but it prevents me from inlining
		 functions that are explicitly defined in header files, sob.) */
	      break;
	    default:
	      setMapPut(i_l,cons,&C_CONS(data->assum));
	    }
	}
      var = Kvarintro(i);
#if DEBUGGING
      if (l!=0 && (C_FLAGS(data->assum) & C_NODBGNAME)==0)
	{
#ifdef OLDDEBUG
	  C_NLOCALS(data->assum)++;
#endif
	  consp = found && (cons->kind!=(CHAR)E_EXCEPTION);
	  excp = found && (cons->kind==(CHAR)E_EXCEPTION);
	  var = wrapDebugName1(MapGetSure(x,A_IDS(data->basis->basisAssum)),
			       x,KEEP(var),data->assum,consp,excp);
	}
#endif
      (void) KEEP(var);
      binding = CONS(var,val);
      bindings = Cons(binding,REFVAL(data->bindings));
      SETREFVAL(data->bindings,bindings);
    }
  QUIT;
  return false;
}

PRIVATE ExprPtr cpBasis(expr,assum,gto)
	/* A_BASIS, return bindings */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER gto;
{
  Basis *basis;
  ExprPtr bindings,def;
  struct export *data;
  ENTER;

  basis = (Basis *)EXTDATA(CAR(ARGS(expr)));
  bindings = resolveImports(expr,C_IMPORTS(basis->basisTable),assum);
#ifdef DEBUGMOD
  {
    INT n;

    n = ListLength(bindings);
    hfprintf(hstderr,"%d imports, ",n);
  }
#endif
#if TRANSLATOR
  def = MakeAppl(basis->bindings,NIL);
  (void) KEEP(def);
  def = CONS(CONS(patany,def),NIL);
#else
  def = basis->bindings;
#endif
  bindings = Reverse1(bindings,def);
  (void) KEEP(bindings);
#ifdef DEBUGMOD
  {
    INT n;

    n = ListLength(def);
    hfprintf(hstderr,"%d defined, ",n);
  }
#endif
  data = (struct export *)alloctmp(sizeof(struct export));
  data->assum = assum;
  data->basis = basis;
  data->bindings = MakeSing(E_REF,NIL);
  (void) KEEP(data->bindings);
  (void) MapSweep(MapDomRestrBy(C_IDS(basis->basisTable),A_IDEXTS(basis->basisAssum)),
		  do_export,(char *)data);
#ifdef DEBUGMOD
  {
    INT n;

    n = ListLength(REFVAL(data->bindings));
    hfprintf(hstderr,"%d exported.\n",n);
    hfflush(hstderr);
  }
#endif
  bindings = MakeLet(CONS(KINT(-1-C_NLOCALS(basis->basisTable)),bindings),
		     REFVAL(data->bindings));
  /* the number of local variables is set to a negative number, so that eval knows
     what to do (allocate a fake copy of GLOBENV before allocating the local environment) */
  (void) KEEP(bindings);
#if MODULE_SYSTEM & STD_ML_MODULE
  setMapMerge(&C_STRUCTS(assum),C_STRUCTS(basis->basisTable));
  setMapMerge(&C_SIGS(assum),C_SIGS(basis->basisTable));
#endif
  setMapMerge(&C_TYTAGS(assum),C_TYTAGS(basis->basisTable)); /* because tytags are global */
  bindings = Cons(bindings,NIL);
  KEEPONLY(bindings);
  return bindings;
}

PUBLIC ExprPtr cp(expr,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr ex;
  ENTER;

  switch (TYPE(expr))
    {
    case A_BASIS: ex = cpBasis(expr,assum,-1); break;
    case A_DECL_LIST: ex = cpDeclList(expr,assum,-1); break;
    default: ex = cpError(expr,assum,-1); break;
    }
  KEEPONLY(ex);
  return ex;
}
