/* Debugger.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Universit Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Mthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h> 
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "type.h"
#include "compile.h"
#include "gram.h"
#include "input.h"
#include "print.h"
#include "bc.h"
#include "rbci.h"
#include "dbgproto.h"
#include "grmproto.h"
#include "evlproto.h"
#include "prtproto.h"
#include "typproto.h"
#include "exeproto.h"
#include "cmpproto.h"
#include "errproto.h"
#include "stkproto.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rdebug.h"
#ifdef HAS_PROTOTYPES
#include "sdebug.h"
#endif
#ifdef HAS_READLINE
#include "linenoise.h"
#endif

#if DEBUGGING
EXTERN EXTENDED zero, fone;
EXTERN SymPtr refsym;
EXTERN extDesc iarrayExtDesc;
EXTERN ExprPtr num_alpha;

EXTERN byteCodeChunk *allocBCs;
EXTERN byteCodeChunk *oldBCs;

EXTERN Boolean stopwanted;
EXTERN Boolean readlineEnabled;
PUBLIC COUNTER debugMode = DBM_NONE;
PUBLIC COUNTER debugNframes = -1;
PUBLIC ExprPtr monitorExprs;
PUBLIC ExprPtr lastPrintExpr;
PUBLIC MapPtr lastPrintVals,lastPrintTypes;
PUBLIC struct breakpoints breakPoints;
PUBLIC ExprPtr dbgName, dbgFile;
PRIVATE COUNTER maxWhere = 10;
PRIVATE int defaultCommand = ' ';

PRIVATE OPCODE opAt (pc)
IN CONST OPCODE *pc;
{
  OPCODE op;
  COUNTER i;

  op = *pc;
  for (i=0; i<breakPoints.n; i++)
    if (breakPoints.pcs[i]==pc)
      {
	op = breakPoints.savedops[i];
	break;
      }
  return op;
}

#define MDBG_DEFCON MAX_TWO_TO_THE_N_INT
#define MDBG_DEFEXC (MDBG_DEFCON >> 1)
#define MDBG_OFF_MASK (~(MDBG_DEFCON | MDBG_DEFEXC))

PRIVATE ExprPtr getValue1(id,snap,typep,info)
IN CONST ExprPtr id;
IN struct snapshot *snap;
OUT ExprPtr *typep;
IN ExprPtr info;
{
  ExprPtr val, type;
  INT n;
  OPCODE dop, *dpc;

  val = NOEXPR;
  if (typep!=NULL)
    {
      type = COUNTEDVAL(info);
      if (type->kind==(CHAR)E_BIAS)
	type = CAR(type);
      *typep = type;
    }
  n = COUNTEDN(info) & MDBG_OFF_MASK;
  dpc = &snap->bc->debug[n];
  switch (dop = *dpc++)
    {
    case DBG_ACCESS:
      {
	COUNTER index;
	
	index = (COUNTER)*dpc++;
	if (index<snap->stackDepth)
	  val = snap->top[index - snap->stackDepth];
      }
    break;
    case DBG_VARREF:
      {
	COUNTER index, l, j;
	ExprPtr env;
	
	index = (COUNTER)*dpc++;
	if (index>=snap->stackDepth)
	  break; /* Bug: incoherent stack information. */
	env = snap->top[index - snap->stackDepth];
	l = (COUNTER)*dpc++;
	j = (COUNTER)*dpc++;
	while (l!=0)
	  {
	    if (env==NIL) /* Bug: incoherent environment information. */
	      break;
	    env = ENEXT(env);
	    l--;
	  }
	if (env==NIL) /* Bug: incoherent environment information. */
	  break;
	if (j<0 || j>=EVARS(env)->n)
	  break; /* Bug: incoherent environment information. */
	val = EVARS(env)->expr[j];
      }
    break;
    }
  return val;
}

PRIVATE ExprPtr getValue(id,snap,typep)
IN CONST ExprPtr id;
IN struct snapshot *snap;
OUT ExprPtr *typep;
{
  ExprPtr info, val, type;
  Boolean found;

  info = NGet(id,AVAL(snap->ids)->expr[0],&found);
  if (found)
    return getValue1 (id, snap, typep, info);
  info = MapGet (id, snap->parent_dbgnames, &found);
  if (found)
    {
      ExprPtr env, env0, nextenv, i_l;
      COUNTER n, i, level;

      env = env0 = snap->top[-snap->stackDepth];
      for (n=0; env->kind==(CHAR)E_ENV; env=nextenv)
	{
	  nextenv = ENEXT(env);
	  if (EVARS(env)==EVARS(GLOBENV) && nextenv->kind==(CHAR)E_ENV)
	    /* this is a fake environment: do not count it. */
	    continue;
	  n++;
	}
      if (n!=0)
	{
	  i_l = LETBIND(info);
	  level = 0;
	  if (i_l->kind==(CHAR)E_CONS)
	    {
	      level = IVAL(CDR(i_l));
	      i_l = CAR(i_l);
	    }
	  i = IVAL(i_l);
	  if (level<n)
	    {
	      n--;
	      for (env = env0; level<n; env=nextenv)
		{
		  nextenv = ENEXT(env);
		  if (EVARS(env)==EVARS(GLOBENV) && nextenv->kind==(CHAR)E_ENV)
		    /* this is a fake environment: skip it. */
		    continue;
		  level++;
		}
	      if (i>=0 && i<EVARS(env)->n)
		{
		  if (typep!=NULL)
		    {
		      type = LETBODY(info);
		      if (type->kind==(CHAR)E_BIAS)
			type = CAR(type);
		      *typep = type;
		    }
		  return EVARS(env)->expr[i];
		}
	    }
	}
    }
  type = MapGet (id, IdTypes(basisAssum), &found);
  if (found)
    {
      if (typep!=NULL)
	*typep = type;
      val = getValueByName (id, &found, basisEnv);
      if (found)
	return val;
    }
  return NOEXPR;
}

PRIVATE void initSnap(snap, bc, top, idp)
OUT struct snapshot *snap;
IN CONST byteCodeChunk *bc;
ExprPtr *top;
IN Boolean idp;
{
  snap->bc = bc;
  snap->pc = 0;
  snap->top = top;
  snap->stackDepth = 0;
  snap->mapStackDepth = 0;
  snap->file = NOEXPR;
  snap->line = -1;
  snap->pos = 0;
  snap->name = NOEXPR;
  {
    ExprPtr prof;
 
    prof = bc->consts[0];
    if (prof->kind!=(CHAR)BUILD_HEADER_KIND ||
	(TYPE(prof)!=EA_PROFILE && TYPE(prof)!=EA_CLR_PROFILE))
      ;
    else
      {
	prof = ARGS(prof);
	snap->name = CAR(prof);
	prof = CDR(prof);
	snap->file = CDR(prof);
      }
  }
  if (idp)
    {
      /* snap->ids is a 3 element array
	 [Nmap-of-visible-variables, list-of-arguments, list-of-locals].
	 */
      snap->ids = MakeArray(3, NIL);
      (void) KEEP(snap->ids);
      snap->parent_dbgnames = NewMap ();
    }
}

PRIVATE Boolean analyzeBc(snap, bc, pc, top, idp, nargsp)
OUT struct snapshot *snap;
IN CONST byteCodeChunk *bc;
OPCODE *pc;
ExprPtr *top;
IN Boolean idp; /* True to also get back names of variables, with their
		   types and values in snap->ids; false to just get stack
		   levels, and pc. */
OUT COUNTER *nargsp; /* number of arguments from caller, if at call site,
			-1 otherwise. */
{
  /* Imitates analyzeVp(). */
  OPCODE *dpc;
  INT pcoff, pc1off;
  OPCODE *dval;
  OPCODE dop;

  pc--;
  pcoff = pc-bc->instr;
  dpc = bc->debug;
  dval = NULL;
  initSnap (snap, bc, top, idp);
  *nargsp = -1;
  while (true)
    {
      switch (dop = *dpc++)
	{
	case DBG_BEGIN:
	  pc1off = (INT)*dpc++;
	  goto anlz_begin;
	case DBG_BEGIN_BIG:
	  pc1off = (((INT)(dpc[0])) << (8*sizeof(OPCODE))) |
	    (((INT)(dpc[1])) & ((1 << (8*sizeof(OPCODE))) - 1));
	  dpc += 2;
	anlz_begin:
	  if (pcoff<pc1off) /* Bug: we didn't find the codepoint. */
	    return false;
	  {
	    OPCODE *dpcend;
	    INT pc2off;

	    dpcend = dpc + (*dpc); /* points to DBG_END or DBG_END_BIG */
	    dpc++;
	    switch (*dpcend++)
	      {
	      case DBG_END:
		pc2off = (INT)*dpcend++;
		goto anlz_begin_end;
	      case DBG_END_BIG:
		pc2off = (((INT)(dpcend[0])) << (8*sizeof(OPCODE))) |
		  (((INT)(dpcend[1])) & ((1 << (8*sizeof(OPCODE))) - 1));
		dpcend += 2;
	      anlz_begin_end:
		if (pcoff>=pc2off) /* Not in this begin...end block. */
		  dpc = dpcend;
		break;
	      default: /* Bug: incorrect debug information. */
		return false;
	      }
	    break;
	  }
	case DBG_END:
	case DBG_END_BIG: /* Bug: codepoint not found. */
	  return false;
	case DBG_WHERE:
	  pc1off = (INT)*dpc++;
	  goto anlz_where;
	case DBG_WHERE_BIG:
	  pc1off = (((INT)(dpc[0])) << (8*sizeof(OPCODE))) |
	    (((INT)(dpc[1])) & ((1 << (8*sizeof(OPCODE))) - 1));
	  dpc += 2;
	anlz_where:
	  if (idp && opAt (&bc->instr[pc1off])==BC_FUNCTION)
	    { /* dump all locals into args list. */
	      ExprPtr l, prev;

	      for (prev = NIL, l = AVAL(snap->ids)->expr[1]; l!=NIL; prev = l, l = CDR(l));
	      if (prev==NIL)
		AVAL(snap->ids)->expr[1] = AVAL(snap->ids)->expr[2];
	      else
		{
		  SETCDR (prev, AVAL(snap->ids)->expr[2]);
		}
	      AVAL(snap->ids)->expr[2] = NIL;
	      HITBARRIER(snap->ids);
	    }
	  if (pcoff<=pc1off) /* normal case */
	    {
	    foundit:
	      snap->pc = pc1off;
	      snap->stackDepth = *dpc++;
	      snap->mapStackDepth = *dpc++;
	      /* obsolete: snap->excStackDepth = *dpc;*/
	      if (pcoff==pc1off) /* We have found it */
		{
		  if (bc->instr[pcoff]==BC_CALL)
		    *nargsp = 1;
		  return true;
		}
	      return false;
	    }
	  if (pcoff==pc1off+1) /* case of BC_N_CALL <n> and BC_CALLBIND <i> */
	    switch (bc->instr[pc1off])
	      {
	      case BC_N_CALL:
		*nargsp = (COUNTER)bc->instr[pcoff];
		pcoff = pc1off;
		goto foundit;
	      case BC_CALLBIND:
		*nargsp = 1;
		pcoff = pc1off;
		goto foundit;
		/* otherwise we have not found it. */
	      }
	  dpc += 3;
	  break;
	case DBG_ACCESS:
	  if (idp)
	    {
	      dval = dpc-1;
	    }
	  dpc++;
	  break;
	case DBG_VARREF:
	  if (idp)
	    {
	      dval = dpc-1;
	    }
	  dpc += 3;
	  break;
	case DBG_DEFNAME:
	case DBG_DEFCON:
	case DBG_DEFEXC:
	  if (idp)
	    {
	      CHAR *s;
	      ExprPtr name, type, info;
	      INT n;
	      ENTER;

	      s = &bc->rawData[*dpc++];
	      type = bc->consts[*dpc++];
	      if (dval==NULL) /* Bug: no value defined, or incoherent value. */
		{
		  dpc += 2;
		  break;
		}
	      name = Kstr(s);
	      (void) KEEP(name);
	      n = dval - bc->debug;
	      switch (dop)
		{
		case DBG_DEFEXC: n |= MDBG_DEFEXC; /*fallthrough*/
		case DBG_DEFCON: n |= MDBG_DEFCON; break;
		}
	      info = MakeCounted(E_COUNTED, type, n);
	      (void) KEEP(info);
	      setNPut(name,info,&AVAL(snap->ids)->expr[0],snap->ids);
	      HITBARRIER(snap->ids);
	      /* now add to list of locals: */
	      info = MakePair (E_CONS, name, AVAL(snap->ids)->expr[2]);
	      AVAL(snap->ids)->expr[2] = info;
	      HITBARRIER(snap->ids);
	      QUIT;
	    }
	  else dpc += 2;
	  break;
	case DBG_FILE:
	  snap->file = bc->consts[*dpc++];
	  break;
	case DBG_LINE:
	  snap->line = (COUNTER)(unsigned short)*dpc++;
	  snap->pos = (COUNTER)(unsigned short)*dpc++;
	  break;
	case DBG_NAME:
	  snap->name = bc->consts[*dpc++];
	  break;
	case DBG_ENV:
	  snap->parent_dbgnames = bc->consts[*dpc++];
	  break;
	case DBG_STOP:
	default:
	  return false;
	}
    }
}

PRIVATE Boolean analyzeBcs(snap, bc, pc, top, bcframes, stacks, up)
OUT struct snapshot *snap;
IN CONST byteCodeChunk *bc;
OPCODE *pc;
ExprPtr *top;
IN CONST struct bcframes *bcframes;
IN CONST struct bcstacks *stacks;
IN COUNTER up;
{
  COUNTER nargs;
  Boolean first;
  struct bcframe *frame;

  first = true;
  frame = bcframes->curBcFrame;
  while (true)
    {
      if (!analyzeBc(snap, bc, pc, top, up==0, &nargs))
	goto error;
      if (!first && nargs>=0)
	{
	  snap->stackDepth -= nargs+1;
	  /* adjust for arguments passed, which we
	   already accounted for in callee. */
	}
      first = false;
      if (up==0)
	break;
      --up;
      if (frame==bcframes->bcframes)
	{
	  bcframes = bcframes->parentBcFrames;
	  if (bcframes==NULL)
	    {
	    error:
	      initSnap (snap, bc, top, true);
	      return false;
	    }
	  frame = bcframes->curBcFrame;
	}
      else
	--frame;
      top -= snap->stackDepth;
      if (top<=stacks->_x.p || top>stacks->_x.p+stacks->_x.n)
	{
	  stacks = stacks->parentStack;
	  if (stacks==NULL)
	    return false;
	  top = &stacks->_x.p[stacks->_x.n];
	  first = true; /* do as though there had been no call */
	}
      bc = frame->bc;
      pc = frame->pc;
    }
  return true;
}

PRIVATE COUNTER nBcs(bcframes)
IN CONST struct bcframes *bcframes;
{
  COUNTER up;
  struct bcframe *frame;

  up = 0;
  frame = bcframes->curBcFrame;
  while (true)
    {
      up++;
      if (frame==bcframes->bcframes)
	{
	  bcframes = bcframes->parentBcFrames;
	  if (bcframes==NULL)
	    break;
	  frame = bcframes->curBcFrame;
	}
      else
	--frame;
    }
  return up;
}

PRIVATE void showVal1(r,lmargin)
IN CONST ExprPtr r;
IN COUNTER lmargin;
{
  COUNTER save_maxlines;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = IVAL(REFVAL(maxLinesRef));
  if (crtab_maxlines>0)
    {
      (void) dynPretty(r,do_printfile,(char *)hstderr,
		       lmargin,IVAL(REFVAL(rightMarginRef)));
      endPretty(do_printfile,(char *)hstderr);
      hfputc('\n',hstderr);
    }
  crtab_maxlines = save_maxlines;
}

PRIVATE void showValNoType1(r,lmargin)
IN CONST ExprPtr r;
IN COUNTER lmargin;
{
  COUNTER save_maxlines;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = IVAL(REFVAL(maxLinesRef));
  if (crtab_maxlines>0)
    {
      (void) valPretty(r,alpha,do_printfile,(char *)hstderr,
		       0,lmargin,IVAL(REFVAL(rightMarginRef)));
      endPretty(do_printfile,(char *)hstderr);
      hfputc('\n',hstderr);
    }
  crtab_maxlines = save_maxlines;
}

PRIVATE void showVal(val,type,lmargin)
IN CONST ExprPtr val,type;
IN COUNTER lmargin;
{
  if (val==NOEXPR)
    {
      hfputs(message(235),hstderr);
      hfputc('\n',hstderr);
    }
  else if (type==NOEXPR)
    showValNoType1(val,lmargin);
  else
    {
      ENTER;

      val = MakePair(E_PACK,val,type);
      (void) KEEP(val);
      showVal1(val,lmargin);
      QUIT;
    }
}

PRIVATE void showSource (filename, zone, f)
IN CONST ExprPtr filename;
IN CONST TZone *zone;
IN hFILE *f;
{
  hFILE *file;
  char *name;
  INT len;
  COUNTER line;
  COUNTER line1,line2;
  COUNTER pos;
  COUNTER pos1, pos2;
  int c;

  if (filename!=NULL && filename!=NIL)
    {
      len = symNameLen(filename);
      name = (char *)alloctmp(len+1);
      safememcpy(name,symName(filename),len);
      name[len] = 0;
      printZone((CHAR *)name,zone,hstderr,true);
      hfputc('\n',hstderr);
      file = hfopen (name, "r");
      if (file==NULL)
	{
	  bmsg(message(74),name);
	}
      else
	{
	  line1 = zone->line1;
	  pos1 = zone->pos1;
	  line2 = zone->line2;
	  pos2 = zone->pos2;
	  line = pos = 0;
	  if (line<line1)
	    {
	      for (;;)
		{
		  c = hgetc (file);
		  switch (c)
		    {
		    case -1: goto end;
		    case '\n':
		      if (++line>=line1)
			goto cont;
		      break;
		    default: break;
		    }
		}
	    cont:
#if OS==UNIX_SYS_V || OS==UNIX_BSD
	      hfprintf (f, "%.5d ", line+1);
#else
	      hfprintf (f, "%.5d ", line);
#endif
	      for (;;)
		{
		  c = hgetc (file);
		  /* Special characters:
		     0xa0:
		     0xc0:
		     0xe0:
		  */
#define START_CHAR 0xab /* left guillemot << */
#define END_CHAR 0xbb /* right guillemot >> */
		  if (line==line1 && pos==pos1)
		    hfputs (">>", f);
		  else if (line==line2 && pos==pos2)
		    hfputs ("<<", f);
		  switch (c)
		    {
		    case -1: goto end;
		    case '\n':
		      pos = 0;
		      hfputc (c, f);
		      if (++line>line2)
			goto end;
#if OS==UNIX_SYS_V || OS==UNIX_BSD
		      hfprintf (f, "%.5d ", line+1);
#else
		      hfprintf (f, "%.5d ", line);
#endif
		      break;
		    default:
		      pos++;
		      hfputc (c, f);
		      break;
		    }
		}
	    }
	end:
	  (void) hfclose(file);
	}
    }
}

PRIVATE ExprPtr getFileName(fun,zonep)
IN CONST ExprPtr fun;
OUT TZone *zonep;
{
  byteCodeChunk *bc;
  ExprPtr prof, args, name;

  if (fun->kind!=(CHAR)E_INTERPRETED_CLOSURE)
    return NOEXPR;
  fun = QENV (fun);
  if (fun->kind!=(CHAR)E_TUPLE_2)
    return NOEXPR;
  bc = (byteCodeChunk *)CAR(fun);
  if (bc->kind!=(CHAR)E_BYTECODE)
    return NOEXPR;
  if (bc->nConsts<1)
    return NOEXPR;
  prof = bc->consts[0];
  if (prof->kind!=(CHAR)BUILD_HEADER_KIND ||
      (TYPE(prof)!=EA_PROFILE && TYPE(prof)!=EA_CLR_PROFILE))
    return NOEXPR;
  if (zonep!=NULL)
    *zonep = *ZONE(prof);
  args = ARGS(prof);
  if (args->kind!=(CHAR)E_CONS)
    return NOEXPR;
  args = CDR(args);
  if (args->kind!=(CHAR)E_CONS)
    return NOEXPR;
  name = CDR(args);
  if (name->kind!=(CHAR)E_STRING)
    return NOEXPR;
  return name;
}

PRINTFUNC(do_prtmax)
{
  struct prtstrmax *data;
  COUNTER len2,i;
  char *s;
  Boolean overflow;

  data = PRINTDATA(struct prtstrmax *);
  if (data->len<data->maxlen)
    {
      len2 = data->maxlen-data->len;
      if ((overflow=(len>len2)))
	len = len2;
      for (i=0, s=(char *)text; i<len; i++, s++)
	hfputc(*s,hstderr);
      data->len += len;
    }
}

PRIVATE void showArgs(snap,lmargin,allvars,istop)
IN struct snapshot *snap;
IN COUNTER lmargin;
IN Boolean allvars,istop;
{
  char *s;
#ifdef SHOW_NAMED_ARGS
  ExprPtr args;
#else
  ExprPtr *argp;
  COUNTER i, nargs;
  INT n;
  ExprPtr params, prms, prevprms, id, info;
  OPCODE *dpc;
  Boolean found;
#endif
  ExprPtr val,type;
  Boolean commap;
  struct showargs *data = (struct showargs *)alloctmp(sizeof(struct showargs));
  ENTER;

  if (snap->name!=NOEXPR)
    lmargin += printSym(snap->name,do_printfile,(char *)hstderr);
  else
    {
      s = istop?message(743):message(236);
      lmargin += strlen(s);
      hfputs(s,hstderr);
    }
  lmargin += printSpace(do_printfile,(char *)hstderr);
  lmargin += printOpenParen(do_printfile,(char *)hstderr);
  data->psmax.maxlen = IVAL(REFVAL(rightMarginRef))-lmargin-7;
  commap = false;

#ifdef SHOW_NAMED_ARGS
  args = Nreverse(AVAL(snap->ids)->expr[1]);
  (void) KEEP (args);
  for (; args!=NIL; args=CDR(args))
    {
      if (commap)
	{
	  COUNTER j;

	  hfputs(",\n",hstderr);
	  for (j=0; j<lmargin; j++)
	    hfputc(' ',hstderr);
	}
      commap = true;
      (void) printSym(CAR(args),do_printfile,(char *)hstderr);
      hfputs(message(239),hstderr);
      val = getValue(CAR(args),snap,&type);
      if (val==NOEXPR)
	hfputs(message(238),hstderr);
      else
	{
	  data->psmax.len = 0;
	  valPrint(val,type,do_prtmax,(char *)&data->psmax,0);
	  if (data->psmax.len>=data->psmax.maxlen)
	    hfputs(message(505),hstderr);
	}
    }
  (void) printCloseParen(do_printfile,(char *)hstderr);
  hfputc('\n',hstderr);
  if (allvars)
    {
      args = Nreverse(AVAL(snap->ids)->expr[2]);
      (void) KEEP (args);
      data->psmax.maxlen = IVAL(REFVAL(rightMarginRef))-5;
      data->snap = snap;
      for (; args!=NIL; args=CDR(args))
	{
	  val = getValue (CAR(args), data->snap, &type);
	  lmargin = printBuf (message(735), do_printfile, (char *)hstderr);
	  lmargin += printSym(CAR(args), do_printfile, (char *)hstderr);
	  lmargin += printBuf (message(239), do_printfile, (char *)hstderr);
	  if (val==NOEXPR)
	    hfputs(message(238),hstderr);
	  else
	    {
	      data->psmax.len = lmargin;
	      valPrint(val,type,do_prtmax,(char *)&data->psmax,0);
	      if (data->psmax.len>=data->psmax.maxlen)
		hfputs(message(505),hstderr);
	      hfputs (";\n", hstderr);
	    }
	}
    }
#else
  argp = snap->top - snap->stackDepth+1;
  nargs = snap->bc->nargs;
  params = Nreverse(AVAL(snap->ids)->expr[1]);
  (void) KEEP (params);
  for (i=0; i<nargs; i++)
    {
      if (commap)
	{
	  COUNTER j;
	  
	  hfputs(",\n",hstderr);
	  for (j=0; j<lmargin; j++)
	    hfputc(' ',hstderr);
	}
      commap = true;
      val = *argp++;
      if (val==NOEXPR)
	hfputs(message(238),hstderr);
      else
	{
	  data->psmax.len = 0;
	  /* try to find actual argument among elements of params list. */
	  found = false;
	  for (prms=params, prevprms=NIL; prms!=NIL;
	       prevprms=prms, prms=CDR(prms))
	    {
	      id = CAR(prms);
	      info = NGet(id,AVAL(snap->ids)->expr[0],&found);
	      if (!found)
		continue;
	      type = COUNTEDVAL(info);
	      if (type->kind==(CHAR)E_BIAS)
		type = CAR(type);
	      n = COUNTEDN(info) & MDBG_OFF_MASK;
	      dpc = &snap->bc->debug[n];
	      if (dpc[0]==DBG_ACCESS &&
		  argp-1 == &snap->top[((COUNTER)dpc[1]) - snap->stackDepth])
		{ /* found, erase entry from params. */
		  if (prevprms==NIL)
		    params = CDR(params);
		  else
		    SETCDR(prevprms,CDR(prms));
		  break;
		}
	      found = false;
	    }
	  if (found)
	    {
	      (void) printSym(CAR(prms),do_printfile,(char *)hstderr);
	      hfputs(message(239),hstderr);
	      valPrint(val,type,do_prtmax,(char *)&data->psmax,0);
	    }
	  else
	    valPrintNoType(val,do_prtmax,(char *)&data->psmax,0);
	  if (data->psmax.len>=data->psmax.maxlen)
	    hfputs(message(505),hstderr);
	}
    }
  (void) printCloseParen(do_printfile,(char *)hstderr);
  hfputc('\n',hstderr);
  if (allvars)
    {
      for (prms=params, prevprms=NIL; prms!=NIL;
	   prevprms=prms, prms=CDR(prms));
      prms = Nreverse(AVAL(snap->ids)->expr[2]);
      (void) KEEP (prms);
      if (prevprms!=NIL)
	{
	  SETCDR(prevprms,prms);
	  prms = params;
	}
      data->psmax.maxlen = IVAL(REFVAL(rightMarginRef))-5;
      data->snap = snap;
      for (; prms!=NIL; prms=CDR(prms))
	{
	  val = getValue (CAR(prms), data->snap, &type);
	  lmargin = printBuf (message(735), do_printfile, (char *)hstderr);
	  lmargin += printSym(CAR(prms), do_printfile, (char *)hstderr);
	  lmargin += printBuf (message(239), do_printfile, (char *)hstderr);
	  if (val==NOEXPR)
	    hfputs(message(238),hstderr);
	  else
	    {
	      data->psmax.len = lmargin;
	      valPrint(val,type,do_prtmax,(char *)&data->psmax,0);
	      if (data->psmax.len>=data->psmax.maxlen)
		hfputs(message(505),hstderr);
	      hfputs (";\n", hstderr);
	    }
	}
    }
#endif

  if (snap->file!=NOEXPR && snap->file!=NIL)
    {
      hfputs (message (744), hstderr);
      (void) printSym (snap->file, do_printfile, (char *)hstderr);
      hfprintf (hstderr, ":%ld\n",
#if OS==UNIX_SYS_V || OS==UNIX_BSD
		snap->line+1
#else
		snap->line
#endif
		);
    }
  QUIT;
}

PRIVATE void showStack(bc, pc, top, bcframes, stacks, up, mw, maxup)
IN CONST byteCodeChunk *bc;
IN CONST OPCODE *pc;
IN CONST ExprPtr *top;
IN CONST struct bcframes *bcframes;
IN CONST struct bcstacks *stacks;
IN COUNTER up, maxup;
IN COUNTER mw;
{
  COUNTER i,start,end;
  Boolean gotBcs;
  struct snapshot snap;
#ifdef SHOW_CALL_AND_RETURN
  OPCODE op;
  COUNTER lmargin;
  COUNTER msg;
  COUNTER n;
  ExprPtr *argp;
#endif

  start = 0;
  if (up>mw/3)
    start = up - mw/3;
  end = start+mw;
  for (i=end; --i>=start; )
    {
      gotBcs = analyzeBcs (&snap, bc, pc, top, bcframes, stacks, i);
      if (gotBcs)
	{
	  hfputc((i==up)?'*':' ',hstderr);
	  hfprintf (hstderr, "%4d| ", i);
	  showArgs(&snap,7,i==up,i==maxup);
#ifdef SHOW_CALL_AND_RETURN
	  op = opAt (snap.bc->instr + snap.pc);
	  switch (op)
	    {
	    case BC_RETURN:
	      msg = 745;
	      n = 1;
	      argp = &snap.top[-1];
	      goto printval;
	    case BC_CALL:
	      msg = 746;
	      n = 1;
	      argp = &snap.top[1];
	      goto printval;
	    case BC_N_CALL:
	      msg = 946;
	      n = opAt (snap.bc->instr + snap.pc + 1);
	      argp = &snap.top[1];
	    printval:
	      while (n!=0)
		{
		  struct showargs *data = (struct showargs *)alloctmp(sizeof(struct showargs));

		  lmargin = printBuf (message(msg), do_printfile, (char *)hstderr);
		  data->psmax.len = lmargin;
		  data->psmax.maxlen = IVAL(REFVAL(rightMarginRef))-4;
		  valPrintNoType (*argp++, do_prtmax, (char *)&data->psmax, 0);
		  if (data->psmax.len>=data->psmax.maxlen)
		    hfputs(message(505),hstderr);
		  hfputs ("\n", hstderr);
		  n--;
		  msg = 947;
		}
	      break;
	    }
#endif
	}
    }
}

#ifdef HAS_READLINE
#define DEBUG_MAX_LEN 4096
PRIVATE char debug_buffer[DEBUG_MAX_LEN];
PRIVATE COUNTER debug_buffer_pos = 0;

IMPORT void readline_completion(const char *buf,linenoiseCompletions *lc,size_t pos); /* in lex.l */

PRIVATE char *debug_hints(buf,color,bold)
const char *buf;
int *color;
int *bold;
{
  char c, c2 = 0, *s;

  /* check that buf is of the form {SPACE}* c {SPACE}*, for a unique
     character c */
  for (s=(char *)buf; c=*s++, isspace(c); );
  *color = 36;
  *bold = 0;
  if (c=='\0') /* empty line */
    return "h(elp)";
  if (c=='-')
    if (c2 = *s, !isspace(c2))
      s++;
  while (isspace(*s)) s++;
  if (*s!='\0')
    return NULL;
  switch (c)
    {
    case 'p':
    case 'm':
    m:
      return " value1, value2, ...";
    case '$':
      if (isspace(s[-1]))
	return NULL;
      return "var=value";
    case 'x':
    x:
      return " exception1, exception2, ...";
    case 'b':
    b:
      return " function[@file][:line]";
    case 'l':
      return " function";
    case '-':
      switch (c2)
	{
	case 'm': goto m;
	case 'b': goto b;
	case 'x': goto x;
	case '$':
	  return "var";
	default:
	  return " m,x,b,$";
	}
    default:
      return NULL;
    }
}

PRIVATE int debug_readline()
{
  int result;
  
  linenoiseSetCompletionCallback(readline_completion);
  linenoiseSetHintsCallback(debug_hints);
  linenoiseSetFreeHintsCallback(NULL);
  linenoiseMaskModeDisable();
  linenoiseSetMultiLine(0);
  result = linenoise_raw((char *)message(90),debug_buffer,DEBUG_MAX_LEN);
  if (result)
    {
      debug_buffer[result-1] = '\0';
      linenoiseHistoryAdd(debug_buffer);
      debug_buffer[result-1] = '\n';
    }
  return result;
}

PRIVATE int debug_next_char()
{
  char c;
  int result;

  c = debug_buffer[debug_buffer_pos++];
  if (c=='\0')
      {
	result = debug_readline();
	debug_buffer_pos = 0;
	if (result==0)
	  return EOF;
	c = debug_buffer[debug_buffer_pos++];
      }
  return (int)c;
}
#define DEBUG_NEXT_CHAR (readlineEnabled?debug_next_char():hgetc(hstdin))
#else
#define DEBUG_NEXT_CHAR hgetc(hstdin)
#endif

PRIVATE int dgetc()
{
  int c;

  c = DEBUG_NEXT_CHAR;
#if CORE_TRACE
  if (coreTraceFile!=NULL && c!=-1)
    {
      hfputc(c,coreTraceFile);
      if (c=='\n')
	{
	  hfflush(coreTraceFile);
	}
    }
#endif
  return c;
}

PRIVATE int skipLine()
{
  int c;

  while (c = dgetc(), c!=-1 && c!='\n');
  return c;
}

PRIVATE int skipLinec(c)
int c;
{
  if (c!='\n' && c!=-1)
    return skipLine();
  return c;
}

PRIVATE int skipBlanks(c)
int c;
{
  while (c==' ' || c=='\t' || c=='\f')
    c = dgetc();
  return c;
}

PRIVATE Boolean is_letter(c)
IN int c;
{
  return (c>='a' && c<='z') || (c>='A' && c<='Z');
}

PRIVATE Boolean is_alpha(c)
IN int c;
{
  return is_letter (c) || (c>='0' && c<='9') || c=='\'' || c=='_';
}

PRIVATE Boolean is_symb(c)
IN int c;
{
  return c=='!' || c=='%' || c=='&' || c=='$' || c=='#' || c=='+'
  || c=='-' || c=='/' || c==':' || c=='<' || c=='=' || c=='>'
  || c=='?' || c=='@' || c=='\\' || c=='~' || c=='\'' || c=='^'
  || c=='|' || c=='*' || c=='`';
}

PRIVATE Boolean is_filechar(c)
IN int c;
{
  return c!=-1 && c!='\n' && c!=' ' && c!='\t' && c!='\f';
}

PRIVATE int skipWord()
{
  int c;

  while (c = dgetc(), is_letter(c) || c=='_');
  return c;
}

PRIVATE int getInt(c,np,gotp)
IN int c;
OUT COUNTER *np;
OUT Boolean *gotp;
{
  *gotp = false;
  *np = 0;
  while (c>='0' && c<='9')
    {
      *np = 10*(*np)+(c-'0');
      c = dgetc();
      *gotp = true;
    }
  return c;
}

PRIVATE int getSignedInt(c,np,gotp)
IN int c;
OUT COUNTER *np;
OUT Boolean *gotp;
{
  Boolean neg;

  *gotp = false;
  if ((neg = (c=='~')))
    {
      c = dgetc ();
    }
  c = getInt (c, np, gotp);
  if (neg)
    *np = -*np;
  return c;
}

PRIVATE int getRealOrInt (c, xp, gotp)
IN int c;
OUT ExprPtr *xp;
OUT Boolean *gotp;
{
  Boolean neg,gotmant,isint;
  REAL x;

  *gotp = false;
  if ((neg = (c=='~')))
    c = dgetc ();
  x = zero;
  gotmant = false;
  isint = true;
  while (c>='0' && c<='9')
    {
      x = 10.0*x+(REAL)(c-'0');
      c = dgetc ();
      gotmant = true;
    }
  if (gotmant)
    {
      REAL base;

      if (c=='.')
	{
	  isint = false;
	  c = dgetc ();
	  base = fone;
	  while (c>='0' && c<='9')
	    {
	      base *= 0.1;
	      x += base * (REAL) (c-'0');
	      c = dgetc ();
	    }
	}
      if (c=='E')
	{
	  COUNTER n;
	  Boolean got;

	  isint = false;
	  c = dgetc ();
	  c = getSignedInt (c, &n, &got);
	  if (got)
	    x = ldexp (x, n);
	  else gotmant = false;
	}
    }
  if (neg)
    x = -x;
  if ((*gotp = gotmant))
    {
      if (isint)
	*xp = Kint ((INT)x);
      else
	*xp = Kreal (&x);
    }
  return c;
}

PRIVATE int getComplexOrInt (c, xp, gotp)
IN int c;
OUT ExprPtr *xp;
OUT Boolean *gotp;
{
  ExprPtr re, im;
  REAL x;
  ENTER;

  c = getRealOrInt (c, &re, gotp);
  if (*gotp)
    {
      c = skipBlanks (c);
      if (c==':')
	{
	  (void) KEEP(re);
	  c = dgetc ();
	  c = skipBlanks (c);
	  c = getRealOrInt (c, &im, gotp);
	  if (*gotp)
	    {
	      (void) KEEP(im);
	      if (re->kind==(CHAR)E_INT)
		{
		  x = (REAL)IVAL(re);
		  re = Kreal (&x);
		  (void) KEEP (re);
		}
	      if (im->kind==(CHAR)E_INT)
		{
		  x = (REAL)IVAL(im);
		  im = Kreal (&x);
		  (void) KEEP (im);
		}
	      x = XVAL(im);
	      if (!ZEROR(x))
		re = Kcomplex (re, im);
	      KEEPONLY (re);
	    }
	}
      *xp = re;
    }
  return c;
}

PRIVATE int getStringCst (c, vp, gotp)
IN int c;
OUT ExprPtr *vp;
OUT Boolean *gotp;
{
  CHAR *buf, *newbuf;
  COUNTER size, i, newsize;
  int cc;

  size = 128;
  buf = (CHAR *)MALLOC(size);
  i = 0;
  c = dgetc ();
  while (true)
    {
      switch (c)
	{
	case '"': /* end of string */
	  *gotp = true;
	  *vp = Kstring(i, buf);
	  MARKFREE(buf);
	  return dgetc ();
	case '\n':
	case -1: /* error */
	error:
	  MARKFREE(buf);
	  return c;
	case '\'': /* escape */
	  c = dgetc ();
	  switch (c)
	    {
	    case '"':
	    case '\\': goto put_c;
	    case 'n': c = '\n'; goto put_c;
	    case 't': c = '\t'; goto put_c;
	    case 'v': c = '\v'; goto put_c;
	    case 'f': c = '\f'; goto put_c;
	    case '^':
	      c = dgetc ();
	      if (c>='@' && c<='_')
		{
		  c -= '@';
		  goto put_c;
		}
	      goto error;
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	      {
		int i;

		i = c-'0';
		c = dgetc ();
		if (c>='0' && c<='9')
		  {
		    i = 10*i+(c-'0');
		    c = dgetc ();
		    if (c>='0' && c<='9')
		      {
			i = 10*i+(c-'0');
			c = i;
			goto put_c;
		      }
		  }
		goto error;
	      }
	    case ' ': case '\t': case '\n': case '\f':
	      while (c = dgetc (), (c==' ' || c=='\t' || c=='\n' || c=='\f'));
	      if (c!='\\')
		goto error;
	      c = dgetc ();
	      break;
	    default:
	      goto error;
	    }
	default:
	put_c:
	  cc = c;
	  c = dgetc ();
	  /*put_char:*/
	  if (i>=size)
	    {
	      newsize = size+128;
	      newbuf = (CHAR *)MALLOC(newsize);
	      safememcpy (newbuf, buf, size);
	      MARKFREE(buf);
	      buf = newbuf;
	      size = newsize;
	    }
	  buf[i++] = cc;
	  break;
	}
    }
}

PRIVATE int getId(c,idp,needsop,filenamep)
IN int c;
OUT ExprPtr *idp;
IN Boolean needsop, filenamep;
{
  CHAR *buf;
  COUNTER i, size;
  Boolean (*isok) P((IN int c));
  Boolean op;

  *idp = NOEXPR;
  size = 128;
  buf = (CHAR *)MALLOC(size);
  if (!needsop)
    goto second_chance;
  if (c=='_')
    {
      *idp = (lastPrintExpr==NIL)?NOEXPR:lastPrintExpr;
      return dgetc();
    }
  if (c=='$') /* debugger variable */
    {
      op = false;
      isok = is_alpha;
      goto parse_id;
    }
  if (is_letter(c))
    {
      op = false;
      isok = is_alpha;
    parse_id:
      buf[0] = (CHAR)c;
      i = 1;
      while (c = dgetc(), (*isok) (c))
	{
	  if (i>=size)
	    {
	      CHAR *newbuf;
	      COUNTER newsize;

	      newsize = size + 128;
	      newbuf = (CHAR *)MALLOC(newsize);
	      safememcpy (newbuf, buf, size);
	      MARKFREE (buf);
	      buf = newbuf;
	      size = newsize;
	    }
	  buf[i++] = (CHAR)c;
	}
      if (!op && i==2 && buf[0]=='o' && buf[1]=='p') /* op <symbol> */
	{
	  c = skipBlanks(c);
	second_chance:
	  op = true;
	  if (filenamep)
	    {
	      isok = is_filechar;
	      goto parse_id;
	    }
	  if (is_letter(c))
	    {
	      isok = is_alpha;
	      goto parse_id;
	    }
	  if (is_symb(c))
	    {
	      isok = is_symb;
	      goto parse_id;
	    }
	  MARKFREE(buf);
	  return c; /* syntax error */
	}
      *idp = Kstring(i,buf);
    }
  MARKFREE(buf);
  return c;
}

PRIVATE int getBasicValue (c, vp, gotp, needsop)
IN int c;
OUT ExprPtr *vp;
OUT Boolean *gotp;
IN Boolean needsop;
{
  ENTER;

  *gotp = false;
  switch (c)
    {
    case '"':
      c = getStringCst (c, vp, gotp);
      break;
    case '~': case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      c = getComplexOrInt (c, vp, gotp);
      break;
    default:
      c = getId (c, vp, needsop, false);
      if (*vp==NOEXPR)
	return c;
      *gotp = true;
      KEEPONLY(*vp);
      return c;
    }
  if (*gotp)
    {
      *vp = Kcounted (E_QUOTE, *vp, 0);
      KEEPONLY(*vp);
    }
  return c;
}

PRIVATE int getSelector(c,vp,gotp)
IN int c;
OUT ExprPtr *vp;
OUT Boolean *gotp;
{
  ENTER;

  *gotp = false;
  switch (c)
    {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      {
	COUNTER i;

	c = getInt(c,&i,gotp);
	if (*gotp)
	  *vp = KINT((INT)i);
	else *vp = NOEXPR;
      }
      break;
    default:
      c = getId (c, vp, false, false);
      if (*vp==NOEXPR)
	return c;
      *gotp = true;
      KEEPONLY(*vp);
      return c;
    }
  if (*gotp)
    {
      *vp = Kcounted (E_QUOTE, *vp, 0);
      KEEPONLY(*vp);
    }
  return c;
}

PRIVATE int parseValue(c,e,codep)
IN int c;
IN CONST ExprPtr e;
OUT ExprPtr *codep;
{
  ExprPtr id, code;
  Boolean got;
  ENTER;

  /* syntax:
     expr := id
     | '(' expr ')'
     | '?' expr expr              (note: '?' is not an id)
     | '#' lab expr               (note: '#' is not an id)
     | '<' constructor '>' expr   (note: '<' is not an id)
     | '!' expr                   (note: '!' is not an id)
     | expr '.(' expr ')'         (array or intarray selection)

     I prefer the more usable, but farther from GimML syntax:
     expr := id
     | '"' string '"'
     | number
     | expr '[' expr ']'    (map read, array select, intarray select, list select)
     | expr '.' id          (record select, tuple select)
     | expr '!'             (dereference, constructor argument extraction)

     We compile:
     id as itself (an E_STRING)
     e1[e2] as E_CONS(e1, e2)
     e1.lab as E_TUPLE_2(e1, lab)
   */
  if (e==NOEXPR)
    {
      *codep = NOEXPR;
      c = getBasicValue (c, &id, &got, true);
      if (!got)
	return c;
      code = id;
      (void) KEEP(code);
    }
  else code = e;
  while (true)
    {
      c = skipBlanks (c);
      switch (c)
	{
	case '.': /* record or tuple or constructor selection */
	  c = dgetc ();
	  c = skipBlanks (c);
	  c = getSelector (c, &id, &got);
	  if (!got) /* syntax error */
	    {
	      QUIT;
	      return c;
	    }
	  if (id->kind==(CHAR)E_QUOTE)
	    id = REFVAL(id);
	  (void) KEEP (id);
	  code = Ktuple2 (code, id);
	  KEEPONLY(code);
	  break;
	case '[': /* map or array or intarray or list selection */
	  c = dgetc ();
	  c = skipBlanks (c);
	  c = parseValue (c, NOEXPR, &id);
	  if (id==NOEXPR || c!=']') /* inherited syntax error, or no proper closing bracket */
	    {
	      QUIT;
	      return c;
	    }
	  code = Kons (code, id);
	  KEEPONLY(code);
	  c = dgetc ();
	  break;
	case '!': /* e! = e.ref, extracting data from constructor */
	  c = dgetc ();
	  code = Kref (E_REF, code);
	  KEEPONLY(code);
	  break;
	case '^': /* undo last selector */
	  switch (code->kind)
	    {
	    case E_TUPLE_2: case E_CONS:
	      code = CAR(code);
	      c = dgetc();
	      break;
	    case E_REF:
	      code = REFVAL(code);
	      c = dgetc();
	      break;
	    default:
	      QUIT;
	      return c;
	    }
	  break;
	default: /* end of expression */
	  KEEPONLY(code);
	  *codep = code;
	  return c;
	}
    }
}

PRIVATE int parseValueList(c,codep)
IN int c;
OUT ExprPtr *codep;
{
  ExprPtr onecode, codes;
  ENTER;

  *codep = NOEXPR;
  codes = NIL;
  while (true)
    {
      c = parseValue(c, NOEXPR, &onecode);
      if (onecode==NOEXPR)
	{
	  QUIT;
	  return c;
	}
      codes = MakePair (E_CONS, onecode, codes);
      KEEPONLY(codes);
      switch (c)
	{
	case ',':
	  c = dgetc ();
	  c = skipBlanks (c);
	  break;
	case '\n':
	case -1:
	  *codep = codes;
	  KEEPONLY(codes);
	  return c;
	default: /* syntax error */
	  QUIT;
	  return c;
	}
    }
}

PRIVATE ExprPtr dbgEvalRef(e1,typep)
IN CONST ExprPtr e1;
OUT ExprPtr *typep;
{
  Boolean found;

  if (typep!=NULL && *typep!=NOEXPR)
    {
      if ((*typep)->kind!=(CHAR)E_TYPE_APPL)
	*typep = NOEXPR;
      else
	{ /* imitated from valPretty() */
	  ExprPtr f;

	  f = APPLFUN(*typep);
	  f = fxSym(f);
	  if (f==stringSym || f==arraySym || f==intSym ||
	      f==funSym || f==boolSym || f==promiseSym ||
	      f==tableSym || f==mapSym)
	    return NOEXPR;
	  else if (f==listSym)
	    {
	      *typep = KONS(CAR(APPLARG(*typep)),*typep);
	      e1 = Ktuple2 (CAR(e1), CDR(e1));
	      (void) KEEP (e1);
	      return e1;
	    }
	  else if (f==refSym)
	    {
	      *typep = CAR(APPLARG(*typep));
	      e1 = REFVAL(e1);
	      return e1;
	    }
	  else if (f==exnSym)
	    {
	      if (ISCONS(e1))
		{
		  *typep = getTypeInExn(CAR(e1));
		  e1 = CDR(e1);
		  return e1;
		}
	      else return NOEXPR;
	    }
	  else if (f==dynSym)
	    {
	      *typep = PACKTYPE(e1);
	      e1 = PACKVAL(e1);
	      return e1;
	    }
	  else /* datatype: imitated from dataPretty() */
	    {
	      ExprPtr tycons,ttm,con,r,tyhdr,cons,type;
	      SHORT magic;

	      tycons = STyCons(basisState);
	      ttm = MapGet(f,TyTags(basisTable),&found);
	      if (found && ttm->kind==E_STRING) /* just one constructor */
		con = ttm;
	      else
		{
		  ttm = CAR(e1);
		  magic = e1->magic;
		  con = AVAL(CAR(ttm))->expr[magic-ttm->magic];
		  e1 = CDR(e1);
		}
	      r = MapGet(f,tycons,&found);
	      if (found)
		{
		  tyhdr = CAR(r);
		  cons = CDR(r);
		  type = MapGetSure(con,cons);
		  if (fxSym(APPLFUN(type))==funSym)
		    {
		      type = CAR(APPLARG(type));
		      type = ApplyType(type,tyhdr,APPLARG(*typep),
				       basisAssum,basisState);
		      *typep = type;
		      return e1;
		    }
		  else return NOEXPR;
		}
	      else /* abstype */
		*typep = NOEXPR;
	    }
	}
    }
  /* here we don't have any info on types */
  if (e1->kind==(CHAR)E_DATA) /* datatype */
    {
      if (e1->magic>=0) /* has an argument */
	return CDR(e1);
    }
  else if (ISCONS(e1))
    {
      if (CAR(e1)->kind==(CHAR)E_EXCEPTION)
	{
	  if (typep!=NULL)
	    *typep = getTypeInExn(CAR(e1));
	  return CDR(e1);
	}
    }
  else if (e1->kind==(CHAR)E_PACK)
    {
      if (typep!=NULL)
	*typep = PACKTYPE(e1);
      return PACKVAL(e1);
    }
  else if (e1->kind==(CHAR)E_REF)
    {
      return REFVAL(e1);
    }
  return NOEXPR;
}

PRIVATE ExprPtr dbgEvalRefIter(e1,typep)
IN CONST ExprPtr e1;
INOUT ExprPtr *typep;
{
  ExprPtr ttm;
  COUNTER magic, m;
  ExprPtr f;

  if (typep!=NULL && *typep!=NOEXPR)
    {
      if ((*typep)->kind!=(CHAR)E_TYPE_APPL)
	return e1;
      f = APPLFUN(*typep);
      f = fxSym(f);
      if (f==stringSym || f==arraySym || f==intSym ||
	  f==funSym || f==boolSym || f==promiseSym ||
	  f==tableSym || f==mapSym ||
	  f==listSym)
	return e1;
      return dbgEvalRef(e1,typep);
    }
  else
    {
    again:
      switch (e1->kind)
	{
	case E_DATA:
	  magic = e1->magic;
	  if (magic<0)
	    return NOEXPR;
	  e1 = CDR(e1);
	  goto again;
	default:
	  return e1;
	}
    }
}

PRIVATE ExprPtr dbgEval(snap,e,typep)
IN struct snapshot *snap;
IN CONST ExprPtr e;
OUT ExprPtr *typep;
{
  switch (e->kind)
    {
    case E_STRING: /* identifier, or debugger variable (starting with _) */
      if (SLEN(e)>=1 && STEXT(e)[0]=='$') /* debugger variable */
	{
	  ExprPtr val;
	  Boolean found;

	  val = MapGet(e,lastPrintVals,&found);
	  if (!found)
	    return NOEXPR;
	  if (typep!=NULL)
	    {
	      *typep = MapGet(e,lastPrintTypes,&found);
	      if (!found)
		*typep = NOEXPR;
	    }
	  return val;
	}
      return getValue (e, snap, typep);
    case E_QUOTE: /* explicit constant */
      e = REFVAL(e);
      if (typep!=NULL)
	switch (e->kind)
	  {
	  case E_STRING: *typep = stringType; break;
	  case E_INT: *typep = intType; break;
	  case E_REAL: case E_COMPLEX: *typep = num_alpha; break;
	  }
      return e;
    case E_TUPLE_2:
      {
	ExprPtr e1, lab;
	Boolean found;

	e1 = dbgEval (snap, CAR (e), typep);
	if (e1==NOEXPR)
	  return NOEXPR;
	lab = CDR(e);
	e1 = dbgEvalRefIter(e1,typep);
	/*!!! deal with E_BIAS, too! */
	switch (e1->kind)
	  {
	  case E_TUPLE_1:
	    if (lab!=ione)
	      return NOEXPR;
	    /* not meant to happen: don't do anything with *typep. */
	    return REFVAL(e1);
	  case E_RECORD_1:
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		switch ((*typep)->kind)
		  {
		  case E_MAPLET: case E_MAP:
		    *typep = MapGet(lab,*typep,&found);
		    if (!found)
		      {
			*typep = NOEXPR;
			return NOEXPR;
		      }
		    break;
		  default:
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    if (lab->kind!=E_INT)
	      {
		lab = MapGet(lab,CAR(e1),&found);
		if (!found)
		  return NOEXPR;
	      }
	    if (lab!=ione)
	      return NOEXPR;
	    return CDR(e1);
	  case E_TUPLE_2:
	    if (lab==ione)
	      {
		if (typep!=NULL && *typep!=NOEXPR)
		  {
		    if ((*typep)->kind==(CHAR)E_CONS)
		      *typep = CAR(*typep);
		    else
		      {
			*typep = NOEXPR;
			return NOEXPR;
		      }
		  }
		return CAR(e1);
	      }
	    if (lab==itwo)
	      {
		if (typep!=NULL && *typep!=NOEXPR)
		  {
		    if ((*typep)->kind==(CHAR)E_CONS &&
			CDR(*typep)->kind==(CHAR)E_CONS)
		      *typep = CAR(CDR(*typep));
		    else
		      {
			*typep = NOEXPR;
			return NOEXPR;
		      }
		  }
		return CDR(e1);
	      }
	    return NOEXPR;
	  case E_RECORD_GNRL:
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		*typep = MapGet(lab,*typep,&found);
		if (!found)
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    if (lab->kind!=(CHAR)E_INT)
	      {
		lab = MapGet(lab,ASHAPE(e1),&found);
		if (!found)
		  return NOEXPR;
	      }
	    goto tuple_gnrl;
	  case E_TUPLE_GNRL:
	    if (lab->kind!=(CHAR)E_INT)
	      return NOEXPR;
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		INT n = IVAL(lab);
		ExprPtr type = *typep;

		while (--n!=0)
		  {
		    if (type->kind==(CHAR)E_CONS)
		      type = CDR(type);
		    else
		      {
			type = NIL;
			break;
		      }
		  }
		if (type->kind==(CHAR)E_CONS)
		  *typep = CAR(type);
		else
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	  tuple_gnrl:
	    {
	      INT i;

	      i = IVAL(lab)-1;
	      if (i<0 || i>=AVAL(e1)->n)
		return NOEXPR;
	      return AVAL(e1)->expr[i];
	    }
	  default:
	    return NOEXPR;
	  }
      }
    case E_CONS:
      {
	ExprPtr e1, e2;
	Boolean found;
	INT i;

	e1 = dbgEval (snap, CAR(e), typep);
	if (e1==NOEXPR)
	  return NOEXPR;
	e1 = dbgEvalRefIter(e1,typep);
	e2 = dbgEval (snap, CDR(e), NULL);
	if (e2==NOEXPR)
	  return NOEXPR;
	switch (e1->kind)
	  {
	  case E_EMPTY:
	    return NOEXPR;
	  case E_MAPLET:
	    if (MAPLETX(e1)!=e2)
	      return NOEXPR;
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		if ((*typep)->kind==(CHAR)E_TYPE_APPL
		    && fxSym(APPLFUN(*typep))==mapSym)
		  *typep = CAR(CDR(APPLARG(*typep)));
		else
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    return MAPLETY(e1);
	  case E_MAP:
	    e1 = MapGet (e2, e1, &found);
	    if (!found)
	      return NOEXPR;
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		if ((*typep)->kind==(CHAR)E_TYPE_APPL
		    && fxSym(APPLFUN(*typep))==mapSym)
		  *typep = CAR(CDR(APPLARG(*typep)));
		else
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    return e1;
	  case E_ARRAY_2:
	    if (e2->kind!=(CHAR)E_INT)
	      return NOEXPR;
	    i = IVAL(e2);
	    if (i<0 || i>=2)
	      return NOEXPR;
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		if ((*typep)->kind==(CHAR)E_TYPE_APPL
		    && fxSym(APPLFUN(*typep))==arraySym)
		  *typep = CAR(APPLARG(*typep));
		else
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    return (i==0)?CAR(e1):CDR(e1);
	  case E_ARRAY_GNRL:
	    if (e2->kind!=(CHAR)E_INT)
	      return NOEXPR;
	    i = IVAL(e2);
	    if (i<0 || i>=AVAL(e1)->n)
	      return NOEXPR;
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		if ((*typep)->kind==(CHAR)E_TYPE_APPL
		    && fxSym(APPLFUN(*typep))==arraySym)
		  *typep = CAR(APPLARG(*typep));
		else
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    return AVAL(e1)->expr[i];
	  case E_EXTERNAL:
	    if (EXTDESC(e1)==&iarrayExtDesc)
	      {
		i = IVAL(e2);
		if (i<0 || i>=((INT *)EXTDATA(e1))[0])
		  return NOEXPR;
		if (typep!=NULL && *typep!=NOEXPR)
		  {
		    if (*typep==iarrayType)
		      *typep = intType;
		    else
		      {
			*typep = NOEXPR;
			return NOEXPR;
		      }
		  }
		e1 = Kint (((INT *)EXTDATA(e1))[i+1]);
		return KEEP(e1);
	      }
	    return NOEXPR;
	  case E_CONS:
	    if (e2->kind!=(CHAR)E_INT)
	      return NOEXPR;
	    i = IVAL(e2);
	    while (i!=0 && e1->kind==(CHAR)E_CONS)
	      {
		i--;
		e1 = CDR(e1);
	      }
	    if (e1->kind!=(CHAR)E_CONS)
	      return NOEXPR;
	    if (typep!=NULL && *typep!=NOEXPR)
	      {
		if ((*typep)->kind==(CHAR)E_TYPE_APPL
		    && fxSym(APPLFUN(*typep))==listSym)
		  *typep = CAR(APPLARG(*typep));
		else
		  {
		    *typep = NOEXPR;
		    return NOEXPR;
		  }
	      }
	    return CAR(e1);
	  default:
	    return NOEXPR;
	  }
      }
    case E_REF:
      {
	ExprPtr e1;

	e1 = dbgEval (snap, REFVAL(e), typep);
	if (e1==NOEXPR)
	  return NOEXPR;
	return dbgEvalRef(e1,typep);
      }
    default:
      return NOEXPR;
    }
}

PRIVATE void printDbgExpr(f, e)
IN hFILE *f;
IN CONST ExprPtr e;
{
  switch (e->kind)
    {
    case E_STRING: (void) printSym (e, do_printfile, (char *)f); break;
    case E_QUOTE: valPrintNoType (REFVAL(e), do_printfile, (char *)f, 0); break;
    case E_TUPLE_2:
      printDbgExpr(f, CAR(e));
      hfputc ('.', f);
      e = CDR(e);
      if (e->kind==(CHAR)E_STRING)
	(void) printSym (e, do_printfile, (char *)f);
      else
	valPrintNoType (e, do_printfile, (char *)f, 0);
      break;
    case E_CONS:
      printDbgExpr(f, CAR(e));
      hfputc ('[', f);
      printDbgExpr(f, CDR(e));
      hfputc (']', f);
      break;
    case E_REF:
      printDbgExpr(f,REFVAL(e));
      hfputc ('!', f);
      break;
    }
}

PRIVATE void dbgView(f, snap, codes, verbose, cnt)
IN hFILE *f;
IN struct snapshot *snap;
IN CONST ExprPtr codes;
IN Boolean verbose, cnt;
{
  ExprPtr code, val, type;
  INT len, lmargin;

  if (cnt)
    {
      len = ListLength(codes);
      lmargin = 5;
    }
  else
    {
      len = 0;
      lmargin = 0;
    }
  for (; codes!=NIL; codes = CDR(codes))
    {
      code = CAR(codes);
      val = dbgEval (snap, code, &type);
      if (val!=NOEXPR || verbose)
	{
	  if (cnt)
	    hfprintf (hstderr, "%3d: ", len);
	  printDbgExpr (f, code);
	  hfputs (message (747), f);
	  showVal (val, type, lmargin);
	  hfflush (f);
	}
      len--;
    }
}

MAPSWEEPFUNC(do_printdebugvars)
{
  ExprPtr type;
  Boolean found;

  printDbgExpr (hstderr, x);
  hfputs ("=", hstderr);
  type = MapGet(x,lastPrintTypes,&found);
  showVal (y,found?type:NOEXPR,SLEN(x)+1);
  return false;
}

PRIVATE void printDebugVars()
{
  (void) MapSweep(lastPrintVals,do_printdebugvars,NOEXPR);
}

MAPSWEEPFUNC(do_possiblerefinements_record)
{
  struct possiblerefinements *ref = MAPDATA(struct possiblerefinements *);
  char *buf;
  ExprPtr s;

  if (ref->n<=0)
    return true;
  if (x->kind!=(CHAR)E_STRING)
    return false;
  ref->n--;
  buf = (char *)alloctmp(SLEN(x)+1);
  buf[0] = '.';
  safememcpy(buf+1,STEXT(x),SLEN(x));
  s = Kstring(SLEN(x)+1,(CHAR *)buf);
  (void) KEEP(s);
  ref->strings = KONS(s,ref->strings);
  freetmp(buf);
  return false;
}

PRINTFUNC(do_possiblerefinements_print)
{
  CHAR **pp;

  pp = PRINTDATA(CHAR **);
  safememcpy(*pp,text,len);
  *pp += len;
}

MAPSWEEPFUNC(do_possiblerefinements_map)
{
  struct possiblerefinements *ref = MAPDATA(struct possiblerefinements *);
  char *buf, *p;
  ExprPtr s;
  INT len;

  if (ref->n<=0)
    return true;
  switch (x->kind)
    {
    case E_STRING: len = 2*SLEN(x)+3; break;
    case E_INT: len = NDIGS_IN_ULONG+10; break;
    case E_REAL: len = 64; break;
    case E_COMPLEX: len = 129; break;
    default: return false;
    }
  ref->n--;
  buf = (char *)alloctmp(len);
  buf[0] = '[';
  p = buf+1;
  valPrintNoType(x,do_possiblerefinements_print,(char *)&p,0);
  *p++ = ']';
  len = p-buf;
  s = Kstring(len,(CHAR *)buf);
  (void) KEEP(s);
  ref->strings = KONS(s,ref->strings);
  freetmp(buf);
  return false;
}

PRIVATE ExprPtr possibleRefinements(e,type,n,completep)
IN CONST ExprPtr e;
IN CONST ExprPtr type;
IN COUNTER n;
OUT Boolean *completep;
{
  *completep = false;
  if (n<=0)
    return NIL;
 again:
  if (type!=NOEXPR)
    switch (type->kind)
      {
      case E_MAPLET: case E_MAP:
	{
	  struct possiblerefinements ref;
	  ENTER;

	  ref.strings = NIL;
	  ref.n = n;
	  if (!MapSweep(type,do_possiblerefinements_record,(CHAR *)&ref))
	    *completep = true;
	  KEEPONLY(ref.strings);
	  return ref.strings;
	}
      case E_CONS:
	{
	  ExprPtr strings, s;
	  char buf[16];
	  COUNTER i;
	  ENTER;

	  strings = NIL;
	  i = 0;
	  while (type->kind==(CHAR)E_CONS && n>0)
	    {
	      n--;
	      i++;
	      sprintf (buf, ".%d", i);
	      s = Kstr((CHAR *)buf);
	      (void) KEEP (s);
	      strings = KONS(s,strings);
	      type = CDR(type);
	    }
	  KEEPONLY(strings);
	  if (type->kind!=(CHAR)E_CONS)
	    *completep = true;
	  return strings;
	}
      case E_TYPE_APPL:
	{
	  ExprPtr f;

	  f = fxSym(APPLFUN(type));
	  if (f==stringSym || f==arraySym || f==intSym ||
	      f==funSym || f==boolSym || f==promiseSym ||
	      f==tableSym || f==mapSym ||
	      f==listSym)
	    /* drop to untyped case */;
	  else if (f==refSym || f==dynSym)
	    {
	      ExprPtr strings, s;
	      ENTER;

	      s = Kstr((CHAR *)"!");
	      (void) KEEP(s);
	      strings = Kons(s,NIL);
	      KEEPONLY(strings);
	      *completep = true;
	      return strings;
	    }
	  else if (f==exnSym)
	    {
	      if (ISCONS(e))
		{
		  type = getTypeInExn(CAR(e));
		  e = CDR(e);
		  goto again;
		}
	    }
	  else /* datatype */
	    {
	      ExprPtr tycons, ttm, con, r, tyhdr, cons, args;
	      SHORT magic;
	      Boolean found;

	      tycons = STyCons(basisState);
	      ttm = MapGet(f,TyTags(basisTable),&found);
	      if (found && ttm->kind==E_STRING) /* just one constructor */
		con = ttm;
	      else
		{
		  ttm = CAR(e);
		  magic = e->magic;
		  con = AVAL(CAR(ttm))->expr[magic-ttm->magic];
		  e = CDR(e);
		}
	      r = MapGet(f,tycons,&found);
	      if (found)
		{
		  args = APPLARG(type);
		  tyhdr = CAR(r);
		  cons = CDR(r);
		  type = MapGetSure(con,cons);
		  if (fxSym(APPLFUN(type))==funSym)
		    {
		      type = CAR(APPLARG(type));
		      type = ApplyType(type,tyhdr,args,
				       basisAssum,basisState);
		      goto again;
		    }
		  else return NIL;
		}
	      else /* abstype */
		return NIL;
	    }
	}
	/* drop to untyped case */
      }
  switch (e->kind)
    {
    case E_RECORD_0: case E_TUPLE_0:
      *completep = true; return NIL;
    case E_RECORD_1:
      {
	struct possiblerefinements ref;
	ENTER;

	ref.strings = NIL;
	ref.n = n;
	(void) do_possiblerefinements_record(CAR(e),NIL,(CHAR *)&ref);
	*completep = true;
	KEEPONLY(ref.strings);
	return ref.strings;
      }
    case E_RECORD_GNRL:
      {
	struct possiblerefinements ref;
	ENTER;

	ref.strings = NIL;
	ref.n = n;
	if (!MapSweep(ASHAPE(e),do_possiblerefinements_record,(CHAR *)&ref))
	  *completep = true;
	KEEPONLY(ref.strings);
	return ref.strings;
      }
    case E_TUPLE_1:
      {
	ExprPtr strings, s;
	ENTER;

	s = Kstr((CHAR *)".1");
	(void) KEEP(s);
	strings = Kons(s,NIL);
	KEEPONLY(strings);
	*completep = true;
	return strings;
      }
    case E_TUPLE_2:
      {
	ExprPtr strings, s;
	ENTER;

	s = Kstr((CHAR *)".1");
	(void) KEEP(s);
	strings = KONS(s,NIL);
	if (n>=2)
	  {
	    *completep = true;
	    s = Kstr((CHAR *)".2");
	    (void) KEEP(s);
	    strings = KONS(s,strings);
	  }
	KEEPONLY(strings);
	return strings;
      }
    case E_TUPLE_GNRL:
      {
	ExprPtr strings, s;
	char buf[16];
	COUNTER i;
	ENTER;

	for (strings=NIL, i=0; n>0 && i<AVAL(e)->n;)
	  {
	    n--;
	    i++;
	    sprintf (buf, ".%d", i);
	    s = Kstr((CHAR *)buf);
	    (void) KEEP (s);
	    strings = KONS(s,strings);
	  }
	KEEPONLY(strings);
	if (n<=0)
	  *completep = true;
	return strings;
      }
    case E_EMPTY: case E_MAPLET: case E_MAP:
      {
	struct possiblerefinements ref;
	ENTER;

	ref.strings = NIL;
	ref.n = n;
	if (!MapSweep(e,do_possiblerefinements_map,(CHAR *)&ref))
	  *completep = true;
	KEEPONLY(ref.strings);
	return ref.strings;
      }
    case E_CONS:
      if (type==NOEXPR && CAR(e)->kind==(CHAR)E_EXCEPTION)
	{
	  type = getTypeInExn(CAR(e));
	  e = CDR(e);
	  goto again;
	}
      /*fallthrough*/
    case E_NIL:
      {
	ExprPtr strings, s;
	char buf[16];
	COUNTER i;
	ENTER;

	strings = NIL;
	i = 0;
	while (e->kind==(CHAR)E_CONS && n>0)
	  {
	    n--;
	    sprintf (buf, "[%d]", i);
	    s = Kstr((CHAR *)buf);
	    (void) KEEP (s);
	    strings = KONS(s,strings);
	    i++;
	    e = CDR(e);
	  }
	KEEPONLY(strings);
	if (e->kind!=(CHAR)E_CONS)
	  *completep = true;
	return strings;
      }
    case E_ARRAY_2:
      {
	ExprPtr strings, s;
	ENTER;

	s = Kstr((CHAR *)"[0]");
	(void) KEEP(s);
	strings = KONS(s,NIL);
	if (n>=2)
	  {
	    *completep = true;
	    s = Kstr((CHAR *)"[1]");
	    (void) KEEP(s);
	    strings = KONS(s,strings);
	  }
	KEEPONLY(strings);
	return strings;
      }
    case E_ARRAY_GNRL:
      {
	ExprPtr strings, s;
	char buf[16];
	COUNTER i;
	ENTER;

	for (strings=NIL, i=0; n>0 && i<AVAL(e)->n;)
	  {
	    n--;
	    sprintf (buf, "[%d]", i);
	    s = Kstr((CHAR *)buf);
	    (void) KEEP (s);
	    strings = KONS(s,strings);
	    i++;
	  }
	KEEPONLY(strings);
	if (n<=0)
	  *completep = true;
	return strings;
      }
    case E_EXTERNAL:
      if (EXTDESC(e)==&iarrayExtDesc)
      {
	ExprPtr strings, s;
	char buf[16];
	COUNTER i;
	ENTER;

	for (strings=NIL, i=0; n>0 && i<((INT *)EXTDATA(e))[0];)
	  {
	    n--;
	    sprintf (buf, "[%d]", i);
	    s = Kstr((CHAR *)buf);
	    (void) KEEP (s);
	    strings = KONS(s,strings);
	    i++;
	  }
	KEEPONLY(strings);
	if (n<=0)
	  *completep = true;
	return strings;
      }
      break;
    case E_DATA:
      if (e->magic>=0) /* has an argument */
	{
	  type = NOEXPR;
	  e = CDR(e);
	  goto again;
	}
      break;
    }
  *completep = true;
  return NIL;
}

PRIVATE void displayPossibleSelectors(snap,e,cnt)
IN struct snapshot *snap;
IN CONST ExprPtr e;
IN COUNTER cnt;
{
  ExprPtr val,refines,type;
  Boolean complete;
  /*ExprPtr s;*/
  ENTER;

  val = dbgEval (snap, e, &type);
  if (val!=NOEXPR)
    {
      lastPrintExpr = e;
      /*
      s = Kstr("_");
      (void) KEEP (s);
      setMapPut(s,val,&lastPrintVals);
      if (type!=NOEXPR)
	setMapPut(s,type,&lastPrintTypes);
      */
      refines = possibleRefinements(val,type,cnt,&complete);
      refines = Reverse(refines);
      if (refines!=NIL)
	{
	  char *comma;

	  hfputs ("  (further selectors: ", hstderr);
	  comma = "";
	  while (refines!=NIL)
	    {
	      hfputs (comma, hstderr);
	      do_printfile (STEXT(CAR(refines)),SLEN(CAR(refines)),
			    (char *)hstderr);
	      refines = CDR(refines);
	      comma = ", ";
	    }
	  if (!complete)
	    hfputs (", ...", hstderr);
	  switch (e->kind)
	    {
	    case E_TUPLE_2: case E_CONS:
	      hfputs (", or ^ to undo last selector", hstderr);
	      break;
	    }
	  hfputs (")\n", hstderr);
	  hfflush (hstderr);
	}
    }
  QUIT;
}

PRIVATE ExprPtr addMonitorExprs (codes, list)
IN CONST ExprPtr codes, list;
{
  ExprPtr code, l;
  ENTER;

  for (; codes!=NIL; codes = CDR(codes))
    {
      code = CAR(codes);
      for (l = list; l!=NIL; l = CDR(l))
	if (code==CAR(l))
	  break;
      if (l!=NIL)
	continue;
      list = MakePair (E_CONS, code, list);
      KEEPONLY(list);
    }
  return list;
}

PRIVATE ExprPtr delMonitorExprs (codes, list)
IN CONST ExprPtr codes, list;
{
  ExprPtr code, l, prev;
  ENTER;

  for (; codes!=NIL; codes = CDR(codes))
    {
      code = CAR(codes);
      for (prev = NIL, l = list; l!=NIL; prev = l, l = CDR(l))
	if (code==CAR(l))
	  break;
      if (l==NIL)
	continue;
      if (prev==NIL)
	list = CDR(list);
      else SETCDR(prev, CDR(list));
      KEEPONLY(list);
    }
  return list;
}

PRIVATE void checkExceptions (snap, codes, checkp, verbose)
IN struct snapshot *snap;
IN ExprPtr codes;
IN Boolean checkp, verbose;
{
  ExprPtr code, val, type;

  for (; codes!=NIL; codes = CDR(codes))
    {
      code = CAR(codes);
      val = dbgEval (snap, code, &type);
      if (val!=NOEXPR)
	{
	  if (val->kind==(CHAR)E_INTERPRETED_CLOSURE && QENV(val)->kind==(CHAR)E_TUPLE_2)
	    val = CDR(QENV(val));
	  if (val->kind==(CHAR)E_EXCEPTION)
	    {
	      if (checkp)
		COUNTEDN(val) &= ~EXC_BENIGN;
	      else
		COUNTEDN(val) |= EXC_BENIGN;
	    }
	  else if (verbose)
	    {
	      printDbgExpr (hstderr, code);
	      hfputs (message (742), hstderr);
	      hfflush (hstderr);
	    }
	}
    }
}

/*
PRIVATE void freeBreaks (breaks)
IN struct breakpoints *breaks;
{
  if (breaks!=NULL)
    {
      MARKFREE(breaks->pcs);
      MARKFREE(breaks->savedops);
      MARKFREE(breaks->breaks);
    }
}
*/

PRIVATE void newBreaks(breaks)
INOUT struct breakpoints *breaks;
{
  OPCODE **pcs, *savedops;
  ExprPtr *bcs;
  INT size;

  size = breaks->size;
  size += 128;
  pcs = (OPCODE **)MALLOC(size * sizeof(OPCODE *));
  safememcpy (pcs, breaks->pcs, breaks->size * sizeof(OPCODE *));
  MARKFREE (breaks->pcs);
  breaks->pcs = pcs;
  savedops = (OPCODE *)MALLOC(size * sizeof(OPCODE));
  safememcpy (savedops, breaks->savedops, breaks->size * sizeof(OPCODE));
  MARKFREE(breaks->savedops);
  breaks->savedops = savedops;
  bcs = (ExprPtr *)MALLOC(size * sizeof(ExprPtr));
  safememcpy (bcs, breaks->bcs, breaks->size * sizeof(ExprPtr));
  MARKFREE(breaks->bcs);
  breaks->bcs = bcs;
  breaks->size = size;
}

PRIVATE COUNTER findBreak (pc, breaks)
IN CONST REGISTER OPCODE *pc;
IN CONST struct breakpoints *breaks;
{
  REGISTER OPCODE **pcp, **pcpend;

  pcp = breaks->pcs;
  pcpend = pcp+breaks->n;
  for (; pcp<pcpend; pcp++)
    if (*pcp==pc)
      return pcp-breaks->pcs;
  return -1;
}

PRIVATE COUNTER addBreak (bc, pc, breaks, checkdups)
INOUT byteCodeChunk *bc;
INOUT OPCODE *pc;
INOUT struct breakpoints *breaks;
IN Boolean checkdups;
{
  if (checkdups)
    {
      COUNTER pos;

      pos = findBreak (pc, breaks);
      if (pos>=0) /* already there */
	return pos;
    }
  if (breaks->n>=breaks->size)
    newBreaks (breaks);
  {
    COUNTER n;

    n = breaks->n;
    breaks->bcs[n] = (ExprPtr)bc;
    breaks->pcs[n] = pc;
    breaks->savedops[n] = *pc;
    *pc = BC_BREAK;
    breaks->n = n+1;
    return n;
  }
}

PUBLIC void restoreBreaks ()
{
  COUNTER n;
  OPCODE savedop, *pc;

  n = breakPoints.n;
  for (; n--!=0; )
    {
      pc = breakPoints.pcs[n];
      savedop = breakPoints.savedops[n];
      breakPoints.savedops[n] = *pc;
      *pc = savedop;
    }
}

PUBLIC void reinstallBreaks ()
{
  COUNTER i, n;
  OPCODE savedop, *pc;

  n = breakPoints.n;
  for (i=0; i<n; i++)
    {
      pc = breakPoints.pcs[i];
      savedop = breakPoints.savedops[i];
      breakPoints.savedops[i] = *pc;
      *pc = savedop;
    }
}

PRIVATE Boolean sweepBreaks1 (fun, file, line, bc, func, data, stoponfirst)
IN CONST ExprPtr fun; /* function name, NOEXPR if not present */
IN CONST ExprPtr file; /* file name, NIL if toplevel, NOEXPR if not present */
IN INT line; /* line number, -1 if not present; at least one of the arguments should be present. */
IN byteCodeChunk *bc;
IN sweepBreaksFun func;
INOUT char *data;
IN Boolean stoponfirst;
{
  OPCODE *pc;
  OPCODE *dpc;
  OPCODE dop;
  ExprPtr current_file;
  ExprPtr current_name;
  INT current_line = -1;
  ExprPtr stack = NIL;
  INT flags;
  Boolean changed;
  ENTER;

  dpc = bc->debug;
#define SWB_FOUNDFILE 0x1
#define SWB_FOUNDNAME 0x2
#define SWB_FOUNDLINE 0x4
#define SWB_FIRSTLINE 0x8
#define SWB_MASK (SWB_FOUNDFILE | SWB_FOUNDNAME | SWB_FOUNDLINE | SWB_FIRSTLINE)
#define SWB_FOUNDFILEP(flags) (flags & SWB_FOUNDFILE)
#define SWB_FOUNDNAMEP(flags) (flags & SWB_FOUNDNAME)
#define SWB_FOUNDLINEP(flags) (flags & SWB_FOUNDLINE)
#define SWB_FIRSTLINEP(flags) (flags & SWB_FIRSTLINE)
#define SWB_SET(flags,mask) flags |= mask
#define SWB_CLR(flags, mask) flags &= ~mask
#define SWB_SETFOUNDFILE(flags) SWB_SET(flags,SWB_FOUNDFILE)
#define SWB_SETFOUNDNAME(flags) SWB_SET(flags,SWB_FOUNDNAME)
#define SWB_SETFOUNDLINE(flags) SWB_SET(flags,SWB_FOUNDLINE)
#define SWB_SETFIRSTLINE(flags) SWB_SET(flags,SWB_FIRSTLINE)
#define SWB_CLRFOUNDFILE(flags) SWB_CLR(flags,SWB_FOUNDFILE)
#define SWB_CLRFOUNDNAME(flags) SWB_CLR(flags,SWB_FOUNDNAME)
#define SWB_CLRFOUNDLINE(flags) SWB_CLR(flags,SWB_FOUNDLINE)
#define SWB_CLRFIRSTLINE(flags) SWB_CLR(flags,SWB_FIRSTLINE)

  current_file = NIL;
  current_name = NIL;
  {
    ExprPtr prof;
 
    prof = bc->consts[0];
    if (prof->kind!=(CHAR)BUILD_HEADER_KIND ||
	(TYPE(prof)!=EA_PROFILE && TYPE(prof)!=EA_CLR_PROFILE))
      ;
    else
      {
	prof = ARGS(prof);
	current_name = CAR(prof);
	prof = CDR(prof);
	current_file = CDR(prof);
      }
  }
  flags = SWB_FIRSTLINE;
  if (file==NOEXPR)
    SWB_SETFOUNDFILE(flags);
  if (fun==NOEXPR)
    SWB_SETFOUNDNAME(flags);
  if (line==-1)
    SWB_SETFOUNDLINE(flags);
  while (true)
    {
      switch (dop = *dpc++)
	{
	case DBG_BEGIN:
	  dpc += 2;
	  goto push;
	case DBG_BEGIN_BIG:
	  dpc += 3;
	push:
	  /* push current information on stack. */
	  stack = MakeCounted (E_COUNTED, stack, flags);
	  KEEPONLY(stack);
	  stack = MakePair (E_CONS, current_name, stack);
	  KEEPONLY(stack);
	  stack = MakePair (E_CONS, current_file, stack);
	  KEEPONLY(stack);
	  break;
	case DBG_END:
	  dpc++;
	  goto pop;
	case DBG_END_BIG:
	  dpc += 2;
	pop:
	  /* pop from the stack: */
	  changed = false;
	  if (CAR(stack)!=current_file)
	    {
	      changed = true;
	      current_file = CAR(stack);
	    }
	  stack = CDR(stack);
	  if (CAR(stack)!=current_name)
	    {
	      changed = true;
	      current_name = CAR(stack);
	    }
	  stack = CDR(stack);
	  if (changed)
	    flags = COUNTEDN(stack);
	  else
	    flags = SWB_CLRFIRSTLINE(COUNTEDN(stack)) | SWB_FIRSTLINEP(flags);
	  /* keep firstline bit if nothing changed during the scope */
	  stack = COUNTEDVAL(stack);
	  break;
	case DBG_WHERE:
	  pc = bc->instr + (INT)*dpc++;
	  goto where;
	case DBG_WHERE_BIG:
	  pc = bc->instr + ((((INT)(dpc[0])) << (8*sizeof(OPCODE))) |
			    (((INT)(dpc[1])) & ((1 << (8*sizeof(OPCODE))) - 1)));
	  dpc += 2;
	where:
	  if (flags==SWB_MASK)
	    {
	      if ((*func) (bc, pc, data))
		{
		  QUIT;
		  return true;
		}
	      if (stoponfirst)
		SWB_CLR(flags, SWB_FIRSTLINE);
	    }
	  dpc += 3;
	  break;
	case DBG_ACCESS:
	  dpc++;
	  break;
	case DBG_VARREF:
	  dpc += 3;
	  break;
	case DBG_DEFNAME:
	case DBG_DEFCON:
	case DBG_DEFEXC:
	  dpc += 2;
	  break;
	case DBG_FILE:
	  current_file = bc->consts[*dpc++];
	  if (file==NOEXPR || file==current_file)
	    SWB_SETFOUNDFILE(flags);
	  else
	    SWB_CLRFOUNDFILE(flags);
	  break;
	case DBG_LINE:
	  current_line = (INT)(unsigned short)*dpc;
	  dpc += 2;
	  if (line==-1 || line==current_line)
	    SWB_SETFOUNDLINE(flags);
	  else
	    SWB_CLRFOUNDLINE(flags);
	  break;
	case DBG_NAME:
	  current_name = bc->consts[*dpc++];
	  if (fun==NOEXPR || fun==current_name)
	    SWB_SETFOUNDNAME(flags);
	  else
	    SWB_CLRFOUNDNAME(flags);
	  break;
	case DBG_ENV:
	  dpc++;
	  break;
	case DBG_STOP:
	  QUIT;
	  return false;
	default:
	  QUIT;
	  return true;
	}
    }
  QUIT;
}

PRIVATE Boolean sweepBreaks (fun, file, line, func, data, stoponfirst)
IN CONST ExprPtr fun; /* function name, NOEXPR if not present */
IN CONST ExprPtr file; /* file name, NIL if toplevel, NOEXPR if not present */
IN INT line; /* line number, -1 if not present; at least one of the arguments should be present. */
IN sweepBreaksFun func;
INOUT char *data;
IN Boolean stoponfirst; /* true if should only sweep through first breaks in each bc. */
{
  byteCodeChunk *bc;
  ENTER;

  for (bc=allocBCs; bc!=NULL; bc=bc->next)
    {
      KEEPONLY ((ExprPtr)bc); /* so that gc() does not destroy this behind our back. */
      if (sweepBreaks1 (fun, file, line, bc, func, data, stoponfirst))
	return true;
    }
  for (bc=oldBCs; bc!=NULL; bc=bc->next)
    {
      KEEPONLY ((ExprPtr)bc); /* so that gc() does not destroy this behind our back. */
      if (sweepBreaks1 (fun, file, line, bc, func, data, stoponfirst))
	return true;
    }
  QUIT;
  return false;
}

PRIVATE Boolean do_setbreak (bc, pc, data)
IN byteCodeChunk *bc;
IN OPCODE *pc;
INOUT char *data;
{
  (void) addBreak (bc, pc, &breakPoints, false);
  return false;
}

PRIVATE Boolean do_setbreak_on_entry (bc, pc, data)
IN byteCodeChunk *bc;
IN OPCODE *pc;
INOUT char *data;
{
  OPCODE op;

  op = opAt (pc);
  if (op==BC_FUNCTION)
    (void) addBreak (bc, pc, &breakPoints, false);
  return false;
}

PRIVATE COUNTER newBreak ()
{
  COUNTER nb;

  nb = breakPoints.nbreaks;
  if (nb>=breakPoints.sizebreaks)
    {
      COUNTER size;
      COUNTER *breaks;

      size = breakPoints.sizebreaks+16;
      breaks = (COUNTER *)MALLOC(size*sizeof(COUNTER));
      safememcpy (breaks, breakPoints.breaks, breakPoints.sizebreaks*sizeof(COUNTER));
      MARKFREE(breakPoints.breaks);
      breakPoints.breaks = breaks;
      breakPoints.sizebreaks = size;
    }
  breakPoints.breaks[nb] = breakPoints.n;
  breakPoints.nbreaks = nb+1;
  return nb;
}

PRIVATE COUNTER setBreak (fun, file, line)
IN CONST ExprPtr fun; /* function name, NOEXPR if not present */
IN CONST ExprPtr file; /* file name, NIL if toplevel, NOEXPR if not present */
IN INT line; /* line number, -1 if not present; at least one of the arguments should be present. */
{
  COUNTER nb;

  nb = newBreak ();
  if (line==-1) /* no line number given: then stop on all entry points to function. */
    (void) sweepBreaks (fun, file, line, do_setbreak_on_entry, NULL, false); /* false means not
										just the first entry
										point. */
  else
    (void) sweepBreaks (fun, file, line, do_setbreak, NULL, true); /* true means just the first breakpoint
								      on the specified line. */
  return nb;
}

PRIVATE int readLocation(snap, c, filep, namep, linep)
IN struct snapshot *snap;
IN int c; /* location is [function][@file][:line] */
OUT ExprPtr *filep, *namep;
OUT COUNTER *linep;
{
  Boolean got;

  *filep = NOEXPR;
  *namep = NOEXPR;
  *linep = -1;
  while (true)
    {
      c = skipBlanks (c);
      switch (c)
	{
	case '@': /* read file name */
	  if (*filep!=NOEXPR)
	    {
	      hfputs (message (748), hstderr);
	      (void) printSym (*filep, do_printfile, (char *)hstderr);
	      hfputs (message (749), hstderr);
	      hfflush (hstderr);
	      return skipLine ();
	    }
	  c = dgetc ();
	  c = skipBlanks (c);
	  if (c=='"')
	    {
	      c = getStringCst (c, filep, &got);
	      if (!got)
		*filep = NOEXPR;
	      else if (c=='"')
		c = dgetc ();
	    }
	  else
	    c = getId (c, filep, false, true);
	  if (*filep==NOEXPR)
	    {
	      hfprintf (hstderr, message(750), c);
	      hfflush (hstderr);
	      return skipLinec (c);
	    }
	  break;
	case ':': /* read line number */
	  if (*linep!=-1)
	    {
	      hfprintf (hstderr, message(751), (INT)*linep);
	      hfflush (hstderr);
	      return skipLine ();
	    }
	  c = dgetc();
	  c = skipBlanks (c);
	  c = getInt (c, linep, &got);
	  if (!got)
	    {
	      hfprintf (hstderr, message(752), c);
	      hfflush (hstderr);
	      *linep = -1;
	      return skipLinec (c);
	    }
#if OS==UNIX_SYS_V || OS==UNIX_BSD
	  --*linep;
#endif
	  break;
	case '\n':
	case -1:
	  return c;
	default:
	  if (*namep!=NOEXPR)
	    {
	      hfputs (message(753), hstderr);
	      (void) printSym (*namep, do_printfile, (char *)hstderr);
	      hfputs (message(754), hstderr);
	      hfflush (hstderr);
	      return skipLine ();
	    }
	  c = skipBlanks (c);
	  c = getId (c, namep, false, false);
	  if (*namep==NOEXPR)
	    {
	      hfprintf (hstderr, message(755), c);
	      hfflush (hstderr);
	      return skipLinec (c);
	    }
	  break;
	}
    }
}

PUBLIC Boolean dbgVetoInline (name, i_l)
IN CONST ExprPtr name;
IN CONST ExprPtr i_l;
{
#ifdef BUGGY
  /* not conservative enough:
     use "Utils/heapsort.ml"
     break();
     b heapsort
     c
     heapsort (op <) {1, 2, 3}
     does not work. */
  ExprPtr env, val;
  byteCodeChunk *bc;
  COUNTER i, j;

  if (i_l->kind==(CHAR)E_CONS)
    return breakPoints.n!=0; /* This is probably a locally defined function, on which
				we haven't possibly put a breakpoint yet, but we don't know:
				inline it provided there are no alive breakpoints. */
  i = IVAL(i_l);
  env = getTopEnv ();
  if (i<EVARS(env)->n)
    {
      val = EVARS(env)->expr[i];
      if (val->kind!=(CHAR)E_INTERPRETED_CLOSURE)
	return false;
      val = QENV(val);
      if (val->kind!=(CHAR)E_TUPLE_2)
	return false;
      bc = (byteCodeChunk *)CAR(val);
      for (j=0; j<breakPoints.n; j++)
	if (bc==(byteCodeChunk *)breakPoints.bcs[j])
	  return true; /* veto it: it is one of the function we have put a breakpoint on. */
    }
  return false;
#endif
#ifdef BUGGY_AS_WELL
  /* Counteexample:
     use "Utils/heapsort.ml"
     break();
     b heap
     c
     heapsort (op <) {1, 2, 3}
     does not stop on heap.
  */
  COUNTER j, nargs;
  struct snapshot snap;
  byteCodeChunk *bc;
  OPCODE *pc;
  Boolean gotBcs;
  ExprPtr e;

  for (j=0; j<breakPoints.n; j++)
    {
      bc = (byteCodeChunk *)breakPoints.bcs[j];
      pc = breakPoints.pcs[j]+1; /* because analyzeBc() subtracts 1 from pc. */
      gotBcs = analyzeBc (&snap, bc, pc, NULL, false, &nargs);
      if (!gotBcs)
	return true; /* if cannot identify function, then it might be the one we're trying to compile. */
      if (snap.name==name)
	return true;
    }
  return false;
#endif
  return breakPoints.n!=0; /* simply forbid any inlining as long as there is one active breakpoint. */
}

PRIVATE void showBreak (n)
IN COUNTER n;
{
  COUNTER k, kmax;
  byteCodeChunk *bc;
  OPCODE *pc;
  struct snapshot snap;
  Boolean gotBcs, first;
  MapPtr seen;
  ExprPtr current, lx;
  ExprPtr currt[3];
  COUNTER nargs;
  Boolean found;
  ENTER;

  k = breakPoints.breaks[n];
  if (n+1<breakPoints.nbreaks)
    kmax = breakPoints.breaks[n+1];
  else
    kmax = breakPoints.n;
  first = true;
  seen = NewMap ();
  for (; k<kmax; k++)
    {
      bc = (byteCodeChunk *)breakPoints.bcs[k];
      pc = breakPoints.pcs[k]+1; /* because analyzeBc() subtracts 1 from pc. */
      gotBcs = analyzeBc (&snap, bc, pc, NULL, false, &nargs);
      if (gotBcs)
	{
	  lx = KINT(snap.line);
	  currt[0] = snap.file;
	  currt[1] = snap.name;
	  currt[2] = lx;
	  current = KtupleGnrl(3, currt);
	  (void) KEEP(current);
	  (void) MapGet (current, seen, &found);
	  if (found)
	    continue;
	  seen = SetAdd (current, seen);
	  KEEPONLY(seen);
	  if (first)
	    {
	      hfprintf (hstderr, "%3d: ", n+1);
	      first = false;
	    }
	  else hfputs (message(756), hstderr);
	  if (snap.name!=NOEXPR)
	    (void) printSym (snap.name, do_printfile, (char *)hstderr);
	  if (snap.file!=NOEXPR && snap.file!=NIL)
	    {
	      hfputc ('@', hstderr);
	      (void) valPrintNoType (snap.file, do_printfile, (char *)hstderr, 0);
	    }
	  if (snap.line!=-1)
	    {
#if OS==UNIX_SYS_V || OS==UNIX_BSD
	      hfprintf (hstderr, ":%d", snap.line+1);
#else
	      hfprintf (hstderr, ":%d", snap.line);
#endif
	    }
	  hfputc ('\n', hstderr);
	}
    }
  hfflush (hstderr);
  QUIT;
}

PRIVATE void eraseBreak(k)
IN COUNTER k;
{
  COUNTER n1, n2, n, delta, i;
  OPCODE *ops, **pcs;
  ExprPtr *bcs;
  COUNTER *breaks;

  restoreBreaks (); /* instead of writing back all savedops between n1 and n2,
		       we restore all breaks, then we shall reinstall them
		       at the end of the code: the reason is that the latter
		       works even in the presence of duplicate breakpoints (i.e.,
		       when there are i<j such that pcs[i]==pcs[j]). */
  n = breakPoints.n;
  n1 = breakPoints.breaks[k];
  k++;
  n2 = (k<breakPoints.nbreaks)?breakPoints.breaks[k]:n;
  delta = n2-n1;
  breakPoints.n -= delta;
  for (i=n2, pcs = &breakPoints.pcs[n1], ops = &breakPoints.savedops[n1], bcs = &breakPoints.bcs[n1];
       i<n; i++, pcs++, ops++, bcs++)
    {
      pcs[0] = pcs[delta];
      ops[0] = ops[delta];
      bcs[0] = bcs[delta];
    }
  for (breaks = &breakPoints.breaks[k]; k<breakPoints.nbreaks; k++)
    breaks[-1] = breaks[0];
  breakPoints.nbreaks--;
  reinstallBreaks ();
}

PRIVATE Boolean installTmpBreaks (bcframes, bc, maxupp, retp)
IN CONST struct bcframes *bcframes;
IN CONST byteCodeChunk *bc;
INOUT COUNTER *maxupp;
IN Boolean retp;
{
  struct bcframe *frame;
  COUNTER n;

  (void) newBreak ();
  n = breakPoints.n;
  frame = bcframes->curBcFrame;
  if (retp)
    goto bump;
  while (true)
    {
      /* install all sorts of temporary breakpoints in the current function. */
      (void) sweepBreaks1 (NOEXPR, NOEXPR, -1, bc, do_setbreak, NULL, false);
      if (breakPoints.n!=n)
	/* managed to install a few breakpoints: return success. */
	return true;
      /* Otherwise, no breakpoint was found;
	 this is typical of the break() function, or of all functions
	 compiled in -nodebug mode. */
    bump:
      if (frame==bcframes->bcframes)
	{
	  bcframes = bcframes->parentBcFrames;
	  if (bcframes==NULL)
	    return false; /* didn't find proper caller to put breakpoints into. */
	  frame = bcframes->curBcFrame;
	}
      else --frame;
      bc = frame->bc;
      --*maxupp;
    }
}

PRIVATE void dbgError(s,c)
IN char *s;
IN int c;
{
  hfputs (s, hstderr);
  hfputc ('\'', hstderr);
  while (c!='\n' && c!=-1)
    {
      hfputc (c, hstderr);
      c = dgetc ();
    }
  hfputc ('\'', hstderr);
  hfputc ('\n', hstderr);
  hfflush (hstderr);
}

PRIVATE void parseError(c)
IN int c;
{
  dbgError(message(740), c);
}

PUBLIC OPCODE DoBreak(bc, pc, top, bcframes, stacks, exc)
IN CONST byteCodeChunk *bc;
IN CONST OPCODE *pc;
IN CONST ExprPtr *top;
IN CONST struct bcframes *bcframes;
IN CONST struct bcstacks *stacks;
IN CONST ExprPtr exc;
{
  COUNTER up, maxup;
  struct snapshot snap;
  Boolean gotBcs, cont, show, view;
  Boolean negated;
  OPCODE op;
  /*int c;*/
  ENTER;

  maxup = nBcs (bcframes)-1;
  switch (debugMode)
    {
    case DBM_NONE:
      debugMode = DBM_IN_DEBUGGER;
      break;
    case DBM_IN_DEBUGGER:
      return BC_FUNCTION; /* occurs when we ask to evaluate an expression from
			     inside the debugger (shouldn't happen). */
    case DBM_STEPPING:
    case DBM_NEXTING:
      if (maxup>debugNframes && exc==NOEXPR)
	goto debug_quit; /* if maxup>debugNframes, then we have tried to step or
			    to next, but we caught a breakpoint at the wrong
			    stack depth: return as though there had been no breakpoint
			    here.
			    Note that it may be the case that maxup<debugNframes, when
			    we do 'n' and a return statement gets executed, or an
			    exception is thrown. */
      debugMode = DBM_IN_DEBUGGER;
      eraseBreak (breakPoints.nbreaks-1);
      break;
    case DBM_STEPPING_EXC:
    case DBM_NEXTING_EXC: /* in these two cases, we have no temporary breakpoint to erase,
			     contrarily to the DBM_STEPPING and DBM_NEXTING cases. */
      if (maxup>debugNframes && exc==NOEXPR)
	goto debug_quit; /* see comment for DBM_NEXTING. */
      debugMode = DBM_IN_DEBUGGER;
      break;
    }
  up = 0;
  if (exc!=NOEXPR) /* about to raise an exception, or exception just raised inside
		    a primitive. */
    {
      struct prtstrmax *data = (struct prtstrmax *)alloctmp(sizeof(struct prtstrmax));
      CHAR *s;

      s = (CHAR *)message (734);
      hfputs ((char *)s, hstderr);
      data->maxlen = IVAL(REFVAL(rightMarginRef)) - strlen ((char *)s) - 4;
      data->len = 0;
      valPrint (exc, exnType, do_prtmax, (char *)data, 0);
      if (data->len>=data->maxlen)
	hfputs(message(505),hstderr);
      hfputc ('\n', hstderr);
      hfflush (hstderr);
    }
  show = true;
  view = true;
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfputc('\n',coreTraceFile);
      hfflush(coreTraceFile);
    }
#endif
  for (cont = true; cont; )
    {
      int c;
      Boolean contl;

      QUIT;
      if (show)
	{
	  if ((gotBcs = analyzeBcs (&snap, bc, pc, top, bcframes, stacks, up)))
	    {
	      TZone tz;

	      showStack (bc, pc, top, bcframes, stacks, up, 1, maxup);
	      tz.line1 = tz.line2 = snap.line;
	      tz.pos1 = tz.pos2 = snap.pos;
	      showSource (snap.file, &tz, hstderr);
	    }
	  show = false;
	}
      if (view)
	dbgView (hstderr, &snap, monitorExprs, false, true);
      view = true;
#ifndef HAS_READLINE
      hfputs(message(90) /*"(debug) "*/,hstderr);
#endif
      hfflush(hstderr);
      stopwanted = false;
      contl = true;
      for (;contl;)
	{
	  c = dgetc();
	  if (c==-1)
	    {
#ifdef EINTR
	      if (errno==EINTR)
		{
		  contl = false;
		  continue;
		}
#endif
	      goto quit;
	    }
	  if (c==' ' || c=='\t' || c=='\f')
	    continue;
	  negated = false;
	again:
	  switch (c)
	    {
	    case 'c': /* continue */
	      defaultCommand = c;
	      (void) skipLine();
	    do_cont:
	      cont = false;
	      contl = false;
	      break;
	    case 'q': /* quit evaluation, and return to toplevel */
	    quit:
	      debugMode = DBM_NONE;
	      Toplevel(1); /* with an error code, so that
			      declarations are not installed */
	      break;
	    case 'w':
	      defaultCommand = c;
	      {
		int c;
		COUNTER mw;
		Boolean got;

		c = skipBlanks(skipWord());
		c = getInt(c,&mw,&got);
		(void) skipLinec(c);
		if (!got)
		  {
		  do_where:
		    mw = maxWhere;
		  }
		showStack(bc,pc,top,bcframes,stacks,up,mw,maxup);
		contl = false;
		break;
	      }
	    case 'u':
	    case 'd':
	      defaultCommand = c;
	      {
		int c;
		COUNTER move;
		Boolean got;

		c = skipBlanks(skipWord());
		c = getInt(c,&move,&got);
		if (!got)
		  {
		  do_up_or_down:
		    move = 1;
		  }
		if (defaultCommand!='u')
		  move = -move;
		show = true;
		up += move;
		if (up<0)
		  up = 0;
		else if (up>maxup)
		  up = maxup;
		contl = false;
		break;
	      }
	    case 'p':
	      defaultCommand = ' ';
	      {
		int c;
		ExprPtr codes;

		c = skipBlanks(skipWord());
		c = parseValueList(c,&codes);
		if (codes==NOEXPR)
		  parseError (c);
		else
		  {
		    codes = Reverse (codes);
		    (void) KEEP (codes);
		    dbgView (hstderr, &snap, codes, true, false);
		    if (codes!=NIL && CDR(codes)==NIL)
		      displayPossibleSelectors(&snap,CAR(codes),3);
		  }
		contl = false;
		break;
	      }
	    case '.': case '^': case '[':
	      if (lastPrintExpr==NIL)
		dbgError ("No 'p' or 'm' command to refine",c);
	      else
	      {
		ExprPtr code;

		defaultCommand = ' ';
		c = parseValue(c,lastPrintExpr,&code);
		if (code==NOEXPR)
		  parseError (c);
		else
		  {
		    dbgView (hstderr, &snap, KONS(code,NIL), true, false);
		    displayPossibleSelectors(&snap,code,3);
		  }
		contl = false;
	      }
	      break;
	    case 'm':
	      defaultCommand = ' ';
	      {
		int c;
		ExprPtr codes;

	      do_monitor:
		c = skipBlanks(skipWord());
		c = parseValueList(c,&codes);
		if (codes==NOEXPR)
		  {
		    if (c=='\n')
		      {
			dbgView (hstderr, &snap, monitorExprs, true, true);
			view = false;
		      }
		    else
		      parseError (c);
		  }
		else
		  monitorExprs = negated?delMonitorExprs (codes, monitorExprs)
		    :addMonitorExprs (codes, monitorExprs);
		if (!negated && codes!=NIL && codes!=NOEXPR && CDR(codes)==NIL)
		  displayPossibleSelectors(&snap,CAR(codes),3);
		contl = false;
		break;
	      }
	    case '$': /* set a debugger variable */
	      {
		ExprPtr x, code, val, type;
		int c;

	      do_setvar:
		defaultCommand = ' ';
		c = '$';
		c = getId(c,&x,true,false);
		c = skipBlanks(c);
		if (negated)
		  {
		    setMapRemove(x,&lastPrintTypes);
		    setMapRemove(x,&lastPrintVals);
		  }
		else if (c!='=')
		  {
		    if (c=='\n')
		      { /* print the debugger variable instead, unless
			   it is just '$', in which case print all
			   debugger variables */
			if (SLEN(x)==1)
			  printDebugVars();
			else
			  {
			    dbgView (hstderr, &snap, KONS(x,NIL), true, false);
			    displayPossibleSelectors(&snap,x,3);
			  }
		      }
		    else
		      parseError(c);
		  }
		else
		  {
		    c = skipBlanks(dgetc());
		    c = parseValue(c,NOEXPR,&code);
		    if (code==NOEXPR)
		      parseError(c);
		    else
		      {
			val = dbgEval (&snap,code,&type);
			if (val!=NOEXPR)
			  {
			    setMapPut(x,val,&lastPrintVals);
			    if (type!=NOEXPR)
			      setMapPut(x,type,&lastPrintTypes);
			  }
			else parseError(c);
		      }
		  }
		contl = false;
		break;
	      }
	    case '-': /* erase some information. */
	      defaultCommand = ' ';
	      {
		int c;

		c = dgetc ();
		c = skipBlanks (c);
		negated = true;
		switch (c)
		  {
		  case 'm': goto do_monitor;
		  case 'x': goto do_checkx;
		  case 'b': goto do_break;
		  case '$': goto do_setvar;
		  default:
		    goto help;
		  }
		break;
	      }
	    case 'x':
	      defaultCommand = ' ';
	      {
		int c;
		ExprPtr codes;

	      do_checkx:
		c = skipBlanks(skipWord());
		c = parseValueList(c,&codes);
		if (codes==NOEXPR)
		  parseError (c);
		else
		  checkExceptions (&snap, codes, !negated, true);
		contl = false;
		break;
	      }
	    case 'l':
	      defaultCommand = ' ';
	      {
		int c;
		ExprPtr code;

		c = skipBlanks(skipWord());
		c = parseValue (c,NOEXPR,&code);
		if (code==NOEXPR)
		  {
		    if (dbgName==NOEXPR || dbgName==NIL)
		      {
			parseError (c);
			contl = false;
			break;
		      }
		    else code = dbgName;
		  }
		{
		  ExprPtr fun, name, type;
		  TZone zone;

		  if (code->kind==(CHAR)E_STRING)
		    dbgName = code;
		  fun = dbgEval (&snap, code, &type);
		  if (fun!=NOEXPR)
		    {
		      name = getFileName (fun, &zone);
		      if (name!=NOEXPR)
			{
			  dbgFile = name;
			  showSource (name, &zone, hstderr);
			}
		    }
		}
		contl = false;
		break;
	      }
	    case 'b':
	      defaultCommand = c;
	      {
		int c;

	      do_break:
		c = skipWord ();
		if (negated)
		  {
		    COUNTER nbrk;
		    Boolean got;

		    c = skipBlanks (c);
		    c = getInt(c, &nbrk, &got);
		    if (got && nbrk>=1 && nbrk<=breakPoints.nbreaks)
		      {
			eraseBreak(nbrk-1);
			goto showbreaks;
		      }
		    hfprintf (hstderr, message(757));
		    (void) skipLinec (c);
		  }
		else
		  {
		    ExprPtr file, name;
		    COUNTER line;
		    COUNTER n;

		    c = readLocation (&snap, c, &file, &name, &line);
		    if (file==NOEXPR && name==NOEXPR && line==-1)
		      {
		      showbreaks:
			(void) skipLinec (c);
		      do_showbreaks:
			for (n=0; n<breakPoints.nbreaks; n++)
			  showBreak (n);
		      }
		    else
		      {
			if (name!=NOEXPR)
			  {
			    dbgName = name;
			    if (file==NOEXPR) /* if name is given but file is not, then do not
						 use default dbgFile. */
			      dbgFile = NOEXPR;
			  }
			if (file!=NOEXPR)
			  dbgFile = file;
			n = setBreak (dbgName, dbgFile, line);
			if (breakPoints.breaks[n]==breakPoints.n) /* no breakpoint set */
			  {
			    hfputs (message(757), hstderr);
			    breakPoints.nbreaks--;
			  }
			else
			  {
			    showBreak (n);
			  }
		      }
		  }
		contl = false;
		break;
	      }
	    case 's':
	      defaultCommand = c;
	      (void) skipLine();
	    do_step:
	      debugMode = DBM_STEPPING;
	      goto do_step_or_next;
	    case 'n':
	      defaultCommand = c;
	      (void) skipLine();
	    do_next:
	      debugMode = DBM_NEXTING;
	    do_step_or_next:
	      defaultCommand = c;
	      op = opAt (pc-1);
	      if (exc!=NOEXPR)
		{ /* was about to launch an exception: simply return, without installing
		   any temporary breakpoint: bcEval0() will monitor the entry into the handler,
		   as part of dealing with the BC_NEXTHANDLE opcode. */
		  debugMode += 2; /* DBM_STEPPING -> DBM_STEPPING_EXC, DBM_NEXTING -> DBM_NEXTING_EXC. */
		  return BC_FUNCTION;
		}
	      switch (op)
		{
		case BC_RAISE: /* just raise the exception, and let bcEval0() monitor
				  entries into the handler as part of dealing with BC_NEXTHANDLE. */
		  break;
		case BC_BREAK: /* Should be impossible, as bytecode compiler does not generate
				  BC_BREAK opcode. */
		  op = BC_FUNCTION;
		  break;
		case BC_RETURN:
		  debugNframes = maxup;
		  if (!installTmpBreaks (bcframes, bc, &debugNframes, true))
		    {
		      /* Try to install temporary breakpoints inside caller.
			 If impossible, then just do as if we has typed 'c'. */
		      debugMode = DBM_NONE;
		      breakPoints.nbreaks--;
		    }
		  break;
		case BC_CALL:
		  if (debugMode==DBM_STEPPING && top[-2]->kind==(CHAR)E_INTERPRETED_CLOSURE)
		    {
		      /* Try to install breakpoints in callee. */
		      COUNTER n;

		      (void) newBreak ();
		      n = breakPoints.n;
		      (void) sweepBreaks1 (NOEXPR, NOEXPR, -1, (byteCodeChunk *)CAR(QENV(top[-2])),
					   do_setbreak, NULL, false);
		      if (breakPoints.n!=n) /* found some breakpoints in callee: OK. */
			{
			  debugNframes = maxup+1; /* only catch breakpoints at current stack depth
						     plus one.  This avoids the following problem.
						     Take a recursive function f: doing n(ext)
						     will install breakpoints throughout f, so
						     execution will stop not on return from f,
						     but on entry into the called instance of f.
						     We avoid this by comparing debugNframes
						     with maxup when entering DoBreak(): if they
						     are not equal, then we have not stopped on
						     the right instance, and we c(ontinue)
						     immediately. */
			  break;
			}
		      /* otherwise, did not find any breakpoints in callee, so do as though
			 we had typed 'n' and fall through. */
		    }
		  /*fallthrough*/
		default:
		  debugNframes = maxup;
		  if (!installTmpBreaks (bcframes, bc, &debugNframes, false))
		    {
		      /* Try to intall temporary breakpoints inside current function.
			 If impossible, then just do as if we has typed 'c'. */
		      debugMode = DBM_NONE;
		      breakPoints.nbreaks--;
		    }
		  if (op==BC_NEXTHANDLE)
		    op = BC_NEXTHANDLE_NODEBUG;
		  break;
		}
	      return op;
	    case '\n': /* repeat last command */
	      switch (c = defaultCommand)
		{
		case 'c': goto do_cont;
		case 'w': goto do_where;
		case 'u':
		case 'd': goto do_up_or_down;
		case 'b': goto do_showbreaks;
		case 's': goto do_step;
		case 'n': goto do_next;
		default: contl = false; break;
		}
	      break;
	    case 'h':
	    default:
	    help:
	      c = skipWord ();
	      c = skipBlanks (c);
	      contl = false;
	      if (c=='\n')
		hfputs(message(795),hstderr);
	      else
		{
		  (void) skipLine();
		  switch (c)
		    {
		    case 'h': hfputs (message(795),hstderr); break;
		    case 's': hfputs (message(950),hstderr); break;
		    case 'n': hfputs (message(951),hstderr); break;
		    case 'c': hfputs (message(952),hstderr); break;
		    case 'p': hfputs (message(953),hstderr); break;
		    case 'm': hfputs (message(954),hstderr); break;
		    case 'x': hfputs (message(955),hstderr); break;
		    case 'b': hfputs (message(956),hstderr); break;
		    case 'w': hfputs (message(957),hstderr); break;
		    case 'u': hfputs (message(958),hstderr); break;
		    case 'd': hfputs (message(959),hstderr); break;
		    case 'l': hfputs (message(960),hstderr); break;
		    case 'q': hfputs (message(961),hstderr); break;
		    case '$': hfputs (message(962),hstderr); break;
		    }
		}
	      hfflush(hstderr);
	      break;
	    }
	}
    }
  /*debug_exit:*/
  debugMode = DBM_NONE;
    debug_quit:
  op = BC_FUNCTION;
  if (exc==NOEXPR)
    {
      op = opAt (pc-1);
      if (op==BC_BREAK)
	op = BC_FUNCTION; /* if DoBreak() was not entered on exception handling,
			     then this was on BC_BREAK or BC_NEXTHANDLE.
			     Normally, BC_BREAK is
			     the result of setting a breakpoint, so return the
			     saved opcode to be executed when we return from here
			     (except if op==BC_BREAK, which is abnormal). */
      if (op==BC_NEXTHANDLE)
	op = BC_NEXTHANDLE_NODEBUG;
    }
  return op;
}

#endif
