/* Definitions for debugger and profiler.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#if DEBUGGING

#define DBG_BEGIN 1
/* BEGIN <pc1> <off> : starts a new scope (let, function, etc.) for all
   opcodes from pc=pc1 until pc=pc2 (excluded, where pc2 is given by
   the END instruction off opcodes later). */
#define DBG_BEGIN_BIG 2
/* BEGIN_BIG <pc1hi> <pc1lo> <off> : same as BEGIN, for large pcs. */
#define DBG_END 3
/* END <pc2> : marks the end of a scope. */
#define DBG_END_BIG 4
/* END_BIG <pc2hi> <pc2lo> */
#define DBG_WHERE 5
/* WHERE <pc> <s> <m> <x> : states that following info describes the state
   at pc, with current stack, map stack and exception stack levels at
   s, m, x respectively.
   There is a WHERE (or WHERE_BIG) info for each codepoint where the
   program can be stopped (mainly on BC_FUNCTION, BC_LOOP and BC_RETURN
   opcodes, but also anywhere else). */
#define DBG_WHERE_BIG 6
/* WHERE_BIG <pchi> <pclo> <s> <m> <x> */
#define DBG_ACCESS 7
/* ACCESS <n> : direct attention to object at position n on the stack. */
#define DBG_VARREF 8
/* VARREF <n> <l> <j> : direct attention to object at position j
   in environment at level l from environment found at position
   n in the stack. */
#define DBG_DEFNAME 9
/* DEFNAME <n> <t> : define identifier (whose name is at &rawdata[n],
   and whose type is consts[t], as being the object where our attention
   has just been directed to. */
#define DBG_DEFCON 10
/* same as DEFNAME but for constructors. */
#define DBG_DEFEXC 11
/* same as DEFNAME but for exceptions. */
#define DBG_FILE 12
/* FILE <f> : states that we've moved to file named consts[f].
   Such info is local to a BEGIN...END block (even BIG), and is only
   produced just before WHERE(_BIG) ops, and when it changes. */
#define DBG_LINE 13
/* LINE <n> <pos> : states that we've moved to line #n, character position pos.
   This info is local to a BEGIN...END block (even BIG). */
#define DBG_NAME 14
/* NAME <n>: states that we've moved to a function named consts[n].
   This info is local the a BEGIN...END block (even BIG). */
#define DBG_ENV 15
/* ENV <n>: states that PARENT_DBGNAMES info is to be found in consts[n]. */
#define DBG_STOP 16
/* STOP: end list of debug instructions. */

#define DBM_NONE 0
#define DBM_IN_DEBUGGER 1
#define DBM_STEPPING 2
#define DBM_NEXTING 3
#define DBM_STEPPING_EXC 4
#define DBM_NEXTING_EXC 5

EXTERN COUNTER debugMode;

/* In the new debugging scheme, we also have E_DEBUGDECL declarations
   of the form E_DEBUGDECL (filename, line) [like E_COUNTED]
   */

struct bcframes;
struct bcstacks;

IMPORT OPCODE DoBreak P((IN CONST struct byteCodeChunk *bc,
			 IN CONST OPCODE *pc,
			 IN CONST ExprPtr *top,
			 IN CONST struct bcframes *bcframes,
			 IN CONST struct bcstacks *stacks,
			 IN CONST ExprPtr exc));

IMPORT Boolean dbgVetoInline P((IN CONST ExprPtr name, IN CONST ExprPtr i_l));
/* This is used in compiling, to solve the following problem.
   Assume that you put a breakpoint on some function f.
   Then the compiler does not know this, and may want to inline
   f; but then your breakpoint has disappeared from the inlined
   version.  So the compiler first asks the debugger, using
   dbgVetoInline(), whether it is OK to inline the function (in
   this case, no).
   Note that, otherwise, the debugger knows perfectly well how
   to put a new breakpoint on all the versions of a given function
   f, even the previously inlined ones.
*/

IMPORT void restoreBreaks P ((void));
IMPORT void reinstallBreaks P ((void));

/* A code expression instrumented for debugging is:
   E_DEBUGZONE (car = code, cdr = BUILD_HEADER(zone,A_CODEPOINT,args)
   where args = [output type, file name].

   (Used to be:
   where type is the number of the codepoint,
   and args = [(trigger on entry,trigger on exit),output type,file name]
   where triggers are functions that must evaluate to trueValue
   when applied on () (on entry) or on the result (on exit)
   if we want a breakpoint (only when type!=DZ_INACTIVE)
   and file name is the name of the file where the code was defined.
   )

   E_DEBUGNAME (car = pat, cdr = counted(n,Let(bind=name,body=type)))
   E_DEBUGNAME is a pattern matcher. When matching it against a value,
   does the same as matching pat against the value, but also store
   Let(name,type) in slot n of the current environment.
   When pat is a varintro(i), then n=i (OLDDEBUG: i+1), so that
   MatchValue (in eval.c) stores the value in slot #i of the array
   of locals (vp) (OLDDEBUG:, and stores the cdr in slot #i+1). This way,
   names are recorded at run time with a small cost. (OLDDEBUG: When trying
   to determine if a vp has been filled up with debugging info,
   it is enough to examine if vp->n is even and if every odd entry
   in it is of kind E_LET (which cannot be the kind of any GimML value).
   Actually, the kind may also be E_APPL (datatype constructor),
   or E_GOTO (reserved).)
*/

EXTERN ExprPtr monitorExprs;
EXTERN ExprPtr lastPrintExpr;
EXTERN MapPtr lastPrintVals, lastPrintTypes;
EXTERN ExprPtr dbgFile;
EXTERN ExprPtr dbgName;

struct breakpoints {
  COUNTER n, size; /* number of breakpoints in pcs and savedops,
		  max-number of same. */
  OPCODE **pcs;
  OPCODE *savedops;
  ExprPtr *bcs;
  COUNTER nbreaks, sizebreaks; /* number of logical breakpoints (each logical breakpoint
				  may correspond to several actual breakpoints, because
				  of inlining;
				  and max-number of same. */
  COUNTER *breaks; /* k_0, ..., k_{nbreaks-1}: letting k_nbreaks be n,
		      logical breakpoint #i includes all actual breakpoints
		      >=k_i and <k_{i+1}. */
};

EXTERN struct breakpoints breakPoints;

IMPORT ExprPtr debugger P((IN CONST ExprPtr info,IN CONST ExprPtr val,IN COUNTER status));

#endif

#if PROFILING

/*
   We define expressions of the form E_PROFILE (car = expr,
   cdr = BUILD_HEADER(zone,type,args)),
   where type may be EA_PROFILE or EA_CLR_PROFILE,
   and args = funcname :: info :: filename
   where info = NIL if no profile information is recorded (EA_CLR_PROFILE),
   or info is an external containing a struct Profile otherwise.
*/

#define prof_tms h_tms

typedef struct Prof1Rec {
  struct prof_tms time;
  struct h_tms gc_time;
  long gc_count;
} Prof1Rec;

typedef struct ProfRec {
  Prof1Rec total; /* total times and counts including called subfunctions */
  Prof1Rec proper; /* times and counts for the function proper */
} ProfRec;

typedef struct Profile {
  INT nCalls; /* number of calls to the function */
  ProfRec stat;
  /* auxiliary data: */
  INT nEntries; /* number of simultaneous active copies of expression */
  ProfRec last;
} Profile;

#define EA_PROFILE (-10)
#define EA_CLR_PROFILE (-11)

EXTERN extDesc *profExtDecs;
IMPORT ExprPtr profAlloc P((void));
IMPORT void resetProfile P((INOUT struct Profile *prof));
#endif
