/* Basic definitions for the whole GimML system (not only the interpreter).
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "portable.h"
/* #include "bignum.h" */
#include <setjmp.h>
#ifdef HAS_STDARG
#include <stdarg.h>
#else
#ifdef HAS_VARARGS
#include <varargs.h>
#endif
#endif

#define HEAP

#define CODE unsigned long

typedef struct TZone {
  unsigned short line1,line2;
  unsigned short pos1,pos2;
} TZone;

struct build_header {
  CHAR kind; /* = BUILD_HEADER_KIND */
  CHAR mark;
  SHORT type;
  TZone z;
  struct Expression *args;
#if BH_PADDING!=0
  char padding[BH_PADDING];
#endif
};

#define ZONE(expr) (&((struct build_header *)expr)->z)
#define TYPE(expr) ((struct build_header *)expr)->type
#define ARGS(expr) ((struct build_header *)expr)->args

typedef enum ExprKind {
/* 0x00 */ E_FREE, E_NIL, E_INT,
/* 0x03 */ E_REAL, E_COMPLEX, E_STRING, E_CONS,
/* 0x07 */ E_EMPTY, E_MAPLET, E_MAP,
/* 0x0a */ E_TUPLE_0, E_TUPLE_1, E_TUPLE_2, E_TUPLE_GNRL,
/* 0x0e */ E_PAT_REVTUPLE_2,
/* 0x0f */ E_PAT_TUPLE_GNRL_CONS, E_PAT_REVTUPLE_GNRL_CONS,
/* 0x11 */ E_PAT_RECORD_GNRL_CONS, E_PAT_REVRECORD_GNRL_CONS,
/* 0x13 */ E_PAT_RECORD_MAPADD, E_PAT_CONS,
/* 0x15 */ E_RECORD_0, E_RECORD_1, E_RECORD_GNRL,
/* 0x18 */ E_PATCHOICE,
/* 0x19 */ E_EXCEPTION,
/* 0x1a */ E_COMPILED_CLOSURE, E_INTERPRETED_CLOSURE, OBSOLETE_E_MEMO_CLOSURE/*FREE*/,
/* 0x1d */ E_POLYFUN, E_MEMO_FN, E_DATACON,
/* 0x20 */ E_NUMTYPE, E_TYPEFX, E_QUOTE, E_REF,
/* 0x24 */ E_DATA, OBSOLETE_E_ARRAY_1/*FREE*/, E_ARRAY_2, E_ARRAY_GNRL,
/* 0x28 */ E_VARINTRO, E_VARREF, E_DEBUGNAME, /* E_DEBUGNAME is reserved for debugging */
/* 0x2b */ E_ENV, E_LET,
/* 0x2d */ E_APPL, E_GOTO, E_TYPE_APPL,
/* 0x30 */ E_PAT_ANY, E_PAT_AS, E_PAT_CHECK, E_PAT_PACK,
/* 0x34 */ E_PAT_MAPADD, E_PAT_CARD_EQ, E_PAT_CARD_GEQ,
/* 0x37 */ E_PAT_REVCONS,
/* 0x38 */ E_GET, E_SET, E_MAPGET,
/* 0x3b */ E_TSELECT, E_TSEL1, E_TSEL2, E_TSELGNRL,
/* 0x3f */ E_RSELECT, E_RSEL1, E_RSELGNRL,
/* 0x42 */ E_RMU1, E_UNSHARED_RMU1, E_RMUX, E_UNSHARED_RMUX,
/* 0x46 */ E_IF, E_PROG, E_AND, E_OR,
/* 0x4a */ E_PROG_OF, E_LIST_OF, E_UNSHARED_LIST_OF, E_SET_OF, E_APPEND_OF,
/* 0x4f */ E_UNSHARED_APPEND_OF, E_UNION_OF, E_UNION_UNDER_OF, E_INTER_OF,
/* 0x53 */ E_MAP_OF, E_MAP_UNDER_OF, E_ALL_OF, E_EXISTS_OF, E_SOME_OF, E_UNSHARED_SOME_OF,
/* 0x59 */ E_PROG_MAP, E_LIST_MAP, E_UNSHARED_LIST_MAP, E_SET_MAP, E_APPEND_MAP,
/* 0x5e */ E_UNSHARED_APPEND_MAP, E_UNION_MAP, E_UNION_UNDER_MAP, E_INTER_MAP,
/* 0x62 */ E_MAP_MAP, E_MAP_UNDER_MAP, E_ALL_MAP, E_EXISTS_MAP, E_SOME_MAP, E_UNSHARED_SOME_MAP,
/* 0x68 */ E_PROG_WHILE, E_LIST_WHILE, E_UNSHARED_LIST_WHILE, E_SET_WHILE, E_APPEND_WHILE,
/* 0x6d */ E_UNSHARED_APPEND_WHILE, E_UNION_WHILE, E_UNION_UNDER_WHILE, E_INTER_WHILE,
/* 0x71 */ E_MAP_WHILE, E_MAP_UNDER_WHILE, E_ALL_WHILE, E_EXISTS_WHILE, E_SOME_WHILE,
/* 0x76 */ E_UNSHARED_SOME_WHILE, E_HANDLE, E_RAISE,
/* 0x79 */ E_DELAY, E_STACK, E_PACK,
/* 0x7c */ E_EXTERNAL,
	/* kinds reserved for instrumenting the code: */
/* 0x7d */ E_DEBUGZONE, E_DEBUGDECL,
/* 0x7f */ E_PROFILE,
	/* special evaluation kinds for non-sharing allocation: */
/* 0x80 */ E_UNSHARED_CONS,
/* 0x81 */ E_UNSHARED_TUPLE_1, E_UNSHARED_TUPLE_2, E_UNSHARED_TUPLE_GNRL,
/* 0x84 */ E_UNSHARED_RECORD_1, E_UNSHARED_RECORD_GNRL,
/* 0x86 */ E_OVERWRITE, E_UNDERWRITE,
	/* kinds of objects that are not Expressions: */
/* 0x88 */ TYPE_ASSUM_KIND, TYPE_STATE_KIND, COMPILE_ASSUM_KIND, BUILD_HEADER_KIND,
/* 0x8c */ E_SLOT_BRANCH, E_COUNTED,
/* 0x8e */ E_EQUAL, E_NEQUAL, E_NOT,
/* 0x91 */ E_BYTECODE, E_N_APPL,/*FREE*/ E_N_LAMBDA,/*FREE*/
/* 0x94 */ E_UNUSED
} ExprKind;

#define E_VAR E_STRING

#define MARK_SHARED 0x40
#define MARK_IN_CAR 0x04
#define MARK_IN_CDR 0x08
#define MARK_IN_REF 0x10
#define MARK_IN_DATA 0x20
#define MARK_REF_COUNT 0x03
#define MARK_GC 0x80

#define GC_MARK(e) (e)->mark |= MARK_GC
#define GC_UNMARK(e) (e)->mark &= ~MARK_GC
#define GC_MARKED(e) ((e)->mark & MARK_GC)

#define NOEXPR (ExprPtr)0L

typedef CONST void (*printFunc) P((IN CONST CHAR *text,
				   IN COUNTER len,
				   INOUT char *printdata));
#define PRINTFUNC(fun) \
PRIVATE void fun P((IN CONST CHAR *text,IN COUNTER len, \
		    INOUT char *_data)); \
PRIVATE void fun(text,len,_data) \
IN CONST CHAR *text; \
IN COUNTER len; \
INOUT char *_data;
#define PRINTDATA(type) ((type)_data)

typedef CONST void (*readFunc) P((OUT CHAR *text,
				  IN COUNTER len,
				  INOUT char *readdata));
#define READFUNC(fun) \
PRIVATE void fun P((INOUT CHAR *text,IN COUNTER len, \
		    INOUT char *_data)); \
PRIVATE void fun(text,len,_data) \
INOUT CHAR *text; \
IN COUNTER len; \
INOUT char *_data;
#define READDATA(type) ((type)_data)

typedef struct StackChunk   {
  struct Expression *protected;
  struct excHandler *excStack;
  char *contents; /* NULL if still on stack */
  char *top; /* overestimated top of stack after capture of continuation */
  char *alloctmp_blocks;
  INT age;
} StackChunk;

typedef struct Vector {
  COUNTER n;
  struct Expression *expr[1];
} Vector,*VecPtr;

#define VECSIZE(n) VARSIZE(Vector,expr,n)

typedef enum doExtTag {
  X_MARK, /* to mark external data; used by GCmark1() */
  X_FREE, /* to free external data; used by GCsweep1() */
  X_MARKDATA, /* to put reference counts on data; used by markData() */
  X_SAVE, /* to save structure to disk or to transmit it over a network */
} doExtTag;

typedef void (*markAction) P((INOUT struct Expression *expr));
typedef struct Expression *(*loadAction) P((INOUT struct Expression *sharedref));

typedef struct doExtDesc {
  doExtTag xtag;
  markAction action;
} doExtDesc;

typedef struct extDesc {
  CHAR *name; /* for identification, and printing when no print method is provided */
  CHAR id[4]; /* for short identification */
  struct extDesc *next;
  unsigned long (*hash) P((IN CONST char *extdata)); /*  hash code function if data must be shared */
  Boolean (*eq) P((IN CONST char *ext1,IN CONST char *ext2)); /* comparison function if data must be shared */
  void (*print) P((IN CONST char *ext,
		   IN printFunc doprint,
		   INOUT char *printdata)); /* print function; may be NULL */
  void (*sweep) P((INOUT char *extdata,IN CONST doExtDesc *doext));
	   /* must do action specified by the xtag on the external data recursively
	   	and call (*action)() on all GimML objects that are in the external */
  struct Expression *(*load) P((IN CONST struct extDesc *desc,IN Boolean shared,
				INOUT struct Expression *sharedref,
				IN CONST loadAction load));
} extDesc,*extDescPtr;

typedef CONST struct Expression *(*compiledAction) P((struct Expression *env, ...));

typedef struct stackDesc {
  struct gcStackView *stack_frames,*heap_frames;
  jmp_buf buf;
} stackDesc;
#define SETJUMP(sd) ((sd).stack_frames = theStackFrames, (sd).heap_frames = theHeapFrames, SETJMP((sd).buf))
#define LONGJUMP(sd,n) (theStackFrames = (sd).stack_frames, theHeapFrames = (sd).heap_frames, LONGJMP((sd).buf,n))

#define OPCODE SHORT

typedef struct Expression {
  /*ExprKind*/
  CHAR kind;
  CHAR mark;
  SHORT magic; /* used in particular for fast switching on datatypes
		  (E_DATA, E_DATACON);
		  stores arity of functions (E_POLYFUN).
	       */
#if EXPR_GAP!=0
  CHAR filler[EXPR_GAP];
#endif
  struct Expression *link;
  union e_what {
    struct Int {
      INT val;
      struct Expression *next;
    } i;
    struct Str {
      INT len;
      CHAR *HEAP text;
    } str;
    struct Pair {
      struct Expression *car,*cdr;
    } pair;
    struct Quote {
      struct Expression *ref;
      INT n;
    } quote;
    struct Array { /* represents arrays (n>=3), tuples (n>=3), records */
      struct Expression *shape; /* map field => index for records; NIL for others */
      Vector *HEAP val;
    } array;
    struct Var {
      COUNTER rank,level;
    } var;
    struct Compiled {
      compiledAction action;
      struct Expression *env; /* closure environment, or any expression in the closure */
    } compiled;
    struct Handler {
      struct Expression *eh; /* cons(env,handlers) or NIL */
      stackDesc *HEAP jb;
    } handler;
    struct Stack {
      StackChunk *chunk; /* a real chunk (current stack) */
      stackDesc *HEAP jb;
    } stack;
    struct ByteCode {
      struct byteCodeChunk *chunk;
      struct Expression *env;
    } bytecode;
    struct  External { /* external data structures unknown to the GimML runtime */
      extDescPtr desc;
      char *data;
    } external;
    REAL x;
  } what;
#if EXPR_PADDING!=0
  char padding[EXPR_PADDING];
#endif
} Expression,*ExprPtr;

#define VEXPRPTR Expression *VOLATILE

EXTERN ExprPtr nihil;
#undef NIL
#define NIL nihil

#define Map Expression
#define MapPtr ExprPtr
#define Symbol Expression
#define SymPtr ExprPtr
#define UNIT tuple0

#define ISCONS(e) (/*(e)!=NIL && */(e)->kind==(CHAR)E_CONS)
#define ISMAP(e) (/*(e)!=NIL && */\
		  ((e)->kind<=(CHAR)E_MAP &&\
		   (e)->kind>=(CHAR)E_EMPTY))
#define ISNEMAP(e) (/*(e)!=NIL && */\
		    ((e)->kind==(CHAR)E_MAP ||\
		     (e)->kind==(CHAR)E_MAPLET))

#define NPROTECTED 512
EXTERN ExprPtr ProtectedListAgainstGC;
EXTERN ExprPtr ProtectedTable[];
EXTERN ExprPtr *ProtectedTableTop;
EXTERN ExprPtr ProtectedAgainstGC;
typedef struct GCMarker {
  ExprPtr pList;
  ExprPtr *pTop;
} GCMarker;

#include "refcount.h"

/* there are two ways to protect your data against GC:
   the first protects *values*, by copying them to special-purpose
   stacks that GC knows about;
   the second protects *variables*, by maintaing a list of explicit
   stack frames in the C stack. While the first method allows for
   more efficient code (since variable values can be put in registers),
   the second produces code that is more easily maintained, or
   more easily generated by the compiler. */
/* first method: */
IMPORT ExprPtr Protect P((IN CONST ExprPtr e));
IMPORT void EndProtect P((GCMarker *save));
IMPORT ExprPtr getProtectWrapper P((void));
IMPORT void setProtectWrapper P((ExprPtr wrapper,ExprPtr keep));
#define ENTER struct GCMarker protectThemAgainstGCs; \
		protectThemAgainstGCs.pList = ProtectedListAgainstGC; \
		protectThemAgainstGCs.pTop = ProtectedTableTop
#define QUIT EndProtect(&protectThemAgainstGCs)
#define KEEP(e) ((ProtectedTableTop<ProtectedTable+NPROTECTED)?(*ProtectedTableTop++=(e)):Protect(e))
IMPORT void EndProtectExcept P((IN CONST GCMarker *save,IN CONST ExprPtr expr));
#define KEEPONLY(e) EndProtectExcept(&protectThemAgainstGCs,e)
#define QUICKKEEP(e) ProtectedAgainstGC = e
#define BEGINQUICKKEEP()
#define ENDQUICKKEEP() ProtectedAgainstGC = NIL
/* second method: */
typedef struct gcStackView {
  COUNTER n;
  struct gcStackView *next; /* next stack frame */
  ExprPtr p[1]; /* actually variable length */
} gcStackView;
EXTERN gcStackView *theStackFrames,*theHeapFrames;
/* theHeapFrames is the same as theStackFrames, except it is for
   frames allocated on the heap, not on the C stack. */

#define ELINK(a) (a)->link

#define IVAL(a) (a)->what.i.val
#define INEXT(a) (a)->what.i.next

#define XVAL(a) (a)->what.x

#define CAR(a) (a)->what.pair.car
#define CDR(a) (a)->what.pair.cdr

#define MAPLEFT(a) CAR(a)
#define MAPRIGHT(a) CDR(a)

#define MAPLETX(a) CAR(a)
#define MAPLETY(a) CDR(a)

#define LETBIND(a) CAR(a)
#define LETBODY(a) CDR(a)

#define MEMOREM(a) CAR(a)
#define MEMOFUN(a) CDR(a)

#define APPLFUN(a) CAR(a)
#define APPLARG(a) CDR(a)

#define PATAS1(a) CAR(a)
#define PATAS2(a) CDR(a)

#define PATMAPADDXY(a) CAR(a)
#define PATMAPADDREST(a) CDR(a)

#define SETREF(a) CAR(a)
#define SETVAL(a) CDR(a)

#define TESTIF(a) CAR(a)
#define TESTFORK(a) CDR(a)

#define PROGDO(a) CAR(a)
#define PROGTHEN(a) CDR(a)

#define OFFP(a) CAR(a)
#define OFDOM(a) CDR(a)

#define WCOND(a) CAR(a)
#define WBODY(a) CDR(a)

#define HBODY(a) CAR(a)
#define HHANDLERS(a) CDR(a)

#define PACKVAL(a) CAR(a)
#define PACKTYPE(a) CDR(a)

#define REPART(a) CAR(a)
#define IMPART(a) CDR(a)

#define DZCODE(a) CAR(a)
#define DZINFO(a) CDR(a)

#define DNPAT(a) CAR(a)
#define DNINFO(a) CDR(a)

#define REFVAL(a) (a)->what.quote.ref
#define COUNTEDVAL(a) (a)->what.quote.ref
#define COUNTEDN(a) (a)->what.quote.n
#define SETCOUNTEDN(a,nn) (a)->what.quote.n = nn
#define FUNARITY(a) (a)->magic
#define SETFUNARITY(a,n) (a)->magic = (SHORT)(n)

#define GETREF(a) REFVAL(a)
#define RBODY(a) REFVAL(a)
#define EXCVAL(a) REFVAL(a)
#define DCODE(a) REFVAL(a)
#define NEXTFREE(a) REFVAL(a)

#define ASHAPE(a) (a)->what.array.shape
#define AVAL(a) (a)->what.array.val

#define ACTION(e) (e)->what.compiled.action
#define QENV(e) (e)->what.compiled.env

#define EVARS(e) AVAL(e)
#define ENEXT(e) ASHAPE(e)

#define STEXT(sym) (sym)->what.str.text
#define SLEN(sym) (sym)->what.str.len

#define VARRANK(a) (a)->what.var.rank
#define VARLEVEL(a) (a)->what.var.level

#define PATCHECKVAL(a) (a)->what.handler.eh
#define PATCHECKBITS(a) (*(CHAR **)&(a)->what.handler.jb)

#define SCHUNK(a) (a)->what.stack.chunk
#define SJMPBUF(a) (a)->what.stack.jb

#define EXTDESC(a) (a)->what.external.desc
#define EXTDATA(a) (a)->what.external.data

#define MAX_ARITY ((1L<<NLOWBITS)-1L)
#define ENV_UNDERLYING(env) (ExprPtr)(((M_TYPE)(env)) & ~MAX_ARITY)
#define ENV_ARITY(env) (((M_TYPE)(env)) & MAX_ARITY)
#define ENV_CONS(env,n) ((ExprPtr) (((M_TYPE)(env)) | (n)))
/*
  A compiledAction f may be called as:
  - (*f) (env ^ 0x1, arg) for unary calls, where arg=(arg1, ..., argn).
  - (*f) (env ^ 0x0) for zero-ary calls.
  - (*f) (env ^ 0x2, arg1, arg2) for binary calls.
  ...
  - (*f) (env ^ (2^NLOWBITS-1), arg1, ..., arg{2^NLOWBITS-1})
  - for n-ary calls, n>=2^NLOWBITS, we always resort to unary calls.
 */

#define ACTION_NAME(name) MERGE(name,Action)
#define ACTION_PROTO(name) ExprPtr ACTION_NAME(name) P((IN CONST ExprPtr env, ...));
#define DECL_ACTION(name) IMPORT ACTION_PROTO(name)
#define DECL_ACTION_NORETURN(name) IMPORT NORETURN ExprPtr ACTION_NAME(name) P((IN CONST ExprPtr env, ...)) _NORETURN;
#ifdef HAS_STDARG
#define FUN_HEADER(name) ExprPtr name(IN CONST ExprPtr env, ...)
#define VA_START va_start(_args,env)
#else
#define FUN_HEADER(name) ExprPtr name(env,va_alist)	\
       IN CONST ExprPtr env;				\
       va_dcl
#define VA_START va_start(_args)
#endif
#define ACTION_HEADER(name) FUN_HEADER(ACTION_NAME(name))
#define FC_KEEP(a) (void) KEEP(a);
#define FC_IGN(a) 
#define FUN_COLLECT(keep) va_list _args; ExprPtr arg, res;		\
    VA_START;								\
    switch (ENV_ARITY(env))						\
      {									\
      case 0:								\
	arg = tuple0;							\
	break;								\
      case 1:								\
	arg = va_arg(_args,ExprPtr);					\
	break;								\
      case 2:								\
	{								\
	  ExprPtr arg1 = va_arg(_args,ExprPtr);				\
	  ExprPtr arg2 = va_arg(_args,ExprPtr);				\
	  arg = Ktuple2(arg1,arg2);					\
	  keep(arg)							\
	  break;							\
	}								\
      default:								\
	{								\
	  COUNTER i, n = ENV_ARITY(env);				\
	  ExprPtr args[MAX_ARITY+1];					\
	  for (i=0; i<n; i++)						\
	    args[i] = va_arg(_args,ExprPtr);				\
	  arg = KtupleGnrl(n,args);					\
	  keep(arg)							\
	  break;							\
	}								\
      }									\
    va_end(_args);
#define ACTION_HEADER_COLLECT(name) \
  ACTION_HEADER(name)  { ENTER; FUN_COLLECT(FC_KEEP) {
#define FUN_COLLECT_WEAK(keep) va_list _args; ExprPtr arg, res;		\
    VA_START;								\
    switch (ENV_ARITY(env))						\
      {									\
      case 0:								\
	arg = tuple0;							\
	break;								\
      case 1:								\
	arg = va_arg(_args,ExprPtr);					\
	break;								\
      case 2:								\
	{								\
	  ExprPtr arg1 = va_arg(_args,ExprPtr);				\
	  ExprPtr arg2 = va_arg(_args,ExprPtr);				\
	  arg = MakePair(E_TUPLE_2,arg1,arg2);				\
	  keep(arg)							\
	  break;							\
	}								\
      default:								\
	{								\
	  COUNTER i, n = ENV_ARITY(env);				\
	  ExprPtr args[MAX_ARITY+1];					\
	  for (i=0; i<n; i++)						\
	    args[i] = va_arg(_args,ExprPtr);				\
	  arg = MakeRecord(E_TUPLE_GNRL,NIL,n,args);			\
	  keep(arg)							\
	  break;							\
	}								\
      }									\
    va_end(_args);
#define ACTION_HEADER_COLLECT_WEAK(name) \
  ACTION_HEADER(name) { ENTER; FUN_COLLECT_WEAK(FC_KEEP) {
#define ACTION_HEADER_0ARG(name) ACTION_HEADER(name) {{
#define ACTION_HEADER_1ARG(name) ACTION_HEADER(name) { va_list _args; ExprPtr arg; \
				   VA_START;				\
				   arg = va_arg(_args,ExprPtr);		\
				   va_end(_args); {
#define ACTION_HEADER_2ARGS(name) ACTION_HEADER(name) { va_list _args; ExprPtr _arg, arg1, arg2; \
				    VA_START;				\
				    if (ENV_ARITY(env)==2) {		\
				      arg1 = va_arg(_args,ExprPtr);	\
				      arg2 = va_arg(_args,ExprPtr);	\
				    }					\
				    else {				\
				      _arg = va_arg(_args,ExprPtr);	\
				      arg1 = CAR(_arg);			\
				      arg2 = CDR(_arg);			\
				    }					\
				    va_end(_args); {
#define ACTION_HEADER_3ARGS(name) ACTION_HEADER(name) {	\
    va_list _args; ExprPtr arg1,arg2,arg3;		\
    VA_START;						\
    if (ENV_ARITY(env)==3) {				\
      arg1 = va_arg(_args,ExprPtr);			\
      arg2 = va_arg(_args,ExprPtr);			\
      arg3 = va_arg(_args,ExprPtr);			\
    }							\
    else {						\
      ExprPtr _arg = va_arg(_args,ExprPtr);		\
      VecPtr vp = AVAL(_arg);				\
      arg1 = vp->expr[0];				\
      arg2 = vp->expr[1];				\
      arg3 = vp->expr[2];				\
    }							\
    va_end(_args); {
#define ACTION_HEADER_4ARGS(name) ACTION_HEADER(name) {	\
    va_list _args; ExprPtr arg1,arg2,arg3,arg4;	\
    VA_START;						\
    if (ENV_ARITY(env)==4) {				\
      arg1 = va_arg(_args,ExprPtr);			\
      arg2 = va_arg(_args,ExprPtr);			\
      arg3 = va_arg(_args,ExprPtr);			\
      arg4 = va_arg(_args,ExprPtr);			\
    }							\
    else {						\
      ExprPtr _arg = va_arg(_args,ExprPtr);		\
      VecPtr vp = AVAL(_arg);				\
      arg1 = vp->expr[0];				\
      arg2 = vp->expr[1];				\
      arg3 = vp->expr[2];				\
      arg4 = vp->expr[3];				\
    }							\
    va_end(_args); {
#define ACTION_HEADER_5ARGS(name) ACTION_HEADER(name) {	\
    va_list _args; ExprPtr _arg,arg1,arg2,arg3,arg4,arg5;	\
    VA_START;						\
    if (ENV_ARITY(env)==5) {				\
      arg1 = va_arg(_args,ExprPtr);			\
      arg2 = va_arg(_args,ExprPtr);			\
      arg3 = va_arg(_args,ExprPtr);			\
      arg4 = va_arg(_args,ExprPtr);			\
      arg5 = va_arg(_args,ExprPtr);			\
    }							\
    else {						\
      ExprPtr _arg = va_arg(_args,ExprPtr);		\
      VecPtr vp = AVAL(_arg);				\
      arg1 = vp->expr[0];				\
      arg2 = vp->expr[1];				\
      arg3 = vp->expr[2];				\
      arg4 = vp->expr[3];				\
      arg5 = vp->expr[4];				\
    }							\
    va_end(_args); {
#define ACTION_HEADER_6ARGS(name) ACTION_HEADER(name) {	\
    va_list _args; ExprPtr arg1,arg2,arg3,arg4,arg5,arg6;	\
    VA_START;						\
    if (ENV_ARITY(env)==6) {				\
      arg1 = va_arg(_args,ExprPtr);			\
      arg2 = va_arg(_args,ExprPtr);			\
      arg3 = va_arg(_args,ExprPtr);			\
      arg4 = va_arg(_args,ExprPtr);			\
      arg5 = va_arg(_args,ExprPtr);			\
      arg6 = va_arg(_args,ExprPtr);			\
    }							\
    else {						\
      ExprPtr _arg = va_arg(_args,ExprPtr);		\
      VecPtr vp = AVAL(_arg);				\
      arg1 = vp->expr[0];				\
      arg2 = vp->expr[1];				\
      arg3 = vp->expr[2];				\
      arg4 = vp->expr[3];				\
      arg5 = vp->expr[4];				\
      arg6 = vp->expr[5];				\
    }							\
    va_end(_args); {
#define ACTION_END }}
#define DEF_ACTION(name) PUBLIC ACTION_HEADER(name)
#define DEF_PRIVATE_ACTION(name) PRIVATE ACTION_HEADER(name)
#define DEF_ACTION_0ARG(name) PUBLIC ACTION_HEADER_0ARG(name)
#define DEF_ACTION_1ARG(name) PUBLIC ACTION_HEADER_1ARG(name)
#define DEF_ACTION_2ARGS(name) PUBLIC ACTION_HEADER_2ARGS(name)
#define DEF_ACTION_3ARGS(name) PUBLIC ACTION_HEADER_3ARGS(name)
#define DEF_ACTION_4ARGS(name) PUBLIC ACTION_HEADER_4ARGS(name)
#define DEF_ACTION_5ARGS(name) PUBLIC ACTION_HEADER_5ARGS(name)
#define DEF_ACTION_6ARGS(name) PUBLIC ACTION_HEADER_6ARGS(name)
#define DEF_ACTION_COLLECT(name) PUBLIC ACTION_HEADER_COLLECT(name)
#define DEF_ACTION_COLLECT_WEAK(name) PUBLIC ACTION_HEADER_COLLECT_WEAK(name)
#define ACTION_END_COLLECT } QUIT; return res; }
#define DEF_PRIVATE_ACTION_0ARG(name) PRIVATE ACTION_HEADER_0ARG(name)
#define DEF_PRIVATE_ACTION_1ARG(name) PRIVATE ACTION_HEADER_1ARG(name)
#define DEF_PRIVATE_ACTION_2ARGS(name) PRIVATE ACTION_HEADER_2ARGS(name)
#define DEF_PRIVATE_ACTION_3ARGS(name) PRIVATE ACTION_HEADER_3ARGS(name)
#define DEF_PRIVATE_ACTION_COLLECT(name) PRIVATE ACTION_HEADER_COLLECT(name)
#define DEF_PRIVATE_ACTION_COLLECT_WEAK(name) PRIVATE ACTION_HEADER_COLLECT_WEAK(name)

IMPORT void GC P((void));
IMPORT void GCgrow P((IN COUNTER urgency, IN CONST ExprPtr tag));

IMPORT ExprPtr Reverse1 P((IN CONST ExprPtr expr,IN CONST ExprPtr tail)) _PURE;
#define Reverse(expr) Reverse1(expr,NIL)
IMPORT ExprPtr Append1 P((IN CONST ExprPtr e1,IN CONST ExprPtr e2)) _PURE;
IMPORT ExprPtr Append0 P((IN CONST ExprPtr e1,IN CONST ExprPtr e2)) _PURE;
IMPORT ExprPtr Append P((IN CONST ExprPtr e1,IN CONST ExprPtr e2)) _PURE;
#define SetAdd(x,map) MapPut(x,UNIT,map)
#define SetRemove(x,map) MapRemove(x,map)
#define SetUnion(map1,map2) MapMerge(map1,map2)
#define SetInter(map1,map2) MapDomRestrTo(map1,map2)
#define SetDiff(map1,map2) MapDomRestrBy(map1,map2)
#define SetDelta(map1,map2) MapDelta(map1,map2)
#define SetInterp(map1,map2) MapInterp(map1,map2)
IMPORT INT ListLength P((IN CONST ExprPtr list)) _PURE;
IMPORT MapPtr Elems0 P((IN CONST ExprPtr list)) _PURE;
IMPORT MapPtr Inds0 P((IN CONST ExprPtr list)) _PURE;
IMPORT MapPtr ListToMap0 P((IN CONST ExprPtr list)) _PURE;

IMPORT ExprPtr Nreverse0 P((INOUT ExprPtr list,IN CONST ExprPtr tail));
IMPORT ExprPtr Nreverse1 P((INOUT ExprPtr list,IN CONST ExprPtr tail));
#define Nreverse(list) Nreverse1(list,NIL)
IMPORT ExprPtr Nappend P((INOUT ExprPtr list1,IN CONST ExprPtr list2));

EXTERN MapPtr empty_map;
#define NewMap() empty_map
#define EmptyMap() NewMap()
#define EmptyMapp(map) ((map)==empty_map)
IMPORT MapPtr MapSingleton P((IN CONST ExprPtr x,IN CONST ExprPtr y)) _PURE;
#define Singleton(x) MapSingleton(x,UNIT)
IMPORT MapPtr MapPair P((IN CONST ExprPtr x1,IN CONST ExprPtr y1,
			 IN CONST ExprPtr x2,IN CONST ExprPtr y2)) _PURE;
#define Pair(x,y) MapPair(x,UNIT,y,UNIT)
IMPORT ExprPtr MapGet1 P((IN CONST ExprPtr elem,IN CONST MapPtr map,
			 IN CONST CODE mask,OUT Boolean *found));
IMPORT ExprPtr MapGet P((IN CONST ExprPtr elem,IN CONST MapPtr map,
			 OUT Boolean *found));
IMPORT ExprPtr MapGetMaplet1 P((IN REGISTER CONST ExprPtr elem,
				IN REGISTER CONST MapPtr map,
				IN REGISTER CODE mask,OUT Boolean *found));
IMPORT ExprPtr MapGetSure P((IN CONST ExprPtr elem,IN CONST MapPtr map)) _PURE;
IMPORT MapPtr MapPutTrie P((IN CONST ExprPtr x,IN CONST ExprPtr y,
			    IN CONST MapPtr map0,IN CODE mask)) _PURE;
IMPORT MapPtr MapPutBehindTrie P((IN CONST ExprPtr x,IN CONST ExprPtr y,
				  IN CONST MapPtr map0,IN CODE mask)) _PURE;
IMPORT MapPtr MapPutMapletTrie P((IN CONST ExprPtr maplet,IN CONST MapPtr map0,
				  IN REGISTER CODE mask)) _PURE;
IMPORT MapPtr MapPutMapletBehindTrie P((IN CONST ExprPtr maplet,IN CONST MapPtr map0,
					IN REGISTER CODE mask)) _PURE;
IMPORT MapPtr MapRemoveTrie P((IN CONST ExprPtr x,IN CONST MapPtr map0,IN CODE mask)) _PURE;
IMPORT ExprPtr MapPut P((IN CONST ExprPtr x,IN CONST ExprPtr y,
			 IN CONST MapPtr map)) _PURE;
IMPORT ExprPtr MapPutBehind P((IN CONST ExprPtr x,IN CONST ExprPtr y,
			       IN CONST MapPtr map)) _PURE;
IMPORT MapPtr MapRemove P((IN CONST ExprPtr x,IN CONST MapPtr map)) _PURE;
typedef CONST Boolean (*mapSweepFunc) P((IN CONST ExprPtr elt,
					 IN CONST ExprPtr val,
					 INOUT CHAR *data));
typedef CONST MapPtr (*mapCollectFunc) P((IN CONST ExprPtr elt,
					  IN CONST ExprPtr val,
					  INOUT CHAR *data));
typedef CONST MapPtr (*mapFoldFunc) P((IN CONST ExprPtr elt,
				       IN CONST ExprPtr val,
				       INOUT CHAR *data));
typedef CONST ExprPtr (*mapFoldOp) P((IN CONST ExprPtr e1,
				      IN CONST ExprPtr e2));
#define MAPFUNC(returntype,fun) \
PRIVATE returntype fun P((IN CONST ExprPtr x,IN CONST ExprPtr y, \
			  INOUT CHAR *_data)); \
PRIVATE returntype fun(x,y,_data) \
IN CONST ExprPtr x,y; \
INOUT CHAR *_data;
#define MAPSWEEPFUNC(fun) MAPFUNC(Boolean,fun)
#define MAPCOLLECTFUNC(fun) MAPFUNC(MapPtr,fun)
#define MAPFOLDFUNC(fun) MAPFUNC(MapPtr,fun)
#define MAPDATA(type) ((type)_data)
#define MAPFOLDOP(op) \
PRIVATE ExprPtr op P((IN CONST ExprPtr e1,IN CONST ExprPtr e2)); \
PRIVATE ExprPtr op(e1,e2) \
IN CONST ExprPtr e1,e2;
IMPORT Boolean MapSweep P((IN CONST MapPtr map,
			   IN mapSweepFunc func,
			   INOUT VOIDPTR data));
IMPORT ExprPtr MapSome P((IN CONST MapPtr map,
			  IN mapSweepFunc func,
			  INOUT VOIDPTR data));
IMPORT MapPtr MapCollect0 P((IN CONST MapPtr map,
			     IN mapCollectFunc func,
			     INOUT VOIDPTR data));
IMPORT MapPtr MapCollect P((IN CONST MapPtr map,
			    IN mapCollectFunc func,
			    INOUT VOIDPTR data));
IMPORT ExprPtr MapFold P((IN CONST MapPtr map,
			  IN mapFoldFunc func,
			  INOUT VOIDPTR data,
			  IN CONST ExprPtr neutral,
			  IN mapFoldOp op
			  ));
IMPORT INT MapCard P((IN CONST MapPtr map)) _PURE;
IMPORT INT MapCardDec P((IN CONST MapPtr map,IN INT n)) _PURE;
#define MapCardEq(map,n) (MapCardDec(map,n)==0)
#define MapCardGeq(map,n) (MapCardDec(map,n)<=0)
#define MapCardLeq(map,n) (MapCardDec(map,n)>=0)
IMPORT MapPtr MapMerge1 P((IN CONST MapPtr map1,IN CONST MapPtr map2,IN CONST CODE mask)) _PURE;
IMPORT MapPtr MapMerge P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT MapPtr MapDomRestrTo1 P((IN CONST MapPtr map1,IN CONST MapPtr map2,IN CODE mask)) _PURE;
IMPORT MapPtr MapDomRestrBy1 P((IN CONST MapPtr map1,IN CONST MapPtr map2,IN CODE mask)) _PURE;
IMPORT MapPtr MapRngRestrTo1 P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT MapPtr MapRngRestrBy1 P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT MapPtr MapDelta1 P((IN CONST MapPtr map1,IN CONST MapPtr map2,IN CODE mask)) _PURE;
IMPORT MapPtr MapDomRestrTo P((IN CONST MapPtr map,IN CONST MapPtr s)) _PURE;
IMPORT MapPtr MapDomRestrBy P((IN CONST MapPtr map,IN CONST MapPtr s)) _PURE;
IMPORT MapPtr MapRngRestrTo P((IN CONST MapPtr map,IN CONST MapPtr s)) _PURE;
IMPORT MapPtr MapRngRestrBy P((IN CONST MapPtr map,IN CONST MapPtr s)) _PURE;
IMPORT MapPtr MapInter P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT MapPtr MapDelta P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT MapPtr MapDom1 P((IN CONST MapPtr map)) _PURE;
IMPORT MapPtr Dom P((IN CONST MapPtr map)) _PURE;
IMPORT MapPtr Rng0 P((IN CONST MapPtr map)) _PURE;
IMPORT Boolean MapMemberp P((IN CONST ExprPtr x,IN CONST MapPtr map)) _PURE;
IMPORT Boolean MapInterp P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT Boolean MapIncludedp P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT Boolean SetIncludedp P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT ExprPtr ChooseInDom P((IN CONST MapPtr map)) _PURE;
IMPORT ExprPtr ChooseInRng P((IN CONST MapPtr map)) _PURE;
IMPORT Boolean Setp P((IN CONST MapPtr map)) _PURE;
IMPORT MapPtr MapCompose P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT MapPtr MapImage P((IN CONST MapPtr map,IN CONST MapPtr s)) _PURE;
IMPORT void MapSplit P((IN CONST MapPtr map,OUT MapPtr *map1,OUT MapPtr *map2));
IMPORT Boolean sameDomains P((IN CONST MapPtr map1,IN CONST MapPtr map2)) _PURE;
IMPORT void setMapPut P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT MapPtr *loc));
IMPORT void setrefMapPut P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT ExprPtr ref));
IMPORT void setcarMapPut P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT ExprPtr pair));
IMPORT void setcdrMapPut P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT ExprPtr pair));
#define setSetAdd(x,sp) setMapPut(x,UNIT,sp)
IMPORT void setMapRemove P((IN CONST ExprPtr x,INOUT MapPtr *map));
IMPORT void setcarMapRemove P((IN CONST ExprPtr x,INOUT ExprPtr pair));
IMPORT void setcdrMapRemove P((IN CONST ExprPtr x,INOUT ExprPtr pair));
IMPORT void setMapMerge P((INOUT MapPtr *loc,IN CONST MapPtr map));
#define setSetUnion(sp,s) setMapMerge(sp,s)
IMPORT void setcarMapMerge P((INOUT ExprPtr pair,IN CONST MapPtr map));
IMPORT void setcdrMapMerge P((INOUT ExprPtr pair,IN CONST MapPtr map));
IMPORT void setMapDomRestrTo P((INOUT MapPtr *loc,IN CONST MapPtr map));
IMPORT void setMapDomRestrBy P((INOUT MapPtr *loc,IN CONST MapPtr map));

     /* a few of the same operations, but not KEEP()ing their results. */
EXTERN ExprPtr mapStack[];
EXTERN ExprPtr *mapStackTop;
#define MASK0 (CODE)(1L<<NLOWBITS)

IMPORT MapPtr MapInverse0 P((IN CONST MapPtr map)) _PURE;
#define MapOverwrite(map1,map2) (map1==empty_map)?map2:((map2==empty_map)?map1:(mapStackTop=mapStack, MapMerge1(map1,map2,MASK0)))
#define MapAdd(x,y,map) (map==empty_map)?MapletKons(x,y):MapPutTrie(x,y,map,MASK0)
#define MapSlip(x,y,map) (map==empty_map)?MapletKons(x,y):MapPutBehindTrie(x,y,map,MASK0)
#define MapKill(x,map) (map==empty_map)?map:MapRemoveTrie(x,map,MASK0)
#define MapDRT(map1,map2) (map1==empty_map || map2==empty_map)?empty_map:(mapStackTop=mapStack, MapDomRestrTo1(map1,map2,MASK0))
#define MapDRB(map1,map2) (map1==empty_map || map2==empty_map)?map1:(mapStackTop=mapStack, MapDomRestrBy1(map1,map2,MASK0))
#define MapRRT(map1,map2) (map1==empty_map || map2==empty_map)?empty_map:(mapStackTop=mapStack, MapRngRestrTo1(map1,map2))
#define MapRRB(map1,map2) (map1==empty_map || map2==empty_map)?map1:(mapStackTop=mapStack, MapRngRestrBy1(map1,map2))
#define MapD(map1,map2) (map1==empty_map)?map2:((map2==empty_map)?map1:MapDelta1(map1,map2,MASK0))
#define Dom0(map) (map==empty_map)?empty_map:(mapStackTop=mapStack, MapDom1(map))

IMPORT void setNPut1 P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT ExprPtr *locp,
			IN CODE mask,IN CONST ExprPtr lastloc));
#define setNPut(x,y,locp,lastloc) setNPut1(x,y,locp,MASK0,lastloc)
IMPORT void setNPutBehind1 P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT ExprPtr *locp,
			IN CODE mask,IN CONST ExprPtr lastloc));
#define setNPutBehind(x,y,locp,lastloc) setNPutBehind1(x,y,locp,MASK0,lastloc)
IMPORT void setNRemove1 P((IN CONST REGISTER ExprPtr x,INOUT REGISTER ExprPtr *locp,
			   IN CONST REGISTER CODE mask,IN CONST ExprPtr lastloc));
#define setNRemove(x,locp,lastloc) setNRemove1(x,locp,MASK0,lastloc)
IMPORT ExprPtr NGet1 P((IN CONST ExprPtr x,IN CONST ExprPtr y,OUT Boolean *foundp,IN CODE mask));
#define NGet(x,loc,foundp) NGet1(x,loc,foundp,MASK0)
IMPORT ExprPtr NGetMaplet1 P((IN CONST ExprPtr x,IN CONST ExprPtr y,OUT Boolean *foundp,IN CODE mask));
#define NGetMaplet(x,loc,foundp) NGetMaplet1(x,loc,foundp,MASK0)
typedef Boolean (*nSweepFunc) P((ExprPtr loc,char *data));
IMPORT Boolean NSweep P((IN CONST ExprPtr loc,IN CONST nSweepFunc fun,IN CONST char *data));
typedef MapPtr (*nCollectFunc) P((ExprPtr loc,char *data));
IMPORT MapPtr NCollect P((IN CONST ExprPtr loc,IN CONST nCollectFunc fun,IN CONST char *data));

IMPORT ExprPtr Ktuple2 P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr Kons P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr MapletKons P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr MapKons P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr Kcomplex P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr KtypeAppl P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr KtypeFx P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr Krecord1 P((IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;

IMPORT ExprPtr Kdata P((IN SHORT magic,IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _PURE;
IMPORT ExprPtr Kvar P((IN ExprKind kind,IN COUNTER counter,IN COUNTER level)) _PURE;
#define Kvarref(i,level) Kvar(E_VARREF,i,level)
#define Kvarintro(i) Kvar(E_VARINTRO,i,0)
IMPORT ExprPtr Kcounted P((IN ExprKind kind,IN CONST ExprPtr expr,IN INT n)) _PURE;
IMPORT ExprPtr KCOUNTED P((IN CONST ExprPtr expr,IN INT n)) _PURE;
#define Kref(kind,expr) Kcounted(kind,expr,0L)
#define KnumType(expr) Kref(E_NUMTYPE,expr)
#define Ktuple1(expr) Kref(E_TUPLE_1,expr)
IMPORT ExprPtr Karray P((IN ExprKind kind,IN CONST ExprPtr shape,IN INT n,IN CONST ExprPtr *exprs)) _PURE;
#define KtupleGnrl(n,exprs) Karray(E_TUPLE_GNRL,NIL,n,exprs)
#define KrecordGnrl(shape,n,exprs) Karray(E_RECORD_GNRL,shape,n,exprs)

IMPORT ExprPtr Kint P((IN INT i)) _PURE;
IMPORT ExprPtr Kreal P((IN CONST REAL *x)) _PURE;
#if EXTP
IMPORT ExprPtr Kext P((IN CONST EXTENDED *x)) _PURE;
#else
#define Kext(x) Kreal(x)
#endif
IMPORT ExprPtr Kstring1 P((IN INT len,IN CONST CHAR *text,IN Boolean copy)) _PURE;
#define Kstring(len,text) Kstring1(len,text,true)
IMPORT ExprPtr Kstr P((IN CONST CHAR *text)) _PURE;
#define Ksymbol(name) Kstr((CHAR *)name)
EXTERN ExprPtr patany,
  tuple0,record0;
IMPORT ExprPtr Kexternal P((IN CONST extDescPtr desc,IN CONST char *data)) _PURE;

#define symName(sym) STEXT(sym)
#define symNameLen(sym) SLEN(sym)

IMPORT ExprPtr KONS P((ExprPtr a,ExprPtr b)) _PURE;
IMPORT ExprPtr KINT P((INT i)) _PURE;

EXTERN ExprPtr zzero,one,mone,two,half,zpi,
#if FLOAT_FORMAT==IEEE754
		pinf,minf,
#endif
		izero,ione,itwo;

EXTERN SymPtr trueValue,falseValue;

typedef enum excTag { X_HANDLER, X_PROF } excTag;

typedef struct excHandler { /* these structures for the exception handler stack,
			       and are alloctmp()ed. */
  struct excHandler *next;
  excTag tag;
  ExprPtr data; /* environment for X_HANDLER,
		   profile data (as an external) for X_PROF. */
  ExprPtr code; /* code and jb are only used for X_HANDLER. */
  stackDesc jb;
} excHandler;
EXTERN excHandler *excStack;
#define NEXTEXCHANDLER(exc) ((exc)->next)

IMPORT ExprPtr Raise P((IN CONST ExprPtr exception)) _NORETURN;
IMPORT NORETURN void Toplevel P((IN int errcond)) _NORETURN;

/*IMPORT ExprPtr Eval P((IN CONST ExprPtr e, IN CONST ExprPtr env));*/

#define N_C_GLOBALS 26
EXTERN ExprPtr CGlobals[];
#define I_GLOBENV 0
#define GLOBENV CGlobals[I_GLOBENV]
#define I_THEEXPR 1
#define THEEXPR CGlobals[I_THEEXPR]
#define I_BINDEXC 2
#define BINDEXC CGlobals[I_BINDEXC]
#define I_MATCHEXC 3
#define MATCHEXC CGlobals[I_MATCHEXC]
#define I_MAPGETEXC 4
#define MAPGETEXC CGlobals[I_MAPGETEXC]
#define I_MAPEXC 5
#define MAPEXC CGlobals[I_MAPEXC]
#define I_EMPTYEXC 6
#define EMPTYEXC CGlobals[I_EMPTYEXC]
#define I_ARITHEXC 7
#define ARITHEXC CGlobals[I_ARITHEXC]
#define I_INCOMPEXC 8
#define INCOMPEXC CGlobals[I_INCOMPEXC]
#define I_TRUEVALUE 9
#define I_FALSEVALUE 10
#define I_NIL 11
#define I_EMPTY 12
#define I_TUPLE0 13
#define I_RECORD0 14
#define I_ZZERO 15
#define I_ONE 16
#define I_MONE 17
#define I_TWO 18
#define I_HALF 19
#define I_ZPI 20
#define I_PINF 21
#define I_MINF 22
#define I_IZERO 23
#define I_IONE 24
#define I_ITWO 25

#if DEBUGGING!=0
IMPORT Boolean checkTypeTag P((IN CONST ExprPtr val,IN CONST ExprPtr type));
#else
#define checkTypeTag(val,type) true
#endif
