/* Definitions for stdio replacement functions.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* why do we need a replacement for stdio functions ?
   well, normally, we don't, but on certain machines
   (like the IBM RS/6000), buffered file functions
   allocate memory for buffers on the fly, and instead
   of doing it through malloc(), they use their own
   inlined version of it. This conflicts with the
   redefinition of the memory management routines in
   mem.c, and produces very obscure bugs.
     This is also necessary, now, to provide the
   suspend/revive facility that we need to build the
   automatic reopening of files.
*/

typedef int pid_type;
#define NO_PROCESS (-1)

#define hFILE_BUFSIZE 2048
#define PWD_BUFSIZE 1024

typedef struct hFILE {
  struct hFILE *next;
  CHAR *dirname; /* directory name at the moment of creation */
  CHAR *filename; /* file name; used to suspend/revive */
  CHAR *ptr; /* pointer inside the buffer buf */
  CHAR *end; /* pointer to the end of data read in buf */
  int fildesc; /* number of file descriptor; -1 if file is suspended */
  pid_type pid; /* process id for hpopen() */
  long pos; /* position in file of character at start of buf */
  long realpos; /* position in file as said by lseek() on fildesc */
  int flags; /* some flags on the buffer and file */
  CHAR buf[hFILE_BUFSIZE];
} hFILE;

#define H_READ 0x01
#define H_WRITE 0x02
#define H_TEXTFILE 0x04
/* used if open without a 'b' mode; currently ignored */
#define H_EOF 0x08
/* set if reached end of file */
#define H_PROCESS 0x10
/* used if opened by hpopen() */

IMPORT hFILE *hfopen P((IN CONST char *name,IN CONST char *mode));
IMPORT int hfclose P((IN hFILE *f));
IMPORT hFILE *hfreopen P((IN CONST char *filename,IN CONST char *mode,IN hFILE *f));

IMPORT int hfsuspend P((IN hFILE *hf));
IMPORT int hfrevive P((IN hFILE *hf));

IMPORT int hfflush P((IN hFILE *f));
IMPORT long hftell P((IN hFILE *f));
IMPORT int hfseek P((IN hFILE *f,IN long offset,IN int whence));

#define hgetc(f) (((f)->ptr<(f)->end)?((int)(*(f)->ptr++)):hReadNewBufGetc(f))
#define hgetcN(f,n) (((f)->ptr<(f)->end)?((int)(*(f)->ptr++)):hReadNewBufGetcN(f,n))
IMPORT int hfgetc P((IN hFILE *f));
#define hputc(c,f) (((f)->ptr<(f)->buf+hFILE_BUFSIZE)?(*(f)->ptr++ = (CHAR)(c), c):hWriteNewBufPutc(c,f))
IMPORT int hfputc P((IN int c,IN hFILE *f));
IMPORT int hfputs P((IN CONST char *s,IN hFILE *f));

/* read at most max_size chars from f, stopping on EOF or '\n';
   final '\n' is stored; return number of characters read. */
IMPORT size_t hfgetline(OUT char *buf,IN size_t max_size,IN hFILE *f);
IMPORT char *hfgets(OUT char *buf, IN size_t size,IN hFILE *f);

IMPORT int hfprintf P((IN hFILE *f,IN CONST char *msg,...));

EXTERN hFILE *hstdin,*hstdout,*hstderr;

IMPORT void InitHIO P((void));
IMPORT void EndHIO P((void));

IMPORT char *ltoasc P((IN unsigned long a));
IMPORT char *rtoasc P((IN CONST EXTENDED *xp));
IMPORT char *timetoasc P((unsigned long t));

/* internal functions: */
IMPORT int hReadNewBufGetcN P((IN hFILE *f,IN int n));
IMPORT int hReadNewBufGetc P((IN hFILE *f));
IMPORT int hWriteNewBufPutc P((IN int c,IN hFILE *f));
IMPORT void unlinkbuf P((IN hFILE *f));
IMPORT void linkbuf P((IN hFILE *f));

IMPORT int hftruncate P((INOUT hFILE *f,IN INT n));

IMPORT int wait_pid P((pid_type pid));
IMPORT int hpopen P((IN CONST char **argv,
		     IN CONST char *mode,OUT hFILE **hff));
/* with argv an array terminated by a NULL entry,
   and argv[0] not NULL. */

