%{
/* GimML grammar (without Standard ML modules).
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "abssyn.h"
#include "expr.h"
#include "gram.h"
#include "type.h"
#include "compile.h"
#include "input.h"
#include "timing.h"
#include "debug.h"
#include "prtproto.h"
#include "grmproto.h"
#include "memproto.h"
#include "evlproto.h"
#include "errproto.h"
#include "cmpproto.h"
#include "typproto.h"
#include "stkproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sgram.h"
#endif

#ifndef bcopy
#define bcopy(src,dst,size) safememcpy(dst,src,(long)size)
#endif

#ifndef YYERROR_VERBOSE
#define YYERROR_VERBOSE
#endif
#undef YYSTACK_USE_ALLOCA
#define YYSTACK_ALLOC(n) alloctmp_anchored(n,&dummy)

#if CORE_TRACE
#define yylex tracingYylex
#endif

  PRIVATE COUNTER yy_gimml_stpcpy (char *yydest, const char *yysrc)
  {
    char c;
    char *yyd = yydest;
    const char *yys = yysrc;

    if (yydest)
      {
	while ((c = *yys++)!='\0')
	  {
	    switch (c)
	      {
	      case '_': c = ' '; break;
	      }
	    *yyd++ = c;
	  }
	*yyd = '\0';
      }
    else while (*yys++!='\0');
    return yys-yysrc-1;
  }

  PRIVATE COUNTER yy_gimml_symcpy (char *yydest, const char *yysrc)
  {
    char c;
    char *yyd = yydest;
    const char *yys = yysrc;

    if (yydest)
      {
	while ((c = *yys++)!='\0')
	  {
	    switch (c)
	      {
	      case '_': c = ' '; break;
	      case 'D': c = '.'; break;
	      case 'P': c = '('; break;
	      case 'Q': c = ')'; break;
	      case 'B': c = '['; break;
	      case 'C': c = ']'; break;
	      case 'L': c = '<'; break;
	      case 'G': c = '>'; break;
	      case 'Y': c = '{'; break;
	      case 'Z': c = '}'; break;
	      case 'I': c = '|'; break;
	      case 'A': c = '+'; break;
	      case 'N': c = '~'; break;
	      case 'F': c = '*'; break;
	      case 'E': c = '='; break;
	      case 'M': c = '-'; break;
	      case 'S': c = ';'; break;
	      case 'T': c = ':'; break;
	      case 'X': c = '^'; break;
	      case 'H': c = '#'; break;
	      case 'U': c = '_'; break;
	      }
	    *yyd++ = c;
	  }
	*yyd = '\0';
      }
    else while (*yys++!='\0');
    return yys-yysrc-1;
  }

 
  #define yytnamerr yy_gimml_namerr
  PRIVATE COUNTER yy_gimml_namerr (char *yyres, const char *yystr)
  {
    COUNTER yyn = 0;
    char *yyr = yyres;
    
    if (yystr[0]=='k' && yystr[1]=='w' && yystr[2]=='_')
      {
	if (yyres)
	  *yyr++ = '`';
	yyn++;
	yyn += yy_gimml_stpcpy (yyr, yystr+3);
	if (yyres)
	  yyres[yyn] = '\'';
	yyn++;
      }
    else if (yystr[0]=='t' && yystr[1]=='_')
      {
	yyn = yy_gimml_stpcpy (yyr, yystr+2);
      }
    else if (yystr[0]=='s' && yystr[1]=='_')
      {
	if (yyres)
	  *yyr++ = '`';
	yyn++;
	yyn += yy_gimml_symcpy (yyr, yystr+2);
	if (yyres)
	  yyres[yyn] = '\'';
	yyn++;
      }
    else
      {
	yyn = yy_gimml_stpcpy (yyres, yystr);
      }
    if (yyres)
      yyres[yyn] = '\0';
    return yyn;
  }
  
EXTERN SymPtr equalSym;
EXTERN Boolean wantCloseFile,syntax_analysis;
EXTERN ExprPtr theSyntax;
EXTERN SymPtr ItSymbol;

EXTERN SymPtr trueSym;

#define yyerror(msg) yy_error(yylloc.first_line,yylloc.first_column,yylloc.last_line,yylloc.last_column,msg)

PRIVATE ExprPtr cons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr aux;

  MAKECONS(a,b,aux);
  return KEEP(aux);
}

PRIVATE ExprPtr list1(a)
IN CONST ExprPtr a;
{
  ExprPtr aux;

  MAKECONS(a,NIL,aux);
  return KEEP(aux);
}

PRIVATE ExprPtr getInt(text, loc)
IN CONST CHAR *text;
IN CONST struct yyltype *loc;
{
  Boolean neg;
  long i;

  errno = 0;
  if ((neg = (text[0]=='~')))
    text++;
  if (text[0]=='0')
    switch (text[1])
      {
      case 'x': case 'X':
	i = (long)strtoul ((char *)text+2, NULL, 16);
	break;
      case 'o': case 'O':
	i = (long)strtoul ((char *)text+2, NULL, 8);
	break;
      case 'b': case 'B':
	i = (long)strtoul ((char *)text+2, NULL, 2);
	break;
      default:
	i = strtol ((char *)text+1, NULL, 10);
	break;
      }
  else i = strtol ((char *)text, NULL, 10);
  if (errno)
    {
      TZone zone;
      makeZone(&zone,loc,loc);
      type_warn(&zone,(CHAR *)message(75));
    }
  if (neg)
    i = -i;
  return KINT(i);
}

PRIVATE void checkLabel(loc)
IN CONST struct yyltype *loc;
{
  if (lexPath!=NIL)
    {
      TZone zone;
      makeZone(&zone,loc,loc);
      type_error(&zone,(CHAR *)message(386));
    }
}

PRIVATE COUNTER quantifier[4][2] = {
  { A_PROG_OF, A_PROG_MAP },
  { A_ALL_OF, A_ALL_MAP },
  { A_EXISTS_OF, A_EXISTS_MAP },
  { A_SOME_OF, A_SOME_MAP }
};

PRIVATE COUNTER qwhile[4] = {
  A_WHILE,
  A_WHILE_ALL,
  A_WHILE_EXISTS,
  A_WHILE_SOME
};

#if FLOAT_FORMAT==IEEE754
#define makeSpecNum(zone,x) makeNum(zone,x)
#else
#define makeSpecNum(zone,x) NIL
#endif

#define checkDebug(loc1,loc2)

PRIVATE ExprPtr do_load (loc1, loc2, s, recompile)
IN CONST struct yyltype *loc1,*loc2;
IN CONST char *s;
IN Boolean recompile;
{
  ExprPtr res;

  (void) getString(loc2,loc2,(CHAR *)s,false);
  {
    int len = strlen(s);
    TZone *zone = (TZone *)alloctmp_anchored(sizeof(TZone),(char *)&len);
    char *text = (char *)alloctmp_anchored(len+1,(char *)&len);

    strcpy(text,s);
    makeZone(zone,loc1,loc2);
    res = moduleLoad(zone,(CHAR *)text,recompile);
    if (res==NIL)
      {
	char *buf = alloctmp_anchored(len+64,(char *)&len);

	sprintf(buf,message(recompile?502:921),text);
	type_error(zone,(CHAR *)buf);
      }
    checkConsEnv(zone,res);
    augmentLexicalEnv(res);
    res = list1(build1(loc1,loc2,A_BASIS,res));
  }
  return res;
}

#if DEBUGGING
PUBLIC Boolean evaluatingp=false,stopwanted=false;
#endif

%}

%union {
  int i;
  ExprPtr e;
}

%token kw_abstype kw_all kw_and kw_andalso kw_as kw_case kw_do kw_datatype
%token kw_else kw_end kw_exception kw_exists kw_fn kw_fun kw_handle kw_if
%token kw_in kw_in_list kw_in_map kw_in_set kw_infix kw_infixr
%token kw_let kw_local kw_nonfix kw_of kw_orelse
%token kw_dimension kw_raise kw_rec kw_some kw_such_that kw_then kw_type
%token kw_scale kw_val kw_with kw_withtype kw_while s_P s_Q
%token s_B s_C s_DP s_DTE
%token s_LY s_Y s_Z
%token s_IB s_CI s_BI s_IC
%token t_comma s_T
%token s_S s_DDD s_U
%token s_I s_II s_E s_EG s_MG s_F
%token precedence_constant string_constant
%token integer_constant real_constant complex_constant
%token hex_integer_constant octal_integer_constant binary_integer_constant
%token t_identifier t_constructor
%token t_type_variable
%token t_left_associative_precedence_0_identifier t_right_associative_precedence_0_identifier
%token t_left_associative_precedence_1_identifier t_right_associative_precedence_1_identifier
%token t_left_associative_precedence_2_identifier t_right_associative_precedence_2_identifier
%token t_left_associative_precedence_3_identifier t_right_associative_precedence_3_identifier
%token t_left_associative_precedence_4_identifier t_right_associative_precedence_4_identifier
%token t_left_associative_precedence_5_identifier t_right_associative_precedence_5_identifier
%token t_left_associative_precedence_6_identifier t_right_associative_precedence_6_identifier
%token t_left_associative_precedence_7_identifier t_right_associative_precedence_7_identifier
%token t_left_associative_precedence_8_identifier t_right_associative_precedence_8_identifier
%token t_left_associative_precedence_9_identifier t_right_associative_precedence_9_identifier
%token t_left_associative_precedence_0_constructor t_right_associative_precedence_0_constructor
%token t_left_associative_precedence_1_constructor t_right_associative_precedence_1_constructor
%token t_left_associative_precedence_2_constructor t_right_associative_precedence_2_constructor
%token t_left_associative_precedence_3_constructor t_right_associative_precedence_3_constructor
%token t_left_associative_precedence_4_constructor t_right_associative_precedence_4_constructor
%token t_left_associative_precedence_5_constructor t_right_associative_precedence_5_constructor
%token t_left_associative_precedence_6_constructor t_right_associative_precedence_6_constructor
%token t_left_associative_precedence_7_constructor t_right_associative_precedence_7_constructor
%token t_left_associative_precedence_8_constructor t_right_associative_precedence_8_constructor
%token t_left_associative_precedence_9_constructor t_right_associative_precedence_9_constructor
%token t_backtick s_X kw_U s_MmG
%token kw_iterate kw_eqtype s_H s_AAIB kw_use
%token s_Ainf s_Ninf
%token s_nan0 s_nan1 s_nan2 s_nan3 s_nan4 s_nan5 s_nan6 s_nan7
%token s_Nnan1 s_Nnan2 s_Nnan3 s_Nnan4 s_Nnan5 s_Nnan6 s_Nnan7
%token kw_sub_map kw_delay kw_pack /* kw_extend suppressed */
%token kw_memofn kw_memofun kw_extern t_pragma t_cpp_line_control_directive

%token kw_sig kw_signature kw_structure kw_sharing t_open s_openF kw_include kw_functor
%token kw_abstraction kw_struct kw_compile

%right s_EG s_MG s_MmG
%left kw_fn
%nonassoc kw_case kw_of
%nonassoc kw_while kw_do
%nonassoc kw_if kw_then kw_else
%left kw_raise
%left kw_handle
%right kw_orelse
%right kw_andalso
%left kw_as

%type <e> real integer number number_or_int
%type <e> scale_id simple_scale scale optional_scale
%type <e> numerical_constant special_constant
%type <e> identifier long_constructor
%type <e> long_id_or_constructor /*long_symbol*/
%type <e> /*constructor id_or_constructor*/ symbol
%type <e> label labeled_expression expression_row optional_expression_row
%type <e> expression_list optional_expression_list
%type <e> expression_sequence implicit_progn
%type <e> domain_sweep domain_sweep_list
%type <e> such_that optional_such_that
%type <e> domain_description
%type <e> pure_maplet maplet maplet_list
%type <i> opening_brace optional_precedence
%type <e> atomic_expression application_expression delay_expression
%type <e> array_expression
%type <e> left_associative_precedence_0_expression right_associative_precedence_0_expression
%type <e> left_associative_precedence_1_expression right_associative_precedence_1_expression
%type <e> left_associative_precedence_2_expression right_associative_precedence_2_expression
%type <e> left_associative_precedence_3_expression right_associative_precedence_3_expression
%type <e> left_associative_precedence_4_expression right_associative_precedence_4_expression
%type <e> left_associative_precedence_5_expression right_associative_precedence_5_expression
%type <e> left_associative_precedence_6_expression right_associative_precedence_6_expression
%type <e> left_associative_precedence_7_expression right_associative_precedence_7_expression
%type <e> left_associative_precedence_8_expression right_associative_precedence_8_expression
%type <e> left_associative_precedence_9_expression right_associative_precedence_9_expression
%type <e> long_left_associative_precedence_0_identifier long_right_associative_precedence_0_identifier
%type <e> long_left_associative_precedence_1_identifier long_right_associative_precedence_1_identifier
%type <e> long_left_associative_precedence_2_identifier long_right_associative_precedence_2_identifier
%type <e> long_left_associative_precedence_3_identifier long_right_associative_precedence_3_identifier
%type <e> long_left_associative_precedence_4_identifier long_right_associative_precedence_4_identifier
%type <e> long_left_associative_precedence_5_identifier long_right_associative_precedence_5_identifier
%type <e> long_left_associative_precedence_6_identifier long_right_associative_precedence_6_identifier
%type <e> long_left_associative_precedence_7_identifier long_right_associative_precedence_7_identifier
%type <e> long_left_associative_precedence_8_identifier long_right_associative_precedence_8_identifier
%type <e> long_left_associative_precedence_9_identifier long_right_associative_precedence_9_identifier
%type <e> infix_expression
%type <e> colon_expression
%type <e> andalso_expression orelse_expression
%type <e> match_expression raise_expression
%type <e> closed_match match match_rule
%type <e> expression semicolon
%type <e> optional_declaration_list declaration_list closed_declaration_list
%type <e> declaration load_declaration simple_declaration fixity_declaration
%type <e> nosemi_declaration extern_declaration
%type <e> value_binding nonrec_value_binding simple_value_binding
%type <e> function_value_binding function_value_binding_l
%type <e> closed_function_value_binding optional_with_fun
%type <e> simple_function_value_binding one_function_value_binding
%type <e> atomic_pattern_list infix_function_header
%type <e> dimension_binding dimension_binding_l simple_dimension_binding
%type <e> scale_binding scale_binding_l simple_scale_binding
%type <e> type_binding type_binding_l simple_type_binding
%type <e> optional_withtype withtype
%type <e> type_variable_sequence type_variable_sequence_l type_variable
%type <e> datatype_binding simple_datatype_binding
%type <e> constructor_binding simple_constructor_binding symbol_list
%type <e> optional_of_type of_type
%type <e> exception_binding exception_binding_l simple_exception_binding
%type <e> optional_type_row type_row labeled_type
%type <e> atomic_type constructor_type
%type <e> simple_dimension_type dimension_type
%type <e> tuple_type tuple_type_l function_type
%type <e> type type_list
%type <e> optional_colon_type colon_type
%type <e> optional_pattern_row pattern_row labeled_pattern
%type <e> optional_as_pattern as_pattern
%type <e> /*optional_pattern_list*/ pattern_list
%type <e> optional_pattern_l_list pattern_l_list
%type <e> optional_union_pattern union_pattern
%type <e> maplet_pattern maplet_pattern_list
%type <e> atomic_pattern application_pattern pack_pattern
%type <e> left_associative_precedence_0_pattern right_associative_precedence_0_pattern
%type <e> left_associative_precedence_1_pattern right_associative_precedence_1_pattern
%type <e> left_associative_precedence_2_pattern right_associative_precedence_2_pattern
%type <e> left_associative_precedence_3_pattern right_associative_precedence_3_pattern
%type <e> left_associative_precedence_4_pattern right_associative_precedence_4_pattern
%type <e> left_associative_precedence_5_pattern right_associative_precedence_5_pattern
%type <e> left_associative_precedence_6_pattern right_associative_precedence_6_pattern
%type <e> left_associative_precedence_7_pattern right_associative_precedence_7_pattern
%type <e> left_associative_precedence_8_pattern right_associative_precedence_8_pattern
%type <e> left_associative_precedence_9_pattern right_associative_precedence_9_pattern
%type <e> long_left_associative_precedence_0_constructor long_right_associative_precedence_0_constructor
%type <e> long_left_associative_precedence_1_constructor long_right_associative_precedence_1_constructor
%type <e> long_left_associative_precedence_2_constructor long_right_associative_precedence_2_constructor
%type <e> long_left_associative_precedence_3_constructor long_right_associative_precedence_3_constructor
%type <e> long_left_associative_precedence_4_constructor long_right_associative_precedence_4_constructor
%type <e> long_left_associative_precedence_5_constructor long_right_associative_precedence_5_constructor
%type <e> long_left_associative_precedence_6_constructor long_right_associative_precedence_6_constructor
%type <e> long_left_associative_precedence_7_constructor long_right_associative_precedence_7_constructor
%type <e> long_left_associative_precedence_8_constructor long_right_associative_precedence_8_constructor
%type <e> long_left_associative_precedence_9_constructor long_right_associative_precedence_9_constructor
%type <e> infix_pattern colon_pattern pattern
%type <e> optional_ellipsis_type ellipsis_type
%type <e> infix_identifier optional_eqtype eqtype_desc simple_eqtype_desc
%type <i> quantifier bar fn
%type <e> main toplevel_declaration

%type <e> value_description value_description_l
%type <e> simple_value_description type_description type_description_l
%type <e> simple_type_description
/*%type <e> exception_description exception_description_l simple_exception_description*/

%expect 2

%start main

%%

main : toplevel_declaration s_S
	{
	  theSyntax = buildrev(&@1,&@2,A_DECL_LIST,$1);
	  YYACCEPT;
	}
	;

	/* a toplevel_declaration is a reversed list of bindings or NIL */
toplevel_declaration : expression /* val it = expr; only form of syntax allowed
				     in debug mode. */
		{ $$ = exprToDecls(&@1,&@1,$1,ItSymbol);
		}
	| toplevel_declaration nosemi_declaration
		{ checkDebug(&@1,&@2);
		    $$ = Nappend($2,$1); }
/*	| toplevel_declaration kw_signature signature_binding
		{ checkDebug(&@1,&@3); $$ = cons($3,$1); }
*/
	|
	   { $$ = NIL; }
	;

load_declaration : t_open string_constant
		{
		  $$ = do_load (&@1, &@2, yytext, true);
		}
		| s_openF string_constant
		{
		  $$ = do_load (&@1, &@2, yytext, false);
		}
		| kw_use string_constant
		{
			ExprPtr name;

			(void) getString(&@2,&@2,(CHAR *)yytext,false);
			if (!OpenFile((CHAR *)yytext))
			  {
			    char *buf = alloctmp_anchored(strlen(yytext)+64,(char *)&name);

			    sprintf(buf,message(389),yytext);
			    {
			      TZone *zone = (TZone *)alloctmp_anchored(sizeof(TZone),(char *)&name);

			      makeZone(zone,&@1,&@2);
			      type_error(zone,(CHAR *)buf);
			    }
			  }
			name = Kstr(yyfilename);
			(void) KEEP(name);
			$$ = list1(build1(&@1,&@2,A_ENTER_FILE,name));
		}
		;

sharp : s_H { pushMarker("#", NULL); }
	;

opening_record : s_IB { pushMarker("|[", "]|"); }
	;

opening_parenthesis : s_P { pushMarker("(", ")"); }
	;

opening_bracket : s_B { pushMarker("[", "]"); }
	;

opening_brace : s_Y { $$ = 0; pushMarker("{", "}"); }
| s_LY { $$ = 1; pushMarker("<{", "}"); }
	;

quantifier : kw_iterate { pushMarker("iterate", "end "); $$ = 0; }
        | kw_all { pushMarker("all", "end "); $$ = 1; }
        | kw_exists { pushMarker("exists", "end "); $$ = 2; }
        | kw_some { pushMarker("some", "end "); $$ = 3; }
	;

bar : s_I { $$ = 0; }
	| s_II { $$ = 1; }
	;

atomic_expression : special_constant
	| long_id_or_constructor
	| opening_record optional_expression_row s_CI
	{ $$ = popMarker(BUILD(&@1,&@3,A_RECORD,$2)); }
	| sharp label
	{ $$ = popMarker(BUILD(&@1,&@2,A_SELECT_FN,ARGS($2))); }
	| sharp integer_constant
	{ $$ = popMarker(build1(&@1,&@2,A_SELECT_FN,getInt((CHAR *)yytext, &@2))); }
	| sharp precedence_constant
	{ $$ = popMarker(build1(&@1,&@2,A_SELECT_FN,getInt((CHAR *)yytext, &@2))); }
	| opening_parenthesis optional_expression_list s_Q
		{ if ($2!=NIL && CDR($2)==NIL)
			{ $$ = CAR($2); }
			else
				$$ = BUILD(&@1,&@3,A_TUPLE,$2);
			$$ = popMarker($$); }
	| opening_parenthesis expression s_S
		expression_sequence s_Q
		{ $$ = popMarker(build1rev(&@1,&@5,A_PROGN,$2,$4)); }
	| opening_bracket s_C
		{ $$ = popMarker(BUILD(&@1,&@2,A_LIST,NIL)); }
	| opening_bracket expression s_C
		{ $$ = popMarker(build1(&@1,&@3,A_LIST,$2)); }
	| opening_bracket expression t_comma
		expression_list s_C
		{ $$ = popMarker(build1rev(&@1,&@5,A_LIST,$2,$4)); }
	| opening_bracket expression bar
		domain_description s_C
		{ $$ = popMarker(build1list(&@1,&@5,
			A_LIST_OF+$3,$2,$4)); }
	| opening_bracket expression bar /* should be s_I really */
		kw_while expression optional_such_that s_C
		{ $$ = popMarker(build3(&@1,&@7,A_WHILE_LIST,$5,$2,$6)); }
        | s_BI { pushMarker("[|", "|]"); } optional_expression_list s_IC
		{ $$ = Kstr((CHAR *)message(357)); (void) KEEP($$);
		  $$ = popMarker(build2(&@1,&@4,A_APPLY,
					build1(&@1,&@4,A_VARIABLE,$$),
					BUILD(&@1,&@4,A_LIST,$3)));
		}
	| opening_brace s_Z
		{ $$ = popMarker(BUILD(&@1,&@2,A_MAP,NIL)); }
	| opening_brace maplet s_Z
		{ $$ = popMarker(build1(&@1,&@3,A_MAP,$2)); }
	| opening_brace maplet t_comma
		maplet_list s_Z
		{ $$ = popMarker(build1rev(&@1,&@5,$1?A_MAP_UNDER:A_MAP,$2,$4)); }
	| opening_brace maplet bar
		domain_description s_Z
		{ $$ = popMarker(build1list(&@1,&@5,
			($1?A_MAP_UNDER_OF:A_MAP_OF)+$3,
			$2,$4)); }
	| opening_brace maplet bar /* should be s_I really */
		kw_while expression optional_such_that s_Z
		{ $$ = popMarker(build3(&@1,&@7,
			($1?A_WHILE_MAP_UNDER:A_WHILE_MAP),$5,$2,$6)); }
	| quantifier expression bar domain_description kw_end
		{ $$ = popMarker(build1list(&@1,&@5,quantifier[$1][$3],$2,$4)); }
	| quantifier expression bar /* should be s_I really */
		kw_while expression optional_such_that kw_end
		{ $$ = popMarker(build3(&@1,&@7,qwhile[$1],$5,$2,$6)); }
	| let_begin
		closed_declaration_list kw_in
		implicit_progn let_end
		{ $$ = popMarker(build2(&@1,&@5,A_LET,$2,$4)); }
	;

	/* maplet_lists are reversed lists of maplets */
maplet_list : maplet
		{ $$ = list1($1); }
	| maplet_list t_comma maplet
		{ $$ = cons($3,$1); }
	;

maplet : expression
	{ $$ = build2(&@1,&@1,A_MAPLET,$1,zeroTuple(&@1)); }
	| pure_maplet
	;

pure_maplet : expression s_EG expression
	{ $$ = build2(&@1,&@3,A_MAPLET,$1,$3); }
	;

	/* domain_descriptions are couples (such_that,list) */
domain_description : { pushMarker(NULL, NULL); }
	domain_sweep_list optional_declaration_list optional_such_that
	{ $$ = popMarker(cons($4,Nreverse1($2,ARGS($3)))); }
	;

	/* domain_sweep_lists are reversed lists */
domain_sweep_list : domain_sweep
	{ $$ = list1($1); }
	| domain_sweep_list kw_and domain_sweep
	{ $$ = cons($3,$1); }
	;

domain_sweep : pattern kw_in_set expression
	{ $$ = build2(&@1,&@3,A_IN_SET,
			build2(&@1,&@1,A_PAT_MAPLET,$1,
				BUILD(&@1,&@1,A_PAT_ANY,NIL)),
			$3); }
	| maplet_pattern kw_in_map expression
	{ $$ = build2(&@1,&@3,A_IN_SET,$1,$3); }
	| pattern kw_in_list expression
	{ $$ = build2(&@1,&@3,A_IN_LIST,$1,$3); }
	| pattern kw_sub_map expression
	{ $$ = build2(&@1,&@3,A_SUB_MAP,$1,$3); }
	;

optional_such_that : such_that
	| empty
	{ $$ = build1(&@1,&@1,A_VARIABLE,trueSym); }
	;

empty :
	;

such_that : kw_such_that expression { $$ = $2; }
	;

let_begin : kw_let { pushEnv(); pushMarker("let", "end "); }
	;

let_end : kw_end { popEnv(); }
	;

implicit_progn : expression_sequence
	{ if (CDR($1)==NIL)
	$$ = CAR($1);
	else
	$$ = buildrev(&@1,&@1,A_PROGN,$1); }
	;

	/* expression sequences are reversed lists of expressions */
expression_sequence : expression
	{ $$ = list1($1); }
	| expression_sequence s_S expression
	{ $$ = cons($3,$1); }
	;

	/* expression lists are lists of expressions */
optional_expression_list : expression_list
	{ $$ = Nreverse($1); }
	| { $$ = NIL; }
	;

	/* expression lists are reversed lists of expressions */
expression_list : expression
	{ $$ = list1($1); }
	| expression_list t_comma expression
	{ $$ = cons($3,$1); }
	;

	/* optional_expression_rows are lists of record maplets */
optional_expression_row : expression_row
	{ $$ = Nreverse($1); }
	| { $$ = NIL; }
	;

	/* expression_rows are reversed lists of record maplets */
expression_row : labeled_expression
	{ $$ = list1($1); }
	| expression_row t_comma labeled_expression
	{ checkUniqueLabel($3,$1);
	$$ = cons($3,$1); }
	;

labeled_expression : label s_E expression
	{ $$ = build2(&@1,&@3,A_RECORD_MAPLET,$1,$3); }
	;

label : label_id
	{ checkLabel(&@1);
		$$ = build1(&@1,&@1,A_LABEL,lexId); }
	;

symbol : label_id
	{ $$ = lexVar(&@1,false); }
	;

label_id : t_identifier
	| t_constructor
	| infix_identifier_1
	| t_left_associative_precedence_0_constructor
	| t_right_associative_precedence_0_constructor
	| t_left_associative_precedence_1_constructor
	| t_right_associative_precedence_1_constructor
	| t_left_associative_precedence_2_constructor
	| t_right_associative_precedence_2_constructor
	| t_left_associative_precedence_3_constructor
	| t_right_associative_precedence_3_constructor
	| t_left_associative_precedence_4_constructor
	| t_right_associative_precedence_4_constructor
	| t_left_associative_precedence_5_constructor
	| t_right_associative_precedence_5_constructor
	| t_left_associative_precedence_6_constructor
	| t_right_associative_precedence_6_constructor
	| t_left_associative_precedence_7_constructor
	| t_right_associative_precedence_7_constructor
	| t_left_associative_precedence_8_constructor
	| t_right_associative_precedence_8_constructor
	| t_left_associative_precedence_9_constructor
	| t_right_associative_precedence_9_constructor
	;

application_expression : atomic_expression
	| application_expression atomic_expression
	{ $$ = build2(&@1,&@2,A_APPLY,$1,$2); }
        | application_expression s_AAIB { pushMarker("++|[", "]|"); }
		optional_expression_row s_CI
		{ $$ = popMarker(build2(&@1,&@5,A_MU,
				 $1,BUILD(&@2,&@5,A_RECORD,$4))); }
	| array_expression
	;

array_expression : application_expression s_DP { pushMarker(".(", ")"); }
			expression s_Q
		{ $$ = Kstr((CHAR *)message(354)); (void) KEEP($$);
		  $$ = popMarker(applyBin(&@1,&@5,
					build1(&@2,&@2,A_VARIABLE,$$),
					$1,$4));
		}
	;

delay_expression : application_expression
        | kw_delay { pushMarker("delay", NULL); }
	delay_expression
	{ $$ = popMarker(build1(&@1,&@3,A_DELAY,$3)); }
        | kw_pack { pushMarker("pack", NULL); }
	delay_expression
	{ $$ = popMarker(build1(&@1,&@3,A_PACK,$3)); }
	;

left_associative_precedence_9_expression : delay_expression
	| left_associative_precedence_9_expression long_left_associative_precedence_9_identifier delay_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_9_identifier : t_left_associative_precedence_9_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_9_constructor
		;

right_associative_precedence_9_expression : left_associative_precedence_9_expression
	| left_associative_precedence_9_expression long_right_associative_precedence_9_identifier right_associative_precedence_9_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_9_identifier : t_right_associative_precedence_9_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_9_constructor
		;

left_associative_precedence_8_expression : right_associative_precedence_9_expression
	| left_associative_precedence_8_expression long_left_associative_precedence_8_identifier right_associative_precedence_9_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_8_identifier : t_left_associative_precedence_8_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_8_constructor
		;

right_associative_precedence_8_expression : left_associative_precedence_8_expression
	| left_associative_precedence_8_expression long_right_associative_precedence_8_identifier right_associative_precedence_8_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_8_identifier : t_right_associative_precedence_8_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_8_constructor
		;

left_associative_precedence_7_expression : right_associative_precedence_8_expression
	| left_associative_precedence_7_expression long_left_associative_precedence_7_identifier right_associative_precedence_8_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_7_identifier : t_left_associative_precedence_7_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_7_constructor
		;

right_associative_precedence_7_expression : left_associative_precedence_7_expression
	| left_associative_precedence_7_expression long_right_associative_precedence_7_identifier right_associative_precedence_7_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_7_identifier : t_right_associative_precedence_7_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_7_constructor
		;

left_associative_precedence_6_expression : right_associative_precedence_7_expression
	| left_associative_precedence_6_expression long_left_associative_precedence_6_identifier right_associative_precedence_7_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_6_identifier : t_left_associative_precedence_6_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_6_constructor
		;

right_associative_precedence_6_expression : left_associative_precedence_6_expression
	| left_associative_precedence_6_expression long_right_associative_precedence_6_identifier right_associative_precedence_6_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_6_identifier : t_right_associative_precedence_6_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_6_constructor
		;

left_associative_precedence_5_expression : right_associative_precedence_6_expression
	| left_associative_precedence_5_expression long_left_associative_precedence_5_identifier right_associative_precedence_6_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_5_identifier : t_left_associative_precedence_5_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_5_constructor
		;

right_associative_precedence_5_expression : left_associative_precedence_5_expression
	| left_associative_precedence_5_expression long_right_associative_precedence_5_identifier right_associative_precedence_5_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_5_identifier : t_right_associative_precedence_5_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_5_constructor
		;

left_associative_precedence_4_expression : right_associative_precedence_5_expression
	| left_associative_precedence_4_expression long_left_associative_precedence_4_identifier right_associative_precedence_5_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_4_identifier : t_left_associative_precedence_4_identifier
		{ $$ = lexVar(&@1,true); }
	| s_E
		{ SETQ(lexId,equalSym);
		  SETQ(lexPath,NIL);
		  $$ = lexVar(&@1,true); }
	| long_left_associative_precedence_4_constructor
	;

right_associative_precedence_4_expression : left_associative_precedence_4_expression
	| left_associative_precedence_4_expression long_right_associative_precedence_4_identifier right_associative_precedence_4_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_4_identifier : t_right_associative_precedence_4_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_4_constructor
		;

left_associative_precedence_3_expression : right_associative_precedence_4_expression
	| left_associative_precedence_3_expression long_left_associative_precedence_3_identifier right_associative_precedence_4_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	| array_expression s_DTE right_associative_precedence_4_expression
	{ $$ = Kstr((CHAR *)message(355)); (void) KEEP($$);
	  $$ = build2(&@1,&@3,A_APPLY,
		      build1(&@1,&@3,A_VARIABLE,$$),
		      build3(&@1,&@3,A_TUPLE,
				CAR(ARGS(CAR(CDR(ARGS($1))))),
				CAR(CDR(ARGS(CAR(CDR(ARGS($1)))))),
				$3));
	}
	;

long_left_associative_precedence_3_identifier : t_left_associative_precedence_3_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_3_constructor
		;

right_associative_precedence_3_expression : left_associative_precedence_3_expression
	| left_associative_precedence_3_expression long_right_associative_precedence_3_identifier right_associative_precedence_3_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_3_identifier : t_right_associative_precedence_3_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_3_constructor
		;

left_associative_precedence_2_expression : right_associative_precedence_3_expression
	| left_associative_precedence_2_expression long_left_associative_precedence_2_identifier right_associative_precedence_3_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_2_identifier : t_left_associative_precedence_2_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_2_constructor
		;

right_associative_precedence_2_expression : left_associative_precedence_2_expression
	| left_associative_precedence_2_expression long_right_associative_precedence_2_identifier right_associative_precedence_2_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_2_identifier : t_right_associative_precedence_2_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_2_constructor
		;

left_associative_precedence_1_expression : right_associative_precedence_2_expression
	| left_associative_precedence_1_expression long_left_associative_precedence_1_identifier right_associative_precedence_2_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_1_identifier : t_left_associative_precedence_1_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_1_constructor
		;

right_associative_precedence_1_expression : left_associative_precedence_1_expression
	| left_associative_precedence_1_expression long_right_associative_precedence_1_identifier right_associative_precedence_1_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_1_identifier : t_right_associative_precedence_1_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_1_constructor
		;

left_associative_precedence_0_expression : right_associative_precedence_1_expression
	| left_associative_precedence_0_expression long_left_associative_precedence_0_identifier right_associative_precedence_1_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_0_identifier : t_left_associative_precedence_0_identifier
		{ $$ = lexVar(&@1,true); }
		| long_left_associative_precedence_0_constructor
		;

right_associative_precedence_0_expression : left_associative_precedence_0_expression
	| left_associative_precedence_0_expression long_right_associative_precedence_0_identifier right_associative_precedence_0_expression
	{ $$ = applyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_0_identifier : t_right_associative_precedence_0_identifier
		{ $$ = lexVar(&@1,true); }
		| long_right_associative_precedence_0_constructor
		;

infix_expression : right_associative_precedence_0_expression
	;

colon_expression : infix_expression
	| colon_expression colon_type
	{ $$ = build2(&@1,&@2,A_CAST,$1,$2); }
	;

andalso_expression : colon_expression
	| colon_expression kw_andalso andalso_expression
	{ $$ = build2(&@1,&@3,A_AND,$1,$3); }
	;

orelse_expression : andalso_expression
	| andalso_expression kw_orelse orelse_expression
	{ $$ = build2(&@1,&@3,A_OR,$1,$3); }
	;

match_expression : orelse_expression
	| orelse_expression
          kw_handle { pushMarker("handle", NULL); }
	closed_match /* shift-reduce on vbar */
	{ TYPE($4) = A_MATCH;
	$$ = popMarker(build2(&@1,&@4,A_HANDLE,$1,$4)); }
	;

raise_expression : match_expression
	| kw_raise raise_expression
	{ $$ = build1(&@1,&@2,A_RAISE,$2); }
	;

while : kw_while { pushMarker("while", NULL); }
	;

expression : raise_expression
        | kw_if {pushMarker("if", NULL);} expression
	kw_then expression
	kw_else expression
	{ $$ = popMarker(build3(&@1,&@7,A_IF,$3,$5,$7)); }
	| while expression
	kw_do expression
	{ $$ = popMarker(build3(&@1,&@4,A_WHILE,$2,$4,
				build1(&@1,&@4,A_VARIABLE,trueSym))); }
        | kw_case {pushMarker("case", NULL);} expression
	kw_of closed_match /* shift-reduce on vbar */
	{ $$ = popMarker(build2(&@1,&@5,A_APPLY,$5,$3)); }
        | fn {pushMarker("fn", NULL);} match /* shift-reduce on vbar */
	{ $$ = popMarker(buildrev(&@1,&@3,$1,$3)); }
	;

fn : kw_fn { $$ = A_FN; }
	| kw_memofn { $$ = A_MEMO_FN; }
	;

closed_match : match
	{ $$ = buildrev(&@1,&@1,A_FN,$1); }
	;

	/* matches are reversed lists of match_rules */
match : match_rule
	{ $$ = list1($1); }
	| match s_I match_rule /* shift-reduce on s_I */
	{ $$ = cons($3,$1); }
	;

match_rule : { pushMarker(NULL,NULL); }
	pattern s_EG expression
	{ $$ = popMarker(build2(&@1,&@4,A_LAMBDA,$2,$4)); }
	;

special_constant : numerical_constant
	| string_constant
	{ $$ = build1(&@1,&@1,A_STRING,
		      getString(&@1,&@1,(CHAR *)yytext,true)); }
	;

numerical_constant : number {unit_recognized=true;}
			optional_scale
		{ unit_recognized = false;
		  $$ = $1;
		  if ($3!=NIL)
		    $$ = build2(&@1,&@3,A_SET_UNIT,$$,$3);
		}
	| integer {unit_recognized=true;}
			optional_scale
		{ unit_recognized = false;
		  $$ = $1;
		  if ($3!=NIL)
		    {
		      TYPE($$) = A_NUMBER;
		      $$ = build2(&@1,&@3,A_SET_UNIT,$$,$3);
		    }
		}
	;

optional_scale : t_backtick scale { $$ = $2; }
	| { $$ = NIL; }
	;

scale : simple_scale
	| scale t_backtick simple_scale
	{ $$ = build2(&@1,&@3,A_UNIT_STAR,$1,$3); }
	;

simple_scale : scale_id
	| scale_id s_X number_or_int
	{ $$ = build2(&@1,&@3,A_UNIT_POWER,$1,$3); }
	;

scale_id : label_id
	{ $$ = lexVar(&@1,true);
	  TYPE($$) = A_UNIT_NAME; }
	;

number : real
	| complex_constant
	{ $$ = makeNum(&@1,getComplex(&@1,&@1,(CHAR *)yytext)); }
	;

number_or_int : integer { TYPE($$) = A_NUMBER; }
     	      | number
	      ;

real : real_constant
	{ $$ = makeNum(&@1,getReal(&@1,&@1,(CHAR *)yytext)); }
	| s_Ainf
	{ $$ = makeSpecNum(&@1,pinf); }
	| s_Ninf
	{ $$ = makeSpecNum(&@1,minf); }
	| s_nan0
	{ $$ = makeSpecNum(&@1,NaNs[0+MAX_NAN]); }
	| s_nan1
	{ $$ = makeSpecNum(&@1,NaNs[1+MAX_NAN]); }
	| s_nan2
	{ $$ = makeSpecNum(&@1,NaNs[2+MAX_NAN]); }
	| s_nan3
	{ $$ = makeSpecNum(&@1,NaNs[3+MAX_NAN]); }
	| s_nan4
	{ $$ = makeSpecNum(&@1,NaNs[4+MAX_NAN]); }
	| s_nan5
	{ $$ = makeSpecNum(&@1,NaNs[5+MAX_NAN]); }
	| s_nan6
	{ $$ = makeSpecNum(&@1,NaNs[6+MAX_NAN]); }
	| s_nan7
	{ $$ = makeSpecNum(&@1,NaNs[7+MAX_NAN]); }
	| s_Nnan1
	{ $$ = makeSpecNum(&@1,NaNs[-1+MAX_NAN]); }
	| s_Nnan2
	{ $$ = makeSpecNum(&@1,NaNs[-2+MAX_NAN]); }
	| s_Nnan3
	{ $$ = makeSpecNum(&@1,NaNs[-3+MAX_NAN]); }
	| s_Nnan4
	{ $$ = makeSpecNum(&@1,NaNs[-4+MAX_NAN]); }
	| s_Nnan5
	{ $$ = makeSpecNum(&@1,NaNs[-5+MAX_NAN]); }
	| s_Nnan6
	{ $$ = makeSpecNum(&@1,NaNs[-6+MAX_NAN]); }
	| s_Nnan7
	{ $$ = makeSpecNum(&@1,NaNs[-7+MAX_NAN]); }
	;

integer : integer_constant
	{ $$ = makeInt(&@1,getReal(&@1,&@1,(CHAR *)yytext)); }
	| hex_integer_constant
	{ $$ = makeInt(&@1, getInt((CHAR *)yytext, &@1)); }
	| octal_integer_constant
	{ $$ = makeInt(&@1, getInt((CHAR *)yytext, &@1)); }
	| binary_integer_constant
	{ $$ = makeInt(&@1, getInt((CHAR *)yytext, &@1)); }
	| precedence_constant
	{ $$ = makeInt(&@1,getReal(&@1,&@1,(CHAR *)yytext)); }
	;

identifier : t_identifier
		       	{ $$ = lexVar(&@1,false); }
			;

long_constructor : t_constructor
						{ $$ = lexVar(&@1,true); }
					;

long_id_or_constructor : t_identifier
							{ $$ = lexVar(&@1,true); }
						| t_constructor
							{ $$ = lexVar(&@1,true); }
						;

/*
long_symbol : label_id
				{ $$ = lexVar(&@1,true); }
			;

constructor : t_constructor
				{ $$ = lexVar(&@1,false); }
	;
*/

optional_declaration_list : closed_declaration_list
	| empty
	{ $$ = BUILD(&@1,&@1,A_DECL_LIST,NIL); }
	;

closed_declaration_list : declaration_list
	{ $$ = buildrev(&@1,&@1,A_DECL_LIST,$1); }
	;

	/* a declaration_list is a reversed list of bindings */
declaration_list : declaration
	| declaration_list declaration
	{ $$ = Nappend($2,$1); }
	;

semicolon : s_S {
		if (wantCloseFile)
		{
		  wantCloseFile = false;
		  CloseFile();
		  if (yyfilename==NULL)
		    $$ = Build(&@1,&@1,A_EXIT_FILE,NIL);
		  else
		    {
		      ExprPtr name;
		      
		      name = Kstr(yyfilename);
		      (void) KEEP(name);
		      $$ = build1(&@1,&@1,A_ENTER_FILE,name);
		    }
		  $$ = list1($$);
		}
		else $$ = NIL;
		}
;

	/* a declaration is a reversed list of bindings */
declaration : nosemi_declaration
	| semicolon
	;

	/* a nosemi_declaration is a reversed list of bindings */
nosemi_declaration : { pushMarker(NULL,NULL); }
	simple_declaration
	{ $$ = popMarker($2); }
        | local_begin { pushMarker("local", "end "); }
	optional_declaration_list kw_in
	{ pushEnv(); }
	optional_declaration_list
	{ setLocalEnv(); popEnv(); }
	local_end
	{ $$ = popMarker(list1(build2(&@1,&@8,A_LOCAL,$3,$6)));
	}
	| kw_extern extern_declaration { checkDebug(&@1,&@2); $$ = $2; }
	;

local_begin : kw_local { pushEnv(); }
	;

local_end : kw_end { popEnv(); }
	;

	/* a simple_declaration is a reversed list of bindings */
simple_declaration : kw_val value_binding
	{ $$ = list1($2); }
	| kw_fun function_value_binding
	{ $$ = list1($2); }
	| kw_memofun function_value_binding
		optional_with_fun
	{ $$ = list1($2); patchMemoFunList(ARGS($2),CAR($3),CDR($3)); }
	| kw_dimension dimension_binding
	{ $$ = list1($2); }
	| kw_scale scale_binding
	{ $$ = list1($2); }
	| kw_type type_binding
	{ $$ = list1($2); }
	| kw_datatype datatype_binding optional_withtype
	{ ExprPtr dtb,prime;
	ENTER;
	dtb = Nreverse($2);
	prime = build1list(&@2,&@3,A_PRIME_DATABIND,$3,dtb);
	checkWithType(dtb,$3);
	$$ = list1(build1(&@1,&@3,A_DATATYPE,prime));
	/* $3 is shared ! */
	if (ARGS($3)!=NIL)
	$$ = cons($3,$$);
	KEEPONLY($$); }
	| kw_abstype {pushEnv(); }
	datatype_binding {pushEnv();}
	optional_withtype
	kw_with
	optional_eqtype
	declaration_list
	kw_end
	{ ExprPtr dtb,prime,declist;
	ENTER;
	setLocalEnv(); popEnv(); popEnv();
	checkEqTypes($7,$3);
	dtb = Nreverse($3);
	prime = build1list(&@3,&@5,A_PRIME_DATABIND,$5,dtb);
	checkWithType(dtb,$5);
	declist = Nreverse($8);
	declist = (ARGS($5)!=NIL)?cons($5,declist):declist;
	/* $5 is shared ! */
	declist = BUILD(&@5,&@8,A_DECL_LIST,declist);
	$$ = list1(build2list(&@1,&@9,A_ABSTYPE,prime,declist,$7));
	KEEPONLY($$); }
	| kw_exception exception_binding
	{ $$ = list1($2); }
	| fixity_declaration
/*	| t_open long_symbol_list
		{ $$ = buildrev(&@1,&@2,A_OPEN_DECL,$2);
			longLexOpenList(ARGS($$),false);
			$$ = list1($$); }
*/
	| load_declaration
	| t_pragma
	{
	$$ = Kstr((CHAR *)yytext);
	(void) KEEP($$);
	$$ = list1(build1(&@1,&@1,A_PRAGMA,$$));
	}
        | t_cpp_line_control_directive
	{ $$ = do_cpp_line_directive(&@1); }
	;

extern_declaration : kw_datatype type_description
			{ TYPE($2) = A_EXT_DATATYPE; $$ = list1($2); }
		| kw_eqtype type_description
			{ TYPE($2) = A_EXT_EQTYPE; $$ = list1($2); }
		| kw_dimension dimension_binding
			{ TYPE($2) = A_EXT_QUANTITY; $$ = list1($2); }
		| kw_val value_description
			{ TYPE($2) = A_EXT_VAL; $$ = list1($2); }
		/*| kw_exception exception_description
			{ TYPE($2) = A_EXT_EXC; $$ = list1($2); }*/
		;

	/* optional_with_fun returns (tyvarseq,functionbindings) */
optional_with_fun : kw_with function_value_binding { $$ = ARGS($2); }
		| empty { $$ = list1(BUILD(&@1,&@1,A_TYPE_VAR_SEQ,NIL)); }
;

fixity_declaration : kw_infix optional_precedence symbol_list
						{ $$ = setFix(&@1,&@3,$3,(COUNTER)($2+1)); }
					| kw_infixr optional_precedence symbol_list
						{ $$ = setFix(&@1,&@3,$3,(COUNTER)(-$2-1)); }
					| kw_nonfix symbol_list
						{ $$ = setFix(&@1,&@2,$2,(COUNTER)0); }
					;

optional_withtype : withtype
	| empty { $$ = BUILD(&@1,&@1,A_TYPE,NIL); }
	;

withtype : kw_withtype type_binding
	{ $$ = $2; }
	;

value_binding : nonrec_value_binding
	{ ExprPtr tyvars;
	tyvars = BUILD(&@1,&@1,A_TYPE_VAR_SEQ,
	listTyVars($1,NIL));
	/* list of unguarded tyvars, not of scoped tyvars */
	$$ = build1rev(&@1,&@1,A_VAL,tyvars,$1); }
	| kw_rec nonrec_value_binding
	{ ExprPtr tyvars;
	tyvars = BUILD(&@2,&@2,A_TYPE_VAR_SEQ,
	listTyVars($2,NIL));
	/* list of unguarded tyvars, not of scoped tyvars */
	$$ = build1rev(&@1,&@2,A_VALREC,tyvars,$2); }
	;

	/* nonrec_value_bindings are reversed lists of bindings */
nonrec_value_binding : simple_value_binding
	{ $$ = list1($1); }
	| nonrec_value_binding kw_and simple_value_binding
	{ /*checkUniqueVal($3,$1);*/
		$$ = cons($3,$1); }
	;

simple_value_binding : pattern s_E expression
	{ $$ = build2(&@1,&@3,A_BIND,$1,$3); }
	;

function_value_binding : function_value_binding_l
	{ ExprPtr tyvars;
	tyvars = BUILD(&@1,&@1,A_TYPE_VAR_SEQ,
	listTyVars($1,NIL));
	/* list of unguarded tyvars, not of scoped tyvars */
	$$ = build1rev(&@1,&@1,A_VALREC,tyvars,$1); }
	;

	/* function_value_binding_ls are reversed lists of bindings */
function_value_binding_l : closed_function_value_binding
	{ $$ = list1($1); }
	| function_value_binding_l kw_and
	closed_function_value_binding
	{ /*checkUniqueVal($3,$1);*/
	$$ = cons($3,$1); }
	;

closed_function_value_binding : simple_function_value_binding
	{ /*checkUniqueValInFunBind($1);*/
	$$ = buildFunBind(&@1,&@1,$1); }
	;

	/* simple_function_value_bindings are reversed lists of
	zonex::f::revpatlist::expr */
simple_function_value_binding : one_function_value_binding
	{ $$ = list1(cons(BUILD(&@1,&@1,-1,NIL),
	$1)); }
	| simple_function_value_binding s_I
	/* shift-reduce on s_I */
	one_function_value_binding
	{ ENTER;
	if (CAR(ARGS(CAR($3)))!=
	CAR(ARGS(CAR(CDR(CAR($1))))))
	type_error(ZONE(CAR($3)),(CHAR *)message(390));
	if (ListLength(CAR(CDR($3)))!=ListLength(CAR(CDR(CDR(CAR($1))))))
	type_error(ZONE(CAR($3)),(CHAR *)message(391));
	$$ = cons(cons(BUILD(&@3,&@3,-1,NIL),
	$3),
	$1);
	KEEPONLY($$); }
	;

	/* one_function_value_bindings are f::revpatlist::expr */
one_function_value_binding : identifier atomic_pattern_list
	optional_colon_type
	s_E expression
	{ ExprPtr expr;
	ENTER;
	TYPE($1) = A_REAL_VAR;
	if ($3==NIL)
		{ expr = $5; }
	else
		expr = build2(&@3,&@5,A_CAST,$5,$3);
	$$ = cons($1,cons($2,expr));
	KEEPONLY($$); }
	| infix_function_header
	optional_colon_type
	s_E expression
	{ ExprPtr expr;
	ENTER;
	if ($2==NIL)
	{ expr = $4; }
	else
	expr = build2(&@2,&@4,A_CAST,$4,$2);
	$$ = cons(CAR($1),
	cons(list1(CDR($1)),
	expr));
	KEEPONLY($$); }
	| opening_parenthesis
	infix_function_header
	s_Q
	optional_colon_type
	s_E expression
	{ ExprPtr expr;
	if ($4==NIL)
	{ expr = $6; }
	else
	expr = build2(&@4,&@6,A_CAST,$6,$4);
	$$ = popMarker(cons(CAR($2),
	cons(list1(CDR($2)),
	expr))); }
	| opening_parenthesis
	infix_function_header
	s_Q
	atomic_pattern_list
	optional_colon_type
	s_E expression
	{ ExprPtr expr;
	if ($5==NIL)
	{ expr = $7; }
	else
	expr = build2(&@5,&@7,A_CAST,$7,$5);
	$$ = popMarker(cons(CAR($2),
	cons(Nappend($4,
	list1(CDR($2))),
	expr))); }
	;

	/* infix_function_headers are f::pattern */
infix_function_header : atomic_pattern infix_identifier atomic_pattern
	{ ENTER;
	TYPE($2) = A_REAL_VAR;
	$$ = cons($2,build2(&@1,&@3,A_PAT_TUPLE,
	$1,$3));
	KEEPONLY($$); }
	;

	/* atomic_pattern_lists are reversed lists of patterns */
atomic_pattern_list : atomic_pattern
	{ $$ = list1($1); }
	| atomic_pattern_list atomic_pattern
	{ $$ = cons($2,$1); }
	;

dimension_binding : dimension_binding_l
	{ $$ = buildrev(&@1,&@1,A_QUANTITY,$1); }
	;

	/* dimension_binding_ls are reversed lists of dimension declarations */
dimension_binding_l : simple_dimension_binding
	{ $$ = list1($1); }
	| dimension_binding_l kw_and simple_dimension_binding
	{ checkUniqueQuantityAndUnit($3,$1);
	$$ = cons($3,$1); }
	;

simple_dimension_binding : symbol
	{ $$ = build2(&@1,&@1,A_NEWQUAN,$1,$1); }
	/* beware : $1 is shared ! */
	| symbol s_P symbol s_Q
	{ $$ = build2(&@1,&@4,A_NEWQUAN,$1,$3); }
	;

scale_binding : scale_binding_l
	{ $$ = buildrev(&@1,&@1,A_UNIT,$1); }
	;

	/* scale_binding_ls are reversed lists of scale declarations */
scale_binding_l : simple_scale_binding
	{ $$ = list1($1); }
	| scale_binding_l kw_and simple_scale_binding
	{ checkUniqueUnit($3,$1);
	$$ = cons($3,$1); }
	;

simple_scale_binding : {unit_recognized=true;}
	number_or_int t_backtick symbol s_E numerical_constant
	{ unit_recognized=false;
	  $$ = build3(&@1,&@6,A_NEWUNIT,$4,$2,$6); }
	;

type_binding : type_binding_l
	{ $$ = buildrev(&@1,&@1,A_TYPE,$1); }
	;

	/* type_binding_ls are reversed lists of type bindings */
type_binding_l : simple_type_binding
	{ $$ = list1($1); }
	| type_binding_l kw_and simple_type_binding
	{ checkUniqueIdInType($3,$1);
		$$ = cons($3,$1); }
	;

simple_type_binding : type_variable_sequence symbol
	s_E type
	{ checkQuantifiedType($1,$4);
		$$ = build3(&@1,&@4,A_TYPE_BIND,$1,$2,$4); }
	;

type_variable_sequence : empty { $$ = BUILD(&@1,&@1,A_TYPE_VAR_SEQ,NIL); }
			| type_variable { $$ = build1(&@1,&@1,A_TYPE_VAR_SEQ,$1); }
			| opening_parenthesis type_variable
				type_variable_sequence_l s_Q
				{ checkUniqueTypeSeq($2,$3);
					$$ = build1rev(&@1,&@4,A_TYPE_VAR_SEQ,$2,$3);
					$$ = popMarker($$); }
						;

	/* type_variable_sequence_ls are reversed lists of variables */
type_variable_sequence_l : { $$ = NIL; }
							| type_variable_sequence_l t_comma type_variable
							{ checkUniqueTypeSeq($3,$1);
								$$ = cons($3,$1); }
							;

	/* datatype_bindings are reversed lists of datatype bindings */
datatype_binding : simple_datatype_binding
	{ $$ = list1($1); }
	| datatype_binding kw_and simple_datatype_binding
	{ checkUniqueIdInType($3,$1);
	checkUniqueConstructors($3,$1);
	$$ = cons($3,$1); }
	;

simple_datatype_binding : type_variable_sequence symbol
	s_E constructor_binding
	{ checkQuantifiedDatatype($1,$4);
	$$ = build2rev(&@1,&@4,A_DATATYPE_BIND,
	$1,$2,$4); }
	;

	/* constructor_bindings are reversed lists of constructor bindings */
constructor_binding : simple_constructor_binding
	{ $$ = list1($1); }
	| constructor_binding s_I simple_constructor_binding
	{ checkUniqueCon($3,$1);
	$$ = cons($3,$1); }
	;

/* don't forget to record the new status of the constructor */
simple_constructor_binding : symbol optional_of_type
	{ setConstructor(CAR(ARGS($1)));
	$$ = ($2==NIL)?
	build1(&@1,&@2,A_CONSTRUCTOR,$1):
	build2(&@1,&@2,A_CONSTRUCTOR,$1,$2); }
	;

optional_of_type : of_type
	| { $$ = NIL; }
	;

of_type : kw_of type
	{ $$ = $2; }
	;

	/* optional_eqtypes are lists of type names */
optional_eqtype : { $$ = NIL; }
	| kw_eqtype eqtype_desc { $$ = Nreverse($2); }
	;

	/* type_descs are reversed lists of type names */
eqtype_desc : simple_eqtype_desc
	{ $$ = list1($1); }
	| eqtype_desc kw_and simple_eqtype_desc
	{ $$ = cons($3,$1); }
	;

simple_eqtype_desc : symbol
				{ $$ = build2(&@1,&@1,A_EQTYPE_INFO,$1,
						BUILD(&@1,&@1,A_RECORD,NIL)); }
/*#ifdef QUOTIENT_TYPES
			| symbol s_Y
				expression_row s_Z
				{ $$ = build2(&@1,&@4,A_EQTYPE_INFO,$1,
						buildrev(&@2,&@4,A_RECORD,$3)); }
*/
			;

exception_binding : exception_binding_l
	{ $$ = buildrev(&@1,&@1,A_EXCEPTION,$1); }
	;

	/* exception_binding_ls are reversed lists of exception bindings */
exception_binding_l : simple_exception_binding
	{ $$ = list1($1); }
	| exception_binding_l kw_and simple_exception_binding
	{ checkUniqueException($3,$1);
		$$ = cons($3,$1); }
	;

simple_exception_binding : symbol optional_of_type
	{ setConstructor(CAR(ARGS($1)));
	$$ = ($2==NIL)?
	build1(&@1,&@2,A_NEWEXC,$1):
	build2(&@1,&@2,A_NEWEXC,$1,$2); }
	| symbol s_E long_constructor
	{ setConstructor(CAR(ARGS($1)));
	$$ = build2(&@1,&@3,A_EXC_BIND,$1,$3); }
	;

atomic_pattern : s_U
	{ $$ = BUILD(&@1,&@1,A_PAT_ANY,NIL); }
	| special_constant
	{ $$ = build1(&@1,&@1,A_PAT_CONST,$1); }
	| identifier
	{ TYPE($1) = A_REAL_VAR /*A_PAT_VAR*/; }
	| long_constructor
	{ TYPE($1) = A_PAT_VAR; }
	| opening_record optional_pattern_row s_CI
	{ $$ = popMarker(BUILD(&@1,&@3,A_PAT_RECORD,$2)); }
	| opening_parenthesis pattern s_Q
	{ $$ = popMarker($2); }
	| opening_parenthesis s_Q
	{ $$ = popMarker(BUILD(&@1,&@2,A_PAT_TUPLE,NIL)); }
	| opening_parenthesis pattern t_comma pattern_list s_Q
	{ $$ = popMarker(build1list(&@1,&@3,A_PAT_TUPLE,$2,$4)); }
	| opening_bracket optional_pattern_l_list s_C
	{ $$ = popMarker($2); }
	| opening_brace s_Z
	{ $$ = popMarker(BUILD(&@1,&@2,A_PAT_EMPTY,NIL)); }
	| opening_brace maplet_pattern_list
	{ union_recognized++; } s_Z
	optional_union_pattern
	{ union_recognized--;
	installMapPattern(&@5,$2,$5);
	$$ = popMarker($2); }
	;

optional_union_pattern : union_pattern
	| { $$ = NIL; }
	;

maplet_pattern_list : maplet_pattern
	{ $$ = build2(&@1,&@1,A_PAT_MAPADD,$1,NIL);
	/* to be replaced later, by installMapPattern() */
	}
	| s_DDD
	{ $$ = BUILD(&@1,&@1,A_PAT_ANYMAP,NIL); }
	| maplet_pattern t_comma maplet_pattern_list
	{ $$ = build2(&@1,&@3,A_PAT_MAPADD,$1,$3); }
	;

maplet_pattern : pattern
	{ $$ = build2(&@1,&@1,A_PAT_MAPLET,$1,zeroPatTuple(&@1)); }
	| pattern s_EG pattern
	{ $$ = build2(&@1,&@3,A_PAT_MAPLET,$1,$3); }
	;

union_pattern : kw_U atomic_pattern
	{ $$ = $2; }
	;

optional_pattern_row : pattern_row
	| { $$ = NIL; }
	;

optional_pattern_l_list : pattern_l_list
	| empty
	{ $$ = BUILD(&@1,&@1,A_PAT_NIL,NIL); }
	;

	/* pattern_rows are lists of labeled_patterns or ellipsis */
pattern_row : s_DDD
	{ $$ = list1(BUILD(&@1,&@1,
	A_PAT_RECORD_ELLIPSIS,NIL)); }
	| labeled_pattern
	{ $$ = list1($1); }
	| labeled_pattern t_comma pattern_row
	{ checkUniqueLabel($1,$3);
	$$ = cons($1,$3); }
	;

	/* pattern_lists are lists of patterns or ellipsis */
pattern_list : s_DDD
	{ $$ = list1(BUILD(&@1,&@1,
	A_PAT_TUPLE_ELLIPSIS,NIL)); }
	| pattern
	{ $$ = list1($1); }
	| pattern t_comma pattern_list
	{ $$ = cons($1,$3); }
	;

pattern_l_list : s_DDD
	{ $$ = BUILD(&@1,&@1,A_PAT_ANYLIST,NIL); }
	| pattern
	{ $$ = build2(&@1,&@1,A_PAT_APPLY,
	build1(&@1,&@1,A_PAT_VAR,consSym),
	build2(&@1,&@1,A_PAT_TUPLE,
	$1,patNil(&@1))); }
	| pattern t_comma pattern_l_list
	{ $$ = build2(&@1,&@3,A_PAT_APPLY,
	build1(&@2,&@2,A_PAT_VAR,consSym),
	build2(&@1,&@3,A_PAT_TUPLE,$1,$3)); }
	;

labeled_pattern : label s_E pattern
	{ $$ = build2(&@1,&@3,A_PAT_RECORD_MAPLET,$1,$3); }
	| identifier optional_colon_type optional_as_pattern
	{ ExprPtr pat,label;
	ENTER;
	pat = $1;
	TYPE(pat) = A_REAL_VAR /*A_PAT_VAR*/;
	label = build1(&@1,&@1,A_LABEL,CAR(ARGS(pat)));
	if ($3!=NIL)
	pat = build2(&@1,&@3,A_PAT_AS,pat,$3);
	if ($2!=NIL)
	pat = build2(&@1,&@2,A_PAT_CAST,pat,$2);
	$$ = build2(&@1,&@3,A_PAT_RECORD_MAPLET,
	label,pat);
	KEEPONLY($$); }
	;

optional_as_pattern : as_pattern
	| { $$ = NIL; }
	;

as_pattern : kw_as pattern
	{ $$ = $2; }
	;

application_pattern : atomic_pattern
	| long_constructor atomic_pattern
	{ TYPE($1) = A_PAT_VAR;
	    $$ = build2(&@1,&@2,A_PAT_APPLY,$1,$2); }
	;

pack_pattern : application_pattern
        | kw_pack { pushMarker("pack", NULL); }
		pack_pattern
	{ $$ = popMarker(build1(&@1,&@3,A_PAT_PACK,$3)); }
	;

left_associative_precedence_9_pattern : pack_pattern
	| left_associative_precedence_9_pattern long_left_associative_precedence_9_constructor pack_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_9_constructor : t_left_associative_precedence_9_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_9_pattern : left_associative_precedence_9_pattern
	| left_associative_precedence_9_pattern long_right_associative_precedence_9_constructor right_associative_precedence_9_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_9_constructor : t_right_associative_precedence_9_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_8_pattern : right_associative_precedence_9_pattern
	| left_associative_precedence_8_pattern long_left_associative_precedence_8_constructor right_associative_precedence_9_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_8_constructor : t_left_associative_precedence_8_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_8_pattern : left_associative_precedence_8_pattern
	| left_associative_precedence_8_pattern long_right_associative_precedence_8_constructor right_associative_precedence_8_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_8_constructor : t_right_associative_precedence_8_constructor
							{ $$ = lexVar(&@1,true); }
							;

left_associative_precedence_7_pattern : right_associative_precedence_8_pattern
	| left_associative_precedence_7_pattern long_left_associative_precedence_7_constructor right_associative_precedence_8_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_7_constructor : t_left_associative_precedence_7_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_7_pattern : left_associative_precedence_7_pattern
	| left_associative_precedence_7_pattern long_right_associative_precedence_7_constructor right_associative_precedence_7_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_7_constructor : t_right_associative_precedence_7_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_6_pattern : right_associative_precedence_7_pattern
	| left_associative_precedence_6_pattern long_left_associative_precedence_6_constructor right_associative_precedence_7_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_6_constructor : t_left_associative_precedence_6_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_6_pattern : left_associative_precedence_6_pattern
	| left_associative_precedence_6_pattern long_right_associative_precedence_6_constructor right_associative_precedence_6_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_6_constructor : t_right_associative_precedence_6_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_5_pattern : right_associative_precedence_6_pattern
	| left_associative_precedence_5_pattern long_left_associative_precedence_5_constructor right_associative_precedence_6_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_5_constructor : t_left_associative_precedence_5_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_5_pattern : left_associative_precedence_5_pattern
	| left_associative_precedence_5_pattern long_right_associative_precedence_5_constructor right_associative_precedence_5_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_5_constructor : t_right_associative_precedence_5_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_4_pattern : right_associative_precedence_5_pattern
	| left_associative_precedence_4_pattern long_left_associative_precedence_4_constructor right_associative_precedence_5_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_4_constructor : t_left_associative_precedence_4_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_4_pattern : left_associative_precedence_4_pattern
	| left_associative_precedence_4_pattern long_right_associative_precedence_4_constructor right_associative_precedence_4_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_4_constructor : t_right_associative_precedence_4_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_3_pattern : right_associative_precedence_4_pattern
	| left_associative_precedence_3_pattern long_left_associative_precedence_3_constructor right_associative_precedence_4_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_3_constructor : t_left_associative_precedence_3_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_3_pattern : left_associative_precedence_3_pattern
	| left_associative_precedence_3_pattern long_right_associative_precedence_3_constructor right_associative_precedence_3_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_3_constructor : t_right_associative_precedence_3_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_2_pattern : right_associative_precedence_3_pattern
	| left_associative_precedence_2_pattern long_left_associative_precedence_2_constructor right_associative_precedence_3_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_2_constructor : t_left_associative_precedence_2_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_2_pattern : left_associative_precedence_2_pattern
	| left_associative_precedence_2_pattern long_right_associative_precedence_2_constructor right_associative_precedence_2_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_2_constructor : t_right_associative_precedence_2_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_1_pattern : right_associative_precedence_2_pattern
	| left_associative_precedence_1_pattern long_left_associative_precedence_1_constructor right_associative_precedence_2_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_1_constructor : t_left_associative_precedence_1_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_1_pattern : left_associative_precedence_1_pattern
	| left_associative_precedence_1_pattern long_right_associative_precedence_1_constructor right_associative_precedence_1_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_1_constructor : t_right_associative_precedence_1_constructor
							{ $$ = lexVar(&@1,true); }
						;

left_associative_precedence_0_pattern : right_associative_precedence_1_pattern
	| left_associative_precedence_0_pattern long_left_associative_precedence_0_constructor right_associative_precedence_1_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_left_associative_precedence_0_constructor : t_left_associative_precedence_0_constructor
							{ $$ = lexVar(&@1,true); }
						;

right_associative_precedence_0_pattern : left_associative_precedence_0_pattern
	| left_associative_precedence_0_pattern long_right_associative_precedence_0_constructor right_associative_precedence_0_pattern
	{ $$ = patApplyBin(&@1,&@3,$2,$1,$3); }
	;

long_right_associative_precedence_0_constructor : t_right_associative_precedence_0_constructor
							{ $$ = lexVar(&@1,true); }
						;

infix_pattern : right_associative_precedence_0_pattern
	;

colon_pattern : infix_pattern
	| infix_pattern colon_type
	{ $$ = build2(&@1,&@2,A_PAT_CAST,$1,$2); }
	;

pattern : colon_pattern
	| colon_pattern as_pattern
	{ if (TYPE($1)!=A_REAL_VAR && (TYPE($1)!=A_PAT_CAST ||
				       TYPE(CAR(ARGS($1)))!=A_REAL_VAR))
	    type_error(ZONE($1),(CHAR *)message(612));
	  $$ = build2(&@1,&@2,A_PAT_AS,$1,$2); }
	;

optional_colon_type : colon_type
	| { $$ = NIL; }
	;

colon_type : s_T type
	{ $$ = $2; }
	;

atomic_type : type_variable
	| opening_record optional_type_row s_CI
	{ $$ = popMarker(BUILD(&@1,&@3,A_RECORD_TYPE,$2)); }
	| opening_parenthesis type s_Q
	{ $$ = popMarker($2); }
	;

type_variable : t_type_variable
	{ $$ = build1(&@1,&@1,A_TYPE_VAR,lexId); }
	;

constructor_type : atomic_type
	| constructor_type symbol
	{ $$ = build2(&@1,&@2,A_TYPE_APPLY,$2,$1); }
	| symbol
	{ $$ = build1(&@1,&@1,A_TYPE_APPLY,$1); }
	| opening_parenthesis type t_comma type_list s_Q
		symbol
	{ $$ = popMarker(build2rev(&@1,&@6,A_TYPE_APPLY,$6,$2,$4)); }
	;

simple_dimension_type : constructor_type
	| constructor_type s_X number_or_int
	{ $$ = build2(&@1,&@3,A_TYPE_POWER,$1,$3); }
	;

dimension_type : simple_dimension_type
	| dimension_type t_backtick simple_dimension_type
	{ $$ = build2(&@1,&@3,A_TYPE_STAR,$1,$3); }
	;

tuple_type : tuple_type_l
	{ if (CDR($1)==NIL)
	{
	$$ = CAR($1);
	if (TYPE($$)==A_TYPE_ELLIPSIS)
	type_error(ZONE($$),(CHAR *)message(392));
	}
	else
	$$ = BUILD(&@1,&@1,A_TUPLE_TYPE,$1); }
	;

	/* tuple_type_ls are lists of types */
tuple_type_l : dimension_type
	{ $$ = list1($1); }
	| ellipsis_type
	{ $$ = list1($1); }
	| dimension_type s_F tuple_type_l
	{ $$ = cons($1,$3); }
	;

ellipsis_type : s_DDD s_T type_variable
	{ $$ = build1(&@1,&@3,A_TYPE_ELLIPSIS,$3); }
	;

function_type : tuple_type
	| tuple_type s_MG function_type
	{ $$ = build2(&@1,&@3,A_FUNCTION_TYPE,$1,$3); }
	| tuple_type s_MmG function_type
	{ $$ = build2(&@1,&@3,A_MAP_TYPE,$1,$3); }
	;

type : { star_recognized++;
	 pushMarker(NULL,NULL);
	}
	function_type
	{ star_recognized--;
	$$ = popMarker($2);
	}
	;

	/* type lists are reversed lists of types */
type_list : type
	{ $$ = list1($1); }
	| type_list t_comma type
	{ $$ = cons($3,$1); }
	;

optional_type_row : type_row optional_ellipsis_type
		{ $$ = Nreverse1($1,$2); }
	| optional_ellipsis_type
	;

optional_ellipsis_type : { $$ = NIL; }
	| t_comma ellipsis_type
		{ $$ = list1($2); }
	;

	/* type_rows are reversed lists of labeled types */
type_row : labeled_type
	{ $$ = list1($1); }
	| type_row t_comma labeled_type
	{ checkUniqueLabel($3,$1);
	$$ = cons($3,$1); }
	;

labeled_type : label colon_type
	{ $$ = build2(&@1,&@2,A_MAPLET_TYPE,$1,$2); }
	;

infix_identifier : infix_identifier_1
						{ $$ = lexVar(&@1,false); }
					;

infix_identifier_1 : t_left_associative_precedence_0_identifier
	| t_right_associative_precedence_0_identifier
	| t_left_associative_precedence_1_identifier
	| t_right_associative_precedence_1_identifier
	| t_left_associative_precedence_2_identifier
	| t_right_associative_precedence_2_identifier
	| t_left_associative_precedence_3_identifier
	| t_right_associative_precedence_3_identifier
	| t_left_associative_precedence_4_identifier
	| t_right_associative_precedence_4_identifier
	| t_left_associative_precedence_5_identifier
	| t_right_associative_precedence_5_identifier
	| t_left_associative_precedence_6_identifier
	| t_right_associative_precedence_6_identifier
	| t_left_associative_precedence_7_identifier
	| t_right_associative_precedence_7_identifier
	| t_left_associative_precedence_8_identifier
	| t_right_associative_precedence_8_identifier
	| t_left_associative_precedence_9_identifier
	| t_right_associative_precedence_9_identifier
	;

optional_precedence : precedence_constant { $$ = yytext[0]-'0'; }
	| { $$ = 0; }
	;

	/* symbol_lists are in reverse */
symbol_list : label_id
		{ checkLabel(&@1); $$ = list1(lexId); }
	| s_E
		{ $$ = list1(equalSym); }
	| symbol_list label_id
		{ checkLabel(&@2); $$ = cons(lexId,$1); }
	| symbol_list s_E
		{ $$ = cons(equalSym,$1); }
	;

value_description : value_description_l
					{ $$ = buildrev(&@1,&@1,A_VAL_SPEC,$1); }
					;

	/* value_description_ls are reversed lists */
value_description_l : simple_value_description
						{ $$ = list1($1); }
					| value_description_l kw_and simple_value_description
						{ checkUniqueValDesc($3,$1);
							$$ = cons($3,$1); }
					;

simple_value_description : identifier s_T type
								{ $$ = build2(&@1,&@3,A_VAL_BIND_SPEC,$1,$3); }
							;

type_description : type_description_l
						{ $$ = buildrev(&@1,&@1,A_TYPE_SPEC,$1); }
					;

	/* type_description_ls are reversed lists */
type_description_l : simple_type_description
						{ $$ = list1($1); }
					| type_description_l kw_and simple_type_description
						{ checkUniqueIdInType($3,$1);
							$$ = cons($3,$1); }
					;

simple_type_description : type_variable_sequence symbol
						{ $$ = build2(&@1,&@2,A_TYPE_BIND_SPEC,$1,$2); }
						;

/*
exception_description : exception_description_l
							{ $$ = buildrev(&@1,&@1,A_EXCEPTION_DESC,$1); }
						;
*/

	/* exception_desciption_ls are reversed lists */
/*
exception_description_l : simple_exception_description
							{ $$ = list1($1); }
						| exception_description_l kw_and simple_exception_description
							{ checkUniqueException($3,$1);
								$$ = cons($3,$1); }
						;

simple_exception_description : symbol optional_of_type
									{ setConstructor(CAR(ARGS($1)));
										$$ = ($2==NIL)?
												build1(&@1,&@1,A_NEWEXC_DESC,$1):
												build2(&@1,&@2,A_NEWEXC_DESC,$1,$2); }
								;
*/

%%

  PRIVATE ExprPtr do_cpp_line_directive(loc)
    YYLTYPE *loc;
  {
    char c, *s = yytext+1;
    int lineno;

    while (*s==' ') s++;
    lineno = 0;
    while (c = *s++, c>='0' && c<='9')
      lineno = 10*lineno + (int)(c-'0');
    while (*s==' ') s++;
    lineno--;
    yylloc.first_line = yylloc.last_line = lineno;
    yylloc.first_column = yylloc.last_column = 0;
    return list1(build1(&yylloc, &yylloc, A_SET_LINE,
			getString(loc, loc, (CHAR *)s, true)));
  }
