/* Parsing support functions.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "expr.h"
#include "gram.h"
#include "gimml.tab.h"
#include "type.h"
#include "input.h"
#include "compile.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "grmproto.h"
#include "errproto.h"
#include "typproto.h"
#include "memproto.h"
#include "prtproto.h"
#include "cmpproto.h"
#include "stkproto.h"
#include "evlproto.h"
#include "dirproto.h"
#include "exgproto.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#include <sys/stat.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rgramaux.h"
#ifdef HAS_PROTOTYPES
#include "sgramaux.h"
#endif

#define DEBUG

#define HML_EXTENSION

#if TRANSLATOR
IMPORT ExprPtr Compile P((IN CONST ExprPtr e,IN COUNTER level,IN Boolean bindp));
#endif
EXTERN hFILE *yyin,*yyout;
EXTERN ExprPtr keepLexEnv,keepConsEnv;

PUBLIC ExprPtr parseMarker;
PUBLIC ExprPtr parseMarkerStack;
PUBLIC stackDesc toplevelJmpBuf;
PUBLIC Boolean toplevelJmpBufInstalled=0;
PUBLIC ExprPtr theSyntax;
PUBLIC SymPtr ItSymbol;

PUBLIC ExprPtr theTypes,theBindings;

PUBLIC CHAR *compile_file = NULL;

PUBLIC Boolean wantCloseFile=false,syntax_analysis=false;

EXTERN int yychar;
EXTERN YYSTYPE yylval;
EXTERN YYLTYPE yylloc;
EXTERN int yynerrs;

EXTERN INT currentAge;
EXTERN ExprPtr features,featuresType;
EXTERN ExprPtr specialMarker;

PUBLIC struct cslist *Checksums=NULL;

PRIVATE ExprPtr openModRef = NULL;

PRIVATE DECL_POW10(my_pow10)
#define POW10 (my_pow10+10)

#if DEBUGGING
EXTERN Boolean evaluatingp,stopwanted, debugEnabled;
#endif

PRIVATE void TextToReal(start,end,text,xp) /* could use atof() with a function that transforms
	tildes into minus signs, but atof() is buggy in Think C 4.0 */
IN CONST struct yyltype *start,*end;
IN CONST CHAR *text;
OUT REAL *xp;
{
  EXTENDED x;
  CHAR *s,c;
  COUNTER dot,nacc;
  unsigned long acc;
  Boolean neg;

  s = text;
  c = *s;
  if (c=='~')
    {
      neg = true;
      s++;
    }
  else
    neg = false;
  for (dot = -1, x = zero, nacc = 0, acc = 0L; (c = *s++);)
    switch (c)
      {
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	  if (nacc>=NDIGS_IN_ULONG)
	    { x=x*POW10[NDIGS_IN_ULONG]+(EXTENDED)acc; acc=0L; nacc=0; }
	  acc = acc*10L+(unsigned long)(c-'0');
	  nacc++;
	  if (dot>=0)
	    dot++;
	  break;
	case '.':
	  dot = 0;
	  break;
	case 'e': case 'E':
	  goto doexp;
	default:
	  {
	    TZone zone;
	    CHAR *s;

	    makeZone(&zone,start,end);
	    s = (CHAR *)alloctmp(128);
	    sprintf((char *)s,message(477),(unsigned int)c);
	    type_error(&zone,s);
	  }
      }
doexp:
  x = x*POW10[nacc]+(EXTENDED)acc;
  if (dot>=0)
    dot = -dot;
  else
    dot = 0;
  if (c!=0)
    {
      c = *s;
      if (c=='~')
	{
	  s++;
	  dot -= atoi((char *)s);
	}
      else
	dot += atoi((char *)s);
    }
  while (dot<0)
    {
      if (dot<-10)
	{
	  x = x*POW10[-10];
	  dot += 10;
	}
      else
	{
	  x = x*POW10[dot];
	  dot = 0;
	}
    }
  while (dot>0)
    {
      if (dot>10)
	{
	  x = x*POW10[10];
	  dot -= 10;
	}
      else
	{
	  x = x*POW10[dot];
	  dot = 0;
	}
    }
  if (neg)
    x = -x;
  *xp = x;
}

PRIVATE ExprPtr cons(a,b)
IN CONST ExprPtr a,b;
{
  REGISTER ExprPtr res;

  MAKECONS(a,b,res);
  QUICKKEEP(res);
  return res;
}

PRIVATE ExprPtr Cons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr res;

  MAKECONS(a,b,res);
  return KEEP(res);
}

PUBLIC ExprPtr usePaths; /* string list ref */

PUBLIC void add_path(name,len)
char *name;
int len;
{
  ExprPtr str,paths;
  ENTER;

  str = Kstring1(len,(CHAR *)name,true);
  paths = REFVAL(usePaths);
  paths = KONS(str,paths);
  SETREFVAL(usePaths,paths);
  QUIT;
}

PUBLIC hFILE *xfopen(name,mode,realname)
char *name,*mode;
char **realname;
{
  char *s;
  hFILE *file;
  COUNTER len;
  ExprPtr paths,path;

  len = strlen(name)+1;
  *realname = (char *)MALLOC_FAIL(len);
  if (*realname==NULL)
    return NULL;
  strcpy(*realname,name);
  file = hfopen(name,mode);
  if (file==NULL)
    {
      MARKFREE(*realname);
      for (paths = REFVAL(usePaths); paths!=NIL; paths = CDR(paths))
	{
	  path = CAR(paths);
	  *realname = (char *)MALLOC_FAIL(SLEN(path)+len+1);
	  if (*realname==NULL)
	    return NULL;
	  safememcpy(*realname,STEXT(path),SLEN(path));
	  s = *realname+SLEN(path);
	  if (SLEN(path)!=0 && s[-1]!=DIR_CHAR && s[-1]!=VOL_CHAR)
	    *s++ = DIR_CHAR;
	  strcpy(s,name);
	  file = hfopen(*realname,mode);
	  if (file!=NULL)
	    break;
	  MARKFREE(*realname);
	  *realname = NULL;
	}
    }
  return file;
}

PUBLIC void InitMarkerStack()
{
  SETQ(parseMarkerStack,NIL);
}

PUBLIC void pushMarker(begin,end)
  IN CONST char *begin, *end;
{
  ExprPtr protect,aux;
  struct lexOpenToks *toks;

  protect = getProtectWrapper();
  PUSHQ(parseMarkerStack,protect,aux);
  toks = (struct lexOpenToks *)MALLOC(sizeof(struct lexOpenToks));
  toks->next = lex_open_toks;
  toks->begin = begin;
  toks->end = end;
  lex_open_toks = toks;
  /*
    hfprintf(hstderr,"push:[");hfflush(hstderr);for(toks=lex_open_toks;toks!=NULL;toks=toks->next) if (toks->begin!=NULL) hfprintf(hstderr,"%s",toks->begin); hfprintf(hstderr,"]\n");hfflush(hstderr);
  */
}

PUBLIC ExprPtr popMarker(keep)
IN CONST ExprPtr keep;
{
  ExprPtr mark;
  struct lexOpenToks *toks;

  if (lex_open_toks!=NULL)
    {
      toks = lex_open_toks->next;
      MARKFREE(lex_open_toks);
      lex_open_toks = toks;
    }
  /*
    hfprintf(hstderr,"pop:[");hfflush(hstderr);for(toks=lex_open_toks;toks!=NULL;toks=toks->next) if (toks->begin!=NULL) hfprintf(hstderr,"%s",toks->begin); hfprintf(hstderr,"]\n");hfflush(hstderr);
  */
  mark = CAR(parseMarkerStack);
  setProtectWrapper(mark,keep);
  POPQ(parseMarkerStack);
  return keep;
}

/*
PUBLIC void tildeToMinus(text)
INOUT CHAR *text;
{
  CHAR *s,c;

  for (s=text; (c = *s)!=0; s++)
    if (c=='~')
      *s = '-';
}
*/

PUBLIC ExprPtr getReal(start,end,text)
IN CONST struct yyltype *start,*end;
IN CHAR *text;
{
  ExprPtr aux;
  REAL x;

  /*tildeToMinus(text);*/
  TextToReal(start,end,text,&x);
  aux = Kreal(&x);
  return KEEP(aux);
}

PUBLIC ExprPtr getComplex(start,end,text)
IN CONST struct yyltype *start,*end;
IN CHAR *text;
{
  ExprPtr re,im;
  CHAR *s,c;
  REAL x,y;

  /*tildeToMinus(text);*/
  for (s=text; c = *s, c!=':'; s++);
  *s = 0;
  TextToReal(start,end,text,&x);
  *s = ':';
  while (c = *++s, c==' ' || c=='\t' || c=='\n' || c=='\v' || c=='\f');
  TextToReal(start,end,s,&y);
  if (ZEROR(y))
    {
      re = Kreal(&x);
      return KEEP(re);
    }
  re = Kreal(&x);
  (void) KEEP(re);
  im = Kreal(&y);
  (void) KEEP(im);
  re = Kcomplex(re,im);
  return KEEP(re);
}

PUBLIC ExprPtr getString(start,end,text,build)
IN CONST struct yyltype *start,*end;
IN REGISTER CHAR *text;
IN Boolean build;
{
  ExprPtr str;
  CHAR *begin;
  REGISTER INT len;
  REGISTER CHAR *dest,c;

  begin = text;
  len = 0;
  for (dest = text, text++; c = *text++, c!='\"'; )
    {
      switch (c)
	{
	case '\\':
	  c = *text++;
	  switch (c)
	    {
	    case '\"':   case '\\':
	      *dest++ = c;
	      len++;
	      break;
	    case 'n':
	      *dest++ = '\n';
	      len++;
	      break;
	    case 't':
	      *dest++ = '\t';
	      len++;
	      break;
	    case 'v':
	      *dest++ = '\v';
	      len++;
	      break;
	    case 'f':
	      *dest++ = '\f';
	      len++;
	      break;
	    case '^':
	      *dest++ = (*text++)-'@';
	      len++;
	      break;
	    case '0':	case '1':	case '2':	case '3':
	    case '4':	case '5':	case '6':	case '7':
	    case '8':	case '9':
	      *dest = (c-'0')*100;
	      *dest += ((*text++)-'0')*10;
	      *dest++ += ((*text++)-'0');
	      len++;
	      break;
	    default:
	      while (c = *text++, c!='\\');
	      break;
	    }
	  break;
	case '\n':
	  if (start!=(struct yyltype *)0L && end!=(struct yyltype *)0L)
	    {
	      TZone zone;

	      *dest = 0;
	      makeZone(&zone,start,end);
	      type_error(&zone,(CHAR *)message(646));
	    }
	  break;
	default:
	  *dest++ = c;
	  len++;
	  break;
	}
    }
  *dest = 0;
  return build?(str = Kstring(len,begin), KEEP(str)):NIL;
}

PUBLIC ExprPtr setFix(start,end,ll,p)
IN CONST struct yyltype *start;
IN CONST struct yyltype *end;
IN CONST ExprPtr ll;
IN COUNTER p;
{
  ExprPtr prec;
  ENTER;

  prec = Kint(p);
  (void) KEEP(prec);
  for (; ll!=NIL; ll=CDR(ll))
    {
      setcarMapPut(CAR(ll),prec,lexEnv);
    }
  QUIT;
  return NIL;
}

PUBLIC void pushEnv()
{
  ExprPtr aux;
  ENTER;

  MAKECONS1(NewMap(),lexEnv,aux);
  lexEnv = aux;
  MAKECONS1(NewMap(),consEnv,aux);
  consEnv = aux;
#if MODULE_SYSTEM & STD_ML_MODULE
  MAKECONS1(NewMap(),strEnv,aux);
  strEnv = aux;
  MAKECONS1(NewMap(),sigEnv,aux);
  sigEnv = aux;
#endif
  QUIT;
}

PUBLIC void popEnv()
{
  POPQ(lexEnv);
  POPQ(consEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  POPQ(strEnv);
  POPQ(sigEnv);
#endif
}

PUBLIC void makeZone(zone,yylsp1,yylsp2)
OUT TZone *zone;
IN CONST struct yyltype *yylsp1,*yylsp2;
{
  zone->line1 = (unsigned short)yylsp1->first_line;
  zone->pos1 = (unsigned short)yylsp1->first_column;
  zone->line2 = (unsigned short)yylsp2->last_line;
  zone->pos2 = (unsigned short)yylsp2->last_column;
}

#define BH_BUCKET_SIZE 32

typedef struct bhBucket {
  struct bhBucket *next;
  SHORT firstNewCell;
  SHORT occupancy[MAX_GENERATIONS];
#if BH_BUCKET_GAP!=0
  char filler[BH_BUCKET_GAP];
#endif
  struct build_header bh[BH_BUCKET_SIZE];
  struct build_header *bhcells[BH_BUCKET_SIZE];
} bhBucket;

PRIVATE bhBucket *allocBuckets=(bhBucket *)0L;
PRIVATE struct build_header *freeBuildHeaders=(struct build_header *)0L;
EXTERN SHORT generation_count;

PRIVATE struct build_header *reserve_bh=(struct build_header *)0L;

PRIVATE void newBucket()
{
  bhBucket *bhb;
  REGISTER struct build_header *bh,*oldbh,**bhp;
  REGISTER COUNTER i;

  bhb = (bhBucket *)MALLOC(sizeof(bhBucket));
  bhb->next = allocBuckets;
  bhb->firstNewCell = 0;
  for (i=0; i<MAX_GENERATIONS; i++)
    bhb->occupancy[i] = 0;
  allocBuckets = bhb;
  oldbh = (struct build_header *)0L; /* we add buckets at the end of the list */
  for (i = 0, bh = bhb->bh, bhp = bhb->bhcells; i<BH_BUCKET_SIZE; i++, bh++)
    {
      *bhp++ = bh;
      bh->kind = E_FREE;
      bh->mark = 0;
      /*bh->gcmark = 0;*/
      bh->args = (ExprPtr)oldbh;
      oldbh = bh;
    }
  if (reserve_bh==(struct build_header *)0L)
    {
      freeBuildHeaders = oldbh;
    }
  else
    { /* append at the end */
      reserve_bh->args = (ExprPtr)oldbh;
    }
  reserve_bh = bhb->bh; /* the last allocated cell */
}

PUBLIC void FreeBuildHeaders()
{
  REGISTER struct build_header *bh;
  REGISTER COUNTER i,n,occupancy;
  struct build_header **bhp,**firstbhp;
  bhBucket *buck;

  for (buck=allocBuckets; occupancy = 0, buck!=(bhBucket *)0L; buck = buck->next)
    {
      n = BH_BUCKET_SIZE-buck->firstNewCell;
      for (i=0, firstbhp = bhp = buck->bhcells+buck->firstNewCell; i<n; i++, bhp++)
	{
	  bh = *bhp;
	  if (GC_MARKED(bh))
	    {
	      GC_UNMARK(bh);
	      occupancy++;
	      if (firstbhp!=bhp) /* firstbhp is always equal to bhp, or points to the first
				    free cell after the firstNewCell position */
		{ /* exchange position of bh with that of the free cell *firstbhp */
		  *bhp = *firstbhp;
		  *firstbhp = bh;
		}
	      firstbhp++;
	      /* this all sorts the newcells array, from the firstNewCell position up,
		 so that all addresses of still allocated cells are in an interval
		 [firstNewCell..firstNewCell+buck->occupancy[generation_count] [ */
	    }
	  else if (bh->kind!=E_FREE)
	    {
	      bh->kind = E_FREE;
	      bh->args = (ExprPtr)freeBuildHeaders;
	      freeBuildHeaders = bh;
	    }
	}
      buck->occupancy[generation_count] = occupancy;
    }
}

PUBLIC void FreeBuildHeader(bh)
IN struct build_header *bh;
{
  bh->kind = E_FREE; /* space won't be got back until a FreeBuildHeaders() */
  bh->args = (ExprPtr)freeBuildHeaders;
  freeBuildHeaders = bh;
}

PUBLIC void cutoffInitialHeaders()
{
  freeBuildHeaders = reserve_bh = (struct build_header *)0L;
  allocBuckets = (bhBucket *)0L;
}

#ifdef GC_SLIDING
PUBLIC void slideOldHeaders()
{
  REGISTER bhBucket *bhbp;
  REGISTER COUNTER i;
  REGISTER SHORT *occ;

  for (bhbp = allocBuckets; bhbp!=(bhBucket *)0L; bhbp=bhbp->next)
    {
      bhbp->occupancy[0] += bhbp->occupancy[1];
      for (i=1, occ=&bhbp->occupancy[1]; i<generation_count; i++, occ++)
	occ[0] = occ[1];
      occ[0] = 0;
    }
}
#endif

PUBLIC void cutoffOldHeaders()
{
  bhBucket *bhbp;
  REGISTER struct build_header **bhp;
  REGISTER COUNTER i;
  REGISTER COUNTER n;

  for (bhbp = allocBuckets; bhbp!=(bhBucket *)0L; bhbp=bhbp->next)
    {
      n = bhbp->occupancy[generation_count];
      if (n==0)
	continue;
      /* mark in advance all "newly old" cells */
      for (i=0, bhp=bhbp->bhcells + bhbp->firstNewCell; i<n; i++, bhp++)
	{
	  GC_MARK((*bhp)); /* to speed up future mark phases */
	}
      bhbp->firstNewCell += n;
    }
}

PUBLIC void refreshOldHeaders(oldgen)
IN COUNTER oldgen;
{
  REGISTER COUNTER n;
  REGISTER bhBucket *bhbp;

  for (bhbp=allocBuckets; bhbp!=(bhBucket *)0L; bhbp=bhbp->next)
    {
      n = bhbp->firstNewCell;
      {
	REGISTER COUNTER i;
	REGISTER SHORT *occ;

	for (i = generation_count, occ = bhbp->occupancy+i; i<oldgen; i++)
	  n -= *occ++;
      }
      {
	REGISTER struct build_header **bhp;
	REGISTER COUNTER i;

	for (i = bhbp->firstNewCell, bhp = bhbp->bhcells+i; --i>=n; )
	  GC_UNMARK((*--bhp));
      }
      bhbp->firstNewCell = n;
    }
}

PUBLIC ExprPtr ZBuild(zone,f,info)
IN CONST TZone *zone;
IN COUNTER f;
IN CONST ExprPtr info;
{
  struct build_header *bh;

  if (freeBuildHeaders==reserve_bh)
    newBucket();
  bh = freeBuildHeaders;
  freeBuildHeaders = (struct build_header *)bh->args;
  bh->kind = BUILD_HEADER_KIND;
  bh->mark = 0;
  /*bh->gcmark = 0;*/
  bh->type = f;
  bh->z = *zone;
  bh->args = info;
  return (ExprPtr)bh;
}

PUBLIC ExprPtr Build1(start,end,f,info)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr info;
{
  TZone zone;

  makeZone(&zone,start,end);
  return ZBuild(&zone,f,info);
}

PUBLIC ExprPtr Build(start,end,f,info)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr info;
{
  ExprPtr aux;
  TZone zone;

  makeZone(&zone,start,end);
  aux = ZBuild(&zone,f,info);
  return KEEP(aux);
}

PUBLIC ExprPtr build1(start,end,f,a)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,NIL));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr build2(start,end,f,a,b)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,NIL)));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr build3(start,end,f,a,b,c)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b,c;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,cons(c,NIL))));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr buildrev(start,end,f,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr list;
{
  ENTER;
  list = Build1(start,end,f,Nreverse(list));
  KEEPONLY(list);
  return list;
}

PUBLIC ExprPtr build1rev(start,end,f,a,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,list;
{
  ExprPtr aux;
  ENTER;

  list = Nreverse(list);
  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,list));
  ENDQUICKKEEP();
  KEEPONLY(aux);
  return aux;
}

PUBLIC ExprPtr build2rev(start,end,f,a,b,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b,list;
{
  ExprPtr aux;
  ENTER;

  list = Nreverse(list);
  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,list)));
  ENDQUICKKEEP();
  KEEPONLY(aux);
  return aux;
}

PUBLIC ExprPtr build1list(start,end,f,a,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,list;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,list));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr build2list(start,end,f,a,b,list)
IN CONST struct yyltype *start,*end;
IN COUNTER f;
IN CONST ExprPtr a,b,list;
{
  ExprPtr aux;

  BEGINQUICKKEEP();
  aux = Build1(start,end,f,cons(a,cons(b,list)));
  ENDQUICKKEEP();
  return KEEP(aux);
}

PUBLIC ExprPtr applyBin(start,end,f,arg1,arg2)
IN CONST struct yyltype *start,*end;
IN CONST ExprPtr f,arg1,arg2;
{
  ExprPtr res;
  ENTER;

  BEGINQUICKKEEP();
  res = cons(arg2,NIL);
  res = cons(arg1,res);
  res = Build1(start,end,A_TUPLE,res);
  QUICKKEEP(res);
  res = cons(res,NIL);
  res = cons(f,res);
  res = Build1(start,end,A_APPLY,res);
  ENDQUICKKEEP();
  KEEPONLY(res);
  return res;
}

PUBLIC ExprPtr patApplyBin(start,end,f,arg1,arg2)
IN CONST struct yyltype *start,*end;
IN CONST ExprPtr f,arg1,arg2;
{
  ExprPtr res;
  ENTER;

  TYPE(f) = A_PAT_VAR;
  res = build2(start,end,A_PAT_APPLY,f,
	       build2(start,end,A_PAT_TUPLE,arg1,arg2));
  KEEPONLY(res);
  return res;
}

PUBLIC ExprPtr lexVar(start,longp)
IN CONST struct yyltype *start;
IN CONST Boolean longp;
{
  ExprPtr aux;
  TZone zone;
  ENTER;

  if (!longp && lexPath!=NIL)
    {
      makeZone(&zone,start,start);
      type_error(&zone,(CHAR *)message(1));
    }
  aux = build1list(start,start,A_VARIABLE,lexId,lexPath);
  KEEPONLY(aux);
  return aux;
}

PRIVATE COUNTER gsCounter=0;
PRIVATE SymPtr gensym()
{
  SymPtr sym;
  CHAR buf[16];

  sprintf((char *)buf,"*x%d",gsCounter++);
  sym = Ksymbol(buf);
  return KEEP(sym);
}

PUBLIC ExprPtr genVar(start,end)
IN CONST struct yyltype *start,*end;
{
  ExprPtr aux;
  SymPtr sym;
  ENTER;

  sym = gensym();
  aux = build1(start,end,A_VARIABLE,sym);
  KEEPONLY(aux);
  return aux;
}

PRIVATE ExprPtr build1MatchFun(list)
IN CONST ExprPtr list;
{
  ExprPtr zfpe,pe,val,ll,pat;
  TZone *zone;
  TZone tz;
  ENTER;

  zfpe = CAR(list);
  zone = ZONE(CAR(zfpe));
  tz = *zone;
  pe = CDR(CDR(zfpe));
  ll = CAR(pe);
  val = CDR(pe);
  BEGINQUICKKEEP();
  for (; ll!=NIL; ll=CDR(ll))
    {
      pat = CAR(ll);
      tz.line1 = ZONE(pat)->line1;
      tz.pos1 = ZONE(pat)->pos1;
      val = ZBuild(&tz,A_LAMBDA,cons(pat,cons(val,NIL)));
      QUICKKEEP(val);
      val = ZBuild(&tz,A_FN,cons(val,NIL));
      QUICKKEEP(val);
    }
  ENDQUICKKEEP();
  KEEPONLY(val);
  return val;
}

PRIVATE ExprPtr build1ArgFun(start,end,list)
IN CONST struct yyltype *start,*end;
IN CONST ExprPtr list;
{
  ExprPtr val,ll,zfpe,pe,aux;
  TZone *zone;
  ENTER;

  val = NIL;
  for (ll=list; ll!=NIL; ll=CDR(ll))
    {
      zfpe = CAR(ll);
      zone = ZONE(CAR(zfpe));
      pe = CDR(CDR(zfpe));
      BEGINQUICKKEEP();
      aux = ZBuild(zone,A_LAMBDA,cons(CAR(CAR(pe)),cons(CDR(pe),NIL)));
      QUICKKEEP(aux);
      val = cons(aux,val);
      ENDQUICKKEEP();
      KEEPONLY(val);
    }
  val = Build1(start,end,A_FN,val);
  KEEPONLY(val);
  return val;
}

PUBLIC ExprPtr buildFun(start,end,n,list)
IN CONST struct yyltype *start,*end;
IN COUNTER n;
IN CONST ExprPtr list;
{
  ExprPtr llvar,ll,val,zfpe,pe,patch,var;
  ExprPtr aux,loc;
  TZone *zone;
  ENTER;
  
  for (llvar=NIL; n!=0; n--)
    {
      llvar = Cons(genVar(start,end),llvar);
    }
  val = NIL;
  for (ll=list; ll!=NIL; ll=CDR(ll))
    {
      zfpe = CAR(ll);
      zone = ZONE(CAR(zfpe));
      pe = CDR(CDR(zfpe));
      aux = ZBuild(zone,A_PAT_TUPLE,Nreverse(CAR(pe)));
      (void) KEEP(aux);
      aux = ZBuild(zone,A_LAMBDA,Cons(aux,Cons(CDR(pe),NIL)));
      val = Cons(KEEP(aux),val);
    }
  val = BUILD(start,end,A_FN,val);
  val = build2(start,end,A_APPLY,val,NIL);
  loc = CDR(ARGS(val));
  /* NIL will be replaced by A_TUPLE(Nreverse(llvar)) later */
  for (ll=llvar; ll!=NIL; ll=CDR(ll))
    {
      var = CAR(ll);
      aux = ZBuild(ZONE(var),A_PAT_VAR,ARGS(var));
      val = build1(start,end,A_FN,
		   build2(start,end,A_LAMBDA,KEEP(aux),val));
    }
  patch = buildrev(start,end,A_TUPLE,llvar);
  SETCAR(loc,patch);
  KEEPONLY(val);
  return val;
}

PUBLIC ExprPtr buildFunBind(start,end,list)
	/* list of zonex::f::revpatlist::expr */
IN CONST struct yyltype *start,*end;
IN CONST ExprPtr list;
{
  ExprPtr f,val;
  ExprPtr aux;
  COUNTER n;
  ENTER;
  
  f = CAR(CDR(CAR(list)));
  if (CDR(list)==NIL)
    val = build1MatchFun(list);
  else
    {
      n = ListLength(CAR(CDR(CDR(CAR(list)))));
      if (n==1)
	val = build1ArgFun(start,end,list);
      else
	val = buildFun(start,end,n,list);
    }
  aux = build2(start,end,A_FUNBIND,f,val);
  KEEPONLY(aux);
  return aux;
}

PRIVATE void patchMemoFun P((INOUT ExprPtr fun));

PRIVATE void patchMemoFun(fun)
INOUT ExprPtr fun;
{
  ExprPtr args,arg;

  if (TYPE(fun)==A_FN)
    {
      TYPE(fun) = A_MEMO_FN;
      for (args=ARGS(fun); args!=NIL; args=CDR(args))
	{
	  arg = CAR(args);
	  if (TYPE(arg)==A_LAMBDA)
	    patchMemoFun(CAR(CDR(ARGS(arg))));
	}
    }
}

PUBLIC void patchMemoFunList(args,tyvarseq,withlist)
INOUT ExprPtr args;
IN CONST ExprPtr tyvarseq;
IN CONST ExprPtr withlist;
{
  ExprPtr lastl;
  ExprPtr ll,binds,fun;

  ll = listTyVars(ARGS(tyvarseq),ARGS(CAR(args)));
  if (ll!=ARGS(CAR(args)))
    {
      ll = ZBuild (ZONE (args), A_TYPE_VAR_SEQ, ll);
      SETCAR(args,ll); /* get union of type variable sequences */
    }
  lastl = args;
  while ((ll=CDR(lastl))!=NIL)
    {
      binds = CDR(ARGS(CAR(ll)));
      fun = CAR(binds);
      patchMemoFun(fun);
      lastl = ll;
    }
  SETCDR(lastl,withlist);
}

PUBLIC void installMapPattern(yylsp,mapadds,rest)
IN CONST struct yyltype *yylsp;
IN CONST ExprPtr mapadds,rest;
{
  ExprPtr nextl;
  TZone zone;
  ENTER;

  while (nextl = CDR(ARGS(mapadds)),
	 CAR(nextl)!=NIL && TYPE(CAR(nextl))!=A_PAT_ANYMAP)
    mapadds = CAR(nextl);
  makeZone(&zone,yylsp,yylsp);
  if (CAR(nextl)==NIL)
    {
      if (rest==NIL)
	rest = ZBuild(&zone,A_PAT_EMPTY,NIL);
      SETCAR(nextl,rest);
    }
  else if (rest!=NIL)
    type_error(&zone,(CHAR *)message(2));
  QUIT;
}

PUBLIC ExprPtr zeroTuple(yylsp)
IN CONST struct yyltype *yylsp;
{
  struct yyltype new;

  new.first_line = yylsp->last_line;
  new.first_column = yylsp->last_column;
  return BUILD(&new,yylsp,A_TUPLE,NIL);
}

PUBLIC ExprPtr zeroPatTuple(yylsp)
IN CONST struct yyltype *yylsp;
{
  struct yyltype new;

  new.first_line = yylsp->last_line;
  new.first_column = yylsp->last_column;
  return BUILD(&new,yylsp,A_PAT_TUPLE,NIL);
}

PUBLIC ExprPtr patNil(yylsp)
IN CONST struct yyltype *yylsp;
{
  struct yyltype new;

  new.first_line = yylsp->last_line;
  new.first_column = yylsp->last_column;
  return BUILD(&new,yylsp,A_PAT_NIL,NIL);
}

PUBLIC ExprPtr makeNum(yylsp,num)
IN CONST struct yyltype *yylsp;
IN CONST ExprPtr num;
{
  ENTER;

  num = build1(yylsp,yylsp,A_NUMBER,num);
  KEEPONLY(num);
  return num;
}

PUBLIC ExprPtr makeInt(yylsp,num)
IN CONST struct yyltype *yylsp;
IN CONST ExprPtr num;
{
  ENTER;

  num = build1(yylsp,yylsp,A_INTEGER,num);
  KEEPONLY(num);
  return num;
}

PUBLIC void do_printfile(text,len,stream)
REGISTER IN CONST CHAR *text;
REGISTER IN COUNTER len;
REGISTER INOUT char *stream;
{
  REGISTER CHAR c;

  while (len--!=0)
    {
      c = *text++;
      hputc(c,(hFILE *)stream);
    }
}

PRIVATE char *errorMsg(msg,sym)
IN CONST char *msg;
IN CONST SymPtr sym;
{
#define BUFLEN 256
  static char buf[BUFLEN];
  char *s;
  size_t len;
  
  strcpy(buf,msg);
  s = buf+strlen(buf);
  *s++ = ':';
  *s++ = ' ';
  len = symNameLen(sym);
  if (s-buf+len<BUFLEN)
    {
      safememcpy(s,symName(sym),len);
      s[len] = 0;
    }
  else
    {
      safememcpy(s,symName(sym),(size_t)(BUFLEN-4-(s-buf)));
      s += BUFLEN-4;
      *s++ = '.';
      *s++ = '.';
      *s++ = '.';
      *s = 0;
    }
  return buf;
#undef BUFLEN
}

PUBLIC void checkUniqueCon(con,conbind)
IN CONST ExprPtr con,conbind;
{
  ExprPtr conname;
  TZone *zone;
  
  conname = CAR(ARGS(con));
  zone = ZONE(conname);
  conname = CAR(ARGS(conname));
  while (conbind!=NIL)
    {
      if (conname==CAR(ARGS(CAR(ARGS(CAR(conbind))))))
	type_error(zone,(CHAR *)errorMsg(message(4),conname));
      conbind = CDR(conbind);
    }
}

PUBLIC void checkUniqueConstructors(datbind,datbindlist)
IN CONST ExprPtr datbind,datbindlist;
{
  ExprPtr conbindlist,con,dtblist,dtb;
  
  for (conbindlist = CDR(CDR(ARGS(datbind)));
       conbindlist!=NIL;
       conbindlist=CDR(conbindlist))
    {
      con = CAR(conbindlist);
      for (dtblist=datbindlist; dtblist!=NIL; dtblist=CDR(dtblist))
	{
	  dtb = CAR(dtblist);
	  checkUniqueCon(con,CDR(CDR(ARGS(dtb))));
	}
    }
}

PUBLIC void checkUniqueIdInType(typbind,typbindlist)
IN CONST ExprPtr typbind,typbindlist;
{
  ExprPtr typname;
  TZone *zone;
  
  typname = CAR(CDR(ARGS(typbind)));
  zone = ZONE(typname);
  typname = CAR(ARGS(typname));
  while (typbindlist!=NIL)
    {
      if (typname==CAR(ARGS(CAR(CDR(ARGS(CAR(typbindlist)))))))
	type_error(zone,(CHAR *)errorMsg(message(5),typname));
      typbindlist = CDR(typbindlist);
    }
}

PUBLIC void checkUniqueTypeSeq(typevar,typeseq)
IN CONST ExprPtr typevar,typeseq;
{
  TZone *zone;
  
  zone = ZONE(typevar);
  typevar = CAR(ARGS(typevar));
  while (typeseq!=NIL)
    {
      if (typevar==CAR(ARGS(CAR(typeseq))))
	type_error(zone,(CHAR *)errorMsg(message(6),typevar));
      typeseq = CDR(typeseq);
    }
}

PUBLIC void checkUniqueUnit(unit,unitlist)
IN CONST ExprPtr unit,unitlist;
{
  ExprPtr unitname;
  TZone *zone;

  unitname = CAR(ARGS(unit));
  zone = ZONE(unitname);
  unitname = CAR(ARGS(unitname));
  while (unitlist!=NIL)
    {
      if (unitname==CAR(ARGS(CAR(ARGS(CAR(unitlist))))))
	type_error(zone,(CHAR *)errorMsg(message(7),unitname));
      unitlist = CDR(unitlist);
    }
}

PUBLIC void checkUniqueQuantityAndUnit(quan,quanlist)
IN CONST ExprPtr quan,quanlist;
{
  ExprPtr unit;
  TZone *zu,*zq;
  ExprPtr uu,qq;
  
  unit = CAR(CDR(ARGS(quan)));
  quan = CAR(ARGS(quan));
  zu = ZONE(unit);
  unit = CAR(ARGS(unit));
  zq = ZONE(quan);
  quan = CAR(ARGS(quan));
  while (quanlist!=NIL)
    {
      qq = CAR(quanlist);
      uu = CAR(CDR(ARGS(qq)));
      qq = CAR(ARGS(qq));
      if (unit==CAR(ARGS(uu)))
	type_error(zu,(CHAR *)errorMsg(message(7),unit));
      if (quan==CAR(ARGS(qq)))
	type_error(zq,(CHAR *)errorMsg(message(8),quan));
      quanlist = CDR(quanlist);
    }
}

PUBLIC void checkUniqueLabel(lexpr,lrow)
IN CONST ExprPtr lexpr,lrow;
{
  ExprPtr label,args;
  TZone *zone;
  
  label = CAR(ARGS(lexpr));
  zone = ZONE(label);
  label = CAR(ARGS(label));
  while (lrow!=NIL)
    {
      args = ARGS(CAR(lrow));
      if (args!=NIL)
	if (label==CAR(ARGS(CAR(args))))
	  type_error(zone,(CHAR *)errorMsg(message(9),label));
      lrow = CDR(lrow);
    }
}

PUBLIC void checkQuantifiedType(tyvarseq,type)
	/* checks that all type variables in type are in the list tyvarseq */
IN CONST ExprPtr tyvarseq,type;
{
  ExprPtr ll;

  switch (TYPE(type))
    {
    case A_TYPE_VAR:
      ll = CAR(ARGS(type));
      for (tyvarseq = ARGS(tyvarseq); tyvarseq!=NIL; tyvarseq=CDR(tyvarseq))
	if (ll==CAR(ARGS(CAR(tyvarseq))))
	  break;
      if (tyvarseq==NIL)
	type_error(ZONE(type),(CHAR *)errorMsg(message(10),ll));
      break;
    case A_MAPLET_TYPE:
      checkQuantifiedType(tyvarseq,CAR(CDR(ARGS(type))));
      break;
    case A_RECORD_TYPE:
    case A_TUPLE_TYPE:
    case A_TYPE_STAR:
    case A_FUNCTION_TYPE:
    case A_MAP_TYPE:
      for (ll = ARGS(type); ll!=NIL; ll=CDR(ll))
	checkQuantifiedType(tyvarseq,CAR(ll));
      break;
    case A_TYPE_APPLY:
      for (ll = CDR(ARGS(type)); ll!=NIL; ll=CDR(ll))
	checkQuantifiedType(tyvarseq,CAR(ll));
      break;
    case A_TYPE_POWER:
    case A_TYPE_ELLIPSIS:
      checkQuantifiedType(tyvarseq,CAR(ARGS(type)));
      break;
    default:
      berror(message(11),(int)TYPE(type));
      break;
    }
}

PUBLIC void checkQuantifiedDatatype(tyvarseq,conbindlist)
IN CONST ExprPtr tyvarseq,conbindlist;
{
  ExprPtr conbind,opttype;
  
  while (conbindlist!=NIL)
    {
      conbind = CAR(conbindlist);
      opttype = CDR(ARGS(conbind));
      if (opttype!=NIL)
	checkQuantifiedType(tyvarseq,CAR(opttype));
      conbindlist = CDR(conbindlist);
    }
}

PUBLIC void checkWithType(datbind,typbind)
	/* checks that no type name defined in typbind
	is defined in datbind */
IN CONST ExprPtr datbind,typbind;
{
  ExprPtr typename,ll;
  TZone *zone;
  
  typbind = ARGS(typbind);
  while (typbind!=NIL)
    {
      typename = CAR(CDR(ARGS(CAR(typbind))));
      zone = ZONE(typename);
      typename = CAR(ARGS(typename));
      for (ll=datbind; ll!=NIL; ll=CDR(ll))
	{
	  if (typename==CAR(CDR(ARGS(CAR(ll)))))
	    type_error(zone,(CHAR *)errorMsg(message(12),typename));
	}
      typbind = CDR(typbind);
    }
}

PUBLIC void checkUniqueException(exc,exclist)
IN CONST ExprPtr exc,exclist;
{
  ExprPtr excname;
  TZone *zone;
  
  excname = CAR(ARGS(exc));
  zone = ZONE(excname);
  excname = CAR(ARGS(excname));
  while (exclist!=NIL)
    {
      if (excname==CAR(ARGS(CAR(ARGS(CAR(exclist))))))
	type_error(zone,(CHAR *)errorMsg(message(13),excname));
      exclist = CDR(exclist);
    }
}

PUBLIC ExprPtr listTyVars(ll,tyvars)
IN CONST ExprPtr ll;
IN CONST ExprPtr tyvars;
{
  while (ll!=NIL)
    {
      tyvars = exprTyVars(CAR(ll),tyvars);
      ll = CDR(ll);
    }
  return tyvars;
}

PRIVATE ExprPtr addTyVar(tyvar,tyvars)
IN CONST ExprPtr tyvar;
IN CONST ExprPtr tyvars;
{
  ExprPtr sym,ll;

  sym = CAR(ARGS(tyvar));
  for (ll=tyvars; ll!=NIL; ll=CDR(ll))
    {
      if (sym==CAR(ARGS(CAR(ll))))
	return tyvars;
    }
  return Cons(tyvar,tyvars);
}

PUBLIC ExprPtr exprTyVars(expr,tyvars)
IN CONST ExprPtr expr;
IN CONST ExprPtr tyvars;
{
  switch (TYPE(expr))
    {
    case A_TYPE_VAR:
      return addTyVar(expr,tyvars);
    case A_VARIABLE:
    case A_NUMBER:
    case A_INTEGER:
    case A_STRING:
    case A_TYPE_ELLIPSIS:
    case A_UNIT_NAME:
    case A_REAL_VAR:
    case A_PAT_VAR:
    case A_PAT_CONST:
    case A_SELECT_FN:
    case A_LABEL:
    case A_BASIS:
    case A_ENTER_FILE:
    case A_EXIT_FILE:
    case A_SET_LINE:
    case A_PRAGMA:
      return tyvars;
    case A_VAL:
    case A_VALREC:
    case A_TYPE: /* type variables inside type declarations are */
    case A_DATATYPE: /* quantified, so must not appear in the set */
    case A_ABSTYPE: /* (error in the definition of standard ML, p.20) */
    case A_EXT_VAL:
    case A_EXT_DATATYPE:
    case A_EXT_EQTYPE:
    /*case A_EXT_EXC:*/
      return tyvars;
    default:
      return listTyVars(ARGS(expr),tyvars);
    }
}

PUBLIC void checkEqTypes(eqtylist,datbind)
	/* check that all type names in the list eqtylist are
		type names defined by datbind, and are unique */
IN CONST ExprPtr eqtylist;
IN CONST ExprPtr datbind;
{
  ExprPtr ll,name,type;
  
  for (; eqtylist!=NIL; eqtylist=CDR(eqtylist))
    {
      type = CAR(ARGS(CAR(eqtylist)));
      name = CAR(ARGS(type));
      for (ll=datbind; ll!=NIL; ll=CDR(ll))
	{
	  if (name==CAR(ARGS(CAR(CDR(ARGS(CAR(ll)))))))
	    break;
	}
      if (ll==NIL)
	type_error(ZONE(CAR(eqtylist)),(CHAR *)errorMsg(message(14),name));
      for (ll=CDR(eqtylist); ll!=NIL; ll=CDR(ll))
	{
	  if (name==CAR(ARGS(CAR(ARGS(CAR(ll))))))
	    type_error(ZONE(type),(CHAR *)message(12));
	}
    }
}

#if MODULE_SYSTEM & STD_ML_SYSTEM
PRIVATE void lexOpen1(quadr,overwritep)
IN CONST ExprPtr quadr;
IN Boolean overwritep;
{
  VecPtr vp;
  ExprPtr e,aux;
  
  vp = AVAL(quadr);
  e = vp->expr[0];
  if (overwritep)
    {
      setcarMapMerge(lexEnv,e);
    }
  else
    {
      MAKECONS(e,NIL,aux);
      SETQ(lexEnv,aux);
    }
  e = vp->expr[1];
  if (overwritep)
    {
      setcarMapMerge(consEnv,e);
    }
  else
    {
      MAKECONS(e,NIL,aux);
      SETQ(consEnv,aux);
    }
  e = vp->expr[2];
  if (overwritep)
    {
      setcarMapMerge(strEnv,e);
    }
  else
    {
      MAKECONS(e,NIL,aux);
      SETQ(strEnv,aux);
    }
  e = vp->expr[3];
  if (overwritep)
    {
      setcarMapMerge(sigEnv,e);
    }
  else
    {
      MAKECONS(e,NIL,aux);
      SETQ(sigEnv,aux);
    }
}

PRIVATE Boolean lexOpenp(name,overwritep,env)
IN CONST SymPtr name;
IN Boolean overwritep;
IN CONST ExprPtr env;
{
  ExprPtr strE,qd;
  Boolean found;
  ENTER;
  
  found = false;
  for (strE = env; strE!=NIL; strE = CDR(strE))
    {
      qd = MapGet(name,CAR(strE),&found);
      if (found)
	{
	  if (qd->kind!=(CHAR)E_ARRAY_GNRL)
	    { /* if not a structure, fail */
	      found = false;
	      break;
	    }
	  lexOpen1(qd,overwritep);
	  break;
	}
    }
  QUIT;
  return found;
}

PUBLIC void lexOpen(loc,name,overwritep,env)
IN CONST struct yyltype *loc;
IN CONST SymPtr name;
IN Boolean overwritep;
IN CONST ExprPtr env;
{
  TZone zone;

  if (!lexOpenp(name,overwritep,env))
    {
      makeZone(&zone,loc,loc);
      strError(&zone,name);
    }
}

PRIVATE void lexOpenz(zone,name,overwritep,env)
IN CONST TZone *zone;
IN CONST SymPtr name;
IN Boolean overwritep;
IN CONST ExprPtr env;
{
  if (!lexOpenp(name,overwritep,env))
    strError(zone,name);
}

PRIVATE ExprPtr longLexOpen1(longsym,sigp)
IN CONST ExprPtr longsym;
IN Boolean sigp;
{
  TZone *zone;
  ExprPtr args,spath,env;

  zone = ZONE(longsym);
  args = ARGS(longsym);
  for (spath=CDR(args); spath!=NIL; spath=CDR(spath))
    lexOpenz(zone,CAR(spath),false,strEnv);
  env = sigp?sigEnv:strEnv;
  return KEEP(env);
}

PUBLIC void longLexOpen(longsym,sigp)
IN CONST ExprPtr longsym;
IN Boolean sigp;
{
  ExprPtr env;

  keepEnv();
  env = longLexOpen1(longsym,sigp);
  restoreEnv();
  lexOpenz(ZONE(longsym),CAR(ARGS(longsym)),true,env);
}

PUBLIC void longLexOpenList(longsymlist,sigp)
IN CONST ExprPtr longsymlist;
IN Boolean sigp;
{
  for (; longsymlist!=NIL; longsymlist=CDR(longsymlist))
    longLexOpen(CAR(longsymlist),sigp);
}

PRIVATE ExprPtr makeStruct()
{
  ExprPtr aux[4];
  ExprPtr e;
  ENTER;

  aux[0] = CAR(lexEnv);
  aux[1] = CAR(consEnv);
  aux[2] = CAR(strEnv);
  aux[3] = CAR(sigEnv);
  e = Karray(E_ARRAY_GNRL,NIL,4,aux);
  KEEPONLY(e);
  return e;
}

PRIVATE void popEnvMake(name,var)
IN CONST SymPtr name;
INOUT ExprPtr *var;
{
  ExprPtr aux;
  ENTER;

  aux = makeStruct();
  popEnv();
  setcarMapPut(name,aux,*var);
  QUIT;
}

PUBLIC void popEnvMakeStruct(strname)
IN CONST SymPtr strname;
{
  popEnvMake(strname,&strEnv);
}

PUBLIC void popEnvMakeSig(signame)
IN CONST SymPtr signame;
{
  popEnvMake(signame,&sigEnv);
}

PRIVATE void poppopEnvMake(funname,inname,var)
IN CONST SymPtr funname,inname;
INOUT ExprPtr *var;
{
  ExprPtr str;
  ExprPtr sig;
  ExprPtr aux;
  ENTER;

  str = makeStruct();
  popEnv();
  sig = MapGetSure(inname,CAR(strEnv));
  popEnv();
  aux = KONS(sig,str);
  setcarMapPut(funname,aux,*var);
  QUIT;
}

PUBLIC void poppopEnvMakeFunctor(funname,inname)
IN CONST SymPtr funname,inname;
{
  poppopEnvMake(funname,inname,&strEnv);
}

PUBLIC void poppopEnvMakeFunctorSig(funname,inname)
IN CONST SymPtr funname,inname;
{
  poppopEnvMake(funname,inname,&sigEnv);
}

PUBLIC void saveLexDummy(env)
INOUT ExprPtr env;
{
  Boolean found;
  ExprPtr cdr;
  MapPtr dummy;

  cdr = CDR(env);
  dummy = MapGet(NIL,CAR(cdr),&found);
  if (!found)
    dummy = CAR(env);
  else
    {
      dummy = MapMerge(dummy,CAR(env));
    }
  popEnv();
  setcarMapPut(NIL,dummy,cdr);
  SETCAR(env,NewMap());
}

PUBLIC void initSaveLexDummy()
{
  pushEnv();
  pushEnv();
}

PUBLIC void endSaveLexDummy(env)
INOUT ExprPtr env;
{
  ExprPtr local;

  env = CDR(env);
  popEnv();
  local = MapGetSure(NIL,CAR(env));
  env = CDR(env);
  setcarMapMerge(env,local);
  popEnv();
}

#define MatchCons(cons1,cons2) SetIncludedp(cons2,cons1)?(cons1):NIL
PRIVATE ExprPtr MatchStruct P((IN CONST MapPtr str1,IN CONST MapPtr str2));

PRIVATE ExprPtr MatchLex(lex1,lex2)
IN CONST MapPtr lex1,lex2;
{ /* must check that MapIncludedp(lex2,lex1), except for those identifiers
     that are declared nonFix in lex2;
     if it is declared nonFix in lex2, then it is declared nonFix in the
     coerced lexEnv we get.
     NO: signature coercion should act as new fixity directives,
     and never fail. */
  ExprPtr lex;
  ENTER;
  /*lex = (MapIncludedp(MapRngRestrBy(lex2,Singleton(izero)),lex1))?lex2:NIL;*/
  lex = MapMerge(lex1,lex2);
  QUIT;
  return lex;
}

PRIVATE ExprPtr matchstruct(str1,str2)
IN CONST ExprPtr str1,str2;
{
  VecPtr vp1,vp2;
  ExprPtr lexE,consE,strE,sigE;
  VecPtr vp;
  ENTER;

  vp1 = AVAL(str1);
  vp2 = AVAL(str2);
  lexE = MatchLex(vp1->expr[0],vp2->expr[0]);
  if (lexE==NIL)
    return NIL;
  consE = MatchCons(vp1->expr[1],vp2->expr[1]);
  if (consE==NIL)
    return NIL;
  strE = MatchStruct(vp1->expr[2],vp2->expr[2]);
  if (strE==NIL)
    return NIL;
  sigE = MatchStruct(vp1->expr[3],vp2->expr[3]);
  if (sigE==NIL)
    return NIL;
  CHECKHETEROFREE();
  {
    ExprPtr aux[4],e;

    aux[0] = lexE;
    aux[1] = consE;
    aux[2] = strE;
    aux[3] = sigE;
    e = Karray(E_ARRAY_GNRL,NIL,4,aux);
    KEEPONLY(e);
    return e;
  }
}

MAPSWEEPFUNC(do_matchstruct)
{
  struct matchstruct *data;
  ExprPtr y1,yy,yy1,yy2;
  Boolean found;

  data = MAPDATA(struct matchstruct *);
  y1 = MapGet(x,data->str1,&found);
  if (!found || y->kind!=y1->kind)
    return true;
  if (y->kind==(CHAR)E_ARRAY_GNRL)
    {
      yy = matchstruct(y1,y,data->exit);
      if (yy==NIL)
	return true;
    }
  else
    {
      yy1 = matchstruct(CAR(y),CAR(y1),data->exit);
      if (yy1==NIL)
	return true;
      yy2 = matchstruct(CDR(y1),CDR(y),data->exit);
      if (yy2==NIL)
	return true;
      yy = KONS(yy1,yy2);
    }
  data->res = MapPut(x,yy,data->res);
  return false;
}

DEF_PRIVATE_ACTION_1ARG(matchstruct)
{
  struct matchstruct *data;
  MapPtr str;
  ENTER;

  data = (struct matchstruct *)ENV_UNDERLYING(env);
  data->exit = arg;
  str = MapCollect(data->str2,do_matchstruct,data);
  KEEPONLY(str);
  return str;
}
ACTION_END

PRIVATE ExprPtr MatchStruct(str1,str2)
IN CONST MapPtr str1,str2;
{
  struct matchstruct *data = (struct matchstruct *)alloctmp(sizeof(struct matchstruct));
  MapStr str;
  ENTER;
  
  data->str1 = str1;
  data->str2 = str2;
  data->res = MapEmpty();
  if (MapSweep(str2,do_matchstruct,data))
    str = NIL;
  else str = data->res;
  KEEPONLY(str);
  return str;
}

PRIVATE void lexStructMatch1(zone,sig)
IN CONST TZone *zone;
IN CONST ExprPtr sig;
{
  ExprPtr lexE,consE,strE,sigE;
  VecPtr vp;
  
  if (sig->kind!=(CHAR)E_ARRAY_GNRL)
    goto error;
  vp = AVAL(sig);
  lexE = MatchLex(CAR(lexEnv),vp->expr[0]);
  if (lexE==NIL)
    goto error;
  consE = MatchCons(CAR(consEnv),vp->expr[1]);
  if (consE==NIL)
    goto error;
  strE = MatchStruct(CAR(strEnv),vp->expr[2]);
  if (strE==NIL)
    goto error;
  sigE = MatchStruct(CAR(sigEnv),vp->expr[3]);
  if (sigE==NIL)
    goto error;
  SETCAR(lexEnv,lexE);
  SETCAR(consEnv,consE);
  SETCAR(strEnv,strE);
  SETCAR(sigEnv,sigE);
  return;
error:
  type_error(zone,(CHAR *)message(17));
}

PRIVATE void lexStructMatch(zone,signame)
IN CONST TZone *zone;
IN CONST SymPtr signame;
/* checks that CAR(env)s match the signature named signame in CAR(CDR(sigEnv)),
   and coerces the CAR(env)s to this signature */
{
  ExprPtr sig;

  sig = MapGetSure(signame,CAR(CDR(sigEnv)));
  lexStructMatch1(zone,sig);
}

PUBLIC void popEnvMakeStructMatch(zone,name,signame)
IN CONST TZone *zone;
IN CONST ExprPtr name,signame;
{
  lexStructMatch(zone,signame);
  popEnvMakeStruct(name);
}

PUBLIC void popOtherEnv()
{
  SETCDR(lexEnv,CDR(CDR(lexEnv)));
  SETCDR(consEnv,CDR(CDR(consEnv)));
  SETCDR(strEnv,CDR(CDR(strEnv)));
  SETCDR(sigEnv,CDR(CDR(sigEnv)));
}

PUBLIC void popEnvStructMatch(zone,signame)
IN CONST TZone *zone;
IN CONST ExprPtr signame;
{
  lexStructMatch(zone,signame);
  popOtherEnv();
}

PRIVATE ExprPtr getFunctor(longsym,sigp)
IN CONST ExprPtr longsym;
IN Boolean sigp;
{
  ExprPtr env,fun;
  Boolean found;

  keepEnv();
  env = longLexOpen1(longsym,sigp);
  restoreEnv();
  for (; env!=NIL; env=CDR(env))
    {
      fun = MapGet(CAR(ARGS(longsym)),CAR(env),&found);
      if (found)
	{
	  if (fun->kind==(CHAR)E_ARRAY_GNRL)
	    break;
	  return fun;
	}
    }
  funError(ZONE(longsym),CAR(ARGS(longsym)));
}

PUBLIC void popEnvApplyFunctor(longsym)
IN CONST ExprPtr longsym;
{ /* applies the functor named longsym to CAR(envs) after popping envs */
  ExprPtr fun;
  ExprPtr saveLexEnv,saveConsEnv,saveStrEnv,saveSigEnv;
  
  saveLexEnv = KEEP(lexEnv);
  saveConsEnv = KEEP(consEnv);
  saveStrEnv = KEEP(strEnv);
  saveSigEnv = KEEP(sigEnv);
  popEnv();
  fun = getFunctor(longsym,false); /* searches from CDR(env) */
  SETQ(lexEnv,saveLexEnv);
  SETQ(consEnv,saveConsEnv);
  SETQ(strEnv,saveStrEnv);
  SETQ(sigEnv,saveSigEnv); /* reinstall old environment */
  lexStructMatch1(ZONE(longsym),CAR(fun)); /* to match the input signature */
  popEnv();
  lexOpen1(CDR(fun),true);
}

PUBLIC void checkUniqueSig(sig,siglist)
IN CONST ExprPtr sig,siglist;
{
  ExprPtr signame;
  TZone *zone;
  
  signame = CAR(ARGS(sig));
  zone = ZONE(signame);
  signame = CAR(ARGS(signame));
  while (siglist!=NIL)
    {
      if (signame==CAR(ARGS(CAR(ARGS(CAR(siglist))))))
	type_error(zone,(CHAR *)errorMsg(message(18),signame));
      siglist = CDR(siglist);
    }
}

PUBLIC void checkUniqueStruct(str,strlist)
IN CONST ExprPtr str,strlist;
{
  ExprPtr strname;
  TZone *zone;
  
  strname = CAR(ARGS(str));
  zone = ZONE(strname);
  strname = CAR(ARGS(strname));
  while (strlist!=NIL)
    {
      if (strname==CAR(ARGS(CAR(ARGS(CAR(strlist))))))
	type_error(zone,(CHAR *)errorMsg(message(20),strname));
      strlist = CDR(strlist);
    }
}

PUBLIC void checkUniqueFunctor(fun,funlist)
IN CONST ExprPtr fun,funlist;
{
  ExprPtr funname;
  TZone *zone;
  
  funname = CAR(ARGS(fun));
  zone = ZONE(funname);
  funname = CAR(ARGS(funname));
  while (funlist!=NIL)
    {
      if (funname==CAR(ARGS(CAR(ARGS(CAR(funlist))))))
	type_error(zone,(CHAR *)errorMsg(message(21),funname));
      funlist = CDR(funlist);
    }
}
#endif

PUBLIC void checkUniqueValDesc(val,vallist)
IN CONST ExprPtr val,vallist;
{
  ExprPtr valname;
  TZone *zone;
  
  valname = CAR(ARGS(val));
  zone = ZONE(valname);
  valname = CAR(ARGS(valname));
  while (vallist!=NIL)
    {
      if (valname==CAR(ARGS(CAR(ARGS(CAR(vallist))))))
	type_error(zone,(CHAR *)errorMsg(message(19),valname));
      vallist = CDR(vallist);
    }
}

PRIVATE void basisSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  Basis *b;

  b = (Basis *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	CHAR c;

	c = b->syntax_analysis;
	outChars(&c,1);
	if (errno) return;
#if DEBUGGING
	c = b->evaluatingp;
#else
	c = 0;
#endif
	outChars(&c,1);
	if (errno) return;
#if DEBUGGING
	c = b->stopwanted;
#else
	c = 0;
#endif
	outChars(&c,1);
	if (errno) return;
	c = b->unit_recognized;
	outChars(&c,1);
	if (errno) return;
	outInt(b->star_recognized);
	if (errno) return;
	outInt(b->union_recognized);
	if (errno) return;
      }
      goto markdata;
    case X_MARK:
      { /* We never save the excHandler part of bases,
	   so this code is only for the X_MARK case: */
	REGISTER excHandler *exc;

	for (exc = b->excStack; exc!=NULL; exc = exc->next)
	  switch (exc->tag)
	    {
	    case X_HANDLER:
	      (*doext->action)(exc->code);
	      /*fallthrough*/
	    case X_PROF:
	      (*doext->action)(exc->data);
	      break;
	    }
      }
      /*fallthrough*/
    case X_MARKDATA:
markdata:
      (*doext->action)(b->parseMarker);
      (*doext->action)(b->parseMarkerStack);
      (*doext->action)(b->lexEnv);
      (*doext->action)(b->consEnv);
      (*doext->action)(b->strEnv);
      (*doext->action)(b->sigEnv);
      (*doext->action)(b->basisState);
      (*doext->action)(b->basisAssum);
      (*doext->action)(b->basisTable);
      (*doext->action)(b->basisEnv);
      (*doext->action)(b->bindings);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr basisLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  Basis *b;
  ExprPtr e;
  CHAR c;
  ENTER;

  b = (Basis *)MALLOC(sizeof(Basis));
  inChars(&c,1);
  if (errno)
    {
err:
      MARKFREE(b);
      return NIL;
    }
  b->syntax_analysis = c;
  inChars(&c,1);
  if (errno) goto err;
#if DEBUGGING
  b->evaluatingp = c;
#endif
  inChars(&c,1);
  if (errno) goto err;
#if DEBUGGING
  b->stopwanted = c;
#endif
  inChars(&c,1);
  if (errno) goto err;
  b->unit_recognized = c;
  b->star_recognized = inInt();
  if (errno) goto err;
  b->union_recognized = inInt();
  if (errno) goto err;
  b->parseMarker = (*load)(sharedref);
  if (errno) goto err;
  b->parseMarkerStack = (*load)(sharedref);
  if (errno) goto err;
  b->lexEnv = (*load)(sharedref);
  if (errno) goto err;
  b->consEnv = (*load)(sharedref);
  if (errno) goto err;
  b->strEnv = (*load)(sharedref);
  if (errno) goto err;
  b->sigEnv = (*load)(sharedref);
  if (errno) goto err;
  b->excStack = NULL; /* cannot load an exception stack */
  b->basisState = (*load)(sharedref);
  if (errno) goto err;
  b->basisAssum = (*load)(sharedref);
  if (errno) goto err;
  b->basisTable = (*load)(sharedref);
  if (errno) goto err;
  b->basisEnv = (*load)(sharedref);
  if (errno) goto err;
  b->bindings = (*load)(sharedref);
  if (errno) goto err;
  e = (shared?Kexternal:MakeExternal)(desc,(char *)b);
  KEEPONLY(e);
  return e;
}

PRIVATE extDesc basisDesc = {
  (CHAR *)"basis",
  { 'b','a','s','e' }, NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  basisSweep,
  basisLoad
  };

PUBLIC ExprPtr basisNew(copy)
IN Boolean copy;
{
  Basis *b;
  ExprPtr e,aux;
  ENTER;

  b = (Basis *)MALLOC(sizeof(Basis));
  b->parseMarker = parseMarker;
  b->parseMarkerStack = parseMarkerStack;
  b->syntax_analysis = syntax_analysis;
#if DEBUGGING
  b->evaluatingp = evaluatingp;
  b->stopwanted = stopwanted;
#endif
  b->unit_recognized = unit_recognized;
  b->star_recognized = star_recognized;
  b->union_recognized = union_recognized;
  if (copy)
    {
      b->lexEnv = copyLexEnv();
      b->consEnv = copyConsEnv();
#if MODULE_SYSTEM & STD_ML_MODULE
      b->strEnv = copyStrEnv();
      b->sigEnv = copySigEnv();
#else
      b->strEnv = b->sigEnv = NIL;
#endif
    }
  else
    {
      b->lexEnv = lexEnv;
      b->consEnv = consEnv;
#if MODULE_SYSTEM & STD_ML_MODULE
      b->strEnv = strEnv;
      b->sigEnv = sigEnv;
#else
      b->strEnv = b->sigEnv = NIL;
#endif
    }
  b->excStack = excStack;
  aux = basisState;
  if (copy)
    aux = DupTypeState(aux);
  b->basisState = aux;
  aux = basisAssum;
  if (copy)
    aux = DupTypeAssum(aux);
  b->basisAssum = aux;
  aux = basisTable;
  if (copy)
    aux = DupCompAssum(aux);
  b->basisTable = aux;
  b->basisEnv = basisEnv;
  b->bindings = NIL;
  e = MakeExternal(&basisDesc,(char *)b);
  KEEPONLY(e);
  return e;
}

PUBLIC void basisInstall(e,errp,copy)
IN CONST ExprPtr e;
IN Boolean errp,copy;
{
  Basis *b;
  ExprPtr aux;

  b = (Basis *)EXTDATA(e);
  SETQ(parseMarker,b->parseMarker);
  SETQ(parseMarkerStack,b->parseMarkerStack);
  unit_recognized = b->unit_recognized;
  star_recognized = b->star_recognized;
  union_recognized = b->union_recognized;
  if (errp)
    {
      SETQ(lexEnv,b->lexEnv);
      if (copy)
	{
	  aux = copyLexEnv();
	  SETQ(lexEnv,aux);
	}
      SETQ(consEnv,b->consEnv);
      if (copy)
	{
	  aux = copyConsEnv();
	  SETQ(consEnv,aux);
	}
#if MODULE_SYSTEM & STD_ML_MODULE
      SETQ(strEnv,b->strEnv);
      if (copy)
	{
	  aux = copyStrEnv();
	  SETQ(strEnv,aux);
	}
      SETQ(sigEnv,b->sigEnv);
      if (copy)
	{
	  aux = copySigEnv();
	  SETQ(sigEnv,aux);
	}
#endif
      {
	struct lexOpenToks *toks;

	while (lex_open_toks!=NULL)
	{
	  toks = lex_open_toks->next;
	  MARKFREE(lex_open_toks);
	  lex_open_toks = toks;
	}
      }
      aux = b->basisState;
      if (copy)
	aux = DupTypeState(aux);
      SETQ(basisState,aux);
      aux = b->basisAssum;
      if (copy)
	aux = DupTypeAssum(aux);
      SETQ(basisAssum,aux);
      aux = b->basisTable;
      if (copy)
	aux = DupCompAssum(aux);
      SETQ(basisTable,aux);
      SETQ(basisEnv,b->basisEnv);
    }
  setHandlerStack(b->excStack);
#if DEBUGGING
  evaluatingp = b->evaluatingp;
  stopwanted = debugEnabled && b->stopwanted;
#endif
  syntax_analysis = b->syntax_analysis;
}

PUBLIC ExprPtr exprToDecls(start,end,expr,sym)
IN CONST struct yyltype *start,*end;
IN CONST ExprPtr expr;
IN CONST ExprPtr sym;
{
  ExprPtr decls;
  ENTER;

  decls = Cons(build2(start,end,A_VAL,
		      BUILD(start,end,A_TYPE_VAR_SEQ,exprTyVars(expr,NIL)),
		      build2(start,end,A_BIND,
			     build1(start,end,A_REAL_VAR,sym),
			     expr)),
	       NIL);
  KEEPONLY(decls);
  return decls;
}


PRIVATE void saveBisonState(bs)
OUT struct bisonstate *bs;
{
  bs->yychar = yychar;
  bs->yylval = yylval;
  bs->yylloc = yylloc;
  bs->yynerrs = yynerrs;
  bs->wantCloseFile = wantCloseFile;
}

PRIVATE void loadBisonState(bs)
IN struct bisonstate *bs;
{
  yychar = bs->yychar;
  yylval = bs->yylval;
  yylloc = bs->yylloc;
  yynerrs = bs->yynerrs;
  wantCloseFile = bs->wantCloseFile;
}

DEF_ACTION_1ARG(precompile)
{
  ExprPtr syntax;

  parseMarker = getProtectWrapper();
  if (yyparse()!=0)
    {
      theTypes = NIL;
      theBindings = NIL;
      return ione; /* error */
    }
  InitMarkerStack();
  (void) alloctmp(0L);
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfflush(coreTraceFile);
#endif
  syntax = theSyntax;
  setProtectWrapper(parseMarker,syntax);
  syntax_analysis = false;
  theTypes = tcDeclaration(syntax,basisAssum,basisState);
  (void) alloctmp(0L);
  if (yynerrs!=0)
    {
      theTypes = NIL;
      theBindings = NIL;
      type_error(NULL,NULL); /* abort elaboration */
    }
  if (yy_input_device==YY_INPUT_STDIN)
    {
      COUNTER save_maxlines,maxlines;

      save_maxlines = crtab_maxlines;
      crtab_maxlines = maxlines = IVAL(REFVAL(maxLinesRef));
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      fxReset();
#endif
      (void) (typeEnvPrint(theTypes,do_printfile,hstdout,basisState)
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	      || FxPretty(theFx,do_printfile,hstdout,basisState)
	      || fxFlush(do_printfile,hstdout,basisState)
#endif
	      );
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      resetFx();
#endif
      if (crtab_maxlines<=0)
	hfputs("\n...\n",hstdout);
      (void) hfflush(hstdout);
      crtab_maxlines = save_maxlines;
    }
  (void) alloctmp(0L);
  SETQ(C_TYSTATE(basisTable),basisState);
  theBindings = cp(syntax,basisTable);
  SETQ(S_PACKERS(basisState),NIL);
  SETQ(S_ANNOTATIONS(basisState),NIL);
  SETQ(C_TYSTATE(basisTable),NIL);
  if (yynerrs!=0)
    {
      theTypes = NIL;
      theBindings = NIL;
      type_error(NULL,NULL); /* abort elaboration */
    }
  return izero; /* toplevel phase correctly completed */
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(gimmlparse)
{
  ExprPtr res;
  ExprPtr types,bindings,state,assum,table,envv;
  VecPtr vp;
  COUNTER n;

  res = ACTION_NAME(precompile)(env,arg); /* no need to call ENV_CONS(ENV_UNDERLYING(env),1),
					     since this is already equal to env. */
  types = theTypes;
  bindings = theBindings;
  if (res!=izero)
    {
      theTypes = NIL;
      theBindings = NIL;
      return res;
    }
  resetHandlerStack();
#if DEBUGGING
  evaluatingp = true;
  stopwanted = false;
#endif
  /* save the following; we shall need them after evaluating;
     but we may return from Eval*() functions by LONGJUMP(),
     so we have to reinstall a coherent version of the
     world. */
  state = DupTypeState(basisState);
  assum = DupTypeAssum(basisAssum);
  table = DupCompAssum(basisTable);
  envv = KEEP(basisEnv);
  vp = EVARS(basisEnv);
  n = nLocals(basisTable);
  {
    if (EvalBindingsRCIt(bindings,
			 basisEnv
			 /* there is something which could fail to work here:
			    normally, vp->n should be the max size of vp,
			    otherwise copying might shrink vp to the current
			    vp->n; this is not so, because of the special
			    treatment of GLOBENV in GCmark().
			 */
			 ,vp->n,n))
      {
#if DEBUGGING
	evaluatingp = false;
#endif
	(void) Raise(BINDEXC); /* because evaluatingp is false,
				  Raise won't return */
      }
  }
#if DEBUGGING
  evaluatingp = false;
#endif
  /* reload coherent version of the world. */
  SETQ(basisState,state);
  SETQ(basisAssum,assum);
  SETQ(basisTable,table);
  SETQ(basisEnv,envv);
  /*vp->n = n;*/ /* done by EvalBindingsRC() */
  AddTypeAssum(basisAssum,types);
  (void) alloctmp(0L);
  if (yy_input_device==YY_INPUT_STDIN)
    {
      valsPrint(types,do_printfile,hstdout);
      hfflush(hstdout);
    }
  setProtectWrapper(parseMarker,NIL);
  (void) alloctmp(0L);
  return izero; /* toplevel phase correctly completed */
}
ACTION_END

PUBLIC void ResetCurrentAge(age)
IN INT age;
{
  if (age<currentAge)
    {
      struct cslist *csl,**cslp;

      currentAge = age;
      for (cslp = &Checksums; csl = *cslp, csl!=NULL;)
	{
	  if (csl->age > age)
	  {
	    *cslp = csl->next;
	    MARKFREE(csl);
	  }
	else
	  cslp = &csl->next;
      }
    }
}

PRIVATE int gimmlparse1(action)
IN compiledAction action;
{
  ExprPtr res;
  int code;
  
  if ((code=SETJUMP(toplevelJmpBuf))==0)
    {
      toplevelJmpBufInstalled = true;
      res = (*action)(ENV_CONS(NIL,1), NIL);
      if (res->kind==E_INT)
	code = (int)IVAL(res);
      else code = 0;
    }
  toplevelJmpBufInstalled = false;
  return code;
}

PUBLIC int GimMLParse1(action)
IN compiledAction action;
{
  ExprPtr basis;
  char *mark;
  INT age;
  int code;
  ENTER;

  basis = basisNew(true);
  InitMarkerStack();
  unit_recognized = false;
  star_recognized = 0;
  union_recognized = 0;
#if CORE_TRACE
  resetTracingYyLex();
#endif
  setTopEnv();
  syntax_analysis = true;
  mark = getMarkFiles();
  age = currentAge;
  code = gimmlparse1(action);
  (void) alloctmp(0L);
  if (code==2) /* no memory left */
    GCgrow(10, NULL); /* get back memory urgently */
  ResetCurrentAge(age);
  if (code!=0)
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	hfflush(coreTraceFile);
#endif
      basisInstall(basis,true,false);
      ResetFiles(mark,true);
      yyinit();
    }
  else
    {
      if (wantCloseFile)
	{
	  wantCloseFile = false;
	  CloseFile();
	}
      basisInstall(basis,false,false);
    }
  (void) alloctmp(0L);
  QUIT;
  return code;
}

PUBLIC int GimMLParse()
{
  int res;

  res = GimMLParse1(ACTION_NAME(gimmlparse));
  currentAge++; /* because we evaluated some declarations, our age increases;
		   used to be important in the presence of first-class
		   continuations. */
  return res;
}

DEF_PRIVATE_ACTION_1ARG(gimmlcompile)
{
  ExprPtr basis;
  
  basis = moduleCompile(compile_file);
  return (basis==NIL)?ione:izero;
}
ACTION_END

PUBLIC int GimMLCompile()
{
  int res;

  res = GimMLParse1(ACTION_NAME(gimmlcompile));
  return res;
}

PRIVATE ExprPtr preCompile1(stk)
IN CONST char *stk;
{
  ExprPtr bindings,types,res;
#if CORE_TRACE
  hFILE *savecoretracefile;
#endif
  TZone *zone;
  ENTER;

  bindings = NIL;
  types = NewTypeAssum();
  A_KIND(types) = A_KIND(basisAssum);
  res = Cons(types,bindings);
#if CORE_TRACE
  savecoretracefile = coreTraceFile;
  coreTraceFile = NULL;
#endif
  zone = (TZone *)alloctmp(sizeof(TZone));
  if (minDebugLevel!=0)
    {
      C_LEVEL(basisTable) = C_ILEVEL(basisTable) = minDebugLevel;
      C_NLOCALS(basisTable) = 0;
    }
  while (getMarkFiles()!=stk)
    {
      zone->line1 = yylloc.last_line;
      zone->pos1 = yylloc.last_column;
      if (GimMLParse1(ACTION_NAME(precompile))==0)
	{
	  zone->line2 = yylloc.last_line;
	  zone->pos2 = yylloc.last_column;
	  /* don't increase currentAge, as we did not evaluate anything. */
	  bindings = Nreverse1(theBindings,bindings);
	  SETCDR(res,bindings);
	  if (A_KIND(types)==TA_PRECOMPILE)
	    {
	      MapPtr err;

	      err = MapDomRestrTo (MapDomRestrBy (MapDomRestrTo (A_IDEXTS(types),
								 A_IDS(types)),
						  A_IDEXTS(theTypes)),
				   A_IDS(theTypes));
	      if (err!=NewMap()) /* then there are imported values that are redefined */
		{
		  type_soft_error (zone, (CHAR *)varMsg (message(244), ChooseInDom(err)));
		  goto err;
		}
	    }
	  AddTypeAssum(types,theTypes);
	  AddTypeAssum(basisAssum,theTypes);
	  SETCAR(res,types);
	  KEEPONLY(res);
	}
      else /* error */
	{
	err:
	  ResetFiles(stk,true);
	  yyinit();
	  QUIT;
	  res = NIL;
	  break;
	}
    }
#if CORE_TRACE
  coreTraceFile = savecoretracefile;
#endif
  theTypes = NIL;
  theBindings = NIL;
  return res;
}

PRIVATE ExprPtr getModuleName(filename)
IN CONST CHAR *filename;
{
  CHAR *s,*send,c;
  ExprPtr name;

  for (s=send=filename+strlen((char *)filename); --s>=filename; )
    { /* strip off directory part */
      c = *s;
      if (c==DIR_CHAR || c==VOL_CHAR)
	break;
      if (c=='.' && s[1]=='m' && s[2]=='l' &&
	  (s[3]==0 || s[4]==0))
	send = s; /* strip off .ml? extension at the end */
#ifdef HML_EXTENSION
      else if (c=='.' && s[1]=='h' && s[2]=='m' && s[3]=='l' &&
	       (s[4]==0 || s[5]==0))
	send = s; /* strop off .hml? extension at the end */
#endif
    }
  s++;
  name = Kstring((INT)(send-s),s);
  return KEEP(name);
}

MAPCOLLECTFUNC(do_trimcons)
{
  ExprPtr tag;
  Boolean found;

  tag = MapGet(y,MAPDATA(MapPtr),&found);
  return found?MapSingleton(y,tag):NewMap();
}

PRIVATE void trimTable(table,assum,oldtable)
INOUT ExprPtr table,oldtable; /* a compile_assum */
IN CONST ExprPtr assum; /* a type_assum */
{
  ExprPtr aux;
  ENTER;

  aux = MapDomRestrTo(C_IDS(table),A_IDS(assum));
  SETQ(C_IDS(table),aux);
#if MODULE_SYSTEM & STD_ML_MODULE
  aux = MapDomRestrTo(C_STRUCTS(table),A_STRUCTS(assum));
  SETQ(C_STRUCTS(table),aux);
  aux = MapDomRestrTo(C_SIGS(table),A_SIGS(assum));
  SETQ(C_SIGS(table),aux);
#endif
  aux = MapCollect(C_IDS(table),do_trimcons,C_CONS(table));
  SETQ(C_CONS(table),aux);
  setMapDomRestrBy(&C_TYTAGS(table),C_TYTAGS(oldtable));
  QUIT;
}

PRIVATE void trimState(state,oldstate)
INOUT ExprPtr state; /* a type_state */
IN CONST ExprPtr oldstate; /* a type_state */
{
  ExprPtr aux;
  ENTER;

  aux = MapDomRestrBy(S_NONEQS(state),S_NONEQS(oldstate));
  SETQ(S_NONEQS(state),aux);
  aux = MapDomRestrBy(S_TYCONS(state),S_TYCONS(oldstate));
  SETQ(S_TYCONS(state),aux);
  QUIT;
}

PRIVATE void trimEnv(env,oldenv)
INOUT ExprPtr env; /* a list of maps */
IN CONST ExprPtr oldenv; /* a list of maps */
{
  ExprPtr ll;
  MapPtr old,diff;
  ENTER;

  old = NewMap();
  for (ll=oldenv; ll!=NIL; ll=CDR(ll))
    old = MapMerge(CAR(ll),old);
  KEEPONLY(old);
  for (ll=env; ll!=NIL; ll=CDR(ll))
    {
      diff = MapDelta(CAR(ll),old); /* contains all entries in CAR(ll) that are not also in old,
				       and all entries in old that are not in CAR(ll). */
      diff = MapDomRestrTo(diff,CAR(ll)); /* now diff only contains entries in CAR(ll) that are
					     different from entries in old. */
      SETCAR(ll,diff);
    }
  QUIT;
}

PUBLIC ExprPtr preCompile(filename,realname)
IN CONST CHAR *filename;
OUT CHAR **realname;
/* return basis, or NIL */
{
  struct bisonstate *bs;
  ExprPtr oldbasis,basis,initbasis;
  ExprPtr bindings,types,res;
  ExprPtr saveModuleName;
  COUNTER saveMinDebugLevel;
  Boolean okp;
  char *stk0;
  ENTER;

  basis = NIL;
  bs = (struct bisonstate *)alloctmp(sizeof(struct bisonstate));
  saveBisonState(bs);
  stk0 = getMarkFiles();
  okp = OpenFile(filename);
  if (okp)
    {
      if (realname!=(CHAR **)0L)
	{
	  COUNTER len;

	  len = strlen((char *)yyfilename);
	  *realname = (CHAR *)MALLOC(len+1);
	  strcpy((char *)*realname,(char *)yyfilename);
	}
      saveModuleName = KEEP(moduleName);
      SETQ(moduleName,getModuleName(filename));
      saveMinDebugLevel = minDebugLevel;
      minDebugLevel = 1;
      oldbasis = basisNew(false);
      initbasis = CurrentChecksum()->basis;
      basisInstall(initbasis,true,true);
      A_KIND(basisAssum) = TA_PRECOMPILE;
      wantCloseFile = false;
      res = preCompile1(stk0);
      if (res!=NIL) /* no error */
	{
	  if (S_NONGENS(basisState)!=NewMap())
	    {
	      type_soft_error(NULL,(CHAR *)message(792));
	    }
	  else
	    {
	      bindings = CDR(res);
	      bindings = Nreverse(bindings);
#if TRANSLATOR
	      bindings = Compile(bindings,minDebugLevel,true);
	      /* bindings is now a closure (a tuple2 of an E_BYTECODE
		 and an environment), not a list of bindings anymore. */
#endif
	      types = CAR(res);
	      trimTable(basisTable,types,((Basis *)EXTDATA(initbasis))->basisTable);
	      trimState(basisState,((Basis *)EXTDATA(initbasis))->basisState);
	      trimEnv(consEnv,((Basis *)EXTDATA(initbasis))->consEnv);
	      trimEnv(lexEnv,((Basis *)EXTDATA(initbasis))->lexEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
	      trimEnv(strEnv,((Basis *)EXTDATA(initbasis))->strEnv);
	      trimEnv(sigEnv,((Basis *)EXTDATA(initbasis))->sigEnv);
#endif
	      basis = basisNew(false);
	      SETQ(((Basis *)EXTDATA(basis))->basisAssum,types);
	      /* we hack the following two fields to hold
		 bindings and the module name instead of
		 various environments, which are pointless here. */
	      SETQ(((Basis *)EXTDATA(basis))->bindings,bindings);
	      SETQ(((Basis *)EXTDATA(basis))->basisEnv,moduleName);
	      SETQ(((Basis *)EXTDATA(basis))->parseMarker,NIL);
	      SETQ(((Basis *)EXTDATA(basis))->parseMarkerStack,NIL);
	      HITBARRIER(basis);
	    }
	}
      basisInstall(oldbasis,true,false);
      minDebugLevel = saveMinDebugLevel;
      SETQ(moduleName,saveModuleName);
      loadBisonState(bs);
    }
  KEEPONLY(basis);
  return basis;
}

PRIVATE void outI(n,out)
IN INT n;
IN hFILE *out;
{
  COUNTER i;

  {
    BEGIN_FROM_INT(n);
    for (i=0; i<sizeof(INT); i++)
      hfputc(FROM_INT(n)[i],out);
    END_FROM_INT(n);
  }
}

PRIVATE void outDepend(mll,out)
IN CONST ExprPtr mll;
IN hFILE *out;
{
  INT len,i;
  ExprPtr e;

  len = ListLength(mll);
  outI(len,out);
  while (errno==0 && mll!=NIL)
    {
      hfputc('\n',out);
      if (errno) break;
      e = CAR(mll);
      len = SLEN(e);
      outI(len,out);
      if (errno) break;
      for (i=0; i<len; i++)
	{
	  hfputc(STEXT(e)[i],out);
	  if (errno) break;
	}
      mll = CDR(mll);
    }
  if (errno==0)
    { hfputc('\n',out); }
}

PRIVATE void recordOpenMod(name,dirp)
IN CONST char *name;
IN Boolean dirp;
{
  if (openModRef!=NULL)
    {
      INT len;
      char *s;
      ExprPtr str,e;
      ENTER;

      len = strlen(name);
      s = name+len;
      if (len>=3 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='.'))
	{ len -= 3; /* chop off '.ml' */ }
      else if (len>=4 && (s[-1]=='x' && s[-2]=='l' && s[-3]=='m' && s[-4]=='.'))
	{ len -= 4; /* chop off '.mlx' */ }
#ifdef HML_EXTENSION
      else if (len>=4 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='h' && s[-4]=='.'))
	{ len -= 4; /* chpo off '.hml' */ }
#endif
      if (dirp)
	{
	  CHAR *pwd,*xname;

	  pwd = (CHAR *)alloctmp(PWD_BUFSIZE+len+1);
	  pwd = getDir(pwd,PWD_BUFSIZE);
	  xname = pwd+strlen((char *)pwd);
	  if (xname>pwd && xname[-1]!=VOL_CHAR && xname[-1]!=DIR_CHAR)
	    *xname++ = DIR_CHAR;
	  safememcpy(xname,name,len);
	  str = Kstring((xname-pwd)+len,pwd);
	}
      else
	str = Kstring(len,(CHAR *)name);
      (void) KEEP(str);
      e = Cons(str,REFVAL(openModRef));
      SETREFVAL(openModRef,e);
      QUIT;
    }
}

PUBLIC ExprPtr moduleCompile(filename)
IN CONST CHAR *filename;
/* return basis, or NIL */
{
  ExprPtr basis;
  hFILE *out;
  CHAR *compname,*realname,*s;
  INT len;
  ExprPtr saveOpenModRef;
  ENTER;

  saveOpenModRef = openModRef;
  openModRef = MakeSing(E_REF,NIL);
  (void) KEEP(openModRef);
  basis = preCompile(filename,&realname);
  if (basis!=NIL)
    {
      recordOpenMod((char *)realname,false);
      /* don't want to save parseMarker info or continuations to disk */
      len = strlen((char *)realname);
      compname = (CHAR *)alloctmp_anchored(len+5, (char *)&out);
      strcpy((char *)compname,(char *)realname);
      s = compname+len;
      if (len>=3 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='.'))
	{
	  s[0] = 'x'; s[1] = 0;
	}
      else
	{
	  s[0] = '.'; s[1] = 'm'; s[2] = 'l'; s[3] = 'x'; s[4] = 0;
	}
      MARKFREE(realname);
      errno = 0;
      out = hfopen((char *)compname,"w");
      if (out!=NULL)
	{
	  hfputs("#!",out);
	  hfputs(INSTALL_DIR,out);
	  hfputs("/gimmlrun\n",out);
	  hfputs(message(807),out);
	  if (errno==0)
	    do_printfile(CurrentChecksum()->cs,CHECKSUM_WIDTH,(char *)out);
	  if (errno==0)
	    outDepend(REFVAL(openModRef),out);
	  if (errno==0)
	    outTransfer(basis,do_printfile,(char *)out,false);
	  (void) hfclose(out);
	}
      if (errno)
	{
	  char *s;

	  s = (char *)alloctmp_anchored(128, (char *)&out);
	  sprintf(s,message(803),compname,errno,ioerrormsg(errno));
	  type_warn(NULL,(CHAR *)s);
      }
    }
  if (saveOpenModRef!=NULL)
    {
      ExprPtr ll;

      ll = Append0(REFVAL(openModRef),REFVAL(saveOpenModRef));
      SETREFVAL(saveOpenModRef,ll);
    }
  openModRef = saveOpenModRef;
  KEEPONLY(basis);
  return basis;
}

READFUNC(readfile)
{
  REGISTER hFILE *hf;
  REGISTER COUNTER i;
  REGISTER int c;

  hf = READDATA(hFILE *);
  for (i=0; i<len; i++)
    {
      c = hgetc(hf);
      if (c==-1)
	{
	  errno = EIO; /* so that inTransfer1() detects a problem. */
	  return;
	}
      *text++ = c;
    }
}

PRIVATE INT inI(mod)
IN hFILE *mod;
{
  INT n;
  COUNTER i;

  {
    BEGIN_TO_INT(n);
    for (i=0; i<sizeof(INT); i++)
      TO_INT(n)[i] = hfgetc(mod);
    END_TO_INT(n);
  }
  return n;
}

PRIVATE Boolean checkOpenMods(mlxt,mod)
IN CONST time_t mlxt;
IN hFILE *mod;
{
  INT i,n,len,j;
  char *buf,*s;
  struct stat st;

  n = inI(mod);
  if (errno) return false;
  if (hfgetc(mod)!='\n' || errno) return false;
  for (i=0; i<n; i++)
    {
      len = inI(mod);
      if (errno) return false;
      buf = alloctmp(len+5);
      for (j=0, s=buf; j<len; j++)
	*s++ = hfgetc(mod);
      *s++ = '.';
      *s++ = 'm';
      *s++ = 'l';
      *s = 0;
      if (stat(buf,&st)==0 && st.st_mtime>mlxt)
	return false;
      *s++ = 'x';
      *s = 0;
      if (stat(buf,&st)==0 && st.st_mtime>mlxt)
	return false;
      if (hfgetc(mod)!='\n' || errno) return false;
      recordOpenMod(buf,false);
    }
  return true;
}

PRIVATE Boolean skipOpenMods(mod)
IN hFILE *mod;
{
  INT i,n,len,j;

  n = inI(mod);
  if (errno) return false;
  if (hfgetc(mod)!='\n' || errno) return false;
  for (i=0; i<n; i++)
    {
      len = inI(mod);
      if (errno) return false;
      for (j=0; j<len; j++)
	(void) hfgetc(mod);
      if (hfgetc(mod)!='\n' || errno) return false;
    }
  return true;
}

PRIVATE hFILE *openModule(name,realname,basisp,recompile)
IN CONST CHAR *name;
OUT char **realname;
OUT ExprPtr *basisp;
IN Boolean recompile;
{
  hFILE *module;
  COUNTER len;
  CHAR *xname,*s;

  len = strlen((char *)name);
  /* otherwise, try adding .mlx at the end of the name */
  xname = (CHAR *)alloctmp_anchored(len+5, (char *)&realname);
  strcpy((char *)xname,(char *)name);
  s = xname+len;
  if (recompile)
    {
      if (len>=3 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='.'))
	{ s -= 3; *s = 0; /* chop off '.ml' */ }
      else if (len>=4 && (s[-1]=='x' && s[-2]=='l' && s[-3]=='m' && s[-4]=='.'))
	{ s -= 4; *s = 0; /* chop off '.mlx' */ }
#ifdef HML_EXTENSION
      else if (len>=4 && (s[-1]=='l' && s[-2]=='m' && s[-3]=='h' && s[-4]=='.'))
	{ s -= 4; *s = 0; /* chop off '.hml' */ }
#endif
      *s++ = '.'; *s++ = 'm'; *s++ = 'l'; *s++ = 'x'; *s = 0; /* add '.mlx' */
    }
  *basisp = NIL;
  module = xfopen((char *)xname,"r",realname);
  if (module!=NULL)
    { /* if module exists under this name, then: */
      {
	char *buf;

	{
	  int c;

	  while (c=hfgetc(module), c!=-1 && c!='\n');
	  /* skip first line, which should be of the form #!/usr/local/bin/gimmlrun
	     for Unix systems to be able to launch complete bytecode files. */
	}
	len = strlen(message(807)); /* then, check that it has the magic header */
	buf = alloctmp_anchored(len+1, (char *)&realname);
	readfile((CHAR *)buf,len,(char *)module);
	buf[len] = 0;
	if (errno || strcmp(buf,message(807))!=0)
	  goto wrongmlx;
      }
      { /* then, that it has a valid checksum */
	Checksum cs;

	readfile(cs,sizeof(Checksum),(char *)module);
	if (errno || GetChecksum(cs)==(struct cslist *)0L)
	  goto wrongmlx;
      }
      if (recompile)
	{ /* then, that there no .ml (or .hml) of the same name that would be newer */
	  struct stat buf;
	  time_t mlxt;

	  (void) fstat (module->fildesc,&buf);
	  mlxt = buf.st_mtime;
	  if (!checkOpenMods(mlxt,module))
	    goto wrongmlx;
	}
      else
	{
	  if (!skipOpenMods(module))
	    goto wrongmlx;
	}
      return module;
wrongmlx:
      MARKFREE(*realname);
      hfclose(module);
    }
  if (recompile)
    {
      /* if this does not work, try and find a .ml, compile it and
	 load the resulting .mlx file. */
#ifdef HML_EXTENSION
      {
	/* first try a .hml file */
	struct stat buf;

	s[-3] = 'h'; s[-2] = 'm'; s[-1] = 'l';
	if (stat ((char *)xname, &buf)!=0)
	  /* else try a .ml */
	  {
	    s[-3] = 'm'; s[-2] = 'l'; s[-1] = 0;
	  }
      }
#else
      s[-1] = 0; /* erase the 'x' at the end of xname */
#endif
      *basisp = moduleCompile(xname); /* compile .ml to .mlx */
      (void) KEEP(*basisp);
    }
  else
    {
      *basisp = NIL;
    }
  return NULL; /* don't open the file, we've got the basis already */
}

PUBLIC ExprPtr moduleLoad(zone,name,recompile)
IN CONST TZone *zone;
IN CONST CHAR *name;
IN Boolean recompile;
{
  hFILE *new;
  CHAR *realname;
  ExprPtr basis;

  new = openModule(name,(char **)&realname,&basis,recompile);
  if (new==NULL)
    return basis;
  basis = inTransfer(readfile,(char *)new);
  if (errno)
    {
      CHAR *tmp = (CHAR *)alloctmp(128+strlen((char *)realname));

      (void) hfclose(new);
      sprintf((char *)tmp,message(811),realname,errno,ioerrormsg(errno));
      type_error(zone,tmp);
    }
  (void) hfclose(new);
  return basis;
}

MAPSWEEPFUNC(do_errconsenv)
{
  struct errconsenv *data;
  ExprPtr type;
  Boolean exn;

  data = MAPDATA(struct errconsenv *);
  type = MapGetSure(x,data->ids);
  exn = (type==exnType ||
	 (type->kind==E_TYPE_APPL && APPLFUN(type)==funSym &&
	  CAR(CDR(APPLARG(type)))==exnType));
  type_soft_error(data->zone,(CHAR *)varMsg(message(exn?819:820),x));
  return false;
}

PUBLIC void checkConsEnv(zone,basis)
IN CONST TZone *zone;
IN CONST ExprPtr basis;
{
  Basis *b;
  ExprPtr ll,imports,importcons;
  struct errconsenv *data = (struct errconsenv *)alloctmp(sizeof(struct errconsenv));
  ENTER;

  b = (Basis *)EXTDATA(basis);
  imports = MapRngRestrTo(A_IDEXTS(b->basisAssum),unitset);
  importcons = NewMap();
  for (ll=b->consEnv; ll!=NIL; ll=CDR(ll))
    importcons = MapMerge(MapDomRestrTo(CAR(ll),imports),importcons);
  for (ll=consEnv; ll!=NIL; ll=CDR(ll))
    importcons = MapDomRestrBy(importcons,CAR(ll));
  KEEPONLY(importcons);
  data->ids = A_IDS(b->basisAssum);
  data->zone = zone;
  (void) MapSweep(importcons,do_errconsenv,data);
  QUIT;
}

PRIVATE void augmentLexicalEnv1(env,env2)
INOUT ExprPtr env;
IN CONST ExprPtr env2;
{
  ENTER;

  env2 = Nreverse(env2);
  for (; env2!=NIL; env2 = CDR(env2))
    setcarMapMerge(env,CAR(env2));
  QUIT;
}

PUBLIC void augmentLexicalEnv(basis)
IN CONST ExprPtr basis;
{
  Basis *b;

  b = (Basis *)EXTDATA(basis);
  augmentLexicalEnv1(lexEnv,b->lexEnv);
  augmentLexicalEnv1(consEnv,b->consEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  augmentLexicalEnv1(strEnv,b->strEnv);
  augmentLexicalEnv1(sigEnv,b->sigEnv);
#endif
}

PUBLIC struct cslist *CurrentChecksum()
{
  struct cslist *aux;

  while (Checksums->age>currentAge)
    {
      aux = Checksums->next;
      MARKFREE(Checksums);
      Checksums = aux;
    }
  return Checksums;
}

PUBLIC struct cslist *GetChecksum(cs)
IN CONST Checksum cs;
{
  struct cslist *csl;
  REGISTER COUNTER i;
  REGISTER CHAR *s,*t;

  for (csl = Checksums; csl!=NULL; csl = csl->next)
    {
      for (i=0, s=csl->cs, t=cs; i<CHECKSUM_WIDTH; i++)
	if (*s++ != *t++)
	  break;
      if (i==CHECKSUM_WIDTH)
	break;
    }
  return csl;
}

PUBLIC void StoreInitialBasis()
{
  struct cslist *new;
  ExprPtr assum,state,table,e;
  ENTER;

  new = (struct cslist *)MALLOC(sizeof(struct cslist));
  new->next = Checksums;
  new->age = currentAge;
  new->basis = basisNew(true);
  assum = Cons(MapRemove(itSym,A_IDS(basisAssum)),
	       Cons(A_TYFCNS(basisAssum),
#if MODULE_SYSTEM & STD_ML_MODULE
		    Cons(A_STRUCTS(basisAssum),
			 Cons(A_SIGS(basisAssum),
#endif
			      Cons(A_UNITS(basisAssum),
				   NIL)
#if MODULE_SYSTEM & STD_ML_MODULE
			      ))
#endif
		    ));
  state = Cons(S_NONEQS(basisState),
	       Cons(S_TYCONS(basisState),
		    NIL));
  table = Cons(MapRemove(itSym,C_IDS(basisTable)),
#if MODULE_SYSTEM & STD_ML_MODULE
	       Cons(C_STRUCTS(basisTable),
		    Cons(C_SIGS(basisTable),
#endif
			 Cons(C_CONS(basisTable),
			      NIL)
#if MODULE_SYSTEM & STD_ML_MODULE
			 ))
#endif
	       );
  /* we remove 'it' from the computation of the checksum, as its
     type can vary, although this should have no influence on the
     checksum, since modules are not allowed to depend on the value
     of it. */
  e = Cons(assum,Cons(state,table));
  CheckSum(e,new->cs);
  if (GetChecksum(new->cs)==NULL) /* never record twice the same initial basis */
    Checksums = new;
  else
    {
      MARKFREE(new);
    }
  QUIT;
}

DEF_ACTION_0ARG(stampworld)
{
  StoreInitialBasis();
  return tuple0;
}
ACTION_END

/***** The following are functions that provide a reasonably secure
  hash function for structures in memory (i.e, a reliable checksum).
  This is needed to ensure that things we saved to disk in some environment
  can be reloaded later on, by checking that the environments are the
  same with probability close to 1. *****/

/* The first trick is to encode couples (a,b):
   Let M be a prime, and c, d positive numbers.
   Assume we have checksums for a and b as integers, then we compute
   <a,b> = c*a+b+d mod M. As M is prime, the chance that <a,b>=<a',b'>
   is 1/M. Indeed, this is the chance that cx+y=0 mod M for any two
   x and y; for each x, this is the chance that y = -cx mod M, i.e. 1/M.
   We then choose several different values for c and d (CHECKSUM_WIDTH values),
   which we choose relatively prime, so that the chance of collision
   decreases again.
*/

PRIVATE unsigned short multipliers[] = { 71, 13, 97, 7, 11, 59, 37, 61 };
PRIVATE unsigned short adders[] = { 17, 31, 89, 3, 53, 47, 5, 23 };

/*
   PUBLIC void checksumCouple1(cs1,cs2,cs)
   IN CONST Checksum *cs1,*cs2;
   OUT Checksum *cs;
   {
   cs->line1 = (9821L*(unsigned long)cs1->line1+(unsigned long)cs2->line1+13849L) & 0xFFFFL;
   cs->line2 = (17341L*(unsigned long)cs1->line2+(unsigned long)cs2->line2+38653L) & 0xFFFFL;
   cs->pos1 = (8827L*(unsigned long)cs1->pos1+(unsigned long)cs2->pos1+14279L) & 0xFFFFL;
   cs->pos2 = (18563L*(unsigned long)cs1->pos2+(unsigned long)cs2->pos2+13443L) & 0xFFFFL;
   }
*/

PRINTFUNC(checksumPrint)
{
  struct checksum *data = PRINTDATA(struct checksum *);
  REGISTER COUNTER i,j;
  REGISTER CHAR *cs;
  REGISTER unsigned short *mul,*add;
  REGISTER unsigned short c;

  for (i=0; i<len; i++)
    {
      c = (unsigned short)*text++;
      for (j = 0, cs = data->cs, mul = multipliers, add = adders; j<CHECKSUM_WIDTH; j++)
	{
	  *cs = (*mul++)*((unsigned short)(*cs))+c+(*add++);
	  cs++;
	}
    }
}

PUBLIC void CheckSum(e,cs)
IN CONST ExprPtr e;
OUT Checksum cs;
{
  struct checksum *data = (struct checksum *)alloctmp(sizeof(struct checksum));
  COUNTER i;

  for (i=0; i<CHECKSUM_WIDTH; i++)
    data->cs[i] = 0;
  outTransfer(e,checksumPrint,(char *)data,true);
  memcpy(cs,data->cs,sizeof(Checksum));
}

PRIVATE ExprPtr S(i)
IN INT i;
{
  return Kstr((CHAR *)message(i));
}

PUBLIC void InitLex()
{
  ExprPtr aux;
  ENTER;

  yyin = hstdin;
  yyout = hstdout;
  lexEnv=consEnv=NIL;
#if MODULE_SYSTEM & STD_ML_MODULE
  strEnv=sigEnv=NIL;
#endif
  keepLexEnv=keepConsEnv=NIL;
#if MODULE_SYSTEM & STD_ML_MODULE
  keepStrEnv=keepSigEnv=NIL;
#endif
  lexId=lexPath=NIL;
  lex_open_toks = NULL;
  pushEnv();
  setConstructor(nilSym);
  setConstructor(consSym);
  setConstructor(refsym);
  setConstructor(trueSym);
  setConstructor(falseSym);
  setConstructor(bindSym);
  setConstructor(matchSym);
  setConstructor(emptySym);
  setConstructor(mapgetSym);
  setConstructor(parsweepSym);
  setConstructor(nomemSym);
  setConstructor(stackSym);
  setConstructor(incompSym);
  setConstructor(nthSym);
  setConstructor(rangeSym);
  setConstructor(asciiSym);
  setConstructor(stringnthSym);
  setConstructor(noninvSym);
  setConstructor(nonsenseSym);
  setConstructor(subscriptSym);
  setConstructor(arithSym);
  setConstructor(ioSym);
  setConstructor(reSym);
/*
infix   0 before
infix   3 o := <<< >>>
infix   4 = <> < > <= >= inset inmap submap subset strless
infixr  5 @ (* infix in ML *) O
infixr  5 ::
infix   6 + - ^ ++ U <| <-| |> |->
infix   7 * div mod divmod / & \ delta intersects
infixr  8 **
infix   9 nth to sub
*/
#define I(i) Kint(i)
  aux = MapMerge(MapMerge(MapPair(S(359),I(1), /* infix 0 before */
				  S(301),I(4)), /* infix 3 o */
			  MapPair(S(343),I(4), /* infix 3 := */
				  S(276),I(5))), /* infix 4 = */
		 MapMerge(MapPair(S(277),I(5), /* infix 4 <> */
				  S(579),I(5)), /* infix 4 < */
			  MapPair(S(580),I(5), /* infix 4 > */
				  S(581),I(5)))); /* infix 4 <= */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(582),I(5), /* infix 4 >= */
				  S(328),I(5)), /* infix 4 inset */
			  MapPair(S(329),I(5), /* infix 4 inmap */
				  S(327),I(5)))); /* infix 4 submap */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(326),I(5), /* infix 4 subset */
				  S(358),I(5)), /* infix 4 strless */
			  MapPair(S(306),I(-6), /* infixr 5 @ */
				  S(302),I(-6)))); /* infixr 5 O */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(193),I(-6), /* infixr 5 :: */
				  S(583),I(7)), /* infix 6 + */
			  MapPair(S(584),I(7), /* infix 6 - */
				  S(307),I(7)))); /* infix 6 ^ */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(217),I(7), /* infix 6 ++ */
				  S(223),I(7)), /* infix 6 U */
			  MapPair(S(318),I(7), /* infix 6 <| */
				  S(319),I(7)))); /* infix 6 <-| */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(320),I(7), /* infix 6 |> */
				  S(321),I(7)), /* infix 6 |-> */
			  MapPair(S(222),I(8), /* infix 7 * */
				  S(585),I(8)))); /* infix 7 div */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(586),I(8), /* infix 7 mod */
				  S(587),I(8)), /* infix 7 divmod */
			  MapPair(S(266),I(8), /* infix 7 #/ */
				  S(323),I(8)))); /* infix 7 & */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(324),I(8), /* infix 7 \ */
				  S(325),I(8)), /* infix 7 delta */
			  MapPair(S(622),I(8), /* infix 7 intersects */
				  S(225),I(-9)))); /* infixr 8 #^ */
  aux = MapMerge(aux,
		 MapPair(S(335),I(10), /* infix 9 nth */
			 S(337),I(10))); /* infix 9 to */
			  /* MapSingleton(S(354),I(10)))); // infix 9 sub */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(263),I(7), /* infix 6 #+ */
				  S(264),I(7)), /* infix 6 #- */
			  MapPair(S(265),I(8), /* infix 7 #* */
				  S(315),I(8)))); /* infix 7 #div */
  aux = MapMerge(aux,
		 MapPair(S(316),I(8), /* infix 7 #mod */
			 S(317),I(8))); /* infix 7 #divmod */
  aux = MapMerge(aux,
		 MapMerge(MapPair(S(278),I(5), /* infix 4 #< */
				  S(279),I(5)), /* infix 4 #> */
			  MapPair(S(280),I(5), /* infix 4 #<= */
				  S(281),I(5)))); /* infix 4 #>= */
  setcarMapMerge(lexEnv,aux);
  /*theFileStack = NULL;*/
  yylloc.last_line = 0;
  yylloc.last_column = 0;
  yyfilename = NULL;
  QUIT;
}

PUBLIC void InitGram()
{
  parseMarker=NIL;
  parseMarkerStack=NIL;
  ItSymbol = itSym;
  registerExtDesc(&basisDesc);
  theTypes = NIL;
  theBindings = NIL;
}
