/* Initialization functions.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "eval.h"
#include "mem.h"
#include "applmem.h"
#include "actions.h"
#include "execinit.h"
#include "timing.h"
#include "iniproto.h"
#include "stkproto.h"
#include "exeproto.h"
#include "memproto.h"
#include "conproto.h"
#include "mapproto.h"
#include "symproto.h"
#include "grmproto.h"
#include "typproto.h"
#include "ariproto.h"
#include "evlproto.h"
#include "cmpproto.h"
#include "rndproto.h"
#include "expproto.h"
#include "errproto.h"
#include "dbgproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_READLINE
#include "linenoise.h"
#endif

#undef GETLOOKUPSTATS

#undef SPACE_OK
#if SPECIAL_MALLOC
#define SPACE_OK
#endif
/* SPACE_OK used to be defined on Unix, too,
   but relied on sbrk(0), which is now deprecated: too bad. */

#ifdef HAS_PROTOTYPES
#include "sinit.h"
#endif

PUBLIC actions *doafterinit=NULL;
PUBLIC actions *dobeforeend=NULL;

IMPORT void InitRE P((void));

#if CORE_TRACE
PUBLIC char coreTraceName[32];
PUBLIC hFILE *coreTraceFile;
#define MAKE_CORE_FILE_NAME() strcpy(coreTraceName,"GimML.trace")
PUBLIC Boolean replay_mode = false;
PUBLIC Boolean want_core_trace = false;
#endif
PUBLIC Boolean initializing=true;
PUBLIC INT maxCells = MAX_SIGNED_LONG;

char *coreEnd = "(* END-OF-PROLOG *)";
EXTERN int gcTraceFd;

EXTERN char *codeStatus;
EXTERN ExprPtr *NaNs;

PUBLIC long replay_size=0; /* 0 to get default size */

EXTERN long gc_utime,gc_stime,gc_rtime,gc_count;
PRIVATE struct tms start_tms;

EXTERN COUNTER initCommandMode;

#ifdef SPACE_OK
#if OS==UNIX_SYS_V || OS==UNIX_BSD
/*IMPORT char *sbrk P((int incr));*/
PRIVATE char *start_limit;
#endif
#endif

PRIVATE void lower_init()
{
  InitExprs();
  InitGC();
  InitAllocTables();
  InitMap();
  InitSymbols();
  InitLex();
  InitGram();
  InitTypes();
  InitArith();
  InitEval();
  InitCompile();
  InitExec();
  InitRE();
  InitReadline();
  InitRandom();
#if TRANSLATOR
  InitBC();
#endif
  InitGlex ();
  InitGyacc ();
  InitLIP ();
#if DEBUGGING
  InitDebug();
#endif
#if PROFILING
  InitProfile();
#endif
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfputs(coreEnd,coreTraceFile);
      hfputc('\n',coreTraceFile);
      hfflush(coreTraceFile);
    }
#endif
  {
    actions *doinit;

    for (doinit = doafterinit; doinit!=(actions *)0L; doinit = doinit->next)
      (*doinit->exec)();
  }
  StoreInitialBasis();
  if (initCommandMode!=2)
    /* if no external initialization, put initials away */
    {
      initializing = false;
      cutoffInitials();
    }
  (void) times(&start_tms);
}

PUBLIC void InitML()
{
  long size,replay_begin;
  char line[128],*s;
  int c;
  int intsize,cntsize,ptrsize;
  int Ver,ver,rev;
  char status[128];

  InitHIO();
#ifdef SPACE_OK
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  start_limit = (char *)sbrk(0);
#endif
#endif
#if SPECIAL_MALLOC
  if (replay_size==0)
    {
#if OS==UNIX_SYS_V || OS==UNIX_BSD
      size = MIN_MEMSIZE;
#else
      size = MEM_SIZE;
#endif
    }
  else
    size = replay_size;
#else
  size = -1;
#endif
  replay_begin = 0L;
  intsize = cntsize = ptrsize = 0;
  Ver = ver = rev = 0;
  status[0] = 0;
#if CORE_TRACE
  if (!replay_mode)
    {
      if (want_core_trace)
	{
	  MAKE_CORE_FILE_NAME();
	  coreTraceFile = hfopen(coreTraceName,"w");
	  if (coreTraceFile==NULL)
	    berror(message(133),coreTraceName);
	  if (initCommandMode!=1) /* don't print banner if in -cmd mode */
	    {
	      hfprintf(hstdout,message(134),coreTraceName);
	    }
	}
    }
  else
    {
      coreTraceFile = NULL;
      if (initCommandMode!=1) /* don't print banner if in -cmd mode */
	{
	  hfprintf(hstdout,message(135));
	}
    }
  hfflush(hstdout);
#endif
  if (replay_mode)
    {
      do {
	s = line;
	do {
	  c = hgetc(hstdin);
	  if (c==-1)
	    berror(message(136));
	  if (c=='\n')
	    {
	      *s = 0;
	      break;
	    }
	  *s++ = c;
	  if (s>=line+sizeof(line))
	    {
	      s[-1] = 0;
	      bmsg(message(137));
	      break;
	    }
	} while (true);
	if (sscanf(line,message(138),&replay_size)==1)
	  size = replay_size;
	else if (sscanf(line,message(139),&replay_begin)==1)
	  ;
	else if (sscanf(line,message(140),&intsize)==1)
	  ;
	else if (sscanf(line,message(141),&cntsize)==1)
	  ;
	else if (sscanf(line,message(142),&ptrsize)==1)
	  ;
	else if (sscanf(line,message(143),&Ver,&ver,status,&rev)==4)
	  ;
	else if (!strcmp(line,coreEnd))
	  break;
	else
	  bmsg(message(144),line);
      } while (true);
    }
#if SPECIAL_MALLOC
  if (size==-1L)
    berror(message(787));
  for (; size>=MIN_MEMSIZE && (theMZone = newMZone(size))==(MZPtr)0L; size = (7L*size)/8L);
  if (size<MIN_MEMSIZE)
    berror(message(145));
#else
  if (size!=-1L)
    berror(message(788));
#endif
  if (replay_mode)
    {
      if (replay_size!=0L && size!=replay_size)
	bmsg(message(146),size,replay_size);
#if SPECIAL_MALLOC
      if (replay_begin!=0L && replay_begin!=(long)theMZone)
	bmsg(message(147),(long)theMZone,replay_begin);
      else
	{
#endif
	  if (intsize!=0 && intsize!=sizeof(INT))
	    bmsg(message(148),sizeof(INT),intsize);
	  if (cntsize!=0 && cntsize!=sizeof(INT))
	    bmsg(message(149),sizeof(COUNTER),cntsize);
	  if (ptrsize!=0 && ptrsize!=sizeof(char *))
	    bmsg(message(150),sizeof(char *),ptrsize);
#if SPECIAL_MALLOC
	}
#endif
      if (Ver!=0 && (Ver!=MAJOR_VERSION || ver!=MINOR_VERSION ||
		     rev!=REVISION ||
		     strcmp(status,codeStatus)))
	bmsg(message(151),Ver,ver,status,rev,
	     MAJOR_VERSION,MINOR_VERSION,
	     codeStatus,
	     REVISION);
    }
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfprintf(coreTraceFile,message(152),size);
#if SPECIAL_MALLOC
      hfprintf(coreTraceFile,message(153),(long)theMZone);
#endif
      hfprintf(coreTraceFile,message(154),sizeof(INT));
      hfprintf(coreTraceFile,message(155),sizeof(COUNTER));
      hfprintf(coreTraceFile,message(156),sizeof(char *));
      hfprintf(coreTraceFile,message(157),MAJOR_VERSION,MINOR_VERSION,
	       codeStatus,REVISION);
      hfflush(coreTraceFile);
    }
#endif
  lower_init();
}

PRIVATE void ml_done_clean(doend)
actions *doend;
{
  if (doend!=(actions *)0L)
    {
      ml_done_clean(doend->next);
      (*doend->exec)();
    }
}

PUBLIC void showstats(fd)
int fd;
{
  struct tms now;
  long utime,stime,rtime;
  EXTENDED cent = 100.0;
#ifdef SPACE_OK
  unsigned long space;
#endif
  char buf[512],*s;

  (void) times(&now);
  utime = now.tms_utime-start_tms.tms_utime;
  stime = now.tms_stime-start_tms.tms_stime;
  rtime = utime+stime;
  s = buf;
  strcpy(s,"GimML timing statistics:\n========================\n\n"); s += strlen(s);

  sprintf(s,"total running time: %fs (user: %fs)\n",
	  ((EXTENDED)rtime/((EXTENDED)HERTZ)),
	  ((EXTENDED)utime)/((EXTENDED)HERTZ));
  s += strlen(s);
  sprintf(s,"number of garbage collections (GC): %ld\n",gc_count);
  s += strlen(s);
  sprintf(s,"total GC time: %fs (user: %fs)\n",
	  ((EXTENDED)gc_rtime)/((EXTENDED)HERTZ),
	  ((EXTENDED)gc_utime)/((EXTENDED)HERTZ));
  s += strlen(s);
  sprintf(s,"GC overhead: ");
  s += strlen(s);
  if (rtime==0)
    sprintf(s,"*");
  else
    sprintf(s,"%.1f%%",cent*((EXTENDED)gc_rtime)/((EXTENDED)rtime));
  s += strlen(s);
  sprintf(s," (user: ");
  s += strlen(s);
  if (utime==0)
    sprintf(s,"*");
  else
    sprintf(s,"%.1f%%",cent*((EXTENDED)gc_utime)/((EXTENDED)utime));
  s += strlen(s);
  sprintf(s,")\n");
  s += strlen(s);
#ifdef SPACE_OK
#if SPECIAL_MALLOC
  space = theMZone->MZ_block.MB_size;
#else
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  space = ((char *)sbrk(0))-start_limit;
#endif
#endif
  sprintf(s,"total data space size = %.1f Kb\n",
	  ((EXTENDED)space)/1024.0);
  s += strlen(s);
#endif
  (void) write(fd,buf,s-buf);
}

PUBLIC NORETURN void ml_done(n)
IN int n;
{
#ifdef GETLOOKUPSTATS
  {
    int i;
    EXTERN INT lookupStats[];

    hfprintf(hstdout,"Statistics on levels for lookup():\n");
    for (i=0; i<500; i++)
      if (lookupStats[i]!=0)
	hfprintf(hstdout,"%d lookups at level %d\n",lookupStats[i],i);
    hfflush(hstdout);
  }
#endif
  ml_done_clean(dobeforeend);
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfputs("\n(*\n",coreTraceFile);
      hfflush(coreTraceFile);
      showstats(coreTraceFile->fildesc);
      hfputs("*)\n",coreTraceFile);
      if (initCommandMode!=1) /* don't print banner if in -cmd mode */
	{
	  hfprintf(hstdout,message(158),coreTraceName);
	  hfflush(hstdout);
	}
      hfclose(coreTraceFile);
    }
#endif
  if (gcTraceFd!=-1)
    {
      showstats(gcTraceFd);
      (void) close(gcTraceFd);
    }
  EndRandom();
  EndHIO();
#if SPECIAL_MALLOC
  if (theMZone!=(MZPtr)0L)
    {
      EndMZone(theMZone);
      theMZone = (MZPtr)0;
    }
#endif
  exit(n);
}
