/* Inlining functions for interpreted form (intermediate between parse tree and bytecode).
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "gram.h"
#include "compile.h"
#include "type.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "grmproto.h"
#include "evlproto.h"
#include "cmpproto.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rinline.h"
#ifdef HAS_PROTOTYPES
#include "sinline.h"
#endif

PRIVATE INT exprSizeVec(e,limit)
IN CONST ExprPtr e;
IN INT limit;
{
  COUNTER i,n,s;

  for (i=0, s=0, n=AVAL(e)->n; i<n; i++)
    {
      s += exprSize(AVAL(e)->expr[i],limit);
      if (s>limit)
	break;
    }
  return s;
}

MAPSWEEPFUNC(do_exprsize)
{
  struct exprsize *data;

  data = MAPDATA(struct exprsize *);
  data->size += 1+exprSize(x,data->limit);
  if (data->size>data->limit)
    return true;
  data->size += exprSize(y,data->limit);
  if (data->size>data->limit)
    return true;
  return false;
}

PRIVATE INT exprSizeMap(map,limit)
IN CONST ExprPtr map;
IN INT limit;
{
  struct exprsize data;

  data.size = 0;
  data.limit = limit;
  (void) MapSweep(map,do_exprsize,(VOIDPTR)&data);
  return data.size;
}

PRIVATE INT exprSizeBindings(bindlist,limit)
IN CONST ExprPtr bindlist;
IN INT limit;
{
  ExprPtr binding;
  INT s;

  s = 0;
  while (bindlist!=NIL)
    {
      binding = CAR(bindlist);
      switch (binding->kind)
	{
	case E_LET:
	  s += exprSizeBindings(CDR(LETBIND(binding)),limit);
	  if (s>limit)
	    return s;
	  s += exprSizeBindings(LETBODY(binding),limit);
	  break;
	case E_DEBUGDECL:
	  break;
	default:
	  s += exprSize(binding,limit);
	  break;
	}
      if (s>limit)
	break;
      bindlist = CDR(bindlist);
    }
  return s;
}

PRIVATE INT exprSizeDoms(domlist,limit)
IN CONST ExprPtr domlist;
IN INT limit;
{
  INT s;

  s = 0;
  while (domlist!=NIL)
    {
      switch (domlist->kind)
	{
	case E_PAT_RECORD_GNRL_CONS:
	case E_PAT_REVRECORD_GNRL_CONS:
	case E_PAT_RECORD_MAPADD:
	  s += exprSize(CAR(domlist),limit);
	  domlist = CDR(domlist);
	  break;
	case E_TUPLE_1:
	  s += exprSizeBindings(REFVAL(domlist),limit);
	  domlist = NIL;
	  break;
	}
      if (s>limit)
	break;
    }
  return s;
}

PUBLIC INT exprSize(code,limit)
IN CONST ExprPtr code;
IN INT limit;
{
  INT size;

  size = 0;
again:
  /*if (code==NIL)
    return;*/
  switch (code->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_STRING:
    case E_COMPLEX: case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_VARINTRO: case E_VARREF: case E_QUOTE:
    case E_PAT_ANY:
    case E_BYTECODE:
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
      return size+1;
    case E_CONS: case E_UNSHARED_CONS:
    case E_PAT_REVCONS: case E_PATCHOICE:
    case E_ARRAY_2: case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_REVTUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_RECORD_MAPADD:
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_MAPGET: case E_RMUX: case E_UNSHARED_RMUX:
    case E_TYPE_APPL:
    case E_PAT_AS: case E_PAT_MAPADD:
    case E_SET:
    case E_IF: case E_PROG: case E_AND: case E_OR:
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
    case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
    case E_OVERWRITE: case E_UNDERWRITE: case E_EQUAL: case E_NEQUAL:
    case E_MAPLET: case E_HANDLE:
      size += 1+exprSize(CAR(code),limit);
      if (size>limit)
	return size;
      code = CDR(code);
      goto again;
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
      size += 1+funSize(CAR(code),limit);
      if (size>limit)
	return size;
      code = CDR(code);
      goto again;
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
    case E_DATA:
      size++;
      code = CDR(code);
      goto again;
    case E_RMU1: case E_UNSHARED_RMU1:
      size += 1+exprSize(CAR(code),limit);
      if (size>limit)
	return size;
      code = CDR(CDR(code));
      goto again;
    case E_LET:
      size += 1+exprSizeBindings(CDR(LETBIND(code)),limit);
      if (size>limit)
	return size;
      code = LETBODY(code);
      goto again;
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
    case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
    case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP:
    case E_MAP_MAP: case E_MAP_UNDER_MAP: case E_ALL_MAP:
    case E_EXISTS_MAP: case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      size += 1+exprSizeDoms(CDR(OFDOM(code)),limit);
      if (size>limit)
	return size;
      code = OFFP(code);
      goto again;
    case E_PACK: case E_PAT_PACK:
      size++;
      code = PACKVAL(code);
      goto again;
    case E_DEBUGZONE:
      code = DZCODE(code);
      goto again;
    case E_DEBUGNAME:
      code = DNPAT(code);
      goto again;
    case E_MEMO_FN:
      size++;
      code = MEMOFUN(code);
      goto again;
    case E_REF: case E_GET:
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_RAISE: case E_EXCEPTION: case E_DELAY:
    case E_COUNTED: case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
    case E_PAT_CONS: case E_NOT:
    case E_DATACON:
      size++;
      code = REFVAL(code);
      goto again;
    case E_ARRAY_GNRL:
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
      size += 1+exprSizeVec(code,limit);
      return size;
    case E_PAT_CHECK:
      size++;
      code = PATCHECKVAL(code);
      goto again;
    case E_PROFILE:
      code = CAR(code);
      goto again;
    case E_MAP:
      size += exprSizeMap(code,limit);
      return size;
    default:
      kinderror(message(664),code->kind);
      return 0;
    }
}

PUBLIC INT funSize(e,limit)
IN CONST ExprPtr e;
IN INT limit;
{
again:
  switch (e->kind)
    {
    case E_PROFILE:
      e = CAR(e);
      goto again;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
      return 1+exprSize(COUNTEDVAL(e),limit);
    default:
      return exprSize(e,limit);
    }
}
