%{
/* GimML scanner.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*#include <stdio.h>  : included by flex scanner already */
#include <ctype.h>
#include <fcntl.h>
/*#include "portable.h" : included through the use of flexpatch */
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "gram.h"
#include "gimml.tab.h"
#include "input.h"
#include "expr.h"
#include "grmproto.h"
#include "memproto.h"
#include "errproto.h"
#include "rprint.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#include <unistd.h>
#endif
#ifdef HAS_READLINE
#include "linenoise.h"
#endif
#ifdef HAS_PROTOTYPES
#include "slex.h"
#endif

#define exit(n) ml_done(n)

#define YY_NO_UNPUT
/* We undefined the following, even though we have to include files (this is why
   scanning buffer is), and to read from strings. The reason is that we had to
   do everything by hand before flex offered these features. Now that flex has them,
   it would be difficult to change this behaviour, coded in the functions OpenFile()
   and CloseFile() below. Admittedly, the latter works only because we have intimate
   knowledge of the internals of the scanner flex generates, and may not work with
   other versions of flex.
*/
#define YY_NO_SCAN_BUFFER
#define YY_NO_SCAN_STRING
#define YY_NO_SCAN_BYTES

PUBLIC Boolean readlineEnabled = true;

#define YY_USER_INIT 
EXTERN hFILE *yyin,*yyout;
EXTERN char *yytext;
EXTERN YYLTYPE yylloc; /* location data for the lookahead */
EXTERN Boolean wantCloseFile;
PUBLIC COUNTER initCommandMode=0;
EXTERN Boolean initializing;

EXTERN ExprPtr boolType,specialMarker;

EXTERN COUNTER minDebugLevel;

#if CORE_TRACE
EXTERN hFILE *coreTraceFile;
#endif

#define BUMP bump();
#define BUMPCR bumpcr();

PRIVATE void bump()
{
  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  yylloc.last_column += yyleng;
}

PRIVATE void bumpcr()
{
  register char *s,*send;

  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  for (s=yytext, send=s+yyleng; s<send; s++)
    {
      yylloc.last_column++;
      if (*s=='\n')
	{
	  yylloc.last_line++;
	  yylloc.last_column = 0;
	}
    }
}

PUBLIC int yy_input_device=YY_INPUT_STDIN;
PUBLIC CHAR *yyfilename=NULL;
PUBLIC CHAR *yystring=NULL;
PUBLIC long yystringlen=0;
PUBLIC long yyoffset=0;
PRIVATE CHAR *yy_end_buffer;

#if DEBUGGING
PUBLIC ExprPtr currentFile;

PUBLIC void setCurrentFile()
{
  ExprPtr file;
  ENTER;

  file = (yyfilename==NULL)?NIL:Ksymbol(yyfilename);
  SETQ(currentFile,file);
  QUIT;
}
#else
#define setCurrentFile()
#endif

PUBLIC Boolean unit_recognized=false;
PUBLIC COUNTER star_recognized=0,
	union_recognized=0;

PUBLIC ExprPtr lexEnv, /* maps ids to precedences as E_INT */
	consEnv; /* sets of constructors */
#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC ExprPtr strEnv,  /* map from structure names to lexical structures
		    (arrays of [lexMap,consMap,strMap,sigMap])
		    or to lexical functors (couples (lexical signature,lexical structure))
		    lexical signatures are exactly the same as lexical structures */
	sigEnv,keepStrEnv,keepSigEnv;
#endif
PUBLIC ExprPtr keepLexEnv,keepConsEnv;
PUBLIC SymPtr lexId,lexPath;
struct lexOpenToks *lex_open_toks = NULL;
EXTERN SymPtr starSym,unionSym,concSym;
EXTERN SymPtr nilSym,consSym,refsym,trueSym,falseSym,
	bindSym,matchSym,emptySym,mapgetSym,
	parsweepSym,nomemSym,nonsenseSym,catchSym,
	stackSym,incompSym,noninvSym,nthSym,
	rangeSym,asciiSym,stringnthSym,
	subscriptSym,arithSym,ioSym,reSym,futureSym;

PRIVATE void skipComment P((void));

typedef struct remText {
  struct remText *next;
  int len;
  CHAR text[1];
}   remText;

PRIVATE remText *theRemText=(remText *)0L;

typedef struct fileStack {
  struct fileStack *prev;
  int input_device;
  int saveCondState,saveCondLevel;
  /*
  char *yy_c_buf_p;
  int yy_n_chars;
  */
  union {
    struct {
      hFILE *file;
      CHAR *name;
    } file;
    struct {
      long len;
      CHAR *text;
    } string;
  } what;
  int line,pos;
  long offset;
  remText *rem;
} fileStack;

PRIVATE fileStack *theFileStack=NULL;
void yyinit();

PRIVATE void openRemText1(text,len)
CHAR *text;
int len;
{
  remText *rem;
  
  if (len>0)
    {
      rem = (remText *)MALLOC(sizeof(remText)+len);
      rem->next = theRemText;
      rem->len = len;
      safememcpy(rem->text,text,(size_t)len);
      theRemText = rem;
    }
}

PRIVATE void openRemText()
{
  int len; /* really deeply depends on how flex works internally! */
  CHAR *after;

#ifdef OLD
  after = (CHAR *)(yytext+yyleng-1); /* points to the '"' */
  after[0] = ';';
  after[1] = yy_hold_char;
#else
  after = (CHAR *)(yytext+yyleng); /* points to the NUL */
  after[0] = yy_hold_char;
#endif
  len = yy_end_buffer - after;
  openRemText1(after,len);
}

PRIVATE void closeRemText()
{
  remText *rem;

  if (theRemText!=(remText *)0L)
    {
      rem = theRemText->next;
      MARKFREE(theRemText);
      theRemText = rem;
    }
}

PRIVATE void resetRemText()
{
  while (theRemText!=(remText *)0L)
    closeRemText();
}

PUBLIC Boolean PushFile()
{
  fileStack *next;
  
  openRemText();
  next = (fileStack *)MALLOC_FAIL(sizeof(fileStack));
  if (next==(fileStack *)0L)
    return false;
  next->prev = theFileStack;
  theFileStack = next;
  next->input_device = yy_input_device;
  next->line = yylloc.last_line;
  next->pos = yylloc.last_column;
  next->offset = yyoffset;
  /*
  *yy_c_buf_p = yy_hold_char;
  next->yy_c_buf_p = yy_c_buf_p;
  next->yy_n_chars = yy_n_chars;
  */
  yylloc.last_line = 0;
  yylloc.last_column = 0;
  yyinit();
  yyoffset = 0;
  switch (next->input_device)
    {
    case YY_INPUT_FILE:
    case YY_INPUT_STDIN:
      next->what.file.file = yyin;
      next->what.file.name = yyfilename;
      break;
    case YY_INPUT_STRING0:
    case YY_INPUT_STRING:
      next->what.string.text = yystring;
      next->what.string.len = yystringlen;
      break;
    }
  next->rem = theRemText;
  theRemText = (remText *)0L;
  return true;
}

PUBLIC Boolean OpenFile(name)
IN CONST CHAR *name;
{
  hFILE *new;
  CHAR *realname;
  
  new = xfopen((char *)name,"r",(char **)&realname);
  if (new==NULL)
    return false;
  if (!PushFile())
    {
      MARKFREE(realname);
      return false;
    }
  yy_input_device = YY_INPUT_FILE;
  yyin = YY_CURRENT_BUFFER->yy_input_file = new;
  yyfilename = realname;
  setCurrentFile();
  openRemText1((CHAR *)";",1); /* to end the declarations after 'use' */
  return true;
}

PUBLIC Boolean OpenString(str,len)
IN CONST CHAR *str;
IN COUNTER len;
{
  if (!PushFile())
    return false;
  yy_input_device = YY_INPUT_STRING;
  yystring = str;
  yystringlen = len;
  return true;
}

PUBLIC void CloseFile()
{
  fileStack *prev;

  resetRemText();
  if (theFileStack==NULL)
    switch (initCommandMode)
      {
      case 0: /* happens when doing gimml <file or ... | gimml: exit */
      case 1: /* command file (-cmd) option was given: exit after execution */
	ml_done(0);
        break;
      case 2: /* init file (-i) option was given: switch to stdin */
	yy_input_device = YY_INPUT_STDIN;
	initCommandMode = 0;
	yylloc.last_line = 0;
	yylloc.last_column = 0;
	yyoffset = 0;
	yyin = YY_CURRENT_BUFFER->yy_input_file = hstdin;
	yyfilename = NULL;
	setCurrentFile();
	initializing  = false;
	cutoffInitials(); /* now put away initial values */
	break;
      case 3: /* compilation option (-c) was given: compile file compile_file */
	initCommandMode = 4;
        ml_done (GimMLCompile ()?10:0);
        break;
      case 4: /* happens when -c option was given but end-of-file was reached during
		 compilation; then bison tries to synchronize on semicolons, and
		 each synchronization calls CloseFile().  Then set wantCloseFile back
		 to true; this will be noticed by yywrap(), which will then issue
		 an 'unexpected end of file' error message. */
	wantCloseFile = true;
	break;
      }
  else
    {
      /*
      yy_c_buf_p = theFileStack->yy_c_buf_p;
      yy_n_chars = theFileStack->yy_n_chars;
      */
      if (yy_input_device==YY_INPUT_FILE)
        {
	  hfclose(yyin);
	  if (yyfilename!=NULL)
	    MARKFREE(yyfilename);
	}
      yylloc.last_line = theFileStack->line;
      yylloc.last_column = theFileStack->pos;
      yyoffset = theFileStack->offset;
      switch (yy_input_device = theFileStack->input_device)
	{
	case YY_INPUT_FILE:
	case YY_INPUT_STDIN:
	  yyin = YY_CURRENT_BUFFER->yy_input_file = theFileStack->what.file.file;
	  yyfilename = theFileStack->what.file.name;
	  setCurrentFile();
	  break;
	case YY_INPUT_STRING0:
	case YY_INPUT_STRING:
	  yystring = theFileStack->what.string.text;
	  yystringlen = theFileStack->what.string.len;
	  yyfilename = NULL;
	  setCurrentFile();
	  break;
	}
      prev = theFileStack->prev;
      theRemText = theFileStack->rem;
      MARKFREE(theFileStack);
      theFileStack = prev;
    }
  yyinit();
  /* unput(';'); to end list of declarations at end of file,
      but does not work. This is taken care of by the
      *after = ';' line in openRemText() */
}

PUBLIC char *getMarkFiles()
{
  return (char *)theFileStack;
}

PUBLIC void ResetFiles(mark,flushRemText)
IN CONST char *mark;
IN Boolean flushRemText;
{
  fileStack *fs;

  for (fs=theFileStack; mark!=(char *)fs; fs=fs->prev)
    if (fs==NULL) /* mark not found */
      goto end;
  while (mark!=(char *)theFileStack && theFileStack!=NULL) /*yy_input_device!=YY_INPUT_STDIN*/
    CloseFile();
end:
  if (flushRemText)
    resetRemText();
  wantCloseFile = false;
}

#ifdef HAS_READLINE
struct readline_completion_data {
  CHAR *prefix;
  COUNTER prefix_len;
  CHAR *buf;
  size_t pos;
  linenoiseCompletions *lc;
};

PRIVATE Boolean do_readline_completion(id,val,data)
IN CONST ExprPtr id;
IN CONST ExprPtr val;
INOUT CHAR *data;
{
  struct readline_completion_data *d = (struct readline_completion_data *)data;
  CHAR *s = STEXT(id);
  COUNTER i, len = SLEN(id);
  CHAR *t = (CHAR *)d->buf;

  for (i=0; i<len; i++)
       if (i>=d->pos)
       {
         linenoiseAddCompletionLen(d->lc,(const char *)d->prefix,d->prefix_len,
                                   (const char *)s,len,
                                   (const char *)&d->buf[d->pos],
                                   strlen((const char *)&d->buf[d->pos]));
         break;
       }
       else if (s[i]!=t[i])
       break;
  return false;
}

IMPORT Boolean enumerateIds(IN CONST mapSweepFunc f,
			    INOUT CHAR *data);

PRIVATE int char_class(c)
IN int c;
{
  switch (c)
  {
  case '!': case '%': case '&': case '$': case '#': case '+': case '-':
  case '/': case ':': case '<': case '=': case '>': case '?': case '@':
  case '\\': case '~': case '^': case '|': case '*': case '`':
    return 1;
  case '_': case '\'':
    return 2;
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9':
    return 3;
  default:
  if ((c>='A' && c<='Z') || (c>='a' && c<='z'))
    return 2;
  return 0;
  }
}

PRIVATE struct lexOpenToks *remOpenToks(buf)
const char *buf;
{
  struct lexOpenToks *tok;
  size_t len, tlen;
  char c;

  len = 0;
  for (tok=lex_open_toks; tok!=NULL; tok=tok->next)
       {
       while (c = buf[len], c==' ' || c=='\t') len++;
       if (tok->end==NULL) continue;
       if (c=='\0') break;
       tlen = strlen(tok->end);
       if (memcmp(&buf[len],tok->end,tlen)!=0)
         return NULL;
       len += tlen;
       }
  return tok;
}

PUBLIC void readline_completion(buf,lc,pos)
const char *buf;
linenoiseCompletions *lc;
size_t pos;
{
  struct readline_completion_data *data;
  struct lexOpenToks *toks;

  if (buf[pos]=='\0')
  {
    toks = remOpenToks(buf);
    if (toks!=NULL) { linenoiseAddCompletionLen(lc,buf,pos,toks->end,strlen(toks->end),NULL,0); return; }
  }
  data = (struct readline_completion_data *)alloctmp_anchored(sizeof(struct readline_completion_data),(char *)&buf);
  data->prefix = (CHAR *)buf;
  data->pos = pos;
  while (pos>0 && char_class((int)buf[pos-1])>=2) pos--;
  data->prefix_len = pos;
  data->buf = (CHAR *)&buf[pos];
  data->pos -= pos;
  if (data->buf[0]!='\0')
  {
    data->lc = lc;
    (void) enumerateIds(do_readline_completion,(CHAR *)data);
  }
  freetmp(data);
}

struct readline_hints_data {
  const char *buf;
  ExprPtr val;
};

PRIVATE Boolean do_hints_completion(id,val,data)
IN CONST ExprPtr id;
IN CONST ExprPtr val;
INOUT CHAR *data;
{
  struct readline_hints_data *d = (struct readline_hints_data *)data;
  CHAR *s = STEXT(id);
  COUNTER i, len = SLEN(id);
  CHAR *t = (CHAR *)d->buf;

  for (i=0; i<len; i++)
     if (s[i]!=t[i])
        return false;
  if (t[i]=='\0')
    {
      d->val = val;
      return true;
    }
  return false;
}

IMPORT void TypePrint P((IN CONST ExprPtr type,
			 IN printFunc doprint,
			 INOUT char *printdata,
			 IN CONST ExprPtr state));

PRINTFUNC(do_printhints)
{
  struct printstring *data;
  COUNTER len2;

  data = PRINTDATA(struct printstring *);
  if (data->len<data->maxlen)
    {
      len2 = data->maxlen-data->len;
      if (len>len2)
	len = data->maxlen-data->len;
      safememcpy((char *)data->buf+data->len,(char *)text,(long)len);
      data->len += len;
    }
}

PRIVATE char *readline_hints(buf,color,bold)
const char *buf;
int *color;
int *bold;
{
#define READLINE_HINT_MAXLEN 128
  static CHAR type_buf[READLINE_HINT_MAXLEN+1];
  struct readline_hints_data *data;
  struct printstring *pdata;
  struct lexOpenToks *toks;
  size_t len;

  data = (struct readline_hints_data *)alloctmp_anchored(sizeof(struct readline_hints_data),(char *)&buf);
  pdata = (struct printstring *)alloctmp_anchored(sizeof(struct printstring), (char *)&buf);
  pdata->len = 0;
  pdata->maxlen = READLINE_HINT_MAXLEN;
  pdata->buf = (CHAR *)type_buf;
  toks = remOpenToks(buf);
  if (toks!=NULL)
    {
      do_printhints((CHAR *)" ",1,(char *)pdata);
      for (toks=lex_open_toks; toks!=NULL; toks=toks->next)
	 if (toks->end!=NULL)
           do_printhints((CHAR *)toks->end,strlen(toks->end),(char *)pdata);
      *color = 36;
      *bold = 0;
      freetmp(data);
      return (char *)type_buf;
    }
  else
    {
      len = strlen(buf);
      while (len>0 && char_class((int)buf[len-1])>=2) len--;
      data->buf = &buf[len];
      if (enumerateIds(do_hints_completion,(CHAR *)data))
      { /* print type into a string, then give it to hints */
        type_buf[0] = ':';
        pdata->len = 1;
        (void) TypePrint(data->val,do_printhints,(char *)pdata,NIL);
        type_buf[pdata->len] = '\0';
        *color = 35;
        *bold = 0;
        freetmp(data);
        return (char *)type_buf;
      }
  }
  freetmp(data);
  return NULL;
#undef READLINE_HINT_MAXLEN
}


EXTERN char *ml_prompt;
#define YY_READLINE(buf,result,max_size) \
  do { \
  if (!readlineEnabled) { hfputs(ml_prompt,hstdout); hfflush(stdout); \
                          result = hfgetline(buf,max_size,hstdin); \
                               break;}; \
  linenoiseSetCompletionCallback(readline_completion); \
  linenoiseSetHintsCallback(readline_hints); \
  linenoiseSetFreeHintsCallback(NULL); \
  linenoiseMaskModeDisable(); \
  linenoiseSetMultiLine(0); \
  result = linenoise_raw(ml_prompt,buf,max_size); \
  if (result) \
    { buf[result-1] = '\0'; linenoiseHistoryAdd(buf); buf[result-1] = '\n'; } \
  } while(0)
#else
#define YY_READLINE(buf,result,max_size) \
do { \
hfputs(ml_prompt,hstdout); hfflush(stdout); \
result = hfgetline(buf,max_size,hstdin); \
} while (0)
#endif

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if (theRemText!=(remText *)0L) \
	{ \
	if (max_size>=theRemText->len) \
	{ \
	result = theRemText->len; \
	safememcpy(buf,(char *)theRemText->text,(size_t)result); \
	closeRemText(); \
	} \
	else \
	{ \
	result = max_size; \
	safememcpy(buf,(char *)theRemText->text,(size_t)result); \
	theRemText->len -= result; \
	safememcpy((char *)theRemText->text,(char *)theRemText->text+result, \
	(size_t)theRemText->len); \
	} \
	} \
	else { \
	switch (yy_input_device) \
	{ \
	case YY_INPUT_FILE: \
	  if ( (result = read(yyin->fildesc, buf, max_size )) < 0 ) \
	     yy_error(yylloc.first_line,yylloc.first_column, \
		     yylloc.last_line,yylloc.last_column, \
		     message(242)); \
	  break; \
	case YY_INPUT_STDIN: \
          YY_READLINE(buf,result,max_size); \
          break; \
	case YY_INPUT_STRING0: \
	case YY_INPUT_STRING: \
	  result = yystringlen-yyoffset; \
	  if (result>max_size) \
		result = max_size; \
	  safememcpy(buf,(char *)(yystring+yyoffset),result); \
	  break; \
	} \
	yyoffset += result; \
	} \
	yy_end_buffer = (CHAR *)((buf)+(result));

#undef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) berror(msg)

/* INITIAL is not defined yet, but I know that this is just 0: */
#define MY_INITIAL 0
#undef yywrap
#define YY_SKIP_YYWRAP
int yywrap()
{
  if (YY_START!=MY_INITIAL)
    {
      BEGIN(MY_INITIAL);
      yy_error(yylloc.first_line,yylloc.first_column,
	       yylloc.last_line,yylloc.last_column,
	       message(567));
    }
  if (wantCloseFile)
    {
      wantCloseFile = false;
      yy_error(yylloc.first_line,yylloc.first_column,
	       yylloc.last_line,yylloc.last_column,
	       message(129));
    }
  wantCloseFile = true;
  openRemText1((CHAR *)";",1);
  return 0; /* continue */
}

PUBLIC void setLocalEnv()
{
  ExprPtr ref;

  ENTER;
  ref = CDR(CDR(lexEnv));
  setcarMapMerge(ref,CAR(lexEnv));
  ref = CDR(CDR(consEnv));
  setcarMapMerge(ref,CAR(consEnv));
#if MODULE_SYSTEM & STD_ML_MODULE
  ref = CDR(CDR(strEnv));
  setcarMapMerge(ref,CAR(strEnv));
  ref = CDR(CDR(sigEnv));
  setcarMapMerge(ref,CAR(sigEnv));
#endif
  QUIT;
}

#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC void keepEnv()
{
  SETQ(keepLexEnv,lexEnv);
  SETQ(keepConsEnv,consEnv);
  SETQ(keepStrEnv,strEnv);
  SETQ(keepSigEnv,sigEnv);
}

PUBLIC void restoreEnv()
{
  SETQ(lexEnv,keepLexEnv);
  SETQ(consEnv,keepConsEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  SETQ(strEnv,keepStrEnv);
  SETQ(sigEnv,keepSigEnv);
#endif
  SETQ(keepLexEnv,NIL);
  SETQ(keepConsEnv,NIL);
  SETQ(keepStrEnv,NIL);
  SETQ(keepSigEnv,NIL);
}
#endif

PUBLIC void setTopEnv()
{
  while (CDR(lexEnv)!=NIL)
    {
      POPQ(lexEnv);
    }
  while (CDR(consEnv)!=NIL)
    {
      POPQ(consEnv);
    }
#if MODULE_SYSTEM & STD_ML_MODULE
  while (CDR(strEnv)!=NIL)
    {
      POPQ(strEnv);
    }
  while (CDR(sigEnv)!=NIL)
    {
      POPQ(sigEnv);
    }
#endif
  SETQ(keepLexEnv,NIL);
  SETQ(keepConsEnv,NIL);
#if MODULE_SYSTEM & STD_ML_MODULE
  SETQ(keepStrEnv,NIL);
  SETQ(keepSigEnv,NIL);
#endif
}

PRIVATE ExprPtr copyEnv1(env)
IN CONST ExprPtr env;
{
  env = Nreverse(env);
  return Nreverse(env);
}

PUBLIC ExprPtr copyLexEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(lexEnv);
  KEEPONLY(e);
  return e;
}

PUBLIC ExprPtr copyConsEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(consEnv);
  KEEPONLY(e);
  return e;
}

#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC ExprPtr copyStrEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(strEnv);
  KEEPONLY(e);
  return e;
}

PUBLIC ExprPtr copySigEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(sigEnv);
  KEEPONLY(e);
  return e;
}
#endif

PRIVATE Boolean getConstructor(name)
IN CONST ExprPtr name;
{
  ExprPtr consE;
  Boolean found;
  
  for (found = false, consE=consEnv; consE!=NIL; consE=CDR(consE))
    {
      (void) MapGet(name,CAR(consE),&found);
      if (found)
	break;
    }
  return found;
}

PUBLIC void setConstructor(name)
IN CONST ExprPtr name;
{
  setcarMapPut(name,UNIT,consEnv);
}

PUBLIC int id_class(name)
IN CONST ExprPtr name;
{
  ExprPtr prec,lexE;
  Boolean found;
  Boolean cc;
  
  cc = getConstructor(name);
  for (lexE = lexEnv; lexE!=NIL; lexE=CDR(lexE))
    {
      prec = MapGet(name,CAR(lexE),&found);
      if (found)
	{
	  switch (IVAL(prec))
	    {
	    case 1:  return cc?t_left_associative_precedence_0_constructor:t_left_associative_precedence_0_identifier;
	    case -1: return cc?t_right_associative_precedence_0_constructor:t_right_associative_precedence_0_identifier;
	    case 2:  return cc?t_left_associative_precedence_1_constructor:t_left_associative_precedence_1_identifier;
	    case -2: return cc?t_right_associative_precedence_1_constructor:t_right_associative_precedence_1_identifier;
	    case 3:  return cc?t_left_associative_precedence_2_constructor:t_left_associative_precedence_2_identifier;
	    case -3: return cc?t_right_associative_precedence_2_constructor:t_right_associative_precedence_2_identifier;
	    case 4:  return cc?t_left_associative_precedence_3_constructor:t_left_associative_precedence_3_identifier;
	    case -4: return cc?t_right_associative_precedence_3_constructor:t_right_associative_precedence_3_identifier;
	    case 5:  return cc?t_left_associative_precedence_4_constructor:t_left_associative_precedence_4_identifier;
	    case -5: return cc?t_right_associative_precedence_4_constructor:t_right_associative_precedence_4_identifier;
	    case 6:  return cc?t_left_associative_precedence_5_constructor:t_left_associative_precedence_5_identifier;
	    case -6: return cc?t_right_associative_precedence_5_constructor:t_right_associative_precedence_5_identifier;
	    case 7:  return cc?t_left_associative_precedence_6_constructor:t_left_associative_precedence_6_identifier;
	    case -7: return cc?t_right_associative_precedence_6_constructor:t_right_associative_precedence_6_identifier;
	    case 8:  return cc?t_left_associative_precedence_7_constructor:t_left_associative_precedence_7_identifier;
	    case -8: return cc?t_right_associative_precedence_7_constructor:t_right_associative_precedence_7_identifier;
	    case 9:  return cc?t_left_associative_precedence_8_constructor:t_left_associative_precedence_8_identifier;
	    case -9: return cc?t_right_associative_precedence_8_constructor:t_right_associative_precedence_8_identifier;
	    case 10: return cc?t_left_associative_precedence_9_constructor:t_left_associative_precedence_9_identifier;
	    case -10:return cc?t_right_associative_precedence_9_constructor:t_right_associative_precedence_9_identifier;
	    }
	  break;
	}
    }
  return cc?t_constructor:t_identifier;
}

PRIVATE int longid_class(text,op)
IN CONST CHAR *text;
IN Boolean op;
{
  REGISTER CHAR *s,*s0,*t,c;
  SymPtr strid;
  ExprPtr aux,path;
  int code;
  ENTER;
  
  path = NIL;
  keepEnv();
  s0 = text;
  while (true)
    {
      s = s0;
      while (c = *s, c!=0 && c!=' ' && c!='\t' && c!='\n' && c!='\v' && c!='\f' && c!='.')
	s++;
      for (t=s; (c = *t); t++)
	if (c==0 || c=='.')
	  break;
      if (c==0)
	break;
      strid = Kstring((INT)(s-s0),s0);
      (void) KEEP(strid);
      MAKECONS(strid,path,aux);
      path = KEEP(aux);
      lexOpen(&yylloc,strid,false,strEnv);
      for (s0=t+1; (c = *s0); s0++)
	if (c!=' ' && c!='\t' && c!='\n' && c!='\v' && c!='\f')
	  break;
    }
  SETQ(lexPath,Nreverse(path));
  aux = Ksymbol(s0);
  SETQ(lexId,aux);
  code = op?(getConstructor(lexId)?t_constructor:t_identifier):id_class(lexId);
  restoreEnv();
  QUIT;
  return code;
}

PRIVATE CHAR *skipSpaces(text)
IN REGISTER CONST CHAR *text;
{
  REGISTER CHAR c;

  while (c = *text++, c==' ' || c=='\t' || c=='\n' || c=='\v' || c=='\f');
  return text-1;
}

#if IMPERATIVE_TYPES==STD_ML_TYVAR
PRIVATE void checkNoLevelType(s)
IN CONST CHAR *s;
{
  while (*s == '\'') s++;
  if (*s>='1' && *s<='9')
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	hfputs(yytext,coreTraceFile);
#endif
      yy_error(yylloc.first_line,yylloc.first_column,
	       yylloc.last_line,yylloc.last_column,
	       message(130));
    }
}
#endif

#if FLOAT_FORMAT==IEEE754
#define IEEE_LEX(n) n
#else
PRIVATE NORETURN int infError()
{
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfputs(yytext,coreTraceFile);
#endif
  yy_error(yylloc.first_line,yylloc.first_column,
	   yylloc.last_line,yylloc.last_column,
	   message(131));
}

#define IEEE_LEX(n) infError()
#endif

%}

SPACE [\ \t\f]
DIG [0-9]
INT (\~?{DIG}+)
FMT [\ \t\n\f]
LETTER [A-Za-z]
ALPHA [A-Za-z0-9\'\_]
SYMB [\!\%\&\$\#\+\-\/\:\<\=\>\?\@\\\~\'\^\|\*\`]
ALPHA_ID ({LETTER}{ALPHA}*)
SYMB_ID ({SYMB}+)
ID ({SYMB_ID}|{ALPHA_ID})
XID ({ID}|\_{ID})
PREFIX ({XID}{FMT}*\.{FMT}*)

%%

abstraction { BUMP return kw_abstraction; }
abstype { BUMP return kw_abstype; }
all	{ BUMP return kw_all; }
and { BUMP return kw_and; }
andalso	{ BUMP return kw_andalso; }
as { BUMP return kw_as; }
case { BUMP return kw_case; }
do	{ BUMP return kw_do; }
datatype { BUMP return kw_datatype; }
delay	{ BUMP return kw_delay; }
else { BUMP return kw_else; }
end { BUMP return kw_end; }
eqtype { BUMP return kw_eqtype; }
exception { BUMP return kw_exception; }
exists { BUMP return kw_exists; }
extern { BUMP return kw_extern; }
fn { BUMP return kw_fn; }
fun { BUMP return kw_fun; }
functor { BUMP return kw_functor; }
handle { BUMP return kw_handle; }
if { BUMP return kw_if; }
in { BUMP return kw_in; }
in{SPACE}list/{FMT}	{ BUMPCR return kw_in_list; }
in{SPACE}map/{FMT}	{ BUMPCR return kw_in_map; }
in{SPACE}set/{FMT}	{ BUMPCR return kw_in_set; }
include { BUMP return kw_include; }
infix	{ BUMP return kw_infix; }
infixr { BUMP return kw_infixr; }
iterate { BUMP return kw_iterate; }
let { BUMP return kw_let; }
local	{ BUMP return kw_local; }
memofn { BUMP return kw_memofn; }
memofun { BUMP return kw_memofun; }
nonfix	{ BUMP return kw_nonfix; }
of { BUMP return kw_of; }
open { BUMP return t_open; }
open\* { BUMP return s_openF; }
orelse { BUMP return kw_orelse; }
pack	{ BUMP return kw_pack; }
dimension	{ BUMP return kw_dimension; }
raise { BUMP return kw_raise; }
rec	{ BUMP return kw_rec; }
sharing { BUMP return kw_sharing; }
sig { BUMP return kw_sig; }
signature { BUMP return kw_signature; }
some { BUMP return kw_some; }
struct { BUMP return kw_struct; }
structure { BUMP return kw_structure; }
sub{SPACE}map/{FMT}	{ BUMPCR return kw_sub_map; }
such{SPACE}that/{FMT}  { BUMPCR return kw_such_that; }
then { BUMP return kw_then; }
type { BUMP return kw_type; }
scale { BUMP return kw_scale; }
val	{ BUMP return kw_val; }
while { BUMP return kw_while; }
with	{ BUMP return kw_with; }
withtype { BUMP return kw_withtype; }
use { BUMP return kw_use; }
U	{ BUMP if (union_recognized)
     return kw_U;
     SETQ(lexId,unionSym);
     SETQ(lexPath,NIL);
     return id_class(lexId); }
\(	{ BUMP return s_P; }
\)	{ BUMP return s_Q; }
\[	{ BUMP return s_B; }
\]	{ BUMP return s_C; }
\.\(	{ BUMP return s_DP; }
\<\{	{ BUMP return s_LY; }
\{	{ BUMP return s_Y; }
\}	{ BUMP return s_Z; }
\|\[	{ BUMP return s_IB; }
\]\|	{ BUMP return s_CI; }
\[\|	{ BUMP return s_BI; }
\|\]	{ BUMP return s_IC; }
\+\+\|\[	{ BUMPCR return s_AAIB; }
\,	{ BUMP return t_comma; }
\:	{ BUMP return s_T; }
\;+	{ BUMP return s_S; }
\.\.\.	{ BUMP return s_DDD; }
\_	{ BUMP return s_U; }
\|	{ BUMP return s_I; }
\|\|	{ BUMP return s_II; }
\=	{ BUMP return s_E; }
\.\:\=	{ BUMP return s_DTE; }
\=\>	{ BUMP return s_EG; }
\-\>	{ BUMP return s_MG; }
\-m\>	{ BUMP return s_MmG; }
\*		{ BUMP
		    if (star_recognized)
			return s_F;
			SETQ(lexId,starSym);
			SETQ(lexPath,NIL);
			return id_class(lexId); }
\`		{ BUMP if (star_recognized || unit_recognized)
			  return t_backtick;
			goto symbid; }
\^\~		{ BUMP if (star_recognized || unit_recognized)
			  { /* writing m^~2 instead of m^ ~2 is common,
				and should be detected. */
			    yyless(1); /* recognize only \^ */
			    return s_X;
			  }
			goto symbid; }
\^		{ BUMP if (star_recognized || unit_recognized)
					return s_X;
				SETQ(lexId,concSym);
				SETQ(lexPath,NIL);
				return id_class(lexId); }
\#	{ BUMP return s_H; }
\(\*\$[DPEV][+-]\*\)	{ BUMP return t_pragma; }
<INITIAL>\(\*	{
	BUMP
#if CORE_TRACE
	if (coreTraceFile!=NULL)
	  hfputs(yytext,coreTraceFile);
#endif
	skipComment();
#if CORE_TRACE
	*yytext = 0;
#endif
	}

{DIG}	{ BUMP return precedence_constant; }
{INT}	{ BUMP return integer_constant; }
\~?0[xX][0-9a-fA-F]+ { BUMP return hex_integer_constant; }
\~?0[oO][0-7]+ { BUMP return octal_integer_constant; }
\~?0[bB][01]+ { BUMP return binary_integer_constant; }
{INT}E{INT} { BUMP return real_constant; }
{INT}\.{DIG}+   { BUMP return real_constant; }
{INT}\.{DIG}+E{INT} { BUMP return real_constant; }
\+inf { BUMP return IEEE_LEX(s_Ainf); }
\~inf { BUMP return IEEE_LEX(s_Ninf); }
NAN\(0\)	{ BUMP return IEEE_LEX(s_nan0); }
NAN\(1\)	{ BUMP return IEEE_LEX(s_nan1); }
NAN\(2\)	{ BUMP return IEEE_LEX(s_nan2); }
NAN\(3\)	{ BUMP return IEEE_LEX(s_nan3); }
NAN\(4\)	{ BUMP return IEEE_LEX(s_nan4); }
NAN\(5\)	{ BUMP return IEEE_LEX(s_nan5); }
NAN\(6\)	{ BUMP return IEEE_LEX(s_nan6); }
NAN\(7\)	{ BUMP return IEEE_LEX(s_nan7); }
\~NAN\(1\)  { BUMP return IEEE_LEX(s_Nnan1); }
\~NAN\(2\)  { BUMP return IEEE_LEX(s_Nnan2); }
\~NAN\(3\)  { BUMP return IEEE_LEX(s_Nnan3); }
\~NAN\(4\)  { BUMP return IEEE_LEX(s_Nnan4); }
\~NAN\(5\)  { BUMP return IEEE_LEX(s_Nnan5); }
\~NAN\(6\)  { BUMP return IEEE_LEX(s_Nnan6); }
\~NAN\(7\)  { BUMP return IEEE_LEX(s_Nnan7); }
{INT}(\.{DIG}+)?(E{INT})?{SPACE}*\:{SPACE}*{INT}(\.{DIG}+)?(E{INT})?	{
	BUMP return complex_constant; }
\"([^\"\\]|(\\([\"\\ntvf]|(\^[\@A-Z\[\\\]\^\_])|({DIG}{DIG}{DIG})|({FMT}*\\))))*\"  {
	BUMPCR return string_constant; }
{PREFIX}*{XID}  { BUMPCR
 symbid:
	return longid_class((CHAR *)yytext,false); }
op{SPACE}+{PREFIX}*{XID}	{ BUMPCR
	return longid_class(skipSpaces((CHAR *)yytext+2),true); }
\'\'?(\_|[1-9]{DIG}*)?\#?{ALPHA_ID} { BUMP
#if IMPERATIVE_TYPES==STD_ML_TYVAR
	checkNoLevelType((CHAR *)yytext);
#endif
	{
	  ExprPtr aux;

	  aux = Ksymbol(yytext);
	  SETQ(lexId,aux);
	  SETQ(lexPath,NIL);
	}
	return t_type_variable; }
{SPACE}+	{ BUMP
#if CORE_TRACE
	if (coreTraceFile!=NULL)
	  hfputc(' ',coreTraceFile);
#endif
	}
<*>\n+	{ BUMPCR
#if CORE_TRACE
	if (coreTraceFile!=NULL)
	  hfputc('\n',coreTraceFile);
#endif
}

"# "[0-9]+" \"".+"\""(" "[1-4]){0,4}\n {
  /* cpp output line control, e.g.
     # 37 "gag" 1 3 4
     - 37 is the line number
     - "gag" is the file name
     - the next numbers, if present, are:
	        1: start of a new file
		2: returning to a file
		3: source is a system header file
		4: following text should be treated as an 'extern"C"' block
  */
  BUMPCR
  return t_cpp_line_control_directive;
}

.	{ CHAR buf[80];
	  sprintf((char *)buf,message(564),yytext[0]);
	  yy_error(yylloc.first_line,yylloc.first_column,
		   yylloc.last_line,yylloc.last_column,
		   (char *)buf);
	}

%%

PUBLIC void yyinit()
{
  YY_FLUSH_BUFFER; /* YY_INIT */
}

enum commentState { SKIPPING, AFTER_OPEN, AFTER_STAR, IN_STRING, STRING_ESCAPE };

PRIVATE void skipComment()
{
  int c;
  int level,state;

  level = 1;
  state = SKIPPING;
  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  while (c=input(), c!=-1)
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	{
	  hputc(c,coreTraceFile);
	}
#endif
      if (c=='\n')
	{
	  yylloc.last_line++;
	  yylloc.last_column = 0;
	}
      else
	yylloc.last_column++;
      switch (state)
	{
	case SKIPPING:
	  switch (c)
	    {
	    case '(':
	      state = AFTER_OPEN;
	      break;
	    case '*':
	      state = AFTER_STAR;
	      break;
	    case '"':
	      state = IN_STRING;
	      break;
	    default:
	      break;
	    }
	  break;
	case AFTER_OPEN:
	  switch (c)
	    {
	    case '"':
	      state = IN_STRING;
	      break;
	    case '*':
	      level++;
	      /*fallthrough*/
	    default:
	      state = SKIPPING;
	      break;
	    }
	  break;
	case AFTER_STAR:
	  switch (c)
	    {
	    case '*':
	      break;
	    case '"':
	      state = IN_STRING;
	      break;
	    case ')':
	      if (--level==0)
		return;
	      /*fallthrough*/
	    default:
	      state = SKIPPING;
	      break;
	    }
	  break;
	case IN_STRING:
	  switch (c)
	    {
	    case '"':
	      state = SKIPPING;
	      break;
	    case '\\':
	      state = STRING_ESCAPE;
	      break;
	    default:
	      break;
	    }
	  break;
	case STRING_ESCAPE:
	  state = IN_STRING;
	  break;
	}
    }
  yy_error(yylloc.first_line,yylloc.first_column,
	   yylloc.last_line,yylloc.last_column,
	   message(132));
}

#if CORE_TRACE
PRIVATE int coreState=0;

PUBLIC void resetTracingYyLex()
{
  coreState = 0;
}

PUBLIC int tracingYylex()
{
  int n;

  n = yylex();
  if (coreTraceFile!=NULL)
    {
      if (n==kw_use)
        coreState = 1;
      else if ((n==string_constant && coreState==1) || coreState==2)
        coreState = 0;
      else
        hfputs(yytext,coreTraceFile);
    }
  return n;
}
#endif
