/* List handling functions.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "eval.h"
#include "expr.h"
#include "mem.h"
#ifdef HAS_PROTOTYPES
#include "slists.h"
#endif

EXTERN EXTENDED zero,fone;

PUBLIC ExprPtr Reverse1(e,tail)
IN CONST ExprPtr e,tail;
/* This is used also to convert unshared lists to their shared reversal,
   so we must not deal with the special case length e=1, tail=NIL specially,
   for example. */
/* Note also that Reverse1() does not KEEP its result. */
{
  BEGINQUICKKEEP();
  while (e!=NIL)
    {
      tail = Kons(CAR(e),tail);
      e = CDR(e);
      QUICKKEEP(tail);
    }
  ENDQUICKKEEP();
  return tail;
}

PUBLIC INT ListLength(list)
IN REGISTER CONST ExprPtr list;
{
  REGISTER INT len;

  for (len = 0; list!=NIL; list=CDR(list), len++);
  return len;
}

PUBLIC ExprPtr Append1(list1,list2)
IN CONST REGISTER ExprPtr list1,list2;
{
#define BUF_SIZE 512
  struct buf { struct buf *next; ExprPtr expr[BUF_SIZE]; } *buf,*aux;
  /* we manage recursion explicitly so as not to overflow the C stack. */
  REGISTER ExprPtr *p,*top;

  buf = NULL;
  p = top = NULL;
  for (; list1!=NIL; list1=CDR(list1))
    {
      if (p==top)
	{
	  aux = buf;
	  buf = (struct buf *)MALLOC(sizeof(struct buf));
	  buf->next = aux;
	  p = buf->expr;
	  top = p+BUF_SIZE;
	}
      *p++ = CAR(list1);
    }
  top = buf->expr;
  while (buf!=NULL)
    {
      for (; p!=top; )
	{
	  list2 = Kons(*--p,list2);
	  QUICKKEEP(list2);
	}
      aux = buf->next;
      MARKFREE(buf);
      buf = aux;
      top = buf->expr;
      p = top+BUF_SIZE;
    }
  return list2;
#undef BUF_SIZE
}

PUBLIC ExprPtr Append0(list1,list2)
IN CONST ExprPtr list1,list2;
{
  if (list2==NIL)
    return list1;
  if (list1==NIL)
    return list2;
  BEGINQUICKKEEP();
  list1 = Append1(list1,list2);
  ENDQUICKKEEP();
  return list1;
}

PUBLIC ExprPtr Append(list1,list2)
IN CONST ExprPtr list1,list2;
{
  list1 = Append0(list1,list2);
  return KEEP(list1);
}

PUBLIC ExprPtr Nappend(list1,list2)
INOUT ExprPtr list1;
IN CONST ExprPtr list2;
{
  REGISTER ExprPtr lastl,ll;

  if (list1==NIL)
    return list2;
  if (list2==NIL)
    return list1;
  for (lastl = list1; (ll=CDR(lastl))!=NIL; lastl = ll)
    ;
  SETCDR(lastl,list2);
  return list1;
}

PUBLIC ExprPtr Nreverse0(list,tail)
INOUT ExprPtr list;
IN CONST ExprPtr tail;
{
  ExprPtr next,e;

  if (list==NIL)
    return tail;
  BEGINQUICKKEEP();
  do {
    e = CAR(list);
    MAKECONS(e,tail,next);
    tail = next;
    QUICKKEEP(tail);
    list = CDR(list);
  } while (list!=NIL);
  ENDQUICKKEEP();
  return tail;
}

PUBLIC ExprPtr Nreverse1(list,tail)
INOUT ExprPtr list;
IN CONST ExprPtr tail;
{
  list = Nreverse0(list,tail);
  return KEEP(list);
}

PUBLIC ExprPtr Inds0(list)
IN CONST ExprPtr list;
{
  INT x;
  MapPtr inds;
  ExprPtr aux;
  ENTER;

  x = 0;
  inds = NewMap();
  while (list!=NIL)
    {
      aux = KINT(x);
      inds = SetAdd(aux,inds);
      KEEPONLY(inds);
      x++;
      list = CDR(list);
    }
  QUIT;
  return inds;
}

PUBLIC ExprPtr Elems0(list)
IN CONST ExprPtr list;
{
  MapPtr elems;
  ENTER;

  elems = NewMap();
  while (list!=NIL)
    {
      elems = SetAdd(CAR(list),elems);
      KEEPONLY(elems);
      list = CDR(list);
    }
  QUIT;
  return elems;
}

PUBLIC ExprPtr ListToMap0(list)
IN CONST ExprPtr list;
{
  INT x;
  MapPtr map;
  ExprPtr aux;
  ENTER;

  x = 0;
  map = NewMap();
  while (list!=NIL)
    {
      aux = KINT(x);
      map = MapPut(aux,CAR(list),map);
      KEEPONLY(map);
      x++;
      list = CDR(list);
    }
  QUIT;
  return map;
}
