/* GimML main().
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <fcntl.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "eval.h"
#include "actions.h"
#include "evlproto.h"
#include "errproto.h"
#include "grmproto.h"
#include "expproto.h"
#include "stkproto.h"
#include "prtproto.h"
#include "iniproto.h"
#include "dirproto.h"
#include "input.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <signal.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sml.h"
#endif
#ifdef HAS_READLINE
#include "linenoise.h"
#endif

EXTERN int asprintf(char **strp, const char *fmt, ...);

#define OPT_EXPANDED

EXTERN COUNTER crtab_maxlines;

#if CORE_TRACE
EXTERN char coreTraceName[];
EXTERN hFILE *coreTraceFile;
EXTERN Boolean replay_mode, want_core_trace;
#endif
EXTERN int gcTraceFd;

EXTERN struct Expression *features,*featuresType;
EXTERN char maintenance[],creator[];
EXTERN long replay_size;
EXTERN int yy_input_device;
EXTERN CHAR *yystring;
EXTERN long yystringlen;
EXTERN long yyoffset;
EXTERN COUNTER theRmargin;

EXTERN INT maxCells;
EXTERN unsigned long ktuple2_hash_size,kons_hash_size,
  mapletkons_hash_size,mapkons_hash_size,
  int_hash_size;
EXTERN unsigned long data_hash_size,real_hash_size;
EXTERN unsigned long string_hash_size,array_hash_size;
EXTERN INT theInlineLimit;

EXTERN COUNTER initCommandMode;

EXTERN EXTENDED fone;

EXTERN struct Expression *nihil;

#if DEBUGGING
EXTERN COUNTER debugMode;
EXTERN Boolean debugEnabled;
#endif

EXTERN Boolean readlineEnabled;

EXTERN actions *doafterinit;
EXTERN actions *dobeforeend;

PRIVATE char *prompt_format = NULL;
EXTERN CHAR *compile_file;

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

PUBLIC compiledAction __gimmlx_code [1];
PUBLIC COUNTER _n__gimmlx_code = 0;
PUBLIC char *ml_prompt = NULL;
#ifdef HAS_READLINE
#define HISTORY_FILE ".gimml_history"
PRIVATE void do_init_readline()
{
  linenoiseHistorySetMaxLen(256);
  linenoiseHistoryLoad(HISTORY_FILE);
}
actions readline_init = { NULL, do_init_readline };

PRIVATE void do_end_readline()
{
  linenoiseHistorySave(HISTORY_FILE);
}
actions readline_end = { NULL, do_end_readline };
#endif

PRIVATE void MLParse()
{
  if (yy_input_device==YY_INPUT_STDIN)
    {
      if (prompt_format!=NULL)
	{
	  CHAR buf[MAXPATHLEN+1];

	  errno = 0;
	  (void) getDir(buf,MAXPATHLEN);
	  if (errno)
	    goto normal;
	  {
	    char *s;

	    asprintf(&s, prompt_format, buf, message(91));
	    if (ml_prompt!=NULL)
	      free (ml_prompt);
	    ml_prompt = s;
	  }
	}
      else
	{
normal:
	  if (ml_prompt!=NULL)
	    free (ml_prompt);
	  ml_prompt = malloc (strlen(message(91))+1);
	  strcpy (ml_prompt, message(91));
	}
      hfflush(hstdout);
    }
  (void) GimMLParse();
}

PRIVATE void MLBanner()
{
#if (CODE_STATUS!=STATUS_RELEASE)
  hfprintf(hstdout,message(92),maintenance);
#ifdef PRINT_FULL_BANNER
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfprintf(hstdout,"Bug reports should include the replay file (%s)\n",coreTraceName);
#endif
  hfputs("Installed features:\n",hstdout);
  crtab_maxlines = MAX_SIGNED_INT;
  (void) valPretty(features,featuresType,do_printfile,
		   (char *)hstdout,0,11,theRmargin);
  hputc('\n',hstdout);
  hputc('\n',hstdout);
#endif
#endif
  hfprintf(hstdout,message(95),OS_STRING,creator);
  hfflush(hstdout);
}

PUBLIC int theArgc,theArgi;
PUBLIC char **theArgv;

int thePathc = 0;
char **thePathv = (char **)0;

PRIVATE void ml_getopt P((int argc,char **argv));
PRIVATE void usage P((IN Boolean fatalp));

PRIVATE NORETURN void loop P((int argc, char *argv[])) _NORETURN;
PRIVATE NORETURN void loop(argc,argv)
int argc;
char *argv[];
{
  int i;

#if OS==UNIX_SYS_V || OS==UNIX_BSD
  signal(2,ml_break);
  signal(3,ml_done);
  signal(SIGTERM,ml_done);
#ifdef SIGFPE
  signal(SIGFPE,ml_fpe);
#endif
#endif
  if ((initCommandMode & 1)!=1) /* don't print banner if in -cmd mode or in -c mode*/
    MLBanner();
  for (i=thePathc; --i>=0; )
    add_path(thePathv[i],strlen(thePathv[i]));
  for (;;)
    MLParse();
}

int main(argc,argv)
int argc;
char *argv[];
{
  ml_getopt(argc,argv);
#ifdef HAS_READLINE
  doafterinit = &readline_init;
  dobeforeend = &readline_end;
#endif
  InitML();
  loop(argc,argv);
  return 0;
}

typedef struct Command {
  char *key;
  void (*do_) P((void));
  char *doc;
  long xmask; /* the command (number #j) cannot be used if the command #i
		    has been used before, for 1<<i in xmask;
		    in particular, if j=i, the command may be reused
		    only if 1<<i is not in xmask */
} Command;

PRIVATE void do_initreplay()
{
  char *replay_file;

  if (theArgi>=theArgc)
    {
      fprintf(stderr,message(96));
      ml_done(10);
    }
  replay_file = theArgv[theArgi++];
  if (freopen(replay_file,"r",stdin)==NULL)
    /* this is used before Init, so we cannot use hfreopen on hstdin,
       but we can use the usual I/O library */
    {
      fprintf(stderr,message(97),replay_file);
      ml_done(10);
    }
  replay_mode = true;
}

PRIVATE void do_initmemsize()
{
  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&replay_size)!=1)
    {
      fprintf(stderr,message(98));
      ml_done(10);
    }
}

PRIVATE void do_initcommand()
{
  CHAR *s;

  if (theArgi>=theArgc
#ifndef OPT_EXPANDED
       || (s = (CHAR *)theArgv[theArgi], *s!='\"' || s[strlen((char *)s)-1]!='\"')
#endif
       )
    {
      fprintf(stderr,message(99));
      ml_done(10);
    }
#ifndef OPT_EXPANDED
  (void) getString(NULL,NULL,s,false); /* on non-unix systems, the shell does not expand strings */
#else
  s = (CHAR *)theArgv[theArgi];
#endif
  initCommandMode = 1;
  yy_input_device = YY_INPUT_STRING0;
  yystring = s;
  yystringlen = (long)strlen((char *)s);
  yyoffset = 0L;
  theArgi++;
}

PRIVATE void do_initinit()
{
  CHAR *s;

  if (theArgi>=theArgc
#ifndef OPT_EXPANDED
       || (s = (CHAR *)theArgv[theArgi], *s!='\"' || s[strlen((char *)s)-1]!='\"')
#endif
       )
    {
      fprintf(stderr,message(100));
      ml_done(10);
    }
#ifndef OPT_EXPANDED
  (void) getString(NULL,NULL,s,false); /* on non-unix systems, the shell does not expand strings */
#else
  s = (CHAR *)theArgv[theArgi];
#endif
  initCommandMode = 2;
  yy_input_device = YY_INPUT_STRING0;
  yystring = s;
  yystringlen = (long)strlen((char *)s);
  yyoffset = 0L;
  theArgi++;
}

PRIVATE void do_initusage()
{
  usage(false);
  ml_done(0);
}

PRIVATE int npaths=10;

PRIVATE void do_addpath()
{
  if (theArgi>=theArgc)
    {
      fprintf(stderr,message(101));
      ml_done(10);
    }
  if (thePathv==(char **)0)
    {
      thePathv = (char **)malloc(npaths*sizeof(char *));
    }
  else if (thePathc>=npaths)
    {
      npaths *= 2;
      thePathv = (char **)realloc((char *)thePathv,npaths*sizeof(char *));
    }
  if (thePathv==(char **)0)
    {
      fprintf(stderr,message(102));
      ml_done(10);
    }
  thePathv[thePathc++] = theArgv[theArgi++];
  /* we cannot use add_path() directly because it uses Malloc(),
     but theMZone has not yet been initialized */
}

PRIVATE void do_rmargin()
{
  int n = theRmargin;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%d",&n)!=1)
    {
      fprintf(stderr,message(103));
      ml_done(10);
    }
  theRmargin = (COUNTER)n;
}

PRIVATE void do_growfactor()
{
  double grow;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%lf",&grow)!=1 || grow<=fone)
    {
      fprintf(stderr,message(104));
      ml_done(10);
    }
  installGrowFactor(&grow);
}

PRIVATE void do_maxcells()
{
  long i = maxCells;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1024)
    {
      fprintf(stderr,message(793));
      ml_done(10);
    }
  maxCells = i;
}

PRIVATE void do_gctrace()
{
  char *file;

  if (theArgi>=theArgc)
    {
      fprintf(stderr,message(613));
      ml_done(10);
    }
  file = theArgv[theArgi++];
  gcTraceFd = open(file,O_CREAT | O_TRUNC | O_WRONLY, 0666);
  if (gcTraceFd==-1)
    {
      fprintf(stderr,message(614),file);
      ml_done(10);
    }
}

PRIVATE void do_pair_hash_size()
{
  unsigned long i = ktuple2_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(655),(unsigned long)KTUPLE2_HASH_SIZE);
      ml_done(10);
    }
  ktuple2_hash_size = i;
  kons_hash_size = i;
  mapletkons_hash_size = i;
  mapkons_hash_size = i;
}

PRIVATE void do_data_hash_size()
{
  unsigned long i = data_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(673),(unsigned long)DATA_HASH_SIZE);
      ml_done(10);
    }
  data_hash_size = i;
}

PRIVATE void do_int_hash_size()
{
  unsigned long i = int_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(657),(unsigned long)INT_HASH_SIZE);
      ml_done(10);
    }
  int_hash_size = i;
}

PRIVATE void do_real_hash_size()
{
  unsigned long i = real_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(658),(unsigned long)REAL_HASH_SIZE);
      ml_done(10);
    }
  real_hash_size = i;
}

PRIVATE void do_string_hash_size()
{
  unsigned long i = string_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(659),(unsigned long)STRING_HASH_SIZE);
      ml_done(10);
    }
  string_hash_size = i;
}

PRIVATE void do_array_hash_size()
{
  unsigned long i = array_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(660),(unsigned long)ARRAY_HASH_SIZE);
      ml_done(10);
    }
  array_hash_size = i;
}

PRIVATE void do_pwd_prompt()
{
  if (theArgi>=theArgc)
    {
      fprintf(stderr,message(666),(unsigned long)ARRAY_HASH_SIZE);
      ml_done(10);
    }
  prompt_format = theArgv[theArgi++];
}

PRIVATE void do_core_trace()
{
#if CORE_TRACE
  want_core_trace = true;
#else
  fprintf(stderr,message(668));
  fflush(stderr);
#endif
}

PRIVATE void do_compile()
{
  if (theArgi>=theArgc)
    {
      fprintf (stderr, message(245));
      ml_done(10);
    }
  compile_file = (CHAR *)theArgv[theArgi++];
  initCommandMode = 3;
  yy_input_device = YY_INPUT_STRING0;
  yystring = (CHAR *)";";
  yystringlen = 1;
  yyoffset = 0L;
}

PRIVATE void do_inline_limit()
{
  unsigned long i = theInlineLimit;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<0)
    {
      fprintf(stderr,message(246),(unsigned long)theInlineLimit);
      ml_done(10);
    }
  theInlineLimit = i;
}

PRIVATE void do_nodebug()
{
  debugEnabled = false;
}

PRIVATE void do_noreadline()
{
  readlineEnabled = false;
}

PRIVATE Command command[] = {
/* 0 */ { "-replay", do_initreplay, " replay-file", 0xffffffffL, },
/* 1 */ { "-mem", do_initmemsize, " memory-size", 0x00000003L },
/* 2 */ { "-cmd", do_initcommand, " ML-command-string", 0x0020000dL },
/* 3 */ { "-init", do_initinit, " ML-init-string", 0x0020000dL },
/* 4 */ { "-path", do_addpath, " path", 0x00000001L },
/* 5 */	{ "-col", do_rmargin, " number-of-columns", 0x00000021L },
/* 6 */	{ "-grow", do_growfactor, " memory-grow-factor", 0x00000041L },
/* 7 */ { "-maxcells", do_maxcells, " max-cells", 0x00000080L },
/* 8 */ { "-h", do_initusage, "", 0x000000001L },
/* 9 */ { "-gctrace", do_gctrace, " file-name", 0x00000200L },
/*10 */ { "-pair-hash-size", do_pair_hash_size, " #entries", 0x00000400L },
/*11 */ { "-int-hash-size", do_int_hash_size, " #entries", 0x00000800L },
/*12 */ { "-real-hash-size", do_real_hash_size, " #entries", 0x00001000L },
/*13 */ { "-string-hash-size", do_string_hash_size, " #entries", 0x00002000L },
/*14 */ { "-array-hash-size", do_array_hash_size, " #entries", 0x00004000L },
/*15 */ { "-pwd-prompt", do_pwd_prompt, " format-string", 0x00008000L, },
/*16 */ { "-core-trace", do_core_trace, "", 0x00010000L, },
/*17 */ { "-data-hash-size", do_data_hash_size, " #entries", 0x00020000L, },
/*18 */ { "-c", do_compile, " source-file-name", 0x00040007L, },
/*19 */ { "-inline-limit", do_inline_limit, " max-inlined-size", 0x00080001L, },
/*20 */ { "-nodebug", do_nodebug, "", 0x00100000L, },
/*21 */ { "-noreadline", do_noreadline, "", 0x00200000L, },
	{ NULL, NULL }
};

PRIVATE long requiredMask = 0L; /* the xmask of mandatory options */

PRIVATE char errbuf[1024];

PRIVATE void ml_getopt(argc,argv)
int argc;
char *argv[];
{
  Command *com;
  char *s;
  long xmask,mask,m;
  int i;
  Boolean start;

  theArgc = argc;
  theArgv = argv;
  replay_mode = false;
  xmask = 0L;
  for (theArgi = 1; theArgi<theArgc; )
    {
      if (!strcmp(theArgv[theArgi], "--"))
	{
	  theArgi++;
	  break;
	}
      for (com=command, mask=1L; com->key!=NULL; com++, mask <<= 1)
	if (!strcmp(theArgv[theArgi],com->key))
	  {
	    if ((m = (com->xmask & xmask)))
	      {
		start = true;
		if (xmask & mask)
		  sprintf(errbuf,message(105),theArgv[theArgi]);
		else
		  {
		    sprintf(errbuf,message(106),theArgv[theArgi]);
		    s = errbuf+strlen(errbuf);
		    for (i=0; m!=0L; i++)
		      {
			if (m & 1)
			  {
			    if (start)
			      start = false;
			    else if (m!=0L && (m & (-m))==m)
			      strcpy(s,message(107));
			    else
			    strcpy(s,message(108));
			    s += strlen(s);
			    sprintf(s,"%s",command[i].key);
			    s += strlen(s);
			  }
			m >>= 1;
		      }
		  }
		{
		  fprintf(stderr,errbuf);
		  fprintf(stderr,"\n");
		  ml_done(10);
		}
	      }
	    theArgi++;
	    xmask |= mask;
	    (*com->do_)();
	    break;
	  }
      if (com->key==NULL)
	{
	  fprintf(stderr,message(109),theArgv[theArgi]);
	  usage(false);
	  ml_done(10);
	}
    }
  if ((m = (requiredMask & (~xmask))))
    {
      start = true;
      
      sprintf(errbuf,
	      ((m & (-m))==m)?message(110):message(111));
      s = errbuf+strlen(errbuf);
      for (i=0; m!=0L; i++)
	if (m & 1)
	  {
	    m >>= 1;
	    if (start)
	      start = false;
	    else if (m!=0L && (m & (-m))==m)
	      strcpy(s,message(107));
	    else
	      strcpy(s,message(108));
	    s += strlen(s);
	    sprintf(s,"%s",command[i].key);
	    s += strlen(s);
	  }
      {
	fprintf(stderr,errbuf);
	ml_done(10);
      }
    }
}

PRIVATE void usage(fatalp)
IN Boolean fatalp;
{
  char *s;
  Command *com;
  long mask;

  sprintf(errbuf,message(112));
  s = errbuf+strlen(errbuf);
  for (com=command, mask = 1L; com->key!=NULL; com++, mask <<= 1)
    {
      sprintf(s,(requiredMask & mask)?" %s%s":" [%s%s]",com->key,com->doc);
      s += strlen(s);
    }
  sprintf (s, " [-- arguments...]");
  s += strlen (s);
  *s++ = '\n';
  *s = 0;
  fprintf(stderr,errbuf);
  if (fatalp)
    ml_done(10);
}

#ifdef SIGFPE
void ml_fpe(code)
int code;
{
  while (true)
    Raise(ARITHEXC); /* cannot get back result,
			so raise exception ad infinitum;
			this is not very smart, but what
			else can we do ? */
}
#endif
