/* Translator from bytecode files to C code.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
/*#include <sys/types.h> already included in timing.h*/
#include <sys/stat.h>
#include <sys/errno.h>
#include <unistd.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "eval.h"
#include "gram.h"
#include "gimml.tab.h"
#include "input.h"
#include "type.h"
#include "abssyn.h"
#include "expr.h"
#include "cxaux.h"
#include "transfer.h"
#include "evlproto.h"
#include "errproto.h"
#include "grmproto.h"
#include "expproto.h"
#include "stkproto.h"
#include "prtproto.h"
#include "iniproto.h"
#include "dirproto.h"
#include "cmpproto.h"
#include "typproto.h"
#include "input.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <signal.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sml.h"
#endif

#undef DEBUG

#define OPT_EXPANDED

struct bcstacks;
struct bcframes;

EXTERN COUNTER crtab_maxlines;

EXTERN int gcTraceFd;

EXTERN struct Expression *features,*featuresType;
EXTERN char maintenance[],creator[];
EXTERN long replay_size;
EXTERN COUNTER theRmargin;

EXTERN INT maxCells;
EXTERN unsigned long ktuple2_hash_size,kons_hash_size,
  mapletkons_hash_size,mapkons_hash_size,
  int_hash_size;
EXTERN unsigned long data_hash_size,real_hash_size;
EXTERN unsigned long string_hash_size,array_hash_size;

EXTERN EXTENDED fone;

EXTERN struct Expression *nihil;

#if DEBUGGING
EXTERN COUNTER debugMode;
EXTERN Boolean debugEnabled;
#endif


EXTERN MapPtr bcXlateMap;
EXTERN COUNTER nbcXlate;

PUBLIC int theArgc,theArgi,theArgi0=1;
PUBLIC char **theArgv;

PUBLIC CHAR *funNames[MAX_GLOBALS];

PRIVATE void ml_getopt P((int argc,char **argv));
PRIVATE void usage P((IN Boolean fatalp));

/* Dummy variables needed by the rest of the GimML system: */
PUBLIC Boolean evaluatingp=false,stopwanted=false;
PUBLIC CHAR *yyfilename;
PUBLIC char *yytext = "";
PUBLIC int yynerrs = 0;
PUBLIC long yyoffset=0;
PUBLIC int yy_input_device=0; /* different from YY_INPUT_STDIN so that symbol
				 table is not printed after the bytecode is loaded. */

PUBLIC ExprPtr currentFile;
PUBLIC void setCurrentFile() {}
PUBLIC COUNTER initCommandMode = 1;
PUBLIC ExprPtr lexEnv, /* maps ids to precedences as E_INT */
  consEnv; /* sets of constructors */
PUBLIC SymPtr lexId,lexPath;
PUBLIC ExprPtr keepLexEnv,keepConsEnv;
PUBLIC Boolean unit_recognized=false;
PUBLIC COUNTER star_recognized=0;
PUBLIC COUNTER union_recognized=0;
PUBLIC void resetTracingYyLex() {}
PUBLIC void setTopEnv() {}
PUBLIC void ResetFiles(mark,flushRemText)
IN CONST char *mark;
IN Boolean flushRemText;
{
}
PUBLIC void yyinit() {}
PUBLIC void CloseFile() {}
/* End of dummy variables. */

PUBLIC char *theCc = "cc";
PUBLIC char *theDefaultCflags = "-c -g -O";
PUBLIC char *theCflags = NULL;
PUBLIC char *theLd = "ld";
PUBLIC char *theDefaultLflags = "-L" LIB_DIR " -lm";
PUBLIC char *theLflags = NULL;

PUBLIC void setConstructor(name)
IN CONST ExprPtr name;
{
  setcarMapPut(name,UNIT,consEnv);
}

PRIVATE ExprPtr copyEnv1(env)
IN CONST ExprPtr env;
{
  env = Nreverse(env);
  return Nreverse(env);
}

PUBLIC ExprPtr copyLexEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(lexEnv);
  KEEPONLY(e);
  return e;
}

PUBLIC ExprPtr copyConsEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(consEnv);
  KEEPONLY(e);
  return e;
}

PUBLIC char *getMarkFiles()
{
  return NULL;
}

PUBLIC OPCODE DoBreak(bc, pc, top, bcframes, stacks, exc)
IN CONST struct byteCodeChunk *bc;
IN CONST OPCODE *pc;
IN CONST ExprPtr *top;
IN CONST struct bcframes *bcframes;
IN CONST struct bcstacks *stacks;
IN CONST ExprPtr exc;
{
  exit(2);
}

PUBLIC void restoreBreaks () {}
PUBLIC void reinstallBreaks () {}

PUBLIC compiledAction __gimmlx_code [1];
PUBLIC COUNTER _n__gimmlx_code = 0;

EXTERN ExprPtr theSyntax;

PRIVATE COUNTER indent = 0;
#define MAXINDENT 16

PRIVATE COUNTER nOut = 0;

PRIVATE void do_outtables (text, len, stream)
     IN CONST CHAR *text;
     IN COUNTER len;
     INOUT char *stream;
{
  COUNTER i;
  CHAR c;

  for (i=0; i<len; i++)
    {
      if (++indent>=MAXINDENT)
	{
	  indent = 0;
	  hfputs ("\n  ", (hFILE *)stream);
	}
      c = text[i];
      if (isprint (c))
	{
	  switch (c)
	    {
	    case '\'':
	      hfputs ("'\\'', ", (hFILE *)stream);
	      break;
	    case '\\':
	      hfputs ("'\\\\', ", (hFILE *)stream);
	      break;
	    default:
	      hfprintf ((hFILE *)stream, "'%c', ", c);
	      break;
	    }
	}
      else
	hfprintf ((hFILE *)stream, "%#x, ", c);
      nOut++;
    }
}

PUBLIC void outMain(out)
     IN CONST hFILE *out;
{
#include "mlstart.inc.c"
}

PRIVATE Boolean dash_E = false;

PUBLIC int yyparse P((void))
{
  TZone *zone;
  ExprPtr res,aux,l,lr, res_lr;
  hFILE *out;
  char fname[32], oname[32];
  ENTER;

  zone = (TZone *)alloctmp_anchored(sizeof(TZone),(char *)&res);
  zone->line1 = zone->line2 = zone->pos1 = zone->pos2 = 0;
  yyfilename = (CHAR *)theArgv[1];
  res = moduleLoad (zone, yyfilename, false);
  if (res==NIL)
    {
      int len = strlen (theArgv[1]);
      char *buf = alloctmp_anchored(len+64,(char *)&res);

      sprintf(buf,message(921),theArgv[1]);
      type_error(zone,(CHAR *)buf);
      return 1;
    }
  res = MakePair (E_CONS, res, NIL);
  KEEP(res);
  aux = ZBuild(zone,A_BASIS,res);
  KEEP(aux);
  l = MakePair (E_CONS, aux, NIL);
  KEEP(l);
  res = ZBuild(zone,A_DECL_LIST,l);
  KEEPONLY(res);

  if (dash_E)
    out = hstdout;
  else
    {
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#define FMT "/tmp/gimmlx%d.c"
#define OFMT "/tmp/gimmlx%d.o"
#define ID getpid()
#else
#define FMT "gimmlx%d.c"
#define OFMT "gimmlx%d.o"
#define ID (int)&out
  /* this is hopefully correct on old machines with a shared address
     space between processes, so that two processes should not have
     data at the same address (sort of). */
#endif
      sprintf ((char *)fname, FMT, ID);
      sprintf ((char *)oname, OFMT, ID);
      out = hfopen ((char *)fname, "w");
      if (out==NULL)
	{
	  hfprintf (hstderr, "gimmlx: cannot open temporary file %s. Stop.\n", fname);
	  hfflush (hstderr);
	  exit (2);
	}
#undef ID
#undef FMT
    }

  hfprintf (out, "/* This file created automatically by gimmlx.\n  Modify at our own risks.\n  While gimmlx is part of GimML and is licensed under the GPL,\n  this file carries no rights and no guarantees from the author of GimML, Jean Goubault-Larrecq,\n  or from Université Paris-Saclay, CNRS, ENS Paris-Saclay, or Laboratoire Méthodes Formelles\n  by virtue of it being produced by GimML.\n  It may however be subject to a specific licence\n  from the author of the source GimML file(s) used to generated the present file.\n  In any case:\n  1. Neither Jean Goubault-Larrecq nor Bull S.A, nor INRIA, nor CNRS, nor ENS Paris-Saclay\n    is responsible for the consequences of use of this file,\n    no matter how awful, even if they arise from defects in it.\n  2. The origin of this file must not be misrepresented,\n    either by explicit claim or by omission.\n  3. Altered versions of this file must be plainly marked as such,\n    and must not be misrepresented as being the original file.\n*/\n\n");
  hfprintf (out, "#include <stdlib.h>\n#include <stdio.h>\n#include <fcntl.h>\n");
  hfprintf (out, "#if OS==UNIX_SYS_V || OS==UNIX_BSD\n#include <sys/errno.h>\n#else\n#include <errno.h>\n#endif\n");
  hfprintf (out, "#if OS==UNIX_SYS_V || OS==UNIX_BSD\n#include <signal.h>\n#endif\n#ifdef HAS_STRING_H\n#include <string.h>\n#else\n#include <strings.h>\n#endif\n");

  hfprintf (out, "#include <assert.h>\n\n#include <gimml/gimmlx.h>\n\n");
  hfprintf (out, "#define STRINGIFY(x) #x\n#define GIMMLX_FILENAME STRINGIFY(%s)\n\n", (char *)yyfilename);
#ifndef DEBUG
  hfprintf (out, "#undef NDEBUG\n"); /* #define NDEBUG to disable assert(). */
#endif
  bcXoutProtos (out);
#ifdef DEBUG
  hfputs("void check_stack_frames () { gcStackView *sv, *next;\n  assert(theStackFrames==NULL || theStackFrames>(gcStackView *)&sv);\n  for (sv=theStackFrames; sv!=NULL; sv=next) { next = sv->next; assert (next==NULL || next>sv); } }\n", out);
#endif
  bcXout (out);

  lr = bcXdefs (out);
  (void) KEEP (lr);
  res_lr = MakePair (E_CONS, res, lr);
  (void) KEEP (res_lr);

  hfprintf (out, "PUBLIC compiledAction __gimmlx_code [] = {", nbcXlate);
  bcXoutCode (out);
  hfprintf (out, " NULL };\nPUBLIC COUNTER _n__gimmlx_code = %d;\n\n", nbcXlate);

  indent = 0;
  nOut = 0;
  hfputs ("\nPRIVATE char __gimmlx_rawdata [] = {\n  ", out);
  outTransfer (res_lr, do_outtables, (char *)out, false);
  /* This call to outTransfer() above is a big hack.
     It uses the special version of the transfer.c functions where XLATE
     is #defined, which contain patches to get and put back information on
     byteCodeChunks.
     In particular, it may be the case that outTransfer() marks data,
     but fails to unmark some (because it goes through E_BYTECODEs
     while marking, but does not go through them when outputting).
     This is OK here because we exit afterwards.
     But this would give strange results in general GimML situations.

     Oh, this is just a suspicion.  Running gimmlx under gdb reveals
     that no E_*_CLOSURE is ever traversed by the above call to outTransfer().
  */
  hfprintf (out, " 0 };\n#define N_GIMMLX_RAWDATA %d\n\n", nOut);

  outMain (out);
  hfflush (out);

  if (dash_E)
    ;
  else
    {
      CHAR *cmd;
      char *cflags = (theCflags==NULL)?theDefaultCflags:theCflags;
      char *lflags = (theLflags==NULL)?theDefaultLflags:theLflags;
      int len;
      int err;

      hfclose (out);
      len = strlen(theCc)+1;
      len += strlen(cflags)+1;
      len += strlen(fname)+1;
      len += strlen(oname)+80; /* + 5 should suffice */
      cmd = malloc(len);
      sprintf ((char *)cmd, "%s %s %s -o %s", theCc, cflags, fname, oname);
      err = system ((char *)cmd);
      free (cmd);
      unlink ((char *)fname);
      if (err)
	exit (2);
      len = strlen(theLd)+1;
      len += 8+strlen((char *)yyfilename); /* 8 : "-o ", ".opt " */
      len += strlen(oname)+1;
      len += strlen(cflags)+1;
      len += 80; /* 8 : "-lgimml " would suffice */
      cmd = malloc(len);
      sprintf ((char *)cmd, "%s -o %s.opt %s %s -lgimml", theLd, yyfilename, oname, lflags);
      err = system ((char *)cmd);
      free (cmd);
      unlink ((char *)oname);
      if (err)
	exit (2);
    }

  exit (0);
}

EXTERN SymPtr funSym;

PRIVATE void MLRun()
{
#if DEBUGGING
  currentFile = NIL;
#endif
  lexEnv = consEnv = NIL;
  pushEnv ();
  lexId = lexPath = NIL;
  keepLexEnv = keepConsEnv = NIL;
  (void) GimMLParse();
  /*NOTREACHED*/
}

PRIVATE NORETURN void loop P((int argc, char *argv[])) _NORETURN;

PRIVATE NORETURN void loop(argc,argv)
int argc;
char *argv[];
{
  int i;

  debugEnabled = false;
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  signal(2,ml_break);
  signal(3,ml_done);
  signal(SIGTERM,ml_done);
#ifdef SIGFPE
  signal(SIGFPE,ml_fpe);
#endif
#endif
  bcXlateMap = NewMap ();
  nbcXlate = 0;
  MLRun();
  ml_done(1);
}

int main(argc,argv)
int argc;
char *argv[];
{
  ml_getopt(argc,argv);
  {
    COUNTER i;

    for (i=0; i<MAX_GLOBALS; i++)
      funNames[i] = NULL;
  }
  InitML();
  loop(argc,argv);
  return 0;
}

typedef struct Command {
  char *key;
  void (*do_) P((void));
  char *doc;
  long xmask; /* the command (number #j) cannot be used if the command #i
		    has been used before, for 1<<i in xmask;
		    in particular, if j=i, the command may be reused
		    only if 1<<i is not in xmask */
}   Command;

PRIVATE void do_initmemsize()
{
  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&replay_size)!=1)
    {
      fprintf(stderr,message(98));
      ml_done(10);
    }
}

PRIVATE void do_initusage()
{
  usage(false);
  ml_done(0);
}

PRIVATE void do_rmargin()
{
  int n = theRmargin;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%d",&n)!=1)
    {
      fprintf(stderr,message(103));
      ml_done(10);
    }
  theRmargin = (COUNTER)n;
}

PRIVATE void do_growfactor()
{
  double grow;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%lf",&grow)!=1 || grow<=fone)
    {
      fprintf(stderr,message(104));
      ml_done(10);
    }
  installGrowFactor(&grow);
}

PRIVATE void do_maxcells()
{
  long i = maxCells;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1024)
    {
      fprintf(stderr,message(793));
      ml_done(10);
    }
  maxCells = i;
}

PRIVATE void do_gctrace()
{
  char *file;

  if (theArgi>=theArgc)
    {
      fprintf(stderr,message(613));
      ml_done(10);
    }
  file = theArgv[theArgi++];
  gcTraceFd = open(file,O_CREAT | O_TRUNC | O_WRONLY, 0666);
  if (gcTraceFd==-1)
    {
      fprintf(stderr,message(614),file);
      ml_done(10);
    }
}

PRIVATE void do_pair_hash_size()
{
  unsigned long i = ktuple2_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(655),(unsigned long)KTUPLE2_HASH_SIZE);
      ml_done(10);
    }
  ktuple2_hash_size = i;
  kons_hash_size = i;
  mapletkons_hash_size = i;
  mapkons_hash_size = i;
}

PRIVATE void do_data_hash_size()
{
  unsigned long i = data_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(673),(unsigned long)DATA_HASH_SIZE);
      ml_done(10);
    }
  data_hash_size = i;
}

PRIVATE void do_int_hash_size()
{
  unsigned long i = int_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(657),(unsigned long)INT_HASH_SIZE);
      ml_done(10);
    }
  int_hash_size = i;
}

PRIVATE void do_real_hash_size()
{
  unsigned long i = real_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(658),(unsigned long)REAL_HASH_SIZE);
      ml_done(10);
    }
  real_hash_size = i;
}

PRIVATE void do_string_hash_size()
{
  unsigned long i = string_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(659),(unsigned long)STRING_HASH_SIZE);
      ml_done(10);
    }
  string_hash_size = i;
}

PRIVATE void do_array_hash_size()
{
  unsigned long i = array_hash_size;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&i)!=1 || i<1)
    {
      fprintf(stderr,message(660),(unsigned long)ARRAY_HASH_SIZE);
      ml_done(10);
    }
  array_hash_size = i;
}

PRIVATE void do_E ()
{
  dash_E = true;
}

PRIVATE void do_cc ()
{
  if (theArgi>=theArgc)
    {
      fprintf (stderr, "no C compiler name provided after -cc.\n");
      ml_done(10);
    }
  theCc = theArgv[theArgi++];
}

PRIVATE void do_ld ()
{
  if (theArgi>=theArgc)
    {
      fprintf (stderr, "no linker name provided after -ld.\n");
      ml_done(10);
    }
  theLd = theArgv[theArgi++];
}

PRIVATE void do_cflags ()
{
  int len, newlen;
  
  if (theArgi>=theArgc)
    {
      fprintf (stderr, "missing argument to -cflags.\n");
      ml_done(10);
    }
  newlen = strlen(theArgv[theArgi]);
  if (theCflags==NULL)
    {
      theCflags = malloc(newlen+1);
      strcpy(theCflags,theArgv[theArgi]);
    }
  else
    {
      len = strlen(theCflags);
      theCflags = realloc(theCflags, len+newlen+2);
      theCflags[len] = ' ';
      strcpy(theCflags+len+1, theArgv[theArgi]);
    }
  theArgi++;
}

PRIVATE void do_lflags ()
{
  int len, newlen;
  
  if (theArgi>=theArgc)
    {
      fprintf (stderr, "missing argument to -lflags.\n");
      ml_done(10);
    }
  newlen = strlen(theArgv[theArgi]);
  if (theLflags==NULL)
    {
      theLflags = malloc(newlen+1);
      strcpy(theLflags,theArgv[theArgi]);
    }
  else
    {
      len = strlen(theLflags);
      theLflags = realloc(theLflags, len+newlen+2);
      theLflags[len] = ' ';
      strcpy(theLflags+len+1, theArgv[theArgi]);
    }
  theArgi++;
}


PRIVATE Command command[] = {
/* 0 */ { "-mem", do_initmemsize, " memory-size", 0x00000001L },
/* 1 */	{ "-col", do_rmargin, " number-of-columns", 0x00000002L },
/* 2 */	{ "-grow", do_growfactor, " memory-grow-factor", 0x00000004L },
/* 3 */ { "-maxcells", do_maxcells, " max-cells", 0x00000008L },
/* 4 */ { "-h", do_initusage, "", 0x00000000L },
/* 5 */ { "-gctrace", do_gctrace, " file-name", 0x00000020L },
/* 6 */ { "-pair-hash-size", do_pair_hash_size, " #entries", 0x00000040L },
/* 7 */ { "-int-hash-size", do_int_hash_size, " #entries", 0x00000080L },
/* 8 */ { "-real-hash-size", do_real_hash_size, " #entries", 0x00000100L },
/* 9 */ { "-string-hash-size", do_string_hash_size, " #entries", 0x00000200L },
/*10 */ { "-array-hash-size", do_array_hash_size, " #entries", 0x00000400L },
/*11 */ { "-data-hash-size", do_data_hash_size, " #entries", 0x00000800L, },
/*12 */ { "-cc", do_cc, " c-compiler-name", 0x00001000L, },
/*13 */ { "-ld", do_ld, " c-linker-name", 0x00002000L, },
/*14 */ { "-cflags", do_cflags, "cc-flags", 0x00000000L, },
/*15 */ { "-lflags", do_lflags, "ld-flags", 0x00000000L, },
/*16 */ { "-E", do_E, "", 0x00000000L },
	{ NULL, NULL }
};

PRIVATE long requiredMask = 0L; /* the xmask of mandatory options */

PRIVATE char errbuf[1024];

PRIVATE void ml_getopt(argc,argv)
int argc;
char *argv[];
{
  Command *com;
  char *s;
  long xmask,mask,m;
  int i;
  Boolean start;

  theArgc = argc;
  theArgv = argv;
  /*replay_mode = false;*/
  xmask = 0L;
  for (theArgi = 1; theArgi<theArgc; )
    {
      if (!strcmp(theArgv[theArgi], "--"))
	{
	  while (++theArgi < theArgc)
	    {
	      theArgv[theArgi0++] = theArgv[theArgi];
	    }
	  break;
	}
      for (com=command, mask=1L; com->key!=NULL; com++, mask <<= 1)
	if (!strcmp(theArgv[theArgi],com->key))
	  {
	    if ((m = (com->xmask & xmask)))
	      {
		start = true;
		if (xmask & mask)
		  sprintf(errbuf,message(105),theArgv[theArgi]);
		else
		  {
		    sprintf(errbuf,message(106),theArgv[theArgi]);
		    s = errbuf+strlen(errbuf);
		    for (i=0; m!=0L; i++)
		      {
			if (m & 1)
			  {
			    if (start)
			      start = false;
			    else if (m!=0L && (m & (-m))==m)
			      strcpy(s,message(107));
			    else
			    strcpy(s,message(108));
			    s += strlen(s);
			    sprintf(s,"%s",command[i].key);
			    s += strlen(s);
			  }
			m >>= 1;
		      }
		  }
		{
		  fprintf(stderr,errbuf);
		  fprintf(stderr,"\n");
		  ml_done(10);
		}
	      }
	    theArgi++;
	    xmask |= mask;
	    (*com->do_)();
	    break;
	  }
      if (com->key==NULL)
	{
	  theArgv[theArgi0++] = theArgv[theArgi++];
	}
    }
  if ((m = (requiredMask & (~xmask))))
    {
      start = true;
      
      sprintf(errbuf,
	      ((m & (-m))==m)?message(110):message(111));
      s = errbuf+strlen(errbuf);
      for (i=0; m!=0L; i++)
	if (m & 1)
	  {
	    m >>= 1;
	    if (start)
	      start = false;
	    else if (m!=0L && (m & (-m))==m)
	      strcpy(s,message(107));
	    else
	      strcpy(s,message(108));
	    s += strlen(s);
	    sprintf(s,"%s",command[i].key);
	    s += strlen(s);
	  }
      {
	fprintf(stderr,errbuf);
	ml_done(10);
      }
    }
  theArgc = theArgi0;
  if (theArgc<2)
    usage (true);
  theArgi = 2;
}

PRIVATE void usage(fatalp)
IN Boolean fatalp;
{
  char *s;
  Command *com;
  long mask;

  sprintf(errbuf,"usage: gimmlx");
  s = errbuf+strlen(errbuf);
  for (com=command, mask = 1L; com->key!=NULL; com++, mask <<= 1)
    {
      sprintf(s,(requiredMask & mask)?" %s%s":" [%s%s]",com->key,com->doc);
      s += strlen(s);
    }
  sprintf (s, " file [-- arguments...]");
  s += strlen (s);
  *s++ = '\n';
  *s = 0;
  fprintf(stderr,errbuf);
  if (fatalp)
    ml_done(10);
}

#ifdef SIGFPE
void ml_fpe(code)
int code;
{
  while (true)
    Raise(ARITHEXC); /* cannot get back result,
			so raise exception ad infinitum;
			this is not very smart, but what
			else can we do ? */
}
#endif
