/* Support functions for Standard ML modules (not used), plus a few utilities.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "eval.h"
#include "type.h"
#include "expr.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rmodules.h"
#ifdef HAS_PROTOTYPES
#include "smodules.h"
#endif

#undef DEBUG

EXTERN ExprPtr tdgAlt;

#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC ExprPtr makeFunctor(locals,in,out)
IN CONST MapPtr locals;
IN CONST ExprPtr in;
IN CONST ExprPtr out;
{
  VecPtr vp;
  ExprPtr ep;

  ENTER;
  CHECKHETEROFREE();
  VECALLOC(*(char **)&vp,VECSIZE(3));
  vp->n = 3;
  vp->expr[0] = locals;
  vp->expr[1] = in;
  vp->expr[2] = out;
  EALLOC_HETERO_NOGC(ep);
  ep->kind = (CHAR)E_ARRAY_GNRL;
  AVAL(ep) = vp;
  ASHAPE(ep) = NIL;
  KEEPONLY(ep);
  return ep;
}
#endif

/* a TDG is either true, false
   or KONS(cond,(l,h)), meaning (if cond then h else l),
   or KTUPLE2(cons,(l,h)), meaning not(if cond then h else l)
*/

EXTERN ExprPtr tdgAlt; /* = (false,true) */

PUBLIC Boolean tdgVarLess(x,y)
IN CONST ExprPtr x,y;
{ /* for module saving purposes, we need to have an order on tdg variables
     that is independent of, say, the system ordering <. */
  REGISTER ExprPtr xx,yy;

  xx = (x->kind==(CHAR)E_REF)?REFVAL(x):x;
  yy = (y->kind==(CHAR)E_REF)?REFVAL(y):y;
  switch (CompareStrings(xx,yy))
    {
    case -1:
      return true;
    case 0:
      return x->kind<y->kind;
    default:
      return false;
    }
}

PUBLIC ExprPtr tdgNot(t)
IN CONST ExprPtr t;
{
  if (t==trueValue)
    return falseValue;
  if (t==falseValue)
    return trueValue;
  if (t->kind==(CHAR)E_CONS)
    {
      t = Ktuple2(CAR(t),CDR(t));
      return KEEP(t);
    }
  return KONS(CAR(t),CDR(t));
}

#define tdgChs(sign,t) ((sign)?(t):tdgNot(t))

PRIVATE ExprPtr makeTdg(pos,x,l,h)
IN Boolean pos;
IN CONST ExprPtr x,l,h;
	/* normalize a TDG by ensuring the h part is positive */
{
  if (l==h)
    return tdgChs(pos,l);
  if (h==trueValue || h->kind==(CHAR)E_CONS)
    {
      l = KONS (l,h);
      if (pos)
	l = Kons (x, l);
      else l = Ktuple2 (x, l);
      /*l = Kpair((pos?E_CONS:E_TUPLE_2),x,l);*/
    }
  else
    {
      l = KONS(tdgNot(l),tdgNot(h));
      if (pos)
	l = Ktuple2 (x,l);
      else l = Kons (x,l);
      /*l = Kpair((pos?E_TUPLE_2:E_CONS),x,l);*/
    }
  return KEEP(l);
}

PUBLIC ExprPtr nameToTdg(x)
IN CONST ExprPtr x;
{
  return KONS(x,tdgAlt);
}

PRIVATE ExprPtr tdgOr1 P((IN CONST ExprPtr t1,IN CONST ExprPtr t2,
 			  IN Boolean sign1,IN Boolean sign2));

PRIVATE ExprPtr tdgOr1(t1,t2,sign1,sign2)
IN CONST ExprPtr t1,t2;
IN Boolean sign1,sign2;
/* if we write [sign]t the quantity computed by tdgChs(sign,t),
   that is, t is pos==true, and tdgNot(t) otherwise,
   then this function computes [sign1]t1 or [sign2]t2
   */
{
  ExprPtr x1,x2,v1,v2;
  ExprPtr *ep;
  COUNTER i;

  if (t1==t2)
    return (sign1==sign2)?tdgChs(sign1,t1):trueValue;
  if (t1==(sign1?trueValue:falseValue) ||
      t2==(sign2?falseValue:trueValue))
    return tdgChs(sign1,t1);
  if (t2==(sign2?trueValue:falseValue) ||
      t1==(sign1?falseValue:trueValue))
    return tdgChs(sign2,t2);
  if (t1>t2)
    {
      Boolean auxs;

      x1 = t1; t1 = t2; t2 = x1;
      auxs = sign1; sign1 = sign2; sign2 = auxs;
    }
  i = (((((unsigned long)t1)+((unsigned long)t2))%TDG_OR_HASH_SIZE) << 2)
	+(sign1<<1)+sign2;
  i += (i<<1); /* multiply i by 3 */
  ep = &AVAL(tdgOrHashTable)->expr[i];
  if (ep[0]==t1 && ep[1]==t2)
    return ep[2];
  x1 = CAR(t1);
  v1 = CDR(t1);
  x2 = CAR(t2);
  v2 = CDR(t2);
  if (x1==x2)
    {
      sign1 = ((t1->kind==(CHAR)E_CONS)==sign1);
      sign2 = ((t2->kind==(CHAR)E_CONS)==sign2);
      x1 = makeTdg(true,x1,tdgOr1(CAR(v1),CAR(v2),sign1,sign2),
			   tdgOr1(CDR(v1),CDR(v2),sign1,sign2));
    }
  else if (tdgVarLess(x1,x2))
    {
      sign1 = ((t1->kind==(CHAR)E_CONS)==sign1);
      x1 = makeTdg(true,x1,tdgOr1(CAR(v1),t2,sign1,sign2),
			   tdgOr1(CDR(v1),t2,sign1,sign2));
    }
  else
    {
      sign2 = ((t2->kind==(CHAR)E_CONS)==sign2);
      x1 = makeTdg(true,x2,tdgOr1(CAR(v2),t1,sign2,sign1),
			   tdgOr1(CDR(v2),t1,sign2,sign1));
    }
  ep[0] = t1;
  ep[1] = t2;
  ep[2] = x1;
  HITBARRIER(tdgOrHashTable);
  return x1;
}

PUBLIC ExprPtr tdgOr(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgOr1(t1,t2,true,true);
  KEEPONLY(t1);
  return t1;
}

PUBLIC ExprPtr tdgAnd(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgNot(tdgOr1(t1,t2,false,false));
  KEEPONLY(t1);
  return t1;
}

PUBLIC ExprPtr tdgImply(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgOr1(t1,t2,false,true);
  KEEPONLY(t1);
  return t1;
}

PRIVATE ExprPtr tdgEquiv1 P((IN CONST ExprPtr t1,IN CONST ExprPtr t2,
 			     IN Boolean eqv));

PRIVATE ExprPtr tdgEquiv1(t1,t2,eqv)
IN CONST ExprPtr t1,t2;
IN Boolean eqv;
	/* computes t1<=>t2 if eqv==true, and t1 xor t2 if eqv==false */
{
  ExprPtr x1,x2,v1,v2;
  /*Boolean pos;*/
  ExprPtr *ep;
  COUNTER i;

  if (t1==t2)
    return eqv?trueValue:falseValue;
  if (t1==trueValue)
    return tdgChs(eqv,t2);
  if (t2==trueValue)
    return tdgChs(eqv,t1);
  if (t1==falseValue)
    return tdgChs(!eqv,t2);
  if (t2==falseValue)
    return tdgChs(!eqv,t1);
  if (t1>t2)
    {
      x1 = t1; t1 = t2; t2 = x1;
    }
  i = (((((unsigned long)t1)+((unsigned long)t2))%TDG_EQUIV_HASH_SIZE)<<1)
	+eqv;
  i += (i<<1); /* multiply i by 3 */
  ep = &AVAL(tdgEquivHashTable)->expr[i];
  if (ep[0]==t1 && ep[1]==t2)
    return ep[2];
  x1 = CAR(t1);
  v1 = CDR(t1);
  x2 = CAR(t2);
  v2 = CDR(t2);
  /*pos = (t1->kind==(CHAR)E_CONS);*/
  if (x1==x2)
    {
      eqv = (eqv==(t1->kind==t2->kind));
      x1 = makeTdg(true,x1,tdgEquiv1(CAR(v1),CAR(v2),eqv),
			   tdgEquiv1(CDR(v1),CDR(v2),eqv));
    }
  else if (tdgVarLess(x1,x2))
    {
      eqv = (eqv==(t1->kind==(CHAR)E_CONS));
      x1 = makeTdg(true,x1,tdgEquiv1(CAR(v1),t2,eqv),
			   tdgEquiv1(CDR(v1),t2,eqv));
    }
  else
    {
      eqv = (eqv==(t2->kind==(CHAR)E_CONS));
      x1 = makeTdg(true,x2,tdgEquiv1(CAR(v2),t1,eqv),
			   tdgEquiv1(CDR(v2),t1,eqv));
    }
  ep[0] = t1;
  ep[1] = t2;
  ep[2] = x1;
  HITBARRIER(tdgEquivHashTable);
  return x1;
}

PUBLIC ExprPtr tdgEquiv(t1,t2)
IN CONST ExprPtr t1,t2;
{
  ENTER;

  t1 = tdgEquiv1(t1,t2,true);
  KEEPONLY(t1);
  return t1;
}

PUBLIC ExprPtr tdgInst(t,subst,defselfp)
IN CONST ExprPtr t;
IN CONST ExprPtr subst;
IN Boolean defselfp;
{ /* subst is a map from names to tdgs;
     by default, names are mapped to themselves
     if defselfp==true, otherwise to true */
  ExprPtr x,y,v,l,h;
  Boolean found;
  ENTER;

again:
  if (t==trueValue || t==falseValue)
    return t;
  v = CDR(t);
  x = CAR(t);
  y = MapGet(x,subst,&found);
  if (!found)
    {
      if (defselfp)
	{
	  l = tdgInst(CAR(v),subst,defselfp);
	  h = tdgInst(CDR(v),subst,defselfp);
	  if (l==CAR(v) && h==CDR(v))
	    ;
	  else
	    t = makeTdg((t->kind==(CHAR)E_CONS),x,l,h);
	}
      else
	{
	  t = CDR(v);
	  goto again;
	}
    }
  else if (y==trueValue)
    {
      t = CDR(v);
      goto again;
    }
  else if (y==falseValue)
    {
      t = CAR(v);
      goto again;
    }
  else
    {
      l = tdgInst(CAR(v),subst,defselfp);
      h = tdgInst(CDR(v),subst,defselfp);
      if (y->kind==(CHAR)E_CONS && CAR(y)==x && CDR(y)==tdgAlt)
	{ /* no substitution: x is unknown */
	  if (l==CAR(v) && h==CDR(v))
	    ;
	  else
	    t = makeTdg((t->kind==(CHAR)E_CONS),x,l,h);
	}
      else
	{
	  l = tdgAnd(tdgImply(y,h),tdgOr(y,l));
	  t = tdgChs((t->kind==(CHAR)E_CONS),l);
	}
    }
  KEEPONLY(t);
  return t;
}

PRIVATE Boolean isElemEquiv(t)
IN CONST ExprPtr t;
/* returns true if t is of the form [sign](y?not t':t') */
{
  ExprPtr v,l,h;
  
  if (t==trueValue || t==falseValue)
    return false;
  v = CDR(t);
  l = CAR(v);
  h = CDR(v);
  return /*(l==trueValue && h==falseValue) ||*/ /* cannot happen */
    (l==falseValue && h==trueValue) ||
      /*(l->kind==(CHAR)E_CONS && h->kind==(CHAR)E_TUPLE_2 &&
	CAR(l)==CAR(h) && CDR(l)==CDR(h)) ||*/ /* cannot happen */
      (l->kind==(CHAR)E_TUPLE_2 && h->kind==(CHAR)E_CONS &&
       CAR(l)==CAR(h) && CDR(l)==CDR(h));
}

PRIVATE Boolean isEquivVar(x,t0)
IN CONST ExprPtr x,t0;
{
  if (t0==trueValue || t0==falseValue)
    return false;
  if (CAR(t0)==x && isElemEquiv(t0))
    return true;
  return isEquivVar(x,CDR(CDR(t0))) && isEquivVar(x,CAR(CDR(t0)));
}

PRIVATE ExprPtr tdgGetEquivVar1(t,t0)
IN CONST ExprPtr t,t0;
{
  ExprPtr x;
  
  while (t!=trueValue && t!=falseValue)
    {
      x = CAR(t);
      if (x->kind==(CHAR)E_VAR && isElemEquiv(t) && isEquivVar(x,t0))
	return x;
    }
  return NIL;
}

PUBLIC ExprPtr tdgGetEquivVar(t)
IN CONST ExprPtr t;
/* returns the first propositional variable x that is a type name
   such that t is the formula x<=>t', or NIL if it does not exist.
   x is such that, when descending through t, we come to a node labelled x
   for all branches, and the successors of x are t'' and not t'' */
{
  return tdgGetEquivVar1(t,t);
}

PUBLIC ExprPtr funEqTdg(f,state)
IN CONST SymPtr f;
IN CONST ExprPtr state;
{
  ExprPtr t,tt;
  Boolean found;

  t = MapGet(f,S_NONEQS(state),&found);
  if (found)
    {
      ENTER;
      
      tt = tdgInst(t,S_NONEQS(state),false);
      if (t!=tt)
	{
	  if (tt==trueValue)
	    setMapRemove(f,&S_NONEQS(state));
	  else
	    setMapPut(f,tt,&S_NONEQS(state));
	  t = tt;
	}
      QUIT; /* t is kept in the state */
    }
  else
    t = trueValue;
  return t;
}

PUBLIC ExprPtr funNotRefTdg(f,state)
IN CONST SymPtr f;
IN CONST ExprPtr state;
{
  ExprPtr var,t;
  ENTER;
  
  if (f==refSym || f==arraySym || f==tableSym)
    return falseValue;
  var = Kref(E_REF,f);
  t = funEqTdg(KEEP(var),state);
  QUIT; /* t is kept in the state or is true */
  return t;
}

	/****	auxiliary functions *****/

PUBLIC ExprPtr getTyHdr(tyvarseq,xx,state)
IN CONST ExprPtr tyvarseq;
IN CONST ExprPtr xx;
IN CONST ExprPtr state;
{
  MapPtr eqconds,eqcond,l,x,tyhdr;
  ENTER;
  
  eqconds = NewMap();
  for (l=tyvarseq; l!=NIL; l=CDR(l))
    {
      x = CAR(l);
      if (xx==NIL || x==xx)
	eqcond = eqTypeTdg(x,state,false);
      else
	eqcond = ((varFlags(x) & F_EQ)?trueValue:falseValue);
      eqconds = MapPut(x,eqcond,eqconds);
    }
  tyhdr = KONS(tyvarseq,eqconds);
  KEEPONLY(tyhdr);
  return tyhdr;
}

#if MODULE_SYSTEM & STD_ML_MODULE
PRIVATE Boolean do_checkcond(x,y,data)
IN CONST ExprPtr x,y;
INOUT struct checkcond *data;
{
  return unify1(CAR(x),CDR(x),data->assum,data->state,true);
}

PRIVATE ExprPtr tyfcnRefTdg(assum,tyfcn)
IN CONST ExprPtr assum;
IN CONST ExprPtr tyfcn;
{
  ExprPtr t;

  while (ISCONS(tyfcn)/* && CDR(tyfcn)!=NIL*/ && CDR(tyfcn)->kind==(CHAR)E_TYPE_APPL)
    {
      tyfcn = APPLFUN(CDR(tyfcn));
      tyfcn = fxSym(tyfcn);
    }
  if (tyfcn==refSym || tyfcn==arraySym || tyfcn==tableSym)
    return falseValue;
  if (tyfcn->kind!=(CHAR)E_VAR || rigidp(tyfcn,assum))
    return trueValue;
  t = Kref(E_REF,tyfcn);
  return KEEP(t);
}

PRIVATE MapPtr do_instnoneq(x,y,subst)
IN CONST ExprPtr x,y;
IN CONST ExprPtr subst;
{
  y = tdgInst(y,subst,true);
  return (y==trueValue)?NewMap():MapSingleton(x,y);
}

PRIVATE Boolean do_checkcons(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct checkcons *data;
{
  ExprPtr y2;
  MapPtr s1,s2;
  Boolean found;

  y2 = MapGetSure(x,data->cons2);
  s1 = data->subst1;
  s2 = data->subst2;
  return unify1(newTypeInst1(y,NIL,data->state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			     false,
#endif
			     &s1),
		newTypeInst1(y2,NIL,data->state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			     false,
#endif
			     &s2),
		data->assum,data->state,true);
}

PRIVATE Boolean occurs2p P((IN CONST ExprPtr typevar,IN CONST ExprPtr type,
			    IN CONST ExprPtr state));

PRIVATE Boolean do_occurs2p(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct occurs2p *data;
{
  return occurs2p(data->typevar,y,data->state);
}

PRIVATE Boolean do_occurs2p_num(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct occurs2p *data;
{
  return occurs2p(data->typevar,x,data->state);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE MapPtr do_occur2sp_fx P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT struct occursp *data));

PRIVATE Boolean occurs2pFx1(typevar,sigma,state)
IN CONST ExprPtr typevar;
IN CONST ExprPtr sigma;
IN CONST ExprPtr state;
{
  struct occursp *data = (struct occursp *)alloctmp(sizeof(struct occursp));
  ExprPtr val;

  sigma = fxRealSigma(sigma,state);
  val = CDR(sigma);
  data->typevar = typevar;
  data->state = state;
  return MapSweep(val,do_occur2sp_fx,data);
}

PRIVATE Boolean do_occurs2p_fx(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct occursp *data;
{
  Boolean (*occp) P((ExprPtr tyvar,ExprPtr type,ExprPtr state));

  occp = (x->kind==(CHAR)E_ARRAY_2)?occurs2pFx1:occurs2p;
  return (*occp)(data->typevar,x,data->state);
}

PRIVATE Boolean occurs2pFx(typevar,f,state)
IN CONST ExprPtr typevar;
IN CONST ExprPtr f;
IN CONST ExprPtr state;
{
  if (f->kind==(CHAR)E_TYPEFX)
    return occurs2pFx1(typevar,CDR(f),state);
  return false;
}
#endif

PRIVATE Boolean occurs2p(typevar,type,state)
IN CONST ExprPtr typevar,type;
IN CONST ExprPtr state;
{
  Boolean found;

again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return false;*/
  switch (type->kind)
    {
      case E_NIL:
	return false;
      case E_VAR:
	{
	  Boolean found;

	  type = nongenValue(type,state,&found);
	  if (found)
	    goto again;
	  return false;
	}
      case E_TYPE_APPL:
	if (typevar==fxSym(APPLFUN(type)))
	  return true;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	if (occurs2pFx(typevar,APPLFUN(type),state))
	  return true;
#endif
	for (type = APPLARG(type); type!=NIL; type=CDR(type))
	  if (occurs2p(typevar,CAR(type),state))
	    return true;
	return false;
      case E_CONS:
	if (occurs2p(typevar,CAR(type),state))
	  return true;
	type = CDR(type);
	goto again;
      case E_EMPTY:
	return false;
      case E_MAPLET:
	type = MAPLETY(type);
	goto again;
      case E_MAP:
	{
	  struct occurs2p *data = (struct occurs2p *)alloctmp(sizeof(struct occurs2p));

	  data->typevar = typevar;
	  data->state = state;
	  return MapSweep(type,do_occurs2p,data);
	}
      case E_NUMTYPE:
	{
	  struct occurs2p *data = (struct occurs2p *)alloctmp(sizeof(struct occurs2p));

	  data->typevar = typevar;
	  data->state = state;
	  return MapSweep(REFVAL(type),do_occurs2p_num,data);
	}
      default:
	kinderror(message(78),type->kind);
	break;
    }
  return false;
}

PUBLIC Boolean setTyFcn(f,tyfcn,assum,state)
IN CONST ExprPtr f;
IN CONST ExprPtr tyfcn;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr aux,cond,t1,t2,ref1,ref2,subst,x,cons1,cons2;
  ExprPtr newstate;
  struct checkcond *data;
  struct checkcons *data2;
  Boolean found;
  ENTER;
  
  while (ISCONS(tyfcn)/* && CDR(tyfcn!=NIL)*/ && CDR(tyfcn)->kind==(CHAR)E_TYPE_APPL
	 && APPLARG(CDR(tyfcn))==CAR(CAR(tyfcn)))
    tyfcn = APPLFUN(CDR(tyfcn));
  if (f==tyfcn)
    return false;
  if (ISCONS(tyfcn) && occurs2p(f,CDR(tyfcn),state))
    {
error1:
      QUIT;
      return true;
      /* type_error(zone,"types do not match (recursive definition)"); */
    }
  cons1 = cons2 = NIL;
  cons1 = MapGet(f,S_TYCONS(state),&found);
  if (found) /* f is a datatype variable */
    {
      if (cons1->kind!=(CHAR)E_CONS)
	cons1 = NIL;
      else
	{
	  if (tyfcn->kind!=(CHAR)E_VAR)
	    {
	      goto error1; /* type_error(zone,"datatypes can only be matched by other datatypes"); */
	    }
	  if (rigidp(tyfcn,assum))
	    {
	      cons2 = MapGet(tyfcn,S_TYCONS(state),&found);
	      if (!found || cons2->kind!=(CHAR)E_CONS || Dom(CDR(cons1))!=Dom(CDR(cons2)))
		{
		  goto error1; /* type_error(zone,"datatypes not compatible (constructors do not match)"); */
		}
	      /* types of constructors will be matched at the end of the function,
		 to ensure that all other attributes are handled first */
	    }
	  else
	    {
	      aux = tyfcn;
	      tyfcn = f;
	      f = aux;
	    }
	}
    }
  newstate = DupTypeState(state);
  setNongenValue(f,tyfcn,newstate);
  cond = MapGet(f,S_CONDS(newstate),&found);
  if (found)
    {
      setMapRemove(f,&S_CONDS(newstate));
      data = (struct checkcond *)alloctmp_anchored(sizeof(struct checkcond), (char *)&f);
      data->assum = assum;
      data->state = newstate;
      if (MapSweep(cond,do_checkcond,data))
	goto error1;
    }
  ref1 = Kref(E_REF,f);
  (void) KEEP(ref1);
  ref2 = tyfcnRefTdg(assum,tyfcn);
  subst = MapSingleton(ref1,ref2);
  t1 = MapGet(f,S_NONEQS(newstate),&found);
  if (!found)
    t1 = trueValue;
  t2 = eqTypeTdg(CDR(tyfcn),newstate,false);
  t1 = tdgEquiv(t1,t2); /* must be true */
  if (t1!=trueValue)
    {
      x = tdgGetEquivVar(t1);
      if (x==NIL)
	{
	  aux = tdgAnd(oldeqconds,t1);
	}
      else
	{
	  subst = MapPut(x,tdgInst(t1,MapSingleton(x,trueValue),true),subst);
	  aux = oldeqconds;
	}
      aux = tdgInst(aux,subst,true);
      SETQ(S_EQCONDS(newstate),aux);
      if (aux==falseValue)
	{
	error3:
	  SETQ(S_EQCONDS(newstate),oldeqconds);
	  goto error1; /* type_error(zone,"types cannot both admit equality"); */
	}
    }
  oldnoneqs = KEEP(S_NONEQS(newstate));
  aux = MapCollect(oldnoneqs,do_instnoneq,subst);
  SETQ(S_NONEQS(newstate),aux);
  if (cons1!=NIL && cons2!=NIL)
    {
      data2 = (struct checkcons *)alloctmp_anchored(sizeof(struct checkcons), (char *)&f);
      data2->assum = assum;
      data2->state = newstate;
      data2->cons2 = CDR(cons2);
      getTyHdrSubsts(CAR(cons1),CAR(cons2),&data2->subst1,&data2->subst2,newstate);
      if (MapSweep(CDR(cons1),do_checkcons,data2))
	{
	  SETQ(S_NONEQS(newstate),oldnoneqs);
	  goto error3;
	}
    }
  overwriteTypeState(state,newstate);
  QUIT;
  return false;
}
#endif

PUBLIC char *nameMsg(msg,var)
IN CONST char *msg;
IN CONST SymPtr var;
{
#define MAXSIZE 128
  static char buf[MAXSIZE];
  char *s,*sname;
  long len,len2;
  
  strcpy(buf,msg);
  s = buf+strlen(buf);
  *s++ = ':';
  *s++ = ' ';
  len = (buf+MAXSIZE-2-s);
  len2 = symNameLen(var);
  for (sname=(char *)symName(var); len2--, *sname++!='%'; );
  if (len>len2)
    len = len2;
  strncpy(s,sname,len);
  s[len] = 0;
  s += strlen(s);
  /* *s++ = '.'; */
  *s = 0;
#undef MAXSIZE
  return buf;
}

#if MODULE_SYSTEM & STD_ML_MODULE
PRIVATE ExprPtr freshGeneric(cp,x)
INOUT COUNTER *cp;
IN CONST ExprPtr x;
{
  CHAR buf[16];
  ExprPtr xx;

  buf[0] = '\'';
  while (getNumber(++*cp,buf+1), xx = Ksymbol(buf), xx==x);
  return KEEP(xx);
}

PUBLIC Boolean recoverUnify2(type1,type2,assum,state)
IN CONST ExprPtr type1; /* type f(t1,...,tn) where f is flexible but not all tis are generic vars */
IN CONST ExprPtr type2;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr l,x,aux,eqs,f;
  COUNTER n1,n2,c;
  Boolean found;
  Boolean failp;
  ENTER;
  
  f = APPLFUN(type1);
  if (/*type2!=NIL && */type2->kind==(CHAR)E_VAR && genericp(type2,assum))
    {
      /* f(t1,...,x,...,tn)=x is always soluble, if x appears only once */
      for (l=APPLARG(type1), n1=0; l!=NIL; l=CDR(l), n1++)
	{
	  x = CAR(l);
	  if (x==type2)
	    {
	      for (n2=0; l=CDR(l), l!=NIL; n2++)
		{
		  x = CAR(l);
		  if (x==type2)
		    goto general;
		}
	      /* correct, so f must be Lambda x1,...,xn1,x,x'1,...,x'n2 . x */
	      l = NIL;
	      c = 0;
	      for (; n2!=0; n2--)
		l = KONS(freshGeneric(&c,x),l);
	      l = KONS(type2,l);
	      for (; n1!=0; n1--)
		l = KONS(freshGeneric(&c,x),l);
	      failp = setTyFcn(f,KONS(getTyHdr(l,x,state),type2),
			       assum,state);
	      QUIT;
	      return failp;
	    }
	}
    }
 general:
  eqs = MapGet(f,S_CONDS(state),&found);
  if (!found)
    eqs = NewMap();
  eqs = SetAdd(KONS(type1,type2),eqs);
  setMapPut(f,eqs,&S_CONDS(state));
  QUIT;
  return false;
}

PUBLIC void coerceTyFcn(zone,tyfcn1,tyfcn2,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr tyfcn1,tyfcn2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  MapPtr s1,s2;
  ExprPtr oldeqflags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  ExprPtr oldstrengths,oldmaxstrengths;
#endif
  ENTER;

  if (ISCONS(tyfcn1))
    {
      if (ISCONS(tyfcn2))
	{
	  if (ListLength(CAR(CAR(tyfcn1)))!=ListLength(CAR(CAR(tyfcn2))))
	    incompatibleTypes(zone,CDR(tyfcn1),CDR(tyfcn2),
			      (CHAR *)message(29),(CHAR *)message(30),
			      assum,state);
	}
      else
	{
	  tyfcn2 = KtypeAppl(tyfcn2,CAR(tyfcn1));
	  tyfcn2 = KONS(CAR(tyfcn1),KEEP(tyfcn2));
	}
    }
  else
    {
      if (ISCONS(tyfcn2))
	{
	  tyfcn1 = KtypeAppl(tyfcn1,CAR(tyfcn2));
	  tyfcn1 = KONS(CAR(tyfcn2),KEEP(tyfcn1));
	}
      else
	{ /* should not happen, because we don't know the arities; we assume
	     the arities are 0 */
	  tyfcn1 = KtypeAppl(tyfcn1,NIL);
	  tyfcn1 = KONS(KONS(NIL,NewMap()),KEEP(tyfcn1));
	  tyfcn2 = KtypeAppl(tyfcn2,NIL);
	  tyfcn2 = KONS(KONS(NIL,NewMap()),KEEP(tyfcn2));
	}
    }
  oldeqflags = KEEP(S_EQFLAGS(state));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  oldstrengths = KEEP(S_STRENGTH(state));
  oldmaxstrengths = KEEP(S_MAXSTRENGTH(state));
#endif
  getTyHdrSubsts(CAR(tyfcn1),CAR(tyfcn2),&s1,&s2,state);
  (void) coerceTypes(zone,newTypeInst1(CDR(tyfcn1),NIL,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				       false,
#endif
				       &s1),
		     newTypeInst1(CDR(tyfcn2),NIL,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				  false,
#endif
				  &s2),
		     assum,state,true);
  SETQ(S_EQFLAGS(state),oldeqflags);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  SETQ(S_STRENGTH(state),oldstrengths);
  SETQ(S_MAXSTRENGTH(state),oldmaxstrengths);
#endif
  QUIT;
}

PUBLIC void getLongStruct(zone,path,assump,state)
IN CONST TZone *zone;
IN CONST ExprPtr path;
INOUT ExprPtr *assump;
IN CONST ExprPtr state;
{
  ExprPtr namesig,name,str;
  Boolean found;
  
  for (; path!=NIL; path=CDR(path))
    {
      namesig = MapGet(CAR(path),A_STRUCTS(*assump),&found);
      if (!found)
	goto error;
      str = CDR(namesig); /* we don't care for the actual structure, but only
			     for its interface, that is, its signature */
      if (str->kind!=(CHAR)TYPE_ASSUM_KIND)
	goto error;
      *assump = str;
    }
  return;
 error:
  type_error(zone,varMsg(message(79),CAR(path)));
}

	/***** renamings of flexible names in signatures *****/

PRIVATE MapPtr do_instflex(x,y,state)
IN CONST ExprPtr x,y;
IN CONST ExprPtr state;
{
  ExprPtr m;
  ENTER;
  
  Boolean found;
  
  if (x->kind!=(CHAR)E_VAR
      || symName(x)[0]=='\'' || (nongenValue(x,state,&found), found))
    m = NewMap(); /* do not rename either first-order variables
		     or names bound to something */
  else
    m = MapSingleton(x,copyTyName(x,state));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_instids(x,y,state)
IN CONST ExprPtr x,y;
IN CONST ExprPtr state;
{
  ExprPtr m;
  ENTER;
  
  m = MapSingleton(x,newTypeInst(y,NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				 ,false
#endif
				 ));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_renrefname(x,y,data)
IN CONST ExprPtr x,y;
IN CONST char *data;
{
  ExprPtr m,tx,ty;
  ENTER;

  tx = Kref(E_REF,x);
  (void) KEEP(tx);
  ty = Kref(E_REF,y);
  (void) KEEP(ty);
  m = MapSingleton(tx,ty);
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_instconds1(x,y,state)
IN CONST ExprPtr x,y;
IN CONST ExprPtr state;
{
  ExprPtr m;
  ENTER;
  
  m = Singleton(KONS(newTypeInst(CAR(x),NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				 ,false
#endif
				 ),
		     newTypeInst(CDR(x),NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				 ,false
#endif
				 )));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_instconds(x,y,state)
IN CONST ExprPtr x,y;
IN CONST ExprPtr state;
{
  ExprPtr m;
  ENTER;
  
  m = MapSingleton(newTypeInst(x,NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			       ,false
#endif
			       ),
		   MapCollect(y,do_instconds1,state));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_insteqflags(x,y,subst)
IN CONST ExprPtr x,y;
IN CONST MapPtr subst;
{
  ExprPtr m;
  ENTER;
  
  m = MapSingleton(x,tdgInst(y,subst,false));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_instnoneqs(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct instsig *data;
{
  ExprPtr m;
  ENTER;

  if (/*x!=NIL && */x->kind==(CHAR)E_REF)
    {
      x = Kref(E_REF,newTypeInst(REFVAL(x),NIL,data->state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				,false
#endif
				));
      (void) KEEP(x);
    }
  else
    x = newTypeInst(x,NIL,data->state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			,false
#endif
			);
  m = MapSingleton(x,tdgInst(y,data->subst,false));
  KEEPONLY(m);
  return m;
}

PRIVATE ExprPtr instTyHdr(tyhdr,subst)
IN CONST ExprPtr tyhdr;
IN CONST MapPtr subst;
{
  ExprPtr r;
  ENTER;

  r = KONS(CAR(tyhdr),MapCollect(CDR(tyhdr),do_insteqflags,subst));
  KEEPONLY(r);
  return r;
}

PRIVATE MapPtr instCons(cons,state)
IN CONST MapPtr cons;
IN CONST ExprPtr state;
{
  ExprPtr m;
  ENTER;

  m = MapCollect(cons,do_instids,state);
  KEEPONLY(m);
  return m;
}

PRIVATE ExprPtr instAssum1 P((IN CONST ExprPtr assum,
			      IN CONST ExprPtr state0,
			      IN CONST MapPtr subst));
PRIVATE ExprPtr instFunctor1 P((IN CONST ExprPtr fun,
				IN CONST ExprPtr state0,
				IN CONST MapPtr subst));

PRIVATE MapPtr do_insttycons(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct instsig *data;
{
  ExprPtr tycons;
  ENTER;
  
  x = newTypeInst(x,NIL,data->state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		  ,false
#endif
		  );
  switch (y->kind)
    {
    case E_CONS: /* datatype */
      y = KONS(instTyHdr(CAR(y),data->subst),
	       instCons(CDR(y),data->state));
      break;
    case E_VAR: /* dimension */
      break;
    case E_TUPLE_2: /* scale */
      y = instScale(y,data->state);
      break;
    case TYPE_ASSUM_KIND: /* structure signatures */
      y = instAssum1(y,data->state,data->subst);
      break;
    case TYPE_FUNCTOR_KIND: /* functor signatures */
      y = instFunctor1(y,data->state,data->subst);
      break;
    default:
      kinderror(message(80),y->kind);
      break;
    }
  tycons = MapSingleton(x,y);
  KEEPONLY(tycons);
  return tycons;
}

PRIVATE MapPtr do_insttyfcn(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct instsig *data;
{
  ExprPtr m;
  ENTER;

  m = MapSingleton(x,KONS(instTyHdr(CAR(y),data->subst),
			  newTypeInst(CDR(y),NIL,data->state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				      ,false
#endif
				      )));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_inststruct(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct instsig *data;
{
  ExprPtr m;
  ENTER;
  
  m = MapSingleton(x,KONS(newTypeInst(CAR(y),NIL,data->state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				      ,false
#endif
				      ),
			  /* newTypeInst diverted to rename structure
			     or functor names */
			  instAssum1(CDR(y),data->state,data->subst)));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_instsig(x,y,data)
IN CONST ExprPtr x,y;
IN CONST struct instsig *data;
{
  ExprPtr m;
  ENTER;

  m = MapSingleton(x,instAssum1(y,data->state,data->subst));
  KEEPONLY(m);
  return m;
}

PRIVATE MapPtr do_instnongens(x,y,state)
IN CONST ExprPtr x,y;
IN CONST ExprPtr state;
{
  ExprPtr val;
  Boolean found;
  
  val = nongenValue(x,state,&found);
  if (found && /*val!=NIL && */
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      (val->kind==(CHAR)E_VAR || val->kind==(CHAR)E_ARRAY_2)
#else
      val->kind==(CHAR)E_VAR
#endif
      )
    x = val;
  return MapSingleton(x,y);
}

PRIVATE MapPtr do_instunit(x,y,state)
IN CONST ExprPtr x,y;
IN CONST ExprPtr state;
{
  ExprPtr m;
  ENTER;
  
  m = MapSingleton(x,KONS(instScale(CAR(y),state),
			  newTypeInst(CDR(y),NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				      ,false
#endif
				      )));
  KEEPONLY(m);
  return m;
}

PRIVATE ExprPtr instAssum1(assum,state0,subst)
IN CONST ExprPtr assum;
IN CONST ExprPtr state0;
IN CONST MapPtr subst;
{
  ExprPtr aux;
  MapPtr rename;
  struct instsig *data;
  ENTER;

  data = (struct instsig *)alloctmp(sizeof(struct instsig));
  assum = DupTypeAssum(assum);
  data->subst = subst;
  data->state = state0;
  aux = MapCollect(A_IDS(assum),do_instids,data->state);
  SETQ(A_IDS(assum),aux);
  aux = MapCollect(A_IDEXTS(assum),do_instids,data->state);
  SETQ(A_IDEXTS(assum),aux);
  aux = MapDomRestrTo(aux,A_CONS(assum));
  /*aux = MapCollect(A_CONS(assum),do_instids,data->state);*/
  SETQ(A_CONS(assum),aux);
  aux = MapCollect(A_TYFCNS(assum),do_insttyfcn,data);
  SETQ(A_TYFCNS(assum),aux);
  aux = MapCollect(A_TYEXTS(assum),do_insttyfcn,data);
  SETQ(A_TYEXTS(assum),aux);
  aux = MapCollect(A_STRUCTS(assum),do_inststruct,data);
  SETQ(A_STRUCTS(assum),aux);
  aux = MapCollect(A_SIGS(assum),do_instsig,data);
  SETQ(A_SIGS(assum),aux);
  aux = MapCollect(A_NONGENS(assum),do_instnongens,data->state);
  SETQ(A_NONGENS(assum),aux);
  aux = MapCollect(A_UNITS(assum),do_instunit,data->state);
  SETQ(A_UNITS(assum),aux);
  KEEPONLY(assum);
  return assum;
}

PRIVATE Boolean do_instlist(maplet,data)
IN CONST ExprPtr maplet;
IN CONST struct instlist *data;
{
  ExprPtr x,y,y2;
  ENTER;

  x = MAPLETX(maplet);
  y = MAPLETY(maplet);
  y2 = (y->kind==E_BIAS)?CAR(y):y;
  y2 = newTypeInst(y2,NIL,data->state0
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		   ,false
#endif
		   );
  if (y->kind==E_BIAS)
    {
      y2 = Kpair (E_BIAS, y2, CDR(y));
      KEEP (y2);
    }
  setNPut(x,y2,&data->new,NOEXPR);
  QUIT;
  return false;
}

PRIVATE ExprPtr instList(list,state0,subst)
IN CONST ExprPtr list;
IN CONST ExprPtr state0;
IN CONST MapPtr subst;
{
  struct instlist *data = (struct instlist *)alloctmp(sizeof(struct instlist));
  ENTER;

  data->state0 = state0;
  data->subst = subst;
  data->new = NIL;
  (void) NSweep(list,do_instlist,data);
  KEEPONLY(data->new);
  return data->new;
}

PRIVATE void instState1(state,state0,subst)
IN CONST ExprPtr state;
IN CONST ExprPtr state0;
IN CONST MapPtr subst;
{
  ExprPtr aux;
  MapPtr rename;
  struct instsig *data;
  ENTER;

  data = (struct instsig *)alloctmp(sizeof(struct instsig));
  data->subst = subst;
  data->state = state0;
  aux = MapCollect(S_CONDS(state),do_instconds,data->state);
  SETQ(S_CONDS(state),aux);
  aux = tdgInst(S_EQCONDS(state),data->subst,false);
  SETQ(S_EQCONDS(state),aux);
  aux = MapCollect(S_EQFLAGS(state),do_insteqflags,data->subst);
  SETQ(S_EQFLAGS(state),aux);
  aux = MapCollect(S_NONEQS(state),do_instnoneqs,data);
  SETQ(S_NONEQS(state),aux);
  aux = MapCollect(S_TYCONS(state),do_insttycons,data);
  SETQ(S_TYCONS(state),aux);
  aux = instList(S_PACKERS(state),state0,subst);
  SETQ(S_PACKERS(state),aux);
  aux = instList(S_ANNOTATIONS(state),state0,subst);
  SETQ(S_ANNOTATIONS(state),aux);
  QUIT;
}

PRIVATE MapPtr do_eliminatelocals(x,y,locals)
IN CONST ExprPtr x,y;
IN CONST MapPtr locals;
{
  ExprPtr z;
  Boolean found;

  z = (x->kind==(CHAR)E_REF)?REFVAL(x):x;
  (void) MapGet(z,locals,&found);
  return found?NewMap():MapSingleton(x,y);
}

PRIVATE ExprPtr instFunctor1(fun,state0,subst)
IN CONST ExprPtr fun;
IN CONST ExprPtr state0;
IN CONST MapPtr subst;
{
  MapPtr newstate,aux,locals;
  ENTER;
  
  locals = FUNCTOR_LOCALS(fun);
  /*!!! to change */
  oldtyvarsubs = KEEP(S_TYVARSUBS(state0));
  oldtyvarsubcache = KEEP(S_TYVARSUBCACHE(state0));
  aux = MapDomRestrBy(oldtyvarsubs,locals);
  SETQ(S_TYVARSUBS(state0),aux);
  aux = MapDomRestrBy(oldtyvarsubcache,locals);
  SETQ(S_TYVARSUBCACHE(state0),aux);
  subst = MapCollect(subst,do_eliminatelocals,locals);
  fun = MAKE_FUNCTOR(locals,FUNCTOR_IN(fun),
		     instAssum1(FUNCTOR_OUT(fun),state0,subst));
  SETQ(S_TYVARSUBS(state0),oldtyvarsubs);
  SETQ(S_TYVARSUBCACHE(state0),oldtyvarsubcache);
  KEEPONLY(fun);
  return fun;
}

PUBLIC ExprPtr instSig(sig,state)
IN CONST ExprPtr sig;
IN CONST ExprPtr state;
{
  ExprPtr aux;
  MapPtr rename,oldtyvarsubs,oldtyvarsubcache,subst;
  ENTER;
  
  rename = MapCollect(A_NONGENS(sig),do_instflex,state);
  subst = MapMerge(rename,MapCollect(rename,do_renrefname,(char *)0L));
  oldtyvarsubs = KEEP(S_TYVARSUBS(state));
  oldtyvarsubcache = KEEP(S_TYVARSUBCACHE(state));
  instState1(state,state,subst);
  aux = MapMerge(oldtyvarsubs,rename);
  SETQ(S_TYVARSUBS(state),aux);
  aux = MapMerge(oldtyvarsubcache,rename);
  SETQ(S_TYVARSUBCACHE(state),aux);
  sig = instAssum1(sig,state,subst);
  SETQ(S_TYVARSUBS(state),oldtyvarsubs);
  SETQ(S_TYVARSUBCACHE(state),oldtyvarsubcache);
  KEEPONLY(sig);
  return sig;
}

	/*****	sharing of structures *****/

PRIVATE Boolean do_getadmkind(x,y,kindp)
IN CONST ExprPtr x,y;
INOUT ExprKind *kindp;
{
  switch (y->kind)
    {
    case TYPE_ASSUM_KIND:
    case TYPE_FUNCTOR_KIND:
      if (*kindp==E_FREE)
	{
	  *kindp = (ExprKind)y->kind;
	  return false;
	}
      if (*kindp==(ExprKind)y->kind)
	return false;
      /* fallthrough */
    default:
      return true;
    }
}

PRIVATE ExprPtr admissify P((IN CONST MapPtr sigs,INOUT ExprPtr state,IN Boolean strictp));

PRIVATE ExprPtr admStructs(sigs,state,strictp)
IN CONST MapPtr sigs;
INOUT ExprPtr state;
IN Boolean strictp;
{
  /*!!!*/
}

PRIVATE ExprPtr admFunctors(functors,state,strictp)
IN CONST MapPtr functors;
INOUT ExprPtr state;
IN Boolean strictp;
{
  ExprPtr uniffun;
  MapPtr oldtyvarsubs,oldtyvarsubcache,aux,locals;
  ENTER;

  oldtyvarsubs = KEEP(S_TYVARSUBS(state));
  oldtyvarsubcache = KEEP(S_TYVARSUBCACHE(state));
  /*!!!*/
  SETQ(S_TYVARSUBS(state),oldtyvarsubs);
  SETQ(S_TYVARSUBCACHE(state),oldtyvarsubcache);
  KEEPONLY(uniffun);
  return uniffun;
}

PRIVATE ExprPtr admissify(sigs,state,strictp)
IN CONST MapPtr sigs;
INOUT ExprPtr state;
IN Boolean strictp;
{
  ExprPtr sig;
  ExprKind kind;
  ENTER;
  
  if (MapCardLeq(sigs,(INT)1L))
    sig = ChooseInDom(sigs);
  else
    {
      sig = NewTypeAssum();
      kind = E_FREE;
      if (MapSweep(sigs,do_getadmkind,&kind))
	sig = NIL;
      else switch (kind)
	{
	case TYPE_ASSUM_KIND:
	  sig = admStructs(sigs,state,strictp);
	  break;
	case TYPE_FUNCTOR_KIND:
	  sig = admFunctors(sigs,state,strictp);
	  break;
	}
    }
  KEEPONLY(sig);
  return sig;
}

PRIVATE MapPtr do_sharestore(x,y,unifsig)
IN CONST ExprPtr x,y;
IN CONST ExprPtr unifsig;
{
  return MapSingleton(x,unifsig);
}

PUBLIC Boolean shareStructs(signames,state)
IN CONST MapPtr signames;
INOUT ExprPtr state;
{
  MapPtr sigs,unifsig,aux;
  ENTER;
  
  sigs = Rng0(MapDomRestrTo(S_TYCONS(state),signames));
  (void) KEEP(size);
  unifsig = admissify(sigs,state,false);
  if (unifsig!=NIL)
    {
      aux = MapCollect(signames,do_sharestore,unifsig);
    }
  QUIT;
  return unifsig!=NIL;
}
#endif
