/* Set of macros for useful basic functions.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef PORTABLE_H
#define PORTABLE_H

#include "config.h"

#define EXTERN extern
#define IMPORT extern
#define EXPORT
#define STATIC static
#define PRIVATE static
#define PUBLIC

#define IN
#define OUT
#define INOUT
#define CONST
#define REGISTER register

#ifdef HAS_PROTOTYPES
# define P(list) list
#else
# define P(list) ()
#endif

#define _NORETURN
#define _PURE
#ifdef __GNUC__
#if __GNUC__ >= 2 /* should be >=2.5 */
#undef _NORETURN
#define HAS_NORETURN
#define _NORETURN __attribute__ ((noreturn))
#endif
#if __GNUC__ >= 2 /* should be >=2.96 */
#define HAS_PURE
#undef _PURE
#define _PURE __attribute__ ((pure))
#endif
#define HAS_MALLOC
#define _MALLOC __attribute__ ((malloc))
#endif

#define CHAR unsigned char
#define COUNTER int
#define INT long
#define SHORT short

#ifdef HAS_STDLIB
#include <stdlib.h>
#else
#ifndef malloc
IMPORT MALLOC_T malloc P((long size)) _MALLOC;
#endif
#ifndef realloc
IMPORT MALLOC_T realloc P((MALLOC_T p,long size));
#endif
#ifndef free
IMPORT FREE_T free P((MALLOC_T p));
#endif
#endif

#ifndef NULL
#define NULL 0L
#endif

 /* define the characters that end directory names (DIR_CHAR)
 and volume names (VOL_CHAR): */
#  define DIR_CHAR '/'
#  define VOL_CHAR '/'

typedef enum Boolean { false, true } Boolean;

#ifdef HAS_VOLATILE
#define VOLATILE volatile
#else
#define VOLATILE
#endif

/*
#ifdef __GNUC__
#define PURE const
#define NORETURN volatile
#else
#define PURE
#define NORETURN
#endif
*/
#define PURE
#define NORETURN

IMPORT VOIDPTR alloctmp P((long size));
IMPORT VOIDPTR alloctmp_anchored P((long size, char *anchor));
IMPORT void freetmp P((VOIDPTR b));

IMPORT void safememcpy P((INOUT VOIDPTR to, IN CONST VOIDPTR from,
			  IN INT size));
IMPORT char *newptr P((long size)) _MALLOC;
IMPORT char *newptr_gc P((long size)) _MALLOC;
IMPORT char *newptr_gc_fail P((long size)) _MALLOC;
IMPORT char *newptr_fail P((long size)) _MALLOC;

EXTERN INT currentHeapSize;

#define SAFEMALLOC(size) newptr((long)size)
#define MALLOC(size) newptr_gc((long)size)
#define MALLOC_FAIL(size) newptr_gc_fail((long)size)
#if SPECIAL_MALLOC
#define MARKFREE(addr) markFree(addr)
#define COMMITFREE() commitFree()
IMPORT void Mfree P((char *addr));
#define FREE(addr) Mfree(addr)
IMPORT char *Mrealloc P((char *ptr,unsigned long newsize));
#define REALLOC(addr,size) Mrealloc(addr,(unsigned long)size)
#else
#define MARKFREE(addr) free(addr)
#define COMMITFREE()
#define FREE(addr) free(addr)
#define REALLOC(addr,size) realloc(addr,size)
#endif

#define C_QNAN 0
#define C_SNAN 1
#define C_INFINITE 2
#define C_ZERONUM 3
#define C_NORMALNUM 4
#define C_DENORMALNUM 5

#define MAX_NAN 7

IMPORT int classifyD P((IN CONST REAL *x));
#if EXTP
IMPORT int classifyX P((IN CONST EXTENDED *x));
#else
#define classifyX(x) classifyD(x)
#endif
#if FLOAT_FORMAT==IEEE754
IMPORT int NaNcodeD P((IN CONST REAL *x)); /* *x must be a NaN */
IMPORT void makeNaND P((IN int i,OUT REAL *x));
#endif
IMPORT double floor P((double x));
EXTERN EXTENDED zero,fhalf;
#define ifloor(x) DEFAULT_IFLOOR(x)

#define G(x) ((unsigned short *)&(x))
#define Gg(x) ((unsigned char *)&(x))

 /* NEGR(x) is true if x<0.0
 POSR(x) is true if x>=0.0
 ZEROR(x) is true if x=0.0 or x=-0.0
 SMALLR(x) is true if x is zero or denormalized
 valid only if x is a REAL variable
 */
#if FLOAT_FORMAT==IEEE754
#if FLOAT_LAYOUT==4
# define NEGR(x) ((G(x)[3] & 0x8000)!=0)
# define POSR(x) ((G(x)[3] & 0x8000)==0)
# define HASHR(x) Gg(x)[6]
# define ZEROR(x) ((G(x)[3] & 0x7fff)==0 && G(x)[2]==0 && G(x)[1]==0 && G(x)[0]==0)
# define SMALLR(x) ((G(x)[3] & 0x7ff0)==0)
#else
# define NEGR(x) ((Gg(x)[0] & 0x80)!=0)
# define POSR(x) ((Gg(x)[0] & 0x80)==0)
# define HASHR(x) Gg(x)[1]
# if FLOAT_LAYOUT==1
#  define ZEROR(x) ((G(x)[0] & 0x7fff)==0 && G(x)[1]==0 && G(x)[2]==0 && G(x)[3]==0)
#  define SMALLR(x) ((G(x)[0] & 0x7ff0)==0)
#else
# if FLOAT_LAYOUT==2
#  define ZEROR(x) ((G(x)[0] & 0xff7f)==0 && G(x)[1]==0 && G(x)[2]==0 && G(x)[3]==0)
#  define SMALLR(x) ((G(x)[0] & 0xf07f)==0)
#else
# if FLOAT_LAYOUT==3
#  define ZEROR(x) ((Gg(x)[0] & 0x7f)==0 && Gg(x)[1]==0 && Gg(x)[2]==0 && Gg(x)[3]==0 && Gg(x)[4]==0 && Gg(x)[5]==0 && Gg(x)[6]==0 && Gg(x)[7]==0)
#  define SMALLR(x) ((Gg(x)[0] & 0x7f)==0 && (Gg(x)[1] & 0xf0)==0)
#else
#endif
#endif
#endif
#endif
#else
# define NEGR(x) ((x)<0.0)
# define POSR(x) ((x)>=0.0)
# define HASHR(x) Gg(x)[1]
# define ZEROR(x) ((x)==0.0 || (x)==-0.0)
# define SMALLR(x) ZEROR(x)
#endif

 /* NEGX(x) is true if x<0.0
 POSX(x) is true if x>=0.0
 ZEROX(x) is true if x=0.0 or x=-0.0
 SMALLX(x) is true if x is zero or denormalized
 valid only if x is an EXTENDED variable
 */
#define NEGX(x) NEGR(x)
#define POSX(x) POSR(x)

#if FLOAT_FORMAT==IEEE754
#if EXTP==0
# define ZEROX(x) ZEROR(x)
# define SMALLX(x) SMALLR(x)
#else
# if FLOAT_LAYOUT==1
#  define SMALLX(x) ((G(x)[0] & 0x7fff)==0)
# else
# if FLOAT_LAYOUT==2
#  define SMALLX(x) ((G(x)[0] & 0xff7f)==0)
# else
# if FLOAT_LAYOUT==3
#  define SMALLX(x) ((Gg(x)[0] & 0x7f)==0 && (Gg(x)[1] & 0xff)==0)
# endif
# endif
# endif
# if EXTP==1
#  if FLOAT_LAYOUT==1
#   define ZEROX(x) ((G(x)[0] & 0x7fff)==0 && G(x)[1]==0 && G(x)[2]==0 && G(x)[3]==0 && G(x)[4]==0)
#  else
#  if FLOAT_LAYOUT==2
#   define ZEROX(x) ((G(x)[0] & 0xff7f)==0 && G(x)[1]==0 && G(x)[2]==0 && G(x)[3]==0 && G(x)[4]==0)
#  else
#  if FLOAT_LAYOUT==3
#   define ZEROX(x) ((Gg(x)[0] & 0x7f)==0 && Gg(x)[1]==0 && Gg(x)[2]==0 && Gg(x)[3]==0 && Gg(x)[4]==0 && Gg(x)[5]==0 && Gg(x)[6]==0 && Gg(x)[7]==0 && Gg(x)[8]==0 && Gg(x)[9]==0)
#  else
#  if FLOAT_LAYOUT==4
#   define ZEROX(x) ((G(x)[4] & 0x7fff)==0 && G(x)[3]==0 && G(x)[2]==0 && G(x)[1]==0 && G(x)[0]==0)
#   define SMALLX(x) ((G(x)[4] & 0x7fff)==)
#  endif
#  endif
#  endif
#  endif
# else
# if EXTP==2
#  if FLOAT_LAYOUT==1
#   define ZEROX(x) ((G(x)[0] & 0x7fff)==0 && G(x)[1]==0 && G(x)[2]==0 && G(x)[3]==0 && G(x)[4]==0 && G(x)[5]==0)
#  else
#  if FLOAT_LAYOUT==2
#   define ZEROX(x) ((G(x)[0] & 0xff7f)==0 && G(x)[1]==0 && G(x)[2]==0 && G(x)[3]==0 && G(x)[4]==0 && G(x)[5]==0)
#  else
#  if FLOAT_LAYOUT==3
#   define ZEROX(x) ((Gg(x)[0] & 0x7f)==0 && Gg(x)[1]==0 && Gg(x)[2]==0 && Gg(x)[3]==0 && Gg(x)[4]==0 && Gg(x)[5]==0 && Gg(x)[6]==0 && Gg(x)[7]==0 && Gg(x)[8]==0 && Gg(x)[9]==0 && Gg(x)[10]==0 && Gg(x)[11]==0)
#  else
#  if FLOAT_LAYOUT==4
#   define ZEROX(x) ((G(x)[5] & 0x7fff)==0 && G(x)[4]==0 && G(x)[3]==0 && G(x)[2]==0 && G(x)[1]==0 && G(x)[0]==0)
#   define SMALLX(x) ((G(x)[5] & 0x7fff)==0)
#  endif
#  endif
#  endif
#  endif
# endif
# endif
#endif
#else
# define ZEROX(x) ((x)==0.0 || (x)==-0.0)
# define SMALLX(x) ZEROX(x)
#endif

#if OS==UNIX_SYS_V || OS==UNIX_BSD
# define CRASH() abort()
# define FREEZE()
#else
# define CRASH() ml_done(20);
# define FREEZE()
#endif

#ifdef offsetof
#define OFFSETOF(type,field) ((INT)offsetof(type,field))
#else
#define OFFSETOF(type,field) ((INT)(((char *)&((type *)0L)->field)-((char *)(type *)0L)))
#endif
#define VARSIZE(type,arrayname,n) OFFSETOF(type,arrayname[n])

IMPORT char *message P((int n));
IMPORT void bmsg P((char *msg,...));
IMPORT NORETURN void berror P((char *msg,...)) _NORETURN;
IMPORT NORETURN void bfatal P((char *msg,...)) _NORETURN;

#include "file.h"
#endif
