/* Printing values and types.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "print.h"
#include "type.h"
#include "expr.h"
#include "prtproto.h"
#include "errproto.h"
#include "cmpproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "gimml.tab.h"
#include "rprint.h"
#ifdef HAS_PROTOTYPES
#include "sprint.h"
#endif

PUBLIC COUNTER theRmargin = 80;
EXTERN ExprPtr rightMarginRef;

EXTERN SymPtr stringsym,stringSym,intSym,
  listsym,listSym,arraysym,arraySym,
  refsym,refSym,tablesym,tableSym,funSym,mapSym,
  boolsym,boolSym,exnsym,exnSym,
  promisesym,promiseSym,
  dynsym,dynSym;
#ifdef OLDDATA
EXTERN ExprPtr nilTags;
#endif
EXTERN ExprPtr maxLinesRef;
EXTERN ExprPtr basisAssum,basisState,basisTable,basisEnv;

PUBLIC COUNTER crtab_maxlines;
PUBLIC ExprPtr floatFormatRef;

EXTERN ExprPtr lexEnv;

EXTERN extDesc tableExtDesc;

#define lexToPri(n) (((n)<0)?(22+2*(n)):(21-2*(n)))
	/* 1 -> 19, -1 -> 20, 2 -> 17, -2 -> 18, etc. */

PUBLIC COUNTER getPri(con)
IN CONST ExprPtr con;
{
  COUNTER pri;
  ExprPtr prec,lexE;
  Boolean found;

#ifdef BRAIN_DAMAGED_FIXITY
  COUNTER i;
  ExprPtr *ep;
  VecPtr vp;

  /* first, search for precedences as E_COUNTED couples in theEnvironment
     (local precedence declarations)
     notice that E_COUNTED is not a valid tag for actual values,
     and that it is not E_LET, E_APPL or E_GOTO (debug tags);
     this is crucial. */
  for (lexE = theEnvironment; lexE!=NIL && lexE!=basisEnv; lexE=ENEXT(lexE))
    {
      vp = EVARS(lexE);
      for (i=vp->n, ep = &vp->expr[i]; --ep, --i>=0; )
	{
	  prec = *ep;
	  if (prec!=NOEXPR && prec->kind==(CHAR)E_COUNTED && COUNTEDVAL(prec)==con)
	    {
	      pri = COUNTEDN(prec);
	      return (COUNTER)lexToPri(pri);
	    }
	}
    }
#endif
  /* then, find them in the lexical environment */
  for (lexE = lexEnv; lexE!=NIL; lexE=CDR(lexE))
    {
      prec = MapGet(con,CAR(lexE),&found);
      if (found)
	{
	  pri = IVAL(prec);
	  return (COUNTER)lexToPri(pri);
	}
    }
  return 0;
}

PUBLIC COUNTER printBuf(buf,doprint,printdata)
IN CONST char *buf;
IN printFunc doprint;
INOUT char *printdata;
{
  COUNTER len;

  len = strlen(buf);
  (*doprint)((CHAR *)buf,len,printdata);
  return len;
}

PRIVATE COUNTER printQuote(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("\"",doprint,printdata);
}

PRIVATE COUNTER printOpenBracket(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("[",doprint,printdata);
}

PRIVATE COUNTER printCloseBracket(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("]",doprint,printdata);
}

PUBLIC COUNTER printOpenBrace(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("{",doprint,printdata);
}

PUBLIC COUNTER printCloseBrace(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("}",doprint,printdata);
}

PUBLIC COUNTER printOpenParen(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("(",doprint,printdata);
}

PUBLIC COUNTER printCloseParen(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(")",doprint,printdata);
}

PUBLIC COUNTER printOpenRecord(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("|[",doprint,printdata);
}

PUBLIC COUNTER printCloseRecord(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("]|",doprint,printdata);
}

PUBLIC COUNTER printComma1(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(",",doprint,printdata);
}

PUBLIC COUNTER printComma(doprint,printdata,commap)
IN printFunc doprint;
INOUT char *printdata;
INOUT Boolean *commap;
{
  if (*commap)
    return printComma1(doprint,printdata);
  else
    {
      *commap = true;
      return 0;
    }
}

PUBLIC void printDot(doprint,printdata,commap)
IN printFunc doprint;
INOUT char *printdata;
INOUT Boolean *commap;
{
  if (*commap)
    (void) printBuf("`",doprint,printdata);
  else
    *commap = true;
}

PUBLIC Boolean prettyComma1(doprint,printdata,lmargin)
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin;
{
  (void) printComma1(doprint,printdata);
  return CrTab(doprint,printdata,lmargin);
}

PUBLIC Boolean prettyComma(doprint,printdata,commap,lmargin)
IN printFunc doprint;
INOUT char *printdata;
INOUT Boolean *commap;
IN COUNTER lmargin;
{
  if (*commap)
    return prettyComma1(doprint,printdata,lmargin);
  *commap = true;
  return false;
}

PRIVATE COUNTER printNil(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("[]",doprint,printdata);
}

PRIVATE COUNTER printTuple0(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("()",doprint,printdata);
}

PUBLIC COUNTER printRecord0(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("|[]|",doprint,printdata);
}

PRIVATE COUNTER printColon(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(":",doprint,printdata);
}

PUBLIC COUNTER printOfType(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" : ",doprint,printdata);
}

PUBLIC COUNTER printEqual(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" = ",doprint,printdata);
}

PUBLIC COUNTER printSym(sym,doprint,printdata)
IN CONST SymPtr sym;
IN printFunc doprint;
INOUT char *printdata;
{
  INT len;
  
  len = symNameLen(sym);
  (*doprint)(symName(sym),(COUNTER)len,printdata);
  return len;
}

PRIVATE COUNTER printEmpty(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("{}",doprint,printdata);
}

PRIVATE COUNTER printMapsto(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" => ",doprint,printdata);
}

PRIVATE COUNTER printRef(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("ref ",doprint,printdata);
}

PRIVATE COUNTER printFn(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("<fn>",doprint,printdata);
}

PRIVATE COUNTER printBeginSpecial(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("<",doprint,printdata);
}

PRIVATE COUNTER printEndSpecial(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(">",doprint,printdata);
}

PRIVATE COUNTER printPromise(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("<promise>",doprint,printdata);
}

PRIVATE COUNTER printDynamic(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("pack (",doprint,printdata);
}

PUBLIC COUNTER printSpace(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" ",doprint,printdata);
}

PRIVATE COUNTER printOpenArray(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("[|",doprint,printdata);
}

PRIVATE COUNTER printCloseArray(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("|]",doprint,printdata);
}

PRIVATE COUNTER printEscape(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("\\",doprint,printdata);
}

PUBLIC COUNTER printEtc(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("...",doprint,printdata);
}

PUBLIC COUNTER printPower(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("^",doprint,printdata);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE COUNTER printPure(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf("pure",doprint,printdata);
}

PRIVATE COUNTER printUnion1(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" U",doprint,printdata);
}

PRIVATE struct fxprint {
  MapPtr rename;
  MapPtr toshow;
  COUNTER i;
} fxPrint;

PRIVATE COUNTER fxPrintVar(sigma,doprint,printdata,state)
IN CONST ExprPtr sigma;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  ExprPtr name;
  Boolean found;
  char *buf = alloctmp(16);
  char *s;

  sigma = fxRealSigma(sigma,state);
  name = MapGet(sigma,S_NONGENS(state),&found);
  if (!found)
    name = MapGet(sigma,fxPrint.rename,&found);
  if (!found)
    {
      safememcpy(buf,(char *)STEXT(CAR(sigma)),SLEN(CAR(sigma)));
      s = buf+SLEN(CAR(sigma));
      sprintf(s,"%d",++fxPrint.i);
      name = Ksymbol(buf);
      (void) KEEP(name);
      fxPrint.rename = MapPut(sigma,name,fxPrint.rename);
      fxPrint.toshow = MapPut(sigma,name,fxPrint.toshow);
    }
  else
    safememcpy(buf,STEXT(name),SLEN(name));
  (*doprint)((CHAR *)buf,(COUNTER)SLEN(name),printdata);
  (void) freetmp (buf);
  return SLEN(name);
}

PRIVATE COUNTER printGeq(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" >= ",doprint,printdata);
}

PRIVATE COUNTER fxPrintLatent(f,doprint,printdata,state)
IN CONST ExprPtr f;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  COUNTER len;

  len = 0;
  if (f->kind==(CHAR)E_TYPEFX)
    {
      len += printBuf("[",doprint,printdata);
      len += fxPrintVar(CDR(f),doprint,printdata,state);
      len += printBuf("]",doprint,printdata);
    }
  return len;
}

PUBLIC void fxReset()
{
  fxPrint.rename = fxPrint.toshow = NewMap();
  fxPrint.i = 0;
}
#else
#define fxPrintLatent(f,doprint,printdata,state) 0
#endif

PRIVATE void intPrint(i,doprint,printdata)
IN INT i;
IN printFunc doprint;
INOUT char *printdata;
{
  char *buf = alloctmp(16);
  
  sprintf(buf,"%ld",i);
  if (buf[0]=='-')
    buf[0] = '~';
  (void) printBuf(buf,doprint,printdata);
  (void) freetmp (buf);
}

PUBLIC void realPrint(x,doprint,printdata)
IN CONST REAL *x;
IN printFunc doprint;
INOUT char *printdata;
{
  CHAR format[32],*s,*t,c;
  CHAR *buf = (CHAR *)alloctmp(32);
  ExprPtr form;
  INT len;
  
  switch (classifyD(x))
    {
    case C_QNAN:
    case C_SNAN:
#if FLOAT_FORMAT==IEEE754
      sprintf((char *)buf,"NAN(%d)",NaNcodeD(x));
#endif
      break;
      /*case -C_QNAN:*/
    case -C_SNAN:
#if FLOAT_FORMAT==IEEE754
      sprintf((char *)buf,"~NAN(%d)",-NaNcodeD(x));
#endif
      break;
    case C_INFINITE:
      strcpy((char *)buf,"+inf");
      break;
    case -C_INFINITE:
      strcpy((char *)buf,"~inf");
      break;
    default:
      form = REFVAL(floatFormatRef);
      len = SLEN(form);
      if (len>31)
	len = 31;
      safememcpy(format,STEXT(form),len);
      format[len] = 0;
      sprintf((char *)buf,(char *)format,*x);
      for (t=s=buf; (c = *s++)!=0;)
	{
	  if (c=='-')
	    *t++ = '~';
	  else
	    *t++ = c;
	}
      *t = 0;
      break;
    }
  (void) printBuf((char *)buf,doprint,printdata);
  (void) freetmp (buf);
}

#define MAX_COUNTER (COUNTER)MAX_SIGNED_INT

PUBLIC void stringPrint(text,len,doprint,printdata,quotes)
IN CONST CHAR *text;
IN INT len;
IN printFunc doprint;
INOUT char *printdata;
IN Boolean quotes;
{
  CHAR *s,*t,c,*aux;
  CHAR *buf = (CHAR *)alloctmp(5);
  
  s = text;
  if (quotes)
    (void) printQuote(doprint,printdata);
  while (len!=0)
    {
      c = 0; /* to keep the compiler from compiling that c might be non-
		initialized at the switch (c) below. */
      for (t=s; len!=0; len--, t++)
	{
	  c = *t;
	  if (c<0x20 || (c>=0x80 && c<0xa0) || c=='\\' || c=='\"')
	    break;
	}
      while (s+MAX_COUNTER<t)
	{
	  (*doprint)(s,MAX_COUNTER,printdata);
	  s += MAX_COUNTER;
	}
      (*doprint)(s,(COUNTER)(t-s),printdata);
      s = t+1;
      if (len!=0)
	{
	  len--;
	  switch (c)
	    {
	    case '\\':
	      aux = (CHAR *)"\\\\";
	      break;
	    case '\"':
	      aux = (CHAR *)"\\\"";
	      break;
	    case '\n':
	      aux = (CHAR *)"\\n";
	      break;
	    case '\t':
	      aux = (CHAR *)"\\t";
	      break;
	    case '\v':
	      aux = (CHAR *)"\\v";
	      break;
	    case '\f':
	      aux = (CHAR *)"\\f";
	      break;
	    default:
	      buf[0] = '\\';
	      if (c<0x20)
		{
		  buf[1] = '^';
		  buf[2] = c+'@';
		  buf[3] = 0;
		}
	      else
		{
		  buf[1] = '0'+c/100;
		  c = c%100;
		  buf[2] = '0'+c/10;
		  c = c%10;
		  buf[3] = '0'+c;
		  buf[4] = 0;
		}
	      aux = buf;
	      break;
	    }
	  (void) printBuf((char *)aux,doprint,printdata);
	}
    }
  (void) freetmp (buf);
  if (quotes)
    printQuote(doprint,printdata);
}

PRIVATE void listPrintNoType(val,doprint,printdata)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
{
  Boolean comma;
  
  (void) printOpenBracket(doprint,printdata);
  comma = false;
  while (ISCONS(val))
    {
      (void) printComma(doprint,printdata,&comma);
      valPrintNoType(CAR(val),doprint,printdata,0);
      val = CDR(val);
    }
  if (val!=NIL)
    {
      printSpace(doprint,printdata);
      valPrintNoType(val,doprint,printdata,0);
    }
  (void) printCloseBracket(doprint,printdata);
}

PRIVATE void tuplePrintNoType(val,doprint,printdata)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
{
  Boolean comma;
  VecPtr vp;
  COUNTER i,n;
  ExprPtr *ep;
  
  (void) printOpenParen(doprint,printdata);
  vp = AVAL(val);
  n = vp->n;
  comma = false;
  for (i=0, ep = vp->expr; i<n; i++)
    {
      (void) printComma(doprint,printdata,&comma);
      valPrintNoType(*ep++,doprint,printdata,0);
    }
  (void) printCloseParen(doprint,printdata);
}

MAPSWEEPFUNC(do_printmapnotype)
{
  struct printmapnotype *data;

  data = MAPDATA(struct printmapnotype *);
  (void) printComma(data->doprint,data->printdata,&data->comma);
  valPrintNoType(x,data->doprint,data->printdata,0);
  if (y!=UNIT)
    {
      (void) printMapsto(data->doprint,data->printdata);
      valPrintNoType(y,data->doprint,data->printdata,0);
    }
  return false;
}

MAPSWEEPFUNC(do_dumpshape)
{
  ExprPtr *exprs = MAPDATA(ExprPtr *);

  exprs[IVAL(y)-1] = x;
  return false;
}

PRIVATE void dumpShape(shape,exprs)
IN CONST ExprPtr shape;
OUT ExprPtr *exprs;
{
  (void) MapSweep(shape,do_dumpshape,(CHAR *)exprs);
}

PRIVATE void recordPrintNoType(val,doprint,printdata)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
{
  COUNTER n = AVAL(val)->n;
  ExprPtr *exprs = (ExprPtr *)alloctmp(sizeof(ExprPtr)*n);

  dumpShape(ASHAPE(val),exprs);
  {
    REGISTER COUNTER i;
    Boolean comma;

    comma = false;
    (void) printOpenRecord(doprint,printdata);
    for (i=0; i<n; i++)
      {
	(void) printComma(doprint,printdata,&comma);
	(void) printSym(exprs[i],doprint,printdata);
	(void) printEqual(doprint,printdata);
	valPrintNoType(AVAL(val)->expr[i],doprint,printdata,0);
      }
  }
  (void) freetmp (exprs);
  (void) printCloseRecord(doprint,printdata);
}

PRIVATE void dataPrintNoType(val,doprint,printdata,pri)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
{
  SHORT magic;
  COUNTER newpri;
  ExprPtr ttm,con;
  ENTER;
  
  ttm = CAR(val);
  magic = val->magic;
  con = AVAL(CAR(ttm))->expr[magic-ttm->magic];
  newpri = getPri(con);
  val = CDR(val);
  if (magic>=0) /* has an argument */
    {
      if (pri<newpri)
	(void) printOpenParen(doprint,printdata);
      if (newpri==0) /* not infix */
	{
	  (void) printSym(con,doprint,printdata);
	  (void) printSpace(doprint,printdata);
	  valPrintNoType(val,doprint,printdata,-1);
	}
      else /* infix(left) if newpri odd, infix(right) if newpri even */
	{
	  valPrintNoType(CAR(val),doprint,printdata,newpri);
	  (void) printSpace(doprint,printdata);
	  (void) printSym(con,doprint,printdata);
	  (void) printSpace(doprint,printdata);
	  valPrintNoType(CDR(val),doprint,printdata,newpri-1);
	}
      if (pri<newpri)
	(void) printCloseParen(doprint,printdata);
    }
  else
    (void) printSym(con,doprint,printdata);
  QUIT;
}

PUBLIC void dynPrint(val,doprint,printdata,pri)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
{
  valPrint(PACKVAL(val),PACKTYPE(val),doprint,printdata,pri);
  (void) printOfType(doprint,printdata);
  fxReset();
  TypePrint(PACKTYPE(val),doprint,printdata,NIL);
  fxFlushPrint(doprint,printdata,NIL,true);
}

PRIVATE void printComplexException (val,doprint,printdata,pri)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
{
  if (pri<0)
    (void) printOpenParen(doprint,printdata);
  valPrintNoType(CAR(val),doprint,printdata,0);
  (void) printSpace(doprint,printdata);
  valPrint(CDR(val),getTypeInExn(CAR(val)),doprint,printdata,-1);
  if (pri<0)
    (void) printCloseParen(doprint,printdata);
}

PRIVATE void arrayPrint(val,type,doprint,printdata)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
{
  ExprPtr *ep;
  COUNTER i,n;
  VecPtr vp;
  Boolean comma;
  
  switch (val->kind)
    {
    case E_ARRAY_2:
      valPrint(CAR(val),type,doprint,printdata,0);
      (void) printComma1(doprint,printdata);
      valPrint(CDR(val),type,doprint,printdata,0);
      break;
    case E_ARRAY_GNRL:
      vp = AVAL(val);
      n = vp->n;
      ep = vp->expr;
      comma = false;
      for (i=0; i<n; i++)
	{
	  (void) printComma(doprint,printdata,&comma);
	  valPrint(*ep++,type,doprint,printdata,0);
	}
      break;
    default:
      kinderror(message(84),val->kind);
    }
}

PUBLIC void valPrintNoType(val,doprint,printdata,pri)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
{
  /*if (val==NIL)
    (void) printNil(doprint,printdata);
    else*/ switch (val->kind)
      {
      case E_NIL:
	(void) printNil(doprint,printdata);
	break;
      case E_INT:
	intPrint(IVAL(val),doprint,printdata);
	break;
      case E_REAL:
	realPrint(&XVAL(val),doprint,printdata);
	break;
      case E_COMPLEX:
	valPrintNoType(REPART(val),doprint,printdata,0);
	(void) printColon(doprint,printdata);
	valPrintNoType(IMPART(val),doprint,printdata,0);
	break;
      case E_STRING:
	stringPrint(STEXT(val),SLEN(val),doprint,printdata,true);
	break;
      case E_CONS:
	if (CAR(val)->kind==(CHAR)E_EXCEPTION)
	  {
	    ExprPtr next;

	    next = CDR(val);
	    if (next->kind!=(CHAR)E_NIL && next->kind!=(CHAR)E_CONS)
	      {
		printComplexException (val, doprint, printdata, pri);
		break;
	      }
	  }
	listPrintNoType(val,doprint,printdata);
	break;
      case E_TUPLE_0:
	(void) printTuple0(doprint,printdata);
	break;
      case E_TUPLE_1:
	(void) printOpenParen(doprint,printdata);
	valPrintNoType(REFVAL(val),doprint,printdata,0);
	(void) printCloseParen(doprint,printdata);
	break;
      case E_TUPLE_2:
	(void) printOpenParen(doprint,printdata);
	valPrintNoType(CAR(val),doprint,printdata,0);
	(void) printComma1(doprint,printdata);
	valPrintNoType(CDR(val),doprint,printdata,0);
	(void) printCloseParen(doprint,printdata);
	break;
      case E_TUPLE_GNRL:
	tuplePrintNoType(val,doprint,printdata);
	break;
      case E_RECORD_0:
	(void) printRecord0(doprint,printdata);
	break;
      case E_RECORD_1:
	(void) printOpenRecord(doprint,printdata);
	(void) printSym(CAR(val),doprint,printdata);
	(void) printEqual(doprint,printdata);
	valPrintNoType(CDR(val),doprint,printdata,0);
	(void) printCloseRecord(doprint,printdata);
	break;
      case E_RECORD_GNRL:
	recordPrintNoType(val,doprint,printdata);
	break;
      case E_DATA:
	dataPrintNoType(val,doprint,printdata,pri);
	break;
      case E_EMPTY:
	(void) printEmpty(doprint,printdata);
	break;
      case E_MAPLET:
	if (MAPLETY(val)!=UNIT)
	  (void) printOpenBrace(doprint,printdata);
	valPrintNoType(MAPLETX(val),doprint,printdata,0);
	if (MAPLETY(val)!=UNIT)
	  {
	    (void) printMapsto(doprint,printdata);
	    valPrintNoType(MAPLETY(val),doprint,printdata,0);
	    (void) printCloseBrace(doprint,printdata);
	  }
	break;
      case E_MAP:
	{
	  struct printmapnotype *data = (struct printmapnotype *)alloctmp(sizeof(struct printmapnotype));
	  
	  data->doprint = doprint;
	  data->printdata = printdata;
	  data->comma = false;
	  (void) printOpenBrace(doprint,printdata);
	  (void) MapSweep(val,do_printmapnotype,data);
	  (void) printCloseBrace(doprint,printdata);
	  (void) freetmp (data);
	  break;
	}
      case E_QUOTE:
	printBuf("<quote>",doprint,printdata);
	break;
      case E_REF:
	if (pri<0)
	  (void) printOpenParen(doprint,printdata);
	(void) printRef(doprint,printdata);
	valPrintNoType(REFVAL(val),doprint,printdata,-1);
	if (pri<0)
	  (void) printCloseParen(doprint,printdata);
	break;
      case E_COMPILED_CLOSURE:
      case E_INTERPRETED_CLOSURE:
      case E_POLYFUN:
	/*case E_N_LAMBDA:*/
	(void) printFn(doprint,printdata);
	break;
      case E_ENV:
	(void) printBuf("<env>",doprint,printdata);
	break;
      case E_EXCEPTION:
	(void) printSym(EXCVAL(val),doprint,printdata);
	break;
      case E_ARRAY_2:
      case E_ARRAY_GNRL:
	{
	  if (pri<0)
	    (void) printOpenParen(doprint,printdata);
	  (void) printOpenArray(doprint,printdata);
	  arrayPrint(val,alpha,doprint,printdata);
	  (void) printCloseArray(doprint,printdata);
	  if (pri<0)
	    (void) printCloseParen(doprint,printdata);
	  break;
	}
      case E_PACK:
	(void) printDynamic(doprint,printdata);
	dynPrint(val,doprint,printdata,pri);
	(void) printCloseParen(doprint,printdata);
	break;
      case E_EXTERNAL:
	if (EXTDESC(val)->print!=NULL)
	  (*EXTDESC(val)->print)(EXTDATA(val),doprint,printdata);
	else
	  {
	    (void) printBeginSpecial(doprint,printdata);
	    (void) printBuf((char *)EXTDESC(val)->name,doprint,printdata);
	    (void) printEndSpecial(doprint,printdata);
	  }
	break;
      default:
	kinderror(message(81),val->kind);
      }
}

PRIVATE void listPrint(val,elttype,doprint,printdata)
IN CONST ExprPtr val;
IN CONST ExprPtr elttype;
IN printFunc doprint;
INOUT char *printdata;
{
  Boolean comma;

  (void) printOpenBracket(doprint,printdata);
  comma = false;
  while (val!=NIL)
    {
      (void) printComma(doprint,printdata,&comma);
      valPrint(CAR(val),elttype,doprint,printdata,0);
      val = CDR(val);
    }
  (void) printCloseBracket(doprint,printdata);
}

PRIVATE void tuplePrint(val,type,doprint,printdata)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
{
  Boolean comma;
  VecPtr vp;
  COUNTER i,n;
  ExprPtr *ep;
  ExprPtr elttype;
  
  (void) printOpenParen(doprint,printdata);
  switch (val->kind)
    {
    case E_TUPLE_0:
      break;
    case E_TUPLE_1:
      elttype = ISCONS(type)?CAR(type):type;
      valPrint(REFVAL(val),elttype,doprint,printdata,0);
      break;
    case E_TUPLE_2:
      if (ISCONS(type))
	{
	  elttype = CAR(type);
	  type = CDR(type);
	}
      else
	elttype = type; /* a type variable */
      valPrint(CAR(val),elttype,doprint,printdata,0);
      (void) printComma1(doprint,printdata);
      elttype = ISCONS(type)?CAR(type):type;
      valPrint(CDR(val),elttype,doprint,printdata,0);
      break;
    case E_TUPLE_GNRL:
      vp = AVAL(val);
      n = vp->n;
      comma = false;
      for (i=0, ep = vp->expr; i<n; i++)
	{
	  (void) printComma(doprint,printdata,&comma);
	  if (ISCONS(type))
	    {
	      elttype = CAR(type);
	      type = CDR(type);
	    }
	  else
	    elttype = type; /* a type variable */
	  valPrint(*ep++,elttype,doprint,printdata,0);
	}
      break;
    default:
      kinderror(message(82),val->kind);
    }
  (void) printCloseParen(doprint,printdata);
}

MAPSWEEPFUNC(do_printmap)
{
  struct printmap *data;

  data = MAPDATA(struct printmap *);
  (void) printComma(data->doprint,data->printdata,&data->comma);
  valPrint(x,data->domtype,data->doprint,data->printdata,0);
  if (data->rngtype!=NIL)
    {
      (void) printMapsto(data->doprint,data->printdata);
      valPrint(y,data->rngtype,data->doprint,data->printdata,0);
    }
  return false;
}

PRIVATE void mapPrint(val,domtype,rngtype,doprint,printdata)
IN CONST ExprPtr val;
IN CONST ExprPtr domtype,rngtype;
IN printFunc doprint;
INOUT char *printdata;
{
  struct printmap *data = (struct printmap *)alloctmp(sizeof(struct printmap));

  (void) printOpenBrace(doprint,printdata);
  data->doprint = doprint;
  data->printdata = printdata;
  data->domtype = domtype;
  data->rngtype = rngtype;
  data->comma = false;
  (void) MapSweep(val,do_printmap,data);
  (void) printCloseBrace(doprint,printdata);
  (void) freetmp (data);
}

PRIVATE void recordPrint(val,type,doprint,printdata)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
{
  (void) printOpenRecord(doprint,printdata);
  switch (val->kind)
    {
    case E_RECORD_0:
      break;
    case E_RECORD_1: /* then type is a maplet from lab<>REST_TAG to type */
      (void) printSym(CAR(val),doprint,printdata);
      (void) printEqual(doprint,printdata);
      valPrint(CDR(val),MAPLETY(type),doprint,printdata,0);
      break;
    case E_RECORD_GNRL:
      {
	COUNTER n = AVAL(val)->n;
	ExprPtr *exprs = (ExprPtr *)alloctmp(sizeof(ExprPtr)*n);

	dumpShape(ASHAPE(val),exprs);
	{
	  REGISTER COUNTER i;
	  Boolean comma,found;
	  ExprPtr x,xtype;

	  comma = false;
	  for (i=0; i<n; i++)
	    {
	      (void) printComma(doprint,printdata,&comma);
	      x = exprs[i];
	      (void) printSym(x,doprint,printdata);
	      (void) printEqual(doprint,printdata);
	      xtype = MapGet(x,type,&found);
	      if (!found)
		xtype = MapGetSure(REST_TAG,type);
	      valPrint(AVAL(val)->expr[i],xtype,doprint,printdata,0);
	    }
	}
	(void) freetmp (exprs);
      }
      break;
    default:
      kinderror(message(83),val->kind);
    }
  (void) printCloseRecord(doprint,printdata);
}

#ifdef OLDDATA
MAPSWEEPFUNC(do_getdatacon)
{
  struct datacon *data;
  Boolean found;

  data = MAPDATA(struct datacon *);
  if (CDR(y)==NIL)
    found = true;
  else
    (void) MapGet(data->tag,CDR(y),&found);
  if (found)
    {
      if (data->con==NIL)
	{
	  data->con = x;
	  if (!data->possiblemultiple)
	    return true; /* found the only possible one */
	}
      else
	{
	  if (data->multiple)
	    return true; /* quit: this is a multiple tag */
	  data->multiple = true;
	}
    }
  return false;
}

PRIVATE ExprPtr getDataCon(val,tyname,boxedp)
IN CONST ExprPtr val;
IN CONST ExprPtr tyname;
OUT Boolean *boxedp;
{
  MapPtr tytagmap;
  MapPtr tytags;
  struct datacon *data = (struct datacon *)alloctmp(sizeof(struct datacon));
  
  tytagmap = MapGetSure(tyname,TyTags(basisTable));
  data->tag = /*(val==NIL)?MAPLETX(nilTags):*/KINT(val->kind);
  data->con = NIL;
  data->possiblemultiple = (ISCONS(val) &&
			    /*CAR(val)!=NIL && */
			    CAR(val)->kind==(CHAR)E_VAR) ||
    val==NIL || val->kind==(CHAR)E_VAR;
  data->multiple = false;
  (void) MapSweep(tytagmap,do_getdatacon,data);
  /* data->con should be returned non NIL */
  if (data->multiple)
    data->con = (val->kind==(CHAR)E_CONS)?CAR(val):val;
  /* if several constructors match, then the data itself
     must contain the constructor */
  tytags = MapGetSure(data->con,tytagmap);
  *boxedp = (tytags->kind==(CHAR)E_CONS);
  val = data->con;
  (void) freetmp (data);
  return val;
}
#endif

PRIVATE void dataPrint(val,tyname,argtypes,doprint,printdata,pri)
IN CONST ExprPtr val;
IN CONST ExprPtr tyname,argtypes;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
{
  ExprPtr tycons,tyhdr,cons,ttm,con,type,r;
  COUNTER len,newpri;
  Boolean found;
#ifdef OLDDATA
  Boolean boxed;
#endif
  ENTER;

#ifdef OLDDATA
  con = getDataCon(val,tyname,&boxed);
#else
  tycons = STyCons(basisState);
  ttm = MapGet(tyname,TyTags(basisTable),&found);
  if (found && ttm->kind==E_STRING) /* special case of datatype with only one, non-constant
				       constructor. */
    {
      con = ttm;
    }
  else
    {
      SHORT magic;

      ttm = CAR(val);
      magic = val->magic;
      con = AVAL(CAR(ttm))->expr[magic-ttm->magic];
      val = CDR(val);
    }
#endif
  newpri = getPri(con);
  r = MapGet(tyname,tycons,&found);
  if (found)
    {
      tyhdr = CAR(r);
      cons = CDR(r);
      type = MapGetSure(con,cons);
#ifdef OLDDATA
      if (boxed && ISCONS(val)) /* boxed is true when the constructor is present,
				   including the case 'con' without an inner type */
	val = CDR(val);
#endif
      if (fxSym(APPLFUN(type))==funSym) /* should be equivalent to magic<0 */
	{
	  if (pri<newpri)
	    (void) printOpenParen(doprint,printdata);
	  type = CAR(APPLARG(type));
	  type = ApplyType(type,tyhdr,argtypes,basisAssum,basisState);
	  if (newpri==0) /* not infix */
	    {
	      (void) printSym(con,doprint,printdata);
	      (void) printSpace(doprint,printdata);
	      valPrint(val,type,doprint,printdata,-1);
	    }
	  else /* infix(left) if newpri odd, infix(right) if newpri even */
	    {
	      valPrint(CAR(val),CAR(type),doprint,printdata,newpri);
	      (void) printSpace(doprint,printdata);
	      (void) printSym(con,doprint,printdata);
	      (void) printSpace(doprint,printdata);
	      valPrint(CDR(val),CAR(CDR(type)),doprint,printdata,newpri-1);
	    }
	  if (pri<newpri)
	    (void) printCloseParen(doprint,printdata);
	}
      else
	(void) printSym(con,doprint,printdata);
    }
  else /* abstype */
    {
      len = SLEN(tyname);
      {
	CHAR *s;

	s = STEXT(tyname);
	for (;;)
	  {
	    len--;
	    if (*s++=='%')
	      break;
	  }
	(void) printBeginSpecial(doprint,printdata);
	(*doprint)(s,len,printdata);
	(void) printEndSpecial(doprint,printdata);
      }
    }
  QUIT;
}

MAPSWEEPFUNC(do_printnum)
{
  struct printnum *data;
  ExprPtr tycons,unit;

  data = MAPDATA(struct printnum *);
  printDot(data->doprint,data->printdata,&data->comma);
  if (x->kind==(CHAR)E_VAR)
    (void) printSym(x,data->doprint,data->printdata);
  else
    {
      x = applFun(x,basisState);
      x = fxSym(x); /* the dimension name */
      tycons = STyCons(basisState);
      unit = MapGetSure(x,tycons);
      (void) printSym(unit,data->doprint,data->printdata);
    }
  if (y!=one)
    {
      (void) printPower(data->doprint,data->printdata);
      valPrintNoType(y,data->doprint,data->printdata,0);
    }
  return false;
}

PRIVATE ExprPtr getValue(i,l,env)
IN COUNTER i;
IN REGISTER COUNTER l;
IN CONST ExprPtr env;
{
  REGISTER ExprPtr ee, nextenv;
  REGISTER COUNTER n;

  for (ee=env, n=0; ee!=NIL; ee=nextenv)
    {
      nextenv = ENEXT(ee);
      if (EVARS(ee)==EVARS(GLOBENV) && nextenv!=NIL)
	continue; /* fake env. */
      n++;
    }
  n -= l;
  for (ee=env, l=1; l<n; ee=nextenv)
    {
      nextenv = ENEXT(ee);
      if (EVARS(ee)==EVARS(GLOBENV) && nextenv!=NIL)
	continue; /* fake env. */
      l++;
    }
  return EVARS(ee)->expr[i];
}

PUBLIC ExprPtr getValueByName(name,foundp,env)
IN CONST ExprPtr name;
OUT Boolean *foundp;
IN CONST ExprPtr env;
{
  ExprPtr i_l;
  COUNTER i,l;

  i_l = MapGet(name,IdLocs(basisTable),foundp);
  if (!*foundp)
    return NIL;
  if (i_l->kind==(CHAR)E_CONS)
    {
      i = IVAL(CAR(i_l));
      l = IVAL(CDR(i_l));
    }
  else
    {
      i = IVAL(i_l);
      l = 0;
    }
  return getValue(i,l,env);
}

PUBLIC ExprPtr getTypeInExn(exc)
IN CONST ExprPtr exc;
{
  ExprPtr name,type,val;
  Boolean found;

  name = REFVAL(exc);
  type = MapGet(name,IdTypes(basisAssum),&found);
  if (!found)
    return _alpha;
  val = getValueByName(name,&found,basisEnv);
  if (!found)
    return _alpha;
  if (val==exc ||
      (val->kind==(CHAR)E_INTERPRETED_CLOSURE && QENV(val)->kind==(CHAR)E_TUPLE_2
       && CDR(QENV(val))==exc))
    return CAR(APPLARG(type));
  return _alpha;
}

#if DEBUGGING

PRIVATE Boolean dataCheckTypeTag(val,tyname)
IN CONST ExprPtr val;
IN CONST ExprPtr tyname;
{
#ifdef OLDDATA
  MapPtr tytagmap;
  struct datacon *data = (struct datacon *)alloctmp(sizeof(struct datacon));
  Boolean found;

  tytagmap = MapGetSure(tyname,TyTags(basisTable));
  data->tag = /*(val==NIL)?MAPLETX(nilTags):*/KINT(val->kind);
  data->con = NIL;
  data->possiblemultiple = (ISCONS(val) &&
			    /*CAR(val)!=NIL && */
			    CAR(val)->kind==(CHAR)E_VAR) ||
			      val==NIL ||
				val->kind==(CHAR)E_VAR;
  data->multiple = false;
  (void) MapSweep(tytagmap,do_getdatacon,data);
  return data->con!=NIL &&
	 (data->multiple?((/*val!=NIL && */val->kind==(CHAR)E_CONS &&
			   ((void) MapGet(CAR(val),tytagmap,&found), found)
			   ) ||
			  ((void) MapGet(val,tytagmap,&found), found))
	  :true);
  /* if several constructors match, then the data itself
     must contain the constructor */
#else
  MapPtr ttm;
  Boolean found;

  if (val->kind==(CHAR)E_DATA)
    return true;
  ttm = MapGet(tyname,TyTags(basisTable),&found);
  return (found && ttm->kind!=E_DATA);
#endif
}

PRIVATE Boolean checkTupleTag(val,type)
IN CONST ExprPtr val,type;
{
  type = CDR(type);
  /*if (type==NIL) return val!=NIL && val->kind==(CHAR)E_TUPLE_1;*/
  switch (type->kind)
    {
      case E_VAR:
	return /*val!=NIL && */val->kind==(CHAR)E_TUPLE_1 ||
				val->kind==(CHAR)E_TUPLE_2 ||
				val->kind==(CHAR)E_TUPLE_GNRL;
      case E_NIL:
	return /*val!=NIL && */val->kind==(CHAR)E_TUPLE_1;
      case E_CONS:
	type = CDR(type);
	/*if (type==NIL) return val!=NIL && val->kind==(CHAR)E_TUPLE_2;*/
	switch (type->kind)
	  {
	    case E_VAR:
	      return /*val!=NIL && */val->kind==(CHAR)E_TUPLE_2 ||
				     val->kind==(CHAR)E_TUPLE_GNRL;
	    case E_NIL:
	      return /*val!=NIL && */val->kind==(CHAR)E_TUPLE_2;
	    case E_CONS:
	      return /*val!=NIL && */val->kind==(CHAR)E_TUPLE_GNRL;
	  }
    }
  return false;
}

PUBLIC Boolean checkTypeTag(val,type)
IN CONST ExprPtr val,type;
{
  /*if (type==NIL)
    return val==tuple0;
    else*/ switch (type->kind)
      {
      case E_NIL:
	return val==tuple0;
      case E_VAR:
	return true;
      case E_TYPE_APPL:
	{
	  ExprPtr f;
	  
	  f = APPLFUN(type);
	  f = fxSym(f);
	  if (f==stringSym)
	    return (/*val!=NIL && */val->kind==(CHAR)E_STRING);
	  else if (f==intSym)
	    return (/*val!=NIL && */val->kind==(CHAR)E_INT);
	  else if (f==listSym)
	    return (val==NIL || val->kind==(CHAR)E_CONS);
	  else if (f==refSym)
	    return (/*val!=NIL && */val->kind==(CHAR)E_REF);
	  else if (f==tableSym)
	    return (/*val!=NIL && */val->kind==(CHAR)E_EXTERNAL && EXTDESC(val)==&tableExtDesc);
	  else if (f==arraySym)
	    return (/*val!=NIL && */(
				     val->kind==(CHAR)E_ARRAY_2 ||
				     val->kind==(CHAR)E_ARRAY_GNRL));
	  else if (f==funSym || f==promiseSym)
	    return (/*val!=NIL && */val->kind==(CHAR)E_COMPILED_CLOSURE ||
		    val->kind==(CHAR)E_INTERPRETED_CLOSURE
				    );
	  else if (f==mapSym)
	    return ISMAP(val);
	  else if (f==boolSym)
	    return val==trueValue || val==falseValue;
	  else if (f==exnSym)
	    return ((ISCONS(val) &&
			/*CAR(val)!=NIL && */CAR(val)->kind==(CHAR)E_EXCEPTION) ||
		    (/*val!=NIL && */val->kind==(CHAR)E_EXCEPTION));
	  else if (f==dynSym)
	    return (/*val!=NIL && */val->kind==(CHAR)E_PACK);
	  else
	    return dataCheckTypeTag(val,f);
	  break;
	}
      case E_CONS:
	return checkTupleTag(val,type);
      case E_EMPTY:
	return val==record0;
      case E_MAPLET:
	return /*val!=NIL && */((MAPLETX(type)==REST_TAG &&
				 (val->kind==(CHAR)E_RECORD_0 ||
				  val->kind==(CHAR)E_RECORD_1 ||
				  val->kind==(CHAR)E_RECORD_GNRL)) ||
				(val->kind==(CHAR)E_RECORD_1 &&
				 CAR(val)==MAPLETX(type)));
      case E_MAP:
	{
	  Boolean found;
	  MapPtr fields;

	  (void) MapGet(REST_TAG,type,&found);
	  if (found)
	    {
	      fields = MapRemove(REST_TAG,type);
	      return (val->kind==(CHAR)E_RECORD_1 &&
		      fields->kind==(CHAR)E_MAPLET &&
		      CAR(val)==MAPLETX(fields)) ||
		     (val->kind==(CHAR)E_RECORD_GNRL &&
		      SetIncludedp(fields,ASHAPE(val)));
	    }
	  else
	    {
	      return val->kind==(CHAR)E_RECORD_GNRL &&
		     sameDomains(ASHAPE(val),type);
	    }
	}
	break;
      case E_NUMTYPE:
	return /*val!=NIL && */(val->kind==(CHAR)E_REAL || val->kind==(CHAR)E_COMPLEX);
      }
  return false;
}
#endif

PUBLIC void valPrint(val,type,doprint,printdata,pri)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
{
  if (!checkTypeTag(val,type))
    valPrintNoType(val,doprint,printdata,pri);
  else
  /*if (type==NIL)
    (void) printTuple0(doprint,printdata);
    else*/ switch (type->kind)
      {
      case E_NIL:
	(void) printTuple0(doprint,printdata);
	break;
      case E_VAR:
	valPrintNoType(val,doprint,printdata,pri);
	break;
      case E_TYPE_APPL:
	{
	  ExprPtr f;
	  
	  f = APPLFUN(type);
	  f = fxSym(f);
	  if (f==stringSym)
	  stringPrint(STEXT(val),SLEN(val),doprint,printdata,true);
	  else if (f==intSym)
	    intPrint(IVAL(val),doprint,printdata);
	  else if (f==listSym)
	    listPrint(val,CAR(APPLARG(type)),doprint,printdata);
	  else if (f==refSym)
	    {
	      if (pri<0)
		(void) printOpenParen(doprint,printdata);
	      (void) printRef(doprint,printdata);
	      valPrint(REFVAL(val),CAR(APPLARG(type)),doprint,printdata,-1);
	      if (pri<0)
		(void) printCloseParen(doprint,printdata);
	    }
	  else if (f==arraySym)
	    {
	      if (pri<0)
		(void) printOpenParen(doprint,printdata);
	      (void) printOpenArray(doprint,printdata);
	      arrayPrint(val,CAR(APPLARG(type)),doprint,printdata);
	      (void) printCloseArray(doprint,printdata);
	      if (pri<0)
		(void) printCloseParen(doprint,printdata);
	    }
	  else if (f==tableSym)
	    valPrintNoType(val,doprint,printdata,pri);
	  else if (f==funSym)
	    (void) printFn(doprint,printdata);
	  else if (f==mapSym)
	    mapPrint(val,CAR(APPLARG(type)),CAR(CDR(APPLARG(type))),
		     doprint,printdata);
	  else if (f==exnSym)
	    {
	      if (ISCONS(val))
		printComplexException (val, doprint, printdata, pri);
	      else
		valPrintNoType(val,doprint,printdata,0);
	    }
	  else if (f==promiseSym)
	    (void) printPromise(doprint,printdata);
	  else if (f==dynSym)
	    valPrintNoType(val,doprint,printdata,false);
	  else
	    dataPrint(val,f,APPLARG(type),doprint,printdata,pri);
	  break;
	}
      case E_CONS:
	tuplePrint(val,type,doprint,printdata);
	break;
      case E_EMPTY:
	(void) printRecord0(doprint,printdata);
	break;
      case E_MAPLET:
      case E_MAP:
	recordPrint(val,type,doprint,printdata);
	break;
      case E_NUMTYPE:
	valPrintNoType(val,doprint,printdata,0);
	type = REFVAL(type);
	if (ISNEMAP(type))
	  {
	    struct printnum *data = (struct printnum *)alloctmp(sizeof(struct printnum));
	    
	    /*(void) printSpace(doprint,printdata);*/
	    data->doprint = doprint;
	    data->printdata = printdata;
	    data->comma = true;
	    (void) MapSweep(type,do_printnum,data);
	    (void) freetmp (data);
	  }
	break;
      default:
	kinderror(message(85),type->kind);
	break;
      }
}

PRINTFUNC(printoverflow)
{
  struct overflow *data;

  data = PRINTDATA(struct overflow *);
  data->lmargin += len;
  if (data->lmargin>=data->rmargin)
    LONGJUMP(data->exit,1);
}

PRIVATE Boolean printOverflow(val,type,lmargin,rmargin,pri)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN COUNTER lmargin,rmargin;
IN COUNTER pri;
{
  struct overflow *VOLATILE data = (struct overflow *)alloctmp(sizeof(struct overflow));
  int ret;

  data->lmargin = lmargin;
  data->rmargin = rmargin;
  if (SETJUMP(data->exit)==0)
    {
      valPrint(val,type,printoverflow,(char *)data,pri);
      ret = 0;
    }
  else ret = 1;
#ifdef HAS_VOLATILE
  (void) freetmp (data);
#endif
  return ret;
}

PUBLIC void endPretty(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  if (crtab_maxlines<=0)
    printBuf("\n...\n",doprint,printdata);
}

PUBLIC Boolean CrTab(doprint,printdata,lmargin)
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin;
{
#define BUFSIZE 128
  CHAR *buf;
  CHAR *s,*send;

  if (--crtab_maxlines<0)
    return true;
  buf = (CHAR *)alloctmp(BUFSIZE);
  buf[0] = '\n';
  lmargin++;
  for (s=buf+1, send = (lmargin<BUFSIZE)?(buf+lmargin):(buf+BUFSIZE); s<send;)
    *s++ = ' ';
  while (lmargin>BUFSIZE)
    {
      (*doprint)(buf,BUFSIZE,printdata);
      buf[0] = ' ';
      lmargin -= BUFSIZE;
    }
  (*doprint)(buf,lmargin,printdata);
  (void) freetmp (buf);
  return false;
#undef BUFSIZE
}

PRIVATE Boolean strprint(data)
     IN struct strcut *data;
{
  if (SETJUMP(data->data.exit)==0)
    {
      stringPrint(data->text,data->len,printoverflow,(char *)&data->data,true);
      return 0;
    }
  else return 1;
}

PRIVATE INT stringCut(text,len,lmargin,rmargin)
IN CONST CHAR *text;
IN INT len;
IN COUNTER lmargin,rmargin;
{
  struct strcut *data;
  INT len2;
  INT len3;
  CHAR c;
  
  if (len<=1)
    return len;
  data = (struct strcut *)alloctmp(sizeof(struct strcut));
  data->data.rmargin = rmargin-1;
  for (len2 = 1, len3 = 0; len2<len; )
    {
      c = text[len2-1];
      if (c>='0' && c<='9')
	while (len2<len)
	  {
	    c = text[len2];
	    if (!(c>='0' && c<='9'))
	      break;
	    len2++;
	  }
      else if ((c>='A' && c<='Z') || (c>='a' && c<='z'))
	while (len2<len)
	  {
	    c = text[len2];
	    if (!((c>='A' && c<='Z') || (c>='a' && c<='z')))
	      break;
	    len2++;
	  }
      data->data.lmargin = lmargin+len3;
      data->text = text+len3;
      data->len = len2-len3;
      if (strprint(data)==0)
	{
	  len3 = len2;
	  len2++;
	  continue;
	}
      if (len3==0)
	break;
      len = len3;
      break;
    }
  (void) freetmp (data);
  return len;
}

PRIVATE Boolean stringPretty(text,len,doprint,printdata,lmargin,rmargin)
IN CONST CHAR *text;
IN INT len;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin,rmargin;
{
  INT cut;
  
  (void) printQuote(doprint,printdata);
  while (cut = stringCut(text,len,lmargin,rmargin), cut<len)
    {
      stringPrint(text,cut,doprint,printdata,false);
      (void) printEscape(doprint,printdata);
      if (CrTab(doprint,printdata,lmargin))
	return true;
      (void) printEscape(doprint,printdata);
      text += cut;
      len -= cut;
    }
  stringPrint(text,cut,doprint,printdata,false);
  (void) printQuote(doprint,printdata);
  return false;
}

PRIVATE Boolean listPretty(val,elttype,doprint,printdata,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr elttype;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin,rmargin;
{
  Boolean comma;
  
  lmargin++;
  rmargin--;
  (void) printOpenBracket(doprint,printdata);
  comma = false;
  while (val!=NIL)
    {
      if (prettyComma(doprint,printdata,&comma,lmargin) ||
	  valPretty(CAR(val),elttype,doprint,printdata,0,lmargin,rmargin))
	return true;
      val = CDR(val);
    }
  (void) printCloseBracket(doprint,printdata);
  return false;
}

PRIVATE Boolean tuplePretty(val,type,doprint,printdata,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin,rmargin;
{
  Boolean comma;
  VecPtr vp;
  COUNTER i,n;
  ExprPtr *ep;
  ExprPtr elttype;
  
  (void) printOpenParen(doprint,printdata);
  lmargin++;
  rmargin--;
  switch (val->kind)
    {
    case E_TUPLE_0:
      break;
    case E_TUPLE_1:
      elttype = ISCONS(type)?CAR(type):type;
      if (valPretty(REFVAL(val),elttype,doprint,printdata,0,lmargin,rmargin))
	return true;
      break;
    case E_TUPLE_2:
      if (ISCONS(type))
	{
	  elttype = CAR(type);
	  type = CDR(type);
	}
      else
	elttype = type; /* a type variable */
      if (valPretty(CAR(val),elttype,doprint,printdata,0,lmargin,rmargin) ||
	  prettyComma1(doprint,printdata,lmargin))
	return true;
      elttype = ISCONS(type)?CAR(type):type;
      if (valPretty(CDR(val),elttype,doprint,printdata,0,lmargin,rmargin))
	return true;
      break;
    case E_TUPLE_GNRL:
      vp = AVAL(val);
      n = vp->n;
      comma = false;
      for (i=0, ep = vp->expr; i<n; i++)
	{
	  if (prettyComma(doprint,printdata,&comma,lmargin))
	    return true;
	  if (ISCONS(type))
	    {
	      elttype = CAR(type);
	      type = CDR(type);
	    }
	  else
	    elttype = type; /* a type variable */
	  if (valPretty(*ep++,elttype,doprint,printdata,0,lmargin,rmargin))
	    return true;
	}
      break;
    default:
      kinderror(message(86),val->kind);
    }
  (void) printCloseParen(doprint,printdata);
  return false;
}

PRIVATE int do_pmap_overflow(pdata)
     IN struct prettymap *pdata;
{
  int overflow;

  if (SETJUMP(pdata->data.exit)==0)
    {
      (void) do_printmap(pdata->x,pdata->y,(CHAR *)&pdata->data2);
      overflow = 0;
    }
  else overflow = 1;
  return overflow;
}

MAPSWEEPFUNC(do_prettymap)
{
  struct printmap *data;
  struct prettymap *pdata = (struct prettymap *)alloctmp(sizeof(struct prettymap));
  COUNTER len;
  int overflow;

  data  = MAPDATA(struct printmap *);
  if (prettyComma(data->doprint,data->printdata,&data->comma,data->lmargin))
    {
      (void) freetmp (pdata);
      return true;
    }
  pdata->x = x;
  pdata->y = y;
  pdata->data.lmargin = data->lmargin;
  pdata->data.rmargin = data->rmargin;
  pdata->data2.doprint = printoverflow;
  pdata->data2.printdata = (char *)&pdata->data;
  pdata->data2.domtype = data->domtype;
  pdata->data2.rngtype = data->rngtype;
  pdata->data2.comma = false;
  overflow = do_pmap_overflow(pdata);
  if (valPretty(x,data->domtype,data->doprint,data->printdata,0,data->lmargin,data->rmargin))
    {
      (void) freetmp (pdata);
      return true;
    }
  if (data->rngtype!=NIL)
    {
      if (overflow)
	if (CrTab(data->doprint,data->printdata,data->lmargin))
	  {
	    (void) freetmp (pdata);
	    return true;
	  }
      len = printMapsto(data->doprint,data->printdata);
      if (valPretty(y,data->rngtype,data->doprint,data->printdata,0,
		    overflow?(data->lmargin+len):data->lmargin,data->rmargin))
	{
	  (void) freetmp (pdata);
	  return true;
	}
    }
  (void) freetmp (pdata);
  return false;
}

PRIVATE Boolean mapPretty(val,domtype,rngtype,doprint,printdata,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr domtype,rngtype;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin,rmargin;
{
  struct printmap *data = (struct printmap *)alloctmp(sizeof(struct printmap));

  (void) printOpenBrace(doprint,printdata);
  data->doprint = doprint;
  data->printdata = printdata;
  data->domtype = domtype;
  data->rngtype = rngtype;
  data->comma = false;
  data->lmargin = lmargin+1;
  data->rmargin = rmargin-1;
  if (MapSweep(val,do_prettymap,data))
    {
      (void) freetmp (data);
      return true;
    }
  (void) freetmp (data);
  (void) printCloseBrace(doprint,printdata);
  return false;
}

PRIVATE Boolean dataPretty(val,tyname,argtypes,doprint,printdata,pri,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr tyname,argtypes;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
IN COUNTER lmargin,rmargin;
{
  ExprPtr tycons,tyhdr,cons,r,ttm,con,type;
  COUNTER len,newpri;
  SHORT magic;
  Boolean found;
#ifdef OLDDATA
  Boolean boxed;
#endif
  ENTER;

#ifdef OLDDATA
  con = getDataCon(val,tyname,&boxed);
#else
  tycons = STyCons(basisState);
  ttm = MapGet(tyname,TyTags(basisTable),&found);
  if (found && ttm->kind==E_STRING) /* special case of datatype with only one, non-constant
				       constructor. */
    {
      con = ttm;
    }
  else
    {
      SHORT magic;

      ttm = CAR(val);
      magic = val->magic;
      con = AVAL(CAR(ttm))->expr[magic-ttm->magic];
      val = CDR(val);
    }
#endif
  newpri = getPri(con);
  r = MapGet(tyname,tycons,&found);
  if (found)
    {
      tyhdr = CAR(r);
      cons = CDR(r);
      type = MapGetSure(con,cons);
#ifdef OLDDATA
      if (boxed && ISCONS(val)) /* boxed is true when the constructor is present,
				   including the case 'con' without an inner type */
	val = CDR(val);
#endif
      if (fxSym(APPLFUN(type))==funSym)
	{
	  if (pri<newpri)
	    {
	      (void) printOpenParen(doprint,printdata);
	      lmargin++;
	      rmargin--;
	    }
	  type = CAR(APPLARG(type));
	  type = ApplyType(type,tyhdr,argtypes,basisAssum,basisState);
	  if (newpri==0) /* not infix */
	    {
	      lmargin += printSym(con,doprint,printdata);
	      lmargin += printSpace(doprint,printdata);
	      if (valPretty(val,type,doprint,printdata,-1,lmargin,rmargin))
		return true;
	    }
	  else /* infix(left) if newpri odd, infix(right) if newpri even */
	    {
	      if (valPretty(CAR(val),CAR(type),doprint,printdata,newpri,lmargin,rmargin))
		return true;
	      (void) printSpace(doprint,printdata);
	      (void) printSym(con,doprint,printdata);
	      if (CrTab(doprint,printdata,lmargin))
		return true;
	      if (valPretty(CDR(val),CAR(CDR(type)),doprint,printdata,newpri-1,lmargin,rmargin))
		return true;
	    }
	  if (pri<newpri)
	    (void) printCloseParen(doprint,printdata);
	}
      else
	(void) printSym(con,doprint,printdata);
    }
  else /* abstype */
    {
      len = SLEN(tyname);
      {
	CHAR *s;

	s = STEXT(tyname);
	for (;;)
	  {
	    len--;
	    if (*s++=='%')
	      break;
	  }
	(void) printBeginSpecial(doprint,printdata);
	(*doprint)(s,len,printdata);
	(void) printEndSpecial(doprint,printdata);
      }
    }
  QUIT;
  return false;
}

PRIVATE Boolean recordPretty(val,type,doprint,printdata,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN CONST COUNTER lmargin,rmargin;
{
  (void) printOpenRecord(doprint,printdata);
  lmargin += 2;
  rmargin += 2;
  switch (val->kind)
    {
    case E_RECORD_1: /* then type is a maplet from lab<>REST_TAG to type */
      lmargin += printSym(CAR(val),doprint,printdata);
      lmargin += printEqual(doprint,printdata);
      if (valPretty(CDR(val),MAPLETY(type),doprint,printdata,0,lmargin,rmargin))
	return true;
      break;
    case E_RECORD_GNRL:
      {
	COUNTER n = AVAL(val)->n;
	ExprPtr *exprs = (ExprPtr *)alloctmp(sizeof(ExprPtr)*n);

	dumpShape(ASHAPE(val),exprs);
	{
	  REGISTER COUNTER i;
	  COUNTER len;
	  Boolean comma,found;
	  ExprPtr x,xtype;

	  comma = false;
	  for (i=0; i<n; i++)
	    {
	      if (prettyComma(doprint,printdata,&comma,lmargin))
		{
		  (void) freetmp (exprs);
		  return true;
		}
	      x = exprs[i];
	      len = printSym(x,doprint,printdata);
	      len += printEqual(doprint,printdata);
	      xtype = MapGet(x,type,&found);
	      if (!found)
		xtype = MapGetSure(REST_TAG,type);
	      if (valPretty(AVAL(val)->expr[i],xtype,doprint,printdata,0,lmargin+len,rmargin))
		{
		  (void) freetmp (exprs);
		  return true;
		}
	    }
	}
	(void) freetmp (exprs);
      }
      break;
    default:
      kinderror(message(87),val->kind);
    }
  (void) printCloseRecord(doprint,printdata);
  return false;
}

PRIVATE Boolean arrayPretty(val,type,doprint,printdata,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin,rmargin;
{
  ExprPtr *ep;
  COUNTER i,n;
  VecPtr vp;
  Boolean comma;
  
  switch (val->kind)
    {
    case E_ARRAY_2:
      if (valPretty(CAR(val),type,doprint,printdata,0,lmargin,rmargin) ||
	  prettyComma1(doprint,printdata,lmargin) ||
	  valPretty(CDR(val),type,doprint,printdata,0,lmargin,rmargin))
	return true;
      break;
    case E_ARRAY_GNRL:
      vp = AVAL(val);
      n = vp->n;
      ep = vp->expr;
      comma = false;
      for (i=0; i<n; i++)
	{
	  if (prettyComma(doprint,printdata,&comma,lmargin) ||
	      valPretty(*ep++,type,doprint,printdata,0,lmargin,rmargin))
	    return true;
	}
      break;
    default:
      kinderror(message(88),val->kind);
    }
  return false;
}

PRIVATE Boolean dynPrintOverflow(val,lmargin,rmargin,pri)
IN CONST ExprPtr val;
IN COUNTER lmargin,rmargin;
IN COUNTER pri;
{
  struct overflow *VOLATILE d = (struct overflow *)alloctmp(sizeof(struct overflow));
  Boolean res;

  d->lmargin = lmargin;
  d->rmargin = rmargin;
  if (SETJUMP(d->exit)==0)
    {
      dynPrint(val,printoverflow,(char *)d,pri);
      res = 0;
    }
  else res = 1;
#ifdef HAS_VOLATILE
  freetmp(d);
#endif
  return res;
}

PUBLIC Boolean dynPretty(val,doprint,printdata,lmargin,rmargin)
IN CONST ExprPtr val;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER lmargin,rmargin;
{
  Boolean cr;

  if (dynPrintOverflow(val,lmargin,rmargin,0))
    {
      if (valPretty(PACKVAL(val),PACKTYPE(val),doprint,printdata,0,lmargin,rmargin) ||
	  CrTab(doprint,printdata,lmargin))
	return true;
      lmargin += printOfType(doprint,printdata);
      fxReset();
      if (TypePretty(PACKTYPE(val),doprint,printdata,NIL,lmargin,rmargin) ||
	  fxFlushPretty(doprint,printdata,NIL,lmargin,rmargin,true,&cr))
	return true;
    }
  else
    dynPrint(val,doprint,printdata,0);
  return false;
}

PUBLIC Boolean valPretty(val,type,doprint,printdata,pri,lmargin,rmargin)
IN CONST ExprPtr val;
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN COUNTER pri;
IN COUNTER lmargin,rmargin;
{
  if (!checkTypeTag(val,type))
    valPrintNoType(val,doprint,printdata,pri);
  else
  /*if (type==NIL)
    valPrint(val,type,doprint,printdata,pri);
    else*/ if (printOverflow(val,type,lmargin,rmargin,pri))
      {
	if (lmargin>=rmargin-4) /* break infinite loops */
	  (void) printEtc(doprint,printdata);
	else
	  switch (type->kind)
	    {
	    case E_NIL:
	      valPrint(val,type,doprint,printdata,pri);
	      break;
	    case E_TYPE_APPL:
	      {
		ExprPtr f;
		
		f = APPLFUN(type);
		f = fxSym(f);
		if (f==stringSym)
		  {
		    if (stringPretty(STEXT(val),SLEN(val),doprint,printdata,lmargin,rmargin))
		      return true;
		  }
		else if (f==listSym)
		  {
		    if (listPretty(val,CAR(APPLARG(type)),doprint,printdata,lmargin,rmargin))
		      return true;
		  }
		else if (f==refSym)
		  {
		    if (pri<0)
		      {
			(void) printOpenParen(doprint,printdata);
			lmargin++;
			rmargin--;
		      }
		    lmargin += printRef(doprint,printdata);
		    if (valPretty(REFVAL(val),CAR(APPLARG(type)),doprint,printdata,
				  -1,lmargin,rmargin))
		      return true;
		    if (pri<0)
		      (void) printCloseParen(doprint,printdata);
		  }
		else if (f==arraySym)
		  {
		    if (pri<0)
		      (void) printOpenParen(doprint,printdata);
		    lmargin += printOpenArray(doprint,printdata);
		    rmargin--;
		    if (arrayPretty(val,CAR(APPLARG(type)),doprint,printdata,lmargin,rmargin))
		      return true;
		    (void) printCloseArray(doprint,printdata);
		    if (pri<0)
		      (void) printCloseParen(doprint,printdata);
		  }
		else if (f==intSym || f==funSym || f==boolSym || f==promiseSym
			 || f==tableSym)
		  valPrint(val,type,doprint,printdata,pri);
		else if (f==exnSym)
		  {
		    if (ISCONS(val))
		      {
			if (pri<0)
			  {
			    (void) printOpenParen(doprint,printdata);
			    lmargin++;
			    rmargin--;
			  }
			valPrintNoType(CAR(val),doprint,printdata,0);
			lmargin += symNameLen(REFVAL(CAR(val)));
			lmargin += printSpace(doprint,printdata);
			if (valPretty(CDR(val),getTypeInExn(CAR(val)),doprint,printdata,-1,lmargin,rmargin))
			  return true;
			if (pri<0)
			  (void) printCloseParen(doprint,printdata);
		      }
		    else
		      valPrintNoType(val,doprint,printdata,0);
		  }
		else if (f==mapSym)
		  {
		    if (mapPretty(val,CAR(APPLARG(type)),CAR(CDR(APPLARG(type))),
				  doprint,printdata,lmargin,rmargin))
		      return true;
		  }
		else if (f==dynSym)
		  {
		    lmargin += printDynamic(doprint,printdata);
		    rmargin--;
		    if (dynPretty(val,doprint,printdata,lmargin,rmargin))
		      return true;
		    (void) printCloseParen(doprint,printdata);
		  }
		else
		  if (dataPretty(val,f,APPLARG(type),doprint,printdata,pri,lmargin,rmargin))
		    return true;
		break;
	      }
	    case E_CONS:
	      if (tuplePretty(val,type,doprint,printdata,lmargin,rmargin))
		return true;
	      break;
	    case E_MAPLET:
	    case E_MAP:
	      if (recordPretty(val,type,doprint,printdata,lmargin,rmargin))
		return true;
	      break;
	    default:
	      valPrint(val,type,doprint,printdata,pri);
	      break;
	    }
      }
    else
      valPrint(val,type,doprint,printdata,pri);
  return false;
}

MAPSWEEPFUNC(do_valprint)
{
  struct valprint *data;
  ExprPtr i_l,val;
  COUNTER i,l,len;

  data = MAPDATA(struct valprint *);
  len = printSym(x,data->doprint,data->printdata);
  len += printEqual(data->doprint,data->printdata);
  i_l = MapGetSure(x,data->locs);
  if (i_l->kind==(CHAR)E_CONS)
    {
      i = IVAL(CAR(i_l));
      l = IVAL(CDR(i_l));
    }
  else
    {
      i = IVAL(i_l);
      l = 0;
    }
  val = getValue(i,l,basisEnv);
  if (val==NOEXPR)
    (void) printBuf("<undef>",data->doprint,data->printdata);
  else
    if (valPretty(val,y,data->doprint,data->printdata,0,len,
		  IVAL(REFVAL(rightMarginRef))))
      return true;
  return CrTab(data->doprint,data->printdata,0);
}

PUBLIC void valsPrint(assum,doprint,printdata)
IN CONST ExprPtr assum; /* type environment */
IN printFunc doprint;
INOUT VOIDPTR printdata;
{
  struct valprint *data = (struct valprint *)alloctmp(sizeof(struct valprint));
  COUNTER save_maxlines;
  ExprPtr ids;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = IVAL(REFVAL(maxLinesRef));
  if (crtab_maxlines>0)
    {
      ids = IdTypes(assum);
      data->doprint = doprint;
      data->printdata = (char *)printdata;
      data->locs = IdLocs(basisTable);
      (void) MapSweep(ids,do_valprint,data);
      endPretty(doprint,(char *)printdata);
    }
  crtab_maxlines = save_maxlines;
  (void) freetmp (data);
}

	/***** type printing *****/

PRIVATE void TypePrint1 P((IN CONST ExprPtr type,
			   IN printFunc doprint,
			   INOUT char *printdata,
			   IN CONST ExprPtr state,
			   IN COUNTER pri));

MAPSWEEPFUNC(do_dumprectype)
{
  field **tpp = MAPDATA(field **);

  (*tpp)->label = x;
  (*tpp)->val = y;
  (*tpp)++;
  return false;
}

PRIVATE int cmpstr(_m1,_m2)
IN CONST ANSI_CONST VOIDPTR _m1;
IN CONST ANSI_CONST VOIDPTR _m2;
{
  REGISTER field *m1 = (field *)_m1;
  REGISTER field *m2 = (field *)_m2;

  return (m1->label==REST_TAG?
	  (m2->label==REST_TAG?0:1):
	  (m2->label==REST_TAG?(-1):CompareStrings(m1->label,m2->label)));
}

PRIVATE void dumpRecTypes(rec,n,types)
IN CONST ExprPtr rec;
IN INT n;
OUT field *types;
{
  field *typep;

  typep = types;
  (void) MapSweep(rec,do_dumprectype,(CHAR *)&typep);
  qsort(types,(long)n,(long)sizeof(field),cmpstr);
}

MAPSWEEPFUNC(do_typeprintnum)
{
  struct typeprintmap *data;

  data = MAPDATA(struct typeprintmap *);
  (void) printDot(data->doprint,data->printdata,&data->comma);
  TypePrint1(x,data->doprint,data->printdata,data->state,0);
  if (y!=one)
    {
      (void) printPower(data->doprint,data->printdata);
      if ((y->kind==(CHAR)E_REAL && NEGR(XVAL(y))) ||
	  (y->kind==(CHAR)E_COMPLEX && NEGR(XVAL(REPART(y)))))
	(void) printSpace(data->doprint,data->printdata);
      valPrintNoType(y,data->doprint,data->printdata,0);
    }
  return false;
}

PUBLIC COUNTER namePrint(name,doprint,printdata)
IN CONST SymPtr name;
IN printFunc doprint;
INOUT char *printdata;
{
  INT len;
  
  len = symNameLen(name);
  {
    CHAR *s;

    s = symName(name);
    for (; len--, *s++!='%';);
    (*doprint)(s,(COUNTER)len,printdata);
  }
  return len;
}

PRIVATE void ellipsisPrint(type,doprint,printdata)
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
{
  (void) printEtc(doprint,printdata);
  (void) printOfType(doprint,printdata);
  TypePrint(type,doprint,printdata,NIL);
}

PRIVATE COUNTER printSet(doprint,printdata)
IN printFunc doprint;
INOUT char *printdata;
{
  return printBuf(" set",doprint,printdata);
}

PRIVATE void TypePrint1(type,doprint,printdata,state,pri)
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN COUNTER pri;
{
  if (state!=NIL)
    type = correctTypeAppl(type,state);
  /*if (type==NIL)
    (void) printBuf("unit",doprint,printdata);
    else*/ switch (type->kind)
      {
      case E_NIL:
	(void) printBuf("unit",doprint,printdata);
	break;
      case E_VAR:
	{
	  ExprPtr rename;
	  Boolean found;
	  
	  if (state!=NIL && (rename = nongenValue(type,state,&found), found))
	    TypePrint1(rename,doprint,printdata,state,pri);
	  else
	    {
	      /* if the type variables is of the form 'x?, but the current state
		 shows (through eqTypeTdg, accessing S_NONEQS) that it admits
		 equality, we add a tick in front, so that it prints ''x?.
		 This way, when coming across type errors due to violation of
		 equality constraints, the messages are clearer. Try:
		 fun f {x,y} = x (op @) y;
		 */
	      if (state!=NIL &&
		  symName(type)[1]!='\'' && symName(type)[symNameLen(type)-1]=='?' &&
		  eqTypeTdg(type,state,false))
		(*doprint)((CHAR *)"'",(COUNTER)1,printdata);
	      (void) printSym(type,doprint,printdata);
	    }
	}
	break;
      case E_TYPE_APPL:
	{
	  ExprPtr args,f,f0;
	  
	  f0 = APPLFUN(type); /* correct because of correctTypeAppl() */
	  f = fxSym(f0);
	  args = APPLARG(type);
	  if (f==mapSym && CAR(CDR(args))==NIL)
	    {
	      TypePrint1(CAR(args),doprint,printdata,state,3);
	      (void) printSet(doprint,printdata);
	      (void) fxPrintLatent(f0,doprint,printdata,state);
	    }
	  else if (f==funSym || f==mapSym)
	    {
	      if (pri>=1)
		(void) printOpenParen(doprint,printdata);
	      TypePrint1(CAR(args),doprint,printdata,state,1);
	      (void) namePrint(f,doprint,printdata);
	      (void) fxPrintLatent(f0,doprint,printdata,state);
	      (void) printSpace(doprint,printdata);
	      TypePrint1(CAR(CDR(args)),doprint,printdata,state,0);
	      if (pri>=1)
		(void) printCloseParen(doprint,printdata);
	    }
	  else
	    {
	      if (args!=NIL)
		{
		  if (CDR(args)==NIL)
		    {
		      TypePrint1(CAR(args),doprint,printdata,state,3);
		    }
		  else
		    {
		      (void) printOpenParen(doprint,printdata);
		      TypePrint1(CAR(args),doprint,printdata,state,0);
		      while (args=CDR(args), args!=NIL)
			{
			  (void) printComma1(doprint,printdata);
			  TypePrint1(CAR(args),doprint,printdata,state,0);
			}
		      (void) printCloseParen(doprint,printdata);
		    }
		  (void) printSpace(doprint,printdata);
		}
	      (void) namePrint(f,doprint,printdata);
	      (void) fxPrintLatent(f0,doprint,printdata,state);
	    }
	  break;
	}
      case E_CONS:
	if (pri>=2)
	  (void) printOpenParen(doprint,printdata);
	for (; /*type!=NIL && */type->kind==(CHAR)E_CONS; type=CDR(type))
	  {
	    TypePrint1(CAR(type),doprint,printdata,state,2);
	    if (CDR(type)!=NIL)
	      (void) printBuf(" * ",doprint,printdata);
	  }
	if (type!=NIL)
	  ellipsisPrint(type,doprint,printdata);
	if (pri>=2)
	  (void) printCloseParen(doprint,printdata);
	break;
      case E_EMPTY:
	(void) printRecord0(doprint,printdata);
	break;
      case E_MAPLET:
	(void) printOpenRecord(doprint,printdata);
	if (MAPLETX(type)==REST_TAG)
	  ellipsisPrint(MAPLETY(type),doprint,printdata);
	else
	  {
	    (void) printSym(MAPLETX(type),doprint,printdata);
	    (void) printOfType(doprint,printdata);
	    TypePrint1(MAPLETY(type),doprint,printdata,state,0);
	  }
	(void) printCloseRecord(doprint,printdata);
	break;
      case E_MAP:
	{
	  INT n = MapCard(type);
	  Boolean comma;
	  field *types0 = (field *)alloctmp(sizeof(field)*n);
	  field *types = types0;

	  dumpRecTypes(type,n,types);
	  (void) printOpenRecord(doprint,printdata);
	  comma = false;
	  for (; n--!=0; types++)
	    {
	      (void) printComma(doprint,printdata,&comma);
	      if (types->label==REST_TAG)
		(void) printEtc(doprint,printdata);
	      else
		(void) printSym(types->label,doprint,printdata);
	      (void) printOfType(doprint,printdata);
	      TypePrint1(types->val,doprint,printdata,state,0);
	    }
	  (void) freetmp (types0);
	  (void) printCloseRecord(doprint,printdata);
	  break;
	}
      case E_NUMTYPE:
	{
	  struct typeprintmap *data = (struct typeprintmap *)alloctmp(sizeof(struct typeprintmap));
	  
	  type = REFVAL(type);
	  if (type==NewMap())
	    (void) printBuf("num",doprint,printdata);
	  else if (type->kind==(CHAR)E_MAPLET && MAPLETY(type)==one)
	    TypePrint1(MAPLETX(type),doprint,printdata,state,pri);
	  else
	    {
	      if (pri>=3)
		(void) printOpenParen(doprint,printdata);
	      data->doprint = doprint;
	      data->printdata = printdata;
	      data->state = state;
	      data->comma = false;
	      (void) MapSweep(type,do_typeprintnum,data);
	      if (pri>=3)
		(void) printCloseParen(doprint,printdata);
	    }
	  (void) freetmp (data);
	  break;
	}
      default:
	kinderror(message(89),type->kind);
	break;
      }
}

PUBLIC void TypePrint(type,doprint,printdata,state)
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  TypePrint1(type,doprint,printdata,state,0);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE void printUnion(doprint,printdata,commap)
IN printFunc doprint;
INOUT char *printdata;
INOUT Boolean *commap;
{
  if (*commap)
    {
      (void) printUnion1(doprint,printdata);
      (void) printSpace(doprint,printdata);
    }
  else
    *commap = true;
}

MAPSWEEPFUNC(do_typeprintfx)
{
  struct typeprintmap *data;

  data = MAPDATA(struct typeprintmap *);
  printUnion(data->doprint,data->printdata,&data->comma);
  if (x->kind==(CHAR)E_ARRAY_2)
    (void) fxPrintVar(x,data->doprint,data->printdata,data->state);
  else
    TypePrint1(x,data->doprint,data->printdata,data->state,0);
  return false;
}

PRIVATE void FxPrint1(constr,doprint,printdata,state,pri)
IN CONST ExprPtr constr;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN COUNTER pri;
{
  struct typeprintmap *data;
  
  if (constr==NewMap())
    (void) printPure(doprint,printdata);
  else if (constr->kind==(CHAR)E_MAPLET)
    {
      if (MAPLETX(constr)->kind==(CHAR)E_ARRAY_2)
	(void) fxPrintVar(MAPLETX(constr),doprint,printdata,state);
      else
	TypePrint1(MAPLETX(constr),doprint,printdata,state,pri);
    }
  else
    {
      if (pri>=3)
	(void) printOpenParen(doprint,printdata);
      data = (struct typeprintmap *)alloctmp(sizeof(struct typeprintmap));
      data->doprint = doprint;
      data->printdata = printdata;
      data->state = state;
      data->comma = false;
      (void) MapSweep(constr,do_typeprintfx,data);
      (void) freetmp (data);
      if (pri>=3)
	(void) printCloseParen(doprint,printdata);
    }
}

PUBLIC void fxFlushPrint(doprint,printdata,state,comma)
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN Boolean comma;
{
  ExprPtr fx;

  while (fxPrint.toshow!=NewMap())
    {
      fx = ChooseInDom(fxPrint.toshow);
      fxPrint.toshow = SetRemove(fx,fxPrint.toshow);
      if (CDR(fx)!=NewMap())
	{
	  (void) printComma(doprint,printdata,&comma);
	  (void) fxPrintVar(fx,doprint,printdata,state);
	  (void) printGeq(doprint,printdata);
	  FxPrint1(CDR(fx),doprint,printdata,state,0);
	}
    }
}

PUBLIC void FxPrint(constr,doprint,printdata,state)
IN CONST ExprPtr constr;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  FxPrint1(constr,doprint,printdata,state,0);
}
#endif

PRIVATE void nilprint(text,len,data)
IN CONST CHAR *text;
IN COUNTER len;
INOUT char *data;
{
}

PRIVATE Boolean printTypeOverflow(type,lmargin,rmargin,state,pri)
IN CONST ExprPtr type;
IN COUNTER lmargin,rmargin;
IN CONST ExprPtr state;
IN COUNTER pri;
{
  struct overflow *VOLATILE data = (struct overflow *)alloctmp(sizeof(struct overflow));
  Boolean res;

  data->lmargin = lmargin;
  data->rmargin = rmargin;
  if (SETJUMP(data->exit)==0)
    {
      TypePrint1(type,printoverflow,(char *)data,state,pri);
      res = 0;
    }
  else res = 1;
#ifdef HAS_VOLATILE
  freetmp(data);
#endif
  return res;
}

PRIVATE Boolean TypePretty1 P((IN CONST ExprPtr type,
			       IN printFunc doprint,
			       INOUT char *printdata,
			       IN CONST ExprPtr state,
			       IN COUNTER pri,
			       IN COUNTER lmargin,
			       IN COUNTER rmargin));

PRIVATE Boolean TypePretty1 P((IN CONST ExprPtr type,IN printFunc doprint,
			       INOUT char *printdata,IN CONST ExprPtr state,
			       IN COUNTER pri,IN COUNTER lmargin,IN COUNTER rmargin));

PRIVATE Boolean TypePretty1(type,doprint,printdata,state,pri,lmargin,rmargin)
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN COUNTER pri;
IN COUNTER lmargin,rmargin;
{
  if (state!=NIL)
    type = correctTypeAppl(type,state);
  if (lmargin>=rmargin-10)
    rmargin = lmargin+20;
  /*if (type==NIL)
    TypePrint1(type,doprint,printdata,state,pri);
    else*/ if (printTypeOverflow(type,lmargin,rmargin,state,pri))
      switch (type->kind)
	{
	case E_NIL:
	  TypePrint1(type,doprint,printdata,state,pri);
	  break;
	case E_TYPE_APPL:
	  {
	    ExprPtr args,f,f0;
	    
	    f0 = APPLFUN(type); /* correct because of correctTypeAppl() */
	    f = fxSym(f0);
	    args = APPLARG(type);
	    if (f==mapSym && CAR(CDR(args))==NIL)
	      {
		if (TypePretty1(CAR(args),doprint,printdata,state,3,lmargin,rmargin-4))
		  return true;
		(void) printSet(doprint,printdata);
		(void) fxPrintLatent(f0,doprint,printdata,state);
	      }
	    else if (f==funSym || f==mapSym)
	      {
		if (pri>=1)
		  {
		    (void) printOpenParen(doprint,printdata);
		    lmargin++;
		    rmargin--;
		  }
		if (TypePretty1(CAR(args),doprint,printdata,state,1,lmargin,rmargin) ||
		    CrTab(doprint,printdata,lmargin))
		  return true;
		lmargin += namePrint(f,doprint,printdata);
		lmargin += fxPrintLatent(f0,doprint,printdata,state);
		lmargin += printSpace(doprint,printdata);
		if (TypePretty1(CAR(CDR(args)),doprint,printdata,state,0,lmargin,rmargin))
		  return true;
		if (pri>=1)
		  (void) printCloseParen(doprint,printdata);
	      }
	    else
	      {
		if (args!=NIL)
		  {
		    rmargin -= namePrint(f,nilprint,(char *)0L);
		    if (CDR(args)==NIL)
		      {
			rmargin--;
			if (TypePretty1(CAR(args),doprint,printdata,state,3,lmargin,rmargin))
			  return true;
		      }
		    else
		      {
			(void) printOpenParen(doprint,printdata);
			lmargin++;
			/*rmargin--;*/
			TypePrint1(CAR(args),doprint,printdata,state,3);
			while (args=CDR(args), args!=NIL)
			  {
			    if (prettyComma1(doprint,printdata,lmargin))
			      return true;
			    TypePrint1(CAR(args),doprint,printdata,state,0);
			  }
			(void) printCloseParen(doprint,printdata);
		      }
		    (void) printSpace(doprint,printdata);
		  }
		(void) namePrint(f,doprint,printdata);
		(void) fxPrintLatent(f0,doprint,printdata,state);
	      }
	    break;
	  }
	case E_CONS:
	  if (pri>=2)
	    {
	      (void) printOpenParen(doprint,printdata);
	      lmargin++;
	      rmargin--;
	    }
	  (void) printBuf("  ",doprint,printdata);
	  for (; /*type!=NIL && */type->kind==(CHAR)E_CONS; type=CDR(type))
	    {
	      if (TypePretty1(CAR(type),doprint,printdata,state,2,lmargin+2,rmargin))
		return true;
	      if (CDR(type)!=NIL)
		{
		  if (CrTab(doprint,printdata,lmargin))
		    return true;
		  (void) printBuf("* ",doprint,printdata);
		}
	    }
	  if (type!=NIL)
	    ellipsisPrint(type,doprint,printdata);
	  if (pri>=2)
	    (void) printCloseParen(doprint,printdata);
	  break;
	case E_MAPLET:
	  (void) printOpenRecord(doprint,printdata);
	  lmargin += 2;
	  rmargin -= 2;
	  if (MAPLETX(type)==REST_TAG)
	    ellipsisPrint(MAPLETY(type),doprint,printdata);
	  else
	    {
	      lmargin += printSym(MAPLETX(type),doprint,printdata);
	      lmargin += printOfType(doprint,printdata);
	      if (TypePretty1(MAPLETY(type),doprint,printdata,state,0,lmargin,rmargin))
		return true;
	    }
	  (void) printCloseRecord(doprint,printdata);
	  break;
	case E_MAP:
	  {
	    INT n = MapCard(type);
	    COUNTER len;
	    Boolean comma;
	    field *types0 = (field *)alloctmp(sizeof(field)*n);
	    field *types = types0;

	    dumpRecTypes(type,n,types);
	    (void) printOpenRecord(doprint,printdata);
	    lmargin += 2;
	    rmargin -= 2;
	    comma = false;
	    for (; n--!=0; types++)
	      {
		if (prettyComma(doprint,printdata,&comma,lmargin))
		  return true;
		if (types->label==REST_TAG)
		  len = printEtc(doprint,printdata);
		else
		  len = printSym(types->label,doprint,printdata);
		len += printOfType(doprint,printdata);
		if (TypePretty1(types->val,doprint,printdata,state,0,lmargin+len,rmargin))
		  return true;
	      }
	    (void) freetmp (types0);
	    (void) printCloseRecord(doprint,printdata);
	    break;
	  }
	default:
	  TypePrint1(type,doprint,printdata,state,pri);
	  break;
	}
    else
      TypePrint1(type,doprint,printdata,state,pri);
  return false;
}

PUBLIC Boolean TypePretty(type,doprint,printdata,state,lmargin,rmargin)
IN CONST ExprPtr type;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN COUNTER lmargin,rmargin;
{
  return TypePretty1(type,doprint,printdata,state,0,lmargin,rmargin);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE Boolean prettyUnion(doprint,printdata,commap,lmargin)
IN printFunc doprint;
INOUT char *printdata;
INOUT Boolean *commap;
IN COUNTER lmargin;
{
  if (*commap)
    {
      (void) printUnion1(doprint,printdata);
      return CrTab(doprint,printdata,lmargin);
    }
  *commap = true;
  return false;
}

PRIVATE Boolean printFxOverflow(constr,lmargin,rmargin,state,pri)
IN CONST ExprPtr constr;
IN COUNTER lmargin,rmargin;
IN CONST ExprPtr state;
IN COUNTER pri;
{
  struct overflow *VOLATILE data = (struct overflow *)alloctmp(sizeof(struct overflow));
  Boolean res;

  data->lmargin = lmargin;
  data->rmargin = rmargin;
  if (SETJUMP(data->exit)==0)
    {
      FxPrint1(constr,printoverflow,(char *)data,state,pri);
      res = 0;
    }
  else res = 1;
#ifdef HAS_VOLATILE
  (void) freetmp (data);
#endif
  return res;
}

MAPSWEEPFUNC(do_typeprettyfx)
{
  struct typeprintmap *data;
  COUNTER len;

  data = MAPDATA(struct typeprintmap *);
  if (prettyUnion(data->doprint,data->printdata,&data->comma,data->lmargin))
    return true;
  if (x->kind==(CHAR)E_ARRAY_2)
    (void) fxPrintVar(x,data->doprint,data->printdata,data->state);
  else
    return TypePretty1(x,data->doprint,data->printdata,data->state,0,
		       data->lmargin,data->rmargin);
  return false;
}

PRIVATE Boolean FxPretty1(constr,doprint,printdata,state,pri,lmargin,rmargin)
IN CONST ExprPtr constr;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN COUNTER pri;
IN COUNTER lmargin,rmargin;
{
  struct typeprintmap *data;

  if (!printFxOverflow(constr,lmargin,rmargin,state,pri))
    FxPrint1(constr,doprint,printdata,state,pri);
  else if (constr==NewMap())
    (void) printPure(doprint,printdata);
  else if (constr->kind==(CHAR)E_MAPLET)
    return TypePretty1(MAPLETX(constr),doprint,printdata,state,pri,lmargin,rmargin);
  else
    {
      if (pri>=3)
	(void) printOpenParen(doprint,printdata);
      lmargin++;
      rmargin -= 2;
      data = (struct typeprintmap *)alloctmp(sizeof(struct typeprintmap));
      data->doprint = doprint;
      data->printdata = printdata;
      data->state = state;
      data->comma = false;
      data->lmargin = lmargin;
      data->rmargin = rmargin;
      (void) MapSweep(constr,do_typeprettyfx,data);
      (void) freetmp (data);
      if (pri>=3)
	(void) printCloseParen(doprint,printdata);
    }
  return false;
}

PUBLIC Boolean fxFlushPretty(doprint,printdata,state,lmargin,rmargin,comma,crp)
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
IN COUNTER lmargin,rmargin;
IN Boolean comma;
OUT Boolean *crp;
{
  ExprPtr fx;

  *crp = false;
  while (fxPrint.toshow!=NewMap())
    {
      fx = ChooseInDom(fxPrint.toshow);
      fxPrint.toshow = SetRemove(fx,fxPrint.toshow);
      if (CDR(fx)!=NewMap())
	{
	  if (prettyComma(doprint,printdata,&comma,lmargin))
	    return true;
	  lmargin += fxPrintVar(fx,doprint,printdata,state);
	  lmargin += printGeq(doprint,printdata);
	  if (FxPretty1(CDR(fx),doprint,printdata,state,0,lmargin,rmargin))
	    return true;
	  *crp = true;
	}
    }
  return false;
}

PUBLIC Boolean fxFlush(doprint,printdata,state)
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  Boolean cr;

  if (fxFlushPretty(doprint,printdata,state,0,
		    IVAL(REFVAL(rightMarginRef)),false,&cr))
    return true;
  return cr && CrTab(doprint,printdata,0);
}

PUBLIC Boolean FxPretty(constr,doprint,printdata,state)
IN CONST ExprPtr constr;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  return FxPretty1(constr,doprint,printdata,state,0,0,
		   IVAL(REFVAL(rightMarginRef))) ||
	 CrTab(doprint,printdata,0);
}
#endif

MAPSWEEPFUNC(do_idsprint)
{
  struct typeprintmap *data;
  COUNTER len;

  data = MAPDATA(struct typeprintmap *);
  len = printSym(x,data->doprint,data->printdata);
  len += printOfType(data->doprint,data->printdata);
  return TypePretty(y,data->doprint,data->printdata,data->state,len,
		    IVAL(REFVAL(rightMarginRef))) ||
    CrTab(data->doprint,data->printdata,0);
}

PRIVATE Boolean idsPrint(ids,doprint,printdata,state)
IN CONST MapPtr ids;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  struct typeprintmap *data = (struct typeprintmap *)alloctmp(sizeof(struct typeprintmap));

  data->doprint = doprint;
  data->printdata = printdata;
  data->state = state;
  {
    Boolean res = MapSweep(ids,do_idsprint,data);

    (void) freetmp (data);
    return res;
  }
}

MAPSWEEPFUNC(do_tyfcnsprint)
{
  struct typeprintmap *data;
  ExprPtr tyvars,tyvar,type;
  COUNTER lmargin;
  Boolean need_paren, comma;

  data = MAPDATA(struct typeprintmap *);
  tyvars = CAR(CAR(y));
  type = CDR(y);
  lmargin = printBuf("type ",data->doprint,data->printdata);
  need_paren = tyvars!=NIL && CDR(tyvars)!=NIL;
  if (need_paren)
    lmargin += printOpenParen(data->doprint,data->printdata);
  comma = false;
  for (;tyvars!=NIL; tyvars=CDR(tyvars))
    {
      lmargin += printComma(data->doprint,data->printdata,&comma);
      tyvar = CAR(tyvars);
      lmargin += printSym(tyvar,data->doprint,data->printdata);
      /*lmargin += printSpace(data->doprint,data->printdata);*/
    }
  if (need_paren)
    lmargin += printCloseParen(data->doprint,data->printdata);
  if (CAR(CAR(y))!=NIL)
    lmargin += printSpace(data->doprint,data->printdata);
  lmargin += printSym(x,data->doprint,data->printdata);
  lmargin += printEqual(data->doprint,data->printdata);
  return TypePretty(type,data->doprint,data->printdata,data->state,lmargin,
		    IVAL(REFVAL(rightMarginRef))) ||
	 CrTab(data->doprint,data->printdata,0);
}

PRIVATE Boolean tyfcnsPrint(tyfcns,doprint,printdata,state)
IN CONST MapPtr tyfcns;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  struct typeprintmap *data = (struct typeprintmap *)alloctmp(sizeof(struct typeprintmap));

  data->doprint = doprint;
  data->printdata = printdata;
  data->state = state;
  {
    Boolean res = MapSweep(tyfcns,do_tyfcnsprint,data);

    (void) freetmp (data);
    return res;
  }
}

MAPSWEEPFUNC(do_unitsprint)
{
  struct typeprintmap *data;
  COUNTER len;

  data = MAPDATA(struct typeprintmap *);
  len = printBuf("scale ",data->doprint,data->printdata);
  len += printSym(x,data->doprint,data->printdata);
  len += printOfType(data->doprint,data->printdata);
  return TypePretty(CDR(y),data->doprint,data->printdata,data->state,len,
		    IVAL(REFVAL(rightMarginRef))) ||
	 CrTab(data->doprint,data->printdata,0);
}

PRIVATE Boolean unitsPrint(units,doprint,printdata,state)
IN CONST MapPtr units;
IN printFunc doprint;
INOUT char *printdata;
IN CONST ExprPtr state;
{
  struct typeprintmap *data = (struct typeprintmap *)alloctmp(sizeof(struct typeprintmap));

  data->doprint = doprint;
  data->printdata = printdata;
  data->state = state;
  {
    Boolean res = MapSweep(units,do_unitsprint,data);

    (void) freetmp (data);
    return res;
  }
}

PUBLIC Boolean typeEnvPrint(assum,doprint,printdata,state)
IN CONST ExprPtr assum;
IN printFunc doprint;
INOUT VOIDPTR printdata;
IN CONST ExprPtr state;
{
#if MODULE_SYSTEM & STD_ML_MODULE
  /* !!! print A_STRUCTS(assum) and A_SIGS(assum) */
#endif
  return (tyfcnsPrint(A_TYFCNS(assum),doprint,(char *)printdata,state) ||
	  unitsPrint(A_UNITS(assum),doprint,(char *)printdata,state) ||
	  idsPrint(A_IDS(assum),doprint,(char *)printdata,state));
}

PRINTFUNC(do_printstring)
{
  struct printstring *data;
  COUNTER len2;

  data = PRINTDATA(struct printstring *);
  if (data->len<data->maxlen)
    {
      len2 = data->maxlen-data->len;
      if (len>len2)
	len = data->maxlen-data->len;
      safememcpy((char *)data->buf+data->len,(char *)text,(long)len);
      data->len += len;
    }
}

#define PRINTBUFSIZE 4096
#define PRINTBUFSLOP 5

PRIVATE void terminateString(data)
INOUT struct printstring *data;
{
  CHAR *s;

  if (data->len>=data->maxlen)
    {
      s = (CHAR *)message(462);
      data->len = data->maxlen;
      data->maxlen += PRINTBUFSLOP;
      do_printstring(s,(COUNTER)strlen((char *)s)+1,(char *)data); /* include the NUL character */
      data->len--; /* to allow overwriting NUL */
    }
  else
    {
      data->buf[data->len] = 0;
    }
}

PRIVATE void crString(data)
INOUT struct printstring *data;
{
  CHAR c = '\n';

  do_printstring(&c,1,(char *)data);
  data->buf[data->len] = 0;
}

PUBLIC void incompatibleTypes(zone,type1,type2,msg1,msg2,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr type1,type2;
IN CHAR *msg1, *msg2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  struct printstring *data;
  COUNTER save_maxlines;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = MAX_SIGNED_INT;
  {
    CHAR *buf;

    buf = (CHAR *)alloctmp_anchored(PRINTBUFSIZE, (char *)&zone);
    data = (struct printstring *)alloctmp_anchored(sizeof(struct printstring), (char *)&zone);
    data->len = 0;
    data->buf = buf;
    type1 = developType(type1,state);
    type2 = developType(type2,state);
    data->maxlen = (PRINTBUFSIZE/2)-PRINTBUFSLOP;
    do_printstring(msg1,(COUNTER)strlen((char *)msg1),(char *)data);
    (void) TypePretty(type1,do_printstring,
		      (char *)data,state,0,
		      IVAL(REFVAL(rightMarginRef)));
    terminateString(data);
    crString(data);
    data->maxlen = PRINTBUFSIZE-PRINTBUFSLOP;
    do_printstring(msg2,(COUNTER)strlen((char *)msg2),(char *)data);
    (void) TypePretty(type2,do_printstring,
		      (char *)data,state,0,
		      IVAL(REFVAL(rightMarginRef)));
    terminateString(data);
    crtab_maxlines = save_maxlines;
    type_soft_error(zone,buf); /* the text is actually NUL-terminated */
    (void) freetmp (buf);
  }
}

PUBLIC void typeError(zone,msg,type,state)
IN CONST TZone *zone;
IN CONST CHAR *msg;
IN CONST ExprPtr type;
INOUT ExprPtr state;
{
  struct printstring *data = (struct printstring *)alloctmp_anchored(sizeof(struct printstring), (char *)&zone);
  CHAR *s;
  COUNTER save_maxlines;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = MAX_SIGNED_INT;
  data->len = 0;
  {
    CHAR *buf;

    buf = (CHAR *)alloctmp_anchored(PRINTBUFSIZE, (char *)&zone);
    data->maxlen = PRINTBUFSIZE-PRINTBUFSLOP;
    data->buf = buf;
    type = developType(type,state);
    do_printstring(msg,(COUNTER)strlen((char *)msg),(char *)data);
    s = (CHAR *)message(31);
    do_printstring(s,(COUNTER)strlen((char *)s),(char *)data);
    (void) TypePretty(type,do_printstring,(char *)data,state,data->len,
		      IVAL(REFVAL(rightMarginRef)));
    terminateString(data);
    crString(data);
    crtab_maxlines = save_maxlines;
    type_soft_error(zone,buf); /* text is actually NUL-terminated */
    (void) freetmp (data);
  }
}
