/* Profiler.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "type.h"
#include "compile.h"
#include "gram.h"
#include "input.h"
#include "print.h"
#include "bc.h"
#include "rbci.h"
#include "dbgproto.h"
#include "grmproto.h"
#include "evlproto.h"
#include "prtproto.h"
#include "typproto.h"
#include "exeproto.h"
#include "cmpproto.h"
#include "errproto.h"
#include "stkproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sprofile.h"
#endif

#if PROFILING
EXTERN ExprPtr voidString;
EXTERN ExprPtr profileExprs;
EXTERN ExprPtr ncallsSym,properSym,totalSym,ngcSym,gctimeSym,timeSym,locSym;

PRIVATE void profSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  Profile *prof;

  prof = (Profile *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      /* We don't save anything in Profiles. */
      break;
    case X_MARK:
    case X_MARKDATA:
      break;
    case X_FREE:
      MARKFREE(prof);
      break;
    }
}

PRIVATE ExprPtr profLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  Profile *prof;
  ExprPtr e;
  
  prof = (Profile *)MALLOC(sizeof(Profile));
  resetProfile(prof);
  e = MakeExternal(desc,(char *)prof);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc profExtDesc = {
  (CHAR *)"profile",
  { 'p','r','o','f' }, NULL,
  NULL, /* no hash function, profiles are not shared */
  NULL, /* no comparison function, profiles are not shared */
  NULL, /* no print function */
  profSweep,
  profLoad
};

PUBLIC ExprPtr profAlloc()
{
  Profile *prof;
  ExprPtr e;
  
  prof = (Profile *)MALLOC(sizeof(Profile));
  resetProfile(prof);
  e = MakeExternal(&profExtDesc,(char *)prof);
  (void) KEEP(e);
  return e;
}

PUBLIC void resetProfile(prof)
INOUT Profile *prof;
{
  prof->nCalls = 0;
  prof->stat.total.time.tms_utime = prof->stat.total.time.tms_stime =
    prof->stat.total.gc_time.tms_utime = prof->stat.total.gc_time.tms_stime =
      prof->stat.total.gc_count = 0;
  prof->stat.proper.time.tms_utime = prof->stat.proper.time.tms_stime =
    prof->stat.proper.gc_time.tms_utime = prof->stat.proper.gc_time.tms_stime =
    prof->stat.proper.gc_count = 0;
  prof->nEntries = 0;
}

DEF_ACTION_0ARG(reset_profiles)
{
  ExprPtr ll;

  for (ll = profileExprs; ll!=NIL; ll = CDR(ll))
    {
      resetProfile((Profile *)EXTDATA(CAR(CDR(ARGS(CAR(ll))))));
    }
  return tuple0;
}
ACTION_END

DEF_ACTION_0ARG(clear_profiles)
{
  ExprPtr ll,info;

  for (ll = profileExprs; ll!=NIL; ll = CDR(ll))
    {
      TYPE(CAR(ll)) = (CHAR)EA_CLR_PROFILE; /* disable profiling here */
      info = CDR(ARGS(CAR(ll)));
      SETCAR(info,NIL); /* erase Profile */
    }
  profileExprs = NIL;
  return tuple0;
}
ACTION_END

PRIVATE ExprPtr mkZone(zone,funcname,filename)
IN CONST ExprPtr zone;
IN CONST ExprPtr funcname,filename;
{
  INT l1,l2,p1,p2;
  ExprPtr e;
  ExprPtr *buf;
  ENTER;

  l1 = ZONE(zone)->line1+1;
  l2 = ZONE(zone)->line2+1;
  p1 = ZONE(zone)->pos1;
  p2 = ZONE(zone)->pos2;
  buf = (ExprPtr *)alloctmp(4*sizeof(ExprPtr));
  buf[0] = funcname;
  buf[1] = filename;
  if (buf[1]==NIL)
    buf[1] = voidString;
  buf[2] = Ktuple2(KINT(l1),KINT(p1));
  (void) KEEP(buf[2]);
  buf[3] = Ktuple2(KINT(l2),KINT(p2));
  (void) KEEP(buf[3]);
  e = MakeRecord(E_TUPLE_GNRL,NIL,4,buf);
  KEEPONLY(e);
  return e;
}

#define report1proftms(tms) report1tms(tms)

PRIVATE ExprPtr report1tms(tms)
IN CONST struct h_tms *tms;
{
  REAL u,s;
  ExprPtr auxu,auxs,e;

  u = ((EXTENDED)tms->tms_utime)/((EXTENDED)HERTZ);
  s = ((EXTENDED)tms->tms_stime)/((EXTENDED)HERTZ);
  auxu = Kreal(&u);
  (void) KEEP(auxu);
  auxs = Kreal(&s);
  (void) KEEP(auxs);
  e = Ktuple2(auxu,auxs);
  return e;
}

PRIVATE ExprPtr report1prof1rec(rec)
IN CONST Prof1Rec *rec;
{
  field fields[3];
  ExprPtr e;
  ENTER;

  fields[0].label = gctimeSym;
  fields[0].val = report1tms(&rec->gc_time);
  (void) KEEP (fields[0].val);
  fields[1].label = ngcSym;
  fields[1].val = KINT(rec->gc_count);
  fields[2].label = timeSym;
  fields[2].val = report1proftms(&rec->time);
  (void) KEEP (fields[2].val);
  sortFieldArray(3,fields); /* we could sort it in advance */
  e = MakeRecordGnrl(3,fields,fieldsToShape(3,fields));
  QUIT;
  return e;
}

PRIVATE ExprPtr report1prof(zone,funcname,filename,prof)
IN CONST ExprPtr zone;
IN CONST ExprPtr funcname,filename;
IN CONST Profile *prof;
{
  field fields[4];
  ExprPtr e;
  ENTER;

  fields[0].label = locSym;
  fields[0].val = mkZone(zone,funcname,filename);
  fields[1].label = ncallsSym;
  fields[1].val = KINT(prof->nCalls);
  fields[2].label = properSym;
  fields[2].val = report1prof1rec(&prof->stat.proper);
  (void) KEEP (fields[2].val);
  fields[3].label = totalSym;
  fields[3].val = report1prof1rec(&prof->stat.total);
  (void) KEEP (fields[3].val);
  sortFieldArray(4,fields); /* we could sort it in advance */
  e = MakeRecordGnrl(4,fields,fieldsToShape(4,fields));
  QUIT;
  return e;
}

DEF_ACTION_0ARG(report_profiles)
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;
  ExprPtr ll,prof;
  ExprPtr args,funcname,filename;
  Profile *profile;

  _x.n = 2;
  _x.next = theStackFrames;
  _x.p[0] = NewMap();
  _x.p[1] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;
  for (ll = profileExprs; ll!=NIL; ll = CDR(ll))
    {
      prof = CAR(ll);
      args = ARGS(prof);
      funcname = CAR(args);
      args = CDR(args);
      if (CAR(args)==NIL || (profile = (Profile *)EXTDATA(CAR(args)),
			     profile->nCalls==0))
	continue; /* don't report functions that have never been called */
      filename = CDR(args);
      _x.p[1] = report1prof(prof,funcname,filename,profile);
      _x.p[0] = MapAdd(_x.p[1],UNIT,_x.p[0]);
    }
  theStackFrames = _x.next;
  return _x.p[0];
}
ACTION_END
#endif
