/* GimML regexp functions.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#undef NFA_DEBUG

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <time.h>
#include <limits.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "transfer.h"
#include "type.h"
#include "exeproto.h"
#include "re.h"

#define RE_MALLOC(sz) (void *)MALLOC(sz)
#define RE_REALLOC(p,sz) REALLOC(p,sz)
#define RE_FREE(p) MARKFREE(p)

EXTERN ExprPtr reException, stringnthException, subscriptException;
EXTERN ExprPtr none,optttm;
EXTERN extDesc iarrayExtDesc;
EXTERN ExprPtr voidString;

struct regex_s {
  char kind;
#define RE_EPSILON 0 /* empty word */
#define RE_CHAR 1 /* single letter */
#define RE_RANGE 2 /* [...] ranges */
#define RE_CONCAT 3 /* concatenation */
#define RE_CHOICE 4 /* | */
#define RE_STAR 5 /* * */
#define RE_PLUS 6 /* + */
#define RE_GROUP 7 /* (...) groups */
#define RE_CATCH_ALL 8 /* .* */
  char nullable;
  union {
    /* nothing: RE_EPSILON, RE_CATCH_ALL */
    char letter; /* RE_CHAR */
    char *range; /* RE_RANGE: 33 byte array (bitfields); we need 32 bytes to
		    represent a bitfield of 256 bits, one for each char;
		    The last byte contains fields for 'special characters';
		    for now, only 'end of string', encoded as character #256.
		 */
    struct pair_s { struct regex_s *left, *right; } pair; /* RE_CONCAT, RE_CHOICE */
    struct regex_s *sub; /* RE_STAR, RE_PLUS */
    struct re_group_s { struct regex_s *sub; int n; } group; /* RE_GROUP */
  } what;
};

#define SET(bitfield,i) (bitfield)[(i)>>3] |= (1 << ((i) & 0x7))
#define CLR(bitfield,i) (bitfield)[(i)>>3] &= ~(1 << ((i) & 0x7))
#define TEST(bitfield,i) ((bitfield)[(i)>>3] & (1 << ((i) & 0x7)))

void re_free (regex_t *re)
{
  if (re==NULL)
    return;
  switch (re->kind)
    {
    case RE_CHAR:
      //fprintf(stderr, "free char '%c'\n", re->what.letter);
      break;
    case RE_CATCH_ALL:
      //fprintf (stderr, "free catch all\n");
      break;
    case RE_RANGE:
      //fprintf (stderr, "free range\n");
      RE_FREE(re->what.range);
      break;
    case RE_CONCAT: case RE_CHOICE:
      //fprintf (stderr, "free concat or choice (\n");
      re_free (re->what.pair.left);
      re_free (re->what.pair.right);
      //fprintf (stderr, ")\n");
      break;
    case RE_STAR: case RE_PLUS:
      //fprintf (stderr, "free * or + (\n");
      re_free (re->what.sub);
      //fprintf (stderr, ")\n");
      break;
    case RE_GROUP:
      //fprintf (stderr, "free group (\n");
      re_free (re->what.group.sub);
      //fprintf (stderr, ")\n");
      break;
    case RE_EPSILON:
      //fprintf (stderr, "free epsilon\n");
      break;
    default:
      //fprintf (stderr, "free what?\n");
      break;
    }
  RE_FREE(re);
}

static regex_t *re_alloc (void)
{
  regex_t *re;
  
  re = RE_MALLOC(sizeof(regex_t));
  re->nullable = 0; /* for now */
  return re;
}

PRIVATE void regexSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  regex_t *re;
  
  switch (doext->xtag)
    {
    case X_SAVE:
      re = (regex_t *)extdata;
    save_again:
      outChars((CHAR *)&re->kind,1);
      if (errno) return;
      switch (re->kind)
	{
	case RE_CHAR:
	  outChars((CHAR *)&re->what.letter,1);
	  if (errno) return;
	  break;
	case RE_RANGE:
	  outChars((CHAR *)&re->what.range,33);
	  if (errno) return;
	  break;
	case RE_CONCAT: case RE_CHOICE:
	  regexSweep ((char *)re->what.pair.left);
	  if (errno) return;
	  re = re->what.pair.right;
	  goto save_again;
	case RE_STAR: case RE_PLUS:
	  re = re->what.sub;
	  goto save_again;
	case RE_GROUP:
	  outInt(re->what.group.n);
	  if (errno) return;
	  re = re->what.group.sub;
	  goto save_again;
	}
      break;
    case X_MARK:
    case X_MARKDATA:
      break;
    case X_FREE:
      re_free ((regex_t *)extdata);
      break;
    }
}

static regex_t *re_make_epsilon (void)
{
  regex_t *re;
  
  re = re_alloc ();
  re->kind = RE_EPSILON;
  re->nullable = 1;
  return re;
}

static regex_t *re_make_catch_all (void)
{
  regex_t *re;
  
  re = re_alloc ();
  re->kind = RE_CATCH_ALL;
  re->nullable = 1;
  return re;
}

static regex_t *re_make_group (regex_t *sub, int n)
{
  regex_t *re;

  re = re_alloc ();
  re->kind = RE_GROUP;
  re->nullable = sub->nullable;
  re->what.group.sub = sub;
  re->what.group.n = n;
  return re;
}

static regex_t *re_make_letter (char c)
{
  regex_t *re;

  re = re_alloc ();
  re->kind = RE_CHAR;
  re->nullable = 0;
  re->what.letter = c;
  return re;
}

static regex_t *re_make_choice (regex_t *left, regex_t *right)
{
  regex_t *re;

  if (left->kind==RE_EPSILON && right->nullable)
    {
      re_free (left);
      return right;
    }
  if (right->kind==RE_EPSILON && left->nullable)
    {
      re_free (right);
      return left;
    }
  re = re_alloc ();
  re->kind = RE_CHOICE;
  re->nullable = left->nullable || right->nullable;
  re->what.pair.left = left;
  re->what.pair.right = right;
  return re;
}

static regex_t *re_make_concat (regex_t *left, regex_t *right)
{
  regex_t *re;

  if (right->kind==RE_EPSILON)
    {
      re_free (right);
      return left;
    }
  if (left->kind==RE_EPSILON)
    {
      re_free (left);
      return right;
    }
  re = re_alloc ();
  re->kind = RE_CONCAT;
  re->nullable = left->nullable && right->nullable;
  re->what.pair.left = left;
  re->what.pair.right = right;
  return re;
}

static regex_t *re_make_dot (void)
{
  char *bitfield;
  int i;
  regex_t *re;

  bitfield = RE_MALLOC(33);
  for (i=0; i<32; i++)
    bitfield[i] = 0xff;
  bitfield[32] = 0; /* do not match end of string */
  re = re_alloc ();
  re->kind = RE_RANGE;
  //re->nullable = 0;
  re->what.range = bitfield;
  return re;
}

static regex_t *re_denull (regex_t *re)
/* return a regex re' such that re* = re'*, but which
   is not nullable (does not recognize epsilon);
   works by modifying re, and deallocating the irrelevant
   parts of re.
*/
{
  regex_t *re2;
  
  if (re==NULL)
    return NULL;
  if (!re->nullable)
    return re;
  switch (re->kind)
    {
    case RE_EPSILON: return NULL; /* NULL is taken to denote the empty regex */
    case RE_CONCAT: /* A.B, where A.B is nullable; that means that both A and B
		       are nullable, so (A.B)* = (A|B)* = (denull(A)|denull(B))*
		       That is, denull(A.B) = denull(A)|denull(B).
		    */
    case RE_CHOICE: /* A|B, where A|B is nullable: (A|B)* = (denull(A)|denull(B))*.
		     Hence denull(A|B) = denull(A)|denull(B). */
      re2 = re_make_choice (re_denull (re->what.pair.left),
			    re_denull (re->what.pair.right));
      RE_FREE(re);
      return re2;
    case RE_STAR: /* denull(A*) = denull(A) */
    case RE_PLUS: /* denull(A+) = denull(A.A*) = denull(A)|denull(A*) = denull(A)
		     if A is nullable;
		     otherwise A+ is not nullable. */
      re2 = re_denull (re->what.sub);
      RE_FREE(re);
      return re2;
    case RE_CATCH_ALL: /* should not happen, but anyway, just in case: denull(.*) = . */
      re2 = re_make_dot ();
      re_free (re);
      return re2;
    case RE_CHAR: case RE_RANGE: /* should not happen, because those are not nullable */
    case RE_GROUP: /* should not happen, because groups cannot be put inside stars,
		      and re_denull() is only called from re_make_star() */
    default:
      return NULL;
    }
}

static regex_t *re_make_star (regex_t *sub)
{
  regex_t *re;

  sub = re_denull (sub);
  if (sub==NULL) /* \emptyset* = epsilon */
    return re_make_epsilon ();
  re = re_alloc ();
  re->kind = RE_STAR;
  re->nullable = 1;
  re->what.sub = sub;
  return re;
}

static unsigned int re_parse_int (char *regex, size_t *posp, size_t len)
{
  unsigned int i;
  char c;

  i = 0;
  while (*posp<len && (c = regex[*posp], isdigit(c)))
    {
      i = 10*i + (c - '0');
      ++*posp;
    }
  return i;
}

static regex_t *re_copy_range (char *bitfield)
{
  char *newbf;
  regex_t *newre;

  newbf = RE_MALLOC(33);
  memcpy (newbf, bitfield, 33);
  newre = re_alloc ();
  newre->kind = RE_RANGE;
  //newre->nullable = 0;
  newre->what.range = newbf;
  return newre;
}

static regex_t *re_make_plus (regex_t *sub)
{
  regex_t *re;

  if (sub->nullable) /* if A is nullable, then A+ = A*.
			Indeed, A+ is always included in A*.  Conversely,
			we must show that A* = epsilon | A+ is included
			in A+, and that boils down to showing that epsilon
			is in A+, which is obvious since epsilon is in A,
			as A is nullable. */
    return re_make_star (sub);
  switch (sub->kind)
    {
      // case RE_EPSILON: /* epsilon+ = epsilon; case already dealt with by the sub->nullable case above */
    case RE_PLUS: /* A++ = A+ */
      //case RE_STAR: /* A*+ = A*; case already dealt with by the sub->nullable case above */
      //case RE_CATCH_ALL: /* should not happen, but anyway, just in case: .*+ = .* */
      // /* No: case already dealt with by the sub->nullable case above */
      return sub;
    }
  re = re_alloc ();
  re->kind = RE_PLUS;
  re->nullable = sub->nullable;
  re->what.sub = sub;
  return re;
}

static regex_t *re_copy (regex_t *re)
{
  if (re==NULL)
    return NULL;
  switch (re->kind)
    {
    case RE_EPSILON: return re_make_epsilon ();
    case RE_CHAR: return re_make_letter (re->what.letter);
    case RE_RANGE: return re_copy_range (re->what.range);
    case RE_CONCAT: return re_make_concat (re_copy (re->what.pair.left),
					   re_copy (re->what.pair.right));
    case RE_CHOICE: return re_make_choice (re_copy (re->what.pair.left),
					   re_copy (re->what.pair.right));
    case RE_STAR: return re_make_star (re_copy (re->what.sub));
    case RE_PLUS: return re_make_plus (re_copy (re->what.sub));
    case RE_GROUP: return re_make_group (re_copy (re->what.group.sub),
					 re->what.group.n);
    case RE_CATCH_ALL: return re_make_catch_all ();
    }
  return NULL;
}

static regex_t *re_load_range()
{
  char *range;
  regex_t *re;
  int i;

  range = (char *)MALLOC(33);
  inChars((CHAR *)range,33);
  if (errno)
    {
      MARKFREE(range);
      return NULL;
    }
  for (i=0; i<33; i++)
    if (range[i])
      break;
  re = re_alloc();
  re->kind = RE_RANGE;
  if (i==33) /* range is completely empty; should not happen */
    re->nullable = 1;
  re->what.range = range;
  return re;
}

static regex_t *re_load()
{
  char kind, c;
  INT n;
  regex_t *sub, *right, *re = NULL;
  
  inChars((CHAR *)&kind,1);
  if (errno) return NULL;
  switch (kind)
    {
    case RE_EPSILON:
      re = re_make_epsilon();
      break;
    case RE_CHAR:
      inChars((CHAR *)&c,1);
      if (errno) return NULL;
      re = re_make_letter(c);
      break;
    case RE_RANGE:
      re = re_load_range();
      break;
    case RE_GROUP:
      n = inInt();
      if (errno) return NULL;
      sub = re_load();
      if (sub==NULL) return NULL;
      re = re_make_group(sub,n);
      break;
    case RE_CONCAT:
      sub = re_load();
      if (sub==NULL) return NULL;
      right = re_load();
      if (right==NULL) return NULL;
      re = re_make_concat(sub,right);
      break;
    case RE_CHOICE:
      sub = re_load();
      if (sub==NULL) return NULL;
      right = re_load();
      if (right==NULL) return NULL;
      re = re_make_choice(sub,right);
      break;
    case RE_STAR:
      sub = re_load();
      if (sub==NULL) return NULL;
      re = re_make_star(sub);
      break;
    case RE_PLUS:
      sub = re_load();
      if (sub==NULL) return NULL;
      re = re_make_plus(sub);
      break;
    case RE_CATCH_ALL:
      re = re_make_catch_all();
      break;
    }
  return re;
}

PRIVATE ExprPtr regexLoad (desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  regex_t *re;
  ExprPtr e;

  re = re_load();
  if (re==NULL) return NIL;
  e = MakeExternal (desc, (char *)re);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc regexExtDesc = {
  (CHAR *)"regex",
  { 'r', 'e', 'g', 'x' },
  NULL,
  NULL, /* no hash function, regexes are not shared */
  NULL, /* no comparison function, regexes are not shared */
  NULL, /* no print function */
  regexSweep,
  regexLoad
};

static regex_t *re_remove_groups (regex_t *re)
{
  regex_t *sub;
  
  if (re==NULL)
    return NULL;
  switch (re->kind)
    {
    case RE_EPSILON: case RE_CHAR: case RE_RANGE: case RE_CATCH_ALL:
      return re;
    case RE_CONCAT: case RE_CHOICE:
      re->what.pair.left = re_remove_groups (re->what.pair.left);
      re->what.pair.right = re_remove_groups (re->what.pair.right);
      return re;
    case RE_STAR: case RE_PLUS:
      re->what.sub = re_remove_groups (re->what.sub);
      return re;
    case RE_GROUP:
      sub = re->what.group.sub; /* re_remove_groups (re->what.group.sub) ? */
      RE_FREE(re);
      return sub;
    }
  return re;
}

static regex_t *re_make_repetition (regex_t *atom, unsigned int i, unsigned int j)
{
  regex_t *re;

  re = re_make_epsilon ();
  while (j > i)
    {
      re = re_make_choice (re_make_epsilon (),
			   re_make_concat (re_copy (atom), re));
      --j;
    }
  while (i > 0)
    {
      re = re_make_concat (re_copy (atom), re);
    }
  re_free (atom);
  return re;
}

static void re_parse_bracket_expr_core (char *regex, size_t *posp, size_t len, int *err,
					char *bitfield)
{
  unsigned int i, j;
  char c;

  *err = 0;
  if (*posp>=len)
    {
      *err = RE_PARSE_PREMATURE_END;
      return;
    }
  for (i=0; i<33; i++)
    bitfield[i] = 0;
  c = regex[*posp];
  if (c==']' || c=='-') /* include ']' in bitfield and proceed */
    {
      SET(bitfield,(unsigned int)(unsigned char)c);
      ++*posp;
    }
  while (*posp<len && (c = regex[*posp], c!=']'))
    {
      ++*posp;
      if (regex[*posp]=='-' && *posp+2<=len && regex[*posp+1]!=']') /* range */
	{
	  for (i=(unsigned int)(unsigned char)c, j=(unsigned int)(unsigned char)regex[*posp+1];
	       i<=j; i++)
	    SET(bitfield,i);
	  *posp += 2;
	}
      else if (c=='[' && regex[*posp]==':')
	{
	  if (*posp+8<=len)
	    {
	      if (memcmp(&regex[*posp], ":alnum:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isalnum(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":alpha:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isalpha(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":blank:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isblank(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":cntrl:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (iscntrl(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":digit:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isdigit(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":graph:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isgraph(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":lower:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (islower(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":print:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isprint(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":punct:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (ispunct(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":space:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isspace(i)) SET(bitfield, i);
		}
	      else if (memcmp(&regex[*posp], ":upper:]", 8)==0)
		{
		  for (i=0; i<128; i++)
		    if (isupper(i)) SET(bitfield, i);
		}
	      else if (*posp+9<=len && memcmp(&regex[*posp], ":xdigit:]", 9)==0)
		{
		  for (i=0; i<128; i++)
		    if (isxdigit(i)) SET(bitfield, i);
		  ++*posp;
		}
	      else goto no_char_class;
	      *posp += 8;
	    }
	}
      else
	{
	no_char_class:
	  i = (unsigned int)(unsigned char)c;
	  SET(bitfield,i);
	}
    }
  if (*posp>=len)
    {
      *err = RE_PARSE_PREMATURE_END;
      return;
    }
  ++*posp;
}

static void re_parse_bracket_expr (char *regex, size_t *posp, size_t len, int *err,
				   char *bitfield)
{
  int i;

  *err = 0;
  if (*posp>=len)
    {
      *err = RE_PARSE_PREMATURE_END;
      return;
    }
  if (regex[*posp]=='^')
    {
      ++*posp;
      re_parse_bracket_expr_core (regex, posp, len, err, bitfield);
      if (*err)
	return;
      for (i=0; i<32; i++)
	bitfield[i] = ~bitfield[i];
      /* do not negate bitfield[32], since that holds only 'special
	 characters', which are not counted in bracket expressions */
    }
  else
    re_parse_bracket_expr_core (regex, posp, len, err, bitfield);
}

static regex_t *re_parse_bracket_expression (char *regex, size_t *posp, size_t len, int *err)
{
  char *bitfield;
  regex_t *re;

  bitfield = RE_MALLOC(33);
  re_parse_bracket_expr (regex, posp, len, err, bitfield);
  if (*err!=0)
    {
      RE_FREE(bitfield);
      return NULL;
    }
  re = re_alloc ();
  re->kind = RE_RANGE;
  re->what.range = bitfield;
  return re;
}

static regex_t *re_parse_core (char *regex, size_t *posp, size_t len, int *err,
			       int *group_n, int does_caret);

static regex_t *re_parse_atom (char *regex, size_t *posp, size_t len, int *err,
			       int *group_n)
{
  regex_t *re;
  int n;
  
  if (*posp>=len)
    {
      *err = RE_PARSE_PREMATURE_END;
      return NULL;
    }
  switch (regex[*posp])
    {
    case '(':
      ++*posp;
      if (group_n!=NULL)
	n = (*group_n)++;
      else n = -1;
      re = re_parse_core (regex, posp, len, err, group_n, 0);
      if (re==NULL)
	return NULL;
      if (*posp>=len || regex[*posp]!=')')
	{
	  *err = RE_PARSE_EXPECTED_CLOSING_PARENTHESIS;
	  re_free (re);
	  return NULL;
	}
      ++*posp;
      if (group_n!=NULL)
	{
	  re = re_make_group (re, n);
	}
      break;
    case '[':
      ++*posp;
      re = re_parse_bracket_expression (regex, posp, len, err);
      break;
    case '.':
      ++*posp;
      re = re_make_dot ();
      break;
    case '\\':
      ++*posp;
      if (*posp>=len)
	{
	  *err = RE_PARSE_NO_LETTER_AFTER_BACKSLASH;
	  return NULL;
	}
      re = re_make_letter (regex[*posp]);
      ++*posp;
      break;
    default:
      re = re_make_letter (regex[*posp]);
      ++*posp;
      break;
    }
  return re;
}

static regex_t *re_parse_piece (char *regex, size_t *posp, size_t len, int *err,
				int *group_n)
{
  regex_t *atom, *re;
  unsigned int i, j;
  int n;

  if (group_n!=NULL)
    n = *group_n;
  else n = -1;
  atom = re_parse_atom (regex, posp, len, err, group_n);
  if (atom==NULL)
    return NULL;
  if (*posp>=len)
    re = atom;
  else
    switch (regex[*posp])
      {
      case '+':
	atom = re_remove_groups (atom);
	*group_n = n;
	re = re_make_plus (atom);
	++*posp;
	break;
      case '*':
	atom = re_remove_groups (atom);
	*group_n = n;
	re = re_make_star (atom);
	++*posp;
	break;
      case '?':
	atom = re_remove_groups (atom);
	*group_n = n;
	re = re_make_choice (re_make_epsilon (),
			     atom);
	++*posp;
	break;
      case '{':
	++*posp;
	i = re_parse_int (regex, posp, len);
	if (*posp>=len)
	  {
	    *err = RE_PARSE_PREMATURE_END;
	    return NULL;
	  }
	switch (regex[*posp])
	  {
	  case '}':
	    j = i; /* match exactly i repetitions of the atom */
	    break;
	  case ',':
	    ++*posp;
	    j = re_parse_int (regex, posp, len);
	    if (*posp>=len)
	      {
		*err = RE_PARSE_PREMATURE_END;
		return NULL;
	      }
	    if (regex[*posp]!='}')
	      {
		*err = RE_PARSE_EXPECTED_CLOSING_BRACE;
		return NULL;
	      }
	    break;
	  default:
	    *err = RE_PARSE_EXPECTED_COMMA_OR_CLOSING_BRACE;
	    return NULL;
	  }
	if (j < i)
	  {
	    *err = RE_PARSE_WRONG_BOUND;
	    return NULL;
	  }
#if 0
	if (j>=1024)
	  {
	    *err = RE_PARSE_BOUND_TOO_HIGH;
	    return NULL;
	  }
#endif
	++*posp;
	atom = re_remove_groups (atom);
	*group_n = n;
	re = re_make_repetition (atom, i, j);
	break;
      default:
	re = atom;
	break;
      }
  return re;
}

static regex_t *re_parse_branch (char *regex, size_t *posp, size_t len, int *err,
				 int *group_n)
{
  regex_t *left, *right, *re;

  if (*posp>=len || regex[*posp]=='|' || regex[*posp]==')' || regex[*posp]=='$')
    {
      re = re_make_epsilon ();
    }
  else
    {
      left = re_parse_piece (regex, posp, len, err, group_n);
      if (left==NULL)
	return NULL;
      right = re_parse_branch (regex, posp, len, err, group_n);
      if (right==NULL)
	{
	  re_free (left);
	  return NULL;
	}
      re = re_make_concat (left, right);
    }
  return re;
}

static regex_t *re_parse_branch_handling_dollar (char *regex, size_t *posp, size_t len,
						 int *err, int *group_n)
{
  regex_t *re;
  int n;

  if (group_n!=NULL)
    n = (*group_n)++;
  else n = -1;
  re = re_parse_branch (regex, posp, len, err, group_n);
  if (re!=NULL)
    {
      re = re_make_group (re, n);
      if (*posp<len && regex[*posp]=='$')
	{
	  ++*posp;
	}
      else
	{ /* Compile R as R .*$ */
	  re = re_make_concat (re, re_make_star (re_make_dot ()));
	  /* instead of re_make_concat (re, re_make_catch_all ()) */
	}
    }
  return re;
}

static regex_t *re_parse_branch_handling_caret (char *regex, size_t *posp, size_t len,
						int *err, int *group_n)
{
  regex_t *re;
  
  if (*posp<len && regex[*posp]=='^')
    {
      ++*posp;
      re = re_parse_branch_handling_dollar (regex, posp, len, err, group_n);
    }
  else
    {
      re = re_parse_branch_handling_dollar (regex, posp, len, err, group_n);
      /* Compile R as .* followed by R */
      if (re!=NULL)
	re = re_make_concat (re_make_star (re_make_dot ()), re);
    }
  return re;
}

regex_t *re_parse_core (char *regex, size_t *posp, size_t len, int *err,
			int *group_n, int does_caret)
{
  regex_t *left, *right, *re;
  int n, *np;

  if (group_n!=NULL)
    n = *group_n;
  if (does_caret)
    left = re_parse_branch_handling_caret (regex, posp, len, err, group_n);
  else left = re_parse_branch (regex, posp, len, err, group_n);
  if (left==NULL)
    return NULL;
  if (*posp < len && regex[*posp]=='|')
    {
      ++*posp;
      if (group_n!=NULL)
	np = &n;
      else np = NULL;
      right = re_parse_core (regex, posp, len, err, np, does_caret);
      if (right==NULL)
	{
	  re_free (left);
	  return NULL;
	}
      if (group_n!=NULL && *group_n!=n)
	{
	  re_free (left);
	  re_free (right);
	  *err = RE_PARSE_SUBMATCH_DISCREPANCY;
	  return NULL;
	}
      re = re_make_choice (left, right);
    }
  else
    re = left;
  return re;
}

regex_t *re_parse_more_control (char *regexp, size_t len, int *errp, int *group_n,
				size_t *posp, int does_caret)
{
  size_t pos;

  if (posp==NULL)
    posp = &pos;
  *posp = 0;
  *errp = 0;
  return re_parse_core (regexp, posp, len, errp, group_n, does_caret);
}

regex_t *re_parse (char *regexp, size_t len, int *errp)
{
  size_t pos;
  int group;

  pos = 0;
  group = 0;
  *errp = 0;
  return re_parse_core (regexp, &pos, len, errp, &group, 1);
}


typedef size_t re_state_num;
#define RE_STATE_UNKNOWN SIZE_MAX

typedef struct re_state_s re_state_t;
/* a re_state_s is actually a state plus some info on outgoing transitions.
 */

struct re_directed_s {
  re_state_num alt;
  char next_bitfield[33];
  char alt_bitfield[33];
};

struct re_state_s {
  char kind;
#define RE_S_MATCH_CHAR 1
#define RE_S_MATCH_RANGE 2
#define RE_S_SPLIT 3
#define RE_S_ACCEPT 4
#define RE_S_MEMO_START 5
#define RE_S_MEMO_END 6
#define RE_S_CHECK_END 7
#define RE_S_DIRECTED_SPLIT 8
#define RE_S_EXCLUSIVE_SPLIT 9
  char presence;
#define RE_PRESENCE_1 1
  /* presence | RE_PRESENCE_1 is true if state is in nfa->buf1
     presence | RE_PRESENCE_2 is true if state is in nfa->buf2
  */
#define RE_PRESENCE_2 2
  re_state_num next;
  size_t lastidx; /* used in re_nfa_run (), filled in only if state is present in nfa->cur
		   (i.e., in nfa->buf2 on even positions, in nfa->buf1 on odd positions) */
  union {
    char letter; /* RE_S_MATCH_CHAR, then go to next */
    char *range; /* RE_S_MATCH_RANGE: 33 byte array (bitfields); if match goto next */
    re_state_num alt; /* RE_S_SPLIT: either go to next, or to what.alt */
    struct memo_s { int nmatch; } memo;
    int nbranch; /* RE_S_ACCEPT, RE_S_CHECK_END */
    struct re_directed_s *dir; /* RE_S_DIRECTED_SPLIT, RE_S_EXCLUSIVE_SPLIT */
  } what;
};

typedef struct re_nfa_thread_s re_nfa_thread_t;
struct re_nfa_thread_s {
  re_state_num q;
  re_nfa_env_t *env;
};

struct re_nfa_s {
  size_t n, sz;
  re_state_t *q;
  re_state_num start;
  /* The following is used by re_nfa_run (): */
  re_nfa_thread_t *buf1, *buf2;
  re_nfa_thread_t *cur, *prev; /* will point to buf1 and buf2, or buf2 and buf1, respectively */
  size_t ncur, nprev;
  int final_branch;
  re_nfa_env_t *final_env;
};

typedef struct re_nfa_actions_s {
  re_nfa_t *nfa;
  size_t nre;
  ExprPtr actions[1]; /* really nre of them */
} re_nfa_actions_t;

PRIVATE void nfaSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  re_nfa_actions_t *nfa_actions;
  re_nfa_t *nfa;
  re_state_t *q;
  COUNTER i;
  size_t nre;

  nfa_actions = (re_nfa_actions_t *)extdata;
  nre = nfa_actions->nre;
  nfa = nfa_actions->nfa;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt(nre);
      if (errno) return;
      if (nre!=0)
	{
	  outInt(nfa->n);
	  if (errno) return;
	  outInt(nfa->sz);
	  if (errno) return;
	  outInt(nfa->start);
	  for (i=0, q=nfa->q; i<nfa->n; i++, q++)
	    {
	      outChars((CHAR *)&q->kind,1);
	      if (errno) return;
	      outInt(q->next);
	      if (errno) return;
	      switch (q->kind)
		{
		case RE_S_MATCH_CHAR:
		  outChars((CHAR *)&q->what.letter,1);
		  break;
		case RE_S_MATCH_RANGE:
		  outChars((CHAR *)q->what.range,33);
		  break;
		case RE_S_SPLIT:
		  outInt(q->what.alt);
		  break;
		case RE_S_MEMO_START: case RE_S_MEMO_END:
		  outInt(q->what.memo.nmatch);
		  break;
		case RE_S_ACCEPT: case RE_S_CHECK_END:
		  outInt(q->what.nbranch);
		  break;
		case RE_S_DIRECTED_SPLIT: case RE_S_EXCLUSIVE_SPLIT:
		  outInt(q->what.dir->alt);
		  if (errno) return;
		  outChars((CHAR *)q->what.dir->next_bitfield,33);
		  if (errno) return;
		  outChars((CHAR *)q->what.dir->alt_bitfield,33);
		  break;
		}
	    }
	}
      for (i=0; i<nre; i++)
	{
	  (*doext->action)(nfa_actions->actions[i]);
	  if (errno) return;
	}
      break;
    case X_MARK:
    case X_MARKDATA:
      for (i=0; i<nre; i++)
	(*doext->action)(nfa_actions->actions[i]);
      break;
    case X_FREE:
      if (nfa!=NULL)
	re_nfa_free (nfa);
      MARKFREE(nfa_actions);
      break;
    }
}

static re_nfa_t *re_nfa_new (void)
{
  re_nfa_t *nfa;

  nfa = RE_MALLOC(sizeof(re_nfa_t));
  nfa->n = 0;
  nfa->sz = 10;
  nfa->q = RE_MALLOC(nfa->sz*sizeof(struct re_state_s));
  nfa->start = -1;
  nfa->buf1 = nfa->buf2 = nfa->cur = nfa->prev = NULL;
  nfa->ncur = nfa->nprev = 0;
  nfa->final_branch = INT_MAX;
  nfa->final_env = NULL;
  return nfa;
}

void re_nfa_free (re_nfa_t *nfa)
{
  size_t i, n;

  if (nfa==NULL)
    return;
  n = nfa->n;
  for (i=0; i<n; i++)
    switch (nfa->q[i].kind)
      {
      case RE_S_MATCH_RANGE:
	RE_FREE(nfa->q[i].what.range);
	break;
      case RE_S_DIRECTED_SPLIT:
      case RE_S_EXCLUSIVE_SPLIT:
	RE_FREE(nfa->q[i].what.dir);
	break;
      }
  if (nfa->buf1!=NULL)
    RE_FREE(nfa->buf1);
  if (nfa->buf2!=NULL)
    RE_FREE(nfa->buf2);
  if (nfa->final_env!=NULL)
    RE_FREE(nfa->final_env);
  RE_FREE(nfa);
}

static re_nfa_actions_t *re_nfa_actions_load(sharedref,load)
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  INT nre, n, sz, start;
  COUNTER i;
  re_state_t *states, *q;
  re_nfa_t *nfa;
  re_nfa_actions_t *nfa_actions;
  struct re_directed_s *dir;

  nre = inInt();
  if (errno) return NULL;
  if (nre!=0)
    {
      n = inInt();
      if (errno) return NULL;
      sz = inInt();
      if (errno) return NULL;
      if (sz<n)
	sz = n;
      start = inInt();
      if (errno) return NULL;
    
      nfa = RE_MALLOC(sizeof(re_nfa_t));
      nfa->n = 0;
      nfa->sz = sz;
      nfa->q = states;
      nfa->start = -1;
      nfa->buf1 = nfa->buf2 = nfa->cur = nfa->prev = NULL;
      nfa->ncur = nfa->nprev = 0;
      nfa->final_branch = INT_MAX;
      nfa->final_env = NULL;

      states = RE_MALLOC(sz*sizeof(struct re_state_s));
      for (i=0, q=states; i<n; i++, q++)
	{
	  inChars((CHAR *)&q->kind,1);
	  q->presence = 0;
	  if (errno)
	    {
	    err_q:
	      re_nfa_free(nfa);
	      return NULL;
	    }
	  nfa->n++;
	  q->next = inInt();
	  if (errno) goto err_q;
	  q->lastidx = 0; /* useless */
	  switch (q->kind)
	    {
	    case RE_S_MATCH_CHAR:
	      inChars((CHAR *)&q->what.letter,1);
	      break;
	    case RE_S_MATCH_RANGE:
	      q->what.range = RE_MALLOC(33);
	      inChars((CHAR *)&q->what.range,33);
	      break;
	    case RE_S_SPLIT:
	      q->what.alt = inInt();
	      break;
	    case RE_S_MEMO_START: case RE_S_MEMO_END:
	      q->what.memo.nmatch = inInt();
	      break;
	    case RE_S_ACCEPT: case RE_S_CHECK_END:
	      q->what.nbranch = inInt();
	      break;
	    case RE_S_DIRECTED_SPLIT: case RE_S_EXCLUSIVE_SPLIT:
	      q->what.dir = dir = RE_MALLOC(sizeof(struct re_directed_s));
	      dir->alt = inInt();
	      if (errno) goto err_q;
	      inChars((CHAR *)dir->next_bitfield,33);
	      if (errno) goto err_q;
	      inChars((CHAR *)dir->alt_bitfield,33);
	      break;
	    default:
	      errno = EIO;
	      break;
	    }
	  if (errno) goto err_q;
	}
    }
  else nfa = NULL;
  nfa_actions = (re_nfa_actions_t *)MALLOC(OFFSETOF(re_nfa_actions_t,actions[nre]));
  nfa_actions->nre = nre;
  nfa_actions->nfa = nfa;
  for (i=0; i<nre; i++)
    {
      nfa_actions->actions[i] = (*load)(sharedref);
      if (errno)
	{
	  MARKFREE(nfa_actions);
	  re_nfa_free(nfa);
	  return NULL;
	}
    }
  return nfa_actions;
}

PRIVATE ExprPtr nfaLoad (desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  re_nfa_actions_t *nfa_actions;
  ExprPtr e;

  nfa_actions = re_nfa_actions_load(sharedref,load);
  if (nfa_actions==NULL) return NIL;
  e = MakeExternal (desc, (char *)nfa_actions);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc nfaExtDesc = {
  (CHAR *)"nfa",
  { 'n', 'f', 'a', ' ' },
  NULL,
  NULL, /* no hash function, nfas are not shared */
  NULL, /* no comparison function, nfas are not shared */
  NULL, /* no print function */
  nfaSweep,
  nfaLoad
};

int re_matching_branch (re_nfa_t *nfa)
{
  if (nfa==NULL)
    return -1;
  return nfa->final_branch;
}

re_nfa_env_t *re_matches (re_nfa_t *nfa)
{
  if (nfa==NULL)
    return NULL;
  return nfa->final_env;
}

#define RE_NFA_ENV_COPY(env) ((env)==NULL?NULL:((env)->refcnt++, (env)))
#define RE_NFA_ENV_FREE(env) do { \
  if ((env)!=NULL && --(env)->refcnt==0) \
    RE_FREE(env);		 \
  } while (0)

static void re_nfa_clean_buf (re_nfa_thread_t *buf, size_t n)
{
  size_t i;
  re_nfa_env_t *env;

  for (i=0; i<n; i++)
    {
      env = buf[i].env;
      RE_NFA_ENV_FREE(env);
    }
}

static re_state_num re_nfa_new_state (re_nfa_t *nfa)
{
  size_t n, sz;

  n = nfa->n;
  sz = nfa->sz;
  if (n>=sz)
    {
      sz = n;
      sz <<= 1;
      nfa->q = RE_REALLOC(nfa->q, sz*sizeof(struct re_state_s));
      nfa->sz = sz;
    }
  nfa->n = n+1;
  return n;
}

void re_state_first (re_nfa_t *nfa, re_state_num q, char bitfield[33])
{
  re_state_t *st;
  int i;
  struct re_directed_s *dir;

 again:
  st = &nfa->q[q];
  switch (st->kind)
    {
    case RE_S_MATCH_CHAR:
      SET(bitfield,(unsigned int)(unsigned char)st->what.letter);
      break;
    case RE_S_MATCH_RANGE:
      for (i=0; i<33; i++)
	bitfield[i] |= st->what.range[i];
      break;
    case RE_S_SPLIT:
      re_state_first (nfa, st->next, bitfield);
      q = st->what.alt;
      goto again;
    case RE_S_ACCEPT:
      for (i=0; i<32; i++)
	bitfield[i] = 0xff;
      /*fallthrough*/
    case RE_S_CHECK_END:
      SET(bitfield,(unsigned int)256); /* 'end of string' */
      break;
    case RE_S_MEMO_START: case RE_S_MEMO_END:
      q = st->next;
      goto again;
    case RE_S_DIRECTED_SPLIT: case RE_S_EXCLUSIVE_SPLIT:
      dir = st->what.dir;
      for (i=0; i<33; i++)
	bitfield[i] |= dir->next_bitfield[i] | dir->alt_bitfield[i];
      break;
    }
}


static void re_optimize_split (re_nfa_t *nfa, re_state_num q)
{
  re_state_t *st;
  char next_bitfield[33], alt_bitfield[33];
  struct re_directed_s *dir;
  int i;
  char cnext, calt;
  int has_next_and_alt, has_next_no_alt, has_alt_no_next;

  st = &nfa->q[q];
  if (st->kind==RE_S_SPLIT)
    {
      re_optimize_split (nfa, st->next);
      re_optimize_split (nfa, st->what.alt);
      memset (next_bitfield, 0, 33);
      memset (alt_bitfield, 0, 33);
      re_state_first (nfa, st->next, next_bitfield);
      re_state_first (nfa, st->what.alt, alt_bitfield);
      has_next_and_alt = 0;
      has_next_no_alt = 0;
      has_alt_no_next = 0;
      for (i=0; i<33; i++)
	{
	  cnext = next_bitfield[i];
	  calt = alt_bitfield[i];
	  has_next_and_alt |= (cnext & calt);
	  has_next_no_alt |= (cnext & ~calt);
	  has_alt_no_next |= (calt & ~cnext);
	}
      if (!has_next_and_alt) /* when next_bitfield inter alt_bitfield is empty: */
	st->kind = RE_S_EXCLUSIVE_SPLIT;
      else if (has_next_no_alt || has_alt_no_next) /* at least one letter will lead
						      to a deterministic choice */
	st->kind = RE_S_DIRECTED_SPLIT;
      else return; /* otherwise, the two branches split on exactly the same letters:
		      keep that a RE_S_SPLIT. */
      dir = RE_MALLOC(sizeof(struct re_directed_s));
      dir->alt = st->what.alt;
      memcpy (dir->next_bitfield, next_bitfield, 33);
      memcpy (dir->alt_bitfield, alt_bitfield, 33);
      st->what.dir = dir;
    }
}

static re_state_num re_make_nfa_split (re_nfa_t *nfa, re_state_num left, re_state_num right)
{
  re_state_num q;
  re_state_t *st;
  
  q = re_nfa_new_state (nfa);
  st = &nfa->q[q];
  st->kind = RE_S_SPLIT;
  st->presence = 0;
  st->next = left;
  st->what.alt = right;
  return q;
}

static re_state_num re_make_nfa_split_nary (re_nfa_t *nfa, size_t nbranch, re_state_num q[])
{
  size_t mid;

  if (nbranch==1)
    return q[0];
  mid = nbranch >> 1;
  return re_make_nfa_split (nfa,
			    re_make_nfa_split_nary (nfa, mid, q),
			    re_make_nfa_split_nary (nfa, nbranch-mid, q+mid));
}

static size_t re_count_choices (regex_t *re)
{
  if (re==NULL)
    return 0;
  if (re->kind==RE_CHOICE)
    return re_count_choices (re->what.pair.left) + re_count_choices (re->what.pair.right);
  return 1;
}

#if 0
static size_t re_count_choices (regex_t *re)
{
  regex_t **re_stk;
  size_t n, sz;
  size_t nchoices;
  
  if (re==NULL)
    return 0;
  sz = 10;
  /* allocate stack by hand, in order to avoid recursion */
  re_stk = (regex_t **)MALLOC(sz*sizeof(regex_t *));
  n = 0;
  nchoices = 0;
  do {
    if (re->kind==RE_CHOICE)
      {
	nchoices++;
	if (n>=sz)
	  {
	    sz <<= 1;
	    re_stk = (regex_t **)REALLOC(re_stk, sz*sizeof(regex_t));
	  }
	re_stk[n++] = re->what.pair.right;
	re = re->what.pair.left;
      }
  } while (n!=0);
  MARKFREE(re_stk);
  return nchoices;
}
#endif

static re_state_num re_make_nfa_aux (regex_t *re, re_nfa_t *nfa, re_state_num cont,
				     int nbranch);

static re_state_num *re_fill_nfa_aux_choice (regex_t *re, re_nfa_t *nfa, re_state_num cont,
					     int nbranch, re_state_num *q)
{
  if (re==NULL)
    return 0;
  switch (re->kind)
    {
    case RE_CHOICE:
      q = re_fill_nfa_aux_choice (re->what.pair.left, nfa, cont, nbranch, q);
      q = re_fill_nfa_aux_choice (re->what.pair.right, nfa, cont, nbranch, q);
      break;
    default:
      q[0] = re_make_nfa_aux (re, nfa, cont, nbranch);
      q++;
      break;
    }
  return q;
}

/* I want to compare sets of natural numbers with a linear preorder that guarantees that
   if max A < min B, then A is strictly less than B.
   One possibility is the multiset ordering, i.e., the numerical ordering
   when A and B are seen as natural numbers written in binary.
   This is also the lexicographic ordering on characters, from most significant
   to least significant bytes.
*/

static int cmp_bitfields (char *A, char *B)
{
  signed char diff;
  int i;

  for (i=0; i<33; i++)
    {
      diff = ((signed char)B[i])-((signed char)A[i]);
      if (diff) return diff;
    }
  return 0;
}

#define WGT(A,B) (cmp_bitfields(B,A)>0)
#define WLT(A,B) (cmp_bitfields(A,B)>0)
#define WEQ(A,B) (cmp_bitfields(A,B)==0)

static void re_sort_nfa_aux_choice_rec (re_state_num *q, char **weights,
					size_t start, size_t end)
/* A (naive) implementation of quicksort, specialized to arrays of states.
   I could use qsort_r() instead, but this does not exist in all distributions of Linux.
   I don't want to use qsort(), which would force me to pass weights in a global variable.
 */
{
  size_t left;
  size_t right;
  size_t mid;
  re_state_num qaux;
  char *pivot_weight, *waux, *w[3];
#define WSWAP(w1,w2) do { waux=w1; w1=w2; w2=waux; } while (0)
#define QSWAP(q1,q2) do { qaux=q1; q1=q2; q2=qaux; } while (0)

  switch (end-start)
    {
    case 0: /* empty table */
    case 1: /* one element, already sorted */
      break;
    case 2:
      weights += start;
      q += start;
      if (WGT(weights[0],weights[1]))
	{
	  WSWAP(weights[0],weights[1]);
	  QSWAP(q[0],q[1]);
	}
      break;
    case 3:
      weights += start;
      q += start;
      if (WGT(weights[0],weights[2]))
	{
	  WSWAP(weights[0],weights[2]);
	  QSWAP(q[0],q[2]);
	}
      if (WLT(weights[1],weights[0]))
	{
	  WSWAP(weights[0],weights[1]);
	  QSWAP(q[0],q[1]);
	}
      else if (WGT(weights[1],weights[2]))
	{
	  WSWAP(weights[1],weights[2]);
	  QSWAP(q[1],q[2]);
	}
      break;
    default:
      left = start-1;
      right = end;
      mid = start + ((end-start)>>1);
      w[0] = weights[start];
      w[1] = weights[mid];
      w[2] = weights[end-1];
      if (WGT(w[0],w[2]))
	WSWAP(w[0],w[2]);
      if (WLT(w[1],w[0]))
	WSWAP(w[0],w[1]);
      if (WGT(w[1],w[2]))
	WSWAP(w[1],w[2]);
      pivot_weight = w[1]; /* take median of threes */
      while (1)
	{
	  do
	    right--;
	  while (WGT(weights[right],pivot_weight));
	  do
	    left++;
	  while (WLT(weights[left],pivot_weight));
	  /* Now all elements at indices start, start+1, ..., left-1
	     have weights <= pivot_weight,
	     all elements at indices right+1, ..., end-2, end-1
	     have weights >= pivot_weight;
	     the element at index left has weight >= pivot_weight,
	     the element at index right has weight <= pivot_weight.
	  */
	  if (left>=right)
	    break;
	  WSWAP(weights[left],weights[right]);
	  QSWAP(q[left],q[right]);
	}
      /* All elements at indices start, ..., left-1 have weights <= pivot_weight.
	 If left==right, the element at index left==right has weight exactly pivot_weight
	 (otherwise left=right+1).
	 All elements at indices right+1, ..., end-1 have weights >= pivot_weight.
      */
      while (left>start && WEQ(weights[left-1],pivot_weight))
	left--;
      do
	right++;
      while (right<end && WEQ(weights[right],pivot_weight));
      re_sort_nfa_aux_choice_rec (q, weights, start, left);
      re_sort_nfa_aux_choice_rec (q, weights, right, end);
      break;
    }
#undef WSWAP
#undef QSWAP
}

static void re_sort_nfa_aux_choice (re_nfa_t *nfa, size_t n, re_state_num *q)
{
  char **weights;
  char *cc;
  char *bitfield;
  size_t i;

  weights = RE_MALLOC(n*sizeof(char *));
  cc = RE_MALLOC(n*33);
  for (i=0; i<n; i++)
    {
      bitfield = weights[i] = cc+33*i;
      memset (bitfield, 0, 33);
      re_state_first (nfa, q[i], bitfield);
    }
  re_sort_nfa_aux_choice_rec (q, weights, 0, n);
  RE_FREE(cc);
  RE_FREE(weights);
}

static re_state_num re_make_nfa_aux_choice (regex_t *re, re_nfa_t *nfa, re_state_num cont,
					    int nbranch)
{
  re_state_num *q;
  re_state_num choice;
  size_t n;

  n = re_count_choices (re);
  q = RE_MALLOC(n*sizeof(re_state_num));
  re_fill_nfa_aux_choice (re, nfa, cont, nbranch, q);
  re_sort_nfa_aux_choice (nfa, n, q);
  choice = re_make_nfa_split_nary (nfa, n, q);
  RE_FREE(q);
  return choice;
}

static re_state_num re_make_nfa_aux (regex_t *re, re_nfa_t *nfa, re_state_num cont,
				     int nbranch)
{
  re_state_t *st;
  re_state_num q, left, right;
  
  if (re==NULL)
    return -1;
  q = -1;
  switch (re->kind)
    {
    case RE_EPSILON:
      q = cont;
      break;
    case RE_CHAR:
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_MATCH_CHAR;
      st->presence = 0;
      st->next = cont;
      st->what.letter = re->what.letter;
      break;
    case RE_RANGE:
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_MATCH_RANGE;
      st->presence = 0;
      st->next = cont;
      st->what.range = RE_MALLOC(33);
      memcpy (st->what.range, re->what.range, 33);
      break;
    case RE_CONCAT:
      right = re_make_nfa_aux (re->what.pair.right, nfa, cont, nbranch);
      q = re_make_nfa_aux (re->what.pair.left, nfa, right, nbranch);
      break;
    case RE_CHOICE:
      q = re_make_nfa_aux_choice (re, nfa, cont, nbranch);
      break;
    case RE_STAR: /* For A*, build the following:
		      ________
		     |        |
		     |        v
		     q <--A-- .
		     |
                     |---------> cont
		   */
      /* It is important that A is not nullable (see denull()), otherwise
	 this would create cycles of epsilon transitions, something we don't want. */
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_SPLIT;
      st->presence = 0;
      st->next = cont;
      st->what.alt = RE_STATE_UNKNOWN; /* will be set later (below) */
      right = re_make_nfa_aux (re->what.sub, nfa, q, nbranch);
      nfa->q[q].what.alt = right;
      /* st->what.alt = re_make_nfa_aux (...)
	 would be incorrect, as the nfa->q array may have been
	 reallocated inside re_make_nfa_aux(), hence st may have moved. */
      break;
    case RE_PLUS: /* For A+, build the following:
		      ________
		     |        |
		     v        |
		     q --A--> . ----> cont
		   */
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_SPLIT;
      st->presence = 0;
      st->next = cont;
      st->what.alt = RE_STATE_UNKNOWN; /* will be set later (below) */
      cont = q;
      q = re_make_nfa_aux (re->what.sub, nfa, cont, nbranch);
      nfa->q[cont].what.alt = q; /* not st->what.alt = q, since the nfa->q array may have been
				    reallocated inside re_make_nfa_aux(), hence st may have moved. */
      break;
    case RE_GROUP:
      right = re_nfa_new_state (nfa);
      st = &nfa->q[right];
      st->kind = RE_S_MEMO_END;
      st->presence = 0;
      st->next = cont;
      st->what.memo.nmatch = re->what.group.n;
      left = re_make_nfa_aux (re->what.group.sub, nfa, right, nbranch);
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_MEMO_START;
      st->presence = 0;
      st->next = left;
      st->what.memo.nmatch = re->what.group.n;
      break;
    case RE_CATCH_ALL:
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_ACCEPT; /* accept immediately without reading rest of string */
      st->presence = 0;
      st->next = -1; /* useless */
      st->what.nbranch = nbranch;
      break;
    }
  return q;
}

static void re_optimize_nfa (re_nfa_t *nfa)
{
  re_state_num i, n;

  n = nfa->n;
  for (i=0; i<n; i++)
    re_optimize_split (nfa, i);
}

static size_t re_nfa_count_splits (re_nfa_t *nfa, re_state_num q)
{
  re_state_t *st;

 again:
  st = &nfa->q[q];
  switch (st->kind)
    {
    case RE_S_SPLIT:
      return re_nfa_count_splits (nfa, st->next) + re_nfa_count_splits (nfa, st->what.alt);
    case RE_S_DIRECTED_SPLIT: case RE_S_EXCLUSIVE_SPLIT:
      return re_nfa_count_splits (nfa, st->next) + re_nfa_count_splits (nfa, st->what.dir->alt);
    case RE_S_MEMO_START: case RE_S_MEMO_END:
      q = st->next;
      goto again;
    default: /* RE_S_MATCH_CHAR, RE_S_MATCH_RANGE, RE_S_CASES, RE_S_ACCEPT, RE_S_CHECK_END */
      return 1;
    }
}

static re_state_num *re_nfa_make_splits (re_nfa_t *nfa, re_state_num q,
					 re_state_num *choices)
{
  re_state_t *st, *stt;
  re_state_num *end;

  st = &nfa->q[q];
  switch (st->kind)
    {
    case RE_S_SPLIT:
      choices = re_nfa_make_splits (nfa, st->next, choices);
      choices = re_nfa_make_splits (nfa, st->what.alt, choices);
      return choices;
    case RE_S_DIRECTED_SPLIT: case RE_S_EXCLUSIVE_SPLIT:
      choices = re_nfa_make_splits (nfa, st->next, choices);
      choices = re_nfa_make_splits (nfa, st->what.dir->alt, choices);
      return choices;
    case RE_S_MEMO_START: case RE_S_MEMO_END:
      end = re_nfa_make_splits (nfa, st->next, choices);
      for (; choices<end; choices++)
	{
	  if (st->next != *choices) /* recreate the RE_S_MEMO_START or the RE_S_MEMO_END,
				       unless st->next has not changed. */
	    {
	      q = re_nfa_new_state (nfa);
	      stt = &nfa->q[q];
	      stt->kind = st->kind;
	      stt->presence = st->presence;
	      stt->next = *choices;
	      stt->what.memo.nmatch = st->what.memo.nmatch;
	    }
	  *choices = q;
	}
      return end;
    default:
      *choices++ = q;
      return choices;
    }
}

static re_state_num *re_nfa_make_splits_array (re_nfa_t *nfa,
					       size_t n, re_state_num *q,
					       re_state_num *choices)
{
  size_t i;

  for (i=0; i<n; i++)
    choices = re_nfa_make_splits (nfa, q[i], choices);
  return choices;
}

re_nfa_t *re_make_nfa (size_t nre, regex_t *re[])
{
  re_nfa_t *nfa;
  re_state_num q;
  re_state_t *st;
  size_t nbranch;
  re_state_num *choices, *splitchoices;
  size_t nsplits;

  if (nre==0)
    return NULL;
  nfa = re_nfa_new ();
  choices = RE_MALLOC(nre*sizeof(re_state_num));
  nsplits = 0;
  for (nbranch = 0; nbranch<nre; nbranch++)
    {
      q = re_nfa_new_state (nfa);
      st = &nfa->q[q];
      st->kind = RE_S_CHECK_END; /* accept if and only if rest of string is empty */
      st->presence = 0;
      st->next = -1; /* useless */
      st->what.nbranch = nbranch;
      choices[nbranch] = re_make_nfa_aux (re[nbranch], nfa, q, nbranch);
      nsplits += re_nfa_count_splits (nfa, choices[nbranch]);
    }
  if (nre==1)
    {
      nfa->start = choices[0];
      RE_FREE(choices);
    }
  else
    {
      splitchoices = RE_MALLOC(nsplits*sizeof(re_state_num));
      (void) re_nfa_make_splits_array (nfa, nre, choices, splitchoices);
      RE_FREE(choices);
      re_sort_nfa_aux_choice (nfa, nsplits, splitchoices);
      nfa->start = re_make_nfa_split_nary (nfa, nsplits, splitchoices);
      RE_FREE(splitchoices);
    }
  re_optimize_nfa (nfa);
  if (nfa->n < nfa->sz) /* trim nfa->q state table, to save space */
    {
      nfa->sz = nfa->n;
      nfa->q = RE_REALLOC(nfa->q, nfa->sz*sizeof(struct re_state_s));
    }
  
  nfa->buf1 = RE_MALLOC(nfa->n*sizeof(re_nfa_thread_t));
  nfa->buf2 = RE_MALLOC(nfa->n*sizeof(re_nfa_thread_t));
  return nfa;
}

PUBLIC ExprPtr raiseREExc(err)
     int err;
{
  ExprPtr e,exc;

  e = KINT((INT)err);
  MAKECONS(reException,e,exc);
  return Raise(KEEP(exc));
}

/*
  re_make_nfa : (regex * (string * intarray -> 'a)) list -> 'a nfa
 */
DEF_ACTION_1ARG(reMakeNfa)
{
  size_t nre;
  re_nfa_t *nfa;
  re_state_num q;
  re_state_t *st;
  size_t nbranch;
  re_state_num *choices, *splitchoices;
  size_t nsplits;
  ExprPtr l;
  re_nfa_actions_t *nfa_actions;
  ExprPtr e;

  nre = ListLength(arg);
  nfa_actions = RE_MALLOC(OFFSETOF(re_nfa_actions_t,actions[nre]));
  nfa_actions->nre = nre;
  if (nre!=0)
    {
      nfa = re_nfa_new ();
      choices = RE_MALLOC(nre*sizeof(re_state_num));
      nsplits = 0;
      for (nbranch = 0, l=arg; nbranch<nre; nbranch++, l=CDR(l))
	{
	  q = re_nfa_new_state (nfa);
	  st = &nfa->q[q];
	  st->kind = RE_S_CHECK_END; /* accept if and only if rest of string is empty */
	  st->presence = 0;
	  st->next = -1; /* useless */
	  st->what.nbranch = nbranch;
	  nfa_actions->actions[nbranch] = CDR(CAR(l));
	  choices[nbranch] = re_make_nfa_aux ((regex_t *)EXTDATA(CAR(CAR(l))),
					      nfa, q, nbranch);
	  nsplits += re_nfa_count_splits (nfa, choices[nbranch]);
	}
      if (nre==1)
	{
	  nfa->start = choices[0];
	  RE_FREE(choices);
	}
      else
	{
	  splitchoices = RE_MALLOC(nsplits*sizeof(re_state_num));
	  (void) re_nfa_make_splits_array (nfa, nre, choices, splitchoices);
	  RE_FREE(choices);
	  re_sort_nfa_aux_choice (nfa, nsplits, splitchoices);
	  nfa->start = re_make_nfa_split_nary (nfa, nsplits, splitchoices);
	  RE_FREE(splitchoices);
	}
      re_optimize_nfa (nfa);
      if (nfa->n < nfa->sz) /* trim nfa->q state table, to save space */
	{
	  nfa->sz = nfa->n;
	  nfa->q = RE_REALLOC(nfa->q, nfa->sz*sizeof(struct re_state_s));
	}
      nfa->buf1 = RE_MALLOC(nfa->n*sizeof(re_nfa_thread_t));
      nfa->buf2 = RE_MALLOC(nfa->n*sizeof(re_nfa_thread_t));
    }
  else nfa = NULL;
  
  nfa_actions->nfa = nfa;
  e = MakeExternal (&nfaExtDesc, (char *)nfa_actions);
  return e;
}
ACTION_END

static re_nfa_env_t *re_nfa_match_add (re_nfa_t *nfa,
				       re_nfa_env_t *env, size_t idx, size_t pos)
{
  size_t sz, newsz, n;
  re_nfa_env_t *oldenv;
  /* We shall reallocate an re_nfa_env_t, and store start at
     index 2*nmatch in it. */

  if (env==NULL)
    {
      n = 0;
      sz = idx+2; /* preallocate two entries (start, end) directly */
      env = RE_MALLOC(offsetof(re_nfa_env_t, pos[sz]));
      env->refcnt = 1;
      env->sz = sz;
    }
  else if (env->refcnt>1) /* do an actual copy */
    {
      n = env->nmatches;
      sz = env->sz;
      if (sz<=idx)
	newsz = idx+2; /* make sure to preallocate two entries (start, end) directly */
      else newsz = sz;
      oldenv = env;
      env = RE_MALLOC(offsetof(re_nfa_env_t, pos[newsz]));
      oldenv->refcnt--;
      env->refcnt = 1;
      env->nmatches = n;
      env->sz = newsz;
      memcpy (env->pos, oldenv->pos, sz*sizeof(size_t));
    }
  else
    {
      n = env->nmatches;
      sz = env->sz;
      if (sz <= idx)
	{
	  sz <<= 1; /* double size by default, */
	  newsz = idx+2; /* but make sure to preallocate two entries (start, end) directly */
	  if (sz<newsz)
	    sz = newsz;
	  env = RE_REALLOC(env, offsetof(re_nfa_env_t, pos[sz]));
	  env->sz = sz;
	}
    }
  /*
  if (env!=NULL)
    memcpy (newenv->pos, env->pos, n*sizeof(re_nfa_env_t));
  */
  if (n<=idx)
    {
      env->nmatches = idx+1;
      for (; n<idx; n++)
	env->pos[n] = (n & 1)?SIZE_MAX:0;
    }
  env->pos[idx] = pos;
  return env;
}

#define RE_NFA_MATCH_START(nfa,env,nmatch,start) re_nfa_match_add(nfa,env,((size_t)nmatch) << 1,start)
#define RE_NFA_MATCH_END(nfa,env,nmatch,end) re_nfa_match_add(nfa,env,(((size_t)nmatch) << 1)+1,end)

static int re_nfa_env_compare (re_nfa_env_t *env, re_nfa_env_t *newenv)
{
  size_t i, n;
  int cmp;

  if (env==NULL)
    return 1;
  if (newenv==NULL)
    return -1;
  n = env->nmatches;
  cmp = newenv->nmatches - n;
  if (cmp)
    return cmp;
  for (i=0; i<n; i++)
    {
      cmp = newenv->pos[i] - env->pos[i];
      if (cmp)
	return (i&1)?cmp:(-cmp); /* prefer largest end values (i.e., i&1!=0),
				    and smallest start values (i.e., i&1==0). */
    }
  return 0;
}

static int re_nfa_insert_state (re_nfa_t *nfa, re_state_num q, re_nfa_env_t *env,
				size_t pos, size_t len)
/* return least number of an accepting state, or INT_MAX if no state accepts */
{
  re_state_t *st;
  re_nfa_thread_t *thr;
  int res = 0;
  struct re_directed_s *dir;
  char cur_presence_mask;

 again:
  st = &nfa->q[q];
  switch (st->kind)
    {
    case RE_S_SPLIT:
      res |= re_nfa_insert_state (nfa, st->next,
				  RE_NFA_ENV_COPY (env), pos, len);
      res |= re_nfa_insert_state (nfa, st->what.alt, env, pos, len);
      return res;
    case RE_S_ACCEPT:
      if (st->what.nbranch < nfa->final_branch)
	{
	  nfa->final_branch = st->what.nbranch;
	  RE_NFA_ENV_FREE(nfa->final_env);
	  nfa->final_env = env;
	}
      else if (st->what.nbranch == nfa->final_branch)
	{
	  int cmp = re_nfa_env_compare (nfa->final_env, env);

	  if (cmp > 0) /* if new environment is better, then update */
	    {
	      RE_NFA_ENV_FREE(nfa->final_env);
	      nfa->final_env = env;
	      nfa->final_branch = st->what.nbranch;
	    }
	  else RE_NFA_ENV_FREE(env);
	}
      return 1;
    case RE_S_MEMO_START:
      env = RE_NFA_MATCH_START (nfa, env, st->what.memo.nmatch, pos);
      q = st->next;
      goto again;
    case RE_S_MEMO_END:
      env = RE_NFA_MATCH_END (nfa, env, st->what.memo.nmatch, pos);
      q = st->next;
      goto again;
    case RE_S_CHECK_END:
      if (pos==len)
	{
	  if (st->what.nbranch < nfa->final_branch)
	    {
	      nfa->final_branch = st->what.nbranch;
	      RE_NFA_ENV_FREE(nfa->final_env);
	      nfa->final_env = env;
	    }
	  else if (st->what.nbranch == nfa->final_branch)
	    {
	      int cmp = re_nfa_env_compare (nfa->final_env, env);

	      if (cmp > 0) /* if new environment is better, then update */
		{
		  RE_NFA_ENV_FREE(nfa->final_env);
		  nfa->final_env = env;
		  nfa->final_branch = st->what.nbranch;
		}
	      else RE_NFA_ENV_FREE(env);
	    }
	  return 1;
	}
      RE_NFA_ENV_FREE(env);
      return res;
    case RE_S_DIRECTED_SPLIT:
      if (pos!=len)
	goto push;
      /* at the end of string, do as though we had read
	 the special 'end of string' character, 256 */
      dir = st->what.dir;
      if (TEST(dir->next_bitfield, 256)!=0)
	{
	  if (TEST(dir->alt_bitfield, 256)!=0)
	    { /* may match both branches */
	      res |= re_nfa_insert_state (nfa, st->next,
					  RE_NFA_ENV_COPY (env),
					  pos, len);
	      q = st->what.dir->alt;
	      goto again;
	    }
	  q = st->next;
	  goto again;
	}
      if (TEST(dir->alt_bitfield, 256)!=0)
	{
	  q = st->what.dir->alt;
	  goto again;
	}
      RE_NFA_ENV_FREE(env);
      return res;
    case RE_S_EXCLUSIVE_SPLIT:
      if (pos!=len)
	goto push;
      dir = st->what.dir;
      if (TEST(dir->next_bitfield, 256)!=0)
	{
	  q = st->next;
	  goto again;
	}
      if (TEST(dir->alt_bitfield, 256)!=0)
	{
	  q = st->what.dir->alt;
	  goto again;
	}
      RE_NFA_ENV_FREE(env);
      return res; /* ignore this state, as above */
    default: /* RE_S_MATCH_CHAR, RE_S_MATCH_RANGE */
      if (pos==len) /* will never match end of string */
	{
	  RE_NFA_ENV_FREE(env);
	  return res;
	}
    push:
      if (pos & 1)
	cur_presence_mask = RE_PRESENCE_1;
      else
	cur_presence_mask = RE_PRESENCE_2;
      if (st->presence & cur_presence_mask)
	{ /* already inserted; then find it, and compare its environment */
	  thr = &nfa->cur[st->lastidx];
	  if (re_nfa_env_compare (thr->env, env) > 0) /* if new environment is better, then update */
	    {
	      RE_NFA_ENV_FREE(thr->env);
	      thr->env = env;
	    }
	  else RE_NFA_ENV_FREE(env);
	  return res;
	}
      thr = &nfa->cur[nfa->ncur];
      thr->q = q;
      thr->env = env;
      st->presence |= cur_presence_mask;
      st->lastidx = nfa->ncur++;
      return res;
    }
}

static int re_nfa_init (re_nfa_t *nfa, size_t len)
{
#ifdef NFA_DEBUG
  size_t i;
#endif
  
  if (nfa->final_env!=NULL)
    {
      RE_NFA_ENV_FREE(nfa->final_env);
      nfa->final_env = NULL;
    }
  nfa->final_branch = INT_MAX;
#ifdef NFA_DEBUG
  for (i=0; i<nfa->n; i++)
    if (nfa->q[i].presence)
      abort();
#endif
  nfa->cur = nfa->buf1;
  nfa->prev = nfa->buf2;
  nfa->ncur = 0;
  nfa->nprev = 0;
  /*
  for (i=0, n=nfa->n; i<n; i++)
    nfa->q[i].lastpos = -1;
  */
  return re_nfa_insert_state (nfa, nfa->start, NULL, 0, len);
}

static int re_nfa_goto (re_nfa_t *nfa, size_t pos, size_t len, char c,
			re_state_num q, re_nfa_env_t *env)
{
  re_state_t *st;
  struct re_directed_s *dir;
  int accept = 0;

 again:
  st = &nfa->q[q];
  switch (st->kind)
    {
    case RE_S_MATCH_CHAR:
#ifdef NFA_DEBUG
      fprintf (stderr, "state %zd, expecting letter '%c'", q, st->what.letter);
#endif
      if (st->what.letter!=c)
	    {
#ifdef NFA_DEBUG
	      fprintf (stderr, ": no\n");
#endif
	      RE_NFA_ENV_FREE(env);
	      return accept; /* ignore this state: go through the transition only
			    if the letters match */
	    }
      break;
    case RE_S_MATCH_RANGE:
#ifdef NFA_DEBUG
      fprintf (stderr, "state %zd, expecting range", q);
#endif
      if (TEST(st->what.range, (unsigned int)(unsigned char)c)==0)
	{
#ifdef NFA_DEBUG
	  fprintf (stderr, ": no\n");
#endif
	  RE_NFA_ENV_FREE(env);
	  return accept; /* ignore this state, as above */
	}
      break;
    case RE_S_EXCLUSIVE_SPLIT:
#ifdef NFA_DEBUG
      fprintf (stderr, "state %zd: exclusive split, ", q);
#endif
      dir = st->what.dir;
      if (TEST(dir->next_bitfield, (unsigned int)(unsigned char)c)!=0)
	{
	  q = st->next;
#ifdef NFA_DEBUG
	  fprintf (stderr, "goto %zd (next).\n", q);
#endif
	  goto again;
	}
      if (TEST(dir->alt_bitfield, (unsigned int)(unsigned char)c)!=0)
	{
	  q = st->what.dir->alt;
#ifdef NFA_DEBUG
	  fprintf (stderr, "goto %zd (alt).\n", q);
#endif
	  goto again;
	}
#ifdef NFA_DEBUG
      fprintf (stderr, " no match.\n");
#endif
      RE_NFA_ENV_FREE(env);
      return accept; /* ignore this state, as above */
    case RE_S_DIRECTED_SPLIT:
#ifdef NFA_DEBUG
      fprintf (stderr, "state %zd: directed split, ", q);
#endif
      dir = st->what.dir;
      if (TEST(dir->next_bitfield, (unsigned int)(unsigned char)c)!=0)
	{
	  if (TEST(dir->alt_bitfield, (unsigned int)(unsigned char)c)!=0)
	    { /* may match both branches */
#ifdef NFA_DEBUG
	      fprintf (stderr, "either goto %zd (next).\n", st->next);
#endif
	      accept |= re_nfa_goto (nfa, pos, len, c, st->next,
				     RE_NFA_ENV_COPY (env));
#ifdef NFA_DEBUG
	      fprintf (stderr, "state %zd: directed split, or goto %zd (alt).\n", q, st->what.dir->alt);
#endif
	      q = st->what.dir->alt;
	      goto again;
	    }
#ifdef NFA_DEBUG
	  fprintf (stderr, "goto %zd (next).\n", st->next);
#endif
	  q = st->next;
#ifdef NFA_DEBUG
	  fprintf (stderr, "goto %zd (next).\n", q);
#endif
	  goto again;
	}
      if (TEST(dir->alt_bitfield, (unsigned int)(unsigned char)c)!=0)
	{
	  q = st->what.dir->alt;
#ifdef NFA_DEBUG
	  fprintf (stderr, "goto %zd (alt).\n", q);
#endif
	  goto again;
	}
#ifdef NFA_DEBUG
      fprintf (stderr, " no match.\n");
#endif
      RE_NFA_ENV_FREE(env);
      return accept; /* ignore this state, as above */
    case RE_S_MEMO_START:
      env = RE_NFA_MATCH_START (nfa, env, st->what.memo.nmatch, pos-1);
      q = st->next;
      goto again;
    case RE_S_MEMO_END:
      env = RE_NFA_MATCH_END (nfa, env, st->what.memo.nmatch, pos-1);
      q = st->next;
      goto again;
    case RE_S_SPLIT:
#ifdef NFA_DEBUG
      fprintf (stderr, "state %zd: split, either goto %zd (next).\n", q, st->next);
#endif
      accept |= re_nfa_goto (nfa, pos, len, c, st->next,
			     RE_NFA_ENV_COPY (env));
#ifdef NFA_DEBUG
      fprintf (stderr, "state %zd: split, or goto %zd (alt).\n", q, st->what.alt);
#endif
      q = st->what.alt;
      goto again;
    case RE_S_CHECK_END: case RE_S_ACCEPT:
      accept |= re_nfa_insert_state (nfa, q, env, pos, len);
      return accept;
    default:				    
#ifdef NFA_DEBUG
      fprintf (stderr, "Oops, unknown kind: %d.\n", (int)st->kind);
#endif
      return accept;
    }
#ifdef NFA_DEBUG
  fprintf (stderr, ": match\n");
#endif
  accept |= re_nfa_insert_state (nfa, st->next, env, pos, len);
  return accept;
}

static int re_nfa_advance (re_nfa_t *nfa, size_t pos, size_t len, char c)
{
  size_t i, n;
  int res;
  re_state_num q;
  re_nfa_thread_t *thr;
  re_nfa_thread_t *swp;
  re_nfa_env_t *env;
  char prev_neg_presence_mask;
  
  if (pos & 1)
    prev_neg_presence_mask = ~RE_PRESENCE_2;
  else
    prev_neg_presence_mask = ~RE_PRESENCE_1;
  /* swap cur and prev */
  swp = nfa->prev;
  nfa->prev = nfa->cur;
  nfa->cur = swp;
  nfa->nprev = nfa->ncur;
  nfa->ncur = 0;
  res = 0;
#ifdef NFA_DEBUG
  fprintf (stderr, "read '%c', %zd states in prev queue\n", c, nfa->nprev);
#endif
  for (i=0, n=nfa->nprev; i<n; i++)
    {
      thr = &nfa->prev[i];
      q = thr->q;
      env = thr->env;
      nfa->q[q].presence &= prev_neg_presence_mask;
      res |= re_nfa_goto (nfa, pos, len, c, q, env);
    }
  return res;
}

static void re_nfa_clean_presence (re_nfa_t *nfa)
{
  size_t i, n;
  re_nfa_thread_t *thr;
  re_state_num q;
  re_state_t *st;

  n = nfa->ncur;
  for (i=0; i<n; i++)
    {
      thr = &nfa->cur[i];
      q = thr->q;
      st = &nfa->q[q];
      st->presence = 0;
    }
#ifdef NFA_DEBUG
  for (i=0; i<nfa->n; i++)
    if (nfa->q[i].presence)
      abort();
#endif
}

int re_nfa_run (re_nfa_t *nfa, char *s, size_t len)
{
  int res;
  size_t i;
  char c;

  if (nfa==NULL)
    return 0;
  res = re_nfa_init (nfa, len);
  if (res)
    {
    accept:
      re_nfa_clean_buf (nfa->cur, nfa->ncur);
      re_nfa_clean_presence (nfa);
      nfa->cur = NULL;
      nfa->prev = NULL;
      /* so only nfa->final_env is the only remaining allocated env */
#ifdef NFA_DEBUG
  for (i=0; i<nfa->n; i++)
    if (nfa->q[i].presence)
      abort();
#endif
      return res;
    }
  for (i=0; i<len; )
    {
      c = s[i++];
      res = re_nfa_advance (nfa, i, len, c);
      if (res)
	goto accept;
      if (nfa->ncur==0) /* no need to read the rest: will not match */
	break;
    }
#ifdef NFA_DEBUG
  for (i=0; i<nfa->n; i++)
    if (nfa->q[i].presence)
      abort();
#endif
  return res;
}

re_nfa_t *re_make_one_nfa (char *regexp, size_t len, int *errp)
{
  regex_t *re;
  re_nfa_t *nfa;

  re = re_parse (regexp, len, errp);
  if (*errp!=0)
    return NULL;
  nfa = re_make_nfa (1, &re);
  re_free (re);
  return nfa;
}

DEF_ACTION_1ARG(reParse)
{
  regex_t *re;
  int err;
  ExprPtr e;

  err = 0;
  re = re_parse((char *)STEXT(arg),SLEN(arg),&err);
  if (re==NULL)
    return raiseREExc(err);
  e = MakeExternal(&regexExtDesc, (char *)re);
  return e;
}
ACTION_END

DEF_ACTION_1ARG(reParseFloat)
{
  regex_t *re;
  size_t pos;
  int err, group;
  ExprPtr e;

  err = 0;
  pos = 0;
  group = 0;
  re = re_parse_more_control ((char *)STEXT(arg),SLEN(arg),&err,&group,&pos,1);
  if (re==NULL)
    return raiseREExc(err);
  e = MakeExternal(&regexExtDesc, (char *)re);
  return e;
}
ACTION_END

/* nfa_run : 'a nfa * string -> 'a option */
DEF_ACTION_2ARGS(reNfaRun)
{
  re_nfa_actions_t *nfa_actions;
  ExprPtr action;
  re_nfa_env_t *env;
  INT *pos;
  ExprPtr matches, e;
  int n;
  COUNTER i, nmatches;
  ENTER;

  nfa_actions = (re_nfa_actions_t *)EXTDATA(arg1);
  if (nfa_actions->nre==0)
    e = none;
  else
    {
      n = re_nfa_run (nfa_actions->nfa, (char *)STEXT(arg2), SLEN(arg2));
      if (n==0)
	e = none;
      else
	{
	  n = re_matching_branch (nfa_actions->nfa);
	  action = nfa_actions->actions[n];
	  env = re_matches (nfa_actions->nfa);
	  nmatches = env->nmatches;
	  pos = (INT *)MALLOC((nmatches+1)*sizeof(INT));
	  pos[0] = nmatches;
	  for (i=0; i<nmatches; i++)
	    pos[i+1] = env->pos[i];
	  matches = MakeExternal(&iarrayExtDesc, (char *)pos);
	  KEEP(matches);
	  e = (*ACTION(action)) (ENV_CONS(QENV(action),2), arg2, matches);
	  KEEPONLY(e);
	  e = Kdata(0,optttm,e); /* SOME e */
	}
    }
  QUIT;
  return e;
}
ACTION_END

/* nfa_run_substr : 'a nfa * string * int * int -> 'a option */
DEF_ACTION_4ARGS(reNfaRunSubstr)
{
  re_nfa_actions_t *nfa_actions;
  ExprPtr action;
  re_nfa_env_t *env;
  INT *pos;
  ExprPtr matches, e;
  int n;
  COUNTER i, nmatches;
  INT start,end;
  ENTER;

  nfa_actions = (re_nfa_actions_t *)EXTDATA(arg1);
  if (nfa_actions->nre==0)
    e = none;
  else
    {
      start = IVAL(arg3);
      end = IVAL(arg4);
      if (start<0)
	e = Raise(stringnthException);
      else if (end>SLEN(arg1))
	e = Raise(stringnthException);
      else
	{
	  if (end<start)
	    end = start;
	  n = re_nfa_run (nfa_actions->nfa, ((char *)STEXT(arg2))+start,
			  end-start);
	  if (n==0)
	    e = none;
	  else
	    {
	      n = re_matching_branch (nfa_actions->nfa);
	      action = nfa_actions->actions[n];
	      env = re_matches (nfa_actions->nfa);
	      nmatches = env->nmatches;
	      pos = (INT *)MALLOC((nmatches+1)*sizeof(INT));
	      pos[0] = nmatches;
	      for (i=0; i<nmatches; i++)
		pos[i+1] = env->pos[i]+start;
	      matches = MakeExternal(&iarrayExtDesc, (char *)pos);
	      KEEP(matches);
	      e = (*ACTION(action)) (ENV_CONS(QENV(action),2), arg2, matches);
	      KEEPONLY(e);
	      e = Kdata(0,optttm,e); /* SOME e */
	    }
	}
    }
  QUIT;
  return e;
}
ACTION_END

/* re_subst : string * intarray * int -> string */
/* equivalent to:
   fun re_subst (s,a,i) =
       substr (s, isub(a,2*i), isub(a,2*i+1));
 */
DEF_ACTION_3ARGS(resubst)
{
  INT i,k,n,start,end;
  INT *ip;
  ExprPtr val;

  ip = (INT *)EXTDATA(arg2);
  n = ip[0];
  i = IVAL(arg3);
  k = 2*i+1;
  if (i<0 || k>=n)
    return Raise(subscriptException);
  start = ip[k];
  end = ip[k+1];
  if (start<0 || end>SLEN(arg1))
    return Raise(stringnthException);
  if (end<start)
    end = start;
  val = Kstring(end-start,&STEXT(arg1)[start]);
  return val;
}
ACTION_END
  
DEF_ACTION_1ARG(remsg)
{
  INT i;

  i = IVAL(arg);
  if (i<1 || i>=RE_PARSE_NERRS)
    return voidString;
  return Kstr((CHAR *)message(548+i));
}
ACTION_END

PRIVATE ExprPtr KSYMBOL(n)
IN COUNTER n;
{
  ExprPtr sym;

  sym = Ksymbol(message(n));
  return sym;
}

PUBLIC void InitRE()
{
  SymPtr regexSym, nfaSym;
  ExprPtr regexType, nfaType, optType, type;
  ENTER;

  registerExtDesc(&regexExtDesc);
  registerExtDesc(&nfaExtDesc);
  /* re_parse : string -> regex */
  regexSym = KSYMBOL(970);
  regexType = KtypeAppl(regexSym,NIL);
  type = funType(stringType,regexType);
  ADDFUN(971,type,reParse);

  nfaSym = KSYMBOL(973);
  nfaType = KtypeAppl(nfaSym,Kons(alpha,NIL));
  /* re_make_nfa : (regex * (string * intarray -> 'a)) list -> 'a nfa */
  type = funType(coupleType(stringType,iarrayType),alpha);
  type = funType(listType(coupleType(regexType,type)),nfaType);
  ADDFUN(974,type,reMakeNfa);

  /* nfa_run : 'a nfa * string -> 'a option */
  optType = KtypeAppl(optSym,Kons(alpha,NIL));
  type = funType(coupleType(nfaType,stringType),optType);
  ADDFUN(975,type,reNfaRun);
  /* nfa_run_substr : 'a nfa * string * int * int -> 'a option */
  type = Kons(nfaType,Kons(stringType,coupleType(intType,intType)));
  type = funType(type,optType);
  ADDFUN(972,type,reNfaRunSubstr);

  /* re_subst : string * intarray * i -> string */
  type = Kons(stringType,coupleType(iarrayType,intType));
  type = funType(type,stringType);
  ADDFUN(976,type,resubst);
  QUIT;
}

/* faire re_msg : int -> string */
 
