/* GimML regexp headers.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/**
 ** Error codes.  See re_parse().
 **/
#define RE_PARSE_PREMATURE_END 1
#define RE_PARSE_EXPECTED_CLOSING_BRACE 2
#define RE_PARSE_EXPECTED_COMMA_OR_CLOSING_BRACE 3
#define RE_PARSE_WRONG_BOUND 4
#define RE_PARSE_BOUND_TOO_HIGH 5
#define RE_PARSE_EXPECTED_CLOSING_PARENTHESIS 6
#define RE_PARSE_NO_LETTER_AFTER_BACKSLASH 7
#define RE_PARSE_SUBMATCH_DISCREPANCY 8
#define RE_PARSE_NERRS 9

/**
 ** Two ways of using the re regular expression library.
 ** 1. The simple way, to match just one regexp:
 ** call re_make_one_nfa() to build a non-deterministic finite automaton (NFA),
 ** then re_nfa_run() to match it against as many strings as you wish;
 ** when you are finished, free the NFA by calling re_nfa_free().
 ** 2. The comprehensive way, to match several regexps at once:
 ** call re_parse() on each one of your regexps: you get a pointer,
 ** of type regex_t *, to something that is essentially a abstract
 ** syntax tree for each;
 ** put the resulting regex_t * pointers in a table re[], which you pass
 ** to re_make_nfa(), getting one (just one) NFA;
 ** free all the regex_t pointers, by calling re_free() on each and every
 ** element of re[];
 ** use re_nfa_run() as above, and finish by calling re_nfa_free().
*/

/**
 ** @struct regex_s
 ** The type of abstract syntax trees for regular expressions.
 **/
typedef struct regex_s regex_t;

/** 
 ** Given a regular expression regexp, of length len, parse it
 ** and build an abstract syntax tree, of type regex_t *.
 ** If an error occurs, the error code will be put into *errp,
 ** and NULL will be returned.
 ** Error codes vary between 1 and RE_PARSE_NERRS-1, and
 ** are defined by the RE_PARSE_* macros above in this file.
 ** If no error occurs, then *errp will be set to 0, and
 ** a valid regex_t * is returned.
 ** NUL characters are not taken as terminators in regexp.
 **
 ** @param regexp The regular expression to parse.
 ** @param len The length of the regular expression to parse.
 ** @param errp Pointer to an int that will hold the error code, or 0, on return.
*/
regex_t *re_parse (char *regexp, size_t len, int *errp);

/**
 ** A version of re_parse() that gives slightly more control.
 **
 ** The additional parameter group_n is a pointer to an int that you set
 ** prior to calling re_parse_more_control(); all submatches will be numbered starting
 ** from that int; on exit, *group_n will hold the first unused submatch number.
 ** In particular, if you set *group_n to 0 before calling re_parse_more_control(),
 ** then it will hold the number of submatches in the regexp on exit.
 **
 ** If group_n is the NULL pointer, then re_parse_more_control() will parse
 ** the regexp and interpret no (...) group as a submatch, rather as just
 ** ordinary, parenthesized regexps.
 **
 ** The additional parameter does_caret, if non-zero, will allow re_parse()
 ** to recognize ^ at the beginning of a branch (see man re_format for what
 ** a branch is) and $ at the end of a branch.  In that case, ^ is interpreted
 ** as matching the beginning of the string, and $ matches the end of the string.
 ** If does_caret is zero, ^ and $ are handled as though they were normal characters.
 ** 
 ** @param regexp The regular expression to parse.
 ** @param len The length of the regular expression to parse.
 ** @param errp Pointer to an int that will hold the error code, or 0, on return.
 ** @param group_n Pointer to an int holding submatch numbers.
 ** @param posp Pointer to an int that holds position where parsing stopped.
 ** @param does_caret Control the meaning of ^ and $.
 **/
regex_t *re_parse_more_control (char *regex, size_t len, int *err,
				int *group_n, size_t *posp, int does_caret);

/** re_free(re)
 ** Free the space used by re.
 **
 ** @param re The regular expression to free.
 **/
void re_free (regex_t *re);

/**
 ** @struct re_nfa_s
 ** Non-deterministic finite automata (NFA).
 **/
typedef struct re_nfa_s re_nfa_t;

/**
 ** Wrapper function, which calls re_parse(), re_make_nfa() and re_free()
 ** so as to convert a regular expression regexp, of size len,
 ** directly into an NFA.
 ** For error codes, see re_parse().
 ** NUL characters are not taken as terminators in regexp.
 **
 ** @param regexp The regular expression to parse.
 ** @param len The length of the regular expression to parse.
 ** @param errp Pointer to an int that will hold the error code, or 0, on return.
 **/
re_nfa_t *re_make_one_nfa (char *regexp, size_t len, int *errp);

/**
 ** Convert an array re of regex_t * pointers, with nre entries,
 ** into a single, combined NFA, which is then returned.
 **
 ** @param nre The number of elements in the array re[].
 ** @param re The array of regex_t * pointers.
 **/
re_nfa_t *re_make_nfa (size_t nre, regex_t *re[]);

/**
 ** Run the NFA nfa on input string s.  The string s contains
 ** len characters.  NUL characters are not taken as terminators in s.
 ** Return a non-zero number if s is recognized by nfa, 0 otherwise.

 ** On acceptance, and assuming that nfa was produced by re_make_nfa(nre,re),
 ** re_matching_branch(nfa) will return the number i of the first
 ** regex_t * re[i] which recognizes s, and
 ** re_matches(nfa) will return a pointer to a structure describing
 ** the positions of the submatch boundaries, or NULL if there is no submatch.
 **
 ** @param nfa The input NFA.
 ** @param s The string to be matched.
 ** @param len The length of the string to be matched.
 **/

int re_nfa_run (re_nfa_t *nfa, char *s, size_t len);
/**
 ** Free all memory used by nfa.
 ** This includes the structure returned by re_matches(), in particular,
 ** so that calling re_matches(nfa) - or even re_matching_branch(nfa) -
 ** after calling re_nfa_free(nfa) is meaningless.
 **
 ** @param nfa The NFA to free.
 **/
void re_nfa_free (re_nfa_t *nfa);

/**
 ** @struct re_nfa_env_s
 ** Structures that hold information about submatches, after a succesful match.
 **/
typedef struct re_nfa_env_s re_nfa_env_t;
struct re_nfa_env_s {
  size_t refcnt;
  size_t nmatches, sz;
  size_t pos[10]; /* actually pos[sz], of which only pos[nmatches] are filled in */
};

/**
 ** Called after re_nfa_run(nfa,s,len), and assuming that re_nfa_run() returned
 ** non-zero and that nfa was produced by re_make_nfa(nre,re),
 ** re_matching_branch(nfa) will return the number i of the first
 ** regex_t * re[i] which recognizes s.
 ** Returns nonsense if re_nfa_run() was not run or if re_nfa_run()
 ** returned 0.
 **
 ** @param nfa The input NFA.
 **/
int re_matching_branch (re_nfa_t *nfa);

/**
 ** Called after re_nfa_run(nfa,s,len), and assuming that re_nfa_run() returned
 ** non-zero and that nfa was produced by re_make_nfa(nre,re),
 ** re_matches(nfa) returns a pointer to a structure of type re_nfa_env_t
 ** holding the boundaries of all submatches - or NULL, if there is no submatch.
 ** That structure is allocated by re_nfa_run(), and is no longer valid
 ** after a further call to re_nfa_run(...,nfa,...) or a call to
 ** re_nfa_free(nfa).
 ** 
 ** Calling env the returned pointer, env->nmatches will always be an
 ** even natural number, equal to twice the number of submatches.
 ** For example, in a(b(.*)d)e(f), there are three submatches, and
 ** correspondingly, env->nmatches will equal 6.

 ** env->pos[0] and env->pos[1] will always hold the start and end positions,
 ** respectively of the whole match.  Hence if you match the above regexp
 ** against the string zzzabkkkdefggg, those values will be equal to 3 and
 ** 10 respectively.

 ** Contrarily to most regexp libraries, re_nfa_run() stops on the
 ** shortest, not longest match (among those which start earliest).
 ** Hence if you match the same regexp against
 ** abdefzbdef, you will get 0 and 5, not 0 and 10.

 ** Then env->pos[2] and env->pos[3] will hold the start and end positions
 ** of the first submatch, env->pos[4] and env->pos[5] those of the second
 ** submatch, etc.
 ** For example, if you run the above regexp against zzzabkkkdefggg, you
 ** will obtain 3 and 9 (matching abkkkd) for the first submatch,
 ** 5 and 8 (matching kkk) for the second submatch, and
 ** 10 and 11 (matching f) for the third submatch.

 ** If there are several ways of mapping submatches to positions (among all
 ** those matches that start and end earliest), only one will be returned.
 ** It should be one that tries to make the first submatch start earliest
 ** and end latest, and once the first submatch is fixed, it chooses
 ** the second submatch to start earliest and to end latest, and so on.
 **
 ** @param nfa The input NFA.
 */
re_nfa_env_t *re_matches (re_nfa_t *nfa);
