/* Line editing.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "type.h"
#include "exeproto.h"
#include <fcntl.h>
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_READLINE
#include "linenoise.h"
#else
#include "file.h"
IMPORT ExprPtr readline P((INOUT hFILE *file));
#endif

EXTERN ExprPtr voidString;
EXTERN ExprPtr stringType,intType,boolType;
EXTERN ExprPtr optSym;

IMPORT ExprPtr Select P((IN CONST ExprPtr label, IN CONST ExprPtr arg)); /* fromexec.c */

PRIVATE ExprPtr promptsym, completionsym, hintssym, maskedsym, multilinesym,
  historysym;

#ifdef HAS_READLINE
PRIVATE ExprPtr do_lineedit_completion = NULL;
/* a completion function;
   the type of completions is:
   string * string -> (string * string * string) list
*/
PRIVATE ExprPtr do_lineedit_hints = NULL;
/* a hints function, of type:
   string -> (string * int * bool) option
 */

PRIVATE void lineeditCompletion(buf,lc,pos)
const char *buf;
linenoiseCompletions *lc;
size_t pos;
{
  size_t len;
  ExprPtr prefix, suffix, completions, l;
  VecPtr vp;
  ENTER;

  if (do_lineedit_completion!=NULL)
    {
      len = strlen(buf);
      prefix = Kstring(pos,(CHAR *)buf);
      (void) KEEP (prefix);
      suffix = Kstring(len-pos,(CHAR *)&buf[pos]);
      (void) KEEP (suffix);
      completions = (*ACTION(do_lineedit_completion)) (ENV_CONS(QENV(do_lineedit_completion),2), prefix, suffix);
      KEEPONLY(completions);
      for (l=completions; l!=NIL; l=CDR(l))
	{
	  vp = AVAL(CAR(l));
	  linenoiseAddCompletionLen(lc,(const char *)STEXT(vp->expr[0]),
				    SLEN(vp->expr[0]),
				    (const char *)STEXT(vp->expr[1]),
				    SLEN(vp->expr[1]),
				    (const char *)STEXT(vp->expr[2]),
				    SLEN(vp->expr[2]));
	}
    }
  QUIT;
}

PRIVATE char *lineeditHints(buf,color,bold)
     const char *buf;
     int *color;
     int *bold;
{
  char *hint;
  size_t len;
  ExprPtr buffer, res;
  VecPtr vp;
  ENTER;

  hint = NULL;
  if (do_lineedit_hints!=NULL)
    {
      len = strlen(buf);
      buffer = Kstring(len,(CHAR *)buf);
      (void) KEEP (buffer);
      res = (*ACTION(do_lineedit_hints)) (ENV_CONS(QENV(do_lineedit_hints),1), buffer);
      if (res->magic==0) /* SOME */
	{
	  res = CDR(res); /* : string * int * bool */
	  vp = AVAL(res);
	  *color = (int)IVAL(vp->expr[1]);
	  if (vp->expr[2]==trueValue)
	    *bold = 1;
	  else *bold = 0;
	  len = SLEN(vp->expr[0]);
	  hint = (char *)MALLOC(len+1);
	  memcpy (hint, STEXT(vp->expr[0]), len);
	  hint[len] = '\0';
	}
    }
  QUIT;
  return hint;
}

PRIVATE void lineeditFreeHints(p)
     void *p;
{
  MARKFREE(p);
}

PRIVATE ExprPtr lineeditGetHistory()
{
  linenoiseHistory *lh;
  char *h;
  ExprPtr l, s, m, res;
  int j, len;
  INT slen;
  ENTER;

  lh = linenoiseHistoryExport();
  len = lh->len;
  for (l=NIL, j=len; --j>=0; )
    {
      h = lh->history[j];
      slen = strlen(h);
      h[slen] = '\n';
      s = Kstring1(slen+1,(CHAR *)h,false);
      (void) KEEP(s);
      l = Kons (s, l);
      KEEPONLY(l);
    }
  linenoiseHistoryFree(lh,0);
  m = KINT(lh->max_len);
  res = Ktuple2(m,l);
  QUIT;
  return res;
}

PRIVATE void lineeditSetHistory(history)
     IN CONST ExprPtr history;
{
  ExprPtr m, l, s;
  int max_len, j;
  linenoiseHistory *lh;
  char *h;
  INT i, len;
  CHAR *text;

  m = CAR(history);
  max_len = IVAL(m);
  lh = linenoiseHistoryAlloc(max_len);
  for (l = CDR(history), j=0; j<max_len && l!=NIL; l=CDR(l), j++)
    {
      s = CAR(l);
      /* find length of string s, chopped off at first newline if any */
      for (i=0, len=SLEN(s), text=STEXT(s); i<len && text[i]!='\n'; i++);
      h = (char *)MALLOC(i+1);
      memcpy(h,text,i);
      h[i] = '\0';
      lh->history[j] = h;
    }
  lh->len = j;
  linenoiseHistoryImport(lh);
}
#endif

/* readline :
   |[prompt : string,
     completion : (string * string -> (string * string * string) list) option,
     hints : (string -> (string * int * bool) option) option,
     history : (int * string list) ref option,
     masked : bool,
     multiline : bool
     ]| -> string
*/

DEF_ACTION_1ARG(lineedit)
{
  ExprPtr prompt;
  char *prompt_s;
  
  prompt = Select(promptsym,arg);
  prompt_s = (char *)alloctmp_anchored(SLEN(prompt)+1,prompt_s);
  memcpy(prompt_s,STEXT(prompt),SLEN(prompt));
  prompt_s[SLEN(prompt)] = '\0';

#ifdef HAS_READLINE
  ExprPtr completion, hints, masked, multiline, history, href, e;
  char *result;
  ExprPtr res;
  COUNTER len;
  linenoiseHistory *lh;

  completion = Select(completionsym,arg);
  if (completion->magic==0) /* SOME */
    {
      do_lineedit_completion = CDR(completion);
      linenoiseSetCompletionCallback(lineeditCompletion);
    }
  else linenoiseSetCompletionCallback(NULL);
  hints = Select(hintssym,arg);
  if (hints->magic==0) /* SOME */
    {
      do_lineedit_hints = CDR(hints);
      linenoiseSetHintsCallback(lineeditHints);
    }
  else linenoiseSetHintsCallback(NULL);
  linenoiseSetFreeHintsCallback(NULL);
  masked = Select(maskedsym,arg);
  if (masked==trueValue)
    linenoiseMaskModeEnable();
  else linenoiseMaskModeDisable();
  multiline = Select(multilinesym,arg);
  linenoiseSetMultiLine((multiline==trueValue));
  history = Select(historysym,arg);
  lh = NULL;
  href = NIL;
  if (history->magic==0) /* SOME */
    {
      lh = linenoiseHistoryExport();
      href = CDR(history);
      e = REFVAL(href);
      lineeditSetHistory(e);
    }
  result = linenoise(prompt_s);
  res = voidString;
  if (result)
    {
      len = strlen(result);
      if (masked!=trueValue)
	{	
	  result[len-1] = '\0';
	  linenoiseHistoryAdd(result);
	  result[len-1] = '\n';
	}
      res = Kstring1(len,(CHAR *)result,false);
    }
  if (history->magic==0) /* SOME */
    {
      e = lineeditGetHistory();
      SETREFVAL(href,e);
      linenoiseHistoryImport(lh);
    }
  return res;
#else
  hfprintf (hstdin, "%s", prompt_s);
  return readline(hstdin);
#endif
}
ACTION_END

/*******
 * Initialization
 *******/

/* there is no garbage collection during initialization,
	so KEEP() is a noop */
#undef KEEP
#define KEEP(x) x
#undef KONS
#define KONS Kons

PRIVATE ExprPtr K(n)
IN CONST COUNTER n;
{
  return Kstr((CHAR *)message(n));
}

PUBLIC void InitReadline()
{
  ExprPtr completiontype,stringpairtype,hintstype,historytype,type;
  
  promptsym = K(963);
  completionsym = K(964);
  stringpairtype = coupleType(stringType,stringType);
  completiontype = funType(stringpairtype,
			   listType(Kons(stringType,stringpairtype)));
  completiontype = KtypeAppl(optSym,Kons(completiontype,NIL));
  hintssym = K(965);
  hintstype = funType(stringType,
		      KtypeAppl(optSym,Kons(Kons(stringType,
						 Kons(intType,
						      Kons(boolType,NIL))),
					    NIL)));
  hintstype = KtypeAppl(optSym,Kons(hintstype,NIL));
  historysym = K(969);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  /* unimplemented */
#else
  historytype = KtypeAppl(optSym,Kons(KtypeAppl(refSym,
						Kons (coupleType(intType,listType(stringType)),
						      NIL)),
				      NIL));
#endif
  maskedsym = K(966);
  multilinesym = K(967);
  type = MapMerge(MapPair(promptsym,stringType,
			  completionsym,completiontype),
		  MapMerge(MapPair(hintssym,hintstype,
				   historysym,historytype),
			   MapPair(maskedsym,boolType,
				   multilinesym,boolType)));
  type = funType(type,stringType);
  ADDFUN(968,type,lineedit);
}
