/* Stack and continuation handling functions.
   This is mostly black magic.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "eval.h"
#include "expr.h"
#include "mem.h"
#include "timing.h"
#include "debug.h"
#include "evlproto.h"
#include "stkproto.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#undef GETSTATS

#ifdef GETSTATS
typedef struct stackStat {
  INT count;
  clock_t time;
  clock_t lastentered;
} stackStat;

stackStat *stackStats = NULL;
#endif

#define MAX_BLOCK_SIZE 1024

typedef struct block1 {
  char *sp;
  struct block1 *prev;
} block1;

typedef struct block {
  struct block *next; /* going up (from most recent to oldest) */
  struct block *prev; /* going down */
  block1 *b1;
  char *end;
#if BLOCK_GAP!=0
  char filler[BLOCK_GAP];
#endif
  struct bcontents {
    char data[MAX_BLOCK_SIZE];
  } b;
} block;

#ifdef HAS_PROTOTYPES
#include "sstack.h"
#endif

#ifdef sparc
#define SPARC_SYSTEM
#endif

#ifdef __sparc__
#define SPARC_SYSTEM
#endif

#undef DEBUG
#undef DEBUG2
#undef DEBUG3

#define ALLOCTMP_FREE

INT alloctmpBytes = 0;

PUBLIC INT currentAge = 0L;

PRIVATE block *alloctmp_block = (block *)0L;

#define BLOCK1_SIZE (MODALIGN(sizeof(block1)+ALIGNMENT-1))

PUBLIC VOIDPTR alloctmp(n)
long n; /* note that this code should remain essentially the same as the following one,
	   for alloctmp_anchored()! */
{
  REGISTER block1 *p1 = NULL;
  REGISTER block *p;

  while ((p = alloctmp_block)!=(block *)0)
    {
      p1 = p->b1;
unstack:
      if (p1!=(block1 *)0)
	{
	  if (p1->sp
#ifdef ASCENDING_STACK
	      <=
#else
	      >=
#endif
	      (char *)&n)
	    {
	      p->b1 = p1;
	      break;
	    }
	  p1 = p1->prev;
	  goto unstack;
	}
      alloctmp_block = p->next;
#ifdef ALLOCTMP_FREE
      alloctmpBytes -= p->end - (char *)p;
      MARKFREE(p);
#endif
    }
  if (n==0)
    return (VOIDPTR)NULL;
  n = MODALIGN(n+ALIGNMENT-1);
  {
    REGISTER char *b;
    REGISTER block1 *p2;

    if (p==(block *)0 ||
	(b = ((char *)p1)+BLOCK1_SIZE,
	 p2 = (block1 *)(b+n),
	 ((char *)p2)+BLOCK1_SIZE>p->end))
      {
	if (n+BLOCK1_SIZE>MAX_BLOCK_SIZE)
	  {
	    INT size = VARSIZE(block,b.data,n+BLOCK1_SIZE);

	    p = (block *)MALLOC(size);
	    alloctmpBytes += size;
	    p->end = &p->b.data[n+BLOCK1_SIZE];
	  }
	else
	  {
	    p = (block *)MALLOC(sizeof(block));
	    alloctmpBytes += sizeof(block);
	    p->end = &p->b.data[MAX_BLOCK_SIZE];
	  }
	p->next = alloctmp_block;
	alloctmp_block = p;
	b = p->b.data;
	p2 = (block1 *)(b+n);
	p1 = (block1 *)0;
      }
    p2->sp = (char *)&n;
    p2->prev = p1;
    p->b1 = p2;
    return (VOIDPTR)b;
  }
}

PUBLIC VOIDPTR alloctmp_anchored(n,anchor)
REGISTER long n; /* note that this code should remain essentially the same as the previous one,
	   for alloctmp()! */
REGISTER char *anchor;
{
  REGISTER block1 *p1 = NULL;
  REGISTER block *p;

  while ((p = alloctmp_block)!=(block *)0)
    {
      p1 = p->b1;
unstack:
      if (p1!=(block1 *)0)
	{
	  if (p1->sp
#ifdef ASCENDING_STACK
	      <=
#else
	      >=
#endif
	      anchor)
	    {
	      p->b1 = p1;
	      break;
	    }
	  p1 = p1->prev;
	  goto unstack;
	}
      alloctmp_block = p->next;
#ifdef ALLOCTMP_FREE
      alloctmpBytes -= p->end - (char *)p;
      MARKFREE(p);
#endif
    }
  if (n==0)
    return (VOIDPTR)NULL;
  n = MODALIGN(n+ALIGNMENT-1);
  {
    REGISTER char *b;
    REGISTER block1 *p2;

    if (p==(block *)0 ||
	(b = ((char *)p1)+BLOCK1_SIZE,
	 p2 = (block1 *)(b+n),
	 ((char *)p2)+BLOCK1_SIZE>p->end))
      {
	if (n+BLOCK1_SIZE>MAX_BLOCK_SIZE)
	  {
	    INT size;

	    size = VARSIZE(block,b.data,n+BLOCK1_SIZE);
	    p = (block *)MALLOC(size);
	    alloctmpBytes += size;
	    p->end = &p->b.data[n+BLOCK1_SIZE];
	  }
	else
	  {
	    p = (block *)MALLOC(sizeof(block));
	    alloctmpBytes += sizeof(block);
	    p->end = &p->b.data[MAX_BLOCK_SIZE];
	  }
	p->next = alloctmp_block;
	alloctmp_block = p;
	b = p->b.data;
	p2 = (block1 *)(b+n);
	p1 = (block1 *)0;
      }
    p2->sp = anchor;
    p2->prev = p1;
    p->b1 = p2;
    return (VOIDPTR)b;
  }
}

PUBLIC void freetmp(b)
IN CONST VOIDPTR b;
{
  REGISTER block1 *p1,*p0;
  REGISTER block *p;

  p0 = (block1 *)(((char *)b)-BLOCK1_SIZE);
  while ((p = alloctmp_block)!=(block *)0)
    {
      p1 = p->b1;
unstack:
      if (p1!=(block1 *)0)
	{
	  if (p1==p0)
	    {
	      p->b1 = p1;
	      return;
	    }
	  p1 = p1->prev;
	  goto unstack;
	}
      alloctmp_block = p->next;
      alloctmpBytes -= p->end - (char *)p;
      if (b==(VOIDPTR)p->b.data)
	{
	  MARKFREE(p);
	  return;
	}
      MARKFREE(p);
    }
  abort();
}

PRIVATE void alloctmpFree(blocks)
IN REGISTER block *blocks;
{
  REGISTER block *p;

  while (blocks!=(block *)0L)
    {
      p = blocks->next;
      alloctmpBytes -= blocks->end - (char *)blocks;
      MARKFREE(blocks);
      blocks = p;
    }
}

#ifdef SPARC_SYSTEM
#ifdef __GNUC__
#define FLUSH_REGS() __asm__("t 0x3")
#else
#define FLUSH_REGS() (void) sparcFlushRegs(32)
  /* this 32 number is pure guessing. */

PRIVATE int sparcFlushRegs(n)
int n;
{
  int a; /* used so that the compiler does not optimize the recursion away. */

  if (n==0)
    return 0;
  a = sparcFlushRegs(n-1);
  return a+1;
}
#endif
#else
#define FLUSH_REGS()
#endif

EXTERN ExprPtr basisEnv;
EXTERN gcStackView *theStackFrames,*theHeapFrames;
EXTERN ExprPtr nonsenseException;

PUBLIC void stackChunkFree(chunk)
     INOUT StackChunk *chunk;
{
  alloctmpFree((block *)chunk->alloctmp_blocks);
  chunk->alloctmp_blocks = (char *)0L;
  if (chunk->contents!=(char *)0L && chunk->contents!=(char *)1L)
    {
      MARKFREE(chunk->contents);
      chunk->contents = (char *)0L;
    }
}

#ifdef DEBUG2
#include "prtproto.h"
#endif
