/* Dumping data to a file, and conversely.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "eval.h"
#include "bc.h"
#include "expr.h"
#include "gram.h"
#include "compile.h"
#include "type.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "mem.h"
#include "errproto.h"
#include "grmproto.h"

#undef DEBUG1
#define DEBUG2

DECL_ACTION(apply_bytecode_closure)
EXTERN compiledAction __gimmlx_code [];
EXTERN COUNTER _n__gimmlx_code;

typedef CONST void (*markFunc) P((ExprPtr e));

#ifdef HAS_PROTOTYPES
#include "strans.h"
#endif

#ifdef XLATE
PUBLIC MapPtr bcXlateMap; /* byteCodeChunk -m> string counted */
PUBLIC COUNTER nbcXlate=0;
#endif

EXTERN EXTENDED zero;

PRIVATE MapPtr outShared;
PRIVATE printFunc outDoprint;
PRIVATE char *outPrintdata;
PRIVATE readFunc inDoread;
PRIVATE char *inReaddata;
PRIVATE Boolean sortMaps;
PRIVATE ExprPtr sortedMapRef;

PRIVATE ExprPtr waitingref;

EXTERN ExprPtr lexEnv,consEnv;
EXTERN ExprPtr strEnv,sigEnv;
EXTERN ExprPtr stdinval,stdoutval,stderrval;

EXTERN ExprPtr
	nomemException,
	stackException,noninvException,
	nthException,rangeException,
	asciiException,stringnthException,
	nonsenseException,
	subscriptException,ioException,
	reException,
  flexErrExc, gyaccErrExc, parseErrExc;

#if PROFILING
EXTERN ExprPtr profileExprs;
#endif
#if DEBUGGING!=0
EXTERN ExprPtr benignExcRegexpRef;
#endif
EXTERN ExprPtr leftMarginRef,rightMarginRef,maxLinesRef,floatFormatRef;
EXTERN ExprPtr maxInlineRef;

/* The specialTransfer[] trick consists in outputting special codes
   whenever output() encounters these values.
   It works only if these values are all pairwise distinct, so that
   we can identify them. */
PRIVATE ExprPtr *specialTransfer[] = {
  &lexEnv, &consEnv,
  &GLOBENV,
  &stdinval, &stdoutval, &stderrval,
  &BINDEXC, &MATCHEXC, &EMPTYEXC, &MAPGETEXC,
  &MAPEXC, &nomemException, &stackException, &INCOMPEXC,
  &nthException, &rangeException, &asciiException,
  &stringnthException, &noninvException, &subscriptException,
  &ARITHEXC, &nonsenseException,
  &ioException, &reException,
#if MODULE_SYSTEM & STD_MODULE_SYSTEM
  &strEnv, &sigEnv,
#endif
  &flexErrExc, &gyaccErrExc, &parseErrExc,
  &maxInlineRef, &benignExcRegexpRef, &leftMarginRef, &rightMarginRef, &maxLinesRef,
  &floatFormatRef,
  NULL
  };

PUBLIC void (*markDataHook) P((ExprPtr e)) = NULL;

/* IncRefTable[i] = (i & ~3) | ((i & 3)==3?3:((i & 3)+1)) */
PUBLIC CHAR IncRefTable[] = {
  0x01, 0x02, 0x03, 0x03,
  0x05, 0x06, 0x07, 0x07,
  0x09, 0x0a, 0x0b, 0x0b,
  0x0d, 0x0e, 0x0f, 0x0f,
  0x11, 0x12, 0x13, 0x13,
  0x15, 0x16, 0x17, 0x17,
  0x19, 0x1a, 0x1b, 0x1b,
  0x1d, 0x1e, 0x1f, 0x1f,
  0x21, 0x22, 0x23, 0x23,
  0x25, 0x26, 0x27, 0x27,
  0x29, 0x2a, 0x2b, 0x2b,
  0x2d, 0x2e, 0x2f, 0x2f,
  0x31, 0x32, 0x33, 0x33,
  0x35, 0x36, 0x37, 0x37,
  0x39, 0x3a, 0x3b, 0x3b,
  0x3d, 0x3e, 0x3f, 0x3f,
  0x41, 0x42, 0x43, 0x43,
  0x45, 0x46, 0x47, 0x47,
  0x49, 0x4a, 0x4b, 0x4b,
  0x4d, 0x4e, 0x4f, 0x4f,
  0x51, 0x52, 0x53, 0x53,
  0x55, 0x56, 0x57, 0x57,
  0x59, 0x5a, 0x5b, 0x5b,
  0x5d, 0x5e, 0x5f, 0x5f,
  0x61, 0x62, 0x63, 0x63,
  0x65, 0x66, 0x67, 0x67,
  0x69, 0x6a, 0x6b, 0x6b,
  0x6d, 0x6e, 0x6f, 0x6f,
  0x71, 0x72, 0x73, 0x73,
  0x75, 0x76, 0x77, 0x77,
  0x79, 0x7a, 0x7b, 0x7b,
  0x7d, 0x7e, 0x7f, 0x7f,
  0x81, 0x82, 0x83, 0x83,
  0x85, 0x86, 0x87, 0x87,
  0x89, 0x8a, 0x8b, 0x8b,
  0x8d, 0x8e, 0x8f, 0x8f,
  0x91, 0x92, 0x93, 0x93,
  0x95, 0x96, 0x97, 0x97,
  0x99, 0x9a, 0x9b, 0x9b,
  0x9d, 0x9e, 0x9f, 0x9f,
  0xa1, 0xa2, 0xa3, 0xa3,
  0xa5, 0xa6, 0xa7, 0xa7,
  0xa9, 0xaa, 0xab, 0xab,
  0xad, 0xae, 0xaf, 0xaf,
  0xb1, 0xb2, 0xb3, 0xb3,
  0xb5, 0xb6, 0xb7, 0xb7,
  0xb9, 0xba, 0xbb, 0xbb,
  0xbd, 0xbe, 0xbf, 0xbf,
  0xc1, 0xc2, 0xc3, 0xc3,
  0xc5, 0xc6, 0xc7, 0xc7,
  0xc9, 0xca, 0xcb, 0xcb,
  0xcd, 0xce, 0xcf, 0xcf,
  0xd1, 0xd2, 0xd3, 0xd3,
  0xd5, 0xd6, 0xd7, 0xd7,
  0xd9, 0xda, 0xdb, 0xdb,
  0xdd, 0xde, 0xdf, 0xdf,
  0xe1, 0xe2, 0xe3, 0xe3,
  0xe5, 0xe6, 0xe7, 0xe7,
  0xe9, 0xea, 0xeb, 0xeb,
  0xed, 0xee, 0xef, 0xef,
  0xf1, 0xf2, 0xf3, 0xf3,
  0xf5, 0xf6, 0xf7, 0xf7,
  0xf9, 0xfa, 0xfb, 0xfb,
  0xfd, 0xfe, 0xff, 0xff
};

/* DecRefTable[i] = (i & ~3) | ((i & 3)==0?0:(i & 3)==3?3:((i & 3)-1)) */
PUBLIC CHAR DecRefTable[] = {
  0x00, 0x00, 0x01, 0x03,
  0x04, 0x04, 0x05, 0x07,
  0x08, 0x08, 0x09, 0x0b,
  0x0c, 0x0c, 0x0d, 0x0f,
  0x10, 0x10, 0x11, 0x13,
  0x14, 0x14, 0x15, 0x17,
  0x18, 0x18, 0x19, 0x1b,
  0x1c, 0x1c, 0x1d, 0x1f,
  0x20, 0x20, 0x21, 0x23,
  0x24, 0x24, 0x25, 0x27,
  0x28, 0x28, 0x29, 0x2b,
  0x2c, 0x2c, 0x2d, 0x2f,
  0x30, 0x30, 0x31, 0x33,
  0x34, 0x34, 0x35, 0x37,
  0x38, 0x38, 0x39, 0x3b,
  0x3c, 0x3c, 0x3d, 0x3f,
  0x40, 0x40, 0x41, 0x43,
  0x44, 0x44, 0x45, 0x47,
  0x48, 0x48, 0x49, 0x4b,
  0x4c, 0x4c, 0x4d, 0x4f,
  0x50, 0x50, 0x51, 0x53,
  0x54, 0x54, 0x55, 0x57,
  0x58, 0x58, 0x59, 0x5b,
  0x5c, 0x5c, 0x5d, 0x5f,
  0x60, 0x60, 0x61, 0x63,
  0x64, 0x64, 0x65, 0x67,
  0x68, 0x68, 0x69, 0x6b,
  0x6c, 0x6c, 0x6d, 0x6f,
  0x70, 0x70, 0x71, 0x73,
  0x74, 0x74, 0x75, 0x77,
  0x78, 0x78, 0x79, 0x7b,
  0x7c, 0x7c, 0x7d, 0x7f,
  0x80, 0x80, 0x81, 0x83,
  0x84, 0x84, 0x85, 0x87,
  0x88, 0x88, 0x89, 0x8b,
  0x8c, 0x8c, 0x8d, 0x8f,
  0x90, 0x90, 0x91, 0x93,
  0x94, 0x94, 0x95, 0x97,
  0x98, 0x98, 0x99, 0x9b,
  0x9c, 0x9c, 0x9d, 0x9f,
  0xa0, 0xa0, 0xa1, 0xa3,
  0xa4, 0xa4, 0xa5, 0xa7,
  0xa8, 0xa8, 0xa9, 0xab,
  0xac, 0xac, 0xad, 0xaf,
  0xb0, 0xb0, 0xb1, 0xb3,
  0xb4, 0xb4, 0xb5, 0xb7,
  0xb8, 0xb8, 0xb9, 0xbb,
  0xbc, 0xbc, 0xbd, 0xbf,
  0xc0, 0xc0, 0xc1, 0xc3,
  0xc4, 0xc4, 0xc5, 0xc7,
  0xc8, 0xc8, 0xc9, 0xcb,
  0xcc, 0xcc, 0xcd, 0xcf,
  0xd0, 0xd0, 0xd1, 0xd3,
  0xd4, 0xd4, 0xd5, 0xd7,
  0xd8, 0xd8, 0xd9, 0xdb,
  0xdc, 0xdc, 0xdd, 0xdf,
  0xe0, 0xe0, 0xe1, 0xe3,
  0xe4, 0xe4, 0xe5, 0xe7,
  0xe8, 0xe8, 0xe9, 0xeb,
  0xec, 0xec, 0xed, 0xef,
  0xf0, 0xf0, 0xf1, 0xf3,
  0xf4, 0xf4, 0xf5, 0xf7,
  0xf8, 0xf8, 0xf9, 0xfb,
  0xfc, 0xfc, 0xfd, 0xff
};

PRIVATE void markVector(e,mark)
INOUT ExprPtr e;
IN markFunc mark;
{
  REGISTER COUNTER i,n;
  VecPtr vp;
  REGISTER ExprPtr *ep;

  vp = AVAL(e);
  n = vp->n;
  for (i=0, ep=vp->expr; i<n; i++)
    (*mark)(*ep++);
}

PRIVATE void markByteCode(bc,mark)
IN CONST byteCodeChunk *bc;
IN markFunc mark;
{
  REGISTER ExprPtr *ep,*epend;

  for (ep = bc->consts, epend = ep + bc->nConsts; ep<epend; )
    (*mark)(*ep++);
}

PRIVATE int cmpint(e1,e2)
IN CONST ExprPtr e1,e2;
{
  return (IVAL(e1)<IVAL(e2))?(-1):(IVAL(e1)>IVAL(e2))?1:0;
}

PRIVATE int cmpgnrl(_ep1,_ep2)
IN CONST ANSI_CONST VOIDPTR _ep1;
IN CONST ANSI_CONST VOIDPTR _ep2;
{
  REGISTER ExprPtr *ep1 = (ExprPtr *)_ep1;
  REGISTER ExprPtr *ep2 = (ExprPtr *)_ep2;
  ExprPtr e1,e2;
  int res;

  e1 = *ep1;
  e2 = *ep2;
  if (e1->kind<e2->kind)
    return -1;
  if (e1->kind>e2->kind)
    return 1;
  switch (e1->kind)
    {
    case E_STRING:
      res = CompareStrings(e1,e2);
      break;
    case E_INT:
      res = cmpint(e1,e2);
      break;
    case E_CONS:
      res = cmpint(CAR(e1),CAR(e2));
      if (res==0)
	res = cmpint(CDR(e1),CDR(e2));
      break;
    case E_TYPE_APPL: /* happens with numerical types, which map dimensions (variables
			 or applications q(), where q is a dimension name) to exponents. */
      res = CompareStrings(APPLFUN(e1),APPLFUN(e2));
      break;
    default:
      res = 0;
      kinderror(message(809),e1->kind);
      break;
    }
  return res;
}

MAPSWEEPFUNC(flushmap)
{
  ExprPtr **epp;

  epp = MAPDATA(ExprPtr **);
  *(*epp)++ = x;
  *(*epp)++ = y;
  return false;
}

PRIVATE ExprPtr sortMap(e)
IN CONST ExprPtr e; /* an E_MAP, which is always shared */
/* this is a hack that we only use to compute checksums (see CheckSum() in gramaux.c).
   Indeed, the order in which we visit elements of maps should be canonical (it should
   not depend on the system ordering), so we sort their elements first.
   We need compare only strings, integers and couples of two integers. */
{
  INT n,n2;
  ExprPtr *p,**epp;
  VecPtr vp;
  ExprPtr ee;
  ENTER;

  n = MapCard(e);
  n2 = n+n;
  CHECKHETEROFREE();
  VECALLOC(*(char **)&vp,VECSIZE(n2));
  p = vp->expr;
  epp = &p;
  (void) MapSweep(e,flushmap,epp);
  qsort(vp->expr,(long)n,(long)(2*sizeof(ExprPtr)),cmpgnrl);
  vp->n = n2;
  MAKETUPLE_NOGC(vp,ee);
  (void) KEEP(ee);
  setrefMapPut(e,ee,sortedMapRef);
  QUIT;
  return ee;
}

PRIVATE doExtDesc doExtMarkData = {
  X_MARKDATA,
  markData
};

PUBLIC void markData(e)
INOUT ExprPtr e;
{
again:
  INCREF(e);
  if (REFCOUNT(e)==1) /* has just been marked */
    switch (e->kind)
      {
      case E_FREE:
	break;
      case E_NIL:
      case E_INT:
      case E_REAL:
      case E_STRING:
      case E_VARINTRO:
      case E_VARREF:
      case E_PAT_ANY:
      case E_EMPTY:
      case E_TUPLE_0:
      case E_RECORD_0:
	break;
      case E_MAP:
	if (sortMaps)
	  {
	    e = sortMap(e);
	    goto again;
	  }
	/*fallthrough*/
      case E_COMPLEX:
      case E_CONS: case E_UNSHARED_CONS:
      case E_PAT_REVCONS: case E_PATCHOICE:
      case E_ARRAY_2:
      case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
      case E_PAT_REVTUPLE_2:
      case E_DATA:
      case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
      case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
      case E_PAT_RECORD_MAPADD:
      case E_RECORD_1: case E_UNSHARED_RECORD_1:
      case E_MAPGET:
      case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
      case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
      case E_LET:
      case E_TYPE_APPL:
      case E_APPL: case E_GOTO:
	/*case E_N_APPL:*/
      case E_PAT_AS:  case E_PAT_MAPADD:
      case E_SET:
      case E_IF: case E_PROG: case E_AND: case E_OR:
      case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
      case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
      case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
      case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
      case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
      case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
      case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
      case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP:
      case E_MAP_MAP: case E_MAP_UNDER_MAP: case E_ALL_MAP:
      case E_EXISTS_MAP: case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
      case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
      case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
      case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
      case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
      case E_PACK: case E_PAT_PACK:
      case E_EQUAL: case E_NEQUAL:
      case E_DEBUGZONE:
      case E_DEBUGNAME:
      case E_TYPEFX:
      case E_OVERWRITE: case E_UNDERWRITE:
      case E_PROFILE:
	markData(CAR(e));
	e = CDR(e);
	goto again;
      case E_MAPLET:
      case E_MEMO_FN:
      case E_HANDLE:
	markData(CDR(e));
	e = CAR(e);
	goto again;
      case E_NUMTYPE:
      case E_QUOTE:
      case E_REF:
      case E_GET:
      case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
      case E_RAISE:
      case E_EXCEPTION:
      case E_DELAY:
      case E_COUNTED:
      case E_DEBUGDECL:
      case E_PAT_CARD_EQ:
      case E_PAT_CARD_GEQ:
      case E_POLYFUN:
	/*case E_N_LAMBDA:*/
      case E_PAT_CONS:
      case E_DATACON:
      case E_NOT:
	e = REFVAL(e);
	goto again;
      case E_ARRAY_GNRL:
      case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
      case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
      case E_ENV:
	markVector(e,markData);
	e = ASHAPE(e);
	goto again;
      case E_COMPILED_CLOSURE:
      case E_INTERPRETED_CLOSURE:
	e = QENV(e);
	goto again;
      case E_PAT_CHECK:
	e = PATCHECKVAL(e);
	goto again;
      case E_STACK:
	/* markings are only used for data structures that are about to be
	   saved or exported; but we never save or export stacks (continuations). */
	break;
      case E_BYTECODE:
	markByteCode((byteCodeChunk *)e,markData);
	break;
      case E_EXTERNAL:
	(*EXTDESC(e)->sweep)(EXTDATA(e),&doExtMarkData);
	break;
      case TYPE_ASSUM_KIND:
	markData(A_IDS(e));
	markData(A_IDLOCS(e));
	markData(A_IDEXTS(e));
	markData(A_CONS(e));
	markData(A_TYFCNS(e));
	markData(A_TYEXTS(e));
#if MODULE_SYSTEM & STD_ML_MODULE
	markData(A_STRUCTS(e));
	markData(A_SIGS(e));
#endif
	markData(A_NONGENS(e));
	markData(A_UNITS(e));
	markData(A_TYVARS(e));
	break;
      case TYPE_STATE_KIND:
	markData(S_TYVARSUBS(e));
	markData(S_TYVARSUBCACHE(e));
	markData(S_CONDS(e));
	markData(S_EQCONDS(e));
	markData(S_EQFLAGS(e));
	markData(S_NONEQS(e));
	markData(S_TYCONS(e));
	markData(S_PACKERS(e));
	markData(S_ANNOTATIONS(e));
	markData(S_NONGENS(e));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	markData(S_STRENGTH(e));
	markData(S_MAXSTRENGTH(e));
#endif
	break;
      case COMPILE_ASSUM_KIND:
	markData(C_IDS(e));
#if MODULE_SYSTEM & STD_ML_MODULE
	markData(C_STRUCTS(e));
	markData(C_SIGS(e));
#endif
	markData(C_CONS(e));
	markData(C_TYSTATE(e));
	markData(C_TYTAGS(e));
	markData(C_IMPORTS(e));
	markData(C_FUNDEF(e));
	break;
      case BUILD_HEADER_KIND:
	e = ((struct build_header *)e)->args;
	goto again;
      default:
	kinderror(message(791),e->kind);
	break;
      }
  else if (markDataHook!=NULL)
    (*markDataHook)(e);
}

PRIVATE void unmarkBucketList1(bucklist)
INOUT BucketPtr bucklist;
{
  REGISTER ExprPtr ep;

  {
    struct Bucket *buck;
    REGISTER COUNTER i;

    for (buck=bucklist; buck!=(BucketPtr)0L; buck=buck->next)
      for (i=0, ep=buck->cell; i<BUCKET_SIZE; i++, ep++)
	{
	  SETREFCOUNT(ep,0);
	}
  }
}

PUBLIC void unmarkBucketLists()
{
  unmarkBucketList1(theHomoBucketList);
  unmarkBucketList1(theHeteroBucketList);
  unmarkBucketList1(oldHomoBucketList);
  unmarkBucketList1(oldHeteroBucketList);
}

PRIVATE void initWait()
{
  REGISTER ExprPtr queue;

  EALLOC_HOMO(queue);
  queue->kind = (CHAR)E_CONS;
  CAR(queue) = queue;
  CDR(queue) = NIL;
  (void) KEEP(queue);
  waitingref = queue;
}

PRIVATE void pushWait(e)
IN CONST ExprPtr e;
{
  REGISTER ExprPtr queue,last;

  BEGINQUICKKEEP();
  EALLOC_HOMO(queue);
  queue->kind = (CHAR)E_CONS;
  CAR(queue) = e;
  CDR(queue) = NIL;
  QUICKKEEP(queue);
  last = CAR(waitingref);
  SETCDR(last,queue);
  SETCAR(waitingref,queue);
  ENDQUICKKEEP();
}

PRIVATE ExprPtr popWait()
{
  REGISTER ExprPtr e,ll;

  ll = CDR(waitingref);
  if (ll==NIL)
    return NIL;
  e = CAR(ll);
  ll = CDR(ll);
  SETCDR(waitingref,ll);
  if (ll==NIL)
    {
      SETCAR(waitingref,waitingref);
    }
  return e;
}

PUBLIC void outInt(n)
IN INT n;
{
  BEGIN_FROM_INT(n);
  (*outDoprint)(FROM_INT(n),sizeof(INT),outPrintdata);
  END_FROM_INT(n);
}

PUBLIC INT inInt()
{
  INT n;

  {
    BEGIN_TO_INT(n);
    (*inDoread)(TO_INT(n),sizeof(INT),inReaddata);
    END_TO_INT(n);
  }
  return n;
}

#ifdef HAS_PROTOTYPES
PUBLIC void outShort(IN SHORT n)
#else
PUBLIC void outShort(n)
IN SHORT n;
#endif
{
  BEGIN_FROM_SHORT(n);
  (*outDoprint)(FROM_SHORT(n),sizeof(SHORT),outPrintdata);
  END_FROM_SHORT(n);
}

PUBLIC SHORT inShort()
{
  SHORT n;

  {
    BEGIN_TO_SHORT(n);
    (*inDoread)(TO_SHORT(n),sizeof(SHORT),inReaddata);
    END_TO_SHORT(n);
  }
  return n;
}

PUBLIC void outChars(s,len)
IN CONST CHAR *s;
IN COUNTER len;
{
  (*outDoprint)(s,len,outPrintdata);
}

PUBLIC void inChars(s,len)
OUT CHAR *s;
IN COUNTER len;
{
  (*inDoread)(s,len,inReaddata);
}

PUBLIC void outShorts (p, n)
IN CONST SHORT *p;
IN INT n;
{
  while (n--!=0)
    {
      outShort (*p++);
      if (errno) return;
    }
}

PUBLIC void outInts (p, n)
IN CONST INT *p;
IN INT n;
{
  while (n--!=0)
    {
      outInt (*p++);
      if (errno) return;
    }
}

PUBLIC void inShorts (p, n)
INOUT SHORT *p;
IN INT n;
{
  while (n--!=0)
    {
      *p++ = inShort ();
      if (errno) return;
    }
}

PUBLIC void inInts (p, n)
INOUT INT *p;
IN INT n;
{
  while (n--!=0)
    {
      *p++ = inInt ();
      if (errno) return;
    }
}

PUBLIC void outReal(x)
IN REAL *x;
{
#if FLOAT_FORMAT==1
  /* IEEE754, dump real (64 bits) as is */
  (*outDoprint)((CHAR *)x,sizeof(REAL),outPrintdata);
#else
  /* otherwise, convert to something looking like IEEE754 format and output */
  /* format = sign(1 bit) exponent(11 bits) mantissa without first bit (52 bits) */
  CHAR aux[8];
  int n;
  double mant;
  Boolean neg;

  if (*x==0.0)
    { aux[0] = aux[1] = aux[2] = aux[3] = aux[4] = aux[5] = aux[6] = aux[7] = 0; }
  else
    {
      mant = frexp((double)*x,&n);
      neg = (mant<0.0);
      if (neg)
	mant = -mant;
      n += 0x3ff;
      if (n<=0)
	{
	  n = 1;
	  aux[2] = aux[3] = aux[4] = aux[5] = aux[6] = aux[7] = 0;
	}
      else if (n>=0x800)
	{
	  n = 0x7ff; /* avoid exponent underflows/overflows */
	  aux[2] = aux[3] = aux[4] = aux[5] = aux[6] = aux[7] = 0xff;
	}
      else
	{
	  mant -= 0.5;
	  mant = ldexp(mant,5);
	  aux[1] = (CHAR)DEFAULT_IFLOOR(mant);
	  mant -= (double)aux[1];
	  {
	    COUNTER i;
	    
	    for (i=2; i<8; i++)
	      {
		mant = ldexp(mant,8);
		aux[i] = (CHAR)DEFAULT_IFLOOR(mant);
		mant -= (double)aux[i];
	      }
	  }
	}
      aux[0] = n>>4;
      if (neg)
	aux[0] |= 0x80;
      aux[1] |= (n & 0x0f) << 4;
    }
  (*outDoprint)(aux,8,outPrintdata);
#endif
}

PUBLIC void inReal(x)
OUT REAL *x;
{
#if FLOAT_FORMAT==1
  /* IEEE754 */
  (*inDoread)((CHAR *)x,sizeof(REAL),inReaddata);
#else
  CHAR aux[8];
  int n;
  COUNTER i;
  double mant;

  (*inDoread)(aux,8,inReaddata);
  if (aux[0]==0 && aux[1]==0)
    *x = zero;
  else
    {
      n = (((unsigned int)aux[0] & 0x7f) << 4) + (((unsigned int)aux[1]) >> 4) - 0x3ff;
      mant = zero;
      for (i=8; --i>=2; )
	{
	  mant += (double)aux[i];
	  mant = ldexp(mant,-8);
	}
      mant += (double)((aux[1] & 0x0f) + 0x10);
      mant = ldexp(mant,n-5);
      if (aux[0] & 0x80)
	mant = -mant;
      *x = mant;
    }
#endif
}

PRIVATE void outZone(zone)
IN CONST TZone *zone;
{
  outShort(zone->line1);
  if (errno) return;
  outShort(zone->line2);
  if (errno) return;
  outShort(zone->pos1);
  if (errno) return;
  outShort(zone->pos2);
}

PRIVATE void inZone(zone)
OUT TZone *zone;
{
  zone->line1 = inShort();
  if (errno) return;
  zone->line2 = inShort();
  if (errno) return;
  zone->pos1 = inShort();
  if (errno) return;
  zone->pos2 = inShort();
}

PRIVATE ExprPtr inTransfer1 P((INOUT ExprPtr sharedref));

PRIVATE ExprPtr sharedRef;
PRIVATE INT nShared;

PRIVATE void addShared(e)
IN CONST ExprPtr e;
{
  if (REFCOUNT(e)==2)
    {
      ExprPtr n;
      ENTER;

#ifdef DEBUG2
      {
	Boolean found;

	n = MapGet(e,REFVAL(sharedRef),&found);
	if (found)
	  {
	    hfprintf(hstderr,"addShared: %d already taken.\n", IVAL(n));
	    hfflush(hstderr);
	  }
      }
#endif
      n = KINT(++nShared);
      setrefMapPut(e,n,sharedRef);
      QUIT;
    }
}

PRIVATE MapPtr getShared(e)
INOUT ExprPtr e;
/* marks all reachable data from e, and returns map numbering all those
   reachable expressions that have at least two parents. */
{
  MapPtr shared;
  ENTER;

  nShared = 0;
  MAKEREF(NewMap(),sharedRef);
  (void) KEEP(sharedRef);
  { /* mark predefined standard variables so that they don't get marked
       any further by markData() below. */
    ExprPtr **epp,*ep;

    for (epp=specialTransfer; (ep = *epp++)!=NULL; )
      {
	if (ep==NULL)
	  continue;
	SETREFCOUNT(*ep,2);
	addShared(*ep);
      }
  }
  markDataHook = addShared;
  markData(e);
  markDataHook = NULL;
  shared = REFVAL(sharedRef);
  { /* record predefined standard variables, and unmark them, so that
     outTransfer1 knows that it only has to emit a crossref. */
    ExprPtr **epp,*ep;

    for (epp=specialTransfer; (ep = *epp++)!=NULL; )
      {
	if (ep==NULL)
	  continue;
	SETREFCOUNT(*ep,0);
      }
  }
  KEEPONLY(shared);
  return shared;
}

PRIVATE void outTransfer1 P((IN CONST ExprPtr e));

PRIVATE void outTransferVec(e)
IN CONST ExprPtr e;
{
  REGISTER COUNTER i,n;

  n = AVAL(e)->n;
  outInt(n);
  for (i=0; i<n; i++)
    {
      outTransfer1(AVAL(e)->expr[i]);
      if (errno) return;
    }
}

#ifndef XLATE
PRIVATE INT outInstrs(pc,pcend,forreal)
REGISTER OPCODE *pc,*pcend;
Boolean forreal;
{
  REGISTER OPCODE op;
#define BC_BUF_LEN 128
  CHAR bc_buf[BC_BUF_LEN];
  REGISTER CHAR *pbc;
#define BC_FLUSH() { if (forreal) { (*outDoprint)(bc_buf, pbc-bc_buf, outPrintdata); if (errno) return len; pbc = bc_buf; } }
#define BC_PUT(c) { if (forreal) {if (pbc>=bc_buf+BC_BUF_LEN) BC_FLUSH(); *pbc++ = (CHAR)(c); } len++; }
  REGISTER INT len;

  len = 0;
  for (pbc = bc_buf; pc<pcend;)
    {
      op = *pc++;
      /* op in [-128,125] => send op as one byte;
	 op in [126, 381] => send 126, op-126 (as two bytes)
	 otherwise        => send 127, op-hi, op-lo (as three bytes)
	 */
      if (op>=-128 && op<=125) /* then send only one byte */
	{ BC_PUT(op & 0xFF); }
      else if (op>=126 && op<=381)
	{
	  BC_PUT(126);
	  BC_PUT(op-126);
	}
      else
	{
	  BC_PUT(127);
	  BC_PUT(op >> 8);
	  BC_PUT(op & 0xFF);
	}
    }
  BC_FLUSH();
  return len;
#undef BC_PUT
#undef BC_FLUSH
#undef BC_BUF_LEN
}
#endif

PRIVATE void outByteCode(bc)
IN CONST byteCodeChunk *bc;
{
#ifdef XLATE
  {
    ExprPtr fname;
    INT n;

    fname = MapGetSure ((ExprPtr)bc, bcXlateMap);
    n = COUNTEDN(fname);
    outInt (n+1); /* 0 is reserved for interpreted closures. */
  }
#else
  outInt(bc->size);
  if (errno) return;
  outInt(bc->nargs);
  if (errno) return;
  outInt(bc->ntargs);
  if (errno) return;
  outInt(bc->stackSize);
  if (errno) return;
  outInt(bc->mapStackOff);
  if (errno) return;
  outInt(bc->excStackOff);
  if (errno) return;
  outInt(bc->nConsts);
  if (errno) return;
  {
    REGISTER COUNTER i,n;

    n = bc->nConsts;
    for (i=0; i<n; i++)
      {
	outTransfer1(bc->consts[i]);
	if (errno) return;
      }
  }
  {
    INT rawoff,instroff,debugoff;

    rawoff = ((char *)bc->rawData)-((char *)bc);
    outInt(rawoff);
    if (errno) return;
    instroff = ((char *)bc->instr)-((char *)bc);
    outInt(instroff);
    debugoff = ((char *)bc->debug)-((char *)bc);
    outInt(debugoff);
    if (errno) return;
    {
      INT len;

      len = outInstrs(bc->instr, (OPCODE *)(((char *)bc)+bc->size), false);
      outInt(len);
      if (errno) return;
    }
    /* (*outDoprint)(bc->rawData, bc->size-rawoff, outPrintdata); */
    (*outDoprint)(bc->rawData, instroff-rawoff, outPrintdata);
    if (errno) return;
  }
  (void) outInstrs(bc->instr, (OPCODE *)(((char *)bc)+bc->size), true);
#endif
}

PRIVATE void outCrossRef(e)
IN CONST ExprPtr e;
{
  INT i;
  CHAR c;

  i = IVAL(MapGetSure(e,outShared));
  if (i==(INT)(SHORT)i)
    {
      c = SHORT_CROSS_REF;
      (*outDoprint)(&c,1,outPrintdata);
      if (errno) return;
      outShort(i);
    }
  else
    {
      c = CROSS_REF;
      (*outDoprint)(&c,1,outPrintdata);
      if (errno) return;
      outInt(i);
    }
}

PRIVATE void outCrossDef(e)
IN CONST ExprPtr e;
{
  INT i;
  CHAR c;

  i = IVAL(MapGetSure(e,outShared));
  if (i==(INT)(SHORT)i)
    {
      c = SHORT_CROSS_DEF;
      (*outDoprint)(&c,1,outPrintdata);
      if (errno) return;
      outShort(i);
    }
  else
    {
      c = CROSS_DEF;
      (*outDoprint)(&c,1,outPrintdata);
      if (errno) return;
      outInt(i);
    }
}

PRIVATE doExtDesc doExtSave = {
  X_SAVE,
  outTransfer1
};

PRIVATE void outKindMark(e)
IN CONST ExprPtr e;
{
  CHAR c;

#ifdef XLATE
  if (e->kind==(CHAR)E_TUPLE_2 && CAR(e)->kind==(CHAR)E_BYTECODE) /* E_BYTECODEs are usually found
								     in the form of E_TUPLE_2(E_BYTECODE(bc), env)
								     in bc->consts arrays and in
								     EXTDATA(basis)->bindings for compiled modules.
								  */
    {
      c = (CHAR)E_COMPILED_CLOSURE;
      (*outDoprint)(&c,1,outPrintdata); /* output tag, except interpreted closures appear as compiled
					 when XLATE is #defined. */
    }
  else
    (*outDoprint)(&e->kind,1,outPrintdata); /* output tag */
#else
  (*outDoprint)(&e->kind,1,outPrintdata); /* output tag */
#endif
  if (errno) return;
  c = (e->mark & MARK_SHARED);
  (*outDoprint)(&c,1,outPrintdata); /* output mark, since we want to keep
				       the MARK_SHARED flag. */
  /* but we don't just output e->mark as is, since when we use outTransfer1()
     to compute checksums, the checksum would depend on irrelevant bits
     like MARK_GC or MARK_IN_CAR. So we hide them. */
}

PRIVATE void outTransfer1(e)
IN CONST ExprPtr e;
{
again:
  switch (REFCOUNT(e))
    {
    case 1: /* simple data, output it and clear refcount */
output:
      SETREFCOUNT(e,0);
      outKindMark(e);
      if (errno) return;
      switch (e->kind)
	{
	case E_NIL:
	case E_PAT_ANY:
	case E_EMPTY:
	case E_TUPLE_0:
	case E_RECORD_0:
	  break;
	case E_INT:
	  outInt(IVAL(e));
	  break;
	case E_REAL:
	  outReal(&XVAL(e));
	  break;
	case E_STRING:
	  outInt(SLEN(e));
	  if (errno) return;
	  (*outDoprint)(STEXT(e),SLEN(e),outPrintdata);
	  break;
	case E_VARREF:
	  outInt(VARLEVEL(e));
	  /*fallthrough*/
	case E_VARINTRO:
	  outInt(VARRANK(e));
	  break;
	case E_TUPLE_2:
#ifdef XLATE
	  /* If XLATE is #defined, we find all byteCodeChunks as CARs of E_TUPLE2s,
	     mostly in bc->consts arrays, but also in EXTDATA(basis)->bindings.
	     We output them as though they were E_COMPILED_CLOSUREs.
	     Whence the hack writing E_COMPILED_CLOSURE instead of
	     E_TUPLE_2 in outKindMark() in this precise case.
	  */
	  if (CAR(e)->kind==(CHAR)E_BYTECODE)
	    {
	      outByteCode((byteCodeChunk *)CAR(e));
	      if (errno) return;
	      e = CDR(e);
	      goto again;
	    }
#endif
	  goto tuple2;
	case E_MAP:
	  if (sortMaps)
	    {
	      e = MapGetSure(e,REFVAL(sortedMapRef));
	      goto again;
	    }
	  /*fallthrough*/
	tuple2:
	case E_COMPLEX:
	case E_CONS: case E_UNSHARED_CONS:
	case E_PAT_REVCONS: case E_PATCHOICE:
	case E_ARRAY_2: case E_MAPGET:
	case E_UNSHARED_TUPLE_2:
	case E_PAT_REVTUPLE_2:
	case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
	case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
	case E_PAT_RECORD_MAPADD:
	case E_RECORD_1: case E_UNSHARED_RECORD_1:
	case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
	case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
	case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
	case E_LET:
	case E_TYPE_APPL:
	case E_APPL: case E_GOTO:
	  /*case E_N_APPL:*/
	case E_PAT_AS: case E_PAT_MAPADD:
	case E_SET:
	case E_IF: case E_PROG: case E_AND: case E_OR:
	case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
	case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
	case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
	case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
	case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
	case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
	case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
	case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP:
	case E_MAP_MAP: case E_MAP_UNDER_MAP: case E_ALL_MAP:
	case E_EXISTS_MAP: case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
	case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
	case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
	case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
	case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
	case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
	case E_PACK: case E_PAT_PACK:
	case E_EQUAL: case E_NEQUAL:
	case E_DEBUGZONE:
	case E_DEBUGNAME:
	case E_TYPEFX:
	case E_OVERWRITE: case E_UNDERWRITE:
	case E_PROFILE:
	  if (e->mark & MARK_SHARED)
	    {
	      outTransfer1(CAR(e));
	      if (errno) return;
	      e = CDR(e);
	      goto again;
	    }
	  /* otherwise, we break possible cycles by putting the car and the cdr
	     in the waitingref list. */
	  pushWait(e);
	  break;
	case E_DATA:
	  if (e->mark & MARK_SHARED)
	    {
	      outShort(e->magic);
	      if (errno) return;
	      outTransfer1(CAR(e));
	      if (errno) return;
	      e = CDR(e);
	      goto again;
	    }
	  pushWait(e);
	  break;
	case E_MAPLET:
	case E_MEMO_FN:
	case E_HANDLE:
	  if (e->mark & MARK_SHARED)
	    {
	      outTransfer1(CDR(e));
	      if (errno) return;
	      e = CAR(e);
	      goto again;
	    }
	  /* otherwise, we break possible cycles by putting the car and the cdr
	     in the waitingref list. */
	  pushWait(e);
	  break;
	case E_POLYFUN:
	  /*case E_N_LAMBDA:*/
	  outShort(e->magic);
	  if (errno) return;
	  /*fallthrough*/
	case E_NUMTYPE:
	case E_QUOTE:
	case E_REF:
	case E_GET:
	case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
	case E_RAISE:
	case E_EXCEPTION:
	case E_DELAY:
	case E_COUNTED:
	case E_DEBUGDECL:
	case E_PAT_CARD_EQ:
	case E_PAT_CARD_GEQ:
	case E_PAT_CONS:
	case E_DATACON:
	case E_NOT:
	  outInt(COUNTEDN(e));
	  if (errno) return;
	  if (e->mark & MARK_SHARED)
	    {
	      e = REFVAL(e);
	      goto again;
	    }
	  /* otherwise, we break possible cycles by putting the car and the cdr
	     in the waitingref list. */
	  pushWait(e);
	  break;
	case E_ARRAY_GNRL:
	case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
	case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
	case E_ENV:
	  if (e->mark & MARK_SHARED)
	    outTransferVec(e);
	  else
	    {
	      outInt(AVAL(e)->n);
	      pushWait(e);
	    }
	  if (errno) return;
	  e = ASHAPE(e);
	  goto again;
	case E_COMPILED_CLOSURE:
	  {
	    COUNTER i;
	    compiledAction action = ACTION(e);

	    for (i=0; i<_n__gimmlx_code; i++)
	      if (action==__gimmlx_code[i])
		break;
	    if (i<_n__gimmlx_code)
	      {
		outInt ((INT)i+1); /* 0 is reserved for interpreted closures. */
		if (errno) return;
		e = QENV(e);
		goto again;
	      }
	    else
	      {
#ifdef DEBUG2
		hfprintf (hstderr, "outTransfer1: could not recognize action\n");
		hfflush (hstderr);
#endif
		errno = EADDRNOTAVAIL;
		break;
	      }
	  }
	case E_INTERPRETED_CLOSURE:
	  /* instead of outputting an address or a name for the action in
	     closures, which are useless or difficult to do, we output the
	     offset between the address of the action and the address of a
	     known function. This allows us to read back the same primitive
	     provided we did not change versions inbetween.
	     Note that this probably does not work on the Amiga with the SAS/C
	     compiler, where strange, optimized representations for function
	     pointers are used (see documentation; this can be made to work
	     with the proper compiler option, forcing all function references
	     to be absolute.)  On other machines, like RS/6000-based ones (or,
	     for that matter, PowerPC), where the processor does not have an
	     indirect branch-to-subroutine instruction, addresses of functions
	     are not the addresses of the actual code but of stubs located in
	     an array.  With this scheme, taking offsets as we do still works.
	     */
#ifdef DEBUG2
	  if (ACTION(e)!=apply_bytecode_closureAction)
	    {
	      hfprintf (hstderr, "outTransfer1: compiled function!=apply_bytecode_closureAction\n");
	      hfflush (hstderr);
	      abort ();
	    }
#endif
#ifdef XLATE
	  e = QENV(e);
#ifdef DEBUG2
	  if (e->kind!=(CHAR)E_BYTECODE)
	    {
	      hfprintf (hstderr, "outTransfer1: CAR(e) should be an E_BYTECODE\n");
	      hfflush (hstderr);
	      abort ();
	    }
#endif
	  outByteCode((byteCodeChunk *)CAR(e));
	  if (errno) return;
	  e = CDR(e);
	  goto again;
#else
	  /*outInt((INT)(((char *)ACTION(e))-((char *)Eval)));*/
	  outInt ((INT)0);
	  if (errno) return;
	  e = QENV(e);
	  goto again;
#endif
	case E_PAT_CHECK:
	  (*outDoprint)(PATCHECKBITS(e),(E_UNUSED+7) >> 3,outPrintdata);
	  if (errno) return;
	  e = PATCHECKVAL(e);
	  goto again;
	case E_STACK:
	  /* don't save anything; it is almost impossible to read and write
	     continuations to disk, anyway. But we would still have to unmark all
	     data that can be reached only by this structure, if we didn't avoid
	     the marking in the first place (see markData() in refcount.c) */
	  bmsg(message(631));
	  break;
	case E_BYTECODE:
	  outByteCode((byteCodeChunk *)e);
	  break;
	case E_EXTERNAL:
	  (*outDoprint)(EXTDESC(e)->id,4,outPrintdata);
	  if (errno) return;
	  (*EXTDESC(e)->sweep)(EXTDATA(e),&doExtSave);
	  break;
	case TYPE_ASSUM_KIND:
	  outInt(A_FNLEVEL(e));
	  if (errno) return;
	  outChars(&A_KIND(e),1);
	  if (errno) return;
	  outChars(&A_FLAGS(e),1);
	  if (errno) return;
	  outTransfer1(A_IDS(e));
	  if (errno) return;
	  outTransfer1(A_IDLOCS(e));
	  if (errno) return;
	  outTransfer1(A_IDEXTS(e));
	  if (errno) return;
	  outTransfer1(A_CONS(e));
	  if (errno) return;
	  outTransfer1(A_TYFCNS(e));
	  if (errno) return;
	  outTransfer1(A_TYEXTS(e));
	  if (errno) return;
	  outTransfer1(A_STRUCTS(e));
	  if (errno) return;
	  outTransfer1(A_SIGS(e));
	  if (errno) return;
	  outTransfer1(A_NONGENS(e));
	  if (errno) return;
	  outTransfer1(A_UNITS(e));
	  if (errno) return;
	  outTransfer1(A_TYVARS(e));
	  break;
	case TYPE_STATE_KIND:
	  outTransfer1(S_TYVARSUBS(e));
	  if (errno) return;
	  outTransfer1(S_TYVARSUBCACHE(e));
	  if (errno) return;
	  outTransfer1(S_CONDS(e));
	  if (errno) return;
	  outTransfer1(S_EQCONDS(e));
	  if (errno) return;
	  outTransfer1(S_EQFLAGS(e));
	  if (errno) return;
	  outInt(S_NEXGEN(e));
	  if (errno) return;
	  outTransfer1(S_NONEQS(e));
	  if (errno) return;
	  outTransfer1(S_TYCONS(e));
	  if (errno) return;
	  outTransfer1(S_PACKERS(e));
	  if (errno) return;
	  outTransfer1(S_ANNOTATIONS(e));
	  if (errno) return;
	  outTransfer1(S_NONGENS(e));
	  if (errno) return;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  outTransfer1(S_STRENGTH(e));
	  if (errno) return;
	  outTransfer1(S_MAXSTRENGTH(e));
#else
	  outTransfer1(NewMap());
	  if (errno) return;
	  outTransfer1(NewMap());
#endif
	  break;
	case COMPILE_ASSUM_KIND:
	  outChars(&C_FLAGS(e),1);
	  if (errno) return;
	  outTransfer1(C_IDS(e));
	  if (errno) return;
	  outTransfer1(C_STRUCTS(e));
	  if (errno) return;
	  outTransfer1(C_SIGS(e));
	  if (errno) return;
	  outInt(C_LEVEL(e));
	  if (errno) return;
	  outInt(C_ILEVEL(e));
	  if (errno) return;
	  outInt(C_NLOCALS(e));
	  if (errno) return;
	  outTransfer1(C_CONS(e));
	  if (errno) return;
	  outTransfer1(C_TYSTATE(e));
	  if (errno) return;
	  outTransfer1(C_TYTAGS(e));
	  if (errno) return;
	  outTransfer1(C_IMPORTS(e));
	  if (errno) return;
	  outTransfer1(C_FUNDEF(e));
	  if (errno) return;
	  outInt(C_DBGCNT(e));
	  break;
	case BUILD_HEADER_KIND:
	  outShort(TYPE(e));
	  if (errno) return;
	  outZone(ZONE(e));
	  if (errno) return;
	  e = ARGS(e);
	  goto again;
	default:
	  kinderror(message(416),e->kind);
	  break;
	}
      break;
    case 0: /* shared data, not the first time we see it (refcount has been cleared on first
	       time); output tag CROSS_REF followed by number of this node. */
      outCrossRef(e);
      break;
    default: /* shared data, first time we see it; output tag CROSS_DEF saying this is shared,
		followed by the number of this node; then clear refcount, and output it. */
      outCrossDef(e);
      if (errno) return;
      goto output;
    }
}

PRIVATE void outPatch(w)
INOUT ExprPtr w;
{
  switch (w->kind)
    {
    case E_DATA:
      outShort(w->magic);
      if (errno) return;
      outTransfer1(CAR(w));
      if (errno) return;
      outTransfer1(CDR(w));
      break;
    case E_MAPLET:
    case E_MEMO_FN:
    case E_HANDLE:
      outTransfer1(CDR(w));
      if (errno) return;
      outTransfer1(CAR(w));
      break;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
    case E_NUMTYPE:
    case E_QUOTE:
    case E_REF:
    case E_GET:
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_RAISE:
    case E_EXCEPTION:
    case E_DELAY:
    case E_COUNTED:
    case E_DEBUGDECL:
    case E_PAT_CARD_EQ:
    case E_PAT_CARD_GEQ:
    case E_PAT_CONS:
    case E_DATACON:
    case E_NOT:
      outTransfer1(REFVAL(w));
      break;
    case E_ARRAY_GNRL:
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
    case E_ENV:
      {
	COUNTER i,n;

	n = AVAL(w)->n;
	for (i=0; i<n; i++)
	  {
	    outTransfer1(AVAL(w)->expr[i]);
	    if (errno) return;
	  }
      }
      break;
    default: /* otherwise, we have to write a regular couple */
      outTransfer1(CAR(w));
      if (errno) return;
      outTransfer1(CDR(w));
      break;
    }
}

PUBLIC void outTransfer(e,doprint,printdata,sortmaps)
IN CONST ExprPtr e;
IN printFunc doprint;
INOUT char *printdata;
IN Boolean sortmaps;
{
  ExprPtr shared,savesortedmaps;
  printFunc savedoprint;
  char *savedata;
  Boolean savesortmaps;
  ENTER;

  shared = outShared;
  savedoprint = outDoprint;
  savedata = outPrintdata;
  savesortmaps = sortMaps;
  savesortedmaps = sortedMapRef;
  outDoprint = doprint;
  outPrintdata = printdata;
  sortMaps = sortmaps;
  if (sortmaps)
    {
      sortedMapRef = MakeSing(E_REF,NewMap());
      (void) KEEP(sortedMapRef);
    }
#if DEBUGGING
  restoreBreaks (); /* this restoreBreaks()/reinstallBreaks() pair to put back
		       all actual opcodes inside byteCodeChunks before outputting
		       data. */
#endif
  outShared = getShared(e);
  initWait();
  errno = 0;
  outTransfer1(e);
  while ((e = popWait()) != NIL && errno==0)
    outPatch(e);
  { /* unmark predefined standard variables */
    ExprPtr **epp,*ep;

    for (epp=specialTransfer; (ep = *epp++)!=NULL; )
      {
	if (ep==NULL)
	  continue;
	SETREFCOUNT(*ep,0);
      }
  }
#if DEBUGGING
  reinstallBreaks ();
#endif
  outShared = shared;
  outDoprint = savedoprint;
  outPrintdata = savedata;
  sortMaps = savesortmaps;
  sortedMapRef = savesortedmaps;
  QUIT;
}

PRIVATE ExprPtr inLabeled(i,sharedref)
IN int i;
INOUT ExprPtr sharedref;
{
  INT n;
  ExprPtr nn,e;
  ENTER;

  n = (i==CROSS_DEF)?inInt():inShort();
  if (errno)
    {
err:
      QUIT;
      return NIL;
    }
  nn = KINT(n);
  e = inTransfer1(sharedref);
  if (errno) goto err;
  setrefMapPut(nn,e,sharedref);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr inShared(i,sharedref)
IN int i;
IN CONST ExprPtr sharedref;
{
  INT n;
  ExprPtr nn,e;
  ENTER;

  n = (i==CROSS_REF)?inInt():inShort();
  if (errno)
    e = NIL;
  else
    {
      nn = KINT(n);
      e = MapGetSure(nn,REFVAL(sharedref));
    }
  QUIT;
  return e;
}

PRIVATE ExprPtr inIntT(mark)
IN int mark;
{
  INT n;
  ExprPtr e;

  n = inInt();
  if (errno) return NIL;
  if (mark & MARK_SHARED)
    return KINT(n);
  MAKEINT(n,e);
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr inRealT(mark)
IN int mark;
{
  REAL x;
  ExprPtr e;

  inReal(&x);
  if (errno) return NIL;
  if (mark & MARK_SHARED)
    e = Kreal(&x);
  else
    {
      MAKEREAL(x,e);
    }
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr inString(mark)
IN int mark;
{
  INT n;
  ExprPtr e;

  n = inInt();
  if (errno) return NIL;
  {
    CHAR *buf;

    VECALLOC(buf,n);
    (*inDoread)(buf,n,inReaddata);
    if (errno)
      {
	VECFREE(buf,n);
	return NIL;
      }
    if (mark & MARK_SHARED)
      {
	e = Kstring1(n,buf,false);
      }
    else
      {
	MAKESTRING(n,buf,e);
      }
  }
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr inVarRef()
{
  INT i,l;
  ExprPtr e;

  l = inInt();
  if (errno) return NIL;
  i = inInt();
  if (errno) return NIL;
  e = Kvarref(i,l); /* mark should always be shared */
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr inVarIntro()
{
  INT i;
  ExprPtr e;

  i = inInt();
  if (errno) return NIL;
  e = Kvarintro(i); /* mark should always be shared */
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr inShCouple(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  ExprPtr car,cdr;
  ENTER;

  car = inTransfer1(sharedref);
  if (errno) return NIL;
  cdr = inTransfer1(sharedref);
  if (errno) return NIL;
  switch (kind)
    {
    case E_TUPLE_2: car = Ktuple2(car,cdr); break;
    case E_CONS: car = Kons(car,cdr); break;
    case E_MAPLET: car = MapletKons(car,cdr); break;
    case E_MAP: car = MapKons(car,cdr); break;
    case E_COMPLEX: car = Kcomplex(car,cdr); break;
    case E_TYPE_APPL: car = KtypeAppl(car,cdr); break;
    case E_TYPEFX: car = KtypeFx(car,cdr); break;
    case E_RECORD_1: car = Krecord1(car,cdr); break;
      /*case E_BIAS: car = Kbias(car,cdr); break;*/
    default: errno = EILSEQ; return NIL;
    }
  /*car = Kpair(kind,car,cdr);*/
  KEEPONLY(car);
  return car;
}

PRIVATE ExprPtr inShData(sharedref)
INOUT ExprPtr sharedref;
{
  SHORT magic;
  ExprPtr car,cdr;
  ENTER;

  magic = inShort();
  if (errno)
    return NIL;
  car = inTransfer1(sharedref);
  if (errno) return NIL;
  cdr = inTransfer1(sharedref);
  if (errno) return NIL;
  car = Kdata(magic,car,cdr);
  KEEPONLY(car);
  return car;
}

PRIVATE ExprPtr inData()
{
  ExprPtr res;

  res = MakeData(0,NIL,NIL);
  (void) KEEP(res);
  pushWait(res);
  return res;
}

PRIVATE ExprPtr inShCoupleRev(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  ExprPtr car,cdr;
  ENTER;

  cdr = inTransfer1(sharedref);
  if (errno) return NIL;
  car = inTransfer1(sharedref);
  if (errno) return NIL;
  switch (kind)
    {
    case E_TUPLE_2: car = Ktuple2(car,cdr); break;
    case E_CONS: car = Kons(car,cdr); break;
    case E_MAPLET: car = MapletKons(car,cdr); break;
    case E_MAP: car = MapKons(car,cdr); break;
    case E_COMPLEX: car = Kcomplex(car,cdr); break;
    case E_TYPE_APPL: car = KtypeAppl(car,cdr); break;
    case E_TYPEFX: car = KtypeFx(car,cdr); break;
    case E_RECORD_1: car = Krecord1(car,cdr); break;
      /*case E_BIAS: car = Kbias(car,cdr); break;*/
    default: errno = EILSEQ; return NIL;
    }
  /*car = Kpair(kind,car,cdr);*/
  KEEPONLY(car);
  return car;
}

PRIVATE ExprPtr inShSingle(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  INT n;
  ExprPtr e;

  n = inInt();
  if (errno) return NIL;
  e = inTransfer1(sharedref);
  if (errno) return NIL;
  e = Kcounted(kind,e,n);
  KEEP(e);
  return e;
}

PRIVATE ExprPtr inSingle0(kind)
IN ExprKind kind;
{
  ExprPtr e;
  INT n;

  n = inInt();
  if (errno) return NIL;
  e = MakeCounted(kind,NIL,n);
  return KEEP(e);
}

PRIVATE ExprPtr inShSingleMagic(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  INT n;
  ExprPtr e;
  SHORT magic;

  magic = inShort();
  if (errno) return NIL;
  n = inInt();
  if (errno) return NIL;
  e = inTransfer1(sharedref);
  if (errno) return NIL;
  e = Kcounted(kind,e,n);
  e->magic = magic; /* so hashcodes for Kcounted()s should not depend on magic!
		       (we only use magic for E_POLYFUNs anyway, which are never
		       shared.)
		    */
  KEEP(e);
  return e;
}

PRIVATE ExprPtr inSingle0Magic(kind)
IN ExprKind kind;
{
  ExprPtr e;
  INT n;
  SHORT magic;

  magic = inShort();
  if (errno) return NIL;
  n = inInt();
  if (errno) return NIL;
  e = MakeCounted(kind,NIL,n);
  e->magic = magic;
  return KEEP(e);
}

PRIVATE ExprPtr inShVec(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  INT i,n;
  ExprPtr *tmp;
  ExprPtr *ep,shape,e;
  ENTER;

  n = inInt();
  if (errno) return NIL;
  tmp = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (i=0, ep=tmp; i<n; i++, ep++)
    {
      *ep = inTransfer1(sharedref);
      if (errno)
	{
err:
	  return NIL;
	}
    }
  shape = inTransfer1(sharedref);
  if (errno) goto err;
  e = Karray(kind,shape,n,tmp);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr inVec0(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  INT i,n;
  VecPtr vp;
  ExprPtr *ep,shape,e;
  ENTER;

  n = inInt();
  if (errno) return NIL;
  shape = inTransfer1(sharedref);
  if (errno) return NIL;
  CHECKHETEROFREE();
  VECALLOC(*(char **)&vp,VECSIZE(n));
  for (i=0, ep=vp->expr; i<n; i++)
    *ep++ = NIL;
  vp->n = n;
  MAKEKRECORD_NOGC(kind,shape,vp,e);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr inClosure(kind,sharedref)
IN ExprKind kind;
INOUT ExprPtr sharedref;
{
  ExprPtr env;
  INT offset;

  offset = inInt(); /* should be 0 in most cases (another integer possibly in gimmlx compiled code). */
  if (errno) return NIL;
  env = inTransfer1(sharedref);
  if (errno) return NIL;
  env = MakeCompiled(kind, apply_bytecode_closureAction, env);
  return KEEP(env);
}

PRIVATE ExprPtr inPatCheck(sharedref)
INOUT ExprPtr sharedref;
{
  ExprPtr e,pat;
  char *bits;

  VECALLOC(bits,(E_UNUSED+7) >> 3);
#define TMP_BITS bits

  (*inDoread)((CHAR *)TMP_BITS,(E_UNUSED+7) >> 3,inReaddata);
  if (errno)
    {
err:
      VECFREE(bits,((E_UNUSED+7) >> 3));
      return NIL;
    }
  pat = inTransfer1(sharedref);
  if (errno) goto err;
  CHECKHETEROFREE();
  EALLOC_HETERO_NOGC(e);
  e->kind = (CHAR)E_PAT_CHECK;
  PATCHECKVAL(e) = pat;
  PATCHECKBITS(e) = (CHAR *)bits;
  (void) KEEP(e);
  return e;
#undef TMP_BITS
}

PRIVATE ExprPtr fakeStack()
{
  ExprPtr stack;
  stackDesc *new;

  CHECKHETEROFREE();
  new = (stackDesc *)MALLOC(sizeof(stackDesc));
  /* cannot use VECALLOC, since we need new to be aligned. */
  EALLOC_HETERO_NOGC(stack);
  stack->kind = (CHAR)E_STACK;
  SJMPBUF(stack) = new;
  SCHUNK(stack) = (StackChunk *)0L;
  (void) KEEP(stack);
  return stack;
}

PRIVATE ExprPtr inTypeAssum(sharedref)
INOUT ExprPtr sharedref;
{
  ExprPtr e;

  e = NewTypeAssum();
  A_FNLEVEL(e) = inInt();
  if (errno) return NIL;
  inChars(&A_KIND(e),1);
  if (errno) return NIL;
  inChars(&A_FLAGS(e),1);
  if (errno) return NIL;
  A_IDS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_IDLOCS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_IDEXTS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_CONS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_TYFCNS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_TYEXTS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
#if MODULE_SYSTEM & STD_ML_MODULE
  A_STRUCTS(e) =
#else
    (void)
#endif
      inTransfer1(sharedref);
  if (errno) return NIL;
#if MODULE_SYSTEM & STD_ML_MODULE
  A_SIGS(e) =
#else
    (void)
#endif
      inTransfer1(sharedref);
  if (errno) return NIL;
  A_NONGENS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_UNITS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  A_TYVARS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  return e;
}

PRIVATE ExprPtr inTypeState(sharedref)
INOUT ExprPtr sharedref;
{
  ExprPtr e;

  e = NewTypeState();
  S_TYVARSUBS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_TYVARSUBCACHE(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_CONDS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_EQCONDS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_EQFLAGS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_NEXGEN(e) = inInt();
  if (errno) return NIL;
  S_NONEQS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_TYCONS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_PACKERS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_ANNOTATIONS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_NONGENS(e) = inTransfer1(sharedref);
  if (errno) return NIL;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  S_STRENGTH(e) = inTransfer1(sharedref);
  if (errno) return NIL;
  S_MAXSTRENGTH(e) = inTransfer1(sharedref);
#else
  (void) inTransfer1(sharedref);
  if (errno) return NIL;
  (void) inTransfer1(sharedref);
#endif
  if (errno) return NIL;
  return e;
}

PRIVATE ExprPtr inCompAssum(sharedref)
INOUT ExprPtr sharedref;
{
  ExprPtr e;
  ExprPtr ids/*,structs,sigs*/,cons,state,tags,imports,fundef;
  COUNTER level,ilevel,nlocals,dbgcnt;
  CHAR flags;

  inChars(&flags,1);
  if (errno) return NIL;
  ids = inTransfer1(sharedref);
  if (errno) return NIL;
  /*structs =*/ (void) inTransfer1(sharedref);
  if (errno) return NIL;
  /*sigs =*/ (void) inTransfer1(sharedref);
  if (errno) return NIL;
  level = inInt();
  if (errno) return NIL;
  ilevel = inInt();
  if (errno) return NIL;
  nlocals = inInt();
  if (errno) return NIL;
  cons = inTransfer1(sharedref);
  if (errno) return NIL;
  state = inTransfer1(sharedref);
  if (errno) return NIL;
  tags = inTransfer1(sharedref);
  if (errno) return NIL;
  imports = inTransfer1(sharedref);
  if (errno) return NIL;
  fundef = inTransfer1(sharedref);
  if (errno) return NIL;
  dbgcnt = inInt();
  if (errno) return NIL;
  e = MAKE_TABLE(ids,structs,sigs,level,ilevel,nlocals,cons,state,tags,imports,fundef,dbgcnt);
  C_FLAGS(e) = flags;
  return e;
}

PRIVATE ExprPtr inBuildHeader(sharedref)
INOUT ExprPtr sharedref;
{
  TZone tz;
  SHORT type;
  ExprPtr args,e;

  type = inShort();
  if (errno) return NIL;
  inZone(&tz);
  if (errno) return NIL;
  args = inTransfer1(sharedref);
  if (errno) return NIL;
  e = ZBuild(&tz,type,args);
  (void) KEEP(e);
#if PROFILING
  if (type==EA_PROFILE)
    profileExprs = MakePair(E_CONS,e,profileExprs);
#endif
  return e;
}

PRIVATE void unknownSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  switch (doext->xtag)
    {
    case X_SAVE:
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr unknownLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  return NIL;
}

PRIVATE extDesc unknownDesc = {
  (CHAR *)"?",
  { '?','?','?','?' }, NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  unknownSweep,
  unknownLoad
  };

PRIVATE ExprPtr inExternal(mark,sharedref)
IN int mark;
INOUT ExprPtr sharedref;
{
  CHAR id[5];
  extDesc *desc;

  (*inDoread)(id,4,inReaddata);
  if (errno) return NIL;
  id[4] = 0;
  desc = getExtDesc(id);
  if (desc==NULL)
    {
      bmsg(message(799),id);
      desc = &unknownDesc; /* should then refuse to load */
    }
  return (*desc->load)(desc,(mark & MARK_SHARED)!=0,sharedref,inTransfer1);
}

PRIVATE ExprPtr inMap1(sharedref)
INOUT ExprPtr sharedref;
/* version of inTransfer1(), specialized to maps. */
{
  CHAR c,mark;
  ExprPtr x,y;
  ENTER;

  (*inDoread)(&c,1,inReaddata);
  if (errno) return NIL;
  switch (c)
    {
    case CROSS_DEF: /* defining an index to reference a shared data */
    case SHORT_CROSS_DEF:
      x = inLabeled(c,sharedref);
      break;
    case CROSS_REF: /* getting the shared data */
    case SHORT_CROSS_REF:
      x = inShared(c,sharedref);
      break;
    case E_MAP:
      (*inDoread)(&mark,1,inReaddata);
      if (errno) return NIL;
      x = inMap1(sharedref);
      if (errno) return NIL;
      y = inMap1(sharedref);
      if (errno) return NIL;
      x = MapMerge(x,y);
      break;
    case E_MAPLET:
      (*inDoread)(&mark,1,inReaddata);
      if (errno) return NIL;
      y = inTransfer1(sharedref);
      if (errno) return NIL;
      x = inTransfer1(sharedref);
      if (errno) return NIL;
      x = MapSingleton(x,y);
      break;
    default:
      x = NIL;
      kinderror(message(417),c);
      break;
    }
  if (errno) return NIL;
  KEEPONLY(x);
  return x;
}

PRIVATE ExprPtr inMap(sharedref)
INOUT ExprPtr sharedref;
{
  ExprPtr car,cdr;
  ENTER;

  car = inMap1(sharedref);
  if (errno) return NIL;
  cdr = inMap1(sharedref);
  if (errno) return NIL;
  car = MapMerge(car,cdr);
  KEEPONLY(car);
  return car;
}

PRIVATE ExprPtr inByteCode(sharedref)
INOUT ExprPtr sharedref;
{
  INT size,rawoff,instroff,debugoff,nconsts,len;
  byteCodeChunk *bc;

    {
      ENTER;

      size = inInt();
      if (errno) return NIL;
      bc = bcAlloc(size);
      bc->nConsts = 0;
      (void) KEEP((ExprPtr)bc);
      bc->size = size;
      if (errno)
	{
	err:
	  QUIT;
	  return NIL;
	}
      bc->nargs = inInt();
      if (errno) goto err;
      bc->ntargs = inInt();
      if (errno) goto err;
      bc->stackSize = inInt();
      if (errno) goto err;
      bc->mapStackOff = inInt();
      if (errno) goto err;
      bc->excStackOff = inInt();
      if (errno) goto err;
      nconsts = inInt();
      if (errno) goto err;
      {
	REGISTER COUNTER i;
	
	for (i=0; i<nconsts; i++)
	  {
	    bc->consts[i] = inTransfer1(sharedref);
	    if (errno) goto err;
	  }
      }
      bc->nConsts = nconsts;
      rawoff = inInt();
      if (errno) goto err;
      bc->rawData = (CHAR *)(((char *)bc)+rawoff);
      instroff = inInt();
      if (errno) goto err;
      debugoff = inInt();
      if (errno) goto err;
      len = inInt();
      if (errno) goto err;
      bc->instr = (OPCODE *)(((char *)bc)+instroff);
      bc->debug = (OPCODE *)(((char *)bc)+debugoff);
      /* (*inDoread)(bc->rawData, bc->size-rawoff, inReaddata); */
      (*inDoread)(bc->rawData, instroff-rawoff, inReaddata);
      if (errno) goto err;
      {
	REGISTER OPCODE *pc,*pcend,op;
#define BC_BUF_LEN 128
	CHAR bc_buf[BC_BUF_LEN];
	REGISTER CHAR *pbc,*pbcend,c;
	REGISTER int escape;
	
	pc = bc->instr;
	pcend = (OPCODE *)(((char *)bc)+bc->size);
	escape = 0;
	while (len>0)
	  {
	    INT llen;
	    
	    llen = (len<BC_BUF_LEN)?len:BC_BUF_LEN;
	    (*inDoread)(bc_buf, llen, inReaddata);
	    if (errno) goto err;
	    len -= llen;
	    /*op = 0;*/
	    for (pbc = bc_buf, pbcend = bc_buf+llen; pbc<pbcend; )
	      {
		if (pc>=pcend)
		  {
		    errno = ENOEXEC;
		    goto err;
		  }
		c = *pbc++;
		/* for compression format, see outInstrs() */
		switch (escape)
		  {
		  case 0: /* normal case */
		    op = (OPCODE)(char)c;
		    switch (op)
		      {
		      case 126: escape = 1; break;
		      case 127: escape = 2; break;
		      default: *pc++ = op; break;
		      }
		    break;
		  case 1: /* second character of two-character sequence */
		    op = ((OPCODE)(unsigned)c)+126;
		    *pc++ = op;
		    escape = 0;
		    break;
		  case 2: /* second character of three-character sequence */
		    op = (OPCODE)(((unsigned)c) << 8);
		    escape = 3;
		    break;
		  case 3: /* third character of three-character sequence */
		    op |= (OPCODE)(unsigned)c;
		    *pc++ = op;
		    escape = 0;
		    break;
		  }
	      }
	  }
#undef BC_BUF_LEN
      }
#ifdef XLATE
      {
	ExprPtr name;

	name = MakeCounted (E_COUNTED, NIL, nbcXlate++);
	(void) KEEP (name);
	setMapPut ((ExprPtr) bc, name, &bcXlateMap);
      }
#endif
      KEEPONLY((ExprPtr)bc);
    }
  return (ExprPtr)bc;
}

PRIVATE ExprPtr inTransfer1(sharedref)
INOUT ExprPtr sharedref;
{
  CHAR c,mark;
  ExprPtr res;
  ENTER;

    {
      (*inDoread)(&c,1,inReaddata);
      if (errno) return NIL;
      switch (c)
	{
	case CROSS_DEF: /* defining an index to reference a shared data */
	case SHORT_CROSS_DEF:
	  res = inLabeled(c,sharedref);
	  break;
	case CROSS_REF: /* getting the shared data */
	case SHORT_CROSS_REF:
	  res = inShared(c,sharedref);
	  break;
	default:
	  (*inDoread)(&mark,1,inReaddata);
	  if (errno) return NIL;
	  switch (c)
	    {
	    case E_NIL: res = NIL; break;
	    case E_PAT_ANY: res = patany; break;
	    case E_EMPTY: res = NewMap(); break;
	    case E_TUPLE_0: res = tuple0; break;
	    case E_RECORD_0: res = record0; break;
	    case E_INT: res = inIntT(mark); break;
	    case E_REAL: res = inRealT(mark); break;
	    case E_STRING: res = inString(mark); break;
	    case E_VARREF: res = inVarRef(); break;
	    case E_VARINTRO: res = inVarIntro(); break;
	    case E_COMPLEX:
	    case E_CONS: case E_UNSHARED_CONS:
	    case E_PAT_REVCONS: case E_PATCHOICE:
	    case E_ARRAY_2: case E_MAPGET:
	    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
	    case E_PAT_REVTUPLE_2:
	    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
	    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
	    case E_PAT_RECORD_MAPADD:
	    case E_RECORD_1: case E_UNSHARED_RECORD_1:
	    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
	    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
	    case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
	    case E_LET:
	    case E_TYPE_APPL:
	    case E_APPL: case E_GOTO:
	      /*case E_N_APPL:*/
	    case E_PAT_AS: case E_PAT_MAPADD:
	    case E_SET:
	    case E_IF: case E_PROG: case E_AND: case E_OR:
	    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
	    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
	    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
	    case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
	    case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
	    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
	    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
	    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP:
	    case E_MAP_MAP: case E_MAP_UNDER_MAP: case E_ALL_MAP:
	    case E_EXISTS_MAP: case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
	    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
	    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
	    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
	    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
	    case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
	    case E_PACK: case E_PAT_PACK:
	    case E_EQUAL: case E_NEQUAL:
	    case E_DEBUGZONE:
	    case E_DEBUGNAME:
	    case E_TYPEFX:
	    case E_OVERWRITE: case E_UNDERWRITE:
	    case E_PROFILE:
	      if (mark & MARK_SHARED)
		res = inShCouple((ExprKind)c,sharedref);
	      else
		{
		  res = MakePair((ExprKind)c,NIL,NIL);
		  (void) KEEP(res);
		  pushWait(res);
		}
	      break;
	    case E_DATA:
	      if (mark & MARK_SHARED)
		res = inShData(sharedref);
	      else
		res = inData();
	      break;
	    case E_MAP: /* maps are different, since they may have to be reordered. */
	      res = inMap(sharedref);
	      break;
	    case E_MAPLET:
	    case E_MEMO_FN:
	    case E_HANDLE:
	      if (mark & MARK_SHARED)
		res = inShCoupleRev((ExprKind)c,sharedref);
	      else
		{
		  res = MakePair((ExprKind)c,NIL,NIL);
		  (void) KEEP(res);
		  pushWait(res);
		}
	      break;
	    case E_POLYFUN:
	      /*case E_N_LAMBDA:*/
	      if (mark & MARK_SHARED)
		res = inShSingleMagic((ExprKind)c,sharedref);
	      else
		{
		  res = inSingle0Magic((ExprKind)c);
		  pushWait(res);
		}
	      break;
	    case E_NUMTYPE:
	    case E_QUOTE:
	    case E_REF:
	    case E_GET:
	    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
	    case E_RAISE:
	    case E_EXCEPTION:
	    case E_DELAY:
	    case E_COUNTED:
	    case E_DEBUGDECL:
	    case E_PAT_CARD_EQ:
	    case E_PAT_CARD_GEQ:
	    case E_PAT_CONS:
	    case E_DATACON:
	    case E_NOT:
	      if (mark & MARK_SHARED)
		res = inShSingle((ExprKind)c,sharedref);
	      else
		{
		  res = inSingle0((ExprKind)c);
		  pushWait(res);
		}
	      break;
	    case E_ARRAY_GNRL:
	    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
	    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
	    case E_ENV:
	      if (mark & MARK_SHARED)
		res = inShVec((ExprKind)c,sharedref);
	      else
		{
		  res = inVec0((ExprKind)c,sharedref);
		  pushWait(res);
		}
	      break;
	    case E_COMPILED_CLOSURE:
	      {
		INT i;
		ExprPtr e;

		i = inInt()-1;
		if (errno) return NIL;
		e = inTransfer1(sharedref);
		if (errno) return NIL;
		res = MakeCompiled (E_COMPILED_CLOSURE, __gimmlx_code [i], e);
		KEEPONLY(res);
		break;
	      }
	    case E_INTERPRETED_CLOSURE:
	      res = inClosure((ExprKind)c,sharedref);
	      break;
	    case E_PAT_CHECK:
	      res = inPatCheck(sharedref);
	      break;
	    case E_STACK:
	      bmsg(message(801));
	      res = fakeStack(); /* so that, if Throw() attempts to throw it, the exception
				    NonSense will be raised. */
	      break;
	    case E_BYTECODE:
	      res = inByteCode(sharedref);
	      break;
	    case E_EXTERNAL:
	      res = inExternal(mark,sharedref);
	      break;
	    case TYPE_ASSUM_KIND:
	      res = inTypeAssum(sharedref);
	      break;
	    case TYPE_STATE_KIND:
	      res = inTypeState(sharedref);
	      break;
	    case COMPILE_ASSUM_KIND:
	      res = inCompAssum(sharedref);
	      break;
	    case BUILD_HEADER_KIND:
	      res = inBuildHeader(sharedref);
	      break;
	    default:
	      res = 0;
	      kinderror(message(432),(int)c);
	      break;
	    }
	  break;
	}
    }
  KEEPONLY(res);
  return res;
}

PRIVATE void inPatch(sharedref,w)
INOUT ExprPtr sharedref;
INOUT ExprPtr w;
{
  ExprPtr e;
  ENTER;

  switch (w->kind)
    {
    case E_DATA:
      {
	w->magic = inShort();
	if (errno) return;
	e = inTransfer1(sharedref);
	if (errno) return;
	SETCAR(w,e);
	e = inTransfer1(sharedref);
	if (errno) return;
	SETCDR(w,e);
	break;
      }
    case E_MAPLET:
    case E_MEMO_FN:
    case E_HANDLE:
      e = inTransfer1(sharedref);
      if (errno) return;
      SETCDR(w,e);
      e = inTransfer1(sharedref);
      if (errno) return;
      SETCAR(w,e);
      break;
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
    case E_NUMTYPE:
    case E_QUOTE:
    case E_REF:
    case E_GET:
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_RAISE:
    case E_EXCEPTION:
    case E_DELAY:
    case E_COUNTED:
    case E_DEBUGDECL:
    case E_PAT_CARD_EQ:
    case E_PAT_CARD_GEQ:
    case E_PAT_CONS:
    case E_DATACON:
    case E_NOT:
      e = inTransfer1(sharedref);
      if (errno) return;
      SETREFVAL(w,e);
      break;
    case E_ARRAY_GNRL:
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
    case E_ENV:
      {
	COUNTER i,n;

	n = AVAL(w)->n;
	for (i=0; i<n; i++)
	  {
	    e = inTransfer1(sharedref);
	    if (errno) return;
	    SETQ(AVAL(w)->expr[i],e);
	    QUIT;
	    HITBARRIER(w);
	  }
      }
      break;
    default: /* otherwise, we have to fill in a regular couple */
      e = inTransfer1(sharedref);
      if (errno) return;
      SETCAR(w,e);
      e = inTransfer1(sharedref);
      if (errno) return;
      SETCDR(w,e);
      break;
    }
  QUIT;
}

PUBLIC ExprPtr inTransfer(doread,readdata)
IN readFunc doread;
INOUT char *readdata;
{
  ExprPtr e,w,sharedref;
  readFunc savedoread;
  char *savedata;
  ENTER;

  savedoread = inDoread;
  savedata = inReaddata;
  inDoread = doread;
  inReaddata = readdata;
  MAKEREF(NewMap(),sharedref);
  (void) KEEP(sharedref);
  initWait();
  errno = 0;
  { /* record predefined standard variables */
    ExprPtr **epp,*ep;
    INT n;

    n = 0;
    for (epp=specialTransfer; (ep = *epp++)!=NULL; )
      if (ep!=NULL)
	setrefMapPut(KINT(++n),*ep,sharedref);
  }
  e = inTransfer1(sharedref);
  while ((w = popWait()) != NIL && errno==0)
    {
      inPatch(sharedref,w);
    }
  inDoread = savedoread;
  inReaddata = savedata;
  KEEPONLY(e);
  return e;
}
