/* Support functions for type checking.
   Copyright (C) 2021 Jean Goubault-Larrecq
   and Université Paris-Saclay, CNRS, ENS Paris-Saclay,
   Laboratoire Méthodes Formelles, 91190, Gif-sur-Yvette, France.

   This file is part of GimML.

   GimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "type.h"
#include "expr.h"
#include "typproto.h"
#include "memproto.h"
#include "prtproto.h"
#include "errproto.h"
#include "ariproto.h"
#include "zproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#endif
#include "rtypeaux.h"
#ifdef HAS_PROTOTYPES
#include "stypeaux.h"
#endif

#undef DEBUG
#undef DEBUG1
#undef DEBUG2
#undef DEBUG_EXPLICIT

#define TYPE_CACHE

#define FAILUNIF return true

#if CORE_TRACE
EXTERN hFILE *coreTraceFile;
#endif

EXTERN COUNTER theRmargin;
EXTERN SHORT generation_count;

EXTERN int yynerrs;

PUBLIC ExprPtr moduleName;

PRIVATE void warnGen P((IN CONST TZone *zone,IN CONST ExprPtr tyvar));

#define TP(type) TypePrint(type,do_printfile,(char *)hstderr,NIL)
#define EP(expr) valPrint(expr,alpha,do_printfile,(char *)hstderr,0)

#ifdef DEBUG1
PRIVATE void dump(assum)
IN CONST ExprPtr assum;
{
  hfprintf(hstderr,"\nvariables that are non generic:\n");
  EP(A_NONGENS(assum));
  hfprintf(hstderr,"\n");
  /*CRASH();*/
}
#endif

#ifdef DEBUG_EXPLICIT
PRIVATE void dump(assum)
IN CONST ExprPtr assum;
{
  hfprintf(hstderr,"\nvariables that are non generic:\n");
  EP(A_NONGENS(assum));
  hfprintf(hstderr,"\n");
  /*CRASH();*/
}
#endif

PRIVATE ExprPtr Cons(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr res;

  MAKECONS(a,b,res);
  return KEEP(res);
}

PRIVATE struct type_assum *allocTypeAssums=(struct type_assum *)0L;
PRIVATE struct type_assum *oldTypeAssums = (struct type_assum *)0L;

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr MakeTypeAssum(IN CONST COUNTER fnlevel,IN CONST ExprPtr ids,
			     IN CONST ExprPtr idlocs,IN CONST ExprPtr idexts,
			     IN CONST ExprPtr cons,
			     IN CONST ExprPtr tyfcns,IN CONST ExprPtr tyexts,
#if MODULE_SYSTEM & STD_ML_MODULE
			     IN CONST ExprPtr structs,IN CONST ExprPtr sigs,
#endif
			     IN CONST ExprPtr nongens,IN CONST ExprPtr units,
			     IN CONST ExprPtr tyvars)
#else
PUBLIC ExprPtr MakeTypeAssum(fnlevel,ids,idlocs,idexts,cons,tyfcns,tyexts,
#if MODULE_SYSTEM & STD_ML_MODULE
			     structs,sigs,
#endif
			     nongens,units,tyvars)
IN CONST COUNTER fnlevel;
IN CONST ExprPtr ids,idlocs,idexts,cons,tyfcns,tyexts;
#if MODULE_SYSTEM & STD_ML_MODULE
IN CONST ExprPtr structs,sigs;
#endif
IN CONST ExprPtr nongens,units,tyvars;
#endif
{
  struct type_assum *new;

  new = (struct type_assum *)MALLOC(sizeof(struct type_assum));
  new->kind = TYPE_ASSUM_KIND;
  new->mark = 0;
  /*new->gcmark = 0;*/
  new->next = allocTypeAssums;
  allocTypeAssums = new;
  A_FNLEVEL(new) = fnlevel;
  A_IDS(new) = ids;
  A_IDLOCS(new) = idlocs;
  A_IDEXTS(new) = idexts;
  A_CONS(new) = cons;
  A_TYFCNS(new) = tyfcns;
  A_TYEXTS(new) = tyexts;
#if MODULE_SYSTEM & STD_ML_MODULE
  A_STRUCTS(new) = structs;
  A_SIGS(new) = sigs;
#endif
  A_NONGENS(new) = nongens;
  A_UNITS(new) = units;
  A_TYVARS(new) = tyvars;
  A_KIND(new) = TA_REGULAR;
  A_FLAGS(new) = 0;
  return KEEP((ExprPtr)new);
}

PUBLIC ExprPtr NewTypeAssum()
{
  return MAKE_ASSUM((COUNTER)0,NewMap(),NewMap(),NewMap(),NewMap(),
		    NewMap(),NewMap(),NewMap(),NewMap(),NewMap(),NewMap(),NewMap());
}

PUBLIC ExprPtr DupTypeAssum(assum)
IN CONST ExprPtr assum;
{
  ExprPtr new_assum;

  new_assum = MAKE_ASSUM(A_FNLEVEL(assum),A_IDS(assum),A_IDLOCS(assum),
			 A_IDEXTS(assum),
			 A_CONS(assum),A_TYFCNS(assum),
			 A_TYEXTS(assum),
			 A_STRUCTS(assum),
			 A_SIGS(assum),A_NONGENS(assum),A_UNITS(assum),A_TYVARS(assum));
  A_KIND(new_assum) = A_KIND(assum);
  A_FLAGS(new_assum) = A_FLAGS(assum);
  return new_assum;
}

PUBLIC void AddTypeAssum(assum,new)
IN CONST ExprPtr assum,new;
{
  ENTER;

  setMapMerge(&A_IDLOCS(assum),A_IDLOCS(new));
  if (A_KIND(assum)!=TA_PRECOMPILE)
    ; /* don't do anything with externs, there should be none in new */
  else
    {
#ifdef BUGGY
      setMapDomRestrBy(&A_IDEXTS(assum),A_IDS(new)); /* eliminate newly defined ids
							from extern list */
      setMapMerge(&A_IDEXTS(assum),A_IDEXTS(new)); /* add all newly defined externs */
      /* So we have the table:
	 old-status new-status afterwards
	 singular   singular   singular
	 singular   imported   imported
	 singular   regular    absent (?/regular)
	 singular   absent     singular
	 imported   singular   imported
	 imported   imported   imported
	 imported   regular    regular (?/error)
	 imported   absent     imported
	 regular    singular   imported (?/regular)
	 regular    imported   imported (?/regular)
	 regular    regular    regular
	 regular    absent     regular
	 absent     singular   singular
	 absent     imported   singular (?/imported)
	 absent     regular    regular
	 absent     absent     absent
	 */
#endif
      /*
	 old-status new-status afterwards
	 singular   singular   singular
	 singular   imported   imported
	 singular   regular    regular
	 singular   absent     singular
	 imported   singular   imported
	 imported   imported   imported
	 imported   regular    error
	 imported   absent     imported
	 regular    singular   regular
	 regular    imported   regular
	 regular    regular    regular
	 regular    absent     regular
	 absent     singular   singular
	 absent     imported   imported
	 absent     regular    regular
	 absent     absent     absent

	 oldext oldid  newext newid  aftext aftid
	 1      0      1      0      1      0
	 1      0      1      1      1      1
	 1      0      0      1      0      1
	 1      0      0      0      1      0
	 1      1      1      0      1      1
	 1      1      1      1      1      1
	 1      1      0      1      x      x
	 1      1      0      0      1      1
	 0      1      1      0      0      1
	 0      1      1      1      0      1
	 0      1      0      1      0      1
	 0      1      0      0      0      1
	 0      0      1      0      1      0
	 0      0      1      1      1      1
	 0      0      0      1      0      1
	 0      0      0      0      0      0

	 So we can take aftid = oldid \/ newid,
	 aftext = (oldext \ (newid \ newext)) \/ (newext \ oldext \ oldid)
       */
      {
	MapPtr newregs, aux, aux2;

	newregs = MapDomRestrBy (A_IDS(new), A_IDEXTS(new));
	aux = MapDomRestrBy (A_IDEXTS(assum), newregs);
	aux2 = MapDomRestrBy (A_IDEXTS(new), A_IDS(assum));
	aux2 = MapDomRestrBy (aux2, A_IDEXTS(assum));
	aux = MapMerge (aux, aux2);
	SETQ (A_IDEXTS(assum), aux);
      }
    }
  setMapMerge(&A_IDS(assum),A_IDS(new));
  setMapMerge(&A_CONS(assum),A_CONS(new));
  if (A_KIND(assum)!=TA_PRECOMPILE)
    ; /* don't do anything with externs, there should be none in assum */
  else
    {
#ifdef BUGGY
      setMapDomRestrBy(&A_TYEXTS(assum),A_TYFCNS(new)); /* eliminate resolved tyexts */
      setMapMerge(&A_TYEXTS(assum),A_TYEXTS(new)); /* add all new tyexts */
      /* So we have the table:
	 old-status new-status afterwards
	 absent     absent     absent
	 absent     imported   imported
	 absent     regular    regular
	 imported   absent     imported
	 imported   imported   imported
	 imported   regular    regular
	 regular    absent     regular
	 regular    imported   imported (?/regular)
	 regular    regular    regular
       */
#endif
      /*
	 absent     absent     absent
	 absent     imported   imported
	 absent     regular    regular
	 imported   absent     imported
	 imported   imported   imported
	 imported   regular    regular
	 regular    absent     regular
	 regular    imported   regular
	 regular    regular    regular

	 oldext oldfcn newext newfcn aftext aftfcn
	 0      0      0      0      0      0
	 0      0      1      1      1      1
	 0      0      0      1      0      1
	 1      1      0      0      1      1
	 1      1      1      1      1      1
	 1      1      0      1      0      1
	 0      1      0      0      0      1
	 0      1      1      1      0      1
         0      1      0      1      0      1

	 So we can take aftfcn = oldfcn \/ newfcn (as below)
	 aftext = (newext \ oldfcn) \/ (newext /\ oldext) \/ (oldext \ newfcn)
	 // = (newext \ (oldfcn \ oldext)) \/ (oldext \ newfcn)
       */
      ExprPtr aux, aux2;

      aux = MapDomRestrBy (A_TYEXTS(new), A_TYFCNS (assum));
      aux2 = MapDomRestrTo (A_TYEXTS(new), A_TYEXTS (assum));
      setMapDomRestrBy (&A_TYEXTS(assum), A_TYFCNS (new));
      setMapMerge (&A_TYEXTS(assum), aux);
      setMapMerge (&A_TYEXTS(assum), aux2);
    }
  setMapMerge(&A_TYFCNS(assum),A_TYFCNS(new));
#if MODULE_SYSTEM & STD_ML_MODULE
  setMapMerge(&A_STRUCTS(assum),A_STRUCTS(new));
  setMapMerge(&A_SIGS(assum),A_SIGS(new));
#endif
  setMapMerge(&A_NONGENS(assum),A_NONGENS(new));
  setMapMerge(&A_UNITS(assum),A_UNITS(new));
  setMapMerge(&A_TYVARS(assum),A_TYVARS(new));
  QUIT;
}

PUBLIC void FreeTypeAssums()
{
  struct type_assum **assump,*assum;

  for (assump = &allocTypeAssums;
       (assum = *assump)!=(struct type_assum *)0L; )
    {
      if (GC_MARKED(assum))
	{
	  GC_UNMARK(assum);
	  assump = &assum->next;
	}
      else
	{
	  *assump = assum->next;
	  MARKFREE(assum);
	}
    }
}

PUBLIC void FreeTypeAssum(assum)
IN struct type_assum *assum;
{
  struct type_assum **assump;

  for (assump = &allocTypeAssums; *assump!=assum; assump = &(*assump)->next);
  /* should not be too slow because there is certainly few type assums */
  assum->mark = 0;
  *assump = assum->next;
  MARKFREE(assum);
}

PRIVATE struct type_assum *TAReserve[MAX_GENERATIONS];

PUBLIC void cutoffInitialTypeAssums()
{
  REGISTER struct type_assum *assum;

  for (assum = allocTypeAssums; assum!=(struct type_assum *)0L; assum=assum->next)
    {
      GC_MARK(assum);
    }
  TAReserve[0] = oldTypeAssums = allocTypeAssums;
  allocTypeAssums = (struct type_assum *)0L;
}

#ifdef GC_SLIDING
PUBLIC void slideOldTypeAssums()
{
  COUNTER i;

  for (i=1; i<generation_count; i++)
    TAReserve[i] = TAReserve[i+1];
}
#endif

PUBLIC void cutoffOldTypeAssums()
{
  REGISTER struct type_assum **assump,*assum;

  TAReserve[generation_count] = oldTypeAssums;
  for (assump = &allocTypeAssums; (assum = *assump)!=(struct type_assum *)0L;
       assump = &assum->next)
      GC_MARK(assum);
  *assump = oldTypeAssums;
  oldTypeAssums = allocTypeAssums;
  allocTypeAssums = (struct type_assum *)0L;
}

PUBLIC void refreshOldTypeAssums()
{
  REGISTER struct type_assum **assump,*assum,*last;

  last = TAReserve[generation_count];
  for (assump = &oldTypeAssums; assum = *assump, assum!=last; assump = &assum->next)
    {
      GC_UNMARK(assum);
    }
  *assump = allocTypeAssums;
  allocTypeAssums = oldTypeAssums;
  oldTypeAssums = assum;
}

PUBLIC void unmarkOldTypeAssums()
{
  REGISTER struct type_assum *assum;

  for (assum = oldTypeAssums; assum!=(struct type_assum *)0L; assum = assum->next)
    GC_UNMARK(assum);
}

PUBLIC MapPtr IdTypes(assum)
IN CONST ExprPtr assum;
{
  return A_IDS(assum);
}

PUBLIC void addIdTypes(id,type,assum)
IN CONST ExprPtr id;
IN CONST ExprPtr type;
INOUT ExprPtr assum;
{
  setMapPut(id,type,&A_IDS(assum));
}

PUBLIC void addConsTypes(id,type,assum)
IN CONST ExprPtr id;
IN CONST ExprPtr type;
INOUT ExprPtr assum;
{
  setMapPut(id,type,&A_CONS(assum));
}

PUBLIC void remConsTypes(id,assum)
IN CONST ExprPtr id;
INOUT ExprPtr assum;
{
  setMapRemove(id,&A_CONS(assum));
}

PRIVATE struct type_state *allocTypeStates=(struct type_state *)0L;
PRIVATE struct type_state *oldTypeStates=(struct type_state *)0L;

PUBLIC ExprPtr MakeTypeState(conds,eqconds,eqflags,nexgen,noneqs,tycons,packers,
			     annotations,nongens
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			     ,strength,maxstrength
#endif
			     )
IN CONST ExprPtr conds;
IN CONST ExprPtr eqconds;
IN CONST ExprPtr eqflags;
IN COUNTER nexgen;
IN CONST ExprPtr noneqs;
IN CONST ExprPtr tycons;
IN CONST ExprPtr packers;
IN CONST ExprPtr annotations;
IN CONST MapPtr nongens;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN CONST MapPtr strength,maxstrength;
#endif
{
  struct type_state *new;

  new = (struct type_state *)MALLOC(sizeof(struct type_state));
  new->kind = TYPE_STATE_KIND;
  new->mark = 0;
  /*new->gcmark = 0;*/
  new->next = allocTypeStates;
  allocTypeStates = new;
  S_TYVARSUBS(new) = NewMap();
  S_TYVARSUBCACHE(new) = NewMap();
  S_CONDS(new) = conds;
  S_EQCONDS(new) = eqconds;
  S_EQFLAGS(new) = eqflags;
  S_NONEQS(new) = noneqs;
  S_NEXGEN(new) = nexgen;
  S_TYCONS(new) = tycons;
  S_PACKERS(new) = packers;
  S_ANNOTATIONS(new) = annotations;
  S_NONGENS(new) = nongens;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  S_STRENGTH(new) = strength;
  S_MAXSTRENGTH(new) = maxstrength;
#endif
  return KEEP((ExprPtr)new);
}

PUBLIC ExprPtr NewTypeState()
{
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  return MAKE_STATE(NewMap(),trueValue,NewMap(),
		    (COUNTER)0,NewMap(),NewMap(),NIL,NIL,
		    NewMap(),NewMap(),NewMap()
		    );
#else
  return MAKE_STATE(NewMap(),trueValue,NewMap(),
		    (COUNTER)0,NewMap(),NewMap(),NIL,NIL,
		    NewMap()
		    );
#endif
}

PUBLIC ExprPtr DupTypeState(state)
IN CONST ExprPtr state;
{
  ExprPtr newstate;

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  newstate = MakeTypeState(S_CONDS(state),S_EQCONDS(state),
			   S_EQFLAGS(state),S_NEXGEN(state),S_NONEQS(state),
			   S_TYCONS(state),S_PACKERS(state),S_ANNOTATIONS(state),
			   S_NONGENS(state),
			   S_STRENGTH(state),S_MAXSTRENGTH(state)
			   );
#else
  newstate = MakeTypeState(S_CONDS(state),S_EQCONDS(state),
			   S_EQFLAGS(state),S_NEXGEN(state),S_NONEQS(state),
			   S_TYCONS(state),S_PACKERS(state),S_ANNOTATIONS(state),
			   S_NONGENS(state)
			   );
#endif
  SETQ(S_TYVARSUBS(newstate),S_TYVARSUBS(state));
  SETQ(S_TYVARSUBCACHE(newstate),S_TYVARSUBCACHE(state));
  return newstate;
}

PUBLIC void overwriteTypeState(state,new)
INOUT ExprPtr state;
IN CONST ExprPtr new;
{
  SETQ(S_CONDS(state),S_CONDS(new));
  SETQ(S_EQCONDS(state),S_EQCONDS(new));
  SETQ(S_EQFLAGS(state),S_EQFLAGS(new));
  S_NEXGEN(state) = S_NEXGEN(new);
  SETQ(S_NONEQS(state),S_NONEQS(new));
  SETQ(S_TYCONS(state),S_TYCONS(new));
  SETQ(S_PACKERS(state),S_PACKERS(new));
  SETQ(S_ANNOTATIONS(state),S_ANNOTATIONS(new));
  SETQ(S_NONGENS(state),S_NONGENS(new));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  SETQ(S_STRENGTH(state),S_STRENGTH(new));
  SETQ(S_MAXSTRENGTH(state),S_MAXSTRENGTH(new));
#endif
  SETQ(S_TYVARSUBS(state),S_TYVARSUBS(new));
  SETQ(S_TYVARSUBCACHE(state),S_TYVARSUBCACHE(new));
}

PUBLIC void FreeTypeStates()
{
  struct type_state **statep,*state;

  for (statep = &allocTypeStates;
       (state = *statep)!=(struct type_state *)0L; )
    {
      if (GC_MARKED(state))
	{
	  GC_UNMARK(state);
	  statep = &state->next;
	}
      else
	{
	  *statep = state->next;
	  MARKFREE(state);
	}
    }
}

PUBLIC void FreeTypeState(state)
IN struct type_state *state;
{
  struct type_state **statep;

  for (statep = &allocTypeStates; *statep!=state; statep = &(*statep)->next);
  /* should not be too slow because there is certainly few type states */
  state->mark = 0;
  *statep = state->next;
  MARKFREE(state);
}

PRIVATE struct type_state *TSReserve[MAX_GENERATIONS];

PUBLIC void cutoffInitialTypeStates()
{
  REGISTER struct type_state *state;

  for (state = allocTypeStates; state!=(struct type_state *)0L; state=state->next)
    {
      GC_MARK(state);
    }
  TSReserve[0] = oldTypeStates = allocTypeStates;
  allocTypeStates = (struct type_state *)0L;
}

#ifdef GC_SLIDING
PUBLIC void slideOldTypeStates()
{
  COUNTER i;

  for (i=1; i<generation_count; i++)
    TSReserve[i] = TSReserve[i+1];
}
#endif

PUBLIC void cutoffOldTypeStates()
{
  REGISTER struct type_state **statep,*state;

  TSReserve[generation_count] = oldTypeStates;
  for (statep = &allocTypeStates; (state = *statep)!=(struct type_state *)0L;
       statep = &state->next)
    GC_MARK(state);
  *statep = oldTypeStates;
  oldTypeStates = allocTypeStates;
  allocTypeStates = (struct type_state *)0L;
}

PUBLIC void refreshOldTypeStates()
{
  REGISTER struct type_state **statep,*state,*last;

  last = TSReserve[generation_count];
  for (statep = &oldTypeStates; state = *statep, state!=last; statep = &state->next)
    {
      GC_UNMARK(state);
    }
  *statep = allocTypeStates;
  allocTypeStates = oldTypeStates;
  oldTypeStates = state;
}

PUBLIC void unmarkOldTypeStates()
{
  REGISTER struct type_state *state;

  for (state = oldTypeStates; state!=(struct type_state *)0L; state = state->next)
    GC_UNMARK(state);
}

PUBLIC ExprPtr STyCons(state)
IN CONST ExprPtr state;
{
  return S_TYCONS(state);
}

PUBLIC ExprPtr coupleType(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr t;

  ENTER;
  t = KONS(a,KONS(b,NIL));
  KEEPONLY(t);
  return t;
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PUBLIC ExprPtr freshFxVar(constr,sort,assum,state)
IN CONST MapPtr constr;
IN CONST ExprPtr sort;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr sigma,aux;
  ENTER;

  sigma = MakePair(E_ARRAY_2,sort,constr);
  (void) KEEP(sigma);
  setSetAdd(sigma,&A_NONGENS(assum));
  KEEPONLY(sigma);
  return sigma;
}

PUBLIC ExprPtr funTypeFx(a,b,sigma)
IN CONST ExprPtr a,b,sigma;
{
  ExprPtr type,f;
  ENTER;

  f = KtypeFx(funSym,sigma);
  (void) KEEP(f);
  type = KtypeAppl(f,coupleType(a,b));
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr promiseTypeFx(a,sigma)
IN CONST ExprPtr a,sigma;
{
  ExprPtr type,f;
  ENTER;

  f = KtypeFx(promiseSym,sigma);
  (void) KEEP(f);
  type = KtypeAppl(f,KONS(a,NIL));
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr exnTypeFx(rho)
IN CONST ExprPtr rho;
{
  ExprPtr type,f;
  ENTER;

  f = KtypeFx(exnSym,rho);
  (void) KEEP(f);
  type = KtypeAppl(f,NIL);
  KEEPONLY(type);
  return type;
}

PRIVATE ExprPtr initFx1(sym,rho,type)
IN CONST ExprPtr sym;
IN CONST ExprPtr rho,type;
{
  ExprPtr f;
  ENTER;

  f = KtypeFx(sym,rho);
  (void) KEEP(f);
  type = KtypeAppl(f,KONS(type,NIL));
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr initFx(rho,type)
IN CONST ExprPtr rho,type;
{
  return initFx1(fxInitSym,rho,type);
}

PUBLIC ExprPtr exnFx(rho,type)
IN CONST ExprPtr rho,type;
{
  return initFx1(fxExnSym,rho,type);
}

PRIVATE ExprPtr zFx(sym,rho)
IN CONST ExprPtr sym,rho;
{
  ExprPtr type,f;
  ENTER;

  f = KtypeFx(sym,rho);
  (void) KEEP(f);
  type = KtypeAppl(f,NIL);
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr fromFx(rho)
IN CONST ExprPtr rho;
{
  return zFx(fxFromSym,rho);
}

PUBLIC ExprPtr gotoFx(rho)
IN CONST ExprPtr rho;
{
  return zFx(fxGotoSym,rho);
}

PUBLIC ExprPtr readFx(rho)
IN CONST ExprPtr rho;
{
  return zFx(fxReadSym,rho);
}

PUBLIC ExprPtr writeFx(rho)
IN CONST ExprPtr rho;
{
  return zFx(fxWriteSym,rho);
}

PRIVATE ExprPtr dupFxVar(sigma,constr,assum)
IN CONST ExprPtr sigma,constr;
INOUT ExprPtr assum;
{
  ExprPtr sigma2;

  sigma2 = MakePair(E_ARRAY_2,CAR(sigma),constr);
  (void) KEEP(sigma2);
  if (!genericp(sigma,assum))
    setSetAdd(sigma2,&A_NONGENS(assum));
}

PRIVATE void addTopFx1(type,assum,state)
INOUT ExprPtr type; /* when type is an application */
INOUT ExprPtr assum;
IN CONST ExprPtr state;
{
  ExprPtr f,sigma,wrap,t,constr;
  ENTER;

  f = APPLFUN(type);
  if (f->kind==(CHAR)E_TYPEFX)
    {
      sigma = CDR(f);
      if (CAR(sigma)==effectSym) /* add top effects only to effects, that is,
				to effect variables of sort "fx" */
	{
	  wrap = KtypeAppl(CAR(f),APPLARG(type));
	  t = KtypeAppl(fxTopSym,KONS(KEEP(wrap),NIL));
	  constr = SetAdd(KEEP(t),CDR(sigma));
	  SETCDR(sigma,constr);
	}
    }
  QUIT;
}

MAPSWEEPFUNC(do_addtopfx)
{
  struct addtopfx *data;

  data = MAPDATA(struct addtopfx *);
  addTopFx(y,data->assum,data->state);
  return false;
}

MAPSWEEPFUNC(do_addtopfx_num)
{
  struct addtopfx *data;

  data = MAPDATA(struct addtopfx *);
  addTopFx(x,data->assum,data->state);
  return false;
}

PUBLIC void addTopFx(type,assum,state)
INOUT ExprPtr type;
INOUT ExprPtr assum;
IN CONST ExprPtr state;
{
  Boolean found;

again:
  /*if (type==NIL) return;*/
  switch (type->kind)
    {
      case E_VAR:
	type = nongenValue(type,state,&found);
	if (found)
	  goto again;
	break;
      case E_TYPE_APPL:
	addTopFx1(type,assum,state);
	type = APPLARG(type);
	goto again;
      case E_EMPTY:
	break;
      case E_MAPLET:
      case E_MAP:
	{
	  struct addtopfx *data = (struct addtopfx *)alloctmp(sizeof(struct addtopfx));

	  data->assum = assum;
	  data->state = state;
	  (void) MapSweep(type,do_addtopfx,data);
	  break;
	}
      case E_NUMTYPE:
	{
	  struct addtopfx *data = (struct addtopfx *)alloctmp(sizeof(struct addtopfx));

	  data->assum = assum;
	  data->state = state;
	  (void) MapSweep(REFVAL(type),do_addtopfx_num,data);
	  break;
	}
      case E_CONS:
	addTopFx(CAR(type),assum,state);
	type = CDR(type);
	goto again;
      default:
	kinderror(message(471),type->kind);
	break;
    }
}
#else
PUBLIC ExprPtr promiseType(a)
IN CONST ExprPtr a;
{
  ExprPtr type;
  ENTER;

  type = KtypeAppl(promiseSym,KONS(a,NIL));
  KEEPONLY(type);
  return type;
}
#endif

PUBLIC ExprPtr funType(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr type;
  ENTER;

  type = KtypeAppl(funSym,coupleType(a,b));
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr mapType(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr type;
  ENTER;

  type = KtypeAppl(mapSym,coupleType(a,b));
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr listType(a)
IN CONST ExprPtr a;
{
  ExprPtr type;
  ENTER;

  type = KtypeAppl(listSym,KONS(a,NIL));
  KEEPONLY(type);
  return type;
}

	/***** fresh type instantiations *****/

PUBLIC Boolean genericp(typevar,assum)
IN CONST ExprPtr typevar;
IN CONST ExprPtr assum;
{
  Boolean found;

  (void) MapGet(typevar,A_NONGENS(assum),&found);
  return !found;
}

PUBLIC int varFlags(tyvar)
IN CONST SymPtr tyvar;
{
  REGISTER int flags;
  REGISTER INT len;
  REGISTER CHAR *HEAP name;

  flags = 0;
  len = symNameLen(tyvar)-1;
  name = symName(tyvar)+1;
  if (len>0 && *name=='\'')
    {
      flags |= F_EQ;
      while (len>0 && *name=='\'')
	{
	  name++;
	  len--;
	}
    }
  if (len>0 && *name=='_')
    {
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
      flags |= F_IMP;
#endif
      while (len>0 && *name=='_')
	{
	  name++;
	  len--;
	}
    }
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  else if (len>0 && *name>='0' && *name<='9')
    {
      flags |= F_IMP;
      while (++name, *name>='0' && *name<='9');
    }
#endif
  if (len>0 && *name=='#')
    {
      flags |= F_NUM|F_EQ; /* every numeric type admits equality */
    }
  return flags;
}

PUBLIC ExprPtr varEqTdg(var,state)
IN CONST ExprPtr var;
IN CONST ExprPtr state;
{
  ExprPtr t;
  Boolean found;

  t = MapGet(var,S_EQFLAGS(state),&found);
  return found?tdgInst(t,S_NONEQS(state),false):
    ((varFlags(var)&F_EQ)?trueValue:falseValue);
}

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
PUBLIC INT naturalStrength(tyvar)
IN CONST SymPtr tyvar;
{
  REGISTER INT len;
  REGISTER CHAR *HEAP name;
  INT n;

  len = symNameLen(tyvar)-1;
  name = symName(tyvar)+1;
  while (len>0 && *name=='\'')
    {
      name++;
      len--;
    }
  if (len>0 && *name=='_')
    n = 1;
  else if (len>0 && *name>='0' && *name<='9')
    {
      n = (INT)(*name - '0');
      while (++name, *name>='0' && *name<='9')
	n = n*10L + (INT)(*name - '0');
    }
  else
    n = INF_STRENGTH;
  return n;
}

PUBLIC INT tyvarStrength(tyvar,state)
IN CONST SymPtr tyvar;
IN CONST ExprPtr state;
{
  REGISTER INT len;
  REGISTER CHAR *HEAP name;
  INT n;
  ExprPtr s;
  Boolean found;

  len = symNameLen(tyvar)-1;
  name = symName(tyvar)+1;
  while (len>0 && *name=='\'')
    {
      name++;
      len--;
    }
  if (state!=NIL && (s = MapGet(tyvar,S_STRENGTH(state),&found), found))
    n = IVAL(s);
  else if (len>0 && *name=='_')
    n = 1;
  else if (len>0 && *name>='0' && *name<='9')
    {
      n = (INT)(*name - '0');
      while (++name, *name>='0' && *name<='9')
	n = n*10L + (INT)(*name - '0');
    }
  else
    n = INF_STRENGTH;
  return n;
}

PUBLIC INT tyvarMaxStrength(tyvar,state)
IN CONST SymPtr tyvar;
IN CONST ExprPtr state;
{
  ExprPtr aux;
  Boolean found;

  aux = MapGet(tyvar,S_MAXSTRENGTH(state),&found);
  return found?IVAL(aux):INF_STRENGTH;
}

PRIVATE void setMaxStrength(tyvar,maxstrength,state)
IN CONST SymPtr tyvar;
IN INT maxstrength;
INOUT ExprPtr state;
{
  ExprPtr aux;
  Boolean found;
  INT oldmax;

  aux = MapGet(tyvar,S_MAXSTRENGTH(state),&found);
  oldmax = found?IVAL(aux):INF_STRENGTH;
  if (oldmax>maxstrength)
    setMapPut(tyvar,KINT(maxstrength),&S_MAXSTRENGTH(state));
}

#endif

PUBLIC void getNumber(n,buf)
IN COUNTER n;
OUT CHAR *buf;
{
  CHAR *s,buf2[16];

  for (s=buf2; n!=0; )
    {
      *s++ = 'a'+(n%26);
      n /= 26;
    }
  while (--s>=buf2)
    *buf++ = *s;
  *buf++ = '?';
  *buf = 0;
}

#ifdef HAS_PROTOTYPES
PUBLIC SymPtr freshInstfromFlags(IN CONST SHORT flags,
				 IN CONST ExprPtr eqtdg,
				 INOUT ExprPtr assum,INOUT ExprPtr state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				 ,IN INT strength,IN INT maxstrength
#endif
				 )
#else
PUBLIC SymPtr freshInstfromFlags(flags,eqtdg,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				 ,strength,maxstrength
#endif
				 )
IN CONST SHORT flags;
IN CONST ExprPtr eqtdg;
INOUT ExprPtr assum;
INOUT ExprPtr state;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength,maxstrength;
#endif
#endif
{
  SymPtr new;
  CHAR *t,buf[16];
  ENTER;

  if (flags & F_NUM)
    eqtdg = trueValue;
  else
    eqtdg = tdgInst(eqtdg,S_NONEQS(state),false);
  if (eqtdg==trueValue)
    flags |= F_EQ;
  else
    flags &= ~F_EQ;
  t = buf;
  *t++ = '\'';
  if ((flags & F_EQ) && !(flags & F_NUM))
    /* since every numeric type admits equality,
       it is useless to add a tick sign when the numeric flag is on */
    *t++ = '\'';
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
  if (flags & F_IMP)
    *t++ = '_';
#endif
  if (flags & F_NUM)
    *t++ = '#';
  getNumber(++S_NEXGEN(state),t);
  new = Ksymbol(buf);
  (void) KEEP(new);
  /* then make the new variable non-generic */
  setSetAdd(new,&A_NONGENS(assum));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  if (flags & F_IMP)
    {
      setMapPut(new,KINT(strength),&S_STRENGTH(state));
      if (maxstrength!=INF_STRENGTH)
	setMapPut(new,KINT(maxstrength),&S_MAXSTRENGTH(state));
    }
#endif
  if (eqtdg!=trueValue && eqtdg!=falseValue)
    {
      setMapPut(new,eqtdg,&S_EQFLAGS(state));
    }
  KEEPONLY(new);
#ifdef DEBUG1
  {
    hfprintf(hstderr,"creating var ");
    TP(new);
    hfprintf(hstderr,"\n");
  }
#endif
  return new;
}

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr _freshInstfromFlags(IN CONST SHORT flags,IN CONST ExprPtr eqtdg,
				   INOUT ExprPtr assum,INOUT ExprPtr state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				   ,IN INT strength,IN INT maxstrength
#endif
				   )
#else
PUBLIC ExprPtr _freshInstfromFlags(flags,eqtdg,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				   ,strength,maxstrength
#endif
				   )
IN CONST SHORT flags;
IN CONST ExprPtr eqtdg;
INOUT ExprPtr assum;
INOUT ExprPtr state;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength,maxstrength;
#endif
#endif
{
  SymPtr sym;

  sym = freshInstfromFlags(flags,eqtdg,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			   ,strength,maxstrength
#endif
			   );
  return (flags&F_NUM)?(sym = KnumType(MapSingleton(sym,one)), KEEP(sym)):sym;
}

PRIVATE SymPtr freshInst(typevar,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			 noimpp,
#endif
			 fIrename)
IN CONST SymPtr typevar;
INOUT ExprPtr assum;
INOUT ExprPtr state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
INOUT MapPtr *fIrename;
{
  SymPtr new;
  Boolean found;

  new = MapGet(typevar,*fIrename,&found);
  if (found)
    return new;
  new = freshInstfromFlags((
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			    noimpp?(varFlags(typevar) & ~F_IMP):
#endif
			    (varFlags(typevar))),
			   varEqTdg(typevar,state),
			   assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			   ,(noimpp?INF_STRENGTH:tyvarStrength(typevar,state)),INF_STRENGTH
#endif
			   );
  *fIrename = MapPut(typevar,new,*fIrename);
  return new;
}

PUBLIC ExprPtr _nongenValue(typevar,state)
IN CONST ExprPtr typevar;
IN CONST ExprPtr state;
{
  ExprPtr type;
  Boolean found;

  do {
    type = nongenValue(typevar,state,&found);
    if (found)
      {
	typevar = type;
	if (/*typevar!=NIL &&*/
	    typevar->kind==(CHAR)E_NUMTYPE &&
	    REFVAL(typevar)->kind==(CHAR)E_MAPLET &&
	    MAPLETY(REFVAL(typevar))==one &&
	    MAPLETX(REFVAL(typevar))->kind==(CHAR)E_VAR)
	  typevar = MAPLETX(REFVAL(typevar));
      }
  } while (found &&
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	   (typevar->kind==(CHAR)E_VAR || typevar->kind==(CHAR)E_ARRAY_2)
#else
	   typevar->kind==(CHAR)E_VAR
#endif
	   );
  return (/*typevar!=NIL && */typevar->kind==(CHAR)E_VAR && (varFlags(typevar) & F_NUM))?
    MapSingleton(typevar,one):typevar;
}

PUBLIC ExprPtr realName(tyvar,state)
IN CONST ExprPtr tyvar;
IN CONST ExprPtr state;
{
  ExprPtr type;
  Boolean found;

  do {
    type = nongenValue(tyvar,state,&found);
    if (found)
      tyvar = type;
  } while (found);
  return tyvar;
}

PUBLIC ExprPtr applFun(type,state)
IN CONST ExprPtr type;
IN CONST ExprPtr state;
{
  return realName(APPLFUN(type),state);
}

MAPSWEEPFUNC(do_newtypeinst)
{
  struct newtypeinst *data;
  MapPtr ren;

  data = MAPDATA(struct newtypeinst *);
  ren = REFVAL(data->fIrename);
  if (x==REST_TAG)
    {
      if (y->kind==(CHAR)E_NUMTYPE)
	y = REFVAL(y);
      y = newTypeInst1(y,data->assum,data->state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		       data->noimpp,
#endif
		       &ren);
      if (ISMAP(y))
	data->new = MapMerge(data->new,y);
      else
	data->new = MapPut(x,y,data->new);
    }
  else
    {
      data->new = MapPut(x,
			 newTypeInst1(y,data->assum,data->state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				      data->noimpp,
#endif
				      &ren),
			 data->new);
    }
  SETREFVAL(data->fIrename,ren);
  return false;
}

MAPSWEEPFUNC(do_newtypeinst_num)
{
  struct newtypeinst *data;
  ExprPtr new,ren;

  data = MAPDATA(struct newtypeinst *);
  ren = REFVAL(data->fIrename);
  new = newTypeInst1(x,data->assum,data->state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		     data->noimpp,
#endif
		     &ren);
  SETREFVAL(data->fIrename,ren);
  data->new = typeMul1((new->kind==(CHAR)E_NUMTYPE)?typePower1(REFVAL(new),y):MapSingleton(new,y),
		       data->new);
  return false;
}

PRIVATE ExprPtr newTyVarInst(tyvar,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			     noimpp,
#endif
			     fIrename)
IN CONST ExprPtr tyvar;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
INOUT MapPtr *fIrename;
{
  if (assum==NIL)
    return tyvar;
  return freshInst(tyvar,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		   noimpp,
#endif
		   fIrename);
}

PUBLIC MapPtr getSubst1(varlist,arglist)
IN CONST ExprPtr varlist,arglist;
{
  MapPtr s;
  ENTER;

  s = NewMap();
  while (varlist!=NIL && arglist!=NIL)
    {
      s = MapPut(CAR(varlist),CAR(arglist),s);
      varlist = CDR(varlist);
      arglist = CDR(arglist);
    }
  KEEPONLY(s);
  return s;
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE MapPtr do_newtypeinst_fx P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT CHAR *data));

#ifdef HAS_PROTOTYPES
PRIVATE ExprPtr newTypeInstFx1(IN CONST ExprPtr sigma,
			       INOUT ExprPtr assum,INOUT ExprPtr state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			       IN Boolean noimpp,
#endif
			       INOUT MapPtr *fIrename)
#else
PRIVATE ExprPtr newTypeInstFx1(sigma,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			       noimpp,
#endif
			       fIrename)
IN CONST ExprPtr sigma;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
INOUT MapPtr *fIrename;
#endif
{
  struct newtypeinst *data;
  ExprPtr constr,news;
  Boolean found;

  if (assum!=NIL && !genericp(sigma,assum))
    return sigma;
  news = MapGet(sigma,*fIrename,&found);
  if (!found)
    {
      constr = CDR(sigma);
      data = (struct newtypeinst *)alloctmp(sizeof(struct newtypeinst));
      data->assum = assum;
      data->state = state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
      data->noimpp = noimpp;
#endif
      data->fIrename = MakeSing(E_REF,*fIrename);
      (void) KEEP(data->fIrename);
      constr = MapCollect(constr,do_newtypeinst_fx,data);
      if (assum==NIL)
	{
	  SETCDR(sigma,constr);
	  news = sigma;
	}
      else
	news = freshFxVar(constr,CAR(sigma),assum,state);
      *fIrename = MapPut(sigma,news,REFVAL(data->fIrename));
    }
  return news;
}

MAPCOLLECTFUNC(do_newtypeinst_fx)
{
  struct newtypeinst *data;
  ExprPtr ren;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
  ExprPtr (*inst) P((ExprPtr f,ExprPtr assum,ExprPtr state,
		     Boolean noimpp,
		     MapPtr *rename));
#else
  ExprPtr (*inst) P((ExprPtr f,ExprPtr assum,ExprPtr state,
		     MapPtr *rename));
#endif

  data = MAPDATA(struct newtypeinst *);
  ren = REFVAL(data->fIrename);
  inst = (x->kind==(CHAR)E_ARRAY_2)?newTypeInstFx1:newTypeInst1;
  x = (*inst)(x,data->assum,data->state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
	      data->noimpp,
#endif
	      &ren);
  SETREFVAL(data->fIrename,ren);
  return MapSingleton(x,y);
}

#ifdef HAS_PROTOTYPES
PRIVATE ExprPtr newTypeInstFx(IN CONST ExprPtr f,
			      INOUT ExprPtr assum,INOUT ExprPtr state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			      IN Boolean noimpp,
#endif
			      INOUT MapPtr *fIrename)
#else
PRIVATE ExprPtr newTypeInstFx(f,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			      noimpp,
#endif
			      fIrename)
IN CONST ExprPtr f;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
INOUT MapPtr *fIrename;
#endif
{
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
  f = KtypeFx(CAR(f),newTypeInstFx1(CDR(f),assum,state,noimpp,fIrename));
#else
  f = KtypeFx(CAR(f),newTypeInstFx1(CDR(f),assum,state,fIrename));
#endif
  return KEEP(f);
}

PRIVATE ExprPtr decorateFx(f,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			   ,noimpp
#endif
			   )
IN CONST ExprPtr f;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
{
  ExprPtr sigma,rename;
  Boolean found;
  ENTER;

  sigma = MapGet(f,autoDecorations,&found);
  if (found)
    {
      rename = NewMap();
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
      f = KtypeFx(f,newTypeInstFx1(sigma,assum,state,noimpp,&rename));
#else
      f = KtypeFx(f,newTypeInstFx1(sigma,assum,state,&rename));
#endif
      KEEPONLY(f);
    }
  return f;
}
#endif

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr newTypeInst1(IN CONST ExprPtr type,INOUT ExprPtr assum,INOUT ExprPtr state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			    IN Boolean noimpp,
#endif
			    INOUT MapPtr *fIrename)
#else
PUBLIC ExprPtr newTypeInst1(type,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			    noimpp,
#endif
			    fIrename)
IN CONST ExprPtr type;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
INOUT MapPtr *fIrename;
#endif
{
again:
  /*if (type==NIL)
    return NIL;*/
  switch (type->kind)
    {
      case E_NIL:
	return NIL;
      case E_VAR:
	if (assum!=NIL && genericp(type,assum))
	  return newTyVarInst(type,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				noimpp,
#endif
				fIrename);
	else
	  {
	    ExprPtr val;
	    Boolean found;

	    val = nongenValue(type,state,&found);
	    if (!found)
	      return type;
	    type = val;
	    goto again;
	  }
      case E_TYPE_APPL:
	{
	  ExprPtr f,s,l;

	  l = newTypeInst1(APPLARG(type),assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			   noimpp,
#endif
			   fIrename);
	  f = applFun(type,state);
	  switch (f->kind)
	    {
	      case E_CONS: /* f is a tyfcn (tyhdr,tyexpr) */
		s = getSubst1(CAR(CAR(f)),l);
		/* no need to check that the substitution obeys the constraints
		on attributes, because it will */
		return newTypeInst1(CDR(f),NIL,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		      		    noimpp,
#endif
		      		    &s);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	      case E_TYPEFX:
		f = newTypeInstFx(f,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				  noimpp,
#endif
				  fIrename);
		break;
	      default: /* case E_VAR */
		f = decorateFx(f,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			       ,noimpp
#endif
				);
		break;
#endif
	    }
	  type = KtypeAppl(f,l);
	  return KEEP(type);
	}
      case E_CONS:
	return KONS(newTypeInst1(CAR(type),assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				 noimpp,
#endif
				 fIrename),
		    newTypeInst1(CDR(type),assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			    	 noimpp,
#endif
			    	 fIrename));
      case E_EMPTY:
	return type;
      case E_MAPLET:
	return MapSingleton(MAPLETX(type),
			    newTypeInst1(MAPLETY(type),assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
					 noimpp,
#endif
					 fIrename));
      case E_MAP:
	{
	  struct newtypeinst *data = (struct newtypeinst *)alloctmp(sizeof(struct newtypeinst));

	  data->new = NewMap();
	  data->assum = assum;
	  data->state = state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
	  data->noimpp = noimpp;
#endif
	  data->fIrename = MakeSing(E_REF,*fIrename);
	  (void) KEEP(data->fIrename);
	  (void) MapSweep(type,do_newtypeinst,data);
	  *fIrename = REFVAL(data->fIrename);
	  return data->new;
	}
      case E_NUMTYPE:
	if (REFVAL(type)==NewMap())
	  return type;
	{
	  struct newtypeinst *data = (struct newtypeinst *)alloctmp(sizeof(struct newtypeinst));

	  data->new = NewMap();
	  data->assum = assum;
	  data->state = state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
	  data->noimpp = noimpp;
#endif
	  data->fIrename = MakeSing(E_REF,*fIrename);
	  (void) KEEP(data->fIrename);
	  (void) MapSweep(REFVAL(type),do_newtypeinst_num,data);
	  *fIrename = REFVAL(data->fIrename);
	  type = KnumType(data->new);
	  return KEEP(type);
	}
      default:
	kinderror(message(22),type->kind);
	return NIL;
    }
}

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr newTypeInst(IN CONST ExprPtr type,INOUT ExprPtr assum,INOUT ExprPtr state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			   ,IN Boolean noimpp
#endif
			   )
#else
PUBLIC ExprPtr newTypeInst(type,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			   ,noimpp
#endif
			   )
IN CONST ExprPtr type;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
#endif
{
  ExprPtr fIrename;
  ENTER;

  fIrename = NewMap();
  type = newTypeInst1(type,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		      noimpp,
#endif
		      &fIrename);
  KEEPONLY(type);
  return type;
}

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr varType(IN CONST ExprPtr var,INOUT ExprPtr assum,INOUT ExprPtr state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			,IN Boolean noimpp
#endif
			)
#else
PUBLIC ExprPtr varType(var,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			,noimpp
#endif
			)
IN CONST ExprPtr var;
INOUT ExprPtr assum,state;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IN Boolean noimpp;
#endif
#endif
{
  ExprPtr type;
  Boolean found;

  type = MapGet(var,A_IDS(assum),&found);
  if (!found)
    { /* try and see whether it is an external variable */
      type = MapGet(var,A_IDEXTS(assum),&found);
      if (!found)
	return NOEXPR;
      /* if it was indeed external */
      setMapPut(var,type,&A_IDS(assum));
      /* then register it as imported indeed (and not to be redefined) */
    }
  return newTypeInst(type,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
		     ,noimpp
#endif
		     );
}

PUBLIC Boolean enumerateIds(f,data)
     IN CONST mapSweepFunc f;
     INOUT CHAR *data;
{
  return MapSweep(A_IDS(basisAssum),f,data);
}

MAPFOLDFUNC(do_eqtypep)
{
  struct eqtypep *data;
  ExprPtr t;

  data = MAPDATA(struct eqtypep *);
  t = eqTypeTdg(y,data->state,data->eqp);
  if (t==falseValue)
    LONGJUMP(data->exit,1);
  return t;
}

PRIVATE ExprPtr eqtypetdg(type,data)
     IN ExprPtr type;
     IN struct eqtypep *data;
{
  if (SETJUMP(data->exit)==0)
      return MapFold(type,do_eqtypep,data,trueValue,tdgAnd);
  else return falseValue;
}

PUBLIC ExprPtr eqTypeTdg(type,state,eqp)
IN CONST ExprPtr type;
INOUT ExprPtr state; /* returns the tdg expressing when type admits equality */
IN Boolean eqp;
{
again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return trueValue;*/
  switch (type->kind)
    {
      case E_NIL:
	return trueValue;
      case E_VAR:
	{
	  ExprPtr val;
	  Boolean found;

	  if (IN_FLAGS(F_EQ,varFlags(type)))
	    return trueValue;
	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      type = val;
	      goto again;
	    }
	  return eqp?trueValue:tdgInst(varEqTdg(type,state),S_NONEQS(state),false);
	}
      case E_TYPE_APPL:
	{
	  ExprPtr f,t;

	  f = APPLFUN(type); /* correct because of correctTypeAppl() */
	  f = fxSym(f);
	  if (f==refSym || f==arraySym || f==tableSym)
	    return trueValue;
	  t = funEqTdg(f,state);
	  if (t!=falseValue)
	    for (type=APPLARG(type); type!=NIL; type=CDR(type))
	      {
		t = tdgAnd(t,eqTypeTdg(CAR(type),state,eqp));
		if (t==falseValue)
		  break;
	      }
	  t = tdgImply(funNotRefTdg(f,state),t);
		/* it admits equality if either f is ref or array, or
		   f respects equality and all arguments admit equality */
	  return t;
	}
      case E_CONS:
	{
	  ExprPtr t;

	  t = eqTypeTdg(CAR(type),state,eqp);
	  if (t!=falseValue)
	    t = tdgAnd(t,eqTypeTdg(CDR(type),state,eqp));
	  return t;
	}
      case E_EMPTY:
	return trueValue;
      case E_MAPLET:
	return eqTypeTdg(MAPLETY(type),state,eqp);
      case E_MAP:
	{
	  struct eqtypep *data = (struct eqtypep *)alloctmp(sizeof(struct eqtypep));

	  data->state = state;
	  data->eqp = eqp;
	  return eqtypetdg(type,data);
	}
      case E_NUMTYPE:
	return trueValue; /* every numerical type admits equality */
      default:
	kinderror(message(23),type->kind);
      break;
    }
  return falseValue;
}

#if IMPERATIVE_TYPES==STD_ML_TYVAR
MAPSWEEPFUNC(do_imptypep)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
  return !impTypep(y,state);
}

MAPSWEEPFUNC(do_imptypep_num)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
  return !impTypep(x,state);
}

PUBLIC Boolean impTypep(type,state)
IN CONST ExprPtr type;
INOUT ExprPtr state;
{
again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return true;*/
  switch (type->kind)
    {
      case E_NIL:
	return true;
      case E_VAR:
	{
	  ExprPtr val;
	  Boolean found;

	  if (IN_FLAGS(F_IMP,varFlags(type)))
	    return true;
	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      type = val;
	      goto again;
	    }
	  return false;
	}
      case E_TYPE_APPL:
	for (type=APPLARG(type); type!=NIL; type=CDR(type))
	  {
	    if (!impTypep(CAR(type),state))
	      return false;
	  }
	return true;
      case E_CONS:
	if (!impTypep(CAR(type),state))
	  return false;
	type = CDR(type);
	goto again;
      case E_EMPTY:
	return true;
      case E_MAPLET:
	type = MAPLETY(type);
	goto again;
      case E_MAP:
	return !MapSweep(type,do_imptypep,state);
      case E_NUMTYPE:
	return !MapSweep(REFVAL(type),do_imptypep_num,state);
      default:
	kinderror(message(24),type->kind);
	break;
    }
  return false;
}
#endif

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
MAPSWEEPFUNC(do_imptypep)
{
  struct imptypep *data;

  data = MAPDATA(struct imptypep *);
  return !impTypep(y,data->state,data->strength);
}

MAPSWEEPFUNC(do_imptypep_num)
{
  struct imptypep *data;

  data = MAPDATA(struct imptypep *);
  return !impTypep(x,data->state,data->strength);
}

PUBLIC Boolean impTypep(type,state,strength)
IN CONST ExprPtr type;
INOUT ExprPtr state;
IN INT strength;
{
again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return true;*/
  switch (type->kind)
    {
      case E_NIL:
	return true;
      case E_VAR:
	{
	  ExprPtr val;
	  Boolean found;

	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      type = val;
	      goto again;
	    }
	  return (IN_FLAGS(F_IMP,varFlags(type)) &&
		 tyvarStrength(type,state)<=strength);
	}
      case E_TYPE_APPL:
	for (type=APPLARG(type); type!=NIL; type=CDR(type))
	  {
	    if (!impTypep(CAR(type),state,strength))
	      return false;
	  }
	return true;
      case E_CONS:
	if (!impTypep(CAR(type),state,strength))
	  return false;
	type = CDR(type);
	goto again;
      case E_EMPTY:
	return true;
      case E_MAPLET:
	type = MAPLETY(type);
	goto again;
      case E_MAP:
	{
	  struct imptypep *data = (struct imptypep *)alloctmp(sizeof(struct imptypep));

	  data->strength = strength;
	  data->state = state;
	  return !MapSweep(type,do_imptypep,data);
	}
      case E_NUMTYPE:
	{
	  struct imptypep *data = (struct imptypep *)alloctmp(sizeof(struct imptypep));

	  data->strength = strength;
	  data->state = state;
	  return !MapSweep(REFVAL(type),do_imptypep_num,data);
	}
      default:
	kinderror(message(24),type->kind);
	break;
    }
  return false;
}
#endif

PUBLIC Boolean numTypep(type,state)
IN CONST ExprPtr type;
INOUT ExprPtr state;
{
again:
  /*if (type==NIL)
    return false;*/
  switch (type->kind)
    {
      case E_NIL:
	return false;
      case E_VAR:
	{
	  ExprPtr val;
	  Boolean found;

	  if (IN_FLAGS(F_NUM,varFlags(type)))
	    return true;
	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      type = val;
	      goto again;
	    }
	  return false;
	}
      case E_NUMTYPE:
	return true;
      default:
	return false;
    }
}

PRIVATE void nonNumError(zone,type,state)
IN CONST TZone *zone;
IN CONST ExprPtr type;
INOUT ExprPtr state;
{
  typeError(zone,(CHAR *)message(25),type,state);
}

MAPCOLLECTFUNC(do_typemul)
{
  struct typemul *data;
  ExprPtr y1,y2;
  Boolean found;

  data = MAPDATA(struct typemul *);
  y1 = MapGet(x,data->type1,&found);
  if (found)
    {
      y2 = MapGet(x,data->type2,&found);
      if (found)
	{
	  y = z_add(y1,y2);
	  (void) KEEP(y);
	}
      else
	y = y1;
    }
  else
    {
      y2 = MapGet(x,data->type2,&found);
      if (found)
	y = y2;
      else
	y = zzero;
    }
  return (y==zzero)?NewMap():MapSingleton(x,y);
}

PUBLIC MapPtr typeMul1(map1,map2) /* used for numerical types, and also for variable scales */
IN CONST MapPtr map1,map2;
{
  struct typemul *data = (struct typemul *)alloctmp(sizeof(struct typemul));
  ExprPtr dom;

  dom = SetUnion(map1,map2);
  data->type1 = map1;
  data->type2 = map2;
  return MapCollect(dom,do_typemul,data);
}

PUBLIC ExprPtr typeMul(zone1,zone2,type1,type2,state)
IN CONST TZone *zone1,*zone2;
IN CONST ExprPtr type1,type2;
IN CONST ExprPtr state;
{
  ENTER;

  /*if (type1==NIL)
    goto error1;*/
  if (type1->kind==(CHAR)E_VAR)
    type1 = _nongenValue(type1,state);
  if (type1->kind!=(CHAR)E_NUMTYPE)
    {
      /*error1:*/
      nonNumError(zone1,type1,state);
      /*error:*/
      type1 = KnumType(NewMap());
      KEEPONLY(type1);
      return type1;
    }
  /*if (type2==NIL)
    goto error2;*/
  if (type2->kind==(CHAR)E_VAR)
    type2 = _nongenValue(type2,state);
  if (type2->kind!=(CHAR)E_NUMTYPE)
    {
      /*error2:*/
      nonNumError(zone2,type2,state);
      KEEPONLY(type1);
      return type1;
    }
  type1 = KnumType(typeMul1(REFVAL(type1),REFVAL(type2)));
  KEEPONLY(type1);
  return type1;
}

PUBLIC ExprPtr scaleMul(scale1,scale2)
IN CONST ExprPtr scale1,scale2;
{
  ExprPtr s,scale;
  ENTER;

  scale = z_mul(CAR(scale1),CAR(scale2));
  (void) KEEP(scale);
  s = Ktuple2(scale,typeMul1(CDR(scale1),CDR(scale2)));
  KEEPONLY(s);
  return s;
}

MAPCOLLECTFUNC(do_typepower)
{
  ExprPtr z;

  z = z_mul(y,MAPDATA(ExprPtr));
  return (z==zzero)?NewMap():MapSingleton(x,KEEP(z));
}

PUBLIC MapPtr typePower1(map,exp) /* used for numerical types, and also for variable scales */
IN CONST MapPtr map;
IN CONST ExprPtr exp;
{
  return MapCollect(map,do_typepower,exp);
}

PUBLIC ExprPtr typePower(zone,type,exp,state)
IN CONST TZone *zone;
IN CONST ExprPtr type,exp;
IN CONST ExprPtr state;
{
  ENTER;

  /*if (type==NIL)
    goto error;*/
  if (type->kind==(CHAR)E_VAR)
    type = _nongenValue(type,state);
  if (type->kind!=(CHAR)E_NUMTYPE)
    {
      /*error:*/
      nonNumError(zone,type,state);
      type = KnumType(NewMap());
      KEEPONLY(type);
      return type;
    }
  type = KnumType(typePower1(REFVAL(type),exp));
  KEEPONLY(type);
  return type;
}

PUBLIC ExprPtr scalePower(scale,exp)
IN CONST ExprPtr scale,exp;
{
  ExprPtr s,ss;
  ENTER;

  ss = z_pow(CAR(scale),exp);
  s = Ktuple2(ss,typePower1(CDR(scale),exp));
  KEEPONLY(s);
  return s;
}

	/*****  unification *****/

PRIVATE Boolean occursp P((IN CONST ExprPtr typevar,IN CONST ExprPtr type,IN CONST ExprPtr state));

MAPSWEEPFUNC(do_occursp)
{
  struct occursp *data;

  data = MAPDATA(struct occursp *);
  return occursp(data->typevar,y,data->state);
}

MAPSWEEPFUNC(do_occursp_num)
{
  struct occursp *data;

  data = MAPDATA(struct occursp *);
  return occursp(data->typevar,x,data->state);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PUBLIC ExprPtr fxRealSigma(sigma,state)
IN CONST ExprPtr sigma;
IN CONST ExprPtr state;
{
  ExprPtr val;
  Boolean found;

again:
  val = nongenValue(sigma,state,&found);
  if (found)
    {
      sigma = val;
      goto again;
    }
  return sigma;
}

PRIVATE Boolean do_occursp_fx P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT CHAR *data));

PRIVATE Boolean occurspFx1(typevar,sigma,state)
IN CONST ExprPtr typevar;
IN CONST ExprPtr sigma;
IN CONST ExprPtr state;
{
  struct occursp *data;
  ExprPtr val;

  sigma = fxRealSigma(sigma,state);
  if (typevar==sigma)
    return true;
  val = CDR(sigma);
  data = (struct occursp *)alloctmp(sizeof(struct occursp));
  data->typevar = typevar;
  data->state = state;
  return MapSweep(val,do_occursp_fx,data);
}

MAPSWEEPFUNC(do_occursp_fx)
{
  struct occursp *data;
  Boolean (*occp) P((ExprPtr tyvar,ExprPtr type,ExprPtr state));

  data = MAPDATA(struct occursp *);
  occp = (x->kind==(CHAR)E_ARRAY_2)?occurspFx1:occursp;
  return (*occp)(data->typevar,x,data->state);
}

PRIVATE Boolean occurspFx(typevar,f,state)
IN CONST ExprPtr typevar;
IN CONST ExprPtr f;
IN CONST ExprPtr state;
{
  if (f->kind==(CHAR)E_TYPEFX)
    return occurspFx1(typevar,CDR(f),state);
  return false;
}
#endif

PRIVATE Boolean occursp(typevar,type,state)
IN CONST ExprPtr typevar,type;
IN CONST ExprPtr state;
{
  ExprPtr val;
  Boolean found;

again:
  if (typevar==type)
    return true;
  /*if (type==NIL)
    return false;*/
  switch (type->kind)
    {
    case E_NIL:
      return false;
    case E_VAR:
      val = nongenValue(type,state,&found);
      if (found)
	{
	  type = val;
	  goto again;
	}
      return false;
    case E_TYPE_APPL:
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      if (occurspFx(typevar,APPLFUN(type),state))
	return true;
#endif
      for (type = APPLARG(type); type!=NIL; type=CDR(type))
	if (occursp(typevar,CAR(type),state))
	  return true;
      return false;
    case E_CONS:
      if (occursp(typevar,CAR(type),state))
	return true;
      type = CDR(type);
      goto again;
    case E_EMPTY:
      return false;
    case E_MAPLET:
      type = MAPLETY(type);
      goto again;
    case E_MAP:
      {
	struct occursp *data = (struct occursp *)alloctmp(sizeof(struct occursp));
	
	data->typevar = typevar;
	data->state = state;
	return MapSweep(type,do_occursp,data);
      }
    case E_NUMTYPE:
      (void) MapGet(typevar,REFVAL(type),&found);
      if (found)
	return true;
      {
	struct occursp *data = (struct occursp *)alloctmp(sizeof(struct occursp));
	
	data->typevar = typevar;
	data->state = state;
	return MapSweep(REFVAL(type),do_occursp_num,data);
      }
    default:
      kinderror(message(26),type->kind);
      break;
    }
  return false;
}

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
PRIVATE Boolean eqimpType P((int new_flags,ExprPtr eqtdg,
			     INT strength,INT maxstrength,
			     ExprPtr type,ExprPtr assum,ExprPtr state));
#else
PRIVATE Boolean eqimpType P((int new_flags,ExprPtr eqtdg,
			     ExprPtr type,ExprPtr assum,ExprPtr state));
#endif

MAPSWEEPFUNC(do_eqimptype)
{
  struct typeattr *data;

  data = MAPDATA(struct typeattr *);
  return eqimpType(data->new_flags,data->eqtdg,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		   data->strength,data->maxstrength,
#endif
		   y,data->assum,data->state);
}

MAPSWEEPFUNC(do_eqimptype_num)
{
  struct typeattr *data;

  data = MAPDATA(struct typeattr *);
  return eqimpType(data->new_flags,data->eqtdg,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		   data->strength,data->maxstrength,
#endif
		   x,data->assum,data->state);
}

PRIVATE Boolean eqimpType(new_flags,eqtdg, /* new_flags must not include F_EQ */
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			  strength,maxstrength,
#endif
			  type,assum,state)
IN int new_flags;
IN CONST ExprPtr eqtdg;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength,maxstrength;
#endif
IN CONST ExprPtr type;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr aux;
  Boolean failp;

again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return unif;*/
  switch (type->kind)
    {
      case E_NIL:
	return false;
      case E_VAR:
	{
	  ExprPtr val,t;
	  int flags;
	  Boolean found;

	  flags = varFlags(type) & ~F_EQ;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  if (genericp(type,assum))
	    {
	      if (!(IN_FLAGS(new_flags,flags) &&
		  (!IN_FLAGS(F_IMP,new_flags) ||
		   tyvarStrength(type,state)<=strength)))
		FAILUNIF;
	      t = eqTypeTdg(type,state,false);
	      if (eqtdg!=falseValue && t!=trueValue)
		{
		  aux = tdgInst(tdgAnd(tdgImply(eqtdg,t),S_EQCONDS(state)),S_NONEQS(state),false);
		  if (aux==falseValue)
		    FAILUNIF;
		  SETQ(S_EQCONDS(state),aux);
		}
	      return false;
	    }
	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      type = val;
	      goto again;
	    }
	  t = tdgOr(eqtdg,eqTypeTdg(type,state,false));
	  if (IN_FLAGS(new_flags,flags) &&
	      (!IN_FLAGS(F_IMP,new_flags) ||
	       tyvarStrength(type,state)<=strength))
	    {
	      setMapPut(type,t,&S_EQFLAGS(state));
	      if (maxstrength!=INF_STRENGTH)
		setMapPut(type,KINT(maxstrength),&S_MAXSTRENGTH(state));
	      return false;
	    }
	  setNongenValue(type,_freshInstfromFlags(flags | new_flags,
						  t,assum,state,strength,maxstrength),
			 state);
	  return false;
#else
	  if (genericp(type,assum))
	    {
	      if (!(IN_FLAGS(new_flags,flags)))
		FAILUNIF;
	      t = eqTypeTdg(type,state,false);
	      if (eqtdg!=falseValue && t!=trueValue)
		{
		  aux = tdgInst(tdgAnd(tdgImply(eqtdg,t),S_EQCONDS(state)),S_NONEQS(state),false);
		  if (aux==falseValue)
		    FAILUNIF;
		  SETQ(S_EQCONDS(state),aux);
		}
	      return false;
	    }
	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      type = val;
	      goto again;
	    }
	  t = tdgOr(eqtdg,eqTypeTdg(type,state,false));
	  if (IN_FLAGS(new_flags,flags))
	    {
	      setMapPut(type,t,&S_EQFLAGS(state));
	      return false;
	    }
	  setNongenValue(type,_freshInstfromFlags(flags | new_flags,
						  t,assum,state),
			 state);
	  return false;
#endif
	}
      case E_TYPE_APPL:
	{
	  ExprPtr f;

	  f = APPLFUN(type); /* correct because of correctTypeAppl() */
	  f = fxSym(f);
	  if (f==refSym || f==arraySym || f==tableSym)
	    {
	      eqtdg = falseValue; /* any ref type already admits equality,
					so we only have to impose the
					imperative attribute on subtypes */
	    }
	  else if (eqtdg!=falseValue)
	    { /* we must impose on arguments :
		- eqtdg itself if funEqTdg(f,state)=true and
		funNotRefTdg(f,state)=true
		- false (nothing) if funNotRefTdg(f,state)=false
		and it is impossible if funEqTdg(f,state)=false
		*/
	      aux = tdgInst(tdgAnd(funEqTdg(f,state),S_EQCONDS(state)),
			    S_NONEQS(state),false);
	      if (aux==falseValue) /* for example with -> */
		FAILUNIF;
	      SETQ(S_EQCONDS(state),aux);
	      eqtdg = tdgAnd(funNotRefTdg(f,state),eqtdg);
	    }
	  if (new_flags!=0 || eqtdg!=falseValue)
	    for (type=APPLARG(type); type!=NIL; type=CDR(type))
	      {
		failp = eqimpType(new_flags,eqtdg,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				  strength,maxstrength,
#endif
				  CAR(type),assum,state);
		if (failp)
		  FAILUNIF;
	      }
	  return false;
	}
      case E_CONS:
	failp = eqimpType(new_flags,eqtdg,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			  strength,maxstrength,
#endif
			  CAR(type),assum,state);
	if (failp)
	  FAILUNIF;
	type = CDR(type);
	goto again;
      case E_EMPTY:
	return false;
      case E_MAPLET:
	type = MAPLETY(type);
	goto again;
      case E_MAP:
	{
	  struct typeattr *data = (struct typeattr *)alloctmp(sizeof(struct typeattr));

	  data->new_flags = new_flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  data->strength = strength;
	  data->maxstrength = maxstrength;
#endif
	  data->assum = assum;
	  data->state = state;
	  data->eqtdg = eqtdg;
	  return MapSweep(type,do_eqimptype,data);
	}
      case E_NUMTYPE:
	if ((new_flags & ~(F_EQ|F_NUM))!=0)
	  {
	    struct typeattr *data = (struct typeattr *)alloctmp(sizeof(struct typeattr));

	    data->new_flags = new_flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	    data->strength = strength;
	    data->maxstrength = maxstrength;
#endif
	    data->assum = assum;
	    data->state = state;
	    data->eqtdg = eqtdg;
	    return MapSweep(REFVAL(type),do_eqimptype_num,data);
	  }
	return false;
      default:
	kinderror(message(27),type->kind);
	break;
    }
  FAILUNIF;
}

/**********
  Unification
  **********/

#define UNIF_OK 0
#define UNIF_DIF_FUNCS 1
#define UNIF_DIF_SORTS 2
#define UNIF_BIND_GEN 3
#define UNIF_OCCURS 4
#define UNIF_MISS_LBLS 5
#define UNIF_DIF_TAGS 6

typedef struct failureCause {
  COUNTER reason;
  ExprPtr s1,s2;
} failureCause;

PRIVATE failureCause unifyFailure;

PRIVATE void setFailure(reason,s1,s2)
IN COUNTER reason;
IN CONST ExprPtr s1,s2;
{
  unifyFailure.reason = reason;
  unifyFailure.s1 = KEEP(s1);
  unifyFailure.s2 = KEEP(s2);
}

PRIVATE Boolean unify_var(typevar,type,assum,state,higherp)
	/* where typevar may not
	be numeric, because '#a may only appear as {'#a => 1.0} */
IN CONST ExprPtr typevar,type;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean higherp;
{
  ExprPtr val,t1,t2,aux;
  int flags1,flags2;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  INT strength1,strength2,maxstrength1,maxstrength2;
#endif
  Boolean found;

again:
  if (typevar==type)
    return false;
  if (genericp(typevar,assum))
    {
      if (/*type!=NIL && */type->kind==(CHAR)E_VAR)
	{
	  if (genericp(type,assum))
	    {
	      unifyFailure.reason = UNIF_BIND_GEN;
	      unifyFailure.s1 = typevar;
	      unifyFailure.s2 = type;
	      FAILUNIF;
	    }
	  val = nongenValue(type,state,&found);
	  if (found)
	    {
	      if (/*val!=NIL && */val->kind==(CHAR)E_VAR && !genericp(val,assum))
		{
		  type = typevar;
		  typevar = val;
		  goto again;
		}
	      unifyFailure.reason = UNIF_BIND_GEN;
	      unifyFailure.s1 = typevar;
	      unifyFailure.s2 = type;
	      FAILUNIF;
	    }
	  flags1 = varFlags(typevar) & ~F_EQ;
	  flags2 = varFlags(type) & ~F_EQ;
	  t1 = eqTypeTdg(typevar,state,false);
	  t2 = eqTypeTdg(type,state,false);
	  if (IN_FLAGS(flags2,flags1)
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	      && (strength1 = tyvarStrength(typevar,state),
		  (!IN_FLAGS(F_IMP,flags2) ||
		   strength1<=tyvarStrength(type,state)))
#endif
	      )
	    {
	      aux = tdgAnd(tdgImply(t2,t1),S_EQCONDS(state));
	      if (aux==falseValue)
		{
		  unifyFailure.reason = UNIF_DIF_SORTS;
		  unifyFailure.s1 = typevar;
		  unifyFailure.s2 = type;
		  FAILUNIF;
		}
	      SETQ(S_EQCONDS(state),aux);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	      setMaxStrength(type,strength1,state);
#endif
	      setNongenValue(type,typevar,state);
	      return false;
	    }
	}
      unifyFailure.reason = UNIF_DIF_SORTS;
      unifyFailure.s1 = typevar;
      unifyFailure.s2 = type;
      FAILUNIF;
    }
  val = nongenValue(typevar,state,&found);
  if (found)
    return unify1(val,type,assum,state,higherp);
  if (/*type!=NIL && */type->kind==(CHAR)E_VAR && !genericp(type,assum))
    {
      val = nongenValue(type,state,&found);
      if (found)
	{
	  type = val;
	  goto again;
	}
      flags1 = varFlags(typevar) & ~F_EQ;
      flags2 = varFlags(type) & ~F_EQ;
      t1 = eqTypeTdg(typevar,state,false);
      t2 = eqTypeTdg(type,state,false);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
      strength1 = tyvarStrength(typevar,state);
      strength2 = tyvarStrength(type,state);
      maxstrength1 = tyvarMaxStrength(typevar,state);
      maxstrength2 = tyvarMaxStrength(type,state);
#endif
      if (IN_FLAGS(flags1,flags2) && tdgImply(t1,t2)==trueValue
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  && strength2<=strength1
#endif
	  )
	{
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  setMaxStrength(type,maxstrength1,state);
#endif
	  setNongenValue(typevar,type,state);
	  return false;
	}
      if (IN_FLAGS(flags2,flags1) && tdgImply(t2,t1)==trueValue
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  && strength1<=strength2
#endif
	  )
	{
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  setMaxStrength(typevar,maxstrength2,state);
#endif
	  setNongenValue(type,typevar,state);
	  return false;
	}
      /* if no inclusion between the flags,
	 or no weakening of strengths (in the STD_ML_NJ_TYVAR case)
	 create a new variable with all flags and relevant strength */
      val = _freshInstfromFlags(flags1|flags2,tdgOr(t1,t2),assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
				,(strength1<strength2)?strength1:strength2
				,(maxstrength1<maxstrength2)?maxstrength1:maxstrength2
#endif
				);
      setNongenValue(type,val,state);
      setNongenValue(typevar,val,state);
      return false;
    }
  if (occursp(typevar,type,state))
    {
      unifyFailure.reason = UNIF_OCCURS;
      unifyFailure.s1 = typevar;
      unifyFailure.s2 = type;
      FAILUNIF;
    }
  /* modify unifier so that type has all the attributes of typevar */
  if (eqimpType(varFlags(typevar) & ~F_EQ,
		eqTypeTdg(typevar,state,false),
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		tyvarStrength(typevar,state),
		tyvarMaxStrength(typevar,state),
#endif
		type,assum,state))
    {
      unifyFailure.reason = UNIF_DIF_SORTS;
      unifyFailure.s1 = typevar;
      unifyFailure.s2 = type;
      FAILUNIF;
    }
  setNongenValue(typevar,type,state);
  return false;
}

PRIVATE Boolean unify_list(list1,list2,assum,state,higherp)
IN CONST ExprPtr list1,list2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean higherp;
{
  while (ISCONS(list1))
    {
      if (unify1(CAR(list1),CAR(list2),assum,state,higherp))
	FAILUNIF;
      list1 = CDR(list1);
      list2 = CDR(list2);
    }
  return false;
}

MAPFOLDFUNC(do_instnumtype)
{
  struct instnumtype *data;
  ExprPtr val;
  Boolean found,fine;

  data = MAPDATA(struct instnumtype *);
  if (x->kind==(CHAR)E_VAR)
    {
      val = nongenValue(x,data->state,&found);
      if (!found)
	val = MapGet(x,data->subst,&found);
      if (found)
	{
	  if (/*val==NIL || */val->kind!=(CHAR)E_NUMTYPE)
	    {
	      data->fine = false;
	      val = NewMap();
	    }
	  else
	    {
	      val = instNumType(REFVAL(val),data->subst,data->state,&fine);
	      data->fine &= fine;
	      val = typePower1(val,y);
	    }
	}
      else
	val = MapSingleton(x,y);
    }
  else /* dimension (flexible or not) */
    {
      while (true)
	{
	  val = APPLFUN(x);
	  val = nongenValue(fxSym(val),data->state,&found);
	  if (!found)
	    val = MapGet(x,data->subst,&found);
	  if (found)
	    {
	      if (val->kind==(CHAR)E_CONS) /* val = (tyhdr,type), i.e, a type function */
		{
		  if (/*CDR(val)==NIL || */CDR(val)->kind!=(CHAR)E_NUMTYPE)
		    {
		      data->fine = false;
		      val = NewMap();
		    }
		  else
		    {
		      val = instNumType(REFVAL(CDR(val)),
					MapMerge(data->subst,
						 getSubst1(CAR(CAR(val)),APPLARG(x))),
					data->state,&fine);
		      data->fine &= fine;
		      val = typePower1(val,y);
		    }
		  break;
		}
	      else
		{
		  x = KtypeAppl(val,APPLARG(x));
		  (void) KEEP(x);
		}
	    }
	  else
	    {
	      val = MapSingleton(x,y);
	      break;
	    }
	}
    }
  return val;
}

PUBLIC MapPtr instNumType(map,subst,state,finep)
IN CONST MapPtr map;
IN CONST MapPtr subst;
IN CONST ExprPtr state;
OUT Boolean *finep;
{
  struct instnumtype *data;
  ENTER;

  data = (struct instnumtype *)alloctmp(sizeof(struct instnumtype));
  data->state = state;
  data->subst = subst;
  data->fine = true;
  map = MapFold(map,do_instnumtype,data,NewMap(),typeMul1);
  *finep = data->fine;
  KEEPONLY(map);
  return map;
}

MAPSWEEPFUNC(do_unify_num)
{
  struct unify_num *data;
  ExprPtr exp1,exp2,exp;
  Boolean found;

  data = MAPDATA(struct unify_num *);
  exp1 = MapGet(x,data->map1,&found);
  if (!found)
    exp1 = zzero;
  exp2 = MapGet(x,data->map2,&found);
  if (found)
    {
      exp = z_sub(exp1,exp2);
      (void) KEEP(exp);
    }
  else exp = exp1;
  if (exp!=zzero)
    {
      if (x->kind==(CHAR)E_VAR && !genericp(x,data->assum))
	{
	  data->varmap = MapPut(x,exp,data->varmap);
	  if (data->strongest==NIL)
	    data->strongest = x;
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
	  else if (varFlags(x) & F_IMP)
	    {
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	      if ((varFlags(data->strongest) & F_IMP)
		  && tyvarStrength(data->strongest,data->state)<tyvarStrength(x,data->state)
		  )
		data->strongest = x;
#endif
	    }
#endif
	  else
	    data->strongest = x;
	}
      else /* dimension */
	data->quanmap = MapPut(x,exp,data->quanmap);
    }
  return false;
}

PRIVATE ExprPtr mininv(y)
IN CONST ExprPtr y;
{
  y = z_div(mone,y);
  (void) KEEP(y);
  return y;
}

PRIVATE Boolean unify_num(map1,map2,assum,state,higherp)
IN CONST MapPtr map1,map2;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
IN Boolean higherp;
{
  struct unify_num *data;
  ExprPtr x,y,type;
  MapPtr dom;
  Boolean fine;
  Boolean failp;
  ENTER;

  map1 = instNumType(map1,NewMap(),state,&fine); /* fine should be true */
  map2 = instNumType(map2,NewMap(),state,&fine); /* fine should be true */
  dom = SetUnion(map1,map2);
  data = (struct unify_num *)alloctmp_anchored(sizeof(struct unify_num), (char *)&data);
  data->varmap = NewMap();
  data->quanmap = NewMap();
  data->assum = assum;
  data->state = state;
  data->map1 = map1;
  data->map2 = map2;
  data->strongest = NIL;
  data->higherp = higherp;
  (void) MapSweep(dom,do_unify_num,data);
  if (data->varmap==NewMap())
    {
      if (data->quanmap!=NewMap())
	{
#if MODULE_SYSTEM & STD_ML_MODULE
	  struct find_flexappl_num *data2 = (struct find_flexappl_num *)
	    alloctmp_anchored(sizeof(struct find_flexappl_num), (char *)&data);
	  ExprPtr f;

	  if (higherp)
	    {
	      data2->assum = assum;
	      data2->state = state;
	      data2->realflex = NIL;
	      data2->altflex = NIL;
	      data2->altflexp = false;
	      if (MapSweep(data->quanmap,do_find_flexappl_num,data2))
		{
		  x = data2->realflex;
		  y = MapGetSure(x,data->varmap);
		  type = MapRemove(x,data->quanmap);
		  y = mininv(y);
		  type = KnumType(typePower1(type,y));
		  (void) KEEP(type);
		  f = APPLFUN(x);
		  failp = setTyFcn(fxSym(f),
				   KONS(getTyHdr(APPLARG(x),NIL,state),type),
				   assum,state);
		}
	      else if (data2->altflexp) /* there is still something to do */
		{
		  x = data2->altflex;
		  y = MapGetSure(x,data->varmap);
		  type = MapRemove(x,data->quanmap);
		  y = mininv(y);
		  type = KnumType(typePower1(type,y));
		  failp = recoverUnify2(x,KEEP(type),assum,state);
		}
	      else
		failp = true; /* really everything is rigid */
	    }
	  else
#endif
	    failp = true;
	}
      else failp = false;
    }
  else
    {
      x = data->strongest;
	/* if there exists one non-imperative type variable,
		it should be chosen; otherwise, we take the
		strongest one */
      y = MapGetSure(x,data->varmap);
      type = MapMerge(MapRemove(x,data->varmap),data->quanmap);
	/* x^y.type=1, so x=type^(-1/y) */
      y = mininv(y);
      type = KnumType(typePower1(type,y));
      setNongenValue(x,KEEP(type),state);
      failp = false;
    }
  QUIT;
  return failp;
}

PRIVATE MapPtr get_real_map_type(type,assum,state,restp,genp)
IN CONST MapPtr type;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
OUT ExprPtr *restp; /* returns rest variables
		       (not {'#a => 1.0} in the case of '#a) or NIL */
IN Boolean genp;
{
  ExprPtr rest,stk;
  Boolean found;

  stk = NIL;
again:
  switch (type->kind)
    {
    case E_EMPTY:
    case E_MAPLET:
    case E_MAP:
      rest = MapGet(REST_TAG,type,&found);
      if (found)
	{
	  stk = Cons(MapRemove(REST_TAG,type),stk);
	  type = rest;
	  goto again;
	}
      *restp = NIL;
      break;
    case E_NUMTYPE:
      type = ChooseInDom(REFVAL(type));
      goto again;
    case E_VAR:
      if (genp && genericp(type,assum))
	{
	  *restp = NIL;
	  type = MapSingleton(REST_TAG,type);
	  break;
	}
      rest = nongenValue(type,state,&found);
      if (found)
	{
	  type = rest;
	  goto again;
	}
      *restp = type;
      type = NewMap();
      break;
    default:
      kinderror(message(76),type->kind);
    }
  for (; stk!=NIL; stk=CDR(stk))
    type = MapMerge(CAR(stk),type);
  return type;
}

PUBLIC MapPtr overrideMapType(type1,type2,assum,state)
IN CONST MapPtr type1,type2;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  if (type1->kind==(CHAR)E_VAR)
    type1 = _nongenValue(type1,state);
  switch (type1->kind)
    {
    case E_VAR:
      if (genericp(type1,assum))
	{
	  type1 = NIL; /* error: type1 cannot be instantiated to be a record type */
	  break;
	}
      else
	{
	  ExprPtr tyvar;

	  tyvar = type1;
	  type1 = MapSingleton(REST_TAG,freshInstfromFlags(0,falseValue,assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
							   ,INF_STRENGTH,INF_STRENGTH
#endif
							   ));
	  setNongenValue(tyvar,type1,state);
	}
      /*fallthrough*/
    case E_EMPTY:
    case E_MAPLET:
    case E_MAP:
      type1 = MapMerge(type1,type2);
      break;
    default:
      type1 = NIL; /* error: type1 is not a record type */
      break;
    }
  return type1;
}

PRIVATE MapPtr addMapRest(rest,maptype)
IN CONST ExprPtr rest;
IN CONST ExprPtr maptype;
{
  return (maptype==EmptyMap())?rest:MapPut(REST_TAG,rest,maptype);
}

MAPSWEEPFUNC(do_unify_map)
{
  struct unify_map *data;

  data = MAPDATA(struct unify_map *);
  return unify1(MapGetSure(x,data->type1),
		MapGetSure(x,data->type2),
		data->assum,data->state,data->higherp);
}

PRIVATE Boolean unify_map(type1,type2,assum,state,higherp)
IN CONST MapPtr type1,type2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean higherp;
{
  ExprPtr rest1,rest2,ptype1,ptype2,dom,rest,type;
  int flags1,flags2;
  ExprPtr t1,t2;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  INT strength1,strength2,maxstrength1,maxstrength2;
#endif
  struct unify_map *data;

  ptype1 = get_real_map_type(type1,assum,state,&rest1,true);
  ptype2 = get_real_map_type(type2,assum,state,&rest2,true);
  if (rest1==NIL)
    if (!SetIncludedp(ptype2,ptype1))
      {
err1:
	setFailure(UNIF_MISS_LBLS,type2,type1);
	FAILUNIF;
      }
  if (rest2==NIL)
    if (!SetIncludedp(ptype1,ptype2))
      {
err2:
	setFailure(UNIF_MISS_LBLS,type1,type2);
	FAILUNIF;
      }
  data = (struct unify_map *)alloctmp(sizeof(struct unify_map));
  data->type1 = ptype1;
  data->type2 = ptype2;
  if (rest1==NIL)
    {
      dom = ptype2;
      type = MapDomRestrBy(ptype1,ptype2);
      if (rest2==NIL)
	{
	  if (type!=NewMap())
	    goto err2;
	}
      else
	{
	  if (eqimpType(varFlags(rest2) & ~F_EQ,
			eqTypeTdg(rest2,state,false),
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			tyvarStrength(rest2,state),
			tyvarMaxStrength(rest2,state),
#endif
			type,assum,state))
	    {
errsorts:
	      setFailure(UNIF_DIF_SORTS,rest2,type1);
	      FAILUNIF;
	    }
	  if (occursp(rest2,type,state))
	    {
errocc1:
	      setFailure(UNIF_OCCURS,rest2,type1);
	      FAILUNIF;
	    }
	  setNongenValue(rest2,type,state);
	}
    }
  else if (rest2==NIL)
	{
	  dom = ptype1;
	  type = MapDomRestrBy(ptype2,ptype1);
	  if (eqimpType(varFlags(rest1) & ~F_EQ,
			eqTypeTdg(rest1,state,false),
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			tyvarStrength(rest1,state),
			tyvarMaxStrength(rest1,state),
#endif
			type,assum,state))
	    goto errsorts;
	  if (occursp(rest1,type,state))
	    {
errocc2:
	      setFailure(UNIF_OCCURS,rest1,type2);
	      FAILUNIF;
	    }
	  setNongenValue(rest1,type,state);
	}
  else
    { /* p1 + 'a1 unifies with p2 + 'a2 (in the case p1 inter p2={})
	    if 'a1 => p2 + 'c, 'a2 => p1 + 'c */
      dom = SetInter(ptype1,ptype2);
      ptype1 = MapDomRestrBy(ptype1,dom);
      ptype2 = MapDomRestrBy(ptype2,dom);
      if (rest1==rest2)
	{
	  if (ptype1!=NewMap()) /* no hope of ever unifying them */
	    goto err1;
	  if (ptype2!=NewMap()) /* no hope of ever unifying them */
	    goto err2;
	}
      else
	{
	  flags1 = varFlags(rest1) & ~F_EQ;
	  flags2 = varFlags(rest2) & ~F_EQ;
	  t1 = eqTypeTdg(rest1,state,false);
	  t2 = eqTypeTdg(rest2,state,false);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  strength1 = tyvarStrength(rest1,state);
	  strength2 = tyvarStrength(rest2,state);
	  maxstrength1 = tyvarMaxStrength(rest1,state);
	  maxstrength2 = tyvarMaxStrength(rest2,state);
#endif
	  if (ptype1==NewMap()) /* particular cases, to gain speed and space */
	    {
	      ptype2 = addMapRest(rest2,ptype2);
	      if (eqimpType(flags1,t1,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    strength1,maxstrength1,
#endif
			    ptype2,assum,state))
		goto errsorts;
	      if (occursp(rest1,ptype2,state))
		goto errocc2;
	      setNongenValue(rest1,ptype2,state);
	    }
	  else if (ptype2==NewMap()) /* particular cases, to gain speed and space */
	    {
	      ptype1 = addMapRest(rest1,ptype1);
	      if (eqimpType(flags2,t2,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    strength2,maxstrength2,
#endif
			    ptype1,assum,state))
		goto errsorts;
	      if (occursp(rest2,ptype1,state))
		goto errocc1;
	      setNongenValue(rest2,ptype1,state);
	    }
	  else
	    {
	      rest = _freshInstfromFlags(flags1|flags2,tdgOr(t1,t2),assum,state
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
					,(strength1<strength2)?strength1:strength2
					,(maxstrength1<maxstrength2)?maxstrength1:maxstrength2
#endif
					);
	      ptype1 = addMapRest(rest,ptype1);
	      ptype2 = addMapRest(rest,ptype2);
	      if (eqimpType(flags2,t2,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    strength2,maxstrength2,
#endif
			    ptype1,assum,state))
		goto errsorts;
	      if (eqimpType(flags1,t1,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    strength1,maxstrength1,
#endif
			    ptype2,assum,state))
		goto errsorts;
	      if (occursp(rest2,ptype1,state))
		goto errocc1;
	      setNongenValue(rest2,ptype1,state);
	      if (occursp(rest1,ptype2,state))
		goto errocc2;
	      setNongenValue(rest1,ptype2,state);
	    }
	}
    }
  data->assum = assum;
  data->state = state;
  data->higherp = higherp;
  return MapSweep(dom,do_unify_map,data);
}

PUBLIC ExprPtr correctTypeAppl(type,state)
IN CONST ExprPtr type;
IN CONST ExprPtr state;
{
  ExprPtr f,s;

  while (/*type!=NIL && */type->kind==(CHAR)E_TYPE_APPL)
    {
      f = applFun(type,state);
      if (f->kind==(CHAR)E_CONS) /* then f is a tyfcn (tyhdr,tyexpr) */
	{
	  s = getSubst1(CAR(CAR(f)),APPLARG(type));
	  /* no need to check that type arguments have the correct
	     attributes, because they will */
	  type = newTypeInst1(CDR(f),NIL,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			      false,
#endif
			      &s);
	}
      else
	{
	  if (f==APPLFUN(type))
	    break;
	  type = KtypeAppl(f,APPLARG(type));
	  (void) KEEP(type);
	}
    }
  return type;
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE Boolean unify_fx(f1,f2,assum,state)
IN CONST ExprPtr f1,f2;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr sigma1,sigma2;
  MapPtr constr;
  ExprPtr val;
  Boolean found;

  if (f1->kind==(CHAR)E_TYPEFX)
    {
      sigma1 = fxRealSigma(CDR(f1),state);
      if (f2->kind==(CHAR)E_TYPEFX)
	{
	  sigma2 = fxRealSigma(CDR(f2),state);
	  if (genericp(sigma1,assum) || genericp(sigma2,assum))
	    return false;
		/* incorrect (happens in the case of TypeMatch, which
	    	is used when unpacking dynamics; the use of top
	    	effects should correct this; also happens when
	    	type casting, but there, we don't want to consider
	    	the latent effects in the casting type) */
	  setNongenValue(sigma1,sigma2,state);
	  constr = SetUnion(CDR(sigma1),CDR(sigma2));
	  SETCDR(sigma1,constr);
	  SETCDR(sigma2,constr);
	}
    }
  return false;
}
#endif

PUBLIC Boolean unify1(type1,type2,assum,state,higherp)
	/* if higherp=true, this modifies fields in the state:
		S_TYVARSUBS, S_NONEQS,
		S_CONDS, S_EQCONDS and S_EQFLAGS (see setTyFcn() in module.c) */
IN CONST ExprPtr type1,type2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean higherp;
{
again:
  if (type1==type2)
    return false;
  
  if (/*type1!=NIL && */type1->kind==(CHAR)E_VAR && !genericp(type1,assum))
    return unify_var(type1,type2,assum,state,higherp);
  if (/*type2!=NIL && */type2->kind==(CHAR)E_VAR && !genericp(type2,assum))
    return unify_var(type2,type1,assum,state,higherp);
  /*if (type1==NIL || type2==NIL)
    goto fail;*/
  type1 = correctTypeAppl(type1,state);
  type2 = correctTypeAppl(type2,state);
  if (type1==type2)
    return false;
#if MODULE_SYSTEM & STD_ML_MODULE
  if (higherp)
    {
      if (isFlexAppl(type1,assum))
	return setTyFcn(APPLFUN(type1),KONS(getTyHdr(APPLARG(type1),NIL,state),type2),
			assum,state);
      if (isFlexAppl(type2,assum))
	return setTyFcn(APPLFUN(type2),KONS(getTyHdr(APPLARG(type2),NIL,state),type1),
			assum,state);
	}
#endif
  if (ISMAP(type1) && ISMAP(type2))
    return unify_map(type1,type2,assum,state,higherp);
  if (type1->kind!=type2->kind)
    {
      setFailure(UNIF_DIF_TAGS,type1,type2);
      goto fail;
    }
  switch (type1->kind)
    {
      case E_NIL:
	return false;
      case E_VAR: /* generic variables: not unifiable */
	setFailure(UNIF_BIND_GEN,type1,type2);
	FAILUNIF;
      case E_TYPE_APPL:
	if (fxSym(APPLFUN(type1))!=fxSym(APPLFUN(type2))) /* correct because of the correctTypeAppl() */
	  {
	    setFailure(UNIF_DIF_FUNCS,fxSym(APPLFUN(type1)),fxSym(APPLFUN(type2)));
	    goto fail;
	  }
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	if (unify_fx(APPLFUN(type1),APPLFUN(type2),assum,state))
	  FAILUNIF;
#endif
	return unify_list(APPLARG(type1),APPLARG(type2),assum,state,higherp);
      case E_CONS:
	if (unify1(CAR(type1),CAR(type2),assum,state,higherp))
	    FAILUNIF;
	type1 = CDR(type1);
	type2 = CDR(type2);
	goto again;
      case E_NUMTYPE:
	return unify_num(REFVAL(type1),REFVAL(type2),assum,state,higherp);
      default:
	kinderror(message(28),type1->kind);
	break;
    }
fail:
#if MODULE_SYSTEM & STD_ML_MODULE
  if (higherp)
    {
      if (/*type1!=NIL && */type1->kind==(CHAR)E_TYPE_APPL &&
	  !isRigid(type1,assum))
	{
	  return recoverUnify2(type1,type2,assum,state);
	}
      if (/*type2!=NIL && */type2->kind==(CHAR)E_TYPE_APPL &&
	  !isRigid(type2,assum))
	{
	  return recoverUnify2(type2,type1,assum,state);
	}
    }
#endif
  FAILUNIF;
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
MAPSWEEPFUNC(do_cyclicfx)
{
  struct cyclicfx *data;
  Boolean found;

  data = MAPDATA(struct cyclicfx *);
  if (/*y!=NIL && */y->kind==(CHAR)E_ARRAY_2 &&
      ((void) nongenValue(y,data->state,&found), !found))
    return occurspFx1(y,CDR(y),data->state);
  return false;
}

PRIVATE Boolean cyclicFx(unif,assum,state)
IN CONST MapPtr unif;
IN CONST ExprPtr assum,state;
{
  struct cyclicfx *data = (struct cyclicfx *)alloctmp(sizeof(struct cyclicfx));

  data->assum = assum;
  data->state = state;
  return MapSweep(unif,do_cyclicfx,data);
}
#endif

PRIVATE Boolean unify(type1,type2,assum,state,higherp)
IN CONST ExprPtr type1,type2;
IN CONST ExprPtr assum,state;
IN Boolean higherp;
{
  return (unify1(type1,type2,assum,state,higherp)
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	  || cyclicFx(MapDomRestrBy(S_TYVARSUBCACHE(state),oldtyvarsubcache),
		      /* !!! incorrect (we try to isolate the new found unifier,
			 but it is not easy) */
		      assum,state)
#endif
	  );
}

PUBLIC Boolean unify2(type1,type2,assum,state)
IN CONST ExprPtr type1,type2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  return (unify1(type1,type2,assum,state,true)
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	  || cyclicFx(MapDomRestrBy(S_TYVARSUBCACHE(state),oldtyvarsubcache),
		      /* !!! incorrect (we try to isolate the new found unifier,
			 but it is not easy) */
		      assum,state)
#endif
	  );
}

PRIVATE void explainDifFunc(f1,f2)
IN CONST ExprPtr f1,f2;
{
  INT len1,len2;
  CHAR *s1,*s2;
  CHAR c;

  /* first, check whether s1 and s2 have similar names. */
  len1 = symNameLen(f1);
  s1 = symName(f1);
  while (c = *s1++, --len1>0 && c!='%');
  len2 = symNameLen(f2);
  s2 = symName(f2);
  while (c = *s2++, --len2>0 && c!='%');
  if (len1==len2 && memcmp(s1,s2,len1)==0) /* then there is a possibility of confusion */
    {
      hfputs(message(412),hstderr);
      do_printfile(s1,len1,(char *)hstderr);
      hfputs(message(758),hstderr);
    }
  else
    {
      hfputs(message(568),hstderr);
      do_printfile(s1,len1,(char *)hstderr);
      hfputs(message(569),hstderr);
      do_printfile(s2,len2,(char *)hstderr);
      hfputs(message(570),hstderr);
    }
}

PRIVATE void explainDifSorts(type1,type2,state)
IN CONST ExprPtr type1,type2;
INOUT ExprPtr state;
{
  type1 = developType(type1,state);
  type2 = developType(type2,state);
  hfputs(message(568),hstderr);
  (void) TypePrint(type1,do_printfile,(char *)hstderr,state);
  hfputs(message(569),hstderr);
  (void) TypePrint(type2,do_printfile,(char *)hstderr,state);
  hfputs(message(571),hstderr);
}

PRIVATE void explainBindGen(typevar,state)
IN CONST ExprPtr typevar;
INOUT ExprPtr state;
{
  hfputs(message(572),hstderr);
  (void) TypePrint(typevar,do_printfile,(char *)hstderr,state);
  hfputs(message(573),hstderr);
}

PRIVATE void explainOccurs(type1,type2,state)
IN CONST ExprPtr type1,type2;
INOUT ExprPtr state;
{
  type2 = developType(type2,state);
  hfputs(message(574),hstderr);
  (void) TypePrint(type1,do_printfile,(char *)hstderr,state);
  hfputs(message(575),hstderr);
  (void) TypePrint(type2,do_printfile,(char *)hstderr,state);
  hfputs(message(577),hstderr);
}

MAPSWEEPFUNC(do_explainmisslbls)
{
  Boolean *commap = MAPDATA(Boolean *);

  if (*commap)
    hfputc(',',hstderr);
  *commap = true;
  hfputc(' ',hstderr);
  valPrintNoType(x,do_printfile,(char *)hstderr,0);
  return false;
}

PRIVATE void explainMissLbls(type1,type2,state)
IN CONST ExprPtr type1,type2;
INOUT ExprPtr state;
{
  ExprPtr lbls1,lbls2,delta;
  Boolean comma;

  type1 = developType(type1,state);
  lbls1 = Dom(MapRemove(REST_TAG,type1));
  type2 = developType(type2,state);
  lbls2 = Dom(MapRemove(REST_TAG,type2));
  delta = MapDelta(lbls1,lbls2);
  hfputs(message(590),hstderr);
  if (delta->kind!=(CHAR)E_MAPLET)
    hfputc('s',hstderr);
  comma = false;
  (void) MapSweep(delta,do_explainmisslbls,(char *)&comma);
  hfputs(message((delta->kind==(CHAR)E_MAPLET)?670:669),hstderr);
  hfputs(message(591),hstderr);
}

PRIVATE void explainDifTags(type1,type2,state)
IN CONST ExprPtr type1,type2;
INOUT ExprPtr state;
{
  type1 = developType(type1,state);
  type2 = developType(type2,state);
  hfputs(message(568),hstderr);
  (void) TypePrint(type1,do_printfile,(char *)hstderr,state);
  hfputs(message(569),hstderr);
  (void) TypePrint(type2,do_printfile,(char *)hstderr,state);
  hfputs(message(570),hstderr);
}

PRIVATE void explainDif(assum,state)
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  switch (unifyFailure.reason)
    {
    case UNIF_DIF_FUNCS:
      explainDifFunc(unifyFailure.s1,unifyFailure.s2);
      break;
    case UNIF_DIF_SORTS:
      explainDifSorts(unifyFailure.s1,unifyFailure.s2,state);
      break;
    case UNIF_BIND_GEN:
      explainBindGen(unifyFailure.s1,state);
      break;
    case UNIF_OCCURS:
      explainOccurs(unifyFailure.s1,unifyFailure.s2,state);
      break;
    case UNIF_MISS_LBLS:
      explainMissLbls(unifyFailure.s1,unifyFailure.s2,state);
      break;
    case UNIF_DIF_TAGS:
      if (unifyFailure.s1->kind==(CHAR)E_VAR && genericp(unifyFailure.s1,assum))
	explainBindGen(unifyFailure.s1,state);
      else if (unifyFailure.s2->kind==(CHAR)E_VAR && genericp(unifyFailure.s2,assum))
	explainBindGen(unifyFailure.s2,state);
      else explainDifTags(unifyFailure.s1,unifyFailure.s2,state);
      break;
    }
  hfputc('\n',hstderr);
  hfflush(hstderr);
}

PUBLIC Boolean coerceTypes(zone,type1,type2,assum,state,higherp)
IN CONST TZone *zone;
IN CONST ExprPtr type1,type2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN Boolean higherp;
{
  Boolean fail;
  ENTER;

  unifyFailure.reason = UNIF_OK;
  fail = false;
  if (
#if MODULE_SYSTEM & STD_ML_MODULE
      higherp?unify2(type1,type2,assum,state):
#endif
      unify(type1,type2,assum,state,false))
    {
      incompatibleTypes(zone,type1,type2,
			(CHAR *)message(29),(CHAR *)message(30),
			assum,state);
      explainDif(assum,state);
      fail = true;
    }
  QUIT;
  return fail;
}

PUBLIC Boolean coerceTypesImport(zone,type1,type2,assum,state,var,higherp)
IN CONST TZone *zone;
IN CONST ExprPtr type1,type2;
INOUT ExprPtr assum;
INOUT ExprPtr state;
IN CONST ExprPtr var;
IN Boolean higherp;
{
  Boolean fail;
  ENTER;

  unifyFailure.reason = UNIF_OK;
  fail = false;
  if (
#if MODULE_SYSTEM & STD_ML_MODULE
      higherp?unify2(type1,type2,assum,state):
#endif
      unify(type1,type2,assum,state,false))
    {
      INT len = SLEN(var);
      CHAR *name = alloctmp_anchored(len+1, (char *)&zone);
      CHAR *s = alloctmp_anchored(80+len, (char *)&zone);

      safememcpy(name, STEXT(var), len);
      name[len] = 0;
      sprintf((char *)s, message(381), name);
      incompatibleTypes(zone,type2,type1,s,(CHAR *)message(382),assum,state);
      explainDif(assum,state);
      fail = true;
    }
  QUIT;
  return fail;
}

PRIVATE ExprPtr freshTyVar(vars,cp,flags)
IN CONST MapPtr vars;
INOUT COUNTER *cp;
IN int flags;
{
  ExprPtr x;
  Boolean found;
  CHAR buf[16];
  COUNTER n;
  ENTER;

  buf[0] = '\'';
  n = 1;
  if ((flags & F_EQ) & !(flags & F_NUM))
    buf[n++] = '\'';
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
  if (flags & F_IMP)
    buf[n++] = '_';
#endif
  if (flags & F_NUM)
    buf[n++] = '#';
  while (getNumber(++*cp,buf+n), x = Ksymbol(buf),
	 /* don't KEEP x, since MapGet does not allocate */
	 (void) MapGet(x,vars,&found), found);
  KEEPONLY(x);
  return x;

}

PRIVATE void getTyHdrSubsts1(tyvars1,tyvars2,eqflags1,eqflags2,s1p,s2p,vars,cp,state)
IN CONST ExprPtr tyvars1,tyvars2; /* with length(tyvars)=length(tyvars2) */
IN CONST MapPtr eqflags1,eqflags2;
INOUT MapPtr *s1p,*s2p;
IN CONST MapPtr vars;
INOUT COUNTER *cp;
INOUT ExprPtr state;
{
  int flags1,flags2;
  ExprPtr eqtdg1,eqtdg2;
  ExprPtr x1,x2,x,aux;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  INT strength1,strength2,strength;
#endif

  tyvars1 = Nreverse(tyvars1);
  tyvars2 = Nreverse(tyvars2);
  for (; tyvars1!=NIL; tyvars1=CDR(tyvars1), tyvars2=CDR(tyvars2))
    {
      x1 = CAR(tyvars1);
      x2 = CAR(tyvars2);
      flags1 = varFlags(x1) & ~F_EQ;
      flags2 = varFlags(x2) & ~F_EQ;
      eqtdg1 = tdgInst(MapGetSure(x1,eqflags1),S_NONEQS(state),false);
      eqtdg2 = tdgInst(MapGetSure(x2,eqflags2),S_NONEQS(state),false);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
      strength1 = tyvarStrength(x1,state);
      strength2 = tyvarStrength(x2,state);
#endif
      if (IN_FLAGS(flags1,flags2) &&
	  tdgImply(eqtdg1,eqtdg2)==trueValue
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  && strength1>=strength2
#endif
	  )
	{
	  *s1p = MapPut(x1,x2,*s1p);
	}
      else if (IN_FLAGS(flags2,flags1) &&
	       tdgImply(eqtdg2,eqtdg1)==trueValue
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	       && strength2>=strength1
#endif
	       )
	{
	  *s2p = MapPut(x2,x1,*s2p);
	}
      else
	{
	  x = freshTyVar(vars,cp,flags1|flags2);
	  *s1p = MapPut(x1,x,*s1p);
	  *s2p = MapPut(x2,x,*s2p);
	  aux = MapPut(x,tdgOr(eqtdg1,eqtdg2),S_EQFLAGS(state));
	  SETQ(S_EQFLAGS(state),aux);
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  strength = (strength1<strength2)?strength1:strength2;
	  if (strength!=1)
	    {
	      setMapPut(x,KINT(strength),&S_STRENGTH(state));
	    }
#endif
	}
    }
}

PUBLIC void getTyHdrSubsts(tyhdr1,tyhdr2,s1p,s2p,state)
IN CONST ExprPtr tyhdr1,tyhdr2; /* with length(CAR(tyhdr1))=length(CAR(tyhdr2)) */
OUT MapPtr *s1p,*s2p;
INOUT ExprPtr state;
{
  MapPtr vars;
  ExprPtr l,x;
  COUNTER c;
  ENTER;

  *s1p = *s2p = NewMap();
  c = 0;
  vars = NewMap();
  for (l=CAR(tyhdr1); l!=NIL; l=CDR(l))
    {
      x = CAR(l);
      vars = SetAdd(x,vars);
    }
  for (l=CAR(tyhdr2); l!=NIL; l=CDR(l))
    {
      x = CAR(l);
      vars = SetAdd(x,vars);
    }
  getTyHdrSubsts1(CAR(tyhdr1),CAR(tyhdr2),CDR(tyhdr1),CDR(tyhdr2),s1p,s2p,vars,&c,state);
  KEEPONLY(Kons(*s1p,*s2p));
}

PUBLIC void coerceFlagTypes(zone,flags,eqtdg,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			    strength,maxstrength,
#endif
			    type,assum,state)
     /* flags should not contain F_NUM, neither F_EQ (use eqtdg=trueValue) */
IN CONST TZone *zone;
IN int flags;
IN CONST ExprPtr eqtdg;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength,maxstrength;
#endif
IN CONST ExprPtr type;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ENTER;

  if (eqimpType(flags,eqtdg,
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		strength,maxstrength,
#endif
		type,assum,state)
      )
  /* instead we find: */
    typeError(zone,(CHAR *)((eqtdg!=falseValue)?message(34):message(37)),
	      type,state);
  QUIT;
}

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
PRIVATE void Weaken1 P((ExprPtr type,ExprPtr assum,ExprPtr state,INT delta,MapPtr *donep));

PRIVATE void tyvarWeaken(typevar,state,delta,donep)
IN CONST SymPtr typevar;
INOUT ExprPtr state;
IN INT delta;
INOUT MapPtr *donep;
{
  INT strength,maxstrength;
  Boolean found;
  ENTER;

  (void) MapGet(typevar,*donep,&found);
  if (!found)
    {
      if (varFlags(typevar) & F_IMP)
	{
	  strength = tyvarStrength(typevar,state);
	  strength -= delta;
	  if (strength<0)
	    strength = 0;
	  if (delta<0)
	    {
	      maxstrength = tyvarMaxStrength(typevar,state);
	      if (strength>maxstrength)
		strength = maxstrength;
	    }
	  setMapPut(typevar,KINT(strength),&S_STRENGTH(state));
	}
      QUIT;
      *donep = SetAdd(typevar,*donep);
    }
}

MAPSWEEPFUNC(do_weaken)
{
  struct weaken *data;

  data = MAPDATA(struct weaken *);
  Weaken1(y,data->assum,data->state,data->delta,data->donep);
  return false;
}

MAPSWEEPFUNC(do_weaken_num)
{
  struct weaken *data;

  data = MAPDATA(struct weaken *);
  Weaken1(x,data->assum,data->state,data->delta,data->donep);
  return false;
}

PRIVATE void Weaken1(type,assum,state,delta,donep)
IN CONST ExprPtr type;
INOUT ExprPtr assum,state;
IN INT delta;
INOUT MapPtr *donep;
{
again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return;*/
  switch (type->kind)
    {
      case E_NIL:
	return;
      case E_VAR:
	if (genericp(type,assum))
	  tyvarWeaken(type,state,delta,donep);
	else
	  {
	    ExprPtr val;
	    Boolean found;

	    val = nongenValue(type,state,&found);
	    if (found)
	      {
		type = val;
		goto again;
	      }
	    else
	      tyvarWeaken(type,state,delta,donep);
	  }
	break;
      case E_TYPE_APPL:
	type = APPLARG(type);
	goto again;
      case E_CONS:
	Weaken1(CAR(type),assum,state,delta,donep);
	type = CDR(type);
	goto again;
      case E_EMPTY:
	break;
      case E_MAPLET:
	type = MAPLETY(type);
	goto again;
      case E_MAP:
	{
	  struct weaken *data = (struct weaken *)alloctmp(sizeof(struct weaken));

	  data->assum = assum;
	  data->state = state;
	  data->delta = delta;
	  data->donep = donep;
	  (void) MapSweep(type,do_weaken,data);
	  break;
	}
      case E_NUMTYPE:
	{
	  struct weaken *data = (struct weaken *)alloctmp(sizeof(struct weaken));

	  data->assum = assum;
	  data->state = state;
	  data->delta = delta;
	  data->donep = donep;
	  (void) MapSweep(REFVAL(type),do_weaken_num,data);
	  break;
	}
      default:
	kinderror(message(38),type->kind);
	break;
    }
}

PUBLIC ExprPtr Weaken(type,assum,state,delta)
IN CONST ExprPtr type;
INOUT ExprPtr assum,state;
IN INT delta;
{
  ExprPtr done;
  ENTER;

  done = NewMap();
  (void) Weaken1(type,assum,state,delta,&done);
  QUIT;
  return type;
}
#endif

PUBLIC MapPtr getTypeSubst(zone,tyhdr,args,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr tyhdr;
IN CONST ExprPtr args;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr subst,tyvar,e,type,tyvarseq,eqtdg1,eqtdg2,eqflags,aux;
  TZone *z;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
#ifdef WEAKNESS_MATTERS
  INT strength;
#endif
#endif
  INT n;
  ENTER;

  tyvarseq = CAR(tyhdr);
  eqflags = CDR(tyhdr);
  if ((n = ListLength(tyvarseq))!=ListLength(args))
    {
      aux = Ksymbol(message((n==0)?243:40));
      aux = KtypeAppl(KEEP(aux),tyvarseq);
      typeError(zone,(CHAR *)message(39),KEEP(aux),state);
      QUIT;
      return NewMap();
    }
  subst = NewMap();
  for (; args!=NIL; args=CDR(args), tyvarseq=CDR(tyvarseq))
    {
      tyvar = CAR(tyvarseq);
      if (tyvar->kind==(CHAR)E_NUMTYPE) /* '#a is coded as {'#a => 1.0} */
	tyvar = ChooseInDom(REFVAL(tyvar));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
#ifdef WEAKNESS_MATTERS
      strength = tyvarStrength(tyvar,state);
#endif
#endif
      e = CAR(args);
      if (e->kind==(CHAR)BUILD_HEADER_KIND)
	{
	  z = ZONE(e);
	  type = tc(e,assum,state);
	}
      else
	{
	  z = (TZone *)0L;
	  type = e;
	}
      eqtdg1 = MapGetSure(tyvar,eqflags);
      eqtdg2 = eqTypeTdg(type,state,false);
      aux = tdgAnd(tdgImply(eqtdg1,eqtdg2),S_EQCONDS(state));
      if (aux==falseValue)
	typeError(z,(CHAR *)message(41),type,state);
      SETQ(S_EQCONDS(state),aux);
      /* the following test suppressed, because a type variable
	 may be an extension of a record or a tuple; in this
	 case, it would be replaced only by records or tuples,
	 not numbers.
	 if (flags & F_NUM)
	 if (!numTypep(type,state))
	 typeError(z,(CHAR *)message(42),type,state);
	 */
      subst = MapPut(tyvar,type,subst);
    }
  KEEPONLY(subst);
  return subst;
}

MAPSWEEPFUNC(do_mergerecordtypes)
{
  struct mergerecordtypes *data;
  ExprPtr y2;

  data = MAPDATA(struct mergerecordtypes *);
  y2 = MapGetSure(x,data->rec2);
  (void) coerceTypes(data->zone,y2,y,data->assum,data->state,true);
  return false;
}

PRIVATE MapPtr mergeRecordTypes(zone,rec1,rec2,assum,state)
IN CONST TZone *zone;
IN CONST MapPtr rec1,rec2;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  MapPtr clash;
  struct mergerecordtypes *data;

  clash = MapDomRestrTo(rec1,rec2);
  data = (struct mergerecordtypes *)alloctmp(sizeof(struct mergerecordtypes));
  data->zone = zone;
  data->rec2 = rec2;
  data->assum = assum;
  data->state = state;
  (void) MapSweep(clash,do_mergerecordtypes,data);
  return MapMerge(rec1,rec2);
}

MAPSWEEPFUNC(do_typesubst)
{
  struct typesubst *data;
  ExprPtr yy;

  data = MAPDATA(struct typesubst *);
  if (x==REST_TAG && y->kind==(CHAR)E_NUMTYPE)
    y = REFVAL(y);
  y = typeSubst(data->zone,y,data->subst,data->assum,data->state);
  yy = (x==REST_TAG && ISMAP(y))?y:MapSingleton(x,y);
  data->new = mergeRecordTypes(data->zone,data->new,yy,data->assum,data->state);
  return false;
}

PRIVATE ExprPtr typeSubstTuple(zone,type,subst,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr type;
IN CONST MapPtr subst;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  if (ISCONS(type))
    return KONS(typeSubst(zone,CAR(type),subst,assum,state),
		typeSubstTuple(zone,CDR(type),subst,assum,state));
  type = typeSubst(zone,type,subst,assum,state);
  if (type!=NIL && type->kind!=(CHAR)E_CONS)
    typeError(zone,(CHAR *)message(71),type,state);
  return type;
}

PRIVATE ExprPtr typeSubstNum(zone,type,subst,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr type;
IN CONST MapPtr subst;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  ExprPtr new;
  Boolean fine;

  if (REFVAL(type)==NewMap())
    return type;
  new = instNumType(REFVAL(type),subst,state,&fine);
  if (!fine)
    typeError(zone,(CHAR *)message(42),type,state);
  new = KnumType(new);
  return KEEP(new);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE MapPtr do_typesubst_fx P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT CHAR *data));

PRIVATE ExprPtr typeSubstFx1(zone,sigma,subst,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr sigma;
IN CONST MapPtr subst;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  struct typesubst *data;

  sigma = fxRealSigma(sigma,state);
  data = (struct typesubst *)alloctmp(sizeof(struct typesubst));
  data->subst = subst;
  data->assum = assum;
  data->state = state;
  data->zone = zone;
  data->new = MapCollect(CDR(sigma),do_typesubst_fx,data);
  return dupFxVar(sigma,data->new,assum);
}

MAPCOLLECTFUNC(do_typesubst_fx)
{
  struct typesubst *data;
  ExprPtr (*subst) P((TZone *zone,ExprPtr f,MapPtr sbst,ExprPtr assum,ExprPtr state));

  data = MAPDATA(struct typesubst *);
  subst = (x->kind==(CHAR)E_ARRAY_2)?typeSubstFx1:typeSubst;
  x = (*subst)(data->zone,x,data->subst,data->assum,data->state);
  return MapSingleton(x,y);
}

PRIVATE ExprPtr typeSubstFx(zone,f,subst,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr f;
IN CONST MapPtr subst;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  if (/*f!=NIL && */f->kind==(CHAR)E_TYPEFX)
    {
      f = KtypeFx(CAR(f),typeSubstFx1(zone,CDR(f),subst,assum,state));
      return KEEP(f);
    }
  return f;
}
#endif

PUBLIC ExprPtr typeSubst(zone,type,subst,assum,state)
IN CONST TZone *zone;
IN CONST ExprPtr type;
IN CONST MapPtr subst;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
again:
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return NIL;*/
  switch (type->kind)
    {
    case E_NIL:
      return NIL;
    case E_VAR:
      {
	ExprPtr val;
	Boolean found;

	val = nongenValue(type,state,&found);
	if (found)
	  {
	    type = val;
	    goto again;
	  }
	val = MapGet(type,subst,&found);
	if (found)
	  return val;
	return type;
      }
    case E_TYPE_APPL:
#if IMPERATIVE_TYPES==EFFECT_TYVAR
      type = KtypeAppl(typeSubstFx(zone,APPLFUN(type),subst,assum,state), /* correct because of correctTypeAppl() */
			typeSubstTuple(zone,APPLARG(type),subst,assum,state));
#else
      type = KtypeAppl(APPLFUN(type), /* correct because of correctTypeAppl() */
			typeSubstTuple(zone,APPLARG(type),subst,assum,state));
#endif
      return KEEP(type);
    case E_CONS:
      type = typeSubstTuple(zone,type,subst,assum,state);
      if (ISCONS(type) && CDR(type)==NIL)
	typeError(zone,(CHAR *)message(70),CAR(type),state);
      return type;
    case E_EMPTY:
      return type;
    case E_MAPLET: /*too complicated to be treated separately*/
    case E_MAP:
      {
	struct typesubst *data = (struct typesubst *)alloctmp(sizeof(struct typesubst));
	
	data->subst = subst;
	data->assum = assum;
	data->state = state;
	data->zone = zone;
	data->new = NewMap();
	(void) MapSweep(type,do_typesubst,data);
	return data->new;
      }
    case E_NUMTYPE:
      return typeSubstNum(zone,type,subst,assum,state);
    default:
      kinderror(message(43),type->kind);
      return NIL;
    }
}

PUBLIC ExprPtr ApplyType(type,tyhdr,argtypes,assum,state)
IN CONST ExprPtr type;
IN CONST ExprPtr tyhdr;
IN CONST ExprPtr argtypes;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  ExprPtr subst;
  TZone zone;
  ENTER;

  zone.line1 = zone.line2 = 0;
  zone.pos1 = zone.pos2 = 0;
  subst = getTypeSubst(&zone,tyhdr,argtypes,assum,state);
  type = typeSubst(&zone,type,subst,assum,state);
  KEEPONLY(type);
  return type;
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE Boolean obsFxp(fx,vars,state)
IN CONST ExprPtr fx;
IN CONST MapPtr vars;
IN CONST ExprPtr state;
{
  MapPtr constr;
  Boolean found;

  if (/*fx!=NIL && */fx->kind==(CHAR)E_TYPE_APPL)
    {
      fx = APPLFUN(fx);
      if (fx->kind==(CHAR)E_TYPEFX)
	{
	  fx = CDR(fx); /* region variable */
	  goto test;
	}
      return false;
    }
  /* then fx is an effect variable: */
test:
  fx = fxRealSigma(fx,state);
  (void) MapGet(fx,vars,&found);
  if (found)
    return true;
  constr = obsFx(CDR(fx),vars,state);
  if (constr!=CDR(fx))
    {
      SETCDR(fx,constr);
    }
  return constr!=NewMap(); /* may still be observable if its constraint is */
}

MAPCOLLECTFUNC(do_obsfx)
{
  struct obsfx *data;

  data = MAPDATA(struct obsfx *);
  return obsFxp(x,data->vars,data->state)?MapSingleton(x,y):NewMap();
}

PUBLIC MapPtr obsFx(constr,vars,state)
IN CONST MapPtr constr;
IN CONST MapPtr vars;
IN CONST ExprPtr state;
{
  MapPtr newconstr;
  struct obsfx *data = (struct obsfx *)alloctmp(sizeof(struct obsfx));
  ENTER;

  data->vars = vars;
  data->state = state;
  newconstr = MapCollect(constr,do_obsfx,data);
  KEEPONLY(newconstr);
  return newconstr;
}

MAPCOLLECTFUNC(do_convertpatfx)
{
  ExprPtr f,rho;
  Boolean found;

  if (/*x!=NIL && */x->kind==(CHAR)E_TYPE_APPL)
    {
      f = APPLFUN(x); /* no need to correctTypeAppl(), since effects do not occur in modules */
      if (f->kind==(CHAR)E_TYPEFX)
	{
	  rho = CDR(f);
	  f = CAR(f);
	  f = MapGet(f,autoPatFx,&found);
	  if (found)
	    {
	      f = KtypeFx(f,rho);
	      x = KtypeAppl(KEEP(f),NIL);
	    }
	}
    }
  return MapSingleton(x,y);
}

PUBLIC MapPtr convertPatFx(fx)
IN CONST MapPtr fx;
{
  ENTER;

  fx = MapCollect(fx,do_convertpatfx,(char *)0L);
  KEEPONLY(fx);
  return fx;
}
#endif

/***** type renaming *****/

PRIVATE ExprPtr InitTypeRename(zonex,varok)
IN CONST ExprPtr zonex;
IN Boolean varok;
{ /* return initial info (renaming,counter,varok,zone,bias,reject,warnp) */
  VecPtr vp;
  ExprPtr aux,aux1,aux2;
  ENTER;

  MAKEINT(0,aux1); (void) KEEP(aux1);
  MAKEINT(0,aux2); (void) KEEP(aux2);
  CHECKHETEROFREE();
  VECALLOC(*(char **)&vp,VECSIZE(7));
  vp->n = 7;
  vp->expr[0] = NewMap();
  vp->expr[1] = aux1;
  vp->expr[2] = (varok?trueValue:falseValue);
  vp->expr[3] = zonex;
  vp->expr[4] = aux2;
  vp->expr[5] = trueValue;
  vp->expr[6] = trueValue;
  MAKETUPLE_NOGC(vp,aux);
  KEEPONLY(aux);
  return aux;
}

#define I_RENAMING(e) AVAL(e)->expr[0]
#define I_COUNTER(e) IVAL(AVAL(e)->expr[1])
#define I_VAROK(e) AVAL(e)->expr[2]
#define I_ZONEX(e) AVAL(e)->expr[3]
#define I_BIAS(e) IVAL(AVAL(e)->expr[4])
#define I_REJECT(e) AVAL(e)->expr[5]
#define I_WARNP(e) AVAL(e)->expr[6]

PRIVATE void ResetTypeRename(info,assum)
INOUT ExprPtr info;
IN CONST ExprPtr assum;
{
  ExprPtr ren;

  ren = MapRngRestrTo(I_RENAMING(info),A_NONGENS(assum));
  SETQ(I_RENAMING(info),ren);
  HITBARRIER(info);
  I_COUNTER(info) = 0;
}

PRIVATE ExprPtr TypeRename1 P((IN CONST ExprPtr type,INOUT ExprPtr info,
			       IN CONST ExprPtr assum,
			       IN CONST ExprPtr state));

PRIVATE ExprPtr tyvarRename(type,info,assum,state)
IN CONST SymPtr type;
INOUT ExprPtr info;
IN CONST ExprPtr assum,state;
{
  ExprPtr rename,eqtdg,eqtdg0;
  Boolean found;
  Boolean existential;
  CHAR buf[16],*t;
  COUNTER i;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  INT strength;
#endif
  int flags;

  rename = MapGet(type,I_RENAMING(info),&found);
  existential = false;
  if (!found)
    {
      if (!genericp(type,assum))
	{
	  rename = nongenValue(type,state,&found);
	  if (!found)
	    {
#ifdef DEBUG
#if CORE_TRACE
	      if (coreTraceFile!=NULL)
		hfflush(coreTraceFile);
#endif
	      CRASH();
#endif
	      (void) MapGet(type,S_NONGENS(state),&found);
	      if (found)
		return type;
	      existential = true;
	    }
	  else
	    return TypeRename1(rename,info,assum,state);
	}
      /*if (I_VAROK(info)!=trueValue)
	typeError(ZONE(CDR(CDR(info))),
	(CHAR *)message(44),
	type,state);*/
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
      strength = tyvarStrength(type,state)-I_BIAS(info);
      if (strength<=0)
	{
	  strength = 0;
	  if (I_REJECT(info)==trueValue)
	    typeError(ZONE(I_ZONEX(info)),(CHAR *)message(46),type,state);
	}
#endif
      flags = varFlags(type) & ~F_EQ;
      eqtdg = eqTypeTdg(type,state,false);
      do {
	t = buf;
	*t++ = '\'';
	if (eqtdg==trueValue && !(flags & F_NUM)) /* for numerical tyvars, no need to print
						     the equality attribute */
	  *t++ = '\'';
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
	if (flags & F_IMP)
	  {
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	    sprintf((char *)t,"%ld",(long)strength);
	    t += strlen((char *)t);
#else
	    *t++ = '_';
#endif
	  }
#endif
	if (flags & F_NUM)
	  *t++ = '#';
	if (eqtdg!=trueValue && eqtdg!=falseValue)
	  {
	    *t++ = '!';
	    /* means we should find a unique variable having
	       this eqtdg; all variables with a '!' character
	       in their names have special eqtdgs stored in
	       the state */
	  }
	i = existential?(S_NEXGEN(state)++):(I_COUNTER(info)++);
	do {
	  *t++ = 'a'+(i%26);
	  i = i/26;
	} while (i!=0);
	if (existential)
	  *t++ = '?';
	*t = 0;
	rename = Ksymbol(buf);
	(void) KEEP(rename);
	found = false;
	if (eqtdg!=trueValue && eqtdg!=falseValue)
	  {
	    eqtdg0 = MapGet(rename,S_EQFLAGS(state),&found);
	    if (found)
	      found = (tdgInst(eqtdg0,S_NONEQS(state),false)!=eqtdg);
	    /* if the variable we built (rename) is already
	       recorded with a different eqtdg, then we should
	       find another '!' name */
	  }
      } while (found);
      if (existential)
	{
	  if (I_WARNP(info)==trueValue)
	    warnGen(ZONE(I_ZONEX(info)),rename);
	  setSetAdd(rename,&S_NONGENS(state));
	}
      setMapPut(type,rename,&I_RENAMING(info));
      HITBARRIER(info);
#ifdef BUGGY
      if (rename!=type)
	{ /* if type was renamed, kill flags associated with original variable */
	  setMapRemove(type,&S_EQFLAGS(state));
	}
#endif
    }
  return rename;
}

MAPSWEEPFUNC(do_dump_rectype)
{
  field **fpp,*fp;

  fpp = MAPDATA(field **);
  fp = (*fpp)++;
  fp->label = x;
  fp->val = y;
  return false;
}

PRIVATE ExprPtr fieldsToRecType(n,array)
IN COUNTER n;
IN CONST field *array;
{
  if (n==0)
    return NewMap();
  if (n==1)
    return MapSingleton(array->label,array->val);
  {
    COUNTER m;

    m = n>>1;
    return MapMerge(fieldsToRecType(m,array),
		    fieldsToRecType(n-m,array+m));
  }
}

PRIVATE ExprPtr recTypeRename(type,info,assum,state)
IN CONST ExprPtr type;
INOUT ExprPtr info;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  field *fields,*fp;
  ExprPtr rest;
  COUNTER n,i;
  ENTER;

  type = get_real_map_type(type,assum,state,&rest,false);
  n = MapCard(type);
  fields = (field *)alloctmp(n*sizeof(field));
  fp = fields;
  (void) MapSweep(type,do_dump_rectype,&fp);
  sortFieldArray(n,fields); /* we rename fields in order, so that identical types
			       have identical tyvar names at the same positions. */
  for (fp=fields, i=0; i<n; i++, fp++)
    fp->val = TypeRename1(fp->val,info,assum,state);
  rest = TypeRename1(rest,info,assum,state);
  type = fieldsToRecType(n,fields);
  if (rest!=NIL)
    type = addMapRest(rest,type);
  KEEPONLY(type);
  return type;
}

PRIVATE int cmpnum(_m1,_m2)
IN CONST ANSI_CONST VOIDPTR _m1;
IN CONST ANSI_CONST VOIDPTR _m2;
{
  REGISTER field *m1 = (field *)_m1;
  REGISTER field *m2 = (field *)_m2;

  if (m1->label->kind==(CHAR)E_TYPE_APPL)
    {
      if (m2->label->kind==(CHAR)E_TYPE_APPL)
	return CompareStrings(APPLFUN(m1->label),APPLFUN(m2->label));
      return -1; /* dimensions are always less than variables */
    }
  if (m2->label->kind==(CHAR)E_TYPE_APPL)
    return 1; /* variables are always greater than dimensions */
  /* now, if m1 and m2 are variables, compare their exponents: */
  {
    ExprPtr e;
    EXTENDED x;
    int i;
    ENTER;

    e = z_sub(m1->val,m2->val);
    if (e->kind==(CHAR)E_COMPLEX)
      x = XVAL(CAR(e));
    else
      x = XVAL(e);
    i = NEGX(x)?(-1):(ZEROX(x)?0:1);
    if (i==0 && e->kind==(CHAR)E_COMPLEX)
      {
	x = XVAL(CDR(e));
	i = NEGX(x)?(-1):(ZEROX(x)?0:1);
      }
    QUIT;
    return i;
  }
}

PRIVATE ExprPtr numTypeRename(type,info,assum,state)
IN CONST ExprPtr type;
INOUT ExprPtr info;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  field *fields,*fp;
  ExprPtr map;
  COUNTER n,i;
  Boolean fine;
  ENTER;

  map = instNumType(REFVAL(type),NewMap(),state,&fine); /* fine should be true */
  n = MapCard(map);
  fields = (field *)alloctmp(n*sizeof(field));
  fp = fields;
  (void) MapSweep(map,do_dump_rectype,&fp);
  qsort(fields,(long)n,(long)sizeof(field),cmpnum);
  /* we do this so as to provide a canonical renaming of types,
     independent of the system order. */
  for (fp=fields, i=0; i<n; i++, fp++)
    fp->label = TypeRename1(fp->label,info,assum,state);
  map = fieldsToRecType(n,fields);
  type = KnumType(map);
  KEEPONLY(type);
  return type;
}

MAPCOLLECTFUNC(do_typerenamemap_0)
{
  struct typerenamemap *data;
  Boolean found;
  ExprPtr e;

  data = MAPDATA(struct typerenamemap *);
  e = MapGet(x,A_IDLOCS(data->assum),&found);
  if (found)
    {
      ResetTypeRename(data->info,data->assum);
      SETQ(I_ZONEX(data->info),e);
      HITBARRIER(data->info);
      y = TypeRename1(y,data->info,data->assum,data->state);
      if (x==REST_TAG && ISMAP(y))
	return y;
    }
  /* found can only be false when x is a variable that was defined in
     an imported module (its defining location is not recorded), then
     we don't have to rename its type. */
  return MapSingleton(x,y);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE MapPtr do_typerename_fx P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT CHAR *data));

PRIVATE long fxCounter=0L;

PRIVATE ExprPtr typeRenameFx1(sigma,info,assum,state)
IN CONST ExprPtr sigma;
INOUT ExprPtr info;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  struct typerenamemap *data;
  ExprPtr fx,name;
  Boolean found;

  sigma = fxRealSigma(sigma,state);
  (void) MapGet(sigma,I_RENAMING(info),&found);
  if (found)
    return sigma;
  data = (struct typerenamep *)alloctmp(sizeof(struct typerenamep));
  data->info = info;
  data->assum = assum;
  data->state = state;
  fx = MapCollect(CDR(sigma),do_typerename_fx,data);
  SETCDR(sigma,fx);
  setMapPut(sigma,sigma,&I_RENAMING(info));
  HITBARRIER(info);
  (void) MapGet(sigma,A_NONGENS(assum),&found);
  if (found)
    {
      name = MapGet(sigma,S_NONGENS(state),&found);
      if (!found) /* record a default, immutable name */
	{
	  char buf[16],*s;

	  safememcpy(buf,(char *)STEXT(CAR(sigma)),SLEN(CAR(sigma)));
	  s = buf+SLEN(CAR(sigma));
	  sprintf(s,"%ld?",++fxCounter);
	  name = Ksymbol(buf);
	  (void) KEEP(name);
	  setMapPut(sigma,name,&S_NONGENS(state));
	}
    }
  return sigma;
}

MAPCOLLECTFUNC(do_typerename_fx)
{
  struct typerenamemap *data;
  ExprPtr (*ren) P((ExprPtr f,ExprPtr info,ExprPtr assum,ExprPtr state));

  data = MAPDATA(struct typerenamemap *);
  ren = (x->kind==(CHAR)E_ARRAY_2)?typeRenameFx1:TypeRename1;
  x = (*ren)(x,data->info,data->assum,data->state);
  return MapSingleton(x,y);
}

PRIVATE ExprPtr typeRenameFx(f,info,assum,state)
IN CONST ExprPtr f;
INOUT ExprPtr info;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  struct typerenamemap data;
  ExprPtr ff,fx;
  Boolean found;

  if (f->kind==(CHAR)E_TYPEFX)
    (void) typeRenameFx1(CDR(f),info,assum,state);
  return f;
}
#endif

PRIVATE ExprPtr TypeRename1(type,info,assum,state)
IN CONST ExprPtr type;
INOUT ExprPtr info;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  type = correctTypeAppl(type,state);
  /*if (type==NIL)
    return type;
    else*/ switch (type->kind)
      {
      case E_NIL:
	return type;
      case E_VAR:
	return tyvarRename(type,info,assum,state);
      case E_TYPE_APPL:
	{
	  ExprPtr f,l;
	  
	  f = APPLFUN(type);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	  f = typeRenameFx(f,info,assum,state);
#endif
	  l = TypeRename1(APPLARG(type),info,assum,state);
	  type = KtypeAppl(f,l);
	  return KEEP(type);
	}
      case E_CONS:
	return KONS(TypeRename1(CAR(type),info,assum,state),
		    TypeRename1(CDR(type),info,assum,state));
      case E_EMPTY:
	return type;
      case E_MAPLET:
	return MapSingleton(MAPLETX(type),
			    TypeRename1(MAPLETY(type),info,assum,state));
      case E_MAP:
	return recTypeRename(type,info,assum,state);
      case E_NUMTYPE:
	if (REFVAL(type)==NewMap())
	  return type;
	return numTypeRename(type,info,assum,state);
      default:
	kinderror(message(45),type->kind);
	return NIL;
      }
}

PUBLIC ExprPtr newMemoRef()
{
#ifdef TYPE_CACHE
  ExprPtr e;

  e = MakeSing(E_REF,NewMap()); /* old: MakeSing(E_REF,NewMap()) */
  return KEEP(e);
#else
  return NIL;
#endif
}

PRIVATE Boolean do_pack_check P((IN CONST ExprPtr maplet,
				 IN CONST char *_data));
PRIVATE Boolean do_pack_check(maplet,_data)
IN CONST ExprPtr maplet;
IN CONST char *_data;
{
  ExprPtr expr,type,tyvars;
  struct pack_check *data;
  ENTER;

  data = (struct pack_check *)_data;
  expr = MAPLETX(maplet);
  /*if (TYPE(expr)!=A_PACK)
    return false; // no, A_PAT_PACK must be considered, too */
  type = MAPLETY(maplet);
  if (/*type!=NIL && */type->kind==(CHAR)E_BIAS)
    type = CAR(type); /* ignore bias; checking abusively quantified
			 type variables because of a bias will be done in tyvarRename */
  tyvars = tyVars(type,data->assum,data->state,0x80
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		  ,INF_STRENGTH
#endif
		  ,data->memoref);
  if (tyvars!=NewMap()) /* this is not the only error case;
			   see tyvarRename */
    typeError(ZONE(expr),(CHAR *)message(46),ChooseInDom(tyvars),data->state);
  KEEPONLY(data->memoref);
  return false;
}

PRIVATE void pack_check(assum,state)
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  struct pack_check *data = (struct pack_check *)alloctmp(sizeof(struct pack_check));
  ENTER;

  data->assum = assum;
  data->state = state;
  data->memoref = newMemoRef();
  (void) NSweep(S_PACKERS(state),do_pack_check,(char *)data);
  QUIT;
}

MAPCOLLECTFUNC(do_insttyfcn)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
  if (ISCONS(y))
    y = KONS(CAR(y),newTypeInst(CDR(y),NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				,false
#endif
				));
  return MapSingleton(x,y);
}

MAPCOLLECTFUNC(do_killnongens)
{
  MapPtr nongens;
  Boolean found;

  nongens = MAPDATA(MapPtr);
  if (x->kind==(CHAR)E_VAR && symName(x)[symNameLen(x)-1]=='?' && ((void) MapGet(x,nongens,&found), !found))
    return NewMap(); /* erase flags of temporary non generics */
  return MapSingleton(x,y);
}

PRIVATE Boolean do_renamelist P((IN CONST ExprPtr loc,
				 INOUT char *_data));
PRIVATE Boolean do_renamelist(loc,_data)
IN CONST ExprPtr loc;
INOUT char *_data;
{
  struct typerenamemap *data;
  ExprPtr e,ee;
  ENTER;

  data = (struct typerenamemap *)_data;
  ResetTypeRename(data->info,data->assum);
  e = MAPLETX(loc);
  SETQ(I_ZONEX(data->info),e);
  HITBARRIER(data->info);
  e = MAPLETY(loc);
  if (/*e!=NIL && */e->kind==(CHAR)E_BIAS)
    {
      if (CDR(e)->kind==E_INT) /* case of packers list, where entries
				  are E_BIAS (type, bias) */
	I_BIAS(data->info) = IVAL(CDR(e));
      else /* case of annotations list, where entries are
	      E_BIAS (type, E_CONS (bias, sharingp)) */
	I_BIAS(data->info) = IVAL(CAR(CDR(e)));
      ee = CAR(e);
    }
  else ee = e;
  ee = TypeRename1(ee,data->info,data->assum,data->state);
  if (e->kind==(CHAR)E_BIAS)
    {
      ee = Kbias(ee, CDR(e));
      (void) KEEP(ee);
    }
  SETMAPLETY(loc,ee);
  QUIT;
  return false;
}

PRIVATE void renameList(loc,data)
IN CONST ExprPtr loc;
INOUT struct typerenamemap *data;
{
  (void) NSweep(loc,do_renamelist,(char *)data);
}

MAPCOLLECTFUNC(do_insttyvarsubs)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  if (/*y!=NIL && */y->kind==(CHAR)E_ARRAY_2)
    {
      ExprPtr fx;
      MapPtr rename;

      rename = NewMap();
      fx = newTypeInstFx1(CDR(y),NIL,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			  false,
#endif
			  &rename);
      SETCDR(y,fx);
    }
#endif
  return MapSingleton(x,newTypeInst(y,NIL,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
				    ,false
#endif
				    ));
}

PUBLIC ExprPtr tcDeclaration(expr,assum,state)
INOUT ExprPtr expr;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  struct typerenamemap *data;
  ExprPtr new_assum,local_state,aux,oldnongens,newtyvarsubs;
  COUNTER oldnexgen;
  ENTER;

  local_state = DupTypeState(state);
  oldnexgen = S_NEXGEN(local_state);
  oldnongens = KEEP(S_NONGENS(local_state));
  setSetUnion(&A_NONGENS(assum),oldnongens);
  new_assum = tc(expr,assum,local_state);
  pack_check(assum,local_state);
  S_NEXGEN(local_state) = oldnexgen;
  data = (struct typerenamemap *)alloctmp(sizeof(struct typerenamemap));
  data->info = InitTypeRename(expr,true);
  SETQ(I_REJECT(data->info),falseValue);
  SETQ(I_WARNP(data->info),trueValue);
  HITBARRIER(data->info);
  data->assum = assum;
  data->state = local_state;
  SETQ(A_IDLOCS(assum),A_IDLOCS(new_assum));
  /*setMapDomRestrBy(&A_IDEXTS(new_assum), A_IDS(assum));*/
  aux = MapCollect(A_IDS(new_assum),do_typerenamemap_0,data);
  SETQ(A_IDS(new_assum),aux);
  aux = MapCollect(A_IDEXTS(new_assum),do_typerenamemap_0,data);
  SETQ(A_IDEXTS(new_assum),aux);
  SETQ(I_WARNP(data->info),falseValue);
  SETQ(I_REJECT(data->info),trueValue);
  HITBARRIER(data->info);
  /* reject fn x => pack ref x, even though x : '1a, because
     inside the pack, the fnlevel (bias) is 1 and must be
     subtracted from the level of x, which is 0 at runtime,
     hence cannot be quantified */
  renameList(S_PACKERS(local_state),data);
  SETQ(I_REJECT(data->info),falseValue);
  HITBARRIER(data->info);
  /* in annotations (typically for the debugger), we must
     correct the type of, say, x in fn x => ref x, so that
     it will be '0a (instead of '1a viewed from the toplevel).
     The only problem is that '0a will be quantified, which
     makes type-checking unsafe; but it is already unsafe
     under the debugger, so we won't correct it. */
  renameList(S_ANNOTATIONS(local_state),data);
  SETQ(A_IDLOCS(assum),NewMap());
  SETQ(A_IDLOCS(new_assum),NewMap());
  aux = MapDomRestrTo(A_IDS(new_assum),A_CONS(new_assum));
  SETQ(A_CONS(new_assum),aux);
  SETQ(A_TYVARS(new_assum),NewMap());
  /*setMapDomRestrBy(&A_TYEXTS(new_assum), A_TYFCNS(assum));*/
  setMapMerge(&A_TYFCNS(assum),
	      MapCollect(A_TYFCNS(new_assum),do_insttyfcn,local_state));
  setMapMerge(&A_TYEXTS(assum),
	      MapCollect(A_TYEXTS(new_assum),do_insttyfcn,local_state));
  /* we need only instantiate the new type functions, since this is
     the typechecking phase of the *toplevel*; hence, no old tyfcn
     has flexible type constructors in it */
  SETQ(A_NONGENS(new_assum),NewMap());
  SETQ(A_NONGENS(assum),NewMap());
  SETQ(S_TYCONS(state),S_TYCONS(local_state)); /* what do we do with 2nd order substs ? */
  SETQ(S_CONDS(state),S_CONDS(local_state)); /* what do we do with 2nd order substs ? */
  aux = MapCollect(S_EQFLAGS(local_state),do_killnongens,S_NONGENS(state));
  SETQ(S_EQFLAGS(state),aux);
  flushTyVarSubCache(local_state);
  SETQ(S_TYVARSUBCACHE(state),S_TYVARSUBS(local_state));
  flushTyVarSubCache(state);
  newtyvarsubs = MapCollect(MapDomRestrTo(S_TYVARSUBS(state),S_NONGENS(local_state)),
			    do_insttyvarsubs,local_state);
  SETQ(S_TYVARSUBS(state),newtyvarsubs);
  S_NEXGEN(state) = S_NEXGEN(local_state);
  SETQ(S_NONGENS(state),S_NONGENS(local_state));
  SETQ(S_NONEQS(state),S_NONEQS(local_state));
  SETQ(S_EQCONDS(state),S_EQCONDS(local_state));
  SETQ(S_PACKERS(state),S_PACKERS(local_state));
  SETQ(S_ANNOTATIONS(state),S_ANNOTATIONS(local_state));
  /* S_PACKERS, S_ANNOTATIONS kept for compilation */
  /* S_MAXSTRENGTH should always be set to {} */
  KEEPONLY(new_assum);
  return new_assum;
}

MAPFOLDFUNC(do_respectseq)
{
  struct eqtypep *data;
  ExprPtr type,t;

  data = MAPDATA(struct eqtypep *);
  y = correctTypeAppl(y,data->state);
  if (/*y!=NIL && */y->kind==(CHAR)E_TYPE_APPL &&
      (type = APPLFUN(y), fxSym(type)==funSym))
    {
      type = CAR(APPLARG(y));
      t = eqTypeTdg(type,data->state,true);
      /* find out whether type admits equality whenever all
	 its variables admit equality; that is, if it respects
	 equality */
      if (t==falseValue)
	LONGJUMP(data->exit,1);
      return t;
    }
  return trueValue;
}

MAPCOLLECTFUNC(do_initmaxeq)
{
  MapPtr sureeq;
  Boolean found;

  sureeq = MAPDATA(MapPtr);
  (void) MapGet(x,sureeq,&found);
  if (found)
    return NewMap();
  y = APPLFUN(CDR(y));
  y = fxSym(y);
  return MapSingleton(y,nameToTdg(y));
}

PRIVATE ExprPtr geteq(y,data)
     IN ExprPtr y;
     IN struct eqtypep *data;
{
  if (SETJUMP(data->exit)==0)
    return MapFold(y,do_respectseq,data,trueValue,tdgAnd);
  else return falseValue;
}

MAPCOLLECTFUNC(do_geteq)
{
  struct geteq *data;
  /* x is a type identifier, y is a map from constructors to types */
  ExprPtr t;
  ExprPtr type;
  ExprPtr f;
  struct eqtypep *data2 = (struct eqtypep *)alloctmp(sizeof(struct eqtypep));
  Boolean found;

  data = MAPDATA(struct geteq *);
  (void) MapGet(x,data->sureeq,&found);
  if (found)
    return NewMap();
  data2->state = data->state;
  if (data->noeq)
    t = falseValue;
  else
    t = geteq(y,data2);
  /* t is now a condition Ci depending on all tynames,
     which must be true for the tyname ti to admit equality */
  type = MapGetSure(x,data->tyfcns);
  /* type is of the form (tyhdr,tname(tyvarseq)) */
  f = applFun(CDR(type),data->state);
  f = fxSym(f);
  return MapSingleton(f,t);
}

MAPCOLLECTFUNC(do_truesubst)
{
  return MapSingleton(x,trueValue);
}

MAPCOLLECTFUNC(do_repleq)
{
  return MapSingleton(x,tdgInst(y,MAPDATA(MapPtr),true));
}

MAPSWEEPFUNC(do_checkeqconds)
{
  struct checkeqcond *data;
  ExprPtr y1,y2;
  Boolean found;

  data = MAPDATA(struct checkeqcond *);
  y1 = MapGet(x,data->eqs1,&found);
  if (!found)
    y1 = trueValue;
  y2 = MapGet(x,data->eqs2,&found);
  if (!found)
    y2 = trueValue;
  y1 = tdgImply(y1,y2);
  if (y1==falseValue)
    type_error(data->zone,(CHAR *)nameMsg(message(784),x));
  return false;
}

PRIVATE void checkEqConds(spec,eqs1,eqs2,state,zone)
     /* check that all eqconds of tynames of spec in eqs2 are implied by ones in eqs1 */
IN CONST MapPtr spec;
IN CONST MapPtr eqs1,eqs2;
INOUT ExprPtr state;
IN CONST TZone *zone;
{
  struct checkeqcond *data = (struct checkeqcond *)alloctmp(sizeof(struct checkeqcond));

  data->eqs1 = eqs1;
  data->eqs2 = eqs2;
  data->state = state;
  data->zone = zone;
  (void) MapSweep(spec,do_checkeqconds,data);
}

PUBLIC void maximizeEquality(consmap,tyfcns,state,sureeq,spec,noeq,zone)
IN CONST MapPtr consmap,tyfcns; /* consmap : type ids -> cons */
INOUT ExprPtr state;
IN CONST MapPtr sureeq;
IN CONST MapPtr spec;
IN CONST Boolean noeq;
IN CONST TZone *zone;
{
  MapPtr eq0,aux;
  struct geteq *data = (struct geteq *)alloctmp(sizeof(struct geteq));
  MapPtr simpleeqs,oldnoneqs;
  MapPtr subst,newsubst;
  ENTER;

  eq0 = MapCollect(tyfcns,do_initmaxeq,sureeq);
  oldnoneqs = S_NONEQS(state);
  (void) KEEP(oldnoneqs);
  setMapMerge(&S_NONEQS(state),eq0);
  data->state = state;
  data->tyfcns = tyfcns;
  data->sureeq = sureeq;
  data->noeq = noeq;
  simpleeqs = MapCollect(consmap,do_geteq,data);
  /* simpleeqs is a map from tynames Ti to tdgs Ci. We must have
     that, in the context of all Tj's (j<>i) realizing fixpoints,
     Ti must also be a fixed point.
     This means that the real fixpoint equation of Ti is:
     [forall j<>i & Tj=Cj] => Ti=Ci
     */
  newsubst = MapCollect(eq0,do_truesubst,(char *)0L);
  do {
    subst = newsubst;
    newsubst = MapCollect(simpleeqs,do_repleq,subst);
  } while (subst!=newsubst);
  checkEqConds(spec,oldnoneqs,subst,state,zone);
  aux = MapMerge(oldnoneqs,subst);
  SETQ(S_NONEQS(state),aux);
  QUIT;
}

PRIVATE unsigned long tyNameGen=0L;

PRIVATE SymPtr freshTyName1(name,len,assum,state,cnt)
IN CONST CHAR *name;
INT len;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
IN unsigned long cnt;
{
  SymPtr sym;
  char *buf,*s,*t,c;
  INT ofs,modlen,i;

  modlen = (moduleName==NIL)?0:SLEN(moduleName);
  t = buf = (char *)alloctmp(len+modlen+16);
  if (moduleName!=NIL)
    {
      for (i=0, s=(char *)STEXT(moduleName); i<modlen; i++)
	{
	  c = *s++;
	  if (c=='%') /* replace % signs by DEL, which is impossible to type in a module name */
	    c = 0x7f;
	  *t++ = c;
	}
      *t++ = '-';
    }
  sprintf(t,"%ld%%",cnt);
  /* used instead of S_NEXGEN(state) because of generative declarations in structures:
     'datatype t=...' in two different structures must have different names,
     even though structures contain type states with the nexgens reset before tc() */
  ofs = strlen(t);
  t += ofs;
  safememcpy(t,name,len);
  t[len] = 0;
  sym = Ksymbol((CHAR *)buf);
  return KEEP(sym);
}

PUBLIC SymPtr freshTyName(name,assum,state)
IN CONST SymPtr name;
IN CONST ExprPtr assum;
INOUT ExprPtr state;
{
  return freshTyName1(symName(name),symNameLen(name),assum,state,++tyNameGen);
  /* correct even in stop-and-copy, because freshTyName1 does
     not allocate before having copies its first argument */
}

PRIVATE SymPtr concTyName(name,suffix)
IN CONST SymPtr name;
IN CONST CHAR *suffix;
{
  SymPtr sym;
  INT len;

  len = symNameLen(name)+strlen((char *)suffix);
  {
    CHAR *buf;

    buf = (CHAR *)alloctmp(len+1);
    safememcpy(buf,symName(name),symNameLen(name));
    strcpy((char *)(buf+symNameLen(name)),(char *)suffix);
    sym = Kstring(len,buf);
  }
  return KEEP(sym);
}

PUBLIC SymPtr copyTyName(name,state)
IN CONST SymPtr name;
INOUT ExprPtr state;
{
  REGISTER CHAR *HEAP s;
  CHAR c;
  INT len;

  len = symNameLen(name);
  s = symName(name);
  while (c = *s++, --len>0 && c!='%');
  if (len>0)
    return freshTyName1(s,len,NIL,state,++tyNameGen);
  /* correct even in stop-and-copy, because freshTyName1 does
     not allocate before having copies its first argument */
  return freshTyName(name,NIL,state);
}

PUBLIC SymPtr getCdrTyName(name,state)
IN CONST SymPtr name;
IN CONST ExprPtr state;
{
  return concTyName(name,(CHAR *)message(48));
}

PUBLIC SymPtr getCarTyName(name,state)
IN CONST SymPtr name;
IN CONST ExprPtr state;
{
  return concTyName(name,(CHAR *)message(49));
}

MAPCOLLECTFUNC(do_initabseq)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
  y = applFun(CDR(y),state);
  return MapSingleton(fxSym(y),falseValue);
}

PUBLIC MapPtr InitAbsEq(tyfcns,state)
IN CONST MapPtr tyfcns;
IN CONST ExprPtr state;
{
  MapPtr tynames;
  ENTER;

  tynames = MapCollect(tyfcns,do_initabseq,state);
  KEEPONLY(tynames);
  return tynames;
}

PUBLIC Boolean TypeMatch(pat,type)
IN CONST MapPtr pat,type;
{
  ExprPtr assum,state;
  Boolean res;
  ENTER;

  assum = DupTypeAssum(basisAssum);
  state = DupTypeState(basisState);
  res = unify(newTypeInst(pat,assum,state
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			  ,false
#endif
			  ),
	      type,assum,state,false);
  QUIT;
  return res;
}

PUBLIC char *varMsg(msg,var)
IN CONST char *msg;
IN CONST SymPtr var;
{
#define MAXSIZE 128
  static char buf[MAXSIZE];
  char *s;
  long len;

  strcpy(buf,msg);
  s = buf+strlen(buf);
  *s++ = ':';
  *s++ = ' ';
  len = (buf+MAXSIZE-2-s);
  if (len>symNameLen(var))
    len = symNameLen(var);
  strncpy(s,(char *)symName(var),len);
  s[len] = 0;
  s += strlen(s);
  /* *s++ = '.'; */
  *s = 0;
#undef MAXSIZE
  return buf;
}

PRIVATE void warnGen(zone,sym)
IN CONST TZone *zone;
IN CONST ExprPtr sym;
{
  type_warn(zone,(CHAR *)varMsg(message(50),sym));
}

#if IMPERATIVE_TYPES==STD_ML_TYVAR
PUBLIC Boolean expansivep(expr)
IN CONST ExprPtr expr;
{
  while (TYPE(expr)==A_CAST)
    expr = CAR(ARGS(expr));
  switch (TYPE(expr))
    {
    case A_VARIABLE:
    case A_NUMBER:
    case A_SET_UNIT:
    case A_STRING:
    case A_FN:
    case A_SELECT_FN:
      /*case A_DELAY:*/
      return false;
    case A_TUPLE:
    case A_LIST:
    case A_MAP:
      return ARGS(expr)!=NIL;
    default:
      return true;
    }
}
#endif

PUBLIC MapPtr valTyVars(tyvarlist,assum,state)
IN CONST ExprPtr tyvarlist;
INOUT ExprPtr assum;
INOUT ExprPtr state;
{
  MapPtr outer_tyvars,tyvars;
  ExprPtr var;
  Boolean found;
  ENTER;

  outer_tyvars = A_TYVARS(assum);
  tyvars = NewMap();
  for (; tyvarlist!=NIL; tyvarlist=CDR(tyvarlist))
    {
      var = CAR(ARGS(CAR(tyvarlist)));
      (void) MapGet(var,outer_tyvars,&found);
      if (!found)
	(void) freshInst(var,assum,state,
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
			false,
#endif
			&tyvars);
    }
  KEEPONLY(tyvars);
  return tyvars;
}

MAPSWEEPFUNC(do_checkexplicits)
{
  struct checkexplicits *data;
  ExprPtr z, oldx;
  Boolean found;
  int flags;
  CHAR *s, *t;

#ifdef DEBUG_EXPLICIT
  hfprintf(hstderr, "x = ");
  TP(x);
  hfprintf(hstderr,".\n");
#endif
  data = MAPDATA(struct checkexplicits *);
  y = _nongenValue(y,data->state);
  if (/*y!=NIL && */y->kind==(CHAR)E_NUMTYPE)
    {
      z = REFVAL(y);
      if (z->kind==(CHAR)E_MAPLET && MAPLETY(z)==one)
	y = MAPLETX(z);
    }
  if (y->kind==(CHAR)E_VAR && !genericp(y,data->assum))
    {
      t = (CHAR *)message (925);
      goto fail_0;
    }
  if (/*y==NIL || */y->kind!=(CHAR)E_VAR)
    {
    fail0:
      t = (CHAR *)message(922);
    fail_0:
      s = alloctmp (strlen((char *)t)+1);
      strcpy ((char *)s, (char *)t);

    fail:
      incompatibleTypes(data->zone,x,y,(CHAR *)message(59),s,data->assum,data->state);
    fail1:
      data->wrong = x;
      return true;
    }
  oldx = MapGet(y,data->values,&found);
  if (found && oldx!=x)
    {
      /* subtle point: two different explicit variables, even when
	 they are instantiated to generic variables, must not be
	 instantiated to the same generic */
      y = oldx;
      goto fail0;
    }
  flags = varFlags (x);
  if ((flags & F_EQ)==0)
    { /* then check that y does not admit equality, surely. */
      if (eqTypeTdg(y,data->state,false)!=falseValue)
	{
	  typeError (data->zone, (CHAR *)message(923), x, data->state);
	  goto fail1;
	}
    }
  if ((flags & F_IMP)==0)
    { /* then check that y is not imperative, surely (we don't check strengths). */
      if (varFlags (y) & F_IMP)
	{
	  typeError (data->zone, (CHAR *)message(924), x, data->state);
	  goto fail1;
	}
    }
  data->values = MapPut(y,x,data->values);
  return false;
}

PUBLIC ExprPtr checkExplicits(zone,tyvarmap,assum,state)
IN CONST TZone *zone;
IN CONST MapPtr tyvarmap;
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
{
  struct checkexplicits *data = (struct checkexplicits *)alloctmp(sizeof(struct checkexplicits));
  ENTER;

#ifdef DEBUG_EXPLICIT
  if (tyvarmap!=EmptyMap())
    hfprintf(hstderr, "Entering checkExplicits().\n");
#endif
  data->zone = zone;
  data->assum = assum;
  data->state = state;
  data->values = NewMap();
  data->wrong = NIL;
  (void) MapSweep(tyvarmap,do_checkexplicits,data);
#ifdef DEBUG_EXPLICIT
  if (tyvarmap!=EmptyMap())
    hfprintf(hstderr, "Leaving checkExplicits().\n");
#endif
  QUIT;
  return data->wrong;
}

MAPCOLLECTFUNC(do_tyvars)
{
  struct tyvars *data;

  data = MAPDATA(struct tyvars *);
  return tyVars(y,data->assum,data->state,data->flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		,data->strength
#endif
		,data->memoref);
}

MAPCOLLECTFUNC(do_tyvars_num)
{
  struct tyvars *data;

  data = MAPDATA(struct tyvars *);
  return tyVars(x,data->assum,data->state,data->flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		,data->strength
#endif
		,data->memoref);
}

#if IMPERATIVE_TYPES==EFFECT_TYVAR
PRIVATE MapPtr do_tyvars_fx P((IN CONST ExprPtr x,IN CONST ExprPtr y,INOUT CHAR *data));

#ifdef HAS_PROTOTYPES
PRIVATE ExprPtr fxTyVars1(IN CONST ExprPtr sigma,
			  IN CONST ExprPtr assum,
			  IN CONST ExprPtr state,
			  IN SHORT flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			  ,IN INTstrength
#endif
			  ,INOUT ExprPtr memoref)
#else
PRIVATE ExprPtr fxTyVars1(sigma,assum,state,flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		      ,strength
#endif
		      ,memoref)
IN CONST ExprPtr sigma;
IN CONST ExprPtr assum,state;
IN SHORT flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength;
#endif
INOUT ExprPtr memoref;
#endif
{
  struct tyvars *data;
  ExprPtr val;
  Boolean found;

  if (flags==0)
    {
      do {
	val = sigma;
	nongens = SetAdd(val,nongens);
	sigma = nongenValue(sigma,state,&found);
      } while (found);
      sigma = val;
    }
  else if (flags==0x80)
    {
      sigma = fxRealSigma(sigma,state);
      nongens = SetAdd(val,nongens);
    }
  data = (struct tyvars *)alloctmp(sizeof(struct tyvars));
  data->assum = assum;
  data->state = state;
  data->flags = flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  data->strength = strength;
#endif
  data->memoref = memoref;
  return SetUnion(nongens,
		  MapCollect(CDR(sigma),do_tyvars_fx,data));
}

MAPCOLLECTFUNC(do_tyvars_fx)
{
  struct tyvars *data;
  ExprPtr (*tyvars)
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	P((IN CONST ExprPtr f,IN CONST ExprPtr assum,IN CONST ExprPtr state,IN SHORT flags,IN INT strength,INOUT ExprPtr memoref));
#else
	P((IN CONST ExprPtr f,IN CONST ExprPtr assum,IN CONST ExprPtr state,IN SHORT flags,INOUT ExprPtr memoref));
#endif

  data = MAPDATA(struct tyvars *);
  tyvars = (x->kind==(CHAR)E_ARRAY_2)?fxTyVars1:tyVars;
  return (*tyvars)(x,data->assum,data->state,
		   data->flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		   ,data->strength
#endif
		   ,data->memoref);
}

PRIVATE ExprPtr fxTyVars(f,assum,state,flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		      ,strength
#endif
		      ,memoref)
IN CONST ExprPtr f;
IN CONST ExprPtr assum,state;
IN SHORT flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength;
#endif
INOUT ExprPtr memoref;
{
  return (f->kind==(CHAR)E_TYPEFX)?fxTyVars1(CDR(f),assum,state,flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
					     ,strength
#endif
		      			     ,memoref):NewMap();
}
#endif

#ifdef HAS_PROTOTYPES
PUBLIC MapPtr tyVars(IN CONST ExprPtr type,
		      IN CONST ExprPtr assum,IN CONST ExprPtr state,
		      IN SHORT flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		      ,IN INT strength
#endif
		      ,INOUT ExprPtr memoref)
#else
PUBLIC MapPtr tyVars(type,assum,state,flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		      ,strength
#endif
		      ,memoref)
	/* get free type variables with flags containing flags in type */
	/* assume that flags does not contain F_EQ, or else it won't be right */
     /* if flags contains the flags 0x80, gets only non bound variables */
IN CONST ExprPtr type;
IN CONST ExprPtr assum,state;
IN SHORT flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IN INT strength; /* and with strength <= strength */
#endif
INOUT ExprPtr memoref;
#endif
{
  ExprPtr nongens;

again:
  type = correctTypeAppl(type,state);
  nongens = NewMap();
  /*if (type==NIL)
    return nongens;*/
  switch (type->kind)
    {
      case E_NIL:
	break;
      case E_VAR:
#ifdef TYPE_CACHE
	{
	  ExprPtr res;
	  Boolean found;

	  res = MapGet(type,REFVAL(memoref),&found);
	  if (found)
	    return res;
	}
#endif
	{
	  Boolean found;
	  ExprPtr type2;

	  if (genericp(type,assum))
	    break;
	  if (!(flags & 0x80) &&
		IN_FLAGS(flags,varFlags(type))
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		&& tyvarStrength(type,state)<=strength
#endif
		)
	    nongens = SetAdd(type,nongens);
	  type2 = nongenValue(type,state,&found);
	  if (found)
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	    nongens = SetUnion(nongens,tyVars(type2,assum,state,flags,
					      strength,memoref));
#else
	    nongens = SetUnion(nongens,tyVars(type2,assum,state,flags,memoref));
#endif
#ifdef TYPE_CACHE
	  setrefMapPut(type,nongens,memoref);
#endif
	  break;
	}
      case E_TYPE_APPL:
#if IMPERATIVE_TYPES==EFFECT_TYVAR
	nongens = fxTyVars(APPLFUN(type),assum,state,flags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			   ,strength
#endif
			   ,memoref);
#endif
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	nongens = SetUnion(nongens,tyVars(APPLARG(type),assum,state,flags,
					  strength,memoref));
#else
	nongens = SetUnion(nongens,tyVars(APPLARG(type),assum,state,flags,
					  memoref));
#endif
	break;
      case E_CONS:
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	nongens = SetUnion(tyVars(CAR(type),assum,state,flags,strength,memoref),
			   tyVars(CDR(type),assum,state,flags,strength,memoref));
#else
	nongens = SetUnion(tyVars(CAR(type),assum,state,flags,memoref),
			   tyVars(CDR(type),assum,state,flags,memoref));
#endif
	break;
      case E_EMPTY:
	break;
      case E_MAPLET:
	type = MAPLETY(type);
	goto again;
      case E_MAP:
	{
	  struct tyvars *data = (struct tyvars *)alloctmp(sizeof(struct tyvars));

	  data->assum = assum;
	  data->state = state;
	  data->flags = flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  data->strength = strength;
#endif
	  data->memoref = memoref;
	  nongens = MapCollect(type,do_tyvars,data);
	  break;
	}
      case E_NUMTYPE:
	{
	  struct tyvars *data = (struct tyvars *)alloctmp(sizeof(struct tyvars));

	  data->assum = assum;
	  data->state = state;
	  data->flags = flags;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	  data->strength = strength;
#endif
	  data->memoref = memoref;
	  nongens = MapCollect(REFVAL(type),do_tyvars_num,data);
	  break;
	}
      default:
	kinderror(message(51),type->kind);
	break;
    }
  return nongens;
}

#if IMPERATIVE_TYPES!=EFFECT_TYVAR
MAPCOLLECTFUNC(do_unquant_y)
{
  struct unquant *data;

  data = MAPDATA(struct unquant *);
  return tyVars(y,data->assum,data->state,F_IMP
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		,0L
#endif
		,data->memoref);
}

PUBLIC MapPtr unquant_y(exps,assum,state,memoref)
IN CONST MapPtr exps; /* map var -> type */
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
INOUT ExprPtr memoref;
{
  struct unquant *data = (struct unquant *)alloctmp(sizeof(struct unquant));
  ExprPtr nongens;
  ENTER;

  data->assum = assum;
  data->state = state;
  data->memoref = memoref;
  nongens = MapCollect(exps,do_unquant_y,data);
  KEEPONLY(nongens);
  return nongens;
}
#endif

MAPCOLLECTFUNC(do_unquant_x)
{
  struct unquant *data;
  ExprPtr nongens;

  data = MAPDATA(struct unquant *);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  if (x->kind==(CHAR)E_ARRAY_2)
    {
      nongens = fxTyVars1(x,data->assum,data->state,0
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
			  ,INF_STRENGTH
#endif
			  ,data->memoref);
    }
  else
#endif
    {
      nongens = tyVars(x,data->assum,data->state,0
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		       ,INF_STRENGTH
#endif
		       ,data->memoref);
    }
  return nongens;
}

PUBLIC MapPtr unquant_x(exps,assum,state,memoref)
IN CONST MapPtr exps; /* map var -> type */
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
INOUT ExprPtr memoref;
{
  struct unquant *data = (struct unquant *)alloctmp(sizeof(struct unquant));
  ExprPtr nongens;
  ENTER;

  data->assum = assum;
  data->state = state;
  data->memoref = memoref;
  nongens = MapCollect(exps,do_unquant_x,data);
  KEEPONLY(nongens);
  return nongens;
}

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
MAPCOLLECTFUNC(do_getstrengths)
{
  ExprPtr state;

  state = MAPDATA(ExprPtr);
  return (varFlags(x) & F_IMP)?
    MapSingleton(x,KINT(tyvarStrength(x,state))):
      NewMap();
}

PUBLIC MapPtr getStrengths(tyvars,state)
IN CONST MapPtr tyvars;
IN CONST ExprPtr state;
{
  MapPtr res;
  ENTER;

  res = MapCollect(tyvars,do_getstrengths,state);
  KEEPONLY(res);
  return res;
}

MAPSWEEPFUNC(do_setstrengths)
{
  struct setstrengths *data;

  data = MAPDATA(struct setstrengths *);
  if (varFlags(x) & F_IMP)
    {
      setMapPut(x,data->strength,&S_STRENGTH(data->state));
    }
  return false;
}

PUBLIC void setStrengths(tyvars,state,strength)
IN CONST MapPtr tyvars;
INOUT ExprPtr state;
IN INT strength;
{
  struct setstrengths *data = (struct setstrengths *)alloctmp(sizeof(struct setstrengths));
  ENTER;

  data->state = state;
  data->strength = KINT(strength);
  (void) MapSweep(tyvars,do_setstrengths,data);
  QUIT;
}

MAPSWEEPFUNC(do_checkstrengths)
{
  struct checkstrengths *data;

  data = MAPDATA(struct checkstrengths *);
  if ((varFlags(x) & F_IMP) && tyvarStrength(x,data->state)<naturalStrength(x))
    type_error(data->zone,(CHAR *)varMsg(message(52),x));
  return false;
}

PUBLIC void checkStrengths(zone,tyvars,state)
IN CONST TZone *zone;
IN CONST MapPtr tyvars;
IN CONST ExprPtr state;
{
  struct checkstrengths *data = (struct checkstrengths *)alloctmp(sizeof(struct checkstrengths));

  data->zone = zone;
  data->state = state;
  (void) MapSweep(tyvars,do_checkstrengths,data);
}
#endif

PRIVATE MapPtr do_unquant_packers P((IN CONST ExprPtr maplet,
				     INOUT char *_data));
PRIVATE MapPtr do_unquant_packers(maplet,_data)
IN CONST ExprPtr maplet;
INOUT char *_data;
{
  struct unquant *data;
  ExprPtr type;

  data = (struct unquant *)_data;
  type = MAPLETY(maplet);
  if (/*type!=NIL && */type->kind==(CHAR)E_BIAS)
    type = CAR(type);
  return tyVars(type,data->assum,data->state,0
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
		,INF_STRENGTH
#endif
		,data->memoref);
}

PUBLIC MapPtr unquant_packers(assum,state,memoref)
IN CONST ExprPtr assum;
IN CONST ExprPtr state;
INOUT ExprPtr memoref;
{
  struct unquant *data = (struct unquant *)alloctmp(sizeof(struct unquant));
  ExprPtr nongens;
  ENTER;

  data->assum = assum;
  data->state = state;
  data->memoref = memoref;
  /* unquantify all type variables still free in packed types */
  nongens = NCollect(S_PACKERS(state),do_unquant_packers,(char *)data);
  KEEPONLY(nongens);
  return nongens;
}

PUBLIC MapPtr getSureeq(eqlist)
IN CONST ExprPtr eqlist;
{
  MapPtr eqs;
  ExprPtr args,id,methods;
  ENTER;

  eqs = NewMap();
  while (eqlist!=NIL)
    {
      args = ARGS(CAR(eqlist));
      id = CAR(ARGS(CAR(args)));
      methods = ARGS(CAR(CDR(args)));
      if (methods!=NIL)
	eqs = SetAdd(id,eqs);
      eqlist = CDR(eqlist);
    }
  KEEPONLY(eqs);
  return eqs;
}

#ifdef HAS_PROTOTYPES
PUBLIC void addPacker(IN CONST ExprPtr expr,IN CONST ExprPtr type,
			  IN COUNTER fnlevel,INOUT ExprPtr state)
#else
PUBLIC void addPacker(expr,type,fnlevel,state)
IN CONST ExprPtr expr;
IN CONST ExprPtr type;
IN COUNTER fnlevel;
INOUT ExprPtr state;
#endif
{
  ENTER;

  if (fnlevel!=0)
    {
      type = Kbias(type,KINT((INT)fnlevel)),
      (void) KEEP(type);
    }
  setNPut(expr,type,&S_PACKERS(state),NOEXPR);
  QUIT;
}

#ifdef HAS_PROTOTYPES
PUBLIC void addAnnotation1(IN CONST ExprPtr expr,IN CONST ExprPtr type,
			   IN COUNTER fnlevel,IN CONST ExprPtr sharingp,
			   INOUT ExprPtr state)
#else
PUBLIC void addAnnotation1(expr,type,fnlevel,sharingp,state)
IN CONST ExprPtr expr;
IN CONST ExprPtr type;
IN COUNTER fnlevel;
IN CONST ExprPtr sharingp;
INOUT ExprPtr state;
#endif
{
  ENTER;

  if (fnlevel==0 && sharingp==trueValue) /* default */
    ;
  else
    {
      sharingp = KONS (KINT((INT)fnlevel), sharingp);
      type = Kbias(type,sharingp);
      (void) KEEP(type);
    }
#ifdef DEBUG2
  {
    Boolean found;
    ExprPtr oldtype;

    oldtype = NGet(expr,S_ANNOTATIONS(state),&found);
    if (found && oldtype!=type)
      CRASH();
  }
#endif
  setNPut(expr,type,&S_ANNOTATIONS(state),NOEXPR);
  QUIT;
}

#ifdef HAS_PROTOTYPES
PUBLIC void addAnnotation(IN CONST ExprPtr expr,IN CONST ExprPtr type,
			  IN COUNTER fnlevel,INOUT ExprPtr state)
#else
PUBLIC void addAnnotation(expr,type,fnlevel,state)
IN CONST ExprPtr expr;
IN CONST ExprPtr type;
IN COUNTER fnlevel;
INOUT ExprPtr state;
#endif
{
  ExprPtr sharingp;
  ENTER;

  sharingp = eqTypeTdg(type,state,true);
  if (sharingp!=falseValue)
    sharingp = trueValue;
  /* sharingp is used by compile.c/noSharing() to detect
     whether it is necessary or not to share expressions
     that are built of this type. */
  addAnnotation1 (expr, type, fnlevel, sharingp, state);
  QUIT;
}

MAPSWEEPFUNC(do_instscale)
{
  struct instscale *data;
  ExprPtr scale;
  Boolean found;

  data = MAPDATA(struct instscale *);
  scale = MapGet(x,S_TYCONS(data->state),&found);
  if (found)
    {
      scale = instScale(scale,data->state);
      scale = scalePower(scale,y);
      data->coeff = z_mul(data->coeff,CAR(scale));
      (void) KEEP(data->coeff);
      data->var = typeMul1(data->var,CDR(scale));
    }
  else
    {
      data->var = typeMul1(data->var,MapSingleton(x,y));
    }
  return false;
}

PUBLIC ExprPtr instScale(scale,state)
IN CONST ExprPtr scale;
IN CONST ExprPtr state;
{
  struct instscale *data = (struct instscale *)alloctmp(sizeof(struct instscale));
  ENTER;

  data->coeff = CAR(scale);
  data->var = NewMap();
  data->state = state;
  (void) MapSweep(CDR(scale),do_instscale,data);
  scale = Ktuple2(data->coeff,data->var);
  KEEPONLY(scale);
  return scale;
}

PUBLIC ExprPtr nongenValue(typevar,state,foundp)
IN CONST ExprPtr typevar;
IN CONST ExprPtr state;
OUT Boolean *foundp;
{
  ExprPtr type;

  type = MapGet(typevar,S_TYVARSUBCACHE(state),foundp);
  if (*foundp)
    return type;
  type = MapGet(typevar,S_TYVARSUBS(state),foundp);
  return type;
}

PUBLIC void flushTyVarSubCache(state)
INOUT ExprPtr state;
{
  ENTER;

  setMapMerge(&S_TYVARSUBS(state),S_TYVARSUBCACHE(state));
  SETQ(S_TYVARSUBCACHE(state),NewMap());
  QUIT;
}
